
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * Revision 1.47  2016-04-08 GK
 * Fix incorrect end time calculation when loading
 * quality records, which could cause incorrect split
 * of quality records. See ISI-443
 *
 * Revision 1.46  2013-10-17 gerald
 * Add flag parameter to getFrameProduct()
 *
 * Revision 1.45  2013-07-12 gerald
 * Write 7 significant digits for ncalib to wfdisc
 * (see ISI-126)
 *
 * Revision 1.44  2013-07-05 gerald
 * Move 'getODBCProvider' to db_access.c
 *
 * Revision 1.43  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.42  2013-03-02  gerald
 * Log more information when loading of sensor
 * information failed
 *
 * Revision 1.41  2012-06-14  ctbto
 * Ignore error when inserting a clf record
 * also when using idcodbc functions
 *
 * Revision 1.40  2012-01-02 13:55:27  ctbto
 * Fix missing checks and wrong data type for sample rate
 * when using idcodbc functions
 *
 * Revision 1.39  2011-10-18 09:33:08  ctbto
 * Make cd2wng works with idcodbc. Default interface for ODBC is
 * gODBC
 *
 * Revision 1.38  2011-09-28 07:48:31  ctbto
 * Remove checkFrameProduct() function which is not longer used.
 * Fix incorrect cd2wng version in frameproduct table.
 * Fix invalid type ID of waveform files used during start up when
 * loading quality records. Update some too old comments and error
 * messages.
 *
 * Revision 1.37  2011-06-09 14:20:57  ctbto
 * Add function to change default frame product IDs
 *
 * Revision 1.36  2011-05-16 11:43:55  gerald
 * Read frame product type ID from internal variable. Regress
 * modification which loads frame product ID from fpdescription table
 *
 * Revision 1.35  2011-04-29 09:55:46  gerald
 * Load frame product type ID from fpdescription table
 *
 * Revision 1.34  2011-04-27 13:31:47  gerald
 * Rework updateFrameProduct to get rid of dependencies to internal
 * reference list
 *
 * Revision 1.33  2011-01-17 14:59:26  gerald
 * Fix wrong SQL C data type mapping check when loading new
 * quality records. Seen only with old CTBTO ODBC driver. Drop
 * check for relative wfdisc paths
 *
 * Revision 1.32  2010-12-13 16:39:58  gerald
 * Replace dfile with wffid
 *
 * Revision 1.31  2010-11-30 19:06:16  gerald
 * Switch to cdquality table. Add channel attributes, channel attribute
 * mask and file offset. Fix wrong number of samples when loading
 * quality record. Change format string for wfdisc and cdquality
 * records to decrease size of debug messages (e.g. %.5f instead
 * of %17.5f)
 *
 * Revision 1.30  2010-11-11 16:41:27  gerald
 * Log more information when cd2wng start-up fails
 *
 * Revision 1.29  2010-10-27 11:42:02  gerald
 * Rename structure member segSumData to segMean and
 * segSumSquareData to segM2 and load meanCounts instead of
 * sunCounts from file/database
 *
 * Revision 1.28  2010-10-20 09:13:16  gerald
 * Use new gODBC function to open more than one connection
 *
 * Revision 1.27  2010-10-18 09:32:12  gerald
 * Add second DB connection to get rid of ODBCGetNextId() timeouts
 * during start up of cd2wng (caused by serialization of DB requests
 * by unix ODBC driver when selects in the DB are slow:
 * select|select|ID request|select|select|select|select|select -->)
 * This change requires also a fix in libgODBC .
 * Rename initDB() to getDBhandle()
 * Decrease number of requested wfdics IDs from 100 to 20, for file IDs
 * from 50 to 20 and for quality IDs from 1 to 20
 *
 * Revision 1.26  2010-07-07 14:21:53  gerald
 * Improve error logging for quality records
 *
 * Revision 1.25  2010-05-11 15:16:41  gerald
 * Fix channel names loaded from the wrong time period when in fixed
 * interval mode and load also station name from static tables
 *
 * Revision 1.24  2010-04-16 09:01:36  gerald
 * Fix 'assumption for NEW_SEG failed' bug when loading
 * quality records in DB mode
 *
 * Revision 1.23  2010-01-12 10:27:42  gerald
 * Minimize the number of DB calls for getting frame IDs by
 * increasing the number of IDs which are requested from the
 * DB from 1 to 50
 * Add retry code to getNextID(). Getting an ID is now retried 3
 * times before giving up
 *
 * Revision 1.22  2009-06-23 16:39:44  gerald
 * Add channel mask when writing the clf table
 * Add stream and position fields to sensor table
 *
 * Revision 1.21  2009-05-20 15:26:57  gerald
 * Change the NULL values to the values used by the old
 * cd2w when insert a new record in the frameproduct table
 * and change the author string to 'cd2wng-2.1.24'
 * Fix missing update of file size for binary files in the
 * frameproduct table
 *
 * Revision 1.20  2009-05-18 09:39:06  gerald
 * Fix C/SQL mapping problem when loading channel revision
 *
 * Revision 1.19  2009-05-11 13:19:21  gerald
 * Fix wrong revision for some channels when loading sensor info
 * from DB
 * Fix ODBC to C data type conversion problem when
 * running in a 64-bit environment
 *
 * Revision 1.18  2009-04-29 16:07:26  gerald
 * Write clf records to database (only). A clf record is only
 * written if the frame is merged into a waveform segment.
 * No clf record is written for invalid data or data which
 * does not match the filter criteria. Note: This
 * feature should be removed after the squal re-design
 *
 * Add revision to sensor info and remove duplicates when loading
 * the sensor info from database
 *
 * Revision 1.17  2009-04-24 14:00:16  gerald
 * Change data type of sumcounts and sumsquarecounts to
 * double to avoid possible integer overflows when computing
 * the square sum within one frame
 * Add file name to quality records required for correct re-start
 *
 * Revision 1.16  2009-04-06 15:05:52  gerald
 * Obsolete code removed
 *
 * Revision 1.15  2009-03-23 17:45:52  gerald
 * Fix splint warnings which occurred after upgrading
 * to new splint version
 *
 * Revision 1.14  2009-02-12 17:12:08  gerald
 * Add saving/updating/loading of quality records and check
 * wfdiscs when they are loaded from DB (start == endtime,
 * number of samples > 0, ...)
 *
 * Revision 1.13  2008-12-18 16:08:31  gerald
 * Strip off at sign and following characters in config parameter
 * 'dbAccount'. Improve recycleIDs() to recycle also quality IDs
 *
 * Revision 1.12  2008-11-03 16:29:49  gerald
 * Add Elenas 64 bit patches
 *
 * Revision 1.11  2008-10-30 16:01:20  gerald
 * Fix too many records after re-start
 *
 * Revision 1.10  2008-09-23 16:49:31  gerald
 * Replace to_date(sysdate) with sysdate and fix bug
 * when loading S3 wfdisc records in DB mode
 *
 * Revision 1.9  2008-09-18 14:01:21  gerald
 * Log wfid when an insert or update failes
 *
 * Revision 1.8  2008-09-16 18:07:25  gerald
 * Fix wrong station name in frameproduct table
 *
 * Revision 1.7  2008-08-12 16:15:26  gerald
 * Add and update clf receiving time in frameproduct
 *
 * Revision 1.6  2008/07/28 12:14:56  gerald
 * Read intern and extern site names from DB, insert
 * binary files and text files into frameproduct,
 * check for relative wfdisc paths, check if all channels
 * are loaded from the waveform header when
 * reading wfdisc files
 *
 * Revision 1.5  2008/07/07 09:16:58  gerald
 * Fix DB error when creating new wfids caused by an
 * unprotected commit. Fix duplicate wfids caused by an
 * unprotected recycleIds() call. Fix wrong offset when reading
 * wfdiscs for restart. Increase IDs by a count of 100
 * instead of 20
 *
 * Revision 1.4  2008-06-03 15:13:16  gerald
 * Add function for loading wfdiscs from DB
 *
 * Revision 1.3  2008/05/08 14:34:52  gerald
 * Add functions for insert, delete and update wfdiscs. Add
 * function for loading sensor descriptions from database
 *
 * Revision 1.2  2008/04/07 12:50:38  gerald
 * Run some test queries in the DB
 *
 * Revision 1.1  2007/12/10 13:12:21  gerald
 * Add CD2W_NG directory
 *
 */

#include <config.h>

#ifdef HAVE_LIBODBC

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "cd_common.h"
#include "logfiles.h"
#include "events.h"
#include "merge.h"
#include "db_access.h"

#ifndef HAVE_LIBIDCODBC
#include "libgODBC.h"
/*@observer@*/ const char *getODBCProvider() { return("godbc unknown"); }
#else /* HAVE_LIBIDCODBC */
#include "idcodbc.h"
/*@observer@*/ const char *getODBCProvider() { return(libIDCodbcVersion()); }
#endif

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

#define ATTRIB_BUFFER_SIZE 64

static long fpTypeID[5] =  /* Array of CD frame Product Type IDs */
       { 31, 32, 33, 35, 54 }; /* Must correspond to CD_*_FILE defines */

/* Defines are copied from db_access.h to make sure that there are the same */
#define CD_B10_FILE   0
#define CD_B11_FILE   1
#define CD_CLF_FILE   2
#define CD_TXT_FILE   3
#define CD_MWF_FILE   4

/**
 * @short Set frame product IDs from DB
 *
 * @description This function sets the frame product IDs provided in
 * @description a colon separated list
 *
 * @param fpIDs   list of frame product ID's. Position must correspond to
 *                CD_*_FILE defines above
 *
 * @return 0 on success or -1 on error
 */

int
setFpIDs(const char* fpIDs)
{
   unsigned int fields;
   size_t lenOfLine;
   split_t result;
   unsigned int ii;

   DEBUG(2, "Set frame product ID's");

   lenOfLine = strlen(fpIDs);
   fields = splitString(fpIDs, lenOfLine, ':', &result);

   /* check number of fields */
   if (result.separators != 4 && fields != 5)
   {
      logError("Invalid number of separators in fpTypeIDs", fpIDs);
      return (-1);
   }

   /* Convert fields to number */
   for (ii = 0; ii < fields; ii++)
   {
      char *endptr;

      /* strtol() stops on first invalid character which is the delimiter */
      fpTypeID[ii] = strtol(result.pString[ii], &endptr, 10);
      DEBUG3(2, "fpTypeID of %s is %ld", result.pString[ii], fpTypeID[ii]);
      if (endptr == result.pString[ii])
      {
         logError("Invalid frameproduct type ID detected -->",
                   result.pString[ii]);
         return (-1);
      }
   }
   return(0);
}

#ifdef LOAD_FP_IDS_FROM_DB

/**
 * @short Load type IDs for files created by CD tools
 *
 * @description This function loads the type IDs for files created
 * @description by CD tools from the fpdescription table
 *
 * @param hdbc        database descriptor
 * @param pCd2w       pointer to application extension structure
 *
 * @return -1 in case of error, 0 otherwise
 */

static const char *cdFileNames[] =
{
  "CD-1.0 binary frame file", /* 31 */
  "CD-1.1 binary frame file", /* 32 */
  "CD frame text file", /* 33 */
  "CD receiver txt file", /* 35 */
  "CD waveform 2.0 file" /* 54 */
};

int
loadFpIDs(SQLHDBC hdbc)
{
   int ii,
       ret;
   ODBC_Result result;
   char **name;
   double *typeIDd = NULL;
   long   *typeIDl = NULL;
   char buffer[QUERY_BUFFER_SIZE];
   /* set up the query */
   (void) snprintf(buffer, QUERY_BUFFER_SIZE,
         "select typeid,name from fpdescription \
 where name like 'CD%%file' order by name");
   DEBUG(2, buffer);
   ret = ODBCGeneralQuery(hdbc, buffer, &result);

   if (ret < 0)
   {
      /*@-compdestroy@*/
      /*@-mustfreefresh@*/
      /* This is a fatal error. Shutdown must be done in the
         calling function */
      logError("Can't read fpdescription from Database", ODBCErrMsg());
      return (-1);
      /*@=mustfreefresh@*/
      /*@=compdestroy@*/
   }

   if (result.numRows < 5)
   {
      (void)snprintf(buffer, sizeof(buffer),
                     "Loading of fp IDs failed: %d < 5", result.numRows);
      goto end_error;
   }

   /* Read now result from SQL query and set up the ID's.
      Result see above */
   DEBUG2(5, "SQL type for typeID [%d]",
          (int)result.column[0].dataType);
   if (result.column[0].dataType == SQL_DECIMAL ||
       result.column[0].dataType == SQL_NUMERIC ||
       result.column[0].dataType == SQL_FLOAT ||
       result.column[0].dataType == SQL_REAL ||
       result.column[0].dataType == SQL_DOUBLE)
   {
      typeIDd = (double *)result.column[0].value;

   }
   else if (result.column[0].dataType == SQL_INTEGER)
   {
      /* Due a bug in ODBCGetResultCols SQL_INTEGER is mapped
         to long instead of SQLINTEGER and SQL_SMALLINT is
         mapped to int instead of SQLSMALLINT and SQL_BIGINT is
         mapped to long instead of SQLBIGINT */
      typeIDl = (long *)result.column[0].value;
   }
   else
   {
      (void)snprintf(buffer, sizeof(buffer),
                     "Unknown SQL type for typeID [%d]",
                    (int)result.column[0].dataType);
      goto end_error;
   }
   name = (char **)result.column[1].value;

   for (ii = 0; ii < result.numRows; ii++)
   {
      size_t ij;
      long typeID;

      if (result.column[0].dataType == SQL_INTEGER)
      {
         /*@-nullderef@*/
         typeID = typeIDl[ii];
      }
      else
      {
         typeID = (long)typeIDd[ii];
         /*@=nullderef@*/
      }
      /* Find index for file */
      for (ij = 0; ij < (sizeof(cdFileNames)/sizeof(char *)); ij++)
      {
          if (0 == strcmp(cdFileNames[ij], name[ii]))
          {
             fpTypeID[ij] = typeID;
          }
      }
   }

   /* Now check if really all file IDs are loaded or if there was
      a duplicate entry in the fpdescription table */
   for (ii = 0; ii < 5; ii++)
      if (0 == fpTypeID[ii]) goto end_error;

   ODBCFreeResult(&result);
   /*@-compdestroy@*/
   return (0);

end_error:

   ODBCFreeResult(&result);
   logError(buffer, "getFpIDs");
   return(-1);
   /*@-mustfreefresh@*/
   /*@=mustfreefresh@*/
   /*@=compdestroy@*/
}

#endif /* LOAD_FP_IDS_FROM_DB */

/**
 * @short Load channel description from DB
 *
 * @description This function loads the channel description from
 * @description the sensor, channel and channame table for the given
 * @description time period
 *
 * @param appName             application name
 * @param hdbc                database descriptor
 * @param sensorTableCount    number of sensor entries (out)
 * @param startDate           start date and time (if any)
 * @param endDate             end time (if any)
 *
 * @return pointer to sensor table or NULL
 */

/*@only@*//*@null@*/ sensor_t *
loadSensorFromDB(const char *appName, SQLHDBC hdbc,
                 /*@out@*/int *sensorTableCount, time_t startDate,
                 time_t endDate)
{
   int ii,
       ij,
       ret;
   ODBC_Result result;
   sensor_t tmpSensor;
#ifndef HAVE_LIBIDCODBC
   char **station;
   char **stream;
   char **intern_sta;
   char **extern_sta;
   char **intern_chan;
   char **insType;
   double *chanIDd = NULL;
   long   *chanIDl = NULL;
   double *revisionD = NULL;
   long   *revisionL = NULL;
   double *positionD = NULL;
   long   *positionL = NULL;
   double *calib;
   double *calper;
   double *sampleRate;
   double *startTime;
   double *endTime;
#else /* HAVE_LIBIDCODBC */
   size_t highMark;
   /*@-compdef@*/
   reqColumns_t reqColumns[] = {
    {"station", tmpSensor.station, CT_STRING, sizeof(tmpSensor.station), -1},
    {"stream", tmpSensor.stream, CT_STRING, sizeof(tmpSensor.stream), -1},
    {"sta", tmpSensor.intern_sta, CT_STRING, sizeof(tmpSensor.intern_sta), -1},
    {"extern_sta", tmpSensor.extern_sta, CT_STRING,
                   sizeof(tmpSensor.extern_sta), -1},
    {"chan", tmpSensor.intern_chan, CT_STRING,
             sizeof(tmpSensor.intern_chan), -1},
    {"chanid", &(tmpSensor.chanID), CT_LONG, sizeof(tmpSensor.chanID), -1},
    {"revision", &(tmpSensor.revision), CT_INT,
                 sizeof(tmpSensor.revision), -1},
    {"position", &(tmpSensor.position), CT_INT,
                 sizeof(tmpSensor.position), -1},
    {"instype", tmpSensor.insType, CT_STRING, sizeof(tmpSensor.insType), -1},
    {"ncalib", &(tmpSensor.nomCalib), CT_FLOAT,
               sizeof(tmpSensor.nomCalib), -1},
    {"ncalper", &(tmpSensor.nomCalibPer), CT_FLOAT,
               sizeof(tmpSensor.nomCalibPer), -1},
    {"samprate", &(tmpSensor.expSampleRateInDB), CT_FLOAT,
               sizeof(tmpSensor.expSampleRateInDB), -1},
    {"time", &(tmpSensor.startTime), CT_DOUBLE,
               sizeof(tmpSensor.startTime), -1},
    {"endtime", &(tmpSensor.endTime), CT_DOUBLE,
               sizeof(tmpSensor.endTime), -1} };
   /*@=compdef@*/
#endif
   char buffer[QUERY_BUFFER_SIZE*2];
   sensor_t *sensorTable;

   if (0 == startDate)
       startDate = (time_t)time(NULL);
   if (0 == endDate)
       endDate = (time_t)time(NULL);
   *sensorTableCount = 0;

   /* set up the query; Free ORACLE driver has problems for values
      other than null in the max() function */
   (void) snprintf(buffer, sizeof(buffer),
         "select distinct c.station, c.stream, s.sta, c.extern_sta, s.chan,\
 s.chanid, i.instype, i.ncalib, i.ncalper, i.samprate, s.time, s.endtime, \
 max(c.revision) as revision, c.position from\
 sensor s, instrument i, channame c where s.inid = i.inid and\
 c.intern_sta = s.sta and c.intern_chan = s.chan and\
 (%f >= s.time and %f <= s.endtime) and c.revision = (select max(revision)\
 from channame where intern_sta = s.sta and intern_chan = s.chan)\
 group by c.station, c.stream, s.sta, c.extern_sta,\
 s.chan, s.chanid, i.instype, i.ncalib, i.ncalper, i.samprate,\
 s.time, s.endtime, c.position\
 order by s.sta, c.extern_sta, s.chan", (float)startDate, (float)endDate);
   ret = ODBCGeneralQuery(hdbc, buffer, &result);

   DEBUG(2, buffer);
   if (ret < 0)
   {
      /* This is not a fatal error. The sender could will try to
         read the clf files from the index file next time */
      /*@-compdestroy@*/
      /*@-mustfreefresh@*/
      logError("Can't read channel info from Database", ODBCErrMsg());
      return(NULL);
      /*@=mustfreefresh@*/
      /*@=compdestroy@*/
   }

#ifndef HAVE_LIBIDCODBC
   if (result.numRows == 0)
#else /* HAVE_LIBIDCODBC */
   if (SQL_NO_DATA_FOUND == result.sqlRetCode)
#endif
   {
      DEBUG(2, "No rows returned");
      /*@-compdestroy@*/
      ODBCFreeResult(&result);
      return(NULL);
      /*@=compdestroy@*/
   }

#ifndef HAVE_LIBIDCODBC
   /* Allocate buffer for structures */
   sensorTable = (sensor_t *)calloc((size_t)result.numRows, sizeof(sensor_t));
#else /* HAVE_LIBIDCODBC */
   /* Number of channels is not known in advance. Start with
      1000 and increase by 1000 if necessary. The select returns approx.
      1200 channels, so the extension code should be triggered any time
      the sensor table is loaded */
#define SENSOR_TAB_INIT_SZ  1000
   sensorTable = (sensor_t *)calloc(SENSOR_TAB_INIT_SZ, sizeof(sensor_t));
   highMark = SENSOR_TAB_INIT_SZ;
#endif /* HAVE_LIBIDCODBC */

   if (NULL == sensorTable)
   {
      /*@-compdestroy@*/
      /*@-mustfreefresh@*/
      ODBCFreeResult(&result);
      stopProgram(appName, "Can't allocate memory for sensor table", NULL);
      return(NULL);
      /*@=mustfreefresh@*/
      /*@=compdestroy@*/
   }

#ifndef HAVE_LIBIDCODBC
   /* -5 is SQL_BIGINT */
   DEBUG2(2, "Revision SQL type is [%hd]",
             (short)result.column[12].dataType);

   /* Read now result from SQL query and set up the data type mapping */
   station = (char **)result.column[0].value;
   stream = (char **)result.column[1].value;
   intern_sta = (char **)result.column[2].value;
   extern_sta = (char **)result.column[3].value;
   intern_chan = (char **)result.column[4].value;
   if (result.column[5].dataType == SQL_DECIMAL ||
       result.column[5].dataType == SQL_NUMERIC ||
       result.column[5].dataType == SQL_FLOAT ||
       result.column[5].dataType == SQL_REAL ||
       result.column[5].dataType == SQL_DOUBLE)
   {
      chanIDd = (double *)result.column[5].value;
      revisionD = (double *)result.column[12].value;
      positionD = (double *)result.column[13].value;
   }
   else if (result.column[5].dataType == SQL_INTEGER)
   {
      /* Due a bug in ODBCGetResultCols SQL_INTEGER is mapped
         to long instead of SQLINTEGER and SQL_SMALLINT is
         mapped to int instead of SQLSMALLINT */
      chanIDl = (long *)result.column[5].value;
      revisionL = (long *)result.column[12].value;
      positionL = (long *)result.column[13].value;
   }
   else
   {
      DEBUG2(2, "Unknown SQL type for chanID [%d]",
             (int)result.column[5].dataType);
      /*@-compdestroy@*/
      /*@-mustfreefresh@*/
      ODBCFreeResult(&result);
      stopProgram(appName, "Can't determine chanID SQL type", NULL);
      return(NULL);
      /*@=mustfreefresh@*/
      /*@=compdestroy@*/
   }

   insType = (char **)result.column[6].value;
   calib = (double *)result.column[7].value;
   calper = (double *)result.column[8].value;
   sampleRate = (double *)result.column[9].value;
   startTime = (double *)result.column[10].value;
   endTime = (double *)result.column[11].value;

   for (ii = 0; ii < result.numRows; ii++)
   {
      /* Fill the sensor table for each sensor */
      (void)strncpy(tmpSensor.station, station[ii], LEN_SITE_NAME);
      (void)strncpy(tmpSensor.stream, stream[ii], LEN_SITE_NAME);
      (void)strncpy(tmpSensor.intern_sta, intern_sta[ii], LEN_SITE_NAME);
      (void)strncpy(tmpSensor.extern_sta, extern_sta[ii], LEN_SITE_NAME);
      (void)strncpy(tmpSensor.intern_chan, intern_chan[ii],
                    LEN_CHANNEL_NAME);
      if (result.column[5].dataType == SQL_INTEGER)
      {
         /*@-nullderef@*/
         tmpSensor.chanID = chanIDl[ii];
         tmpSensor.revision = (int)revisionL[ii];
         tmpSensor.position = (int)positionL[ii];
      }
      else
      {
         tmpSensor.chanID = (long)chanIDd[ii];
         tmpSensor.revision = (int)revisionD[ii];
         tmpSensor.position = (int)positionD[ii];
         /*@=nullderef@*/
      }

      (void)strncpy(tmpSensor.insType, insType[ii], LEN_INSTRUMENT);
      tmpSensor.nomCalib = (float)calib[ii];
      /* Check nominal calibration value */
      /*@-realcompare@*/
      if (0.0 == tmpSensor.nomCalib)
      {
         DEBUG2(2, "Invalid nominal calibration value, chanID [%ld]",
                tmpSensor.chanID);
         /*@-compdestroy@*/
         /*@-mustfreefresh@*/
         ODBCFreeResult(&result);
         stopProgram(appName, "Invalid nominal calibration value",
                     tmpSensor.intern_sta);
         return(NULL);
         /*@=mustfreefresh@*/
         /*@=compdestroy@*/
      }
      /*@=realcompare@*/
      tmpSensor.nomCalibPer = (float)calper[ii];
      tmpSensor.expSampleRateInDB = (float)sampleRate[ii];
      tmpSensor.startTime = startTime[ii];
      tmpSensor.endTime = endTime[ii];
      sensorTable[ii] = tmpSensor;
   }
#else /* HAVE_LIBIDCODBC */

   ii = 0;
   while (ODBCFetchNextRow(&result))
   {
      /* Fill the sensor table for each sensor */
      tmpSensor.nomCalib = 0.0; /* for splint */
      tmpSensor.chanID = 0L; /* for splint */
      ret = ODBCGetColumns(sizeof(reqColumns), reqColumns, &result);

      /* Check nominal calibration value */
      /*@-realcompare@*/
      if (0.0 == tmpSensor.nomCalib)
      {
         DEBUG2(2, "Invalid nominal calibration value, chanID [%ld]",
                tmpSensor.chanID);
         DEBUG3(2, "loading sensor table stopped at line %d err code(%d)",
                (int)ii, ret);
         /*@-compdestroy@*/
         /*@-mustfreefresh@*/
         /*@-usereleased@*/
         ODBCFreeResult(&result);
         stopProgram(appName, "Invalid nominal calibration value",
                     tmpSensor.intern_sta);
         return(NULL);
         /*@=usereleased@*/
         /*@=mustfreefresh@*/
         /*@=compdestroy@*/
      }
      /*@=realcompare@*/

      if (ret != 0)
      {
         /*@-compdestroy@*/
         /*@-mustfreefresh@*/
         ODBCFreeResult(&result);
         DEBUG3(2, "loading sensor table stopped at line %d err code(%d)",
                (int)ii, ret);
         stopProgram(appName, "Can't load sensor table", NULL);
         return(NULL);
         /*@=mustfreefresh@*/
         /*@=compdestroy@*/
      }

      /* Check if the sensor info table must be extended */
      if ((size_t)ii == (highMark - 1))
      {
         highMark += SENSOR_TAB_INIT_SZ;
         DEBUG2(2, "realloc sensor table %d", (int)highMark);

         sensorTable = (sensor_t *)realloc(sensorTable,
                                           highMark * sizeof(sensor_t));
         if (NULL == sensorTable)
         {
            /*@-compdestroy@*/
            /*@-mustfreefresh@*/
            ODBCFreeResult(&result);
            stopProgram(appName,
                        "Can't reallocate memory for sensor table", NULL);
            return(NULL);
            /*@=mustfreefresh@*/
            /*@=compdestroy@*/
         }
      }
      sensorTable[ii] = tmpSensor;
      ii++;
   }
#endif /* HAVE_LIBIDCODBC */

   *sensorTableCount = ii;
   ODBCFreeResult(&result);

   IF_DEBUG (2)
   {
      for (ij = 0; ij < ii; ij++)
      {
         DEBUG_HDR;
         (void) fprintf(DEBUG_FP,
            "sensor: %.6s %.6s %.6s %.6s %.8s %ld %.6s %f %f %f %f %f %d %d\n",
                        sensorTable[ij].station, sensorTable[ij].stream,
                        sensorTable[ij].intern_sta, sensorTable[ij].extern_sta,
                        sensorTable[ij].intern_chan, sensorTable[ij].chanID,
                        sensorTable[ij].insType, sensorTable[ij].nomCalib,
                        sensorTable[ij].nomCalibPer,
                        sensorTable[ij].expSampleRateInDB,
                        sensorTable[ij].startTime,
                        sensorTable[ij].endTime, sensorTable[ij].revision,
                        sensorTable[ij].position);
         DEBUG_FOOTER;
      }
   }
   /*@-compdestroy@*/
   return(sensorTable);
   /*@=compdestroy@*/
}

/**
 * @short Wrapper function for ODBCRecycleIds
 *
 * @description This function recycles the given ID list.
 * @description This is a wrapper function for ODBCRecycleIds
 * @description protected by a mutex lock
 *
 * @param wfids      Array holding record ID's
 * @param nbr        ID count
 * @param pAppMutex  Pointer to application specific mutex
 * @param what       type of record (wfdisc or quality)
 *
 * @return number of recycled ID's on success or -1 on error
 */

int
#ifndef HAVE_LIBIDCODBC
recycleIds(int *rid, int sR, pthread_mutex_t *pAppMutex, int what)
#else /* HAVE_LIBIDCODBC */
recycleIds(long *rid, int sR, pthread_mutex_t *pAppMutex, int what)
#endif
{
   int ret;

   /* Recycling the ID must be protected to make sure that
      no more than one thread try to recycle its rid's at the
      same time or that rid's are recycled when new rid's are
      created */
   if (0 != pthread_mutex_lock(pAppMutex))
   {
      logError("Error while trying recycle rid",
               "pthread_mutex_lock failed");
      return(-1);
   }
   if (WFDISC_PASS == what)
   {
      ret = ODBCRecycleIds(WFID_STRING, rid, sR);
   }
   else
   {
      ret = ODBCRecycleIds(QUALID_STRING, rid, sR);
   }

   if (0 != pthread_mutex_unlock(pAppMutex))
   {
      logError("Error while trying recycle rid",
               "pthread_mutex_unlock failed");
      return(-1);
   }

   return(ret);
}

/**
 * @short Get next free ID
 *
 * @description This function returns the next free ID for the given
 * @description identification string. This is a wrapper function for
 * @description ODBCGetNextId protected by a mutex lock
 *
 * @param whatID       ID identification in DB
 * @param pConfigData  pointer to configuration data
 *
 * @return next free ID on success or -1 on error
 */

long
getNextID(/*@null@*/SQLHDBC hdbc, const char *whatID,
          configParams_t *pConfigData)
{
   long nextID;
   int retries;

   if (NULL == hdbc)
   {
      logError("Error while trying to get ID", "ID database handle == NULL");
      return(-1L);
   }
   DEBUG2(3, "ID_HANDLE=%lx",(unsigned long)hdbc);

   for (retries = 0; retries < 3; retries++)
   {
      /* ID generation must be protected by a mutex lock, otherwise
         it is possible that more than one thread try to manipulate the
         internal used ID list */
      if (0 != pthread_mutex_lock(&(pConfigData->appMutex)))
      {
         logError("Error while trying to get ID",
                  "pthread_mutex_lock failed");
         return(-1L);
      }

      nextID =  (long)ODBCGetNextId(hdbc, "lastid", whatID);

      if (0 != pthread_mutex_unlock(&(pConfigData->appMutex)))
      {
         logError("Error while trying to get ID",
                  "pthread_mutex_unlock failed");
         return(-1L);
      }

      /* Check if an ID is available, otherwise retry
         Waiting is done in ODBCGetNextId */
      if (-1L == nextID)
      {
         logWarning("Can't get ID, retry", ODBCErrMsg());
      }
      else break;
   }

   if (-1L == nextID)
   {
      logError("Can't get ID", ODBCErrMsg());
   }
   return(nextID);
}

/**
 * @short Update frame product table
 *
 * @description This function updates the frame product table
 * @description for the given file ID.
 *
 * @param hdbc         database handle
 * @param pLastSend    pointer to the entry in the IRL last sent
 * @param cdtype       CD type of station (CD-1.0 or CD-1.1)
 *
 * @return 0 on success or -1 on error
 */

int
updateFrameProduct(/*@null@*/HDBC hdbc, long frid, uint32_t size, double time)
{
   char buffer[QUERY_BUFFER_SIZE];
#ifndef HAVE_LIBIDCODBC
   char *fName = "updateFrameProduct";
   int ret;
   SQLHSTMT hstmt;
#else
   ODBC_Result result;
#endif

   if (NULL == hdbc) return(0);

   /* Check if the file object owns a frame ID. This is not an
      error condition! */
   if (-1 == frid)
   {
      DEBUG(2, "frid is -1");
      return(0);
   }

   /* Set up the query. Size of bin file grows when new data arrives. The
      endtime might not be correct if not in AS_ARRIVED mode */
   (void)snprintf(buffer, sizeof(buffer),
              "update frameproduct set endtime=%.5f, dsize=%ld where frid=%ld",
               time, (long)size, frid);
   DEBUG(2, buffer);

#ifndef HAVE_LIBIDCODBC
   /* Allocate statement handle */
   ret = ODBCAllocStmt(hdbc, &hstmt);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
      (void)ODBCSetErrorMsg(fName, SQL_HANDLE_DBC, hdbc);
      goto end_error;
   }

   /* Execute the query */
   ret = SQLExecDirect(hstmt, (unsigned char *) buffer, SQL_NTS);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
       (void)ODBCSetErrorMsg(fName, SQL_HANDLE_STMT, hstmt);
       if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
       {
          logWarning("SQLFreeHandle failed", fName);
       }
       goto end_error;
   }

   if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
   {
      logWarning("SQLFreeHandle failed", fName);
   }
#else /* HAVE_LIBIDCODBC */

   if (ODBCExecDirect(hdbc, buffer, &result))
   {
      ODBCFreeResult(&result);
      goto end_error;
   }
   ODBCFreeResult(&result);
#endif /* HAVE_LIBIDCODBC */

   /*@-compdestroy@*/
   /*@-mustfreefresh@*/
   return (0);

end_error:
   shutdownApplication("can't update frameproduct table", ODBCErrMsg());
   return(-1);
   /*@=mustfreefresh@*/
   /*@=compdestroy@*/
}

/**
 * @short Get frid from frameproduct table
 *
 * @description This function queries the frame product table
 * @description for the given file. If the file exists in the frame
 * @description product table it returns the frid, otherwise
 * @description it depends on mode if a new frid is created and
 * @description if the file is inserted in the frame product table.
 *
 * @param hdbc          database handle
 * @param pStation      pointer to station structure
 * @param pathAndFile   path and file name
 * @param type          type of file (binary, waveform, ...)
 * @param mode          ONLY_QUERY or query and insert
 * @param startTime     start time value
 * @param flag          set if a new ID has been created
 *
 * @return next free ID on success or -1 on error
 */

long
getFrameProduct(HDBC hdbc, station_t *pStation, char *pathAndFile,
                int type, int mode, double startTime,
                /*@null@*//*@out@*/int *flag)
{
   char buffer[QUERY_BUFFER_SIZE];
   char *sedot,
#define LEN_FILE_NAME    32
#define LEN_FILE_PATH    64
        *pFile,
        file[LEN_FILE_NAME],
        path[LEN_FILE_PATH];
   int ret;
   long frid = -1;
   ODBC_Result result;
   struct stat statBuf;
#ifndef HAVE_LIBIDCODBC
   char *fName = "getFrameProduct";
#endif
   char *errorTxt = "";
   configParams_t *pConfigData =
                  (struct _configParams *)pStation->pConfigData;

   ST_DEBUG2(2, "FNAP=%s", pathAndFile);
   memset(path, 0, sizeof(path));
   memset(file, 0, sizeof(file));

   type = (int)fpTypeID[type];
   if (NULL != flag) *flag = 0;

   /* Split filename in directory path, file name and
      extension. First search for last "/" in string */
   pFile = strrchr(pathAndFile, '/');
   if (NULL == pFile)
   {
      /* No "/" found */
      path[0] = '.';
      /* Keep in mind that the path might be too long for the DB field! */
      (void)strncpy(file, pathAndFile, sizeof(file) - 1);
   }
   else
   {
      *pFile = '\0'; /* terminate path temporary (replace "/" with NIL) */
      (void)strncpy(path, pathAndFile, sizeof(path));
      *pFile = '/';
      (void)strncpy(file, pFile + 1, sizeof(file) - 1);
   }

   /* search for last "." after last "/" */
   sedot = strrchr(file, '.');
   if (NULL == sedot)
   {
      logWarning("Invalid filename (missing .)", pathAndFile);
      return(0);
   }

   /* set up the query */
   (void)snprintf(buffer, sizeof(buffer),
         "select frid from frameproduct where typeid=%d \
 and dir='%s' and dfile='%s'", type, path, file);

   ret = ODBCGeneralQuery((SQLHDBC)hdbc, buffer, &result);
   DEBUG(2, buffer);

   if (ret < 0)
   {
      /* This is treated as a fatal error. CD2W is not able to
         read the frameproduct records */
      /*@-compdestroy@*/
      /*@-mustfreefresh@*/
      shutdownApplication("Can't read frameproduct records from Database",
                          ODBCErrMsg());
      return(-1L);
      /*@=mustfreefresh@*/
      /*@=compdestroy@*/
   }

#ifndef HAVE_LIBIDCODBC
   if (result.numRows == 0)
#else /* HAVE_LIBIDCODBC */
   if (0 == ODBCFetchNextRow(&result))
#endif
   {
#ifndef HAVE_LIBIDCODBC
      SQLHSTMT hstmt;
#endif
      HDBC idHdbc = (HDBC)getHdbc(pStation->pConfigData, ID_HANDLE);

      DEBUG(2, "No rows returned when reading frameproduct");
      ODBCFreeResult(&result);

      if (ONLY_QUERY == mode)
      {
         /*@-compdestroy@*/
         /*@-mustfreefresh@*/
         /* This is only a request for a check, no insert is made */
         return (-1L);
         /*@=mustfreefresh@*/
         /*@=compdestroy@*/
      }

      /* Don't carry if the file exists, behave like an ID will
         be created in any case */
      if (NULL != flag) *flag = CREATE_ID;

      /* Stat the file to get the file size */
      DEBUG2(2,"Stat [%s]", pathAndFile);
      if (0 != stat(pathAndFile, &statBuf))
      {
         if (IGNORE_NONEXIST == mode)
         {
            /*@-compdestroy@*/
            /*@-mustfreefresh@*/
            /* This is an optional file, e.g. text file */
            return (0L);
            /*@=mustfreefresh@*/
            /*@=compdestroy@*/
         }

         /* TODO: this could also happen then the station type has changed */
         /* Shutdown application */
         (void)snprintf(buffer, sizeof(buffer), "Can't stat [%s]", pathAndFile);
         errorTxt = buffer;
         goto error_shutdown;
      }

      frid = getNextID(idHdbc, FRID_STRING, pConfigData);
      if (-1L == frid)
      {
         /* Shutdown application */
         (void)snprintf(buffer, sizeof(buffer), "Can't get frid for [%s]",
                        pathAndFile);
         errorTxt = buffer;
         goto error_shutdown;
      }
      /* Data provider name is the same as the station name for application
         type CD2W */
      (void)snprintf(buffer, sizeof(buffer),
            "insert into frameproduct values (%ld, %d, '%s', '%s', 0, %ld, \
 %.5f, 0, '%.7s', '-', 'cd2wng-%s', 1, 1, 0, sysdate)",
                     frid, type, path, file, (long)statBuf.st_size, startTime,
                     pStation->stationName, VERSION);
      DEBUG(2, buffer);

#ifndef HAVE_LIBIDCODBC
      /* Allocate statement handle */
      ret = ODBCAllocStmt(hdbc, &hstmt);
      if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
      {
         (void)ODBCSetErrorMsg(fName, SQL_HANDLE_DBC, hdbc);
         /* Shutdown application */
         (void)snprintf(buffer, sizeof(buffer), "Can't allocate stmt for [%s]",
                        pathAndFile);
         errorTxt = buffer;
         goto error_shutdown;
      }

      /* Execute the SQL command */
      ret = SQLExecDirect(hstmt, (unsigned char *) buffer, SQL_NTS);
      if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
      {
         (void)ODBCSetErrorMsg(fName, SQL_HANDLE_STMT, hstmt);
         if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
         {
            logWarning("SQLFreeHandle failed", fName);
         }
         /* Shutdown application */
         (void)snprintf(buffer, sizeof(buffer),
               "Can't insert frameproduct for [%s]", pathAndFile);
         errorTxt = buffer;
         goto error_shutdown;
      }

      if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
      {
         logWarning("SQLFreeHandle failed", fName);
      }
#else /* HAVE_LIBIDCODBC */
      /* Result of previous select has been freed already */
      /* Add new record to frameproduct table */
      if (ODBCExecDirect(hdbc, buffer, &result))
      {
         ODBCFreeResult(&result);
         /* Shutdown application */
         (void)snprintf(buffer, sizeof(buffer),
               "Can't insert frameproduct for [%s]", pathAndFile);
         errorTxt = buffer;
         goto error_shutdown;
     }
     ODBCFreeResult(&result);
#endif /* HAVE_LIBIDCODBC */

     /*@-compdestroy@*/
     /*@-mustfreefresh@*/
     return(frid);
     /*@=compdestroy@*/
     /*@=mustfreefresh@*/
   }
#ifndef HAVE_LIBIDCODBC
   else if (result.numRows > 1)
   {
      ODBCFreeResult(&result);

      /* Shutdown application */
      (void)snprintf(buffer, sizeof(buffer),
                     "More than one frid found for [%s]", pathAndFile);
      errorTxt = buffer;
      goto error_shutdown;
   }
   else
   {
      long *fridl;
      double *fridd;
      /* Read now result from SQL query and set up the mapping */
      DEBUG2(5, "SQL type for frid [%d]",
             (int)result.column[0].dataType);
      if (result.column[0].dataType == SQL_DECIMAL ||
          result.column[0].dataType == SQL_NUMERIC ||
          result.column[0].dataType == SQL_FLOAT ||
          result.column[0].dataType == SQL_REAL ||
          result.column[0].dataType == SQL_DOUBLE)
      {
         fridd = (double *)result.column[0].value;
         frid = (long)fridd[0];
      }
      else if (result.column[0].dataType == SQL_INTEGER ||
               result.column[0].dataType == SQL_BIGINT)
      {
         /* Due a bug in ODBCGetResultCols SQL_INTEGER is mapped
            to long instead of SQLINTEGER and SQL_SMALLINT is
            mapped to int instead of SQLSMALLINT */
         fridl = (long *)result.column[0].value;
         frid = fridl[0];
      }
      else
      {
         ODBCFreeResult(&result);
         DEBUG2(2, "Unknown SQL type for frameproduct id [%d]",
                (int)result.column[0].dataType);
         /* Shutdown application */
         errorTxt = "Can't determine chanID SQL type";
         goto error_shutdown;
      }

      ODBCFreeResult(&result);
   }
#else /* HAVE_LIBIDCODBC */
   else
   {
      /* FRID found, read it */
      ret = ODBCGetAsLong(&result, "frid", &frid, NULL);
      if (0 != ret)
      {
         ODBCFreeResult(&result);
         (void)snprintf(buffer, sizeof(buffer), "Can't get frid [%d]", ret);
         errorTxt = buffer;
         goto error_shutdown;
      }

      /* Check for more than one result */
      if (1 == ODBCFetchNextRow(&result))
      {
         ODBCFreeResult(&result);

         /* Shutdown application */
         (void)snprintf(buffer, sizeof(buffer),
                        "More than one frid found for [%s]", pathAndFile);
         errorTxt = buffer;
         goto error_shutdown;
      }
      ODBCFreeResult(&result);
   }
#endif /* HAVE_LIBIDCODBC */

   /*@-compdestroy@*/
   /*@-mustfreefresh@*/
   return (frid);

error_shutdown:
   shutdownApplication(errorTxt, ODBCErrMsg());
   return(-1L);
   /*@=mustfreefresh@*/
   /*@=compdestroy@*/
}


/**
 * @short Commit changes in DB
 *
 * @description This function commits changes in the DB
 *
 * @param hdbc         database handle
 *
 * @return wfid on success or -1 on error
 */

int
commitChanges(SQLHDBC hdbc, configParams_t *pConfigData)
{
   SQLRETURN ret;

   /* It is not allowed to commit changes when the ID generation has a
      'cursor open for update', otherwise a database error is raised!! */
   if (0 != pthread_mutex_lock(&(pConfigData->appMutex)))
   {
      logError("Error while trying to get ID",
               "pthread_mutex_lock failed");
      return(-1);
   }

   ret = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT);

   if (0 != pthread_mutex_unlock(&(pConfigData->appMutex)))
   {
      logError("Error while trying to get ID",
               "pthread_mutex_unlock failed");
      return(-1);
   }

   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
#ifndef HAVE_LIBIDCODBC
      (void)ODBCSetErrorMsg("commitChanges", SQL_HANDLE_DBC, hdbc);
#endif
      logError("Can't commit changes", ODBCErrMsg());
      return(-1);
   }
   return(0);
}

/**
 * @short load quality records from DB
 *
 * @description This function loads all quality records for a given
 * @description day from the database.
 *
 * @param pStation        pointer to station structure
 * @param loadDate        date of quality records to load
 * @param pWaveFormHead   pointer to waveform header
 * @param dayInWindow     day in time window
 *
 * @return NO_ERROR on success or event which caused the error
 */

uint16_t
loadQualityRecordsFromDB(station_t *pStation, uint32_t loadDate,
                         waveFormHead_t *pWaveFormHead, uint32_t dayInWindow)
{
   int ii,
       ret;
   ODBC_Result result;
#ifndef HAVE_LIBIDCODBC
   char **sta;
   char **chan;
   double *chanIDd = NULL;
   long   *chanIDl = NULL;
   double *nsamplesd = NULL;
   long   *nsamplesl = NULL;
   double *qualidd = NULL;
   long   *qualidl = NULL;
   double *firstslotd = NULL;
   long   *firstslotl = NULL;
   double *qualflagd = NULL;
   long   *qualflagl = NULL;
   double *foffd = NULL;
   long   *foffl = NULL;
   double *attributesd = NULL;
   long   *attributesl = NULL;
   double *attMaskd = NULL;
   long   *attMaskl = NULL;
   double *pMean = NULL;
   double *pMaxtime;
   double *pAvgtime;
   double *pMintime;
   double *pStartTime;
   double *pEndTime;
   double *pM2;
#else /* HAVE_LIBIDCODBC */
   double pMintime[1];
   double pAvgtime[1];
   double pMaxtime[1];
   double pEndTime[1];
#endif /* HAVE_LIBIDCODBC */
   char buffer[QUERY_BUFFER_SIZE];
   char *dfile;
   int posInChMask;
   channelHead_t *pChannelHead;
   CDdataChannel_t *channel;
   segmentHead_t *pSegHeader;
   channelWaveSeg_t *pSegment;
   dayHeader_t *pDay;
   size_t lenOfBuffer;
   char *errorTxt = "";
   void *hdbc;
   char pathAndFile[FILE_NAME_SIZE];
   uint16_t event = NO_ERROR;
   long wffId;
   char *fName = "loadQualityRecordsFromDB";

   hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);

   if (NULL == hdbc)
   {
      /* No DB connection */
      shutdownApplication("Internal error: No DB handle available", fName);
      return(E_INTERNAL);
   }

   /* Set up file name and path of waveform file */
   lenOfBuffer = sizeof(pathAndFile);
   ret = getFileName(pStation, (time_t)loadDate, CREATE_USR_FILE, pathAndFile,
                     &lenOfBuffer);

   if (-1 == ret)
   {
      /* Oops! The buffer is too small to hold the whole
         file path plus the file name or the file type was wrong.
         A shutdown is initiated */
      shutdownApplication("Internal error: buffer too small", pathAndFile);
      return(E_INTERNAL);
   }

   /* Query frameproduct table for waveform file */
   wffId = getFrameProduct((SQLHDBC)hdbc, pStation, pathAndFile, CD_MWF_FILE,
                            ONLY_QUERY, (double)loadDate, NULL);
   if (-1L == wffId)
   {
      DEBUG2(2, "[%s] not found in frameproduct", pathAndFile);
      return(NO_ERROR);
   }
   else
   {
      /* Save the waveform file ID */
      pWaveFormHead->waveFileID[dayInWindow] = wffId;
   }

   /* Split filename into path and file */
   pathAndFile[lenOfBuffer] = '\0';
   dfile = &pathAndFile[lenOfBuffer + 1];

   /* set up the query */
   /* NOTE: The select criteria start and end time is not optimal,
      because the start time or end time might be one second in the
      previous/next day. Better is to select based on the file name */
   (void) snprintf(buffer, sizeof(buffer),
         "select sta, chan, time, qualid, chanid, endtime, nsamp, firstslot, \
 qualityflags, mintime, maxtime, avgtime, meancounts, m2, \
 chanstatus, chanstatusmask, foff \
 from cdquality where wffid = %ld order by time",  wffId);
   DEBUG(2, buffer);
   ret = ODBCGeneralQuery((SQLHDBC)hdbc, buffer, &result);

   if (ret < 0)
   {
      /* This is treated as a fatal error. CD2W is not able to
         read the quality records */
      /*@-compdestroy@*/
      /*@-mustfreefresh@*/
      shutdownApplication("Can't load cdquality records from Database",
                          ODBCErrMsg());
      return(E_INTERNAL);
      /*@=mustfreefresh@*/
      /*@=compdestroy@*/
   }

#ifndef HAVE_LIBIDCODBC
   if (result.numRows == 0)
#else /* HAVE_LIBIDCODBC */
   if (SQL_NO_DATA_FOUND == result.sqlRetCode)
#endif /* HAVE_LIBIDCODBC */
   {
      DEBUG2(2, "No rows returned for %s", dfile);
      /*@-compdestroy@*/
      ODBCFreeResult(&result);
      return(NO_ERROR);
      /*@=compdestroy@*/
   }

   channel = &(pWaveFormHead->tmpChannel);

#ifndef HAVE_LIBIDCODBC
   /* Read now result from SQL query and set up the mapping
      in the channel cache */
   sta = (char **)result.column[0].value;
   chan = (char **)result.column[1].value;
   pStartTime = (double *)result.column[2].value;

   if (result.column[4].dataType == SQL_DECIMAL ||
       result.column[4].dataType == SQL_NUMERIC ||
       result.column[4].dataType == SQL_FLOAT ||
       result.column[4].dataType == SQL_REAL ||
       result.column[4].dataType == SQL_DOUBLE)
   {
      /* Driver maps number(x) to float or double */
      qualidd = (double *)result.column[3].value;
      chanIDd = (double *)result.column[4].value;
      nsamplesd = (double *)result.column[6].value;
      firstslotd = (double *)result.column[7].value;
      qualflagd = (double *)result.column[8].value;
      attributesd = (double *)result.column[14].value;
      attMaskd = (double *)result.column[15].value;
      foffd = (double *)result.column[16].value;

   }
   else if (result.column[4].dataType == SQL_INTEGER &&
           (result.column[3].dataType == SQL_INTEGER ||
            result.column[3].dataType == SQL_BIGINT))
   {
      /* Number(8) needs at least 32 bits, number(10) at least 34 bits,
         but both are mapped to long in gODBC */
      qualidl = (long *)result.column[3].value;
      chanIDl = (long *)result.column[4].value;
      nsamplesl = (long *)result.column[6].value;
      firstslotl = (long *)result.column[7].value;
      qualflagl = (long *)result.column[8].value;
      attributesl = (long *)result.column[14].value;
      attMaskl = (long *)result.column[15].value;
      foffl = (long *)result.column[16].value;
   }
   else
   {
      /* TODO: change also wfdisc reading */
      DEBUG3(2, "Unknown SQL type for chanID [%d] or qualID [%d]",
             (int)result.column[4].dataType, (int)result.column[3].dataType);
      errorTxt = "Can't determine chanID SQL type";
      /* Shutdown application */
      event = E_LOAD_RECORD;
      goto error_shutdown;
   }
   pEndTime = (double *)result.column[5].value;
   pMintime = (double *)result.column[9].value;
   pMaxtime = (double *)result.column[10].value;
   pAvgtime = (double *)result.column[11].value;
   pMean = (double *)result.column[12].value;
   pM2 = (double *)result.column[13].value;

   /* Read quality records */
   for (ii = 0; ii < result.numRows; ii++)
#else /* HAVE_LIBIDCODBC */
   ii = 0;
   while (ODBCFetchNextRow(&result))
#endif /* HAVE_LIBIDCODBC */
   {
      uint32_t startSlot;
      int retVal;
      long qualid;
      uint16_t segQual;
      uint32_t slotsOccupied,
               foff,
               attributes,
               segStatusBitMask;
      long samplesInSegment,
           chanId;
      double segMean,
             segM2;

#ifndef HAVE_LIBIDCODBC
      /* Fill the channel description */
      (void)strncpy(channel->siteName, sta[ii], LEN_SITE_NAME);
      (void)strncpy(channel->channelName, chan[ii], LEN_CHANNEL_NAME);
      if (result.column[4].dataType == SQL_INTEGER)
      {
         /*@-nullderef@*/
         chanId = chanIDl[ii];
         samplesInSegment = nsamplesl[ii];
	 qualid = qualidl[ii];
         startSlot = (uint32_t)(firstslotl[ii]);
         segQual = (uint16_t)(qualflagl[ii]);
         attributes = (uint32_t)(attributesl[ii]);
         segStatusBitMask = (uint32_t)(attMaskl[ii]);
         foff = (uint32_t)(foffl[ii]);
      }
      else
      {
         chanId = (long)chanIDd[ii];
         samplesInSegment = (long)nsamplesd[ii];
	 qualid = (long)qualidd[ii];
         startSlot = (uint32_t)(firstslotd[ii]);
         segQual = (uint16_t)(qualflagd[ii]);
         attributes = (uint32_t)(attributesd[ii]);
         segStatusBitMask = (uint32_t)(attMaskd[ii]);
         foff = (uint32_t)(foffd[ii]);
         /*@=nullderef@*/
      }
      channel->timeStamp = pStartTime[ii];
      segMean = pMean[ii];
      /* convert string to long long */
      segM2 = pM2[ii];

#else /* HAVE_LIBIDCODBC */
      /*@-compdef@*/
      reqColumns_t reqColumns[] = {
       {"sta", channel->siteName, CT_STRING, sizeof(channel->siteName), -1},
       {"chan", channel->channelName, CT_STRING,
                sizeof(channel->channelName), -1},
       {"time", &(channel->timeStamp), CT_DOUBLE,
                sizeof(channel->timeStamp), -1},
       {"qualid", &qualid, CT_LONG, sizeof(qualid), -1},
       {"chanid", &chanId, CT_LONG, sizeof(chanId), -1},
       {"endtime", &pEndTime[0], CT_DOUBLE, sizeof(pEndTime[0]), -1},
       {"nsamp", &samplesInSegment, CT_LONG, sizeof(samplesInSegment), -1},
       {"firstslot", &startSlot, CT_INT32U, sizeof(startSlot), -1},
       {"qualityflags", &segQual, CT_INT16U, sizeof(segQual), -1},
       {"mintime", &pMintime[0], CT_DOUBLE, sizeof(pMintime[0]), -1},
       {"maxtime", &pMaxtime[0], CT_DOUBLE, sizeof(pMaxtime[0]), -1},
       {"avgtime", &pAvgtime[0], CT_DOUBLE, sizeof(pAvgtime[0]), -1},
       {"meancounts", &segMean, CT_DOUBLE, sizeof(segMean), -1},
       {"m2", &segM2, CT_DOUBLE, sizeof(segM2), -1},
       {"chanstatus", &attributes, CT_INT32U, sizeof(attributes), -1},
       {"chanstatusmask", &segStatusBitMask, CT_INT32U,
                          sizeof(segStatusBitMask), -1},
       {"foff", &foff, CT_INT32U, sizeof(foff), -1} };
      /*@=compdef@*/

      /* for splint */
      segMean = segM2 = 0.0;
      pEndTime[0] = pMintime[0] = pMaxtime[0] = pAvgtime[0] = 0.0;
      startSlot = foff = attributes = segStatusBitMask = 0;
      qualid = samplesInSegment = chanId = 0;
      segQual = 0;

      /* Read result row */
      ret = ODBCGetColumns(sizeof(reqColumns),reqColumns, &result);

#endif /* HAVE_LIBIDCODBC */
      IF_DEBUG (2)
      {
         S_DEBUG_HDR;
         (void) fprintf(DEBUG_FP,
          "load %.6s %.8s %f %ld %ld %f %ld %u %hu %u %u %u %f %f %f %lf %lf\n",
                     channel->siteName, channel->channelName,
                     channel->timeStamp, qualid, chanId,
                     pEndTime[ii], samplesInSegment, /* Timeliness threshold, */
                     (unsigned int)startSlot, (unsigned short int)segQual,
                     (unsigned int)attributes, (unsigned int)segStatusBitMask,
                     (unsigned int)foff, pMintime[ii], pMaxtime[ii],
                     pAvgtime[ii], /* mean value, */
                     segMean, /* segRms */
                     segM2);
          DEBUG_FOOTER;
      }

#if HAVE_LIBIDCODBC
      /* Check return code from ODBCGetColumns */
      if (ret != 0)
      {
         (void)snprintf(buffer, sizeof(buffer),
                 "Can't ODBCGetColumns cdquality record for ID=%ld", qualid);
         errorTxt = buffer;
         /* Shutdown application */
         event = E_LOAD_RECORD;
         goto error_shutdown;
      }
#endif /* HAVE_LIBIDCODBC */

      /* Now the channel mask should contain an entry for each channel
       * found in the quality records. The problem is, that the waveform
       * file contains external channel names, whereas the quality records
       * contain internal channel names. Therefore the channel mask has to
       * be searched using the external channel name if necessary. It is
       * required that cd2w has at least the same information when it is
       * re-started as it has in the previous run!!
       */
      posInChMask = checkChMaskPosition(channel->siteName,
                                        channel->channelName,
                                        &(pStation->chCache));
      if (-1 == posInChMask)
      {
         ST_DEBUG4(1, "No channel index for %.6s/%.8s - %s",
                   channel->siteName, channel->channelName, fName);
         errorTxt = "No channel index found while reading DB";
         /* Shutdown application */
         event = E_INTERNAL;
         goto error_shutdown;
      }

      pChannelHead = pWaveFormHead->channelHeads[posInChMask];

      /* Check if the channel header exists, otherwise stop
         application. Channels are created in loadChannelDescription()
         when the wfdisc records are loaded and the quality record
         can't refer to data which is not referenced by a wfdisc record */
      if (NULL == pChannelHead)
      {
         /* This happens when the quality records but not the wfdisc records
            have been written (e.g. no wfdisc IDs available). Theoretically
            this should not happen because no quality records are written
            when a problem occurs during saving of wfdisc records */
         (void)snprintf(buffer, sizeof(buffer),
         "No channel header for %.6s/%.8s - %ld while loading cdquality record",
                channel->siteName, channel->channelName, qualid);
         logError(buffer, fName);
         /* Shutdown application */
         errorTxt = "No channel header found while reading cdquality record";
         event = E_LOAD_RECORD;
         goto error_shutdown;
      }

      /* At this point the channel is checked and the quality record
         can be loaded */
      if (startSlot > ONE_DAY)
      {
         /* TODO: This value might be off by one slot!!
            If the channel is sorted by time, that segment must have
            a left overlap with the preceding or a right overlap
            with the following segment  */
         startSlot = ((uint32_t)channel->timeStamp) - loadDate;
	 ST_DEBUG(2, "too high start slot");
      }

      /* Compute number of occupied slots */
      slotsOccupied = (uint32_t)(samplesInSegment)/pChannelHead->expSampleRate;

      pDay = &(pChannelHead->day[dayInWindow]);
      pSegHeader = &(pDay->qualSeg);

      /* Get a segment. There is no need to merge it with other
         segments because a quality record has always the largest
         possible size and this day has no quality records processed
         so far */
      pSegment = getSegment(pSegHeader, (uint32_t)startSlot,
                            startSlot + slotsOccupied - 1, &retVal);

      if (NULL == pSegment)
      {
         /* Shut down is started in the called function */
         DEBUG4(1, "Attempt to get segment [%.6s/%.8s] [%d] failed",
                channel->siteName, channel->channelName, retVal);
         errorTxt = "Attempt to get segment failed";
         event = E_NO_MEM;
         goto error_shutdown;
      }

      IF_DEBUG(2)
      {
         S_DEBUG_HDR;
         (void) fprintf(DEBUG_FP,
                "Get q-segment [%.6s/%.8s] [%d] [%d] [%u] [%d] [%d]\n",
                channel->siteName, channel->channelName, retVal,
                (int)startSlot, (unsigned int)slotsOccupied,
                (int)pSegment->firstSlot, (int)pSegment->lastSlot);

         DEBUG_FOOTER;
      }

      /* Check if a new quality segment has to be created */
      if (NEW_SEG != retVal)
      {
         /* This happens when the quality flag is different or the
            value range for the square counts is exceeded */
         DEBUG3(2, "Must create new q-segment: [%.8s/%.4s]",
                    channel->siteName, channel->channelName);
         pSegment = createSegment(pSegHeader, pSegment);
         if (NULL == pSegment)
         {
            /* Shutdown is down in createSegment() */
            errorTxt = "Attempt to create segment failed";
            event = E_NO_MEM;
            goto error_shutdown;
         }
      }

      pSegment->segState = SEG_NORMAL;
      pSegment->segStartTime = channel->timeStamp;
      /* Internal used end time is the real one, but
         end time in quality record is the start time of
         the last sample. Keep also in mind that the real sample
         rate is used when the quality record has been written, but
         the expected is used here. This might cause a very small
         difference when the segments are compared (in the debug output) */
      pSegment->segEndTime = pEndTime[ii] + 1.0/pChannelHead->expSampleRate;
      pSegment->samplesInSeg = (uint32_t)samplesInSegment;
      pSegment->offsetWaveform = foff;
      pSegment->firstSlot = (uint32_t)startSlot;
      pSegment->lastSlot = (uint32_t)(startSlot + slotsOccupied - 1);
      pSegment->calib = 1.0;
      pSegment->calper = 0.0;
      pSegment->attributes = attributes;
      pSegment->segStatusBitMask = segStatusBitMask;
      pSegment->subrecord.rid[0] = qualid;
      pSegment->segQual = segQual;
      pSegment->segMinTimeliness = pMintime[ii];
      pSegment->segMaxTimeliness = pMaxtime[ii];
      pSegment->segAvgTimeliness = pAvgtime[ii];
      pSegment->segMean = segMean;
      pSegment->segM2 = segM2;

      /* Check for pending shutdown */
      if (pStation->shutdownEnable != SD_NONE) { break; }
   }

   /*@-compdestroy@*/
   ODBCFreeResult(&result);
   return(NO_ERROR);
   /*@=compdestroy@*/

error_shutdown:

   /*@-compdestroy@*/
   ODBCFreeResult(&result);
   shutdownApplication(errorTxt, fName);
   return(event);
   /*@=compdestroy@*/
   /*@-mustfreefresh@*/
   /*@=mustfreefresh@*/
}

/**
 * @short load wfdisc from DB
 *
 * @description This function loads all wfdisc records for a given
 * @description day from the database. This function must be called before a
 * @description new day is merged or the channel offset might be incorrect!!
 *
 * @param pStation        pointer to station structure
 * @param loadDate        date of wfdisc records to load
 * @param pWaveFormHead   pointer to waveform header
 *
 * @return NO_ERROR on success or event which caused the error
 */

uint16_t
loadWfdiscRecordsFromDB(station_t *pStation, uint32_t loadDate,
                        waveFormHead_t *pWaveFormHead)
{
   int ii,
       ret;
   ODBC_Result result;
   uint16_t event = NO_ERROR;
#ifndef HAVE_LIBIDCODBC
   char **sta;
   char **chan;
   double *chanIDd = NULL;
   long   *chanIDl = NULL;
   double *nsamplesd = NULL;
   long   *nsamplesl = NULL;
   double *wfidd = NULL;
   long   *wfidl = NULL;
   double *offsetd = NULL;
   long   *offsetl = NULL;
   double *calib;
   double *calper;
   double *sampleRate;
   double *startTime;
   double *endTime;
   char **datatype,
        **clipped;
#else /* HAVE_LIBIDCODBC */
   char clipped[1][2];
   char datatype[1][3];
   double endTime[1];
   float sampleRate[1];
#endif /* HAVE_LIBIDCODBC */
   long   wfid;
   char buffer[QUERY_BUFFER_SIZE];
   char *dir,
        *dfile;
   int isFirst = 1,
       dayInWindow,
       posInChMask;
   long samplesInSegment,
        chanId;
   channelHead_t *pChannelHead;
   CDdataChannel_t *channel;
   char pathAndFile[FILE_NAME_SIZE];
   size_t lenOfBuffer;
   void *hdbc;

   hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);

   if (NULL == hdbc)
   {
      /* No DB connection */
      shutdownApplication("internal error in loadWfdiscRecordsFromDB",
                          "No DB handle available");
      return(E_INTERNAL);
   }

   /* Compute day in time window */
   dayInWindow = (int)(loadDate - pWaveFormHead->startTimeWindow)/ONE_DAY;
   if (dayInWindow >= MAX_MERGE_DAYS)
   {
      shutdownApplication("Invalid dayInWindow", "loadWfdiscRecordsFromDB");
      return(E_INTERNAL);
   }

   /* Set up file name and path of waveform file */
   lenOfBuffer = sizeof(pathAndFile);
   ret = getFileName(pStation, (time_t)loadDate, CREATE_USR_FILE, pathAndFile,
                     &lenOfBuffer);

   if (-1 == ret)
   {
      /* Oops! The buffer is too small to hold the whole
         file path plus the file name or the file type was wrong.
         A shutdown is initiated */
      shutdownApplication("internal error in loadWfdiscRecordsFromDB",
                          pathAndFile);
      return(E_INTERNAL);
   }

   /* Split filename into path and file */
   pathAndFile[lenOfBuffer] = '\0';

   /* Set dir and dfile for wfdiscs. Absolute wfdisc paths
      are not longer used to simplify configuration. Symbolic
      links or a proper filePath should be used instead */
   dir = pathAndFile;
   dfile = &pathAndFile[lenOfBuffer + 1];

   /* set up the query */
   (void) snprintf(buffer, sizeof(buffer),
         "select sta, chan, time, wfid, chanid, endtime, nsamp, samprate, \
 calib, calper, datatype, clip, foff from wfdisc \
 where dir = '%s' and dfile = '%s' order by time", dir, dfile);
   DEBUG(2, buffer);
   ret = ODBCGeneralQuery((SQLHDBC)hdbc, buffer, &result);

   if (ret < 0)
   {
      /* This is treated as a fatal error. CD2W is not able to
         read the wfdisc records */
      /*@-compdestroy@*/
      /*@-mustfreefresh@*/
      shutdownApplication("Can't load wfdisc records from Database",
                          ODBCErrMsg());
      return(E_LOAD_RECORD);
      /*@=mustfreefresh@*/
      /*@=compdestroy@*/
   }

#ifndef HAVE_LIBIDCODBC
   if (result.numRows == 0)
#else /* HAVE_LIBIDCODBC */
   if (SQL_NO_DATA_FOUND == result.sqlRetCode)
#endif
   {
      DEBUG(2, "No rows returned");
      /*@-compdestroy@*/
      ODBCFreeResult(&result);
      return(NO_ERROR);
      /*@=compdestroy@*/
   }

   channel = &(pWaveFormHead->tmpChannel);

#ifndef HAVE_LIBIDCODBC
   /* Read now result from SQL query and set up the mapping
      in the channel cache */
   sta = (char **)result.column[0].value;
   chan = (char **)result.column[1].value;
   startTime = (double *)result.column[2].value;

   DEBUG2(5, "SQL type for wfid [%d]",
          (int)result.column[3].dataType);
   if (result.column[4].dataType == SQL_DECIMAL ||
       result.column[4].dataType == SQL_NUMERIC ||
       result.column[4].dataType == SQL_FLOAT ||
       result.column[4].dataType == SQL_REAL ||
       result.column[4].dataType == SQL_DOUBLE)
   {
      chanIDd = (double *)result.column[4].value;
      nsamplesd = (double *)result.column[6].value;
      wfidd = (double *)result.column[3].value;
      offsetd = (double *)result.column[12].value;

   }
   else if (result.column[4].dataType == SQL_INTEGER)
   {
      /* Due a bug in ODBCGetResultCols SQL_INTEGER is mapped
         to long instead of SQLINTEGER and SQL_SMALLINT is
         mapped to int instead of SQLSMALLINT and SQL_BIGINT is
         mapped to long instead of SQLBIGINT */
      chanIDl = (long *)result.column[4].value;
      nsamplesl = (long *)result.column[6].value;
      wfidl = (long *)result.column[3].value;
      offsetl = (long *)result.column[12].value;
   }
   else
   {
      (void)snprintf(buffer, sizeof(buffer),
                     "Unknown SQL type for chanID [%d]",
                    (int)result.column[4].dataType);
      /* Shutdown application */
      goto error_shutdown;
   }
   endTime = (double *)result.column[5].value;
   sampleRate = (double *)result.column[7].value;
   calib = (double *)result.column[8].value;
   calper = (double *)result.column[9].value;
   datatype = (char **)result.column[10].value;
   clipped = (char **)result.column[11].value;

   /* Read wfdisc records */
   for (ii = 0; ii < result.numRows; ii++)
#else /* HAVE_LIBIDCODBC */
   ii = 0;
   while (ODBCFetchNextRow(&result))
#endif
   {
      uint32_t startSlot,
               channelTime;
      int retVal;

#ifndef HAVE_LIBIDCODBC
      /* Fill the channel description */
      (void)strncpy(channel->siteName, sta[ii], LEN_SITE_NAME);
      (void)strncpy(channel->channelName, chan[ii], LEN_CHANNEL_NAME);
      if (result.column[4].dataType == SQL_INTEGER)
      {
         /*@-nullderef@*/
         chanId = chanIDl[ii];
         samplesInSegment = nsamplesl[ii];
	 wfid = wfidl[ii];
         channel->channelOffset = (uint32_t)(offsetl[ii]);
      }
      else
      {
         chanId = (long)chanIDd[ii];
         samplesInSegment = (long)nsamplesd[ii];
	 wfid = (long)wfidd[ii];
         channel->channelOffset = (uint32_t)(offsetd[ii]);
         /*@=nullderef@*/
      }
      channel->calibFactor = (float)calib[ii];
      channel->calibPeriod = (float)calper[ii];
      channel->timeStamp = startTime[ii];

#else /* HAVE_LIBIDCODBC */
      /*@-compdef@*/
      reqColumns_t reqColumns[] = {
       {"sta", channel->siteName, CT_STRING, sizeof(channel->siteName), -1},
       {"chan", channel->channelName, CT_STRING,
                sizeof(channel->channelName), -1},
       {"time", &(channel->timeStamp), CT_DOUBLE,
                sizeof(channel->timeStamp), -1},
       {"wfid", &wfid, CT_LONG, sizeof(wfid), -1},
       {"chanid", &chanId, CT_LONG, sizeof(chanId), -1},
       {"endtime", &endTime[0], CT_DOUBLE, sizeof(endTime[0]), -1},
       {"nsamp", &samplesInSegment, CT_LONG, sizeof(samplesInSegment), -1},
       {"samprate", &sampleRate[0], CT_FLOAT, sizeof(sampleRate[0]), -1},
       {"calib", &(channel->calibFactor), CT_FLOAT,
                  sizeof(channel->calibFactor), -1},
       {"calper", &(channel->calibPeriod), CT_FLOAT,
                  sizeof(channel->calibPeriod), -1},
       {"datatype", datatype[0], CT_STRING, sizeof(datatype[0]), -1},
       {"clip", clipped[0], CT_STRING, sizeof(clipped[0]), -1},
       {"foff", &(channel->channelOffset), CT_INT32U,
                  sizeof(channel->channelOffset), -1} };
      /*@=compdef@*/

      /* for splint */
      wfid = chanId = samplesInSegment = 0;
      endTime[0] = sampleRate[0] = 0.0;
      datatype[0][0] = clipped[0][0] = '\0';

      /* Read result row */
      ret = ODBCGetColumns(sizeof(reqColumns),reqColumns, &result);

#endif /* HAVE_LIBIDCODBC */

      IF_DEBUG (2)
      {
         S_DEBUG_HDR;
         (void) fprintf(DEBUG_FP,
         "load record %.6s %.8s %f %ld %ld %f %ld %f %f %f %.2s %c %s %s %ld\n",
                        channel->siteName, channel->channelName,
                        channel->timeStamp, wfid, chanId,
                        /* jdate */ endTime[ii], samplesInSegment,
                        sampleRate[ii],
                        channel->calibFactor, channel->calibPeriod,
                        /* instype */ /* segtype */ datatype[ii],
                        clipped[ii][0], dir, dfile,
                        (long)channel->channelOffset
                        /* commid */ /* lddate */);
          DEBUG_FOOTER;
      }

#if HAVE_LIBIDCODBC
      /* Check return code from ODBCGetColumns */
      if (ret != 0)
      {
         (void)snprintf(buffer, sizeof(buffer),
                       "Can't ODBCGetColumns wfdisc record for ID=%ld", wfid);
         /* Shutdown application */
         goto error_shutdown;
      }
#endif /* HAVE_LIBIDCODBC */

      /* This are some common bugs in records which should not occur
         (theoretically, at least) */
      if (samplesInSegment <= 0)
      {
         logError("0 samples in record, ignoring invalid wfdisc",
                    pStation->conID);
         continue;
      }
      /*@-realcompare@*/
      if (channel->timeStamp == endTime[ii])
      {
         logError("start == end time, ignoring invalid wfdisc record",
                    pStation->conID);
         continue;
      }
      /*@=realcompare@*/

      /* At least one wfdisc record exists. There must be a
         corresponding waveform file and the channel information
         in the waveform file has to be read before the first wfdisc
         record is merged */
      if (isFirst)
      {
         /* Read channel info from waveform file and create all
            channels found in the waveform file. This step is necessary
            to have the correct channel start offset in all possible cases */
         if (loadChannelDescription(pStation, loadDate, pWaveFormHead,
                                    dayInWindow))
         {
            (void)snprintf(buffer, sizeof(buffer),
                          "Can't load channel description for wfid=%ld",
                          wfid);
            /* Shutdown application */
            goto error_shutdown;
         }
         isFirst = 0;
      }

      /* Now the channel mask should contain an entry for each channel
       * found in the wfdisc records. The problem is, that the waveform
       * file contains external channel names, whereas the wfdisc records
       * contain internal channel names. Therefore the channel mask has to
       * be searched using the external channel name if necessary. It is
       * required that cd2w has at least the same information when it is
       * re-started as it has in the previous run!!
       */
      posInChMask = checkChMaskPosition(channel->siteName,
                                        channel->channelName,
                                        &(pStation->chCache));
      if (-1 == posInChMask)
      {
         (void)snprintf(buffer, sizeof(buffer),
                        "No channel index for %.6s/%.8s",
                        channel->siteName, channel->channelName);
         /* Shutdown application */
         goto error_shutdown;
      }

      pChannelHead = pWaveFormHead->channelHeads[posInChMask];

      /* Check if the channel header exists, otherwise stop
         application. Channels are created in loadChannelDescription()
         and the wfdisc record can't refer to non existing channels */
      if (NULL == pChannelHead)
      {
         (void)snprintf(buffer, sizeof(buffer),
                        "No channel header found for %.6s/%.8s",
                        channel->siteName, channel->channelName);
         /* Shutdown application */
         goto error_shutdown;
      }

      /* At this point the channel is initialized and the wfdisc
         can be loaded. First the slot is computed from the offset.
         If the computed slot is higher than slots per day, than
         the slot is computed using the channel start time and the
         segment is marked as not merge-able. */
      startSlot = ((uint32_t)(channel->channelOffset) -
                   pChannelHead->day[dayInWindow].channelWavOffset) /
                   (pChannelHead->expSampleRate * pWaveFormHead->sampleSize);

      ST_DEBUG3(3, "Computed start slot [%d/%d]", (int)startSlot,
                   (int)(((uint32_t)channel->timeStamp) - loadDate));
      if (startSlot > ONE_DAY)
      {
         /* TODO: This value might be off by one slot!!
            If the channel is sorted by time, that segment must have
            a left overlap with the preceding or a right overlap
            with the following segment  */
         startSlot = ((uint32_t)channel->timeStamp) - loadDate;
	 ST_DEBUG(2, "too high start slot");
      }

      /* Compute time this wfdisc record encompasses in milliseconds */
      channelTime = ((uint32_t)(samplesInSegment) / pChannelHead->expSampleRate
                                * 1000);
      channel->nbrOfSamples = (uint32_t)samplesInSegment;

      retVal = tryToMerge(channel, 0, channelTime, pChannelHead, startSlot,
                          pStation, pWaveFormHead, 0, dayInWindow, 1,
                          clipped[ii][0], endTime[ii], wfid, 0.0, &event);
      ST_DEBUG2(2, "tryToMerge returns %d", retVal);

      if (NO_ERROR != event) { break; }

      /* Check for pending shutdown */
      if (pStation->shutdownEnable != SD_NONE) { break; }
   }

   /*@-compdestroy@*/
   ODBCFreeResult(&result);
   return(event);
   /*@=compdestroy@*/

error_shutdown:

   /*@-compdestroy@*/
   ODBCFreeResult(&result);
   shutdownApplication(buffer, "loadWfdiscRecordsFromDB");
   return(E_LOAD_RECORD);
   /*@=compdestroy@*/
   /*@-mustfreefresh@*/
   /*@=mustfreefresh@*/
}

/**
 * @short Delete wfdisc or quality record in DB
 *
 * @description This function deletes a wfdisc or quality record in
 * @description the database with the given ID
 *
 * @param hdbc         database handle
 * @param rid          record ID
 * @param what         type of record (wfdisc or quality)
 *
 * @return 0 on success or -1 on error
 */

int
deleteRecordInDB(SQLHDBC hdbc, long rid, int what)
{
   char query[QUERY_BUFFER_SIZE];
#ifndef HAVE_LIBIDCODBC
   SQLRETURN ret;
   SQLHSTMT hstmt;
#else
   ODBC_Result result;
#endif

   (void) snprintf(query, sizeof(query), "delete from %.10s where %.10s=%ld",
                   what == WFDISC_PASS ? "wfdisc" : "cdquality",
                   what == WFDISC_PASS ? WFID_STRING : QUALID_STRING,
                   rid);
   DEBUG(2, query);

#ifndef HAVE_LIBIDCODBC
   /* Allocate statement handle */
   ret = ODBCAllocStmt(hdbc, &hstmt);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
    (void)ODBCSetErrorMsg("deleteRecordInDB", SQL_HANDLE_DBC, hdbc);
    goto end_error;
   }

   /* Execute the query */
   ret = SQLExecDirect(hstmt, (unsigned char *) query, SQL_NTS);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
       (void)ODBCSetErrorMsg("deleteRecordInDB", SQL_HANDLE_STMT, hstmt);
       if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
       {
          logWarning("SQLFreeHandle failed", "deleteRecordInDB");
       }
       goto end_error;
   }

   if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
   {
      logWarning("SQLFreeHandle failed", "deleteRecordInDB");
   }

#else /* HAVE_LIBIDCODBC */
   if (ODBCExecDirect(hdbc, query, &result))
   {
      ODBCFreeResult(&result);
      goto end_error;
   }
   ODBCFreeResult(&result);
#endif

   /*@-compdestroy@*/
   return(0);

end_error:

   /*@-mustfreefresh@*/
   shutdownApplication("Can't delete wfdisc record", ODBCErrMsg());
   return(-1);
   /*@=mustfreefresh@*/
   /*@=compdestroy@*/
}

/**
 * @short Update or insert wfdisc record in DB
 *
 * @description This function update or insert a wfdisc record in
 * @description the database using the given mode to distinguish
 * @description between an update or an insert.
 *
 * @param hdbc         database handle
 * @param mode         INSERT_WFDISC or UPDATE_WFDISC
 * @param site         site name
 * @param channel      channel name
 * @param startTime    start time wfdisc record
 * @param wfid         wfdisc ID
 * @param channelID    channel ID
 * @param jdate        julian date of data in record
 * @param endTime      end time wfdisc record
 * @param samplesInSegment   number of samples in segment
 * @param calib        calibration value
 * @param calper       calibration periode
 * @param insType      instrument type
 * @param dataType     s3 or s4
 * @param clipped      clipped flag
 * @param dir          directory of corresponding waveform file
 * @param dfile        file name of corresponding waveform file
 * @param offset       offset of first count of record in the waveform file
 *
 * @return 0 on success or -1 on error
 */

int
updateWfdiscInDB(SQLHDBC hdbc, int mode, const char *site, const char *channel,
                 double startTime, long wfid, long channelID, long jDate,
                 double endTime, long samplesInReqSegment, double sampleRate,
                 float calib,  float calper, const char *insType,
                 /* segtype, */ const char *dataType, const char *clipped,
                 const char *dir, const char *dfile, long offset)
{
   char query[QUERY_BUFFER_SIZE];
#ifndef HAVE_LIBIDCODBC
   SQLRETURN ret;
   SQLHSTMT hstmt;
#else
   ODBC_Result result;
#endif

   if (INSERT_WFDISC == mode)
   {
      (void) snprintf(query, sizeof(query), "insert into wfdisc values \
('%.6s', '%.8s', %.5f, %ld, %ld, %ld, %.5f, %ld, %.7f, %.7g, \
 %.6f, '%.6s', '-', '%.2s', '%.1s', '%.64s', '%.32s', %ld, -1, sysdate)",
                      site, channel, startTime, wfid, channelID, jDate,
                      endTime, samplesInReqSegment, sampleRate, calib, calper,
                      insType, dataType, clipped, dir, dfile, offset);
   }
   else
   {
      (void) snprintf(query, sizeof(query), "update wfdisc set \
 time=%.5f, jdate=%ld, endtime=%.5f, nsamp=%ld, samprate=%.7f, \
 foff=%ld, lddate=sysdate where wfid = %ld", startTime, jDate,
                     endTime, samplesInReqSegment, sampleRate, offset, wfid);
   }

   DEBUG(2, query);

#ifndef HAVE_LIBIDCODBC
   /* Allocate statement handle */
   ret = ODBCAllocStmt(hdbc, &hstmt);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
      (void)ODBCSetErrorMsg("updateWfdiscInDB", SQL_HANDLE_DBC, hdbc);
      goto end_error;
   }

   /* Execute the query */
   ret = SQLExecDirect(hstmt, (unsigned char *) query, SQL_NTS);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
       (void)ODBCSetErrorMsg("updateWfdiscInDB", SQL_HANDLE_STMT, hstmt);
       if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
       {
          logWarning("SQLFreeHandle failed", "updateWfdiscInDB");
       }
       goto end_error;
   }

   if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
   {
      logWarning("SQLFreeHandle failed", "updateWfdiscInDB");
   }

#else /* HAVE_LIBIDCODBC */
   if (ODBCExecDirect(hdbc, query, &result))
   {
      ODBCFreeResult(&result);
      goto end_error;
   }
   ODBCFreeResult(&result);
#endif /* HAVE_LIBIDCODBC */

   /*@-compdestroy@*/
   return(0);

end_error:

   /*@-mustfreefresh@*/
   (void) snprintf(query, sizeof(query),
                   "Can't update/insert wfdisc record WFID=%ld", wfid);
   shutdownApplication(query, ODBCErrMsg());
   return(-1);
   /*@=mustfreefresh@*/
   /*@=compdestroy@*/
}

/**
 * @short Insert clf record in DB
 *
 * @description This function inserts a clf record in
 * @description the database. This function is obsolete
 * @description and should be dropped if squal is re-designed
 *
 * @param hdbc           database handle
 * @param pCurrentFrame  pointer to the current IRL element
 * @param clfType        type of frame
 * @param duration       duration of frame
 * @param revision       stream revision
 * @param channelMask    channel mask (with index from channelname)
 *
 * @return 0 on success or -1 on error
 */

int
insertClfInDB(SQLHDBC hdbc, intRefListParams_t *pCurrentFrame,
              const char clfType, const uint32_t duration,
              const int revision, const char *channelMask)
{
   char query[QUERY_BUFFER_SIZE];
#ifndef HAVE_LIBIDCODBC
   SQLRETURN ret;
   SQLHSTMT hstmt;
#else
   ODBC_Result result;
   int ret;
#endif

   (void) snprintf(query, sizeof(query), "insert into clf values \
(%ld, '%.16s', %.5f, %.5f, %.5f, %ld, %ld, %ld, '>%c', \
 '-', '0x%x', '%.64s', %d, sysdate)",
                   pCurrentFrame->pFileObj->fid,
                   pCurrentFrame->aFrameset,
                   pCurrentFrame->clfTime,
                   pCurrentFrame->nominalTime,
                   pCurrentFrame->clfTimeSpan,
                   (long int)pCurrentFrame->offsetBinFile,
                   (long int)pCurrentFrame->frameSize,
                   (long int)duration,
                   clfType,
                   /* verifyStatus, */
                   (unsigned int)pCurrentFrame->fpc,
                   channelMask,
                   revision);
   DEBUG(2, query);

#ifndef HAVE_LIBIDCODBC
   /* Allocate statement handle */
   ret = ODBCAllocStmt(hdbc, &hstmt);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
      (void)ODBCSetErrorMsg("insertClfInDB", SQL_HANDLE_DBC, hdbc);
      goto end_error;
   }

   /* Execute the query */
   ret = SQLExecDirect(hstmt, (unsigned char *) query, SQL_NTS);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
       (void)ODBCSetErrorMsg("insertClfInDB", SQL_HANDLE_STMT, hstmt);
       /* In most cases the clf record does exist. Log only a
          warning in this case. If you want to be on the save side,
          either the return code has to be checked or a select must
          be done before to check if the record exists */
       (void) snprintf(query, sizeof(query),
                   "Warning: can't insert clf record FRID=%ld, offset=%ld %d",
                    pCurrentFrame->pFileObj->fid,
                    (long int)pCurrentFrame->offsetBinFile, ret);
       logWarning(query, ODBCErrMsg());
   }

   if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
   {
      logWarning("SQLFreeHandle failed", "insertClfInDB");
   }

#else /* HAVE_LIBIDCODBC */
   ret = ODBCExecDirect(hdbc, query, &result);
   if (ret)
   {
      /* Not so much control as with libgodbc, so we ignore the error in any case */
      (void) snprintf(query, sizeof(query),
                     "Warning: can't insert clf record FRID=%ld, offset=%ld %d",
                       pCurrentFrame->pFileObj->fid,
                       (long int)pCurrentFrame->offsetBinFile, ret);
      logWarning(query, ODBCErrMsg());
   }
   ODBCFreeResult(&result);
#endif /* HAVE_LIBIDCODBC */

   /*@-compdestroy@*/
   return(0);

#ifndef HAVE_LIBIDCODBC
end_error:

   /*@-mustfreefresh@*/
   (void) snprintf(query, sizeof(query),
                   "Can't insert clf record FRID=%ld, offset=%ld",
                    pCurrentFrame->pFileObj->fid,
                    (long int)pCurrentFrame->offsetBinFile);
   shutdownApplication(query, ODBCErrMsg());
   return(-1);
   /*@=mustfreefresh@*/
   /*@=compdestroy@*/
#endif /* !HAVE_LIBIDCODBC */
}

/**
 * @short Update or insert quality record in DB
 *
 * @description This function update or insert a quality record in
 * @description the database using the given mode to distinguish
 * @description between an update or an insert.
 *
 * @param hdbc         database handle
 * @param mode         INSERT_WFDISC or UPDATE_WFDISC
 * @param qr           pointer to quality record
 *
 * @return 0 on success or -1 on error
 */

int
updateQualityInDB(SQLHDBC hdbc, int mode, cdQuality_t *qr)
{
   char query[QUERY_BUFFER_SIZE];
#ifndef HAVE_LIBIDCODBC
   SQLRETURN ret;
   SQLHSTMT hstmt;
#else
   ODBC_Result result;
#endif

   if (INSERT_WFDISC == mode)
   {
      (void) snprintf(query, sizeof(query), "insert into cdquality values \
('%.6s', '%.8s', %.5f, %ld, %ld, %.5f, %lu, %lu, %hu, %u, %u, %hu, \
 %.5f, %.5f, %.5f, %.5f, %.5f, %.5f, %lu, %ld, -1, sysdate)",
          qr->q_site, qr->q_channel, qr->q_startTime, (long)qr->q_qualID,
          (long)qr->q_chanID, qr->q_endTime, (unsigned long)qr->q_nSamp,
          (unsigned long)qr->q_firstSlot, (unsigned short)qr->q_qualityFlags,
          (unsigned int)qr->q_channelStatus,
          (unsigned int)qr->q_channelStatusMask,
          (unsigned short)qr->q_tThresh,  qr->q_minTime, qr->q_avgTime,
          qr->q_maxTime, qr->q_meanCounts,
          qr->q_rmsDiff, qr->q_M2, (unsigned long)qr->q_fOff,
          (long)qr->q_wffId);
   }
   else
   {
      (void) snprintf(query, sizeof(query), "update cdquality set \
 time=%.5f, endtime=%.5f, nsamp=%lu, firstslot=%u, \
 mintime=%.5f, maxtime=%.5f, avgtime=%.5f, meancounts=%.5f, \
 rmsdiff=%.5f, m2=%.5f, foff=%lu, lddate=sysdate where qualid = %ld",
                qr->q_startTime, qr->q_endTime,
                (unsigned long)qr->q_nSamp, (unsigned int)qr->q_firstSlot,
                qr->q_minTime, qr->q_maxTime, qr->q_avgTime,
                qr->q_meanCounts, qr->q_rmsDiff, qr->q_M2,
                (unsigned long)qr->q_fOff, (long)qr->q_qualID);
   }

   DEBUG(2, query);

#ifndef HAVE_LIBIDCODBC
   /* Allocate statement handle */
   ret = ODBCAllocStmt(hdbc, &hstmt);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
      (void)ODBCSetErrorMsg("updateQualityInDB", SQL_HANDLE_DBC, hdbc);
      goto end_error;
   }

   /* Execute the query */
   ret = SQLExecDirect(hstmt, (unsigned char *) query, SQL_NTS);
   if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
   {
       (void)ODBCSetErrorMsg("updateQualityInDB", SQL_HANDLE_STMT, hstmt);
       if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
       {
          logWarning("SQLFreeHandle failed", "updateQualityInDB");
       }
       goto end_error;
   }

   if (SQLFreeHandle(SQL_HANDLE_STMT, hstmt) != SQL_SUCCESS)
   {
      logWarning("SQLFreeHandle failed", "updateQualityInDB");
   }
#else /* HAVE_LIBIDCODBC */
   if (ODBCExecDirect(hdbc, query, &result))
   {
      ODBCFreeResult(&result);
      goto end_error;
   }
   ODBCFreeResult(&result);
#endif /* HAVE_LIBIDCODBC */

   /*@-compdestroy@*/
   return(0);

end_error:

   /*@-mustfreefresh@*/
   (void) snprintf(query, sizeof(query),
       "Can't update/insert cdquality record QUALID=%ld", (long)qr->q_qualID);
   shutdownApplication(query, ODBCErrMsg());
   return(-1);
   /*@=mustfreefresh@*/
   /*@=compdestroy@*/
}

/*
 * @short Connect to database
 *
 * @description This function returns the SQL handle to the
 * @description The Parameter dataSource, username und password
 * @description taken from the global configuration data.
 *
 * @param pSenderParams The pointer to the configuration data
 *
 * @return 0 on success, -1 on error
 */

/*@null@*/static SQLHDBC
sqlConnect (senderParams_t *pSenderParams)
{
   SQLHDBC hdbc;
   int ii;

   if ('\0' == pSenderParams->dbConnectName[0] ||
       '\0' == pSenderParams->dbAccount[0])
   {
      /* Don't check dbConnectPassword. dbAccount can also
         contain password */
      DEBUG(2, "dbConnectName and/or dbAccount not provided");
      return(NULL);
   }

   /*  Any characters after (and including) the @ are ignored because
       the database name is derived from dbConnectName */
   for (ii = 0; pSenderParams->dbAccount[ii] != '\0'; ii++)
      if ('@' == pSenderParams->dbAccount[ii])
      {
         pSenderParams->dbAccount[ii] = '\0';
         break;
      }

#ifdef HAVE_LIBIDCODBC
   hdbc = ODBCConnect2(pSenderParams->dbConnectName,
                       pSenderParams->dbAccount,
                       pSenderParams->dbConnectPassword, 0, 30);
#else /* !HAVE_LIBIDCODBC */
   hdbc = ODBCConnect3(pSenderParams->dbConnectName,
                       pSenderParams->dbAccount,
                       pSenderParams->dbConnectPassword, 0, 30, 1);
#endif /* !HAVE_LIBIDCODBC */

   if (NULL == hdbc)
   {
      /*@-mustfreefresh@*/
      logError("Can't connect to DB", ODBCErrMsg());
      return(NULL);
      /*@=mustfreefresh@*/
   }

   return(hdbc);
}

/*
 * @short Disconnect from database
 *
 * @description This function disconnects from the database
 *
 * @param hdbc    database handle
 *
 * @return nothing
 */

void
sqlDisconnect(SQLHDBC hdbc)
{
   (void)ODBCDisconnect(hdbc);
}

/*
 * @short Get SQL handle
 *
 * @description This function returns an SQL handle for the
 * @description configured database account.
 * @description The parameters dataSource, username und password
 * @description are taken from the global configuration data.
 *
 * @param pSenderParams The pointer to the configuration data
 *
 * @return SQL handle on success, stops application on error
 */

/*@null@*/SQLHDBC
getDBhandle(const char *appName, senderParams_t *pSenderParams)
{
   SQLHDBC hdbc;

   /* Run in file mode if no account is configured */
   if ('\0' == pSenderParams->dbConnectName[0] ||
       '\0' == pSenderParams->dbAccount[0])
   {
      return(NULL);
   }

   /* Connect to DB */
   hdbc = sqlConnect(pSenderParams);
   if (NULL == hdbc)
   {
      stopProgram(appName, "can't connect to DB", ODBCErrMsg());
      exit(EXIT_FAILURE);
   }
   return (hdbc);
}

/*
 * @short Set ID increment
 *
 * @description This function raises the ID count which are
 * @description taken from the lastid table when a new ID is
 * @description requested and no cached ID is available
 *
 * @return nothing on success, stops application on error
 */

void
raiseIDcache( const char *appName)
{
   /* Set increment count for wfid */
   if(ODBCSetRequestIdIncrement(WFID_STRING, 20))
   {
      stopProgram(appName, "can't set wfid increment count", ODBCErrMsg());
      exit(EXIT_FAILURE);
   }

   /* Set increment count for qualid */
   if(ODBCSetRequestIdIncrement(QUALID_STRING, 20))
   {
      stopProgram(appName, "can't set wfid increment count", ODBCErrMsg());
      exit(EXIT_FAILURE);
   }

   /* Set increment count for frid to avoid timeouts in testbed */
   if(ODBCSetRequestIdIncrement(FRID_STRING, 20))
   {
      stopProgram(appName, "can't set frid increment count", ODBCErrMsg());
      exit(EXIT_FAILURE);
   }
}


#else  /* HAVE_LIBODBC */

/**
 * @short        Get ODBC provider
 *
 * @description  This function returns the name and the version
 * @description  of the ODBC provider library ('godbc', 'libidcodbc' or
 * @description  'none' if compiled with ODBC support)
 *
 * @return name and the version of the ODBC provider library
 */

/*@observer@*/ const char *getODBCProvider() { return("none"); }
#endif /* HAVE_LIBODBC */
