
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: init_cd2w.c,v $
 *
 * Revision 1.17  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.16  2013-04-06 gerald
 * Set default synchronization time for
 * restart/wfdisc/cdquality information.
 * Fix compiler warning
 *
 * Revision 1.15  2011-09-27 08:36:37  ctbto
 * Fix compilation without database access when
 * sqltypes.h is not installed
 *
 * Revision 1.14  2011-06-09 14:24:36  ctbto
 * Add configuration parameter 'fpTypeIDs' for frame product IDs
 *
 * Revision 1.13  2011-01-18 09:36:44  gerald
 * Set default value for minDetSampRate to 2
 *
 * Revision 1.12  2011-01-17 14:46:30  gerald
 * Add new cd2wng configuration parameter minDetSampRate. It defines
 * the minimum sample rate for which a constant data and sensor noise
 * check is done. Used to avoid 'unreal' constant data and sensor
 * noise detections
 *
 * Revision 1.11  2010-11-30 18:30:13  gerald
 * Add channelStatusMask to cd2wng configuration parameters
 *
 * Revision 1.10  2010-05-11 15:24:28  gerald
 * Add function for getting cd2wng extension pointer
 *
 * Revision 1.9  2010-01-28 14:52:43  gerald
 * Add new sample rate tolerance parameter
 *
 * Revision 1.8  2009-06-09 15:52:28  gerald
 * Add maxWfdiscMemoryDuration configuration parameter
 *
 * Revision 1.7  2009-04-29 16:07:52  gerald
 * Add new configuration parameter 'enableClfRecords'
 *
 * Revision 1.6  2009-02-02 16:49:33  gerald
 * Drop enableWaveformFile config parameter. It makes no sense
 * to create wfdiscs without waveform files (re-start!)
 *
 * Revision 1.5  2008-12-10 11:34:52  gerald
 * Introduce new config parameters cd2wProductionLineFile, enableWaveformFile,
 * and enableQualityCheck
 *
 * Revision 1.4  2008-09-16 18:09:48  gerald
 * Read requested waveform type from configuration file and
 * change default to s3
 *
 * Revision 1.3  2008/07/28 09:26:36  gerald
 * Move functions for setting and retrieving sensor information
 * to libCD
 *
 * Revision 1.2  2008-05-08 14:38:54  gerald
 * Add function for loading  sensor descriptions from file
 * and function which searches the sensor description
 *
 * Revision 1.1  2008/03/20 17:01:38  gerald
 * Add functions for initializing cd2w
 *
 *
 */

#include <config.h>

#ifdef S_SPLINT_S
#include "../../splint_hints.h"
#endif /* S_SPLINT_S */
#include <netinet/in.h>  /* Need this to parse code with splint */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "inittables.h"
#include "cd2w.h"
#include "init_cd2w.h"
#include "merge.h"
#include "libpar.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

/*@-fullinitblock@*/
static senderParams_t cd2wParams =
{
   /* path and name to dataconsumer file ...*/ "",
   /* path and name to production line file.*/ "",
   /* Time to sleep Scheduler thread .......*/ 10,
   /* Look back time Scheduler .............*/ 168,
   /* Time to sleep Worker threads .........*/ 10,
   /* Worker thread retry time ........ ....*/ 60,
   /* internal ref. List store interval ....*/ 6,
   /* wfdisc sync time interval ............*/ 0,
   /* internal ref. List store path ........*/ "",
   /* path and name to index file ..........*/ "",
   /* database connect name ................*/ "",
   /* database password ....................*/ "",
   /* database account .....................*/ "",
   /* waveform type (s3/s4).................*/ "s3",
   /* sensor file ..........................*/ "",
   /* frame product type IDs................*/ "",
   /* look back time worker thread .........*/ 7,
   /* sample rate tolerance in percent .....*/ 1,
   /* mask for channel status bits .........*/ (STS_CLOCK_DIFF_TOO_LARGE |
                                               STS_CLIPPED | STS_ZEROED_DATA),
   /* minimum value for constant data det...*/ 2,
   /* ----- rest is not read from config file, compiler sets it to zero --- */
};

/*@=fullinitblock@*/

static nbrParams_t nbrCd2wParams[] =
{
  /* ----------- 16 bit unsigned ---------------------- */
  /* sequenz is irrelevant, a parameter should appear only once */
  {"sleepTimeSched", &(cd2wParams.sleepTimeSched), INT16U},
  {"sleepTimeWork", &(cd2wParams.sleepTimeWork), INT16U},
  {"sleepTimeRetry", &(cd2wParams.sleepTimeRetry), INT16U},
  {"refListStoreInterval", &(cd2wParams.refListStoreInterval), INT16U},
  {"sampleRateTolerance", &(cd2wParams.sampleRateTolerance), INT16U},
  {"syncTimeInterval", &(cd2wParams.syncTimeInterval), INT16U},
  /* ----------- 32 bit signed ------------------------ */
  {"maxWfdiscMemoryDuration", &(cd2wParams.lookBackTimeWorker), INT32S},
  {"lookBackTimeSched", &(cd2wParams.lookBackTimeSched), INT32S},
  /* ----------- 32 bit unsigned ---------------------- */
  {"channelStatusMask", &(cd2wParams.channelStatusMask), INT32U},
  {"minDetSampRate", &(cd2wParams.detectionThreshold), INT32U}
};
static bitParams_t bitCd2wParams[] =
{
  {"enableWfdiscFile", "YES", CREATE_WFDISC_RECORDS, 0},
  {"enableWfdiscFile", "NO", 0, CREATE_WFDISC_RECORDS },
  {"enableQualityCheck", "YES", CREATE_QUAL_RECORDS, 0},
  {"enableQualityCheck", "NO", 0, CREATE_QUAL_RECORDS },
  {"enableClfRecords", "YES", CREATE_CLF_RECORDS, 0}, /* DEPRECATED */
  {"enableClfRecords", "NO", 0, CREATE_CLF_RECORDS }, /* DEPRECATED */
  {"processMode", "FIFO", PROCESS_MODE_FIFO, PROCESS_MODE_LIFO},
  {"processMode", "LIFO", PROCESS_MODE_LIFO, PROCESS_MODE_FIFO},
  {"processMode", "AS_ARRIVED", 0, PROCESS_MODE_FIFO|PROCESS_MODE_LIFO}
};

static strParams_t strCd2wParams[] =
{
  /* ----------- optional value ---------------------- */
  {"sensorFile", cd2wParams.sensorFile, VALUE_OPTIONAL},
  {"fpTypeIDs", cd2wParams.fpIDs, VALUE_OPTIONAL},
  /* ----------- require a value ---------------------- */
  {"refListStorePath", cd2wParams.refListStorePath, VALUE_REQUIRED},
  /* Next value is optional for receiver but required for CD2W */
  {"cdrecvIndexFilePath", cd2wParams.indexFilePath, VALUE_REQUIRED},
  {"cd2wProductionLineFile", cd2wParams.productionLineFile, VALUE_REQUIRED},
  {"dbConnectName", cd2wParams.dbConnectName, VALUE_REQUIRED},
  {"dbConnectPassword", cd2wParams.dbConnectPassword, VALUE_REQUIRED},
  {"dbAccount", cd2wParams.dbAccount, VALUE_REQUIRED},
  {"waveformType", cd2wParams.waveformType, VALUE_REQUIRED}
};

/**
 * @short Set senders configuration extension pointer
 *
 * @description This function sets the senders configuration extension
 * @description in the common configuration structure
 *
 * @param pConfigData  configuration structure
 *
 * @return nothing
 */

static void
setCd2wConfigExtension(configParams_t *pConfigData)
{
   /*@-mustfreeonly@*/
   pConfigData->configExtension = &cd2wParams;
   /*@=mustfreeonly@*/
}

/**
 * @short Get cd2wng configuration extension pointer
 *
 * @description This function returns the cd2wng configuration
 * @description extension in the common configuration structure
 *
 * @return cd2wng configuration extension structure
 */

/*@dependent@*/senderParams_t *
getCd2wngConfigExtension()
{
   return (&cd2wParams);
}

/**
 * @short Read sender specific configuration parameters
 *
 * @description This function reads the sender specific parameters from
 * @description the specified configuration file and stores it in the
 * @description configParams extension object. This function checks also some
 * @description values of the sender extensions
 *
 * @param appName      name of the application
 * @param filename     name of the configuration file
 * @param pConfigData  configuration structure
 *
 * @return nothing
 */

void
readCd2wConfigParams(const char *appName, const char *fileName,
                       configParams_t *pConfigData)
{
   const char *tmpArgv[2];
   char configFile[LIBPAR_BUFSIZ];

   /* Build argv vector for setpar() */
   (void)snprintf(configFile, sizeof(configFile), "par=%s", fileName);
   tmpArgv[0] = appName;
   tmpArgv[1] = configFile;

   /* Read configuration file with (standard) libpar function to
      support reading of environment variables */
   if (setpar(2, tmpArgv) != 0)
   {
      stopProgram(appName, "can't read configuration file", fileName);
      return;
   }

   /* Set CD2W specific default values. File related values are shadowed
      to global sysconf value later */
   cd2wParams.senderMode |= CREATE_WFDISC_RECORDS;

   /* Process configuration file. Numbers have to be
      processed first to set configured debug level */
   (void) processNumberParams(sizeof(nbrCd2wParams), nbrCd2wParams);
   (void) processBitParams(sizeof(bitCd2wParams), bitCd2wParams,
                           &(cd2wParams.senderMode));
   (void) processStringParams(sizeof(strCd2wParams), strCd2wParams);

   endpar();

   DEBUG(5, "+++ CD2W specific parameters read");

   /* Add CD2W extension to config structure
      and check some values of the sender extension */
   setCd2wConfigExtension(pConfigData);

   /* CD2W needs a configured reference list store path */
   if ('\0' == cd2wParams.refListStorePath[0])
   {
      stopProgram(appName, "No refListStorePath configured", NULL);
      return;
   }

   /* Check waveform type */
   if ('s' != cd2wParams.waveformType[0] ||
       ('4' != cd2wParams.waveformType[1] && '3' != cd2wParams.waveformType[1])
       || '\0' != cd2wParams.waveformType[2])

   {
      stopProgram(appName, "Invalid waveform type", cd2wParams.waveformType);
      return;
   }

   /* Create reference list saving directory if it does not exist */
   if (-1 == checkDirectory(cd2wParams.refListStorePath))
   {
      if (-1 == makeDirectoryDeep(cd2wParams.refListStorePath,
                                  DIR_PERMS, (uint16_t) 0))
      {
         stopProgram(appName, "can't create directory",
                     cd2wParams.refListStorePath);
         return;
      }
   }
}

/**
 * @short dump sender specific configuration parameter
 *
 * @description This function dumps the sender specific configuration
 * @description parameter to syslog
 *
 * @return None
 */

void
dumpCd2wConfigurationParameter(configParams_t *pConfigData)
{
   (void) dumpBitParams(sizeof(bitCd2wParams), bitCd2wParams,
                        pConfigData->systemConf, 0);
   (void) dumpNumberParams(sizeof(nbrCd2wParams), nbrCd2wParams);
   (void) dumpStringParams(sizeof(strCd2wParams), strCd2wParams);
}
