
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * Revision 1.33  2015-11-27 gerald
 * Follow change in CDlibVersion()
 *
 * Revision 1.32  2015-09-06 gerald
 * Make loop through production lines more robust
 *
 * Revision 1.31  2015-03-07 gerald
 * Assign production line in initSenderStation()
 *
 * Revision 1.30  2013-10-12 gerald
 * Fix timezone problem on FreeBSD
 *
 * Revision 1.29  2013-06-13 gerald
 * Add CD2WNG and libCD version to usage string. Add
 * libCD version when logging start message to syslog.
 * Add closing libCD. Write usage message to stderr
 *
 * Revision 1.28  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.27  2012-07-25 gerald
 * Add support for black list
 *
 * Revision 1.26  2012-05-06  ctbto
 * Check also for late arrived data when end time is provided
 *
 * Revision 1.25  2011-06-09 14:24:36  ctbto
 * Add configuration parameter 'fpTypeIDs' for frame product IDs
 *
 * Revision 1.24  2011-05-16 11:49:34  gerald
 * Regress modification which loads frame product ID
 * from fpdescription table. Fix compiler warning
 *
 * Revision 1.23  2011-04-29 09:55:46  gerald
 * Load frame product type ID from fpdescription table
 *
 * Revision 1.22  2011-03-15 15:49:28  gerald
 * Break too long line
 *
 * Revision 1.21  2011-03-02 15:29:59  gerald
 * Add ID string and log warning if no sensor information
 * is available
 *
 * Revision 1.20  2010-10-18 09:26:33  gerald
 * Add second DB connection to get rid of ODBCGetNextId() timeouts
 * during start up of cd2wng. This change requires also a fix in
 * libgODBC
 *
 * Revision 1.19  2010-05-11 15:23:22  gerald
 * Fix channel names loaded from the wrong time period
 * when in fixed interval mode
 *
 * Revision 1.18  2009-06-23 16:40:49  gerald
 * Use the data provider name instead of the data consumer
 * name for the station ID. This allows to find all channels
 * related to a given stream and to authenticate received data
 *
 * Revision 1.17  2009-06-15 13:24:56  gerald
 * Fix comment
 *
 * Revision 1.16  2009-06-09 15:53:05  gerald
 * Do not merge data older than current time - maxWfdiscMemoryDuration
 *
 * Revision 1.15  2009-06-09 10:56:42  gerald
 * Add monitor extension
 *
 * Revision 1.14  2009-05-19 13:52:56  gerald
 * Change syslog IDs as requested by Henrik
 *
 * Revision 1.13  2009-04-06 15:08:12  gerald
 * Fix incorrect start and end times in fixed interval mode
 *
 * Revision 1.12  2009-02-18 15:40:22  gerald
 * Fix bug which prevents creation of wfdisc files when
 * configured without ODBC
 *
 * Revision 1.11  2009-02-02 16:49:33  gerald
 * Drop enableWaveformFile config parameter. It makes no sense
 * to create wfdiscs without waveform files (re-start!)
 *
 * Revision 1.10  2008-12-18 16:15:35  gerald
 * Add port number to syslog ID. Add configuration parameter for enable
 * quality check and and change the default for 'enableWfdiscFile'
 * to YES
 *
 * Revision 1.9  2008-11-11 18:56:29  gerald
 * Fix wrong syslog ID
 *
 * Revision 1.8  2008-09-16 18:04:12  gerald
 * Ignore suspend signal (CRTL-Z) and log a warning to
 * stdout when CRTL-Z is pressed
 *
 * Revision 1.7  2008/07/28 09:24:49  gerald
 * Add application name to getSensorDescr() function
 *
 * Revision 1.6  2008-06-03 15:19:14  gerald
 * Update main.c to use the new data consumer autocreate feature
 *
 * Revision 1.5  2008/05/08 15:49:09  gerald
 * Load sensor description if available and disconnect
 * from database (otherwise the process does not exit on Linux)
 *
 * Revision 1.4  2008/04/07 12:44:53  gerald
 * Open DB connection when configured and compiled with
 * ODBC support
 *
 * Revision 1.3  2008/03/20 17:04:39  gerald
 * Call cd2w specific initalization function instead of standard
 * cdsender function
 *
 * Revision 1.2  2008/01/14 13:06:09  gerald
 * Fix problem with sigwait when running under gdb
 *
 * Revision 1.1  2007/12/10 13:12:21  gerald
 * Add CD2W_NG directory
 *
 *
 */

#include <config.h>

#ifdef S_SPLINT_S
 #include "../../splint_hints.h"
#endif /* S_SPLINT_S */

#include <netinet/in.h>  /* Need this to parse code with splint */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "events.h"
#include "cd2w.h"
#include "init.h"
#include "inittables.h"
#include "init_sender.h"
#include "init_cd2w.h"
#include "merge.h"
#include "command.h"
#include "monitor.h"
#include "certificate.h"

#ifdef HAVE_DB_ACCESS
#include "db_access.h"
#endif

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

/*@unused@*//*@observer@*/ const char *appVersion (void)
{
  return ("$Id: main.c,v 1.25 2011-06-09 14:24:36 ctbto Exp $");
}

#define JOIN_WORKER_THREADS joinHandlingThreads
#define SEVEN_DAYS   (60*60*24*7)

/**
 * @short Start or restart worker threads
 *
 * @description This function starts or restart a worker thread for
 * @description each known production line. No thread is started
 * @description when the worker thread is in the 'stopped' state.
 * @description Processing threads are restarted periodically
 *
 * @param pConfigData     Configuration data
 *
 * @return Nothing
 */

static void
startProcessingThread(configParams_t *pConfigData)
{
   station_t *pStation;

   /* Start a worker thread for each production line
      when no worker thread exists */
   for (pStation = pConfigData->pHistory->firstStation;
        pStation != NULL;
        pStation = pStation->nextStation)
   {
      if (NULL == pStation->appExtension ||
          NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine)
      {
         /* Station without production line. This one is
            created from a key preload directive, we can ignore
            it */
         DEBUG2(9, "Station without production line found [%.17s]",
                pStation->conID);
         continue;
      }

      if (STATION_STOPPED == pStation->state)
      {
         /* Stopped station. No worker thread is created for
            such a station. A worker thread is created only if the
            state changes (by a start command) */
         DEBUG2(9, "Station stopped [%.17s]", pStation->conID);

         /* Check for clf files older than N days and remove these entries
            Removing entries is only allowed when no thread is started
            for that production line (STOPPED or UNCONNECTED state) */
         (void)removeTooOld(pStation, SEVEN_DAYS);
         continue;
      }

      if (STATION_UNCONNECTED != pStation->state)
      {
         /* Processing thread is already started */
         DEBUG3(9, "Thread exists [%.17s] [%d]", pStation->conID,
                (int) pStation->state);
         continue;
      }

      DEBUG2(9, "Try to start worker thread for [%s]", pStation->conID );

      /* Check if a shutdown request is pending. Don't check
         this before the debug messages. A printf is a possible
         thread switch! */
      if (SD_NONE != pConfigData->shutdownEnable) break;

#ifdef HAVE_AUTHENTICATION
      /* Check if certificates have to be loaded. Don't
         load the certificates in the handling threads,
         the functions might not be thread save */
      if (HAVE_TO_LOAD_CERTIFCATES(pStation))
      {
         /* No keys are loaded, load all keys for station */
         if (0 != loadCertificates(pStation,
                                  (pStation->systemConf & VERBOSE_LOGGING)))
         {
            return;
         }
      }
#endif /* HAVE_AUTHENTICATION */

      /* Start sender thread */
      pStation->state = STATION_WAIT2SEND;
      if (0 != pthread_create(&(pStation->threadID), NULL, merge,
                              pStation))
      {
         shutdownApplication("can't start worker thread", "schedThread");
      }
      pConfigData->pHistory->nbrOfThreadsStarted++;
   }
}

/**
 * @short Schedule thread
 *
 * @description The scheduler thread is responsible for starting a
 * @description worker thread for each production line and for supplying
 * @description each worker thread with an initial clf file list.
 * @description In continuous mode the initial clf file list is updated
 * @description periodically
 *
 * @param arg  void pointer to config parameters
 *
 * @return Nothing
 */

/*@null@*/ static void *
schedThread(void *arg)
{

   station_t *pStation;
   configParams_t *pConfigData;  /* configuration parameter */
   struct _prodLine *pProdLine;  /* Current production line */
   int indexFilesParsed;
   char conID[LEN_CON_ID + 1];
   time_t startDate,             /* start date for parsing index files */
          endDate,               /* end date for parsing index files */
          nowTime;
   senderParams_t *pCd2wParams;
#ifdef HAVE_AUTHENTICATION
   pthread_t comID;        /* ID command thread */
   station_t *pControl;
#endif /* HAVE_AUTHENTICATION */

   DEBUG(5, "+++ Scheduler thread started");
   /* Retrieve data passed to thread */
   pConfigData = (configParams_t *) arg;
   pConfigData->pHistory->nbrOfThreadsStarted++;

   /* set first production line */
   pCd2wParams = (senderParams_t *)pConfigData->configExtension;
   pProdLine = pCd2wParams->pProdLine;

   /* At least one production line has to exist */
   if (NULL == pProdLine)
   {
      stopProgram(APP_NAME, "Nothing to do", "schedThread");
      return(NULL); /*for splint */
   }

#ifdef HAVE_AUTHENTICATION
   /* Create a station object for the controller station */
   pControl = getControlObject(pConfigData);
   if (NULL != pControl)
   {
      DEBUG(2,"Control object created");

      /* starting of command thread */
      if (0 != pthread_create(&comID, NULL, commandThread, pConfigData))
      {
         pControl = NULL;
         stopProgram(APP_NAME, "can't start command thread", NULL);
      }
   }
#endif /* HAVE_AUTHENTICATION */

   /* Create a station object for each line in the production line file
      which does not contain a wildcard as data provider */
   for (; pProdLine != NULL; pProdLine = pProdLine->pNextprodLine)
   {
      if (NULL != pProdLine->pDataConsumer)
      {
         /* Check for wildcard in Data Provider name. No station object is
            created for a Data Provider with a wildcard now. This is done
            later when the index file is parsed */
         if ('*' == pProdLine->pDpName[0] && '\0' == pProdLine->pDpName[1])
            continue;

         /* Check if the station is on the black list or if one or more
            channel is on the black list. Keep in mind that, if a station
            is on the black list, the data can be processed later. No
            IRL file is created or modified, whereas the data of a blacklisted
            channel can't be processed later (in continuous mode).
            Blacklisted sites/channel are not supported yet. Stations with
            blacklisted site/channels must be created with the wild
            card later */
         if (pProdLine->isBlacklisted)
            continue;
         /* Check if channels/sites have been blacklisted here to
            distinguish between hole station and site/channel blacklists
         if (pProdLine->isBlacklisted && !HAS_CHANNEL_BLACKLISTED(pProdLine))
         {
            continue;
         } */

         /* Create new station object for each production line.
            The ID of the station object is DataproviderName_DataconsumerName
            and as the station name (used to find the right certificates)
            the Dataprovider name is used */
         (void) snprintf(conID, sizeof(conID), "%s_%s", pProdLine->pDpName,
                          pProdLine->pDataConsumer->dcName);
         pStation = initSenderStation(conID, pProdLine->pDpName, pProdLine);

         if (NULL == pStation)
         {
            stopProgram(APP_NAME, "can't create station object", conID);
            return(NULL); /* for splint */
         }
      }
   }

   /* Set parsing start and end time */
   nowTime = time(NULL);

   if (0 == pCd2wParams->startDate)
   {
      /* continuous mode */
      startDate = nowTime - (pCd2wParams->lookBackTimeSched * 3600);
      endDate = 0;
      DEBUG(5, "continuous mode or auxiliary mode");
   }
   else
   {
      /* fix interval mode  */
      startDate = pCd2wParams->startDate;

      /* TODO: Check if end date is after nowTime */
      if (0 == pCd2wParams->endDate) endDate = nowTime;
      else endDate = pCd2wParams->endDate;

      /* Start initial window at
         nominal time of first frame - MAX_MERGE_DAYS */
      pCd2wParams->lookBackTimeWorker = 0;
      DEBUG(5, "fix interval mode");
   }

   /* Parse index file(s). After parsing each station object
      should have assigned a list of clf files to process. If end
      date is provided, check until end date + MAX_MERGE_DAYS to catch
      also late arrived data the first time. This is save, because
      the data time is checked again in the handling thread when the
      IRL is build */
   indexFilesParsed = catIdxDir(pConfigData->indexFilePath,
                                pConfigData->portNumber,
                                pCd2wParams->pProdLine,
                                pCd2wParams->nbrOfprodLines,
                                &startDate, endDate  == 0 ? endDate :
                                            endDate + MAX_WINDOW);

   if (indexFilesParsed < 0)
   {
      DEBUG2(5,"Search for clf files returns [%d]", indexFilesParsed);
      stopProgram(APP_NAME, "unable to parse Index files", "schedThread");
      return(NULL); /*for splint */
   }

   logWrite(LOG_NOTICE, "Scheduler started");

   /* Start or restart worker threads */
   startProcessingThread(pConfigData);

   /* Now all worker threads should start to send data. In continuous
      mode the scheduler thread will check in sleepTimeSched time
      intervals if new clf files are available */
   while (0 == endDate || endDate > time(NULL))
   {
      struct timespec sleeptime;
      int ii;

      /* No authentication or no control port configured */
      /*@-usedef@*/
      sleeptime.tv_sec = 1;
      sleeptime.tv_nsec = 0;
      /*@=usedef@*/

      /* sleep for a sleepTimeSched time interval */
      for (ii = (int)pCd2wParams->sleepTimeSched;
           ii > 0 && SD_NONE == pConfigData->shutdownEnable; ii--)
      {
         /*@-usedef@*/
         (void) nanosleep(&sleeptime, NULL);
         /*@=usedef@*/
      }

      /* Check if a shutdown request is pending */
      if (SD_NONE != pConfigData->shutdownEnable) break;
      DEBUG(2, "Scheduler: check index file(s) for new entries");

      /* (Re-)parse index files. After parsing each station object
         will have an updated list of clf files to process.
         Modifications of the clf file list is protected by
         the stations mutex */
      indexFilesParsed = catIdxDir(pConfigData->indexFilePath,
                                   pConfigData->portNumber,
                                   pCd2wParams->pProdLine,
                                   pCd2wParams->nbrOfprodLines,
                                   &startDate, endDate);
      DEBUG2(5,"catIdxDir returns [%d]", indexFilesParsed);
      if (indexFilesParsed < 0)
      {
         shutdownApplication("unable to (re-)parse index files", NULL);
         break;
      }

      /* Restart worker threads */
      startProcessingThread(pConfigData);
   }

   /* Wait for all working threads to complete */
   JOIN_WORKER_THREADS(pConfigData->pHistory->firstStation);
   DEBUG(5, "+++ Scheduler thread terminating");

   /* Check if the main thread has a signal received before. Otherwise
      signal shutdown condition to main thread. In fixed interval mode
      no signal has been sent to the main thread. The SIGNAL should be
      sent only once.*/
   if (0 != pthread_mutex_lock(&(pConfigData->shutdownMutex)))
   {
      logError("can't set shutdown mutex", "schedThread");
   }

   if ((unsigned char)0 == pConfigData->shutdownEnable)
   {
      (void) pthread_kill(pConfigData->mainID, SIGUSR1);
   }

   if (0 != pthread_mutex_unlock(&(pConfigData->shutdownMutex)))
   {
      logError("can't unset shutdown mutex", "schedThread");
   }

#ifdef HAVE_AUTHENTICATION
   if (NULL != pControl)
   {
      int retVal;

      /* Wait for command thread to complete */
      DEBUG(9,"before joining command thread");
      /*@-usedef@*/
      retVal = pthread_join(comID, NULL);
      /*@=usedef@*/
      DEBUG2(9,"command thread joined [%d]", retVal);
   }
#endif /* HAVE_AUTHENTICATION */

   pthread_exit(NULL);
   return(NULL); /*for splint */
}


/*
 * @short Catch signals
 *
 * @description This function catches a signal and kills the main thread
 * @description with SIGKILL
 *
 * @param signal  signal
 *
 * @return None
 */

static void
catch_Signal(int signal)
{
   char comBuffer[LOGSIZE];

   /* Don't use getpid() on LINUX here!! */
   (void) snprintf(comBuffer, LOGSIZE, "catchSignal: TID=%lu, SIG=%d",
                   (unsigned long)pthread_self(), signal);
   DEBUG(2, comBuffer);

   (void) pthread_kill(pthread_self(), SIGKILL);
}


/**
 * @short Start required threads for CDsender
 *
 * @description This function does all the necessary stuff to start the
 * @description scheduler and monitor thread and waiting for user termination
 *
 * @param sigMask signal mask for sigwait
 * @param pConfigData     configuration data
 *
 * @return terminating signal number
 */

static int
startSThreads(sigset_t *sigMask, configParams_t *pConfigData)
{
   pthread_attr_t tAttr;     /* Thread attributes */
   struct sigaction sigInfo;
   pthread_t schedID;      /* ID schedule thread */
   pthread_t monID;        /* ID monitor thread */
#ifdef HAVE_AUTHENTICATION
   pthread_t certID;       /* ID certificate thread */
#endif /* HAVE_AUTHENTICATION */
   int sigVal,
       retVal;

   memset (&sigInfo, 0, sizeof (struct sigaction)); /* for splint */
   /* initialize attribute */
   if (0 !=  pthread_attr_init(&tAttr))
   {
      stopProgram(APP_NAME,"can't initialize attribute", NULL);
   }
   /* only PTHREAD_SCOPE_SYSTEM works on Linux! */
   if (0 !=  pthread_attr_setscope(&tAttr, PTHREAD_SCOPE_SYSTEM))
   {
      stopProgram(APP_NAME, "can't set scope", NULL);
   }

   /* start monitor thread if configured */
   if (0 != pConfigData->summaryRefreshTime &&
       0 != pthread_create(&monID, &tAttr, monitorThread, pConfigData))
   {
      stopProgram(APP_NAME, "can't start monitor thread", NULL);
   }

#ifdef HAVE_AUTHENTICATION
   /* start certificate thread if configured */
   if (0 != pConfigData->CRLcheckTime &&
       0 != pthread_create(&certID, &tAttr, certificateThread, pConfigData))
   {
      stopProgram(APP_NAME, "can't start certificate thread", NULL);
   }
#endif /* HAVE_AUTHENTICATION */

   /* starting of scheduler thread */
   if (0 != pthread_create(&schedID, &tAttr, schedThread, pConfigData))
   {
      stopProgram(APP_NAME, "can't start scheduler thread", NULL);
   }

   /* Block all signals in main thread */
   (void) sigfillset(sigMask);
   (void) pthread_sigmask(SIG_BLOCK, sigMask, NULL);

#if linux && HAVE_LIBPTHREAD
   /* (void) sigdelset(sigMask, SIGSEGV); */
#endif

   /* Wait until shutdown is signaled */
   sigVal = wait4shutdown(sigMask);

   /* Unblock all signals. This has no affect on LINUX if waiting
      on a pthread_join() call. Install signal handler */
   sigInfo.sa_handler = catch_Signal;
   sigInfo.sa_flags = 0;
   (void) sigaction(SIGINT, &sigInfo, NULL);
   (void) pthread_sigmask(SIG_UNBLOCK, sigMask, NULL);

#ifdef HAVE_AUTHENTICATION
   /* Stop certificate Thread */
   if (0 != pConfigData->CRLcheckTime)
   {
      /* It's save to cancel the certificate thread */
      /*@-usedef@*/
      if (pthread_cancel(certID) != 0)
      /*@=usedef@*/
      {
         logError("can't cancel certificate thread", "startSThreads");
      }
      /* Join the certificate thread
         Under Linux it could happen that joining the certificate thread
         is not possible when a segmentation fault has occurred. In this
         case the CD2W will wait at this point until a signal
         is received */
      DEBUG(9, "Before pthread_join certificate");
      retVal = pthread_join(certID, NULL);
      DEBUG(9, "After pthread_join certificate");
      pConfigData->pHistory->nbrOfThreadsStopped++;
      DEBUG2(5,"Joining certificates thread returns [%d]", retVal);
      logWrite(LOG_NOTICE, "Certificate thread stopped");
   }
#endif /* HAVE_AUTHENTICATION */

   /* Set shutdown flag */
   if (SIGINT == sigVal) setShutdownFlag(SD_NORMAL);
   else setShutdownFlag(SD_FAST);

   /* Wait for scheduler thread to complete */
   DEBUG(9,"before joining scheduler thread");
   retVal = pthread_join(schedID, NULL);
   DEBUG2(9,"scheduler thread joined [%d]", retVal);

   /* Stop Monitor Thread. Don't use pthread_cancel() to stop monitor thread.
      That could lead to incomplete logging entries */
   if (pConfigData->summaryRefreshTime != 0)
   {
      /*@-usedef@*/
      stopMonitor(monID, pConfigData);
      /*@=usedef@*/
      retVal = pthread_join(monID, NULL);
      DEBUG2(5,"Joining monitor thread returns [%d]", retVal);
      logWrite(LOG_NOTICE,"Monitor thread stopped");
   }

   if (0 !=  pthread_attr_destroy(&tAttr))
   {
      logError("can't destroy attribute", NULL);
   }

   return(sigVal);
}

/**
 * @short Check argv vector for start and end date
 *
 * @description This function checks the argv vector for start and end date
 * @description or auxiliary station mode switch and puts the sender in fixed
 * @description interval mode if at least an end date is found, in auxiliary
 * @description station mode if the auxiliary switch is found or in
 * @description continuous mode otherwise
 *
 * @param argc         Number or CLI arguments
 * @param argv         Vector with CLI arguments
 * @param mode         operating mode (out)
 * @param startDate    start date and time (out)
 * @param endDate      end date and time (out)
 *
 * @return nothing
 */

static void
checkArgs4Date(int argc, char *argv[], uint32_t *mode,
               time_t *startDate, time_t *endDate)
{
   /*TODO: move to library */

   if (argc > 2)
   {
      /* Check for auxiliary station mode */
      if (argc == 3 && 0 == strncmp(argv[2],"0/0/0 00:00:00",15))
      {
         /* Set auxiliary station operating mode */
         *mode |= OPER_MODE_AUX;
         DEBUG(2, "Auxiliary station mode");
      }
      else
      {
         /* Set start date */
         *startDate = ascDate2int(argv[2], "%4d/%2d/%2d %2d:%2d:%2d");
         if ((time_t)-1 == *startDate)
         {
            stopProgram(APP_NAME, "Invalid value in start date", NULL);
            return;
         }

         DEBUG2(5, "start date is %u", (unsigned int)*startDate);

         /* Set fixed interval operating mode */
         *mode |= OPER_MODE_FIXED;
      }

      if (argc == 4)
      {
         /* Set end date */
         *endDate = ascDate2int(argv[3], "%4d/%2d/%2d %2d:%2d:%2d");
         if ((time_t)-1 == *endDate)
         {
            stopProgram(APP_NAME, "Invalid value in end date", NULL);
            return;
         }
      }
   }
   else
   {
      /* Set continuous operating mode */
      *mode |= OPER_MODE_CONTINUOUS;
   }
}

/**
 * @short CD sender main function
 *
 * @description This function handles the CLI arguments and calls
 * @description the appropriate functions for data processing
 *
 * @param argc  Number or CLI arguments
 * @param argv  Vector with CLI arguments
 *
 * @return EXIT_FAILURE or EXIT_SUCCESS
 */

int
main (int argc, char *argv[])
{
   int	ii;
   char comBuffer[LOGSIZE + 1];
   char syslogID[48];
   char dateString[18];
   configParams_t *pConfigData;
   int sigVal;
   int sensorTableCount = 0;
   sigset_t  sigMask;        /* Signal mask value */
   senderParams_t *pCd2wConfig;
   sensor_t *sensorDesc = NULL;

#ifdef HAVE_DB_ACCESS
   SQLHDBC hdbc;
#endif /* HAVE_DB_ACCESS */

   /* Check if CD2W don't run under super-user ID (KT231003) */
   ii = (int)getuid();
   if (0 == ii)
   {
      /* If this would be allowed, the sender would be able
         to read *ALL* files in the file system with a proper
         constructed configuration file */
      printf("Please start CD2W-NG under a non super user ID\n");
      exit(EXIT_FAILURE);
   }

   /* Check command line argument(s) */
   if (argc < 2 || argc > 4)
   {
      fprintf(stderr, "CD2W-NG %.10s (%.16s/%.16s)\n",
              VERSION, CDlibVersion(), getODBCProvider());
      fprintf(stderr, "usage: %s parameter_file [start date [end date]]\n",
              argv[0]);
      fprintf(stderr,
              "%s: %s config.par '2004/01/23 00:00:00' '2004/01/25 00:00:00'\n",
              "example", argv[0]);
      exit(EXIT_FAILURE);
   }

   /* get configuration data with default values */
   pConfigData = initConfigParams(APP_NAME, CD2W);

   /* Set different default values for cd2wng */
   pConfigData->systemConf |= CREATE_WFDISC_FILE;

   /* Receiver port number is part of the syslog ID, but the configuration
      file has to be read first to get the configured port number. So
      a temporary syslog ID is used while reading the configuration file.
      Buffer is build with snprintf to avoid security holes! */
   (void)snprintf(syslogID, sizeof(syslogID), "ctbt.acquisition.cd2wng");
   logInit(syslogID, LOG_INFO);

   /* Read the configuration file */
   (void)readConfigurationFile(APP_NAME, argv[1]);

   /* Add sender specific object to station structure */
   pConfigData->sizeOfAppExtension = sizeof(cdsend_t);

#ifdef EXPERIMENTAL
   /* Register application specific commands */
   /* pConfigData->appSpecificCtrl = processSenderCommands; */
#endif

   /* Register application specific monitor function */
   pConfigData->appSpecificMon = cd2wMonitor;

   /* Load CD2W specific config variables */
   readCd2wConfigParams(APP_NAME, argv[1], pConfigData);

   /* Config extension are set in readCd2wConfigParams() */
   pCd2wConfig = (senderParams_t *)pConfigData->configExtension;

   /* Default: switch off the creation of files */
   pConfigData->systemConf &= ~(CREATE_INDEX_FILE | CREATE_TEXT_FILE |
                                CREATE_CLF_FILE | CREATE_BIN_FILE);

   /* Load production line file. No data consumer file is
      read before, because CD2W does not need a data consumer and
      readPLfile() is able to create dummy data consumers on the fly */
   readPLfile(pCd2wConfig, APP_NAME, AUTO_CREATE);

   /* Check for fixed interval or auxiliary mode */
   checkArgs4Date(argc, argv, &(pCd2wConfig->senderMode),
                  &(pCd2wConfig->startDate), &(pCd2wConfig->endDate));

   /* read further files if necessary and do some checks */
   processConfiguration(pConfigData, CD2W, APP_NAME);

   /*@-mustfreeonly@*/
   /*@-memtrans@*/
   pConfigData->usrFileExt = "mwf";
   pConfigData->appFileExt = "qlr";
   /*@=memtrans@*/
   /*@=mustfreeonly@*/

   /* For the waveform file the user specific file from libCD is used.
      It makes no sense to write wfdisc records without waveform files */
   if ((pCd2wConfig->senderMode & CREATE_WFDISC_RECORDS) != 0)
      pConfigData->systemConf |= CREATE_USR_FILE | CREATE_WFDISC_FILE;
   else
      pConfigData->systemConf &= ~(CREATE_USR_FILE | CREATE_WFDISC_FILE);

   /* For the quality file the application specific file
      from libcd is used */
   if ((pCd2wConfig->senderMode & CREATE_QUAL_RECORDS) != 0)
       pConfigData->systemConf |= CREATE_APP_FILE;

#ifdef HAVE_DB_ACCESS

   /* Try open the ODBC connection */
   hdbc = getDBhandle(APP_NAME, pConfigData->configExtension);

   /*@-branchstate@*/
   if (NULL != hdbc)
   {
      /* Check if frame product IDs have to be set */
      if ('\0' != pCd2wConfig->fpIDs[0])
      {
         /* Set the frame product ID's */
         if (-1 == setFpIDs(pCd2wConfig->fpIDs))
         {
            stopProgram(APP_NAME, "Can't set frame product ID's", NULL);
         }
      }

      /* Load the sensor description */
      sensorDesc = loadSensorFromDB(APP_NAME, hdbc, &sensorTableCount,
                                    pCd2wConfig->startDate,
                                    pCd2wConfig->endDate);

      /* Save DB handle for further use */
      /*@-mustfreeonly@*/
      pCd2wConfig->sndAPIAppPointer = (void *)hdbc;
      /* Try open the ODBC connection for retrieving IDs */
      hdbc = getDBhandle(APP_NAME, pConfigData->configExtension);
      pCd2wConfig->sndAPIAppPointer2 = (void *)hdbc;
      /*@=mustfreeonly@*/

      /* Raise number of cached IDs */
      raiseIDcache(APP_NAME);

      /* Turn of writing to wfdisc file and quality file */
      pConfigData->systemConf &= ~CREATE_WFDISC_FILE;
      pConfigData->systemConf &= ~CREATE_APP_FILE;
   }
   else
   {
      /* Load sensor description from file */
      sensorDesc = loadSensorFromFile(APP_NAME, pCd2wConfig->sensorFile,
                                      &sensorTableCount);
   }

   /*@=branchstate@*/
#else /* ! HAVE_DB_ACCESS */

   sensorDesc = loadSensorFromFile(APP_NAME, pCd2wConfig->sensorFile,
                                   &sensorTableCount);
#endif /* ! HAVE_DB_ACCESS */

   DEBUG2(2, "%d lines loaded from sensor table", sensorTableCount);

   if (NULL != sensorDesc)
       setSensorDescr(sensorDesc, sensorTableCount, APP_NAME);

   /* Control port number is now available, so the real syslog ID
      can be set */
   closelog();
   (void) snprintf(syslogID, sizeof(syslogID), "ctbt.acquisition.cd2wng.%u",
                  (unsigned int)pConfigData->appControlPort);
   logInit(syslogID, LOG_INFO);

   getDate(dateString, sizeof(dateString));
   (void) snprintf(comBuffer, LOGSIZE,
                   "CD2W-NG %.10s (%.16s/%.16s) started at %s by User %d",
                   VERSION, CDlibVersion(), getODBCProvider(), dateString, ii);
   logWrite(LOG_NOTICE, comBuffer);

   /* Log warning if no sensor description is available */
   if (NULL == sensorDesc)
      logWarning("No sensor description available. Site name mapping disabled",
                 "");

   /* if verboseEnable is set, dump configuration to syslog */
   if (pConfigData->systemConf & VERBOSE_LOGGING)
   {
       dumpConfigurationParameter();
       dumpCd2wConfigurationParameter(pConfigData);
   }

   /* The preferred way to handle signals is to block all signals in
      all threads and let the main thread waits for and processes
      signals. The following signals are used by the receiver:
      SIGHUP reserved for configuration reload
      SIGINT  orderly shutdown (CRTL C)
      SIGTERM reserved for immediate shutdown
      SIGUSR1 used internal by the receiver */
   (void) sigemptyset(&sigMask);

#if linux && HAVE_LIBPTHREAD
   /* unfortunately the preferred signal behavior will not work under the LINUX
      pthread implementation. If all signals are blocked, a synchronous signal
      is not seen in the main thread if the signal occurs in
      an other thread (don't mix occur with sent!!). This is not the
      POSIX behavior! If synchronous signals are not blocked in the main thread,
      a further blocked asynchronous signal must be send to wake up sigwait().
      And than the (before sent) synchronous signal is delivered which cause
      the receiver to be terminated immediately. Very strange behavior!!
      For LINUX only a couple of asynchronous signals are blocked for the worker
      threads. The main thread blocks all signals. This give the receiver a
      chance to handle synchronous signals as excepted. */
   (void) sigaddset(&sigMask, SIGHUP);
   (void) sigaddset(&sigMask, SIGINT);
   (void) sigaddset(&sigMask, SIGQUIT);
   (void) sigaddset(&sigMask, SIGTERM);
   (void) sigaddset(&sigMask, SIGUSR1);
#else /* linux && HAVE_LIBPTHREAD */
   (void) sigfillset(&sigMask);
#endif /* not linux && HAVE_LIBPTHREAD */

   (void) pthread_sigmask(SIG_BLOCK, &sigMask, NULL);
   pConfigData->mainID = pthread_self();

   /* Start required threads */
   sigVal = startSThreads(&sigMask, pConfigData);

   /* CD sender is shutting down */
   getDate(dateString, sizeof(dateString));
   (void) snprintf(comBuffer, LOGSIZE, "CD2W-NG stopped at %s by signal %d",
                   dateString, sigVal );
   logWrite(LOG_NOTICE, comBuffer);
   closeLibCD();
   closelog();

#ifdef HAVE_DB_ACCESS
      if (NULL != pCd2wConfig->sndAPIAppPointer)
         sqlDisconnect(pCd2wConfig->sndAPIAppPointer);
      if (NULL != pCd2wConfig->sndAPIAppPointer2)
         sqlDisconnect(pCd2wConfig->sndAPIAppPointer2);
#endif /* HAVE_DB_ACCESS */

   exit(EXIT_SUCCESS);
}
