
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: merge.h,v $
 * Revision 1.32  2013-07-05 gerald
 * Move 'getODBCProvider' to db_access.c
 *
 * Revision 1.31  2013-06-26 gerald
 * Replace 'sensor noise' by 'no input from sensor'
 * and clean up defines
 *
 * Revision 1.30  2013-06-17 gerald
 * Add function to query ODBC provider library
 *
 * Revision 1.29  2013-03-25 gerald
 * Rename ONLY_WFDISCS to SYNC_WFDISCS
 *
 * Revision 1.28  2011-09-27 08:42:18  ctbto
 * Move quality structure from db_access.h to merge.h
 * to fix compilation without database access when sqltypes.h
 * is not installed.
 *
 * Revision 1.27  2011-04-27 15:29:56  gerald
 * Remember current waveform file size for frameproduct table
 *
 * Revision 1.26  2010-12-13 16:36:55  gerald
 * Add ngoodsamp and change wffid type to long
 *
 * Revision 1.25  2010-11-30 19:07:08  gerald
 * Switch to cdquality table. Split cdquality records based
 * on different channel attributes. Map CD1 to CD1_1 attributes.
 * Start work on waveform file ID
 *
 * Revision 1.24  2010-10-27 11:42:21  gerald
 * Rename structure member segSumData to segMean and
 * segSumSquareData to segM2
 *
 * Revision 1.23  2010-10-18 09:46:49  gerald
 * Return more detailed error code in case of an error
 * and Add second DB connection to get rid of
 * ODBCGetNextId() timeouts
 *
 * Revision 1.22  2010-05-11 15:25:04  gerald
 * Add description to quality flag values and rename QUA_NA to
 * QUA_NO_DATA_CHECK and QUA_AUTH_OFF to QUA_NOT_AUTHENTICATED
 *
 * Revision 1.21  2010-04-16 09:00:08  gerald
 * Add function prototype for createSegment()
 *
 * Revision 1.20  2010-03-04 12:35:41  gerald
 * Move define for ONE_DAY to misc.h
 *
 * Revision 1.19  2009-06-23 16:45:38  gerald
 * Add number of channels per stream to waveform header
 * (if sensorinfo is available)
 *
 * Revision 1.18  2009-06-15 15:30:23  gerald
 * Move obsolete frame count from channel to frame statistic
 * and add frames dropped by filter rule
 *
 * Revision 1.17  2009-06-09 10:56:42  gerald
 * Add monitor extension
 *
 * Revision 1.16  2009-05-18 09:38:25  gerald
 * Add per thread database handle to sender structure
 *
 * Revision 1.15  2009-04-24 14:02:13  gerald
 * Change data type of sumcounts and sumsquarecounts to
 * double to avoid possible integer overflows when computing
 * the square sum within one frame
 * Fix incorrect waveform header when cd2wng is not stopped with
 * crtl-C. Waveform header files are not longer re-written when
 * cd2wng re-starts
 *
 * Revision 1.14  2009-03-23 17:42:30  gerald
 * Fix splint warnings which occurred after upgrading
 * to new splint version
 *
 * Revision 1.13  2009-02-18 15:41:59  gerald
 * Add constant data threshold value
 *
 * Revision 1.12  2009-02-12 17:41:56  gerald
 * Add function prototype for getSegment()
 *
 * Revision 1.11  2008-12-18 16:16:34  gerald
 * Change data types for statisctic data
 *
 * Revision 1.10  2008-12-10 11:54:22  gerald
 * Add creation of quality records when data is merged.
 * Re-work statistic output
 *
 * Revision 1.9  2008-10-30 15:50:34  gerald
 * Change type of offsetSub to uint32_t
 *
 * Revision 1.8  2008-09-16 18:11:00  gerald
 * Store waveform data also in S3
 *
 * Revision 1.7  2008/07/28 12:16:10  gerald
 * Move functions for setting and retrieving sensor information
 * to libCD
 *
 * Revision 1.6  2008-06-03 15:25:50  gerald
 * Function headers added
 *
 * Revision 1.5  2008/05/08 15:55:16  gerald
 * Add sensor info to channel header and sub wfdiscs info
 * to segment structure
 *
 * Revision 1.4  2008/03/20 17:06:24  gerald
 * Fix wrong offset after an overlapped segment, fix endless loop
 * when shifting for one day and the stations does not send all
 * channels in the same frame. Complete restart code for file mode
 *
 * Revision 1.3  2008/01/14 13:07:29  gerald
 * Add defines for one second slots
 *
 * Revision 1.2  2007/12/11 10:29:13  gerald
 * Decrease slot size to 1 second
 *
 * Revision 1.1  2007/12/10 13:12:21  gerald
 * Add CD2W_NG directory
 *
 *
 */

#ifndef _MERGE_H
#define _MERGE_H

#include "cd_common.h"
#include "init.h"

#define LOW_FRAME_TIME        10   /**< The smallest duration of a frame */
#define WFDISC_MAX_HOUR        4
#define WFDISC_DURATION      (60*60*WFDISC_MAX_HOUR)
#define MAX_SUB_RECORDS      ((24/WFDISC_MAX_HOUR)+1)
#define TIMELY_THRESHOLD     300   /**< Definition of timely data */
#define CONST_DATA_THRESHOLD   5   /**< min. number of constant data samples */
#define DIGITIZER_NOISE_THRESHOLD 5   /**< Definition for digitizer noise */


#define GET_SAMPLE_RATE(nbrOfSamples, frameTimeLength) \
            ((float)(nbrOfSamples)/((float)frameTimeLength/1000))

typedef struct _channelWaveSeg   /**< Describes channel waveform segments */
{
   /*@null@*/
   struct _channelWaveSeg *nextSeg;  /**< Pointer to next channel segment */
   /*@null@*//*@dependent@*/
   struct _channelWaveSeg *prevSeg;  /**< Pointer to previous channel segment */
   uint32_t firstSlot;           /**< First slot covered by this segment */
   uint32_t lastSlot;            /**< last slot covered by this segment */
   uint32_t samplesInSeg;        /**< Number of samples in segment */
   uint32_t offsetWaveform;      /**< Offset in the waveform file for
                                      this segment */
   uint32_t attributes;          /**< Channel status bits */
   uint32_t segStatusBitMask;    /**< Mask for considered status bits */
#define   STS_DEAD_SENSOR_CHANNEL     (uint32_t)(1<<0)
#define   STS_ZEROED_DATA             (uint32_t)(1<<1)
#define   STS_CLIPPED                 (uint32_t)(1<<2)
#define   STS_CALIBRATION_UNDERWAY    (uint32_t)(1<<3)
#define   STS_EQUIP_HOUSING_OPEND     (uint32_t)(1<<8)
#define   STS_DIGI_EQUIP_OPEN         (uint32_t)(1<<9)
#define   STS_VAULT_DOOR_OPEN         (uint32_t)(1<<10)
#define   STS_AUTH_SEAL_BROKEN        (uint32_t)(1<<11)
#define   STS_EQUIP_MOVED             (uint32_t)(1<<12)
#define   STS_CLOCK_DIFF_TOO_LARGE    (uint32_t)(1<<16)
#define   STS_GPS_RECEIVER_OFF        (uint32_t)(1<<17)
#define   STS_GPS_RECEIVER_UNLOCKED   (uint32_t)(1<<18)
#define   STS_DIGI_ANALOG_IN_SHORTEND (uint32_t)(1<<19)
#define   STS_DIGI_CAL_LOOP_BACK      (uint32_t)(1<<20)
#define   STS_MAIN_PWR_FAILURE        (uint32_t)(1<<24)
#define   STS_BACKUP_PWR_UNSTABLE     (uint32_t)(1<<25)
   double segStartTime;          /**< Creation time of first sample  */
   double segEndTime;            /**< End time of last sample  */
   float  calib;                 /**< Calibration factor */
   float  calper;                /**< Calibration period */
   uint16_t segState;            /**< Segment state: see below */
#define         SEG_INIT                 ((uint16_t)0)
#define         SEG_NORMAL               ((uint16_t)1)
#define         SEG_TOO_LONG             ((uint16_t)2)
#define         SEG_SCHEDULED4RELEASE    ((uint16_t)4)
#define         SEG_RELEASED             ((uint16_t)8)
#define         SEG_CACHED               ((uint16_t)16)

#define         SEG_DONT_MERGE (SEG_TOO_LONG|SEG_SCHEDULED4RELEASE)
   uint16_t segQual;            /**< Quality state: see below */
#define   QUA_OK            ((uint16_t)0) /**< perfect data */
#define   QUA_CONST_DATA    ((uint16_t)1) /**< constant data detected in CSF */
#define   QUA_NO_INPUT      ((uint16_t)2) /**< no input detected in CSF;
                                               supersedes constant data */
#define   QUA_NO_DATA_CHECK ((uint16_t)4) /**< data not checked for constant
                                               data and sensor noise, e.g.
                                               SOH channel, reserved for
                                               future use */
#define   QUA_NOT_TIMELY    ((uint16_t)8) /**< data not arrived in time */
#define   QUA_AUTH_FAILED   ((uint16_t)16) /**< data authentication failed */
#define   QUA_NOT_AUTHENTICATED ((uint16_t)32) /**< data not authenticated.
                                            Authentication switched off or
                                            compiled without SSL */
#define   QUA_AUTH_NO_CERT      ((uint16_t)64) /**< no certificate found
                                                    for authentication check */
#define   QUA_AUTH_NOT_SIGNED  ((uint16_t)128) /**< CSF not signed */
#define   QUA_FR_AUTH_FAILED   ((uint16_t)256) /**< frame authentication
                                                    failed */
#define   QUA_FR_NOT_AUTHENTICATED ((uint16_t)512) /**< frame not authenticated.
                                            Authentication switched off or
                                            compiled without SSL */
#define   QUA_FR_AUTH_NO_CERT ((uint16_t)1024) /**< no certificate found for
                                                   frame authentication check */
#define   QUA_FR_AUTH_NOT_SIGNED ((uint16_t)2048) /**< frame not signed */
#define   QUA_FR_AUTH_NA     ((uint16_t)4096) /**< frame authentication not
                                                   applicable  for CD-1.0 */
   double segMinTimeliness;   /**< minimum observed timeliness (seconds) */
   double segMaxTimeliness;   /**< maximum observed timeliness (seconds) */
   double segAvgTimeliness;   /**< average observed timeliness (seconds) */
   double segMean;            /**< mean value of counts in segment */
   double segM2;              /**< square sum of differences in segment */

#define         NEW_SEG            1
#define         DUPLICATE_SEG      2
#define         MERGE_LEFT         4
#define         MERGE_RIGHT        8
#define         OVERLAP_LEFT      16
#define         OVERLAP_RIGHT     32
#define         OVERLAP_COMPLETE  64
   struct
   {
      long rid[MAX_SUB_RECORDS];              /**< Record ID */
      uint32_t offsetSub[MAX_SUB_RECORDS];         /**< offset waveform */
      uint32_t samplesInSub[MAX_SUB_RECORDS];  /**< samples in waveform */
   } subrecord;
}
channelWaveSeg_t;

typedef struct _cdQuality   /**< Describes a quality record */
{
   /*@dependent@*/
   const char *q_site;     /**< internal site name; 6.6s */
   /*@dependent@*/
   const char *q_channel;  /**< internal channel name; 8.8s */
   double q_startTime;     /**< Creation time of first sample; 17.5f  */
   int32_t q_qualID;       /**< Quality ID from DB; 10ld */
   int32_t q_chanID;       /**< Channel ID from DB; 8ld */
   double q_endTime;       /**< End time of last sample; 17.5f  */
   uint32_t q_nSamp;       /**< Number of samples in segment; 8ld */
   uint32_t q_firstSlot;   /**< First slot covered by this segment; 8u */
   uint16_t q_qualityFlags;   /**< Quality state: 6hu, see QUA_* defines
                                   in channelWaveSeg_t above */
   uint32_t q_channelStatus;  /**< Channel status bits with mask applied; 8u,
                                   see STS_* defines in channelWaveSeg_t
                                   above */
   uint32_t q_channelStatusMask; /**< Channel status mask; 8u */
   uint16_t q_tThresh;            /**< Timely threshold; 8u */
   double q_minTime;   /**< minimum observed timeliness (seconds); 13.5f */
   double q_avgTime;   /**< average observed timeliness (seconds); 13.5f */
   double q_maxTime;   /**< maximum observed timeliness (seconds); 13.5f */
   double q_meanCounts; /**< mean value of counts in segment; 13.5f */
   double q_rmsDiff;   /**< square sum of differences in segment; 13.5f */
   double q_M2;        /**< M2 value in segment; ??17.5f */
   uint32_t q_fOff;    /**< Offset in the waveform file for record */
   int32_t q_wffId;    /**< Waveform file ID from DB; 10ld */
}
cdQuality_t;

#define S4              4
#define S3              3

#define MAX_MERGE_DAYS  7
#define MAX_WINDOW  (MAX_MERGE_DAYS * ONE_DAY)

#define GET_CHANNEL_SIZE(sampleRate, slots, dataSize) \
                         ((sampleRate) * (slots) * (dataSize))

#define SYNC_WFDISCS      0
#define ONLY_STATISTICS   1

enum { INSERT_WFDISC,
       UPDATE_WFDISC };

#define WFDISC_PASS   1
#define QUALITY_PASS  2

/**
 *  Quality record structure
 *  sta[7] Station name
 *  chan[9] Channel name
 *  double time: Epoch time of first sample in record.
 *  long quality ID: Record Id in DB, -1 in file mode
 *  long chanid: Channel operation id. Initial value = -1.
 *  double endtime: Epoch time of last sample in record.
            Time + (nsamp-1)/samprate.
 *  unsigned long nsamp: Number of samples.
 *  unsigned int slot: Slot number (needed for restart)
 *  short quality flag: ok, const. data, sensor noise, quality NA, not
                        authenticated and not timely
 *  unsigned int channel status: Channel status as found in the frame with
                                       channel status mask applied
 *  unsigned int channel status mask: mask for the channel status
 *  unsigned short timeliness threshold
 *  double min timeliness: minimum observed timeliness (seconds)
 *  double max timeliness: maximum observed timeliness (seconds)
 *  double average timeliness: averaged observed timeliness (seconds)
 *  double mean: mean value of waveform data
 *  double rmsdiff: root mean square value of the differences to the mean
                         for this record
 *  double M2: M2 value; required for rmsdiff
 *  unsigned long foff: Offset in the waveform
 *  long wffid: waveform file ID in DB, -1 in file mode
 *  unsigned long ngoodsamp: reserved for future use
 *  lddate[18] Load date. Initial value = "-".
 */

#define QUAL_11 "%-6.6s %-8.8s %17.5f %8ld %8ld %17.5f %8lu %8u %6hu %8u %8u %8hu %13.5f %13.5f %13.5f %13.5f %13.5f %17.0f %32s %8lu %-17.17s\n"
#define QUAL_11_IN "%6s %8s %lf %*d %ld %lf %lu %u %hu %u %u %*d %lf %lf %lf %lf %*f %lf %*s %lu %*s\n"
#define QUAL_12 "%-6.6s %-8.8s %17.5f %8ld %8ld %17.5f %8lu %8u %6hu %8u %8u %8hu %13.5f %13.5f %13.5f %13.5f %13.5f %17.0f %8lu %8ld %8ld %-17.17s\n"
#define QUAL_12_IN "%6s %8s %lf %*d %ld %lf %lu %u %hu %u %u %*d %lf %lf %lf %lf %*f %lf %lu %*d %*d %*s\n"

typedef struct _segHead  /**< Describes a segment chain */
{
   /*@null@*/
   channelWaveSeg_t *firstSeg;   /**< First segment */
   /*@null@*//*@dependent@*/
   channelWaveSeg_t *lastSeg;   /**< Last segment */
   uint32_t nbrOfSegments;      /**< Number of segments */
} segmentHead_t;

typedef struct _dayHead  /**< Describes a channel for one day */
{
   segmentHead_t wfdiscSeg; /**< Wfdisc segments */
   segmentHead_t qualSeg;   /**< Quality segments */
   uint32_t gaps;           /**< Number of gaps between segments */
   uint32_t overlaps;       /**< Number of overlapping segments */
   uint32_t duplicates;     /**< Number of duplicate segments */
   uint32_t tooLong;         /**< Number of segments with too many samples */
   uint32_t tooShort;        /**< Number of segments with too few samples */
   uint32_t filledSlots;     /**< Number of filled slots */
   uint32_t channelWavOffset;  /**< Offset of channel in waveform file */
   uint32_t headerIsWritten;   /**< Flag if the channel info in waveform file
                                    has been saved */
} dayHeader_t;

typedef struct _channelHead  /**< Describes a waveform channel header */
{
   uint32_t expSampleRate;        /**< Expected sample rate */
   uint32_t expChannelTimeLng;    /**< Expected channel time length (seconds) */
   uint32_t nbrOfSlots;           /**< Number of segment slots */
   uint32_t channelIndex;         /**< Position in channel mask */
   dayHeader_t day[MAX_MERGE_DAYS];   /**< channel segments for one day */
   /*@null@*//*@dependent@*/
   sensor_t *sensorInfo;          /**< Sensor values like chanID and more */
} channelHead_t;

typedef struct _waveFormHead  /**< Describes a waveform segment header */
{
   channelHead_t *channelHeads[MAX_CHANNELS]; /**< Buffer for channel headers */
   uint32_t startTimeWindow;          /**< Start time for moving window */
   uint32_t lengthWaveFile[MAX_MERGE_DAYS];  /**< Computed length of waveform
                                                  file per day */
   uint32_t curLenWF[MAX_MERGE_DAYS];  /**< Current length of waveform
                                                  file per day */
   uint32_t headerLinesWritten[MAX_MERGE_DAYS];  /**< Number of header
                                     lines written to the waveform file */
   long waveFileID[MAX_MERGE_DAYS];  /**< ID's for waveform file */
   unsigned char *uncompressBuf;            /**< Buffer for uncompress */
   uint32_t uncompressBufLng;      /**< Length of buffer for uncompress */
   CDdataChannel_t tmpChannel;     /**< Channel structure for loading wfdisc
                                        records */
   uint16_t sampleSize;       /**< size of a sample in the waveform file */
   uint16_t channelCnt;  /**< Number of channels per stream if sensor
                               info is available */
   uint32_t obsoleteFrames;        /**< Number of obsolete frames */
   uint32_t framesDropped;          /**< Frames dropped by filter rule */
   uint32_t channelAttrMaskHd;   /**< channel attribute mask */
   /*@null@*/void *localHdbc; /**< pointer for database handle (if any) */
} waveFormHead_t;


#define WAVEFILE_HEAD_LINE  "%-6.6s %-8.8s %10u %4u %1d\n"
#define WAVEFILE_HEAD_LINE_IN  "%6s %8s %u %u %u\n"
#define LEN_WAVEFILE_HEAD_LINE  (7 + 9 + 11 + 5 + 2)

#define CALIB_TOLERANCE   0.05

#define DB_HANDLE 1  /**< Retrieve standard DB handle */
#define ID_HANDLE 2  /**< Retrieve ID DB handle used for generating IDs */

/* ==== Prototypes ========================================================== */

/*@null@*//*@dependent@*/
extern void *getHdbc(const configParams_t *pConfigData, int which);
extern int
loadChannelDescription(station_t *pStation, uint32_t loadDate,
                       waveFormHead_t *pWaveFormHead, int dayInWindow);
extern int
tryToMerge(CDdataChannel_t *channel, double nominalTime,
           uint32_t channelTime, channelHead_t *pChannelHead, uint32_t slot,
           station_t *pStation, waveFormHead_t *pWaveFormHead,
           uint32_t currentDay, int mergeDay, int isRestartWfdisc,
           char clipped, double wfdiscEndTime, long wfid, double clfTime,
           uint16_t *event);
/*@null@*/extern void *merge(void *arg);
/*@null@*//*@dependent@*/ extern channelWaveSeg_t *
getSegment(segmentHead_t *pSegHead, uint32_t startSlot, uint32_t endSlot,
           /*@out@*/int *result);
/*@null@*//*@dependent@*/ extern channelWaveSeg_t *
createSegment(segmentHead_t *pSegHead, /*@null@*/channelWaveSeg_t *leftSeg);
extern int cd2wMonitor(station_t *pStation);
#endif /* _MERGE_H */

