
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: control.c,v $
 *
 * Revision 1.20  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.19  2011-04-18 13:44:09  gerald
 * Fix comment and wrong frame destination
 *
 * Revision 1.18  2011-03-02 15:35:32  gerald
 * Add support for EC signature size
 *
 * Revision 1.17  2007-06-28 07:47:39  gerald
 * Save errno of last system error for logging in text
 * files (PR-0294)
 *
 * Revision 1.16  2006/04/13 15:00:11  gerald
 * Fix typos and rename defines for event types to avoid inadvertently
 * use of E_INTERNAL
 *
 * Revision 1.15  2006/04/05 13:18:57  gerald
 * Debug messages improved
 *
 * Revision 1.14  2005/05/17 15:20:38  gerald
 * Make use of new error/warning statistics structure
 *
 * Revision 1.13  2005/05/11 10:08:10  gerald
 * Call updateStatistics() instead of updateStationStatistics() and
 * pass new statistics structure instead of whole station structur
 *
 * Revision 1.12  2005/02/17 15:15:36  gerald
 * Fix bad sequence number when generating command frames
 *
 * Revision 1.11  2004/12/14 13:26:02  gerald
 * Add sequence number to command frames (current time in usec)
 *
 * Revision 1.10  2004/09/22 15:26:39  gerald
 * Write log files if configured explicitly. Use libCD functions
 * to read command response and send command request
 *
 * Revision 1.9  2004/09/10 10:48:41  gerald
 * Write debug messages to stderr instead of stdout,
 * use station objects frame buffer instead of own buffer
 *
 * Revision 1.8  2004/09/06 09:33:53  gerald
 * Check station specific system configuration parameter when checking
 * if authentication is switched on instead of system wide parameter
 *
 * Revision 1.7  2004/08/23 09:51:09  gerald
 * Change exit code of cdcon to 100 when application to control
 * does not respond. Exit code 1 is now only initialization failure
 *
 * Revision 1.6  2004/06/18 12:40:11  gerald
 * Rename WITHOUT_ defines
 *
 * Revision 1.5  2004/01/14 16:40:46  gerald
 * Fix compiler warnings when compiled without authentication
 *
 * Revision 1.4  2004/01/12 15:24:08  gerald
 * checkFrameAuthentication() has got a more generic interface. Check
 * new return codes
 *
 * Revision 1.3  2003/12/05 08:37:37  gerald
 * Re-work logging of CDcontrol. Add control of non local receivers.
 * Check frame authentication of Command Response frame
 *
 * Revision 1.2  2003/11/14 14:04:46  gerald
 * Parse and check command response trailer
 *
 * Revision 1.1  2003/10/30 15:01:07  gerald
 * Read and print command response. Move code from main.c
 * to control.[ch]
 *
 * Revision 1.1  2003/10/17 14:13:42  gerald
 * Start work on control application
 *
 */

#include <config.h>
#ifdef S_SPLINT_S
#include <netinet/in.h>  /* Need this to parse code with splint */
#endif /* S_SPLINT_S */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <sys/ioctl.h>   /* Don't remove this include !!*/
#ifndef FIONREAD
#include <stropts.h>
#define FIONREAD I_NREAD  /* FIONREAD is used under Linux */
#endif /* FIONREAD */
#include "events.h"
#include "control.h"
#include "readwrite.h"
#include "logfiles.h"
#include "cdcon.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif


/**
 * @short Send a command
 *
 * @description This function builds a Command Request frame and
 * @description sends this frame over a given socket.
 *
 * @param pStation station object
 * @param command  command to send
 *
 *
 * @return 0 on success or event code on error
 */

uint16_t
sendCommand(station_t *pStation, const char *command)
{
   const configParams_t *pConfigData = pStation->pConfigData;
   uint32_t offsetTrailer;
   uint16_t event;
   int len;     /* length of frame */
   char /*@dependent@*/ *buffer;
   uint64_t sequenzNbr;

   sequenzNbr = getSequenceNbr();

   /* TODO: check if the command is not too large for the buffer */
   buffer = pStation->frameBuffer;

   /* Set payload and compute length of payload */
   offsetTrailer = setCORpayload(&buffer[SZ_HDR], pConfigData->frameCreator,
                                 "", "", "", command);

   /* Set header */
   setHeader(buffer, COMMAND_REQUEST, offsetTrailer, pConfigData->frameCreator,
             pStation->stationName,
             sequenzNbr); /* UGLY: The controller should manage a sequence
                             number file for all known frame_creator/destination
                             pairs, but that would be hard to maintain. So
                             it uses the current time as sequence number, which
                             is sufficient to make the frame 'unique' */

#ifdef HAVE_AUTHENTICATION
   if (NULL == pConfigData->privateKey)
   {
      len = (int)setTrailer(buffer, offsetTrailer, 0, NULL, 0);
   }
   else
   {
      char signature[SZ_SIGNATURE_MAX];
      size_t sigLen = sizeof(signature);

      /* Frame authentication is active */
      if (signData(pConfigData->privateKey, buffer, (size_t)offsetTrailer,
                   signature, &sigLen))
      {
         /* something unexpected has happen during signature
            verification. */
         logError("Unexpected error while signing data", "sendCommand");

         if (pConfigData->systemConf & STRICT_MODE_ENABLE)
         {
            return(1);
         }
         /* Send unsigned frame */
         len = (int)setTrailer(buffer, offsetTrailer, 0, NULL, 0);
      }
      else
      {
         /* Set trailer with signature */
         len = (int)setTrailer(buffer, offsetTrailer,
                               pConfigData->privateKeyID, signature, sigLen);
      }
   }
#else

   len = (int)setTrailer(buffer, offsetTrailer, 0, NULL, 0);

#endif /* defined  HAVE_AUTHENTICATION */

   /* statistic is updated and data is saved in sendFrame */
   event = sendFrame(pStation, buffer, len, S_COMRQ_SENT, GOOD_COMMAND_REQ, 0);
   if (NO_ERROR != event)
   {
      if (E_SYSTEM == event)
      {
         pStation->savedSystemErr = (uint32_t)errno;
         perror("Error while sending command frame");
      }
      logCDevent(ET_E_CON, pStation, event, NULL, (uint32_t)errno);
      return(event);
   }

   return(NO_ERROR);
}

/**
 * @short Process command response
 *
 * @description This function processes a command response
 * @description frame and prints the command response to stdout
 *
 * @param pStation       station object
 *
 * @return 0 on success or event code on error
 */

uint16_t
receiveResponse(station_t *pStation )
{
   rwFlags_t flags;
   int bytesRead;
   const configParams_t *pConfigData = pStation->pConfigData;
   CDevent_t actEvent;
   uint16_t event = NO_ERROR;
   char /*@dependent@*/ *buffer;
   char clfType;

   buffer = pStation->frameBuffer;

   /* Initialize read/write flag structure */
   memset((void *) &flags, 0, sizeof (flags));
   flags.shutdownFlag = &(pConfigData->shutdownEnable);
   flags.timeout = pConfigData->CRtimeOut;
   flags.resetDataTimeout = pConfigData->CRtimeOut;

   /* Set event values */
   actEvent.conID = pStation->conID;
   actEvent.portInfo = pStation->portInfo;
   actEvent.is_CR = 1;
   actEvent.type = CD1_1;
   actEvent.lastCRFreceived = 0;

   /* Now try to read the command request frame */
   bytesRead = readCDframe(pStation, &actEvent, &flags, &event, &clfType);

   ST_DEBUG3(2,"readCDframe returns [%d] [%c]", (int)event, clfType);

   if (bytesRead > 0)
   {

      /* Bad events are logged in readCDframe, however there
         could be valid frames of other types received */
      if (IS_GOOD_FRAME(clfType) && S_COMRS_RECEIVED != event)
      {
         /* No valid command response */
         event = E_UNEXPECTED_TYPE;
         MAKE_FRAME_BAD(clfType);
         /* This is not really an invalid frame, but we should remember
            this event somewhere */
         pStation->whenUnconnected.invalidFrames++;
         logCDevent2(ET_L_ERROR, E_UNEXPECTED_TYPE, &actEvent, NO_INFO);
      }

      if (S_COMRS_RECEIVED == event)
      {
         /* Command response received */
         if (bytesRead < (SZ_HDR + SZ_COMMRES_PL + SZ_TRAILER_WO_SIG))
         {
            /* Frame is too small for command response */
            event = E_FRAMESIZE;
            MAKE_FRAME_BAD(clfType);
            pStation->whenUnconnected.invalidFrames++;
            logCDevent2(ET_L_ERROR, E_FRAMESIZE, &actEvent, NO_INFO);
         }
         else
         {
            CDcomRes_t payload;

            if (0 == parseComRes(&buffer[SZ_HDR], &payload,
                                 (uint32_t)(bytesRead - SZ_HDR)))
            {
               (void)fwrite(payload.response, 1,
                            (size_t)(payload.responseSize), stdout);
               event = NO_ERROR;
            }
            else
            {
               /* Invalid sizes in command response payload */
               event = E_FRAMESIZE;
               MAKE_FRAME_BAD(clfType);
               pStation->whenUnconnected.invalidFrames++;
               logCDevent2(ET_L_ERROR, E_FRAMESIZE, &actEvent, NO_INFO);
               DEBUG(5, "Can't parse Connection Response frame");
            }
         }
      }

      /* Update statistics and save data */
      updateStatistics(&(pStation->framesReceived), bytesRead, NO, clfType);
      if (saveData(pStation, pStation->frameBuffer, bytesRead, 0, event,
                   NO, clfType))
      {
         /* SaveData will shutdown the application if data
            could not be saved */
         pStation->whenUnconnected.systemFailures++;
         return(E_SAVE_DATA);
      }
   }

   return(event);
}

/**
 * @short Open a connection to the applications command port
 *
 * @description This function does all the necessary stuff to
 * @description open a connection to the applications command port
 *
 * @param pStation        station object
 * @param ipAddress       IP address for connect
 *
 * @return nothing
 */

void
openConnection(station_t *pStation, const char *ipAddress)
{
   int sendSock,
       retVal,
       sockError = 0;
   in_addr_t inAddr;
   rwFlags_t flags;
   char logBuffer[LOGSIZE];
   const configParams_t *pConfigData = pStation->pConfigData;

   DEBUG(2, "+++ Open Connection");

   /* Open a connection to CD application's control port.
      For a CD receiver this is the well known port */
   inAddr = inet_addr(ipAddress);
   if ((in_addr_t) -1 == inAddr)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "Invalid IP address [%.16s]", ipAddress);
      logError(logBuffer, "openConnection");
      exit(EXIT_FAILURE);
   }

   /* Create and configure sender socket */
   sendSock = createSocket();
   if (-1 == sendSock)
   {
      logError("createSocket failed","openConnection");
      exit(EXIT_FAILURE);
   }

   flags.sockFlags = 0;
   flags.shutdownFlag = &(pConfigData->shutdownEnable);
   flags.timeout = pConfigData->CRtimeOut;
   flags.resetDataTimeout = pConfigData->CRtimeOut;
   flags.firstTime = NULL;

   /* Connect to CD application */
   retVal = connectWithTimeout(sendSock, inAddr,
                               htons(pConfigData->portNumber),
                               &flags, &sockError);

   pStation->portInfo->socket = sendSock;
   /* Set local port and IP address in the portinfo structure */
   if (-1 == setPortInfo(pStation->portInfo, sendSock, NULL))
   {
      logError("setPortInfo failed","openConnection");
      exit(EXIT_FAILURE);
   }

   /* Set remote port and IP address in the portinfo structure,
      remote and local addresses are swapped for log messages */
   pStation->portInfo->remotePort = pStation->portInfo->localPort;
   memcpy(pStation->portInfo->remoteIPaddr, pStation->portInfo->localIPaddr,
             sizeof(pStation->portInfo->localIPaddr));
   pStation->portInfo->localPort = pConfigData->portNumber;
   /*@-mayaliasunique@*/
   memcpy(pStation->portInfo->localIPaddr, ipAddress,
             sizeof(pStation->portInfo->localIPaddr));
   /*@=mayaliasunique@*/

   switch (retVal)
   {
   case 1:
      logCDevent(ET_E_CR, pStation, E_CR_TIMEOUT, NULL, NO_INFO);
      (void) close(sendSock);
      exit(E_CR_TIMEOUT);
   case -1:
      logCDevent(ET_E_CR, pStation, E_CR_FAILED, NULL, (uint32_t)errno);
      /* TRICKY: The next two lines are a little bit tricky and won't work
         on all OS I guess */
      errno = sockError;
      perror("Connect request failed");
      (void) close(sendSock);
      exit(E_CR_FAILED);
   case -2:
      logCDevent(ET_S_CR, pStation, S_SHUTDOWN, NULL, NO_INFO);
      (void) close(sendSock);
      exit(EXIT_SUCCESS);
   }

   /* CDcontrol is now connected to the applications control port */
   pStation->state = STATION_CONNECTING;
}
