
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: main.c,v $
 *
 * Revision 1.29  2013-06-17 gerald
 * Add CD Sender and libCD version to usage string.
 * Add closing libCD. Write usage message to stderr
 *
 * Revision 1.28  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.27  2013-04-05 gerald
 * Syslog ID changed
 *
 * Revision 1.26  2011-03-31 09:22:51  gerald
 * Comment updated
 *
 * Revision 1.25  2008-01-14 12:12:18  gerald
 * Fix core dump in CD-controller when it has no rights
 * to write into filePath
 *
 * Revision 1.24  2007/09/13 10:57:54  gerald
 * Update function header
 *
 * Revision 1.23  2006/05/24 13:45:41  gerald
 * Fix typo
 *
 * Revision 1.22  2006/05/11 08:49:18  gerald
 * Changes treatment of key caches.  A key cache does not longer belong
 * to one station. Instead a list of key caches is managed and each key
 * cache is globally accessable. A reference to a key cache is now assigned
 * to a station structure when the station connects for the first time
 *
 * Revision 1.21  2006/04/13 15:00:11  gerald
 * Fix typos and rename defines for event types to avoid inadvertently
 * use of E_INTERNAL
 *
 * Revision 1.20  2006/02/14 13:01:12  gerald
 * readConfigurationFile() does not longer need to know which
 * application has called it
 *
 * Revision 1.19  2005/09/21 09:22:11  gerald
 * Make use of CDtoolsVersion()
 *
 * Revision 1.18  2005/08/11 09:45:05  gerald
 * Fix compiler warnings
 *
 * Revision 1.17  2005/02/17 15:08:10  gerald
 * Fix wrong user id when command is logged
 *
 * Revision 1.16  2004/11/12 14:20:26  gerald
 * Add app ID to initConfigParams(). Remove message to stdout
 *
 * Revision 1.15  2004/10/13 09:35:10  gerald
 * Fix compilation error when compile without authentication
 *
 * Revision 1.14  2004/09/22 15:25:02  gerald
 * Add station name to CD controller CLI. Use CD event as exit code
 * for CD controller. Exit code 1 is configuration problem and not
 * E_IP_UNKNOWN
 *
 * Revision 1.13  2004/09/16 13:26:16  gerald
 * Add port as a command line argument
 *
 * Revision 1.12  2004/09/10 10:49:59  gerald
 * Use station objects frame buffer instead of own buffer
 *
 * Revision 1.11  2004/08/23 09:51:09  gerald
 * Change exit code of cdcon to 100 when application to control
 * does not respond. Exit code 1 is now only initialization failure
 *
 * Revision 1.10  2004/04/27 15:27:43  gerald
 * Fix missing "ctbt.acquisition..." string in syslog messages when
 * an initialization error is logged. The port number in the syslog
 * ID string is added after the configuration file is read
 *
 * Revision 1.9  2004/03/08 15:41:53  gerald
 * Call processConfiguration() which has been moved from
 * cdreceiver code to libCD
 *
 * Revision 1.8  2004/02/13 15:15:36  gerald
 * CRCinit() has been moved to initConfigParames()
 *
 * Revision 1.7  2004/02/03 08:58:13  gerald
 * Add some comments and fix log message
 *
 * Revision 1.6  2004/01/14 16:40:46  gerald
 * Fix compiler warnings when compiled without authentication
 *
 * Revision 1.5  2003/12/05 08:34:59  gerald
 * Re-work logging of CDcontrol. Add control of non local receivers
 *
 * Revision 1.4  2003/11/20 13:12:01  gerald
 * Move functions reading the configuration file and the configuration
 * structure itself from CDreceiver to libCD. Fix splint warnings
 *
 * Revision 1.3  2003/11/14 14:02:35  gerald
 * Use common configuration structure instead of own
 *
 * Revision 1.2  2003/10/30 15:01:07  gerald
 * Read and print command response. Move code from main.c
 * to control.[ch]
 *
 * Revision 1.1  2003/10/17 14:13:42  gerald
 * Start work on control application
 *
 */

#include <config.h>
#ifdef S_SPLINT_S
#include <netinet/in.h>  /* Need this to parse code with splint */
#endif /* S_SPLINT_S */
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>

#include "inittables.h"
#include "events.h"
#include "control.h"
#include "cdcon.h"
#include "init.h"
#include "command.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

extern int errno;

/*
 * @short catch signals
 *
 * @description This function catches a signal and sets the shutdown
 * @description flag in the config structure
 *
 * @param signal  signal
 *
 * @return None
 */

static void
catchSignal(int signal)
{
   char comBuffer[LOGSIZE];

   /* Don't use getpid() on LINUX here!! */
   (void) snprintf(comBuffer, LOGSIZE, "catchSignal: SIG=%d", signal);
   DEBUG(2, comBuffer);

   /* Command all handing threads to terminate */
   setShutdownFlag(SD_FAST);
}

/**
 * @short Print usage message
 *
 * @description This function prints a usage message
 *
 * @param argv  Vector with CLI arguments
 *
 * @return Nothing
 */

static void
usage(char *argv[])
{
   fprintf(stderr, "CD-controller %.10s (libCD %.10s)\n",
           VERSION, CDlibVersion());
   fprintf(stderr, "usage: %s config_file %s command[=command_parameter]\n",
           argv[0], "station_name:[IP]:port");

   printCommands();

   exit(EXIT_FAILURE);
}

/**
 * @short Check command line arguments
 *
 * @description This function checks the command line arguments
 *
 * @param argc  Number of CLI arguments
 * @param argv  Vector with CLI arguments
 *
 * @return Nothing
 */

static int
checkCommandLine(int argc, char *argv[])
{
   int rVal;

   /* Check if 3 arguments are provided */
   if (argc != 4) usage(argv);

   rVal = checkCommand(argv[3]);
   if (-1 == rVal) usage(argv);

   return(rVal);
}

/**
 * @short Create a station object
 *
 * @description This function creates a station object and loads the
 * @description certificates for the controlled station. This is
 * @description necessary to use libCD functions and to authenticate
 * @description the response. Loading the certificates based on the
 * @description frame creator string in the command response frame
 * @description would be a security risk!!!!
 *
 * @param pConfigData     configuration data
 * @param stationName     name of the station to connect
 *
 * @return NULL on failure or a station object on success
 */

static /*@dependent@*//*@null@*/ station_t *
createStation(configParams_t *pConfigData, char *stationName)
{
   station_t *pStation;
   static portMap_t portInfo;
   char tmpBuffer[LEN_CON_ID];

   DEBUG(2, "+++ Create station object");

   /* Create new station object.
      The ID of the station object is frameCreator_stationName
      and as the station name (used to find the right certificates)
      the given station name is used */
   (void) snprintf(tmpBuffer, sizeof(tmpBuffer), "%s_%s",
                   pConfigData->frameCreator, stationName);
   pStation = getStation(tmpBuffer, stationName);

   if (NULL == pStation)
   {
      logError("no memory for controllers station object", "openConnection");
      return(NULL);
   }

   /* Provide a portinfo structure */
   pStation->portInfo = &portInfo;

   /* Set station state and type */
   pStation->type = CD1_1;
   pStation->state = STATION_UNCONNECTED;

#ifdef HAVE_AUTHENTICATION

   if (HAVE_TO_LOAD_CERTIFCATES(pStation))
   {
      int rVal;

      /* Try to load receivers public certificate */
      rVal = loadCertificates(pStation,
                              (pConfigData->systemConf & VERBOSE_LOGGING));

      if (0 != rVal)
      {
         logError("Unexpected error while loading certificates",
                  "createStation");
         return(NULL);
      }
   }

   /* In strict mode at least a NONE key directive for a
      signing object must exists or the application will stop */
   if (AUTH_NOKEYS == pStation->keyCache->authState)
   {
      /* No keys available */
      if (pConfigData->systemConf & STRICT_MODE_ENABLE)
      {
         logCDevent(ET_E_CR, pStation, E_NO_KEYS, NULL, NO_INFO);
         return(NULL);
      }
      /* warnings are not counted as errors in statistics */
      logCDevent(ET_W_CR, pStation, E_NO_KEYS, NULL, NO_INFO);
   }
   /* AUTH_NONE is no error or warning */

#endif /* defined HAVE_AUTHENTICATION */

   return (pStation);
}

/**
 * @short CD control main function
 *
 * @description This function handles the CLI arguments and calls
 * @description the appropriate functions
 *
 * @param argc  Number of CLI arguments
 * @param argv  Vector with CLI arguments
 *
 * @return EXIT_FAILURE or EXIT_SUCCESS
 */

int
main (int argc, char *argv[])
{
   char comBuffer[LOGSIZE];  /* per Thread or on stack!! */
   char syslogID[48];
#define TMP_LNG 32
   char tmpBuffer[TMP_LNG];
   int ii,
       uid,
       retVal;
   struct sigaction sigInfo;
   configParams_t *pConfigData;
   station_t *pStation;
   char *ipAddress = "127.0.0.1";
   char *endptr;
   long port;
   split_t result;

#ifndef HAVE_AUTHENTICATION
   printf("You have to configure CD Tools with authentication to run CD Controller\n");
   exit(EXIT_FAILURE);
#endif

   /* Check if CDcontrol don't run under super-user ID */
   uid = (int)getuid();
   if (0 == uid)
   {
      /* If this would be allowed, the receiver would be able
         to read *ALL* files in the file system with a proper
         constructed configuration file */
      printf("Please start CD-control under a non super user ID\n");
      exit(EXIT_FAILURE);
   }

   /* Check command line argument(s) */
   (void) checkCommandLine(argc, argv);

   /* Check recipient data */
   if (strlen(argv[2]) > 31) usage(argv);

   /* Copy argv[2] into local buffer to avoid problems when argv
      is write protected */
   strncpy(tmpBuffer, argv[2], sizeof(tmpBuffer));
   ii = (int)splitString(tmpBuffer, 31, ':', &result);
   if (ii < 2 || ii > 3) usage(argv);

   /* Overwrite separator with NIL */
   tmpBuffer[result.lenString[0]] = '\0';
   tmpBuffer[result.lenString[0] + result.lenString[1] + 1] = '\0';

   /* Check station name */
   if (result.lenString[0] > LEN_STATION_NAME
       ||
       result.lenString[0] < 3)
   {
      usage(argv);
   }

   /* Check IP address */
   /*@-branchstate@*/
   if (result.lenString[1] > 0)
   {
      if ((in_addr_t)-1 == inet_addr(result.pString[1])) usage(argv);

      ipAddress = (char *)result.pString[1];
   }
   /*@=branchstate@*/

   /* Check port Number */
   port = strtol(result.pString[2], &endptr, 10);
   if (*endptr != '\0') usage(argv);

   /* get configuration data with default values */
   pConfigData = initConfigParams(APP_NAME, CONTROLLER);

   /* Build syslog prefix.
      Buffer is build with snprintf to avoid security holes! */
   (void) snprintf(syslogID, sizeof(syslogID), "ctbt.acquisition.cdcon");
   logInit(syslogID, LOG_INFO);

   /* Read configuration file */
   (void) readConfigurationFile(APP_NAME, argv[1]);

   /* read further files if necessary and do some checks */
   processConfiguration(pConfigData, CONTROLLER, APP_NAME);

   /* if verboseEnable is set, dump configuration to syslog */
   if (pConfigData->systemConf & VERBOSE_LOGGING)
       dumpConfigurationParameter();

#ifdef HAVE_AUTHENTICATION
   /* Check if a private key is available */
   if (NULL == pConfigData->privateKey)
   {
      printf("You have to have a private key configured to use CD Controller\n");
      exit(EXIT_FAILURE);
   }
#endif /* HAVE_AUTHENTICATION */

   /* Overwrite port with command line argument */
   pConfigData->portNumber = (uint16_t) port;

   /* Log command to send */
   (void) snprintf(comBuffer, LOGSIZE,
      "CDcontrol %.10s (libCD %.10s): Send command [%s] by User %d to %s:%d",
                   VERSION, CDlibVersion(), argv[3], uid, ipAddress, (int)port);
   logWrite(LOG_NOTICE, comBuffer);

   /* Install signal handler */
   memset (&sigInfo, 0, sizeof (struct sigaction)); /* for splint */
   sigInfo.sa_handler = catchSignal;
   sigInfo.sa_flags	= 0;
   (void) sigaction(SIGINT, &sigInfo, NULL);

   /* Needed for shutdownApplication() */
   pConfigData->mainID = pthread_self();

   retVal = EXIT_SUCCESS;
   /* Create a station object and load keys for controlled CD application */
   pStation = createStation(pConfigData, (char *)result.pString[0]);

   if (NULL == pStation) exit(EXIT_FAILURE);

   /* Open Connection to the controlled application */
   openConnection(pStation, ipAddress);

   /*@-compdef@*/
   /* Send command to the controlled application */
   retVal = (int)sendCommand(pStation, argv[3]);
   if (0 == retVal)
   {
      /* Command has been sent without problems,
         read response */
      retVal = (int)receiveResponse(pStation);
   }
   /*@=compdef@*/

   (void)close(pStation->portInfo->socket);
   closeLibCD();
   closelog();
   exit(retVal);
}
