/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: gaps.c,v $
 *
 * Revision 1.17  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.16  2012-06-27 gerald
 * Fix function comment
 *
 * Revision 1.15  2012-02-02  gerald
 * Add comment that frame set information is not persistent
 *
 * Revision 1.14  2007-04-04 14:08:12  gerald
 * Fix purify warning
 *
 * Revision 1.13  2006/05/23 09:30:29  gerald
 * Fix some typos
 *
 * Revision 1.12  2006/04/26 08:28:06  gerald
 * Move receiver specific config and station structure members
 * into config extension respectively station extension structure.
 * Move reading of receiver specific config parameters from libCD
 * to receiver specific function. Update comments
 *
 * Revision 1.11  2006/04/18 09:19:39  gerald
 * Rename defines for event types to avoid inadvertently use of
 * E_INTERNAL and fix typos
 *
 * Revision 1.10  2005/01/24 16:46:13  gerald
 * Revert to old AckNack behaviour (delete oldest gap)
 * F&P conform behavior did not work
 *
 * Revision 1.9  2004/10/13 11:06:23  gerald
 * Fix compiler warning when compiled without authentication
 *
 * Revision 1.8  2004/07/07 14:07:40  gerald
 * Change behavior when MAX_GAPS limit is reached. Now the connection
 * is closed instead of removing the lowest gap. And the gap lists are
 * deleted when the connection is closed.
 *
 * Revision 1.7  2003/11/24 10:56:41  gerald
 * Replace global debug macro with station specific debug macro to
 * support station specific debug levels
 *
 * Revision 1.6  2003/03/24 15:21:25  gerald
 * Remove trailing blanks and fix splint warnings
 *
 * Revision 1.5  2003/02/17 15:54:11  ctbto
 * Reduce messages to syslog if verboseEnabel is set to NO
 *
 * Revision 1.4  2003/02/12 09:10:33  ctbto
 * Fix warnings that occur when compiling with  -Wall
 *
 * Revision 1.3  2002/12/09 12:51:21  ctbto
 * Add new log message "new gap created"
 *
 * Revision 1.2  2002/10/31 12:56:30  ctbto
 * Log sequence warning only if a new gap is created
 *
 * Revision 1.1  2002/10/28 15:45:33  ctbto
 * Initial revision
 *
 *
 */

#include <config.h>
#include <string.h>
#include <stdlib.h>
#include "cdrecv.h"
#include "events.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

/**
 * @short get frame set object
 *
 * @description This function returns a frame set object for a given creator
 * @description and destination name. If no one could be found a new
 * @description frame set object is created
 *
 * @param header    pointer to header information
 * @param pStation  pointer to station object
 *
 * @returns a frame set object on success or NULL on error
 */

/*@dependent@*/ /*@null@*/ static CDframeset_t *
getFrameset(const CDheader_t *header, station_t *pStation)
{
   char framesetName[MAX_FRAMESET_NAME];
   CDframeset_t *actFrameset,
                *prevFrameset;

   buildFramesetName(framesetName, sizeof(framesetName), header->creator,
                     header->destination);

   if (NULL == pStation->appExtension)
   {
      /* application extension object has not been set up */
      shutdownApplication("Missing appExtension", "getFrameset");
      return(NULL);
   }

   actFrameset = ((cdrecv_t *)(pStation->appExtension))->firstFrameset;

   /* Search for frame set */
   prevFrameset = actFrameset;
   while (NULL != actFrameset)
   {
      if (strncmp(framesetName, actFrameset->name, MAX_FRAMESET_NAME) == 0)
          return(actFrameset);

      prevFrameset = actFrameset;
      actFrameset = actFrameset->nextFrameset;
   }

   /* No frame set found, create one */
   if (NULL == (actFrameset = calloc(1, sizeof(CDframeset_t))))
   {
      logError("Can't allocate memory for new frameset", "getFrameset");
      return (NULL);
   }

   DEBUG2(5,"New frameset created [%.20s]", framesetName);
   (void)strncpy(actFrameset->name, framesetName, (size_t)MAX_FRAMESET_NAME);

   /* Keep in mind that all frame sets are deleted when the connection
      is closed! */
   actFrameset->lowestSeqNbr = -1;
   /* highestSeqNbr and gapcount is set proper by calloc */

   /* insert new frame set in linked list */
   if (NULL == prevFrameset)
   {
      ((cdrecv_t *)(pStation->appExtension))->firstFrameset = actFrameset;
   }
   /*@-mustfreeonly@*/
   else prevFrameset->nextFrameset = actFrameset;
   /*@=mustfreeonly@*/

   return (actFrameset);
}

/**
 * @short delete gap from gap list
 *
 * @description This function deletes a gap from the gap list of a given
 * @description frame set. The gap to delete is marked with a value of
 * @description -1 in the gapBegin field
 *
 * @param frameset  frame set to modify
 * @param gapIndex  index in gap list
 *
 * @return nothing
 */

static void
removeGap(CDframeset_t *frameset, unsigned int gapIndex)
{
   if (gapIndex >= MAX_GAPS)
   {
      /* this should never ever happen */
      logError("Internal error [deleteGap]", "gap index out of range");
      return;
   }

   /* Remove gap. It's necessary to move down all
      remaining gaps one entry */
   memmove(&(frameset->gaps[gapIndex]), &(frameset->gaps[gapIndex + 1]),
                  sizeof(CDgaps_t) * (MAX_GAPS - 1 - gapIndex));
   frameset->gapCount--;
   return;
}

/**
 * @short insert new gap into gap list
 *
 * @description This function inserts a new gap into gap list of a given
 * @description frame set. If MAX_GAPS already exists, no new gap is
 * @description created and an error is reported to the calling function
 *
 * @param frameset  frame set to modify
 * @param gapBegin  lowest sequence number of the new gap
 * @param gapEnd    highest sequence number of the new gap
 * @param pStation  station object
 *
 * @return 0 on success or 1 if MAX_GAPS is reached
 */

static int
insertGap(CDframeset_t *frameset, uint64_t gapBegin, uint64_t gapEnd,
          station_t *pStation)
{
   uint32_t ii;

   if (MAX_GAPS == frameset->gapCount)
   {
      /* delete the gap with the lowest sequence number (first gap).
         This should be the oldest data */
      logCDevent(ET_W_CON, pStation, W_DROP_GAP, NULL, NO_INFO);
      ST_DEBUG(5,"remove gap");
      removeGap(frameset, 0);
      return(0);
      /*
      return(1); Will close connection
      */
   }

   if (MAX_GAPS == frameset->gapCount)
   {
      /* this should never ever happen */
      logError("Internal error [insertGap]", "wrong gapCount");
      return(0);
   }

   for (ii = 0; ii < MAX_GAPS; ii++)
   {
       if (frameset->gapCount == ii) break;

       if (frameset->gaps[ii].gapBegin > gapEnd)
       {
          ST_DEBUG(5, "Move gaps up");
          /* Higher gap found, insert new gap before
             this gap. It's necessary to move up all
             remaining gaps one entry */
          memmove(&(frameset->gaps[ii + 1]), &(frameset->gaps[ii]),
                  sizeof(CDgaps_t) * (MAX_GAPS - 1 - ii));
          frameset->gaps[ii].gapBegin = gapBegin;
          frameset->gaps[ii].gapEnd = gapEnd;
          frameset->gapCount++;
          return(0);
       }
   }

   if (frameset->gapCount == ii)
   {
      /* New gap is highest gap. Insert gap at the end */
      frameset->gaps[ii].gapBegin = gapBegin;
      frameset->gaps[ii].gapEnd = gapEnd;
      frameset->gapCount++;
   }
   return(0);
}

/**
 * @short update frame set
 *
 * @description This function updates a gap list for a given frame set
 *
 * @param header    header of received frame
 * @param pStation  station object
 *
 * @return 0 on success, 1 if MAX_GAPS is reached or -1 if no memory
 *         is available
 */

int
updateFrameset(const CDheader_t *header, station_t *pStation)
{
   /*@dependent@*/ CDframeset_t *frameset;
   uint32_t ii;
#define UPDATE_GAPS     0
#define NEW_GAP         1
   int processGaps;
   uint64_t gapBegin = 0,
            gapEnd = 0;

   /* If the frame set does not exist, a new frame set is created */
   frameset = getFrameset(header, pStation);

   if (NULL == frameset) return(-1);

   if ((uint64_t)-1 == frameset->lowestSeqNbr)
   {
      /* This is the first received frame */
      frameset->lowestSeqNbr = header->sequenceNbr;
      frameset->highestSeqNbr = header->sequenceNbr;
      return(0);
   }

   processGaps = UPDATE_GAPS;

   /* Update frame set */
   if (frameset->lowestSeqNbr > header->sequenceNbr)
   {
      if (frameset->lowestSeqNbr == (header->sequenceNbr + 1))
      {
         /* lowest sequence number - 1 is never a gap.
            The receiver will not raise a warning in this case */
         frameset->lowestSeqNbr = header->sequenceNbr;
         return(0);
      }
      processGaps = NEW_GAP;
      gapBegin = header->sequenceNbr + 1;
      gapEnd = frameset->lowestSeqNbr - 1;

      /* set new lowest sequence number */
      frameset->lowestSeqNbr = header->sequenceNbr;
   }
   else if (frameset->highestSeqNbr < header->sequenceNbr)
   {
      if (frameset->highestSeqNbr == (header->sequenceNbr - 1))
      {
         /* highest sequence number + 1 is never a gap */
         frameset->highestSeqNbr = header->sequenceNbr;
         return(0);
      }
      processGaps = NEW_GAP;
      gapBegin = frameset->highestSeqNbr + 1;
      gapEnd = header->sequenceNbr - 1;

      /* set new highest sequence number */
      frameset->highestSeqNbr = header->sequenceNbr;
   }

   ST_DEBUG3(5, "frame set [%.20s], sequence # [%lld]", frameset->name,
             (long long)header->sequenceNbr);
   ST_DEBUG3(5, "lowest [%lld], highest sequence # [%lld]",
             (long long)frameset->lowestSeqNbr,
             (long long)frameset->highestSeqNbr);
   ST_DEBUG2(5, "gapCount [%d]", (int)frameset->gapCount);

   if (UPDATE_GAPS == processGaps)
   {
      ST_DEBUG(5,"Update gap list");
      /* Sequence number is part of an existing gap or
         it's a resent frame */

      for (ii = 0; ii < MAX_GAPS; ii++)
      {
          if (frameset->gapCount == ii) break;
          if (frameset->gaps[ii].gapBegin == header->sequenceNbr)
          {
             /* related gap found */
             if (frameset->gaps[ii].gapEnd == header->sequenceNbr)
             {
                /* This is a one frame gap. Remove it */
                removeGap(frameset, (unsigned int)ii);
             }
             /* shrink gap from begin */
             else (frameset->gaps[ii].gapBegin)++;
             return (0);
          }
          else if (frameset->gaps[ii].gapEnd == header->sequenceNbr)
          {
             /* related gap found; shrink gap from the end */
             (frameset->gaps[ii].gapEnd)--;
             return (0);
          }
          else if (frameset->gaps[ii].gapBegin < header->sequenceNbr &&
                   frameset->gaps[ii].gapEnd > header->sequenceNbr)
          {
             /* related gap found. This gap must be split in two gaps */
             gapBegin = header->sequenceNbr + 1;
             gapEnd = frameset->gaps[ii].gapEnd;
             frameset->gaps[ii].gapEnd = header->sequenceNbr - 1;
             processGaps = NEW_GAP;
             break;
          }
      }

      if (frameset->gapCount == ii)
      {
         /* this sequence number is not seen the first time */
         logCDevent(ET_W_CON, pStation, W_SEQUENCE, NULL, NO_INFO);
         return(0);
      }
   }

   if (NEW_GAP == processGaps)
   {
      /* Sequence number creates a new gap.
         Only if a new gap is created a warning is raised  */
      if (pStation->pConfigData->systemConf & MORE_VERBOSE_LOGGING)
         logCDevent(ET_W_CON, pStation, W_NEW_GAP, NULL, NO_INFO);

      ST_DEBUG3(5, "new gap, lowest [%lld], highest sequence # [%lld]",
                (long long)gapBegin, (long long)gapEnd);

      return (insertGap(frameset, gapBegin, gapEnd, pStation));
   }

   return(0);
}

/**
 * @short delete frame sets
 *
 * @description This function deletes the gap lists for all known
 * @description frame sets
 *
 * @param pStation  station object
 *
 * @return nothing
 */

void
deleteFramesets(station_t *pStation)
{
   CDframeset_t *actFrameset,
                *prevFrameset;

   if (NULL == pStation->appExtension)
   {
      /* application extension object has not been set up */
      shutdownApplication("Missing appExtension", "deleteFramesets");
      return;
   }

   /* Search for frameset */
   actFrameset = ((cdrecv_t *)(pStation->appExtension))->firstFrameset;
   ((cdrecv_t *)(pStation->appExtension))->firstFrameset = NULL;

   while (NULL != actFrameset)
   {
      DEBUG2(2,"Delete frameset [%.20s]", actFrameset->name);
      prevFrameset = actFrameset;
      actFrameset = actFrameset->nextFrameset;
      (void)free((void *)prevFrameset);
   }
}
