/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: handling.c,v $
 *
 * Revision 1.98  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.97  2012-06-27 gerald
 * Fix function comment
 *
 * Revision 1.96  2012-01-02 12:41:26  ctbto
 * Fix invalid frame signature length check which does not
 * consider ECDSA sizes in CD receiver
 *
 * Revision 1.95  2011-09-27 08:33:18  ctbto
 * Allow data frames as first frames received in CD-1.1 connected
 * state in non-strict mode (requested by GP)
 *
 * Revision 1.94  2011-03-02 15:22:43  gerald
 * Add support for EC signature size
 *
 * Revision 1.93  2010-11-09 16:40:56  gerald
 * Raise warning if a data frame with frame destination != "0"
 * is received in non strict mode and raise an error, send
 * an alert frame and terminate the connection in strict mode
 * (requested by GP)
 *
 * Revision 1.92  2009-06-04 07:41:23  gerald
 * Don't send an Option Response Frame in offline mode. This
 * fixes the invalid offset in clf and txt files after parsing
 * a Disconnect Alert frame in cdrecv's offline mode
 *
 * Revision 1.91  2009-03-11 10:11:55  gerald
 * Fix splint warnings which occurred after upgrading
 * to new splint version
 *
 * Revision 1.90  2007/11/22 12:29:39  gerald
 * Add parameter for requested file creation time to syncFiles()
 *
 * Revision 1.89  2007/06/28 08:10:29  gerald
 * Fix wrong 'unsupported frame type' in text files
 *
 * Revision 1.88  2007/06/28 07:47:39  gerald
 * Save errno of last system error for logging in text
 * files (PR-0294)
 *
 * Revision 1.87  2007/05/10 09:09:45  gerald
 * Fix splint warnings
 *
 * Revision 1.86  2007/02/19 14:30:35  gerald
 * Add Idriz modifications to checkCDxxframeType() functions
 *
 * Revision 1.85  2007/02/12 10:17:05  gerald
 * Add virtual connect/disconnect (event) frame to clf file
 *
 * Revision 1.84  2007/01/24 09:11:40  gerald
 * Add computing and logging of send time span (rename
 * startRead to startRoW, endRead to endRoW, drop startWrite)
 *
 * Revision 1.83  2007/01/15 15:01:01  gerald
 * Channel name cache pointer has been replaced by a structure
 *
 * Revision 1.82  2007/01/11 15:55:58  gerald
 * Add channel name cache to parameter list when calling
 * parseAndCheckDFF() and parseAndCheckDataPayload()
 *
 * Revision 1.81  2006/07/20 13:22:56  gerald
 * Check IP address of incoming connection request in handling
 * thread again if a station file is available
 *
 * Revision 1.80  2006/05/24 13:11:56  gerald
 * Add check if frame has been signed by a forward station
 *
 * Revision 1.79  2006/05/11 07:48:38  gerald
 * Changes treatment of key caches.  A key cache does not longer belong
 * to one station. Instead a list of key caches is managed and each key
 * cache is globally accessable. A reference to a key cache is now assigned
 * to a station structure when the station connects for the first time
 *
 * Revision 1.78  2006/04/26 08:28:06  gerald
 * Move receiver specific config and station structure members
 * into config extension respectively station extension structure.
 * Move reading of receiver specific config parameters from libCD
 * to receiver specific function. Update comments
 *
 * Revision 1.77  2006/04/18 09:19:39  gerald
 * Rename defines for event types to avoid inadvertently use of
 * E_INTERNAL and fix typos
 *
 * Revision 1.76  2006/04/05 13:17:46  gerald
 * Move checkParsingAndAuthentication() to libCD
 *
 * Revision 1.75  2006/04/03 11:27:26  gerald
 * Debug messages improved
 *
 * Revision 1.74  2005/05/17 15:19:13  gerald
 * Make use of new error/warning statistics structure
 *
 * Revision 1.73  2005/05/11 09:31:25  gerald
 * Call updateStatistics() instead of updateStationStatistics() and
 * pass new statistics structure instead of whole station structure
 *
 * Revision 1.72  2005/05/04 08:22:46  gerald
 * Drop redirectPortInfo structure
 *
 * Revision 1.71  2005/04/29 10:43:07  gerald
 * Reset frame parsing code before an (unparsed) frame is sent
 *
 * Revision 1.70  2005/04/11 13:01:36  gerald
 * Reset nominal time when receive DF. This fixes logging
 * of nominal time taken from the previous data frame when current
 * frame is not parsed
 *
 * Revision 1.69  2005/02/21 15:42:06  gerald
 * Fix endless loop in cdrecv when a frame with frame length or
 * trailer offset of 0 is received. This fix causes different
 * treatment of frames with an invalid header. Before this fix,
 * a frame with an invalid frame length but an valid type was
 * parsed with the parsing functions for that frame type. Now
 * frames with an invalid frame length are treated as "unknown"
 * and re-acquiring is done instead
 *
 * Revision 1.68  2005/01/21 10:53:09  gerald
 * Move verifyChannelData to libCD
 *
 * Revision 1.67  2004/10/13 11:22:26  gerald
 * Write more accurate authentication state to wfdisc files.
 * This has also an impact on .txt files. The channel authentication
 * state has changed from 'Unknown' to 'Failed (no key)' if no keys
 * are available
 *
 * Revision 1.66  2004/09/16 15:01:42  gerald
 * Add connection state to sendFrame(). Fix splint warnings
 * (use adjustBufferSize())
 *
 * Revision 1.65  2004/09/10 10:37:22  gerald
 * Write debug messages to stderr instead of stdout. Frame buffer
 * is allocated when creating the station object now.
 *
 * Revision 1.64  2004/08/23 09:35:15  gerald
 * Set the 'not parsed flag' only for Data and Data Format Frames
 * after receiving a frame. Set frame type and size/auth offset for
 * data integrity test
 *
 * Revision 1.63  2004/07/07 14:07:40  gerald
 * Change behavior when MAX_GAPS limit is reached. Now the connection
 * is closed instead of removing the lowest gap. And the gap lists are
 * deleted when the connection is closed.
 *
 * Revision 1.62  2004/05/14 16:35:55  gerald
 * Move sendFrame() to libCD. Add check for invalid option type
 *
 * Revision 1.61  2004/05/14 11:35:10  georgian
 * Check for ORF option one
 *
 * Revision 1.60  2004/04/27 15:30:01  gerald
 * Move frame header check from  acquireCD11Framing() to
 * checkCD11frameType()
 *
 * Revision 1.59  2004/04/05 13:58:12  gerald
 * Fix wrong clftype (BAD_CON_RES instead of BAD_PAF) and wrong
 * event (E_SYSTEM instead of E_TIMEOUT)
 *
 * Revision 1.58  2004/03/23 15:27:18  gerald
 * Create only one binary file per day as long as the receiver
 * is not stopped
 *
 * Revision 1.57  2004/03/22 12:11:41  gerald
 * Rename define DATA_CHECK_ENABLE to ONLINE_MODE
 *
 * Revision 1.56  2004/03/09 13:57:59  gerald
 * Rename all variables which have tbs as part of the name and
 * rename WITHOUT_ defines
 *
 * Revision 1.55  2004/02/13 14:09:04  gerald
 * Move checkCD*frameType() to libCD
 *
 * Revision 1.54  2004/01/15 15:24:19  gerald
 * Change behaviour when a handling thread initializes too slow. Now
 * the shutdown flag is set for such a thread and the shutdown is done
 * by the thread itself instead of cancel the handling thread.
 *
 * Revision 1.53  2004/01/14 16:42:54  gerald
 * Fix wrong frame header values for CD1.1 frames in log files
 * when a timeout has occurred
 *
 * Revision 1.52  2004/01/12 15:06:23  gerald
 * checkFrameAuthentication() has got a more generic interface. Check
 * new return codes
 *
 * Revision 1.51  2003/12/05 08:54:19  gerald
 * Move checkFrameAuthentication() to libCD
 *
 * Revision 1.50  2003/12/03 14:43:49  gerald
 * Add station specific shutdown condition. Add station
 * specific system configuration flag. Fix wrong timeout when
 * re-acquire CD1_1 frames.
 *
 * Revision 1.49  2003/11/27 11:29:45  gerald
 * Add connection ID field to station structure
 *
 * Revision 1.48  2003/11/24 10:56:41  gerald
 * Replace global debug macro with station specific debug macro to
 * support station specific debug levels
 *
 * Revision 1.47  2003/11/14 13:46:01  gerald
 * Fix wrong comment
 *
 * Revision 1.46  2003/10/30 13:10:26  gerald
 * Add some comments and include new logfiles.h
 *
 * Revision 1.45  2003/10/17 14:01:17  gerald
 * Check for pending fast user shutdown request after a call to
 * writeWithTimeout()
 *
 * Revision 1.44  2003/08/27 15:49:56  gerald
 * Fix the handling of system errors in the handling thread. System
 * errors should now be handled and reported correctly
 *
 * Fix bug in signature verification of CD1.1 subframe data. Set
 * transformation type to '0' when data is signed before canadian
 * compression
 *
 * Revision 1.43  2003/05/26 08:37:33  gerald
 * Fix wrong data offset in wfdisc records for all channels in the
 * first frame after a binary file switch
 *
 * Revision 1.42  2003/05/23 10:44:38  gerald
 * Fix verification of compressed CD1.1 data if number of samples
 * is not a multiple of 20
 *
 * Revision 1.41  2003/05/21 08:10:23  gerald
 * Fix bug: CD1.1 data which were signed *AFTER* Canadian compression have
 * not been checked correct
 *
 * Revision 1.40  2003/05/14 15:34:15  gerald
 * Try alternate keys if first key does not verify correct
 *
 * Revision 1.39  2003/04/17 15:22:38  gerald
 * Add alternate key check for CD1 channel data. If verification with
 * first key fails try all alternate keys. Alternate keys are set during
 * caching of certificates (or ARM keys)
 *
 * Revision 1.38  2003/04/14 09:28:54  gerald
 * Fix bug: don't log previous frame in txt file when too short
 * frame with valid header is received  and a timeout occurs.
 *
 * Revision 1.37  2003/03/31 14:22:03  gerald
 * Add more debug messages when verifying channel authentication
 *
 * Revision 1.36  2003/03/27 13:59:49  gerald
 * Fix more splint warnings
 *
 * Revision 1.35  2003/03/19 14:52:04  gerald
 * Re-work frame acquire functions according to John's comments
 * Set frame parsing code "Unknown frame Type" when receiving
 * unknown frame
 *
 * Revision 1.34  2003/03/18 15:06:14  gerald
 * Set frame signature check state "Unknown" if no certificates are
 * loaded
 *
 * Revision 1.33  2003/03/18 13:29:20  gerald
 * Log more detailed frame/channel signature check result
 *
 * Revision 1.32  2003/03/17 14:28:13  gerald
 * Add key and site/channel specific NONE directives. Allow NONE
 * directives for unsigned channels.
 *
 * Revision 1.31  2003/03/12 14:21:11  gerald
 * Set correct channel authentication state if no keys available
 * or station NONE directive. Fix wrong channel authentication
 * states when no certificate is found for the previous channel.
 * Add channel authentication summary. Remove trailing blanks.
 *
 * Revision 1.30  2003/03/06 13:19:07  gerald
 * Add authentication states "skipped" and "switched off". Fix wrong
 * frame authentication state when at least a channel authentication
 * failed. Check if channel is signed when channel authentication is
 * switched on. Make verifyChannelData generic useable. Write wfdisc
 * records also for "bad" data frames (eg. authentication failure).
 *
 * Revision 1.29  2003/02/17 15:48:49  ctbto
 * Add check if buffer is large enough for AckNack frame
 * Reduce messages to syslog if verboseEnabel is set to NO
 *
 * Revision 1.28  2003/02/12 09:36:16  ctbto
 * Fix warnings that occur when compiling with  -Wall
 * Fix core dump when check signature for uncompressed or "Canadian
 * before signing" data channels
 * Fix bug in authentication check for CD1.1 channel data
 * Fix incorrect authentication state for frames with an valid
 * frame signature
 *
 * Revision 1.27  2003/02/07 15:55:11  ctbto
 * Add frame/channel parsing code for received frames
 * Add signature verification status for all received CD1.1 frames
 * Fix checks for "bad" frames. A frame becomes invalid if at least
 * a warning occurs during parsing or authentication check
 *
 * Revision 1.26  2003/01/31 12:55:04  ctbto
 * Don't check if the first CD1.1 frame is an option request frame in
 * offline mode
 *
 * Revision 1.25  2003/01/20 09:20:07  ctbto
 * Rename getSignature() to signData()
 *
 * Revision 1.24  2003/01/13 14:57:46  ctbto
 * Fix bug in handleCD11(): authentication warnings were counted to
 * connection request warnings
 *
 * Fix bug that wrong frame destination is sent back sometimes in OSF and
 * AF frames
 *
 * Remember parsed header for further use (write log files for example)
 *
 * Revision 1.23  2002/12/16 13:33:31  ctbto
 * Check signatures of CD1.1 frames and sign outgoing CD1.1 frames
 *
 * Revision 1.22  2002/12/09 13:16:38  ctbto
 * Continue work on authentication (load local PEM keys, check CD1 data
 * signatures, log authentication information)
 *
 * Revision 1.21  2002/11/29 09:51:01  ctbto
 * Add reacquire for CD1.1
 * Call saveInUserFormat() on each received valid data frame if build
 * with user extension (see configure -help)
 *
 * Revision 1.20  2002/11/11 14:54:53  ctbto
 * Fix bug which causes wrong cdreciver keyID is sent back to sender
 *
 * Revision 1.19  2002/10/30 10:01:21  ctbto
 * Fix bug that sent frames have wrong sending time
 * Complete AckNak code (sending heartbeat frames)
 *
 * Revision 1.18  2002/10/15 07:56:40  ctbto
 * Add check for 'could not save data' condition
 * Start work for sending CD-1.1 AckNack frames
 *
 * Revision 1.17  2002/10/10 09:54:38  ctbto
 * Include John's comments
 *
 * Revision 1.16  2002/09/26 11:57:45  ctbto
 * Fix bug in findDataType() that datatype is always s4 for CD-1 frames.
 * Write station connect and disconnect info to txt file
 *
 * Revision 1.15  2002/09/24 14:46:26  ctbto
 * Count station handling warnings (one warning per invalid frame, not per
 * error/warning in frame),  Log valid incoming frames only if MORE verbose
 * mode is set. Do some code cleanup (updateStatistics, check return values)
 *
 * Revision 1.14  2002/09/20 07:32:03  ctbto
 * clean frame.c frame.h from receiver specific code. This allows easy use
 * of frame parsing functions by other applications, but slow down receiver
 * a little bit in non strict mode if many warnings are found during parsing.
 *
 * Revision 1.13  2002/09/11 10:03:51  ctbto
 * Improve shutdown code. Fix bug with wrong filenames in wfdisc records.
 * Write receive time to lddate field in wfdisc records.
 *
 * Revision 1.12  2002/09/05 16:20:52  ctbto
 * Include John's comments (rename functions, move functions to new files,
 * code changes, etc.)
 *
 * Revision 1.11  2002/09/03 07:54:41  ctbto
 * Fix bug that first data frame in bin file was not written to wfdisc file
 * Fix bug that calibration information (calib, calper) was not parsed correct
 *
 * Revision 1.10  2002/09/02 11:34:07  ctbto
 * Use same data structure for CD1 and CD1.1 data information
 * Do more CD1.1 checks and write CD1.1 wfdisc records
 *
 * Revision 1.9  2002/08/26 14:38:52  ctbto
 * Add processing of received AckNack Frames, CRC check for CD_1.1 frames and
 * rename some variables to be closer to F&P names
 *
 * Revision 1.8  2002/07/23 09:05:11  ctbto
 * Add writing wfdisc records and re-acquiring for CD1 connections.
 * Add processing Option Response Frames. Fix some bugs.
 *
 * Revision 1.7  2002/07/12 10:03:03  ctbto
 * Correct some bugs, parse and check DF, add wfdiscs but don't write it
 *
 * Revision 1.6  2002/07/04 15:53:54  ctbto
 * Add duplication of DFF frames. Correct some SPLINT warnings
 *
 * Revision 1.5  2002/07/03 17:53:44  ctbto
 * Use stack instead of stations logBuffer. Add different frame types to
 * updateStatistic() calls. Add information whether or not it is called
 * from a connection request to saveData(). Add more checks for DFF and AF
 * frames.
 *
 * Revision 1.4  2002/06/26 14:54:30  ctbto
 * Add trivial CD11 handling (log frames, send alert frame)
 * Include checkDFF in parseDFF -> checkAndParseDFF(). Check alone
 * makes no sense.
 *
 * Revision 1.3  2002/06/21 12:43:57  ctbto
 * Add functions for parsing and checking DF frames. Add function
 * for parsing Alert frame, send Alert Frame when terminating.
 *
 * Revision 1.2  2002/06/20 07:56:24  ctbto
 * Add DFF parsing
 *
 * Revision 1.1.1.1  2002/06/18 07:28:05  ctbto
 * Import sources
 *
 */

#include <config.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "cdrecv.h"
#include "events.h"
#include "gaps.h"
#include "logfiles.h"
#include "readwrite.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

/**
 * @short prepare trailer
 *
 * @description This function prepares a CD1.1 frame trailer for
 * @description for sending
 *
 * @param pStation       pointer to station object
 * @param buffer         data to send
 * @param length         length of data to send without trailer
 *
 * @return 0 on failure or length of frame to sent
 */

static uint32_t
prepareTrailer(station_t *pStation, char *buffer, size_t length)
{
#ifdef HAVE_AUTHENTICATION

   char signature[SZ_SIGNATURE_MAX];
   const configParams_t *pConfigData = pStation->pConfigData;
   size_t sigLen = sizeof(signature);

   if (NULL == pConfigData->privateKey)
   {
      return(setTrailer(buffer, length, 0, NULL, 0));
   }

   /* Frame authentication is active */
   if (signData(pConfigData->privateKey, buffer, length, signature, &sigLen))
   {
      /* something unexpected has happen during signature
         verification. Shutdown receiver?? */
      logError("Unexpected error while signing data", "prepareTrailer");

      pStation->whenConnected.systemFailures++;
      logCDevent(ET_E_CON, pStation, E_OOPS, NULL, NO_INFO);

      if (pStation->systemConf & STRICT_MODE_ENABLE)
      {
         return (0);
      }
      /* Send unsigned frame */
      return(setTrailer(buffer, length, 0, NULL, 0));
   }

   return(setTrailer(buffer, length, pConfigData->privateKeyID, signature,
                     sigLen));
#else

   return(setTrailer(buffer, length, 0, NULL, 0));

#endif /* HAVE_AUTHENTICATION */
}

/**
 * @short try to re-acquire CD1.1 framing
 *
 * @description This function handles CD11 frames with an invalid ID.
 * @description Current strategy is to search through data until a valid
 * @description frame ID is located. To try and ensure that this sequence
 * @description actually represents a valid frame attempts are made
 * @description to ensure that the next 72 bytes of the frame match what
 * @description is expected i.e. is the frame header valid, has the
 * @description payload reasonable values, etc.
 * @description At this point the data which has been parsed before the frame
 * @description ID was located is stored as an unknown frame and hopefully
 * @description receiving can continue as normal. If the number of bytes parsed
 * @description exceeds the current maximal frame size, reacquire is stopped
 * @description and an error is returned
 *
 * @param sd                socket descriptor
 * @param frameBuffer       input buffer
 * @param frameBufferSize   size of input buffer
 * @param pStation          pointer to station object
 * @param bytes             bytes read (returned)
 *
 * @return event S_ACQUIRE, E_ACQUIRE, E_TIMEOUT, E_SYSTEM, E_CLOSED
 *               or S_SHUTDOWN
 */

static uint16_t
acquireCD11Framing(int sd, char *frameBuffer, uint32_t frameBufferSize,
                  const station_t *pStation, int *bytes)
{
   rwFlags_t flags;
   int bytesRead = 0;
   uint32_t totalPeek,   /* Number of bytes written to frameBuffer */
            intBuffer[18]; /* Buffer for reacquiring */
   const configParams_t *pConfigData = pStation->pConfigData;
   double nomTime;

   ST_DEBUG(5,"Try to re-acquire CD1.1 framing");
   ST_DEBUG2(9, "Timeout = %d", (int) pConfigData->timeOut11);

   totalPeek  = 0;
   flags.shutdownFlag = &(pStation->shutdownEnable);
   memset(intBuffer, 0, sizeof(intBuffer)); /* for SPLINT */

   /*Loop until have found beginning of next valid frame*/
   do
   {
      /* Parse data until locate a possible valid frame ID or until MFS
         bytes are parsed */
      while (totalPeek < frameBufferSize)
      {
         flags.resetDataTimeout = pConfigData->timeOut11;
         /* Read 1 byte and store it in the input buffer */
         flags.timeout = pConfigData->timeOut11;
         flags.sockFlags = 0;
         flags.firstTime = NULL;
         bytesRead = readWithTimeout(sd, &frameBuffer[totalPeek], 1, &flags);

         if (bytesRead != 1)
         {
            /* timeout or connection closed */
            if (0 == flags.timeout) return (E_TIMEOUT);
            else if (-1 == bytesRead) return (E_SYSTEM);
            /* No data is saved on fast user requested shutdown!! */
            else if (-2 == bytesRead) return (S_SHUTDOWN);
            else return (E_CLOSED);
         }

         totalPeek++;
         *bytes = (int) totalPeek;

         /* Now read 72 bytes to check for valid frame IDs,
            but leave them in input buffer */
         flags.sockFlags = MSG_PEEK;
         flags.timeout = pConfigData->timeOut11;
         bytesRead = readWithTimeout(sd, (char *)intBuffer, 72, &flags);

         if (bytesRead != 72)
         {
            /* timeout or connection closed */
            if (0 == flags.timeout) return (E_TIMEOUT);
            else if (-1 == bytesRead) return (E_SYSTEM);
            /* No data is saved on fast user requested shutdown!! */
            else if (-2 == bytesRead) return (S_SHUTDOWN);
            else return (E_CLOSED);
         }

         /* Run some validity checks depending on frame type */
         if (0 == checkCD11frameType(intBuffer, &nomTime)) return(S_ACQUIRE);

      } /* end while */

      if (totalPeek >= frameBufferSize)
      {
         /* Max. current buffer size reached, giving up */
         return(E_ACQUIRE);
      }

   } while ((unsigned char) 0 == pStation->shutdownEnable);

   /* User has requested shutdown */
   return(S_SHUTDOWN);
}

/**
 * @short try to re-acquire framing
 *
 * @description This function handles CD1 frames with an invalid ID.
 * @description Current strategy is to search through data until a valid
 * @description frame ID is located. To try and ensure that this sequence
 * @description actually represents a valid frame boundary, attempts are made
 * @description to ensure that the next 20 bytes of the frame match what
 * @description is expected i.e. are the frame sizes valid, is the nominal
 * @description time reasonable, etc. At this point the data which has
 * @description been parsed before the frame ID was located is stored as an
 * @description unknown frame and hopefully receiving can continue as normal.
 * @description If the number of bytes parsed exceeds the current maximal
 * @description frame size, reacquire is stopped and an error is returned
 *
 * @param sd                socket descriptor
 * @param frameBuffer       input buffer
 * @param frameBufferSize   size of input buffer
 * @param pStation          pointer to station object
 * @param bytes             bytes read (returned)
 *
 * @return event S_ACQUIRE, E_ACQUIRE, E_TIMEOUT, E_SYSTEM, E_CLOSED or
 *               S_SHUTDOWN
 */

static uint16_t
acquireCD1Framing(int sd, char *frameBuffer, uint32_t frameBufferSize,
                  const station_t *pStation, int *bytes)
{
   rwFlags_t flags;
   int bytesRead = 0;
   uint32_t totalPeek,   /* Number of bytes written to frameBuffer */
            intValue,
            intBuffer[5]; /* Buffer for re-acquiring */
   const configParams_t *pConfigData = pStation->pConfigData;
   double nomTime;

   ST_DEBUG(5,"Try to re-acquire CD1 framing");
   ST_DEBUG2(9, "Timeout = %d", (int) pConfigData->timeOut);

   totalPeek  = 0;
   flags.shutdownFlag = &(pStation->shutdownEnable);
   memset(intBuffer, 0, sizeof(intBuffer)); /* for SPLINT */

   /*Loop until have found beginning of next valid frame*/
   do
   {
      /* Parse data until locate a possible valid frame ID or until MFS
         bytes are parsed */
      while (totalPeek < frameBufferSize)
      {
         flags.resetDataTimeout = pConfigData->timeOut;
         /* Read 1 byte and store it in the input buffer */
         flags.timeout = pConfigData->timeOut;
         flags.sockFlags = 0;
         flags.firstTime = NULL;
         bytesRead = readWithTimeout(sd, &frameBuffer[totalPeek], 1, &flags);

         ST_DEBUG4(9,"bytesRead [%d], totalPeek [%u], frameBufferSize [%u]",
                   bytesRead, (unsigned int) totalPeek,
                   (unsigned int) frameBufferSize);

         if (bytesRead != 1)
         {
            /* timeout or connection closed */
            if (0 == flags.timeout) return (E_TIMEOUT);
            else if (-1 == bytesRead) return (E_SYSTEM);
            /* No data is saved on fast user requested shutdown!! */
            else if (-2 == bytesRead) return (S_SHUTDOWN);
            else return (E_CLOSED);
         }

         totalPeek++;
         *bytes = (int) totalPeek;

         /* Now read 20 bytes to check for valid frame ID,
            but leave them in input buffer */
         flags.sockFlags = MSG_PEEK;
         flags.timeout = pConfigData->timeOut;
         bytesRead = readWithTimeout(sd, (char *)intBuffer, 20, &flags);

         ST_DEBUG4(9,"bytesRead [%d], totalPeek [%u], frameBufferSize [%u]",
                   bytesRead, (unsigned int) totalPeek,
                (unsigned int) frameBufferSize);

         if (bytesRead != 20)
         {
            /* timeout or connection closed */
            if (0 == flags.timeout) return (E_TIMEOUT);
            else if (-1 == bytesRead) return (E_SYSTEM);
            /* No data is saved on fast user requested shutdown!! */
            else if (-2 == bytesRead) return (S_SHUTDOWN);
            else return (E_CLOSED);
         }

         intValue = ntohl(intBuffer[0]);

         if (intValue == 1 || intValue == 2 || intValue == 4)
         {
           break;
         }

      } /* end while */

      if (totalPeek >= frameBufferSize)
      {
         /* Max. current buffer size reached, giving up */
         return(E_ACQUIRE);
      }

      /* A valid frame ID has been located check next 20-bytes
         to ensure compliance */
      if (0 == checkCD1frameType(intBuffer, &nomTime)) return(S_ACQUIRE);

   } while ((unsigned char) 0 == pStation->shutdownEnable);

  /* User has requested shutdown */
  return(S_SHUTDOWN);
}

/**
 * @short send AckNack Frame or Frames
 *
 * @description This function sends an AckNack Frame for each received frame
 * @description set to acknowledge received data frames. If no frame set
 * @description is known, an AckNak frame with a frame set named <creator>:
 * @description <stationname> and no gaps is sent
 *
 * @param pStation  pointer to station object
 * @param buffer    receive buffer
 * @param creator   ID of frame creator
 *
 * @return event
 */

static uint16_t
sendAckNack(station_t *pStation, char *buffer, const char *creator)
{
   uint16_t event = NO_ERROR;
   uint32_t ii;
   /*@dependent@*/ CDframeset_t *actFrameset;

   /* Receiver uses the received buffer to build response frame.
      Destination field is set to station name, because frame set
      value is in the payload. */

   ST_DEBUG(5,"Send AckNack Frame(s)");

   if (NULL == pStation->appExtension)
   {
      /* application extension object has not been set up */
      shutdownApplication("Missing appExtension", "sendAckNack");
      return(E_OOPS);
   }

   actFrameset = ((cdrecv_t *)(pStation->appExtension))->firstFrameset;

   do
   {
#if (INIT_BUFFER_SIZE) < MAX_SIZE_ACKNACK_FRAME
     /* Oops! Initial buffer size is too small for largest
        AckNack frame. Stop compilation with an error */
     #error Increase initial CD11 buffer size
#endif

      /* Send one AckNak Frame for every known Frame set */
      ii = setAckNakPayload(buffer + SZ_HDR, actFrameset, creator,
                            pStation->stationName);
      setHeader(buffer, ACK_NACK, SZ_HDR + ii, creator, pStation->stationName,
                (uint64_t) 0);
      /* Copy all needed values in the trailer. This includes also
         the signature if signing is on */
      ii = prepareTrailer(pStation, buffer,  SZ_HDR + ii);
      if (0 == ii) break;

      /* statistic is updated and data is saved in sendFrame */
      pStation->checkPass = 0;
      event = sendFrame(pStation, buffer, (int)ii, S_SEND_ACKNAK,
                        GOOD_ACKNAK, 1);

      if (NO_ERROR != event) break;

      if (NULL != actFrameset) actFrameset = actFrameset->nextFrameset;
   }
   while (NULL != actFrameset);

   return(event);
}

/**
 * @short Process invalid frame
 *
 * @description This function process a frame with an invalid frame
 * @description type or invalid frame size if the receiver is not
 * @description in strict mode. Otherwise it sets only the proper event
 *
 * @param pStation     pointer to station object
 * @param event        event type
 * @param bytesRead    pointer to number of bytes read (out)
 * @param frameLen     frame length field of received frame
 * @param minFrameLen  minimum valid frame length
 *
 * @return event (S_ACQUIRE or an error event)
 */

static uint16_t
processInvalidFrame(station_t *pStation, uint16_t event, int *bytesRead,
                    int frameLen, int minFrameLen)
{
   pStation->whenConnected.invalidFrames++;

   /* Invalid frame type or frame size. Set correct event */
   if (frameLen > MAX_FRAME_LEN)
   {
      event = E_FRAME_TOO_BIG;
      pStation->checkPass |= ER_FRAME2BIG;
   }
   else if (frameLen < minFrameLen || (frameLen % 4) != 0)
   {
      event = E_FRAMESIZE;
      pStation->checkPass |= ER_SIZE;
   }
   else pStation->checkPass |= WA_UNKNOWN_TYPE;

   if (0 != (pStation->systemConf & STRICT_MODE_ENABLE))
   {
      return (event);
   }

   /* log event which cause reacquire */
   logCDevent(ET_W_CON, pStation, event, NULL, (uint32_t)frameLen);

   /* Non strict mode; try to reacquire framing */
   *bytesRead = 0;
   logCDevent(ET_W_CON, pStation, W_ACQUIRE, NULL, NO_INFO);

   if (CD1 == pStation->type)
   {
      event = acquireCD1Framing(pStation->portInfo->socket,
                                pStation->frameBuffer,
                                pStation->frameBufferSize,
                                pStation, bytesRead);
   }
   else
   {
      event = acquireCD11Framing(pStation->portInfo->socket,
                                 pStation->frameBuffer,
                                 pStation->frameBufferSize,
                                 pStation, bytesRead);
   }

   if (S_ACQUIRE != event)
   {
       return (event);
   }

   /* valid frame found. Log invalid frame event, update
      statistic and save data */
   (void) gettimeofday(&(pStation->endRoW), NULL);
   if (pStation->systemConf & VERBOSE_LOGGING)
   {
      logCDevent(ET_S_CON, pStation, event, NULL, NO_INFO);
   }

   updateStatistics(&(pStation->framesReceived), *bytesRead, NO, FRAME_UNKNOWN);
   if (saveData(pStation, pStation->frameBuffer, *bytesRead, 0, event,
                NO, FRAME_UNKNOWN))
   {
       /* If data could not be saved, the connection will be terminated.
          If there is insufficient disc space, the receiver will
          be stopped too. This is done by saveData() */
       event = E_SAVE_DATA;
       pStation->whenConnected.systemFailures++;
       *bytesRead = 0;
    }

    return(event);
}

/**
 * @short handles connection for a CD1.1 station
 *
 * @description This function handles the connection for a CD1.1 station.
 * @description It processes all incoming data, sends an Option Response
 * @description Frame as answer to an Option Request Frame and sends an Alert
 * @description Frame to the connected station if an error occurs or the
 * @description receiver is shutting down
 *
 * @param pStation  pointer to station object
 *
 * @return event
 */

#define HEARTBEAT     ((short)1)
#define LINE_IDLE     ((short)2)

static uint16_t
handleCD11(station_t *pStation)
{
   rwFlags_t flags;
   const configParams_t *pConfigData;
   int sd = pStation->portInfo->socket;
   char *buffer = pStation->frameBuffer,
        clfType;
   uint16_t event = E_FRAME_TYPE,   /* ID of occurred event */
            retValue,
            eventType,
            mustBeORF = 0;   /* First frame must be an Option Request */
   int frameLen,        /* len of received frame */
       sigSize,
       bytesRead;
   uint32_t frameID,
            strictMode,
            addInfo,
            ii,
            *pIntBuffer;
   CDtrailer_t trailer;
   long nextAckNackTime,
        nextConnectionTimeout;
   short nextTimeout;
   struct timeval nowTime;
   CDheader_t *header = &(pStation->header);

   /* Note: There is no station name in the frame header. This can cause a big
      security hole if the receiver has no information which frame sets
      belongs to which station or which site belongs to which station  */

   /* buffer should be aligned by malloc() */
   pIntBuffer = (uint32_t *)buffer;
   pConfigData = pStation->pConfigData;
   flags.shutdownFlag = &(pStation->shutdownEnable);

   /* Only in online mode must the first frame be an option request frame */
   if (pStation->systemConf & ONLINE_MODE) mustBeORF = 1;

   /* Set initial AckNack send and connection timeout time */
   (void) gettimeofday(&nowTime, NULL);
   nextAckNackTime = nowTime.tv_sec + (long)pConfigData->heartBeat;
   nextConnectionTimeout = nowTime.tv_sec + (long)pConfigData->timeOut11;
   nextTimeout = LINE_IDLE;
   /* Set Line Idle Timeout. This timeout is set to resetDataTimeout
      on each received byte by readWithTimeout */
   flags.timeout = pConfigData->timeOut11;

   do
   {
      addInfo = NO_INFO;
      clfType = FRAME_UNKNOWN;
      event = E_FRAME_TYPE;
      eventType = ET_W_CON;
      /* config data could change during runtime, so it is
         re-read on each frame */
      strictMode = pStation->systemConf & STRICT_MODE_ENABLE;
      flags.resetDataTimeout = pConfigData->timeOut11;
      pStation->checkPass = NO_ERROR;
      pStation->authenPassFrame = AUTH_UNKNOWN; /* default value */
      pStation->authenSummary = AUTH_UNKNOWN; /* default value */

      /* Reset saved frame header from previous frame */
      memset(&(pStation->header), 0, sizeof(pStation->header));

      /* Check if AckNack frame should be send */
      if (pConfigData->heartBeat > 0)
      {
         /* Receiver must send AckNack frames. Note that the receiver
            doesn't send heartbeat frames when receiving a frame.
            If the heartbeat interval has to be guaranteed there are
            only two possibilities: send heartbeat in own thread or change
            readWithTimeout() to send heartbeats also while receiving
            a frame (but then readWithTimeout could not be a generic
            function).  */

         (void) gettimeofday(&nowTime, NULL);
         if (nextAckNackTime <= nowTime.tv_sec)
         {
            /* Set next AckNack time */
            nextAckNackTime = nowTime.tv_sec + (long)pConfigData->heartBeat;

            /* Send AckNack Frame */
            if ((event = sendAckNack(pStation, buffer,
                                     pConfigData->frameCreator)))
            {
               break;
            }
         }

         /* Compute next timeout */
         if (nextAckNackTime < nextConnectionTimeout)
         {
            flags.timeout =  pConfigData->heartBeat;
            nextTimeout = HEARTBEAT;
         }
         else
         {
            flags.timeout = (int32_t)(nextConnectionTimeout - nowTime.tv_sec);
            nextTimeout = LINE_IDLE;
         }

         /* There is a small chance that the timeout becomes null.
            In this case a one second timeout is used */
         if (flags.timeout < 1) flags.timeout = 1;
      }

      /* Read first LOOKUP_CD11 bytes of frame but leave them in
         input buffer */

      flags.sockFlags = MSG_PEEK;
      flags.firstTime = &(pStation->startRoW);
      bytesRead = readWithTimeout(sd, buffer, LOOKUP_CD11, &flags);
      flags.firstTime = NULL;

      if (-1 == bytesRead)
      {
         /* Interrupt or system failure */
         pStation->savedSystemErr = (uint32_t)errno;
         pStation->whenConnected.systemFailures++;
         event = E_SYSTEM;
         addInfo = (uint32_t)errno;
         break;
      }
      else if (-2 == bytesRead)
      {
         /* No data is saved on fast user requested shutdown!! */
         event = S_SHUTDOWN;
         break;
      }

      if (LOOKUP_CD11 != bytesRead)
      {
         /* timeout or connection closed */
         if ( 0 == flags.timeout)
         {
            /* check for heartbeat timeout. When at least one
               byte is received this is no heartbeat timeout per
               definition */
            if (0 == bytesRead && HEARTBEAT == nextTimeout) continue;

            pStation->whenConnected.timeouts++;
            event = E_TIMEOUT;
         }
         else
         {
            pStation->whenConnected.closedConnections++;
            event = E_CLOSED;
         }
         break;
      }

      /* Read frame ID and offset to trailer */
      pStation->header.frameType = frameID = ntohl(pIntBuffer[0]);
      /* trailer offset+first two int in trailer+CRC */
      pStation->header.trailerOffset = ntohl(pIntBuffer[1]);
      frameLen = (int)pStation->header.trailerOffset + 8 + 8;

      IF_ST_DEBUG(9)
      {
         S_DEBUG_HDR;
         (void) fprintf(DEBUG_FP, "Receive %02x %02x %02x %02x ",
                       (unsigned int)buffer[0], (unsigned int)buffer[1],
                       (unsigned int)buffer[2], (unsigned int)buffer[3]);
         (void) fprintf(DEBUG_FP, "ID=%u, len=%d\n", (unsigned int)frameID,
                        frameLen);
         DEBUG_FOOTER;
      }

      /* Check if the frame size is valid. Check at least for a
         frame size of zero or an endless loop will occur */
      if (frameLen < MAX_FRAME_LEN && frameLen >= MIN_SZ_CD11_FRAME &&
          (frameLen % 4) == 0)
      {
         /* Set clf type. This is necessary to identify the frame in
            the clf file. Data is written to the configured output files,
            even if an error occurs.  */
         if (DATA == frameID)
         {
            /* It was a requirement that the parsing functions can be
             * used by other applications. This caused that parsing and
             * logging is separated in different functions. But than it's
             * hard for the logging functions to know what data could be
             * logged if an error has been occurred. So an indicator is
             * set that the current frame is not parsed complete here
             * and reset if the frame has parsed complete */
            pStation->checkPass |= ER_NOT_PARSED;
            clfType = BAD_DATA;
            /* reset number of parsed channels */
            if (NULL != pStation->DFinfo)
            {
               pStation->DFinfo->parsedChannels = 0;
               pStation->DFinfo->nominalTime = 0.0;
            }
         }
         else if (ALERT == frameID) clfType = BAD_ALERT;
         else if (ACK_NACK == frameID) clfType = BAD_ACKNAK;
         else if (OPTION_REQUEST == frameID) clfType = BAD_OPTION_REQ;
         else if (OPTION_RESPONSE == frameID) clfType = BAD_OPTION_RES;
         else if (COMMAND_REQUEST == frameID) clfType = BAD_COMMAND_REQ;
         else if (COMMAND_RESPONSE == frameID) clfType = BAD_COMMAND_RES;
         else if (CONNECTION_REQUEST == frameID) clfType = BAD_CRF;
         else if (CONNECTION_RESPONSE == frameID) clfType = BAD_CON_RES;
      }

      /* Check if we have a frame type now */
      if (FRAME_UNKNOWN == clfType)
      {
         /* SZ_SIGNATURE is not considered (counted to the frameLen) here */
         event = processInvalidFrame(pStation, event, &bytesRead, frameLen,
                                     MIN_SZ_CD11_FRAME);

         /* set addInfo if there was a system error */
         if (E_SYSTEM == event)
         {
            addInfo = (uint32_t)errno;
            pStation->savedSystemErr = (uint32_t)errno;
         }

         if (S_ACQUIRE != event)
         {
            /* re-acquire failed or strict mode */
            break;
         }

         /* re-acquire successful. Data is received, so line idle
            timeout is reset. Continue with next frame */
         nextConnectionTimeout = pStation->endRoW.tv_sec +
                                 (long)pConfigData->timeOut11;

         continue;
      }

      /* Check if the buffer size is sufficient. Adjusting buffer size could
         be necessary after receiving a DF */
      if (pStation->frameBufferSize < (uint32_t)(frameLen + SZ_SIGNATURE_MAX))
      {
         ii = (uint32_t)frameLen + SZ_SIGNATURE_MAX;
         if (-1 == adjustBufferSize(pStation, ii, &event))
         {
            addInfo = ii;
            /* Buffer is needed for sending Alert frame. Old buffer is
               not released when no bigger one is allocated */
            break; /* break do loop; SPLINT does not realize this */
         }
         /* buffer should be aligned by malloc() */
         buffer = pStation->frameBuffer;
         pIntBuffer = (uint32_t *)buffer;
      }

      /* Read frame without signature now. When data is received,
         no heartbeat timeout is checked */
      flags.timeout = pConfigData->timeOut11;
      flags.sockFlags = 0;
      bytesRead = readWithTimeout(sd, buffer, frameLen, &flags);
      (void) gettimeofday(&(pStation->endRoW), NULL);

      if (-1 == bytesRead)
      {
         /* Interrupt or system failure */
         pStation->whenConnected.systemFailures++;
         event = E_SYSTEM;
         addInfo = (uint32_t)errno;
         pStation->savedSystemErr = (uint32_t)errno;
         bytesRead = LOOKUP_CD11;
         break;
      }
      else if (-2 == bytesRead)
      {
         /* No data is saved on fast user requested shutdown!! */
         event = S_SHUTDOWN;
         break;
      }

      /* Parse CD1.1 header if enough bytes received */
      if (bytesRead >= SZ_HDR) parseHeader(buffer, &(pStation->header));

      if (frameLen != bytesRead)
      {
         /* timeout or connection closed */
         if ( 0 == flags.timeout)
         {
            pStation->whenConnected.timeouts++;
            event = E_TIMEOUT;
         }
         else
         {
            pStation->whenConnected.closedConnections++;
            event = E_CLOSED;
         }
         break;
      }

      /* Check if frame has a signature. This check works well for
         DSS (DSA) and ECDSA signatures in non ASN.1 format */
      memcpy(&ii, &buffer[frameLen - 12], sizeof(ii));
      sigSize = (int)(ntohl(ii));
      if (IS_VALID_SIG_SIZE(sigSize)) /*trailer->signatureSize*/
      {
         int nbrOfBytes;

         /* Frame has a signature. Read rest of frame.
            First 16 bytes of trailer are already read, so we must read
            only signature size bytes. Note: This additional read tells
            the protocol could be smarter */

         flags.timeout = pConfigData->timeOut11;
         nbrOfBytes = readWithTimeout(sd, buffer + frameLen, sigSize, &flags);
         (void) gettimeofday(&(pStation->endRoW), NULL);

         if (-1 == nbrOfBytes)
         {
            /* Interrupt or system failure */
            pStation->whenConnected.systemFailures++;
            event = E_SYSTEM;
            addInfo = (uint32_t)errno;
            pStation->savedSystemErr = (uint32_t)errno;
            break;
         }
         else if (-2 == bytesRead)
         {
            /* No data is saved on fast user requested shutdown!! */
            event = S_SHUTDOWN;
            break;
         }

         bytesRead += nbrOfBytes;

         if (nbrOfBytes != sigSize)
         {
            ST_DEBUG3(5, "Try to read [%d], read [%d]", sigSize, nbrOfBytes);

            /* timeout or connection closed */
            if ( 0 == flags.timeout)
            {
               pStation->whenConnected.timeouts++;
               event = E_TIMEOUT;
            }
            else
            {
               pStation->whenConnected.closedConnections++;
               event = E_CLOSED;
            }
            break;
         }
      }

      /* Data is received, so line idle timeout is reset */
      nextConnectionTimeout = pStation->endRoW.tv_sec +
                              (long)pConfigData->timeOut11;

      /* Frame received, parse CD1.1 trailer */
      parseTrailer(&buffer[frameLen - 16], &trailer);

      /* Check CRC */
      if (CRC64((unsigned char *)buffer, (uint32_t)bytesRead) != 0)
      {
         /* incorrect CRC */
         pStation->checkPass |= WA_CRC;
         if (strictMode)
         {
            pStation->whenConnected.invalidFrames++;
            event = E_CRC;
            break;
         }
         /* Warnings are not counted as errors in station summary */
         pStation->whenConnected.warnings++;
         logCDevent(ET_W_CON, pStation, E_CRC, NULL, NO_INFO);
      }

#ifdef HAVE_AUTHENTICATION
      /* Verify frame signature */
      ii = (uint32_t)checkFrameAuthentication(pStation, pStation->keyCache,
                          pStation->pStationIP == NULL ?  NULL :
                          pStation->pStationIP->keyCacheSta,
                     strictMode, buffer, bytesRead, &trailer, &event);
      if (2 == ii)
      {
         /* A warning has occurred during frame signature check */
         pStation->whenConnected.warnings++;
         logCDevent(ET_W_CON, pStation, event, NULL, NO_INFO);
      }
      else if (1 == ii)
      {
         /* An error has occurred during frame signature check */
         if (E_OOPS == event || E_NO_MEM == event)
            pStation->whenConnected.systemFailures++;
         else
            pStation->whenConnected.authenticationFailures++;

         break;
      }
#endif /* HAVE_AUTHENTICATION */

      /* Handle frame according to frame ID */
      if (DATA == frameID)
      {
         int retVal;

         /* Update frame set and gap lists */
         retVal = updateFrameset(&(pStation->header), pStation);
         if (-1 == retVal)
         {
            pStation->whenConnected.systemFailures++;
            event = E_NO_MEM;
            break;
         }
         else if (1 == retVal)
         {
            /* According to the F&P, the only way to remove a gap is
               to close the connection */
            event = E_2MANY_GAPS;
            break;
         }

         if (mustBeORF)
         {
            /* First frame must be Option Request Frame,
               but in none strict mode we continue with a warning */
            if (strictMode)
            {
               pStation->whenConnected.invalidFrames++;
               pStation->checkPass |= ER_PROTOCOL;
               event = E_NO_ORF;
               break;
            }

            mustBeORF = 0;
            /* Warnings are not counted as errors in station summary */
            pStation->whenConnected.warnings++;
            logCDevent(ET_W_CON, pStation, E_NO_ORF, NULL, NO_INFO);
         }

         if (pStation->header.destination[0] != '0' ||
             pStation->header.destination[1] != (char)0)
         {
            /* Data frames must have no destination specified, but in none
               strict mode we process such frames with a warning */
            if (strictMode)
            {
               pStation->whenConnected.invalidFrames++;
               pStation->checkPass |= ER_PROTOCOL;
               event = E_DESTINATION;
               break;
            }
            /* Warnings are not counted as errors in station summary */
            pStation->whenConnected.warnings++;
            logCDevent(ET_W_CON, pStation, E_DESTINATION, NULL, NO_INFO);
         }

         /* Parse and check data frame payload */
         pStation->checkPass |=  parseAndCheckDataPayload(pStation->DFinfo,
                                             strictMode,
                                             buffer, pConfigData->maxDeviation,
                                             (uint32_t)bytesRead,
                                             pStation->endRoW.tv_sec,
                                             &(pStation->chCache));

         /* Reset frame not parsed */
         pStation->checkPass &= ~ER_NOT_PARSED;

         /* Check if an error or warning has occurred during parsing
            and check authentication if configured */
         if (checkParsingAndAuthentication(pStation, &event, &eventType,
                                           &clfType))
         {
            break;
         }
      }
      else if (OPTION_REQUEST == frameID)
      {
         mustBeORF = 0;
         /* Check Option Request payload */
         event = checkORFpayload(header, buffer, strictMode);

         if (NO_ERROR == event)
         {
            event = S_OPTION_REQ;
            eventType = ET_S_CON;
            clfType = GOOD_OPTION_REQ;
         }
         else
         {
            pStation->whenConnected.invalidFrames++;
            if (strictMode) break;
         }
      }
      else if (ACK_NACK == frameID)
      {
         event = S_ACKNAK;
         eventType = ET_S_CON;
         clfType = GOOD_ACKNAK;
      }
      else if (ALERT == frameID)
      {
         /* Alert frame terminates connection */
         clfType = GOOD_ALERT;
         eventType = ET_S_CON;
         event = S_CLOSED_BY_AF;
      }
      else
      {
         /* Frame type is not supported yet, raise warning */
         pStation->checkPass |= WA_UNSUPPORTED;
         eventType = ET_W_CON;
         event = E_UNSUPPORTED;
         /* Ugly, but it works */
         pStation->savedSystemErr = (uint32_t)frameID;
         addInfo = frameID;
      }

      /* frame received, log event, update statistic and save data,
         warning and errors are logged before */
      if (0 != (pStation->systemConf & MORE_VERBOSE_LOGGING) ||
           E_UNSUPPORTED == event)
      {
         logCDevent(eventType, pStation, event, NULL, addInfo);
      }
      updateStatistics(&(pStation->framesReceived), bytesRead, NO, clfType);

      /* Save data in specified files (bin, txt, clf, wfdisc, .. */
      if (saveData(pStation, buffer, bytesRead, 0, event, NO, clfType))
      {
         /* If data could not be saved, the connection will be terminated.
            If there is insufficient disc space, the receiver will
            be stopped too. This is done by saveData() */
         event = E_SAVE_DATA;
         pStation->whenConnected.systemFailures++;
         bytesRead = 0;
         break;
      }

      /* There could be an error in further processing, so data must
         be prevented to be saved two times */
      bytesRead = 0;

      /* Alert frame received? Nothing more to do here */
      if (S_CLOSED_BY_AF == event) break;

      /* Don't send an Option Response Frame in offline mode or the
         offset in the clf and txt files created in offline mode becomes
         wrong */
      if (GOOD_OPTION_REQ == clfType &&
         0 != (pStation->systemConf & ONLINE_MODE))
      {
         /* Receiver uses the received buffer to build response frame.
            There is currently only one option, so the payload is
            sent back without inspection  */
         setHeader(buffer, OPTION_RESPONSE, pStation->header.trailerOffset,
                   pConfigData->frameCreator, pStation->stationName,
                   (uint64_t) 0);

         /* Copy all needed values in the trailer. This includes also
            the signature if signing is on */
         ii = prepareTrailer(pStation, buffer, pStation->header.trailerOffset);
         if (0 == ii) break;

         /* statistic is updated and data is saved in sendFrame */
         pStation->checkPass = 0;
         event = sendFrame(pStation, buffer, (int)ii, S_SEND_OPTION_RES,
                           GOOD_OPTION_RES, 1);
         if (NO_ERROR != event)
         {
            break;
         }
         /* Set next AckNack time */
         (void) gettimeofday(&nowTime, NULL);
         nextAckNackTime = nowTime.tv_sec + (long)pConfigData->heartBeat;
      }

      /* Set shutdown event before checking if a user shutdown request
         is pending */
      event = S_SHUTDOWN;

   } while ((unsigned char) 0 == pStation->shutdownEnable);

   if (S_CLOSED_BY_AF != event)
   {
      if (S_SHUTDOWN != event)
      {
         /* Some error has occurred. Connection will be terminated
            log negative event, update statistic and save data */
         logCDevent(ET_E_CON, pStation, event, NULL, addInfo);

         if (bytesRead > 0)
         {
            updateStatistics(&(pStation->framesReceived), bytesRead,
                             NO, clfType);
            /* Note: the receiver log full and partial frames */
            /* There was an other error before, so a data save
               error is only logged here */
            if (saveData(pStation, buffer, bytesRead, 0, event, NO,
                         clfType))
            {
               pStation->whenConnected.systemFailures++;
               logCDevent(ET_E_CON, pStation, E_SAVE_DATA, NULL, NO_INFO);
            }
         }
      }
      else
      {
          logCDevent(ET_S_CON, pStation, event, NULL, NO_INFO);
      }

      if (E_CLOSED != event)
      {
         /* Reset frame not parsed */
         pStation->checkPass &= ~ER_NOT_PARSED;

         /* Maybe an AckNack frame could be sent here to inform the
            sender which frames are received */
         /* Send ALERT FRAME to terminate connection */
         setHeader(buffer, ALERT, (uint32_t) 40, pConfigData->frameCreator,
                   pStation->stationName, (uint64_t)0);
         memset(&buffer[SZ_HDR], 0, sizeof(uint32_t));
         /* Copy all needed values in the trailer. This includes also
            the signature if signing is on */
         ii = prepareTrailer(pStation, buffer, 40);
         if (0 != ii)
         {
            /* statistic is updated and data is saved in sendFrame */
            pStation->checkPass = 0;
            retValue = sendFrame(pStation, buffer, (int)ii, S_SEND_ALERT,
                                 GOOD_ALERT, 1);
            if (NO_ERROR != retValue)
            {
               logCDevent(ET_E_CON, pStation, retValue, NULL, NO_INFO);
            }
         }
      }
   }

   /* Delete all known frame sets */
   deleteFramesets(pStation);
   return(event);
}

/**
 * @short handles connection for a CD1 station
 *
 * @description This function handles the connection for a CD1 station.
 * @description It processes incoming data and send Alert Frames to
 * @description the connected station if an error occurs or the
 * @description receiver is shutting down
 *
 * @param pStation  pointer to station object
 *
 * @return event which terminates connection
 */

static uint16_t
handleCD1(station_t *pStation)
{
   rwFlags_t flags;
   const configParams_t *pConfigData;
   int sd = pStation->portInfo->socket;
   int frameLen,    /* len of received frame */
       bytesRead;
   char *buffer =  pStation->frameBuffer,
        clfType,
        logBuffer[LOGSIZE];
   uint16_t event,       /* ID of occurred event */
       eventType,
       retValue,
       mustBeDFF = 1;    /* First frame must be a DFF */
   uint32_t frameID,
            strictMode,
            addInfo,
            *pIntBuffer;   /* integer pointer pointing to receive buffer */
   CD1alert_t alertFrame;


   /* buffer should be aligned by malloc() */
   pIntBuffer = (uint32_t *)buffer;
   pConfigData = pStation->pConfigData;
   flags.shutdownFlag = &(pStation->shutdownEnable);

   do
   {
      addInfo = NO_INFO;
      clfType = FRAME_UNKNOWN;
      event = E_FRAME_TYPE;
      eventType = ET_W_CON;
      /* config data could change during runtime, so it is
         re-read on each frame */
      strictMode = pStation->systemConf & STRICT_MODE_ENABLE;
      pStation->authenSummary = AUTH_UNKNOWN; /* default value */
      pStation->checkPass = NO_ERROR;

      /* Read first LOOKUP_CD1 bytes of frame but leave them in input buffer */
      flags.sockFlags = MSG_PEEK;
      flags.timeout = pConfigData->timeOut;
      flags.resetDataTimeout = pConfigData->timeOut;
      flags.firstTime = &(pStation->startRoW);
      bytesRead = readWithTimeout(sd, buffer, LOOKUP_CD1, &flags);
      flags.firstTime = NULL;

      if (-1 == bytesRead)
      {
         /* Interrupt or system failure */
         pStation->whenConnected.systemFailures++;
         event = E_SYSTEM;
         addInfo = (uint32_t)errno;
         pStation->savedSystemErr = (uint32_t)errno;
         break;
      }
      else if (-2 == bytesRead)
      {
         /* No data is saved on fast user requested shutdown!! */
         event = S_SHUTDOWN;
         break;
      }

      if (LOOKUP_CD1 != bytesRead)
      {
         /* timeout or connection closed */
         if (0 == flags.timeout)
         {
            pStation->whenConnected.timeouts++;
            event = E_TIMEOUT;
         }
         else
         {
            pStation->whenConnected.closedConnections++;
            event = E_CLOSED;
         }
         break;
      }

      /* Read frame ID and frame length */
      pStation->header.frameType = frameID = ntohl(pIntBuffer[0]);
      pStation->header.trailerOffset = ntohl(pIntBuffer[1]);
      frameLen = (int)pStation->header.trailerOffset;

      IF_ST_DEBUG(9)
      {
         S_DEBUG_HDR;
         (void) fprintf(DEBUG_FP, "ID=%u, len=%u, 0x%x\n",
                       (unsigned int)frameID, (unsigned int)frameLen,
                       (unsigned int)pIntBuffer[2]);
         DEBUG_FOOTER;
      }

      /* Check if the frame size is valid. Check at least for a
         frame size of zero or an endless loop will occur */
      if (frameLen < MAX_FRAME_LEN && frameLen >= MIN_SZ_CD1_FRAME &&
          (frameLen % 4) == 0)
      {
         /* Set clf type. This is necessary to identify the frame in
            the clf file. Data is written to the configured output files,
            even if an error occurs.  */
         if (CD1_DATA == frameID)
         {
            /* It was a requirement that the parsing functions can be
             * used by other applications. This caused that parsing and
             * logging is separated in different functions. But than it's
             * hard for the logging functions to know what data could be
             * logged if an error has been occurred. So an indicator is
             * set that the current frame is not parsed complete here
             * and reset if the frame has parsed complete */
            pStation->checkPass |= ER_NOT_PARSED;
            clfType = BAD_DATA;
            /* reset nbr. of parsed channels */
            if (NULL != pStation->DFinfo)
            {
               pStation->DFinfo->nominalTime = 0.0;
               pStation->DFinfo->parsedChannels = 0;
            }
         }
         else if (CD1_ALERT == frameID)
         {
            clfType = BAD_ALERT;
         }
         else if (CD1_DATA_FORMAT == frameID)
         {
            /* Note: CD1 receiver checks the frame length and corrects it
               to 2820 if bigger and receiver in non strict mode. It also
               checks the maximum DF size and corrects it to 10000000 in non
               strict mode. And last, it checks the number of channels here */
            pStation->checkPass |= ER_NOT_PARSED;
            clfType = BAD_FORMAT;
            /* reset number of parsed data format channels */
            if (NULL != pStation->DFinfo)
            {
               pStation->DFinfo->parsedDFFchannels = 0;
            }
         }
      }

      /* Check if we have a frame type now */
      if (FRAME_UNKNOWN == clfType)
      {
         event = processInvalidFrame(pStation, event, &bytesRead, frameLen,
                                     SZ_CD1_ALERT);

         /* set addInfo if there was a system error */
         if (E_SYSTEM == event)
         {
            addInfo = (uint32_t)errno;
            pStation->savedSystemErr = (uint32_t)errno;
         }

         if (S_ACQUIRE != event)
         {
            /* re-acquire failed or strict mode */
            break;
         }

         /* re-acquire successful. Continue with next frame */
         continue;
      }

      /* Check if the buffer size is sufficient. Adjusting buffer size could
         be necessary after receiving a DF. maxDFsize in DFF is not used
         to make the receiver more robust */
      if (pStation->frameBufferSize < (uint32_t)frameLen)
      {
         if (-1 == adjustBufferSize(pStation, (uint32_t)frameLen, &event))
         {
            addInfo = (uint32_t)frameLen;
            /* Old buffer is not released when no bigger one is allocated */
            break; /* break do loop; SPLINT does not realize this */
         }
         /* buffer should be aligned by malloc() */
         buffer = pStation->frameBuffer;
         pIntBuffer = (uint32_t *)buffer;
      }

      /* Read the whole frame now */
      flags.timeout = pConfigData->timeOut;
      flags.sockFlags = 0;
      bytesRead = readWithTimeout(sd, buffer, frameLen, &flags);
      (void) gettimeofday(&(pStation->endRoW), NULL);

      if (-1 == bytesRead)
      {
         /* Interrupt or system failure */
         pStation->whenConnected.systemFailures++;
         event = E_SYSTEM;
         addInfo = (uint32_t)errno;
         pStation->savedSystemErr = (uint32_t)errno;
         bytesRead = LOOKUP_CD1;
         break;
      }
      else if (-2 == bytesRead)
      {
         /* No data is saved on fast user requested shutdown!! */
         event = S_SHUTDOWN;
         break;
      }

      if (frameLen != bytesRead)
      {
         /* timeout or connection closed */
         if ( 0 == flags.timeout)
         {
            pStation->whenConnected.timeouts++;
            event = E_TIMEOUT;
         }
         else
         {
            pStation->whenConnected.closedConnections++;
            event = E_CLOSED;
         }
         break;
      }

      /* Handle frame according to frame ID */
      if (CD1_DATA == frameID)   /* Data Frame received */
      {
         if (mustBeDFF)
         {
            /* First frame must be a DFF */
            pStation->whenConnected.invalidFrames++;
            event = E_NO_DFF;
            break;
         }

         /* Check and parse frame. No events are logged in this function */
         pStation->checkPass |= parseAndCheckDF(pStation->DFinfo,
                                                strictMode, buffer,
                                                pConfigData->maxDeviation,
                                                (pStation->endRoW).tv_sec,
                                                &(pStation->chCache));

         /* Reset frame not parsed */
         pStation->checkPass &= ~ER_NOT_PARSED;
         /* Check if an error or warning has occurred during parsing
            and check authentication if configured */
         if (checkParsingAndAuthentication(pStation, &event, &eventType,
                                           &clfType))
         {
            break;
         }
      }
      else if (CD1_DATA_FORMAT == frameID)   /* Data Format Frame received */
      {
         /* Process data format frame */
         mustBeDFF = 0;

         /* Check and parse frame. No events are logged in this function */
         pStation->checkPass |= parseAndCheckDFF(pStation->DFinfo,
                                                 strictMode, buffer,
                                                 &(pStation->chCache));

         /* Reset frame not parsed */
         pStation->checkPass &= ~ER_NOT_PARSED;
         /* There could be a lot of warnings and/or errors. These
            events were not logged in the parsing function, because
            parsing should be used by non cd-receiver applications too.
            So we need an addition step to log all invalid values.
            Warnings never stop parsing, so logging is done before
            check for errors */

         logDFFparsingEvents(pStation, strictMode);

         /* If there was a parsing error or a parsing warning and receiver
            is in strict mode, frame processing is aborted */
         if (HAS_DF_ERRORS(pStation->checkPass)
             ||
             /* Some warnings never abort frame processing */
             (0 != strictMode && HAS_DF_WARNINGS(pStation->checkPass)))
         {
            event = E_DATA_FORMAT;
            pStation->whenConnected.invalidFrames++;
            break;
         }

         /* Warnings are not counted as errors in station summary */
         if (HAS_ANY_DF_WARNING(pStation->checkPass))
         {
            pStation->whenConnected.warnings++;
         }

         /* A frame becomes bad if an error or warning has occurred */
         if (NO_ERROR == pStation->checkPass)
         {
            event = S_DATA_FORMAT;
            eventType = ET_S_CON;
            clfType = GOOD_FORMAT;
         }
         else
         {
            event = E_DATA_FORMAT;
            pStation->whenConnected.invalidFrames++;
         }

         /* Check if format buffer size is sufficient. Adjusting buffer size
            could be necessary after receiving a DFF */
         if (pStation->formatBufferSize < (uint32_t)frameLen)
         {
            /* Frame size is checked before */
            if (pStation->systemConf & VERBOSE_LOGGING)
            {
               (void) snprintf(logBuffer, LOGSIZE,
                               "Adjust format buffer size to [%u]",
                               (unsigned int)frameLen);
               logWrite(LOG_INFO, logBuffer);
            }

            /*@-sharedtrans*/
            pStation->formatBuffer = realloc(pStation->formatBuffer,
                                             (size_t)frameLen);
            /*@=sharedtrans*/

            if (NULL == pStation->formatBuffer)
            {
               pStation->whenConnected.systemFailures++;
               addInfo = (uint32_t)frameLen;
               event = E_NO_MEM3;
               break;
            }
            pStation->formatBufferSize = (uint32_t)frameLen;
         }
         memcpy(pStation->formatBuffer, buffer, (size_t)frameLen);
         pStation->lenOfDFF= (size_t)frameLen;
         pStation->clfTypeDFF = clfType;
         pStation->eventDFF = event;
      }
      else if (CD1_ALERT == frameID)   /* Alert Frame received */
      {
         /* Process alert frame */
         clfType = GOOD_ALERT;
         eventType = ET_S_CON;
         parseAF(buffer, &alertFrame);

         /* Reset frame not parsed */
         pStation->checkPass &= ~ER_NOT_PARSED;

         switch (alertFrame.alertType)
         {
            case NO_ACTION:
                 event = S_ALERT;
                 break;
            case TERMINATE:
                 event = S_CLOSED_BY_AF;
                 break;
            case FLUSH_FORMAT:
                 event = S_FLUSH_FORMAT;
                 mustBeDFF = 1;
                 break;
            default:
                clfType = BAD_ALERT;
                event = E_ALERT;
                eventType = ET_W_CON;
                addInfo = alertFrame.alertType;
                pStation->checkPass = WA_INVALID_ALERT_ID;
                break;
         }
         if (E_ALERT == event)
         {
            if (strictMode)
            {
               pStation->whenConnected.invalidFrames++;
               break;
            }
            /* Warnings are not counted as errors in station summary */
            pStation->whenConnected.warnings++;
         }
      }

      /* frame received, log event, update statistic and save data,
         warning and errors are logged before */
      if (pStation->systemConf & MORE_VERBOSE_LOGGING)
      {
         logCDevent(eventType, pStation, event, NULL, addInfo);
      }
      updateStatistics(&(pStation->framesReceived), bytesRead, NO, clfType);

      /* Save data in specified files (bin, txt, clf, wfdisc, .. */
      if (saveData(pStation, buffer, bytesRead, 0, event, NO, clfType))
      {
          /* If data could not be saved, the connection will be terminated.
             If there is insufficient disc space, the receiver will
             be stopped too. This is done by saveData() */
          event = E_SAVE_DATA;
          pStation->whenConnected.systemFailures++;
          bytesRead = 0;
          break;
      }

      /* There could be an error in further processing, so data must
         be prevented to be saved twice */
      bytesRead = 0;

      /* Alert frame received? Nothing more to do here */
      if (S_CLOSED_BY_AF == event) break;

      /* Set shutdown event before checking if a user shutdown request
         is pending */
      event = S_SHUTDOWN;

   } while ((unsigned char) 0 == pStation->shutdownEnable);

   if (S_CLOSED_BY_AF != event)
   {
      if (S_SHUTDOWN != event)
      {
         /* Some error has occurred. Connection will be terminated
            log negative event, update statistic and save data */
         logCDevent(ET_E_CON, pStation, event, NULL, addInfo);

         if (bytesRead > 0)
         {
            updateStatistics(&(pStation->framesReceived), bytesRead,
                                    NO, clfType);
            /* Note: the receiver log full and partial frames */
            if (NULL != buffer)
            {
               /* There was an other error before, so a data save
                  error is only logged here */
               if (saveData(pStation, buffer, bytesRead, 0, event, NO,
                            clfType))
               {
                  pStation->whenConnected.systemFailures++;
                  logCDevent(ET_E_CON, pStation, E_SAVE_DATA, NULL, NO_INFO);
               }
            }
         }
      }
      else
      {
          logCDevent(ET_S_CON, pStation, event, NULL, NO_INFO);
      }

      if (E_CLOSED != event)
      {
         /* Reset frame not parsed */
         pStation->checkPass &= ~ER_NOT_PARSED;
         /* Send ALERT FRAME to terminate connection */
         alertFrame.frameType = htonl(4);
         alertFrame.frameLength = htonl(12);
         alertFrame.alertType = htonl(TERMINATE_REC);
         /* statistic is updated and data is saved in sendFrame */
         pStation->checkPass = 0;
         retValue = sendFrame(pStation, (char *)&alertFrame, 12, S_SEND_ALERT,
                              GOOD_ALERT, 1);
         if (NO_ERROR != retValue)
         {
            logCDevent(ET_E_CON, pStation, retValue, NULL, NO_INFO);
         }
      }
   }

   return(event);
}

/*
 * @short wakeup listener
 *
 * @description This function wakes up the listener while waiting for
 * @description a handling thread to come up
 *
 * @return None
 */

static void
wakeupListener(station_t *pStation, const unsigned char state)
{
   /* pstation->pConfigData is "const configParams_t *", but this
      is the only exception where config data is modified from
      a station thread */
   configParams_t *pConfigData = (configParams_t *) pStation->pConfigData;
   receiverParams_t *pReceiverParams;

   pReceiverParams = (receiverParams_t *)pConfigData->configExtension;

   if (0 != pthread_mutex_lock(&(pReceiverParams->readyMutex)))
   {
      shutdownApplication("Error while wakeup listening thread",
                          "pthread_mutex_lock failed");
      pthread_exit(NULL);
   }

   pStation->state = state; /* STATION_WAITING or STATION_UNCONNECTED */
   ST_DEBUG2(5,"send cond_broadcast [%d]", (int) pStation->state);
   if (0 != pthread_cond_broadcast(&(pReceiverParams->readyCondition)))
   {
      shutdownApplication("Error while wakeup listening thread",
                          "pthread_cond_broadcast failed");
   }

   if (0 != pthread_mutex_unlock(&(pReceiverParams->readyMutex)))
   {
      shutdownApplication("Error while wakeup listening thread",
                          "pthread_mutex_unlock failed");
      pthread_exit(NULL);
   }
}

/**
 * @short handling thread - handles connection for a specific station
 *
 * @description This function handles the connection for a specific station.
 * @description It waits for a connection request (IP, not CD) on
 * @description the redirected port, creates all output files, wakes up
 * @description the listener and calls protocol specific handling functions.
 *
 * @param arg  void pointer to station object
 *
 * @return None
 */

/*@null@*/ void *
portHandle(void *arg)
{
   station_t *pStation;
   const configParams_t *pConfigData;
   int32_t timeOut;
   int newSock;
   struct sockaddr_in from;
   uint16_t event;

   /* Retrieve data passed to thread */
   pStation = (station_t *) arg;
   ST_DEBUG(2, "+++ CD handling thread started");

   pConfigData = pStation->pConfigData;
   pConfigData->pHistory->nbrOfThreadsStarted++;

   /* reset duplication of DFF in new file(s) */
   pStation->lenOfDFF = 0;

#ifdef CREATE_NEW_FILE
   /* force syncFiles() to create new files(s) */
   pStation->files.savedDay = (time_t)0;
#endif

   /* Initialize all needed output files. This will
      open old file(s) if one exists for the current day or
      create new file(s). It's necessary to open the files
      *before* data is read, because the text file contains
      additional information */
   if (-1 == syncFiles(pStation, NO, REOPEN_OLD_FILES, 0))
   {
      pStation->whenConnected.systemFailures++;
      logCDevent(ET_E_CR, pStation, E_FILES, NULL, NO_INFO);
      wakeupListener(pStation, STATION_UNCONNECTED);
      cleanUp(pStation);
      pConfigData->pHistory->nbrOfThreadsStopped++;
      return(NULL);
   }

   /* Set timeout value */
   if (CD1 == pStation->type) { timeOut = pConfigData->timeOut; }
   else { timeOut = pConfigData->timeOut11; }

   /* Allocate buffer for DFF info if this is the first connection for this
      station. Buffer is aligned proper by malloc, calloc() sets allocated
      memory to zero so DFinfo must not be cleared to have default values */
   if (NULL == pStation->DFinfo)
   {
      pStation->DFinfo = calloc(1, sizeof(CDdata_t));

      if (NULL == pStation->DFinfo)
      {
         /* No memory available */
         /*@-compdef@*/
         logCDevent(ET_E_CON, pStation, E_NO_MEM2, NULL, sizeof(CDdata_t));
         /*@=compdef@*/
         pStation->whenConnected.systemFailures++;
         wakeupListener(pStation, STATION_UNCONNECTED);
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         return(NULL);
      }
   }

   /* Allocate buffer for DFF frame if this is the first connection for this
      station. Buffer is aligned proper by malloc */
   if (CD1 == pStation->type && NULL == pStation->formatBuffer)
   {
      /*@-mustfreeonly@*/
      pStation->formatBuffer = (char *) malloc(2820);
      /*@=mustfreeonly@*/
      if (NULL == pStation->formatBuffer)
      {
         /* No memory available */
         /*@-compdef@*/
         logCDevent(ET_E_CON, pStation, E_NO_MEM3, NULL, 2820);
         /*@=compdef@*/
         pStation->whenConnected.systemFailures++;
         wakeupListener(pStation, STATION_UNCONNECTED);
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         return(NULL);
      }
      pStation->formatBufferSize = 2820;
   }

   /* Wake up listener */
   wakeupListener(pStation, STATION_WAITING);

   ST_DEBUG(9, "wait for station to connect");
   /* Wait user defined time for station to connect on redirect port */
   if ((newSock = waitForConnect(pStation->portInfo->socket,
                                 &from, timeOut,
                                 &(pStation->shutdownEnable))) < 0)
   {
      switch (newSock)
      {
      case -2:
         /* user shutdown request is pending (-2) */
         logCDevent(ET_S_CON, pStation, S_SHUTDOWN, NULL, NO_INFO);
         break;
      case -3:
         /* Timeout */
         logCDevent(ET_E_CON, pStation, E_TIMEOUT, NULL, NO_INFO);
         pStation->whenConnected.timeouts++;
         break;
      default:
         /* Interrupt or system failure (-1) */
         pStation->savedSystemErr = (uint32_t)errno;
         logCDevent(ET_E_CON, pStation, E_SYSTEM, NULL, (uint32_t)errno);
         pStation->whenConnected.systemFailures++;
      }
      cleanUp(pStation);
      pConfigData->pHistory->nbrOfThreadsStopped++;
      return(NULL);
   }

   /* Connecting station has been connected to re-direction port.
      Close socket where the handling thread has listened for station
      to connect */
   (void) close(pStation->portInfo->socket);

   /* NAT IP is set wrong here!! */
   if (-1 == setPortInfo(pStation->portInfo, newSock, &from))
   {
      pStation->savedSystemErr = (uint32_t)errno;
      logCDevent(ET_E_CON, pStation, E_SYSTEM, NULL, (uint32_t)errno);
      pStation->whenConnected.systemFailures++;
      cleanUp(pStation);
      pConfigData->pHistory->nbrOfThreadsStopped++;
      return(NULL);
   }

   /* Check if station is allowed to connect */
   if ('\0' != pConfigData->authStationNameFile[0])
   {
      const stationIP_t *pStationIP;

      /* Check IP address. If IP address is found, the corresponding
         station name (or a wildcard) is returned */
      pStationIP = ipAddressIsKnown(pConfigData, from.sin_addr.s_addr);

      if (NULL == pStationIP)
      {
         /* unknown station IP address
            To prevent flooding the disc with invalid data in case of DOS
            attacks, no data is logged. If you want to avoid syslog
            entries in this case too, comment the next line out */
         logCDevent(ET_E_CON, pStation, E_IP_UNKNOWN, NULL, NO_INFO);
         pStation->whenConnected.invalidIPs++;
         /* give valid stations a chance to connect in case of DoS attacks */
         delayedClose(newSock);
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         return(NULL);
      }

      /* Check if the station is allowed to connect from this IP */
      pStationIP = stationNameMatchIP(pConfigData, pStationIP,
                                      pStation->stationName);
      if (NULL == pStationIP)
      {
         /* station not permitted */
         logCDevent(ET_E_CON, pStation, E_STATION_UNKNOWN, NULL, NO_INFO);
         pStation->whenConnected.invalidStations++;
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         return(NULL);
      }
   }

   pStation->state = STATION_CONNECTED;
   logCDevent(ET_S_CON, pStation, S_CONNECTED, NULL, NO_INFO);

   /* Write some additional info to txt and clf file */
   addInfo2Files(pStation, S_CONNECTED);

   /* Call the station type dependent handling functions */
   if (CD1 == pStation->type)
   {
      event = handleCD1(pStation);
   }
   else
   {
      event = handleCD11(pStation);
   }

   /* Write some additional info to txt and clf file */
   addInfo2Files(pStation, event);
   cleanUp(pStation);
   logCDevent(ET_S_CON, pStation, S_DISCONNECTED, NULL, NO_INFO);

   pConfigData->pHistory->nbrOfThreadsStopped++;
   ST_DEBUG(2, "+++ CD handling thread stopped");
   pthread_exit(NULL);
   return(NULL);  /* for SPLINT */
}
