
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: init_receiver.c,v $
 *
 * Revision 1.6  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.5  2013-04-06 gerald
 * Fix compiler warning
 *
 * Revision 1.4  2011-04-18 13:41:54  gerald
 * Change CD Receivers default start port to port+3
 *
 * Revision 1.3  2008-10-30 16:04:04  gerald
 * Fix bug which prevents reading the endPort config parameter when
 * the startPort config parameter was not set
 *
 * Revision 1.2  2007/06/14 08:36:02  gerald
 * Fix comment
 *
 * Revision 1.1  2006/05/11 08:51:52  gerald
 * Move reading of receiver specific config parameters from libCD
 * to receiver specific function
 *
 * Revision 1.2  2006/04/06 12:00:04  gerald
 *
 */

#include <config.h>

#ifdef S_SPLINT_S
#include "../../splint_hints.h"
#endif /* S_SPLINT_S */
#include <netinet/in.h>  /* Need this to parse code with splint */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

#include "inittables.h"
#include "init_receiver.h"
#include "cdrecv.h"
#include "libpar.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif


/*@-fullinitblock@*/
static receiverParams_t receiverParams =
{
   /* lowest data port .............. */ 0,
   /* highest data port ............. */ 0,
   /* ----- rest is not read from config file, compiler sets it to zero --- */
};

/*@=fullinitblock@*/

static nbrParams_t nbrReceiverParams[] =
{
  /* ----------- 16 bit unsigned ---------------------- */
  {"startPort", &(receiverParams.startPort), INT16U},
  {"finishPort", &(receiverParams.endPort), INT16U}
  /* sequence is irrelevant, a parameter should appear only once */
};

/**
 * @short Read receiver specific configuration parameters
 *
 * @description This function reads the receiver specific parameters from
 * @description the specified configuration file and stores it in the
 * @description configParams extension object.
 * @description Values are not checked in this function
 *
 * @param appName      name of the application
 * @param filename     name of the configuration file
 * @param pConfigData  configuration structure
 *
 * @return nothing
 */

void
readReceiverConfigParams(const char *appName, const char *fileName,
                         configParams_t *pConfigData)
{
   const char *tmpArgv[2];
   char configFile[LIBPAR_BUFSIZ];
   uint16_t tmpVal,
            ii;
   /*@shared@*/ portMap_t *portInfoArray;


   /* Build argv vector for setpar() */
   (void)snprintf(configFile, sizeof(configFile), "par=%s", fileName);
   tmpArgv[0] = appName;
   tmpArgv[1] = configFile;

   /* Read configuration file with (standard) libpar function to
      support reading of environment variables */
   if (setpar (2, tmpArgv) != 0)
   {
      stopProgram(appName, "can't read configuration file", fileName);
      return;
   }

   /* Process configuration file */
   (void) processNumberParams(sizeof(nbrReceiverParams), nbrReceiverParams);
   endpar();

   /* Add Sender extension to config structure */
   /*@-mustfreeonly@*/
   pConfigData->configExtension = &receiverParams;
   /*@=mustfreeonly@*/
   DEBUG(5, "+++ Receiver specific parameters read");

   /* Do some configuration checks
      If start_port or finish_port are not set, use build in defaults */
   if (0 == receiverParams.startPort)
   {
      receiverParams.startPort = pConfigData->portNumber + 3;
   }
   if (0 == receiverParams.endPort)
   {
      receiverParams.endPort = pConfigData->portNumber + 99;
   }


   /* Initialize port map array */
   if (receiverParams.startPort > receiverParams.endPort)
   {
      tmpVal = receiverParams.endPort;
      receiverParams.endPort = receiverParams.startPort;
      receiverParams.startPort = tmpVal;
   }

   tmpVal = receiverParams.endPort - receiverParams.startPort + 1;

   /* Allocate port info table */
   if (NULL == (portInfoArray = (portMap_t *)
                calloc(tmpVal, sizeof (portMap_t))))
   {
      stopProgram(appName, "can't allocate memory for port map table",
                  NULL);
      return;
   }
   receiverParams.portInfoArray = portInfoArray;
   receiverParams.nbrOfPorts = tmpVal;

   /* Set initial values for port info table */
   for (ii = receiverParams.startPort; ii <= receiverParams.endPort; ii++)
   {
       portInfoArray->localPort = ii;
       portInfoArray->socket = -1;
       portInfoArray++;
   }

   /* initializes all needed mutex lock variables */
   if (0 != pthread_mutex_init(&(receiverParams.readyMutex), NULL))
   {
      stopProgram(appName, "can't initialize readyMutex", NULL);
   }

   if (0 !=  pthread_cond_init(&(receiverParams.readyCondition), NULL))
   {
      stopProgram(appName, "can't initialize readyCond", NULL);
   }
}

/**
 * @short dump receiver specific configuration parameter
 *
 * @description This function dumps the receiver specific configuration
 * @description parameter to syslog
 *
 * @return None
 */

void
dumpReceiverConfigurationParameter(/*@unused@*/configParams_t *pConfigData)
{
   (void) dumpNumberParams(sizeof(nbrReceiverParams), nbrReceiverParams);
}
