
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: listen.c,v $
 *
 * Revision 1.84  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.83  2012-06-27 gerald
 * Fix function comment
 *
 * Revision 1.82  2012-04-10 gerald
 * Update execute command call parameters. Required
 * for the theoretical fix for ISI-18
 *
 * Revision 1.81  2011-03-02 15:22:43  gerald
 * Add support for EC signature size
 *
 * Revision 1.80  2009-03-11 10:11:55  gerald
 * Fix splint warnings which occurred after upgrading
 * to new splint version
 *
 * Revision 1.79  2007/06/14 08:37:05  gerald
 * Rewrote security level code. Move functions required
 * by the command thread to libCD
 *
 * Revision 1.78  2007/05/31 14:41:08  gerald
 * Remember the key ID to for the security level check
 *
 * Revision 1.77  2007/01/24 09:11:40  gerald
 * Add computing and logging of send time span (rename
 * startRead to startRoW, endRead to endRoW, drop startWrite)
 *
 * Revision 1.76  2006/11/02 13:18:03  gerald
 * Raise a warning when a command request frame is received and
 * compiled without authentication
 *
 * Revision 1.75  2006/06/20 09:19:05  gerald
 * Don't shutdown if ioctl() fails. Treat this event as a
 * bad connection request instead. Improve syslog messages
 * to get more diagnosis info
 *
 * Revision 1.74  2006/05/24 13:11:56  gerald
 * Add check if frame has been signed by a forward station
 *
 * Revision 1.73  2006/05/11 07:48:38  gerald
 * Changes treatment of key caches.  A key cache does not longer belong
 * to one station. Instead a list of key caches is managed and each key
 * cache is globally accessable. A reference to a key cache is now assigned
 * to a station structure when the station connects for the first time
 *
 * Revision 1.72  2006/04/26 08:28:06  gerald
 * Move receiver specific config and station structure members
 * into config extension respectively station extension structure.
 * Move reading of receiver specific config parameters from libCD
 * to receiver specific function. Update comments
 *
 * Revision 1.71  2006/04/18 09:19:39  gerald
 * Rename defines for event types to avoid inadvertently use of
 * E_INTERNAL and fix typos
 *
 * Revision 1.70  2006/04/03 11:27:26  gerald
 * Debug messages improved
 *
 * Revision 1.69  2006/02/24 09:20:13  gerald
 * Add check for invalid station name to processing of CD1.1
 * connection request frames
 *
 * Revision 1.68  2005/10/14 10:57:45  gerald
 * Use station name from connection request payload instead of frame
 * creator name to check if CD-1.1 station is allowed to connect
 *
 * Revision 1.67  2005/09/07 11:01:52  gerald
 * Always behave like frame authentication switched on when receiving a
 * command request by the receiver, regardless if frame authentication
 * is switched on or not
 *
 * Revision 1.66  2005/07/07 08:48:21  gerald
 * Refuse connection request silently when in stopped state
 *
 * Revision 1.65  2005/05/17 15:19:13  gerald
 * Make use of new error/warning statistics structure
 *
 * Revision 1.64  2005/05/11 09:31:25  gerald
 * Call updateStatistics() instead of updateStationStatistics() and
 * pass new statistics structure instead of whole station structure
 *
 * Revision 1.63  2005/05/04 08:22:46  gerald
 * Drop redirectPortInfo structure
 *
 * Revision 1.62  2005/04/29 10:43:07  gerald
 * Reset frame parsing code before an (unparsed) frame is sent
 *
 * Revision 1.61  2005/02/17 15:17:24  gerald
 * Remove E_SIZE event. E_FRAMESIZE is the same event
 *
 * Revision 1.60  2004/09/23 11:33:53  gerald
 * Do not reset CD1 frame number when a Command Request is
 * received. Check signature of Command Request frames regardless
 * if frame signature verification is switched on or not
 *
 * Revision 1.59  2004/09/16 15:00:06  gerald
 * Fix check for allowed control station. Allow more than
 * one station per IP address when filtering IP and station names
 *
 * Revision 1.58  2004/09/10 10:35:07  gerald
 * Move ipAddressIsKnown(), delayedClose() and getListenerObject()
 * to CD library. Write debug messages to stderr instead of stdout,
 * Check if command request comes from an authorized station,
 * Use station objects frame buffer insted of own buffer
 *
 * Revision 1.57  2004/09/06 09:32:18  gerald
 * Check station specific system configuration parameter when checking
 * if authentication is switched on instead of system wide parameter
 *
 * Revision 1.56  2004/08/23 09:33:16  gerald
 * Set frame type and size/auth offset for integrity check
 *
 * Revision 1.55  2004/05/28 17:08:54  gerald
 * Fix missing IP and name check when no valid line is found
 * in the station file
 *
 * Revision 1.54  2004/05/14 16:33:37  gerald
 * Move sendFrame() and cleanUp() function from receiver to libCD
 *
 * Revision 1.53  2004/05/07 15:24:49  gerald
 * Add config parameter "controlStation". Only the control station
 * is authorized to send command frames to the receiver
 *
 * Revision 1.52  2004/04/05 13:55:46  gerald
 * Rename function setCRSpayload() to setCRpayload()
 *
 * Revision 1.51  2004/03/09 13:57:59  gerald
 * Rename all variables which have tbs as part of the name and
 * rename WITHOUT_ defines
 *
 * Revision 1.50  2004/02/03 09:07:05  gerald
 * Don't shutdown receiver when waitForConnect() fails and errno
 * is set to ECONNABORTED. This is no local system error.
 *
 * Revision 1.49  2004/01/15 15:24:19  gerald
 * Change behaviour when a handling thread initializes too slow. Now
 * the shutdown flag is set for such a thread and the shutdown is done
 * by the thread itself instead of cancel the handling thread.
 *
 * Revision 1.48  2004/01/14 16:45:58  gerald
 * Fix compilation error when compiled without LDAP access or
 * without authentication
 *
 * Revision 1.47  2004/01/12 14:07:53  gerald
 * Make IP address which is sent back in Port Assignment or Connection
 * Response Frame configurable. This is needed when Network Address
 * Translation is used in the in-route
 *
 * Use common function for frame authentication checks. That changes the
 * behaviour of the listener thread. Frames are not longer received with
 * an error/warning if no keys are available and frame signature check is
 * switched off (_U_* log files)
 *
 * Fix wrong frame destination name in Connection Response Frames
 *
 * Revision 1.46  2003/12/16 14:41:24  gerald
 * Fix some splint warnings
 *
 * Revision 1.45  2003/12/11 16:23:56  gerald
 * Reset connection specific shutdown flag before handling
 * thread terminates
 *
 * Revision 1.44  2003/12/05 08:57:23  gerald
 * Process Command Request frames only when compiled with authentication
 *
 * Revision 1.43  2003/12/03 14:49:25  gerald
 * Move setPortInfo() to libCD. Add station specific system
 * configuration flag. Rename unconHistory to listenerObj
 *
 * Revision 1.42  2003/11/27 11:22:40  gerald
 * Add connection ID field to station structure
 *
 * Revision 1.41  2003/11/24 10:56:41  gerald
 * Replace global debug macro with station specific debug macro to
 * support station specific debug levels
 *
 * Revision 1.40  2003/11/20 12:36:59  gerald
 * Move utility function shutdownReceiver() from receiver to libCD and
 * rename it to shutdownApplication(). Fix splint warnings. Move
 * getStation() and getStationState to libCD
 *
 * Revision 1.39  2003/11/14 13:43:39  gerald
 * Fix splint warning
 *
 * Revision 1.38  2003/10/30 14:48:37  gerald
 * Add processing of Command Request frames
 *
 * Revision 1.37  2003/10/17 14:01:17  gerald
 * Check for pending fast user shutdown request after a call to
 * writeWithTimeout()
 *
 * Revision 1.36  2003/10/02 15:20:11  gerald
 * Combine station and IP file to one file (station file)
 *
 * Revision 1.35  2003/08/27 14:45:08  gerald
 * Remove old comment
 *
 * Revision 1.34  2003/04/17 14:46:28  gerald
 * Add attempt to load certificates from LDAP first. Fix wrong
 * _U_* filename when well known port higher than 9999
 *
 * Revision 1.33  2003/03/27 14:03:14  gerald
 * Fix more splint warnings
 *
 * Revision 1.32  2003/03/24 16:12:32  gerald
 * Fix splint errors
 *
 * Revision 1.31  2003/03/18 13:29:20  gerald
 * Log more detailed frame/channel signature check result
 *
 * Revision 1.30  2003/03/17 14:28:13  gerald
 * Add key and site/channel specific NONE directives. Allow NONE
 * directives for unsigned channels.
 *
 * Revision 1.29  2003/03/12 14:21:55  gerald
 * Add channel authentication summary. Remove trailing blanks.
 *
 * Revision 1.28  2003/03/07 14:58:17  gerald
 * Make frame authentication state consistent (to handling.c)
 *
 * Revision 1.27  2003/03/06 13:22:17  gerald
 * Include John's comments.  Add authentication states "skipped"
 * and "switched off"
 *
 * Revision 1.26  2003/02/17 16:03:01  ctbto
 * Reduce messages to syslog if verboseEnabel is set to NO
 * Use macro to check if keys should be loaded
 *
 * Revision 1.25  2003/02/12 09:11:00  ctbto
 * Fix warnings that occur when compiling with  -Wall
 *
 * Revision 1.24  2003/02/07 15:55:11  ctbto
 * Add frame/channel parsing code for received frames
 * Add signature verification status for all received CD1.1 frames
 * Fix checks for "bad" frames. A frame becomes invalid if at least
 * a warning occurs during parsing or authentication check
 *
 * Revision 1.23  2003/01/31 13:14:02  ctbto
 * Spilt processConnectionRequests() into two functions (use part of the old
 * function for starting a handling thread in offline mode)
 *
 * Revision 1.22  2003/01/13 15:28:41  ctbto
 * Rework getConnectionRequest (logiScope)
 *
 * Add getListenerObject function
 *
 * Fix bug that nbrOfThreads stopped is incorrect when Listener
 * thread is canceled by the main thread
 *
 * Fix bug that thread ID of new handling thread is set to late (this
 * could cause a thread which not releases its resources or which is
 * not joined when a 'out of disc space' condition occurs). This bug
 * has been found by inspection of the source code!
 *
 * Fix odd behavior that the listener thread starts a handling thread when
 * a 'out of disc space' condition occurs during connection request
 *
 * Revision 1.21  2002/12/16 13:32:13  ctbto
 * Check frame signature of CD1.1 connection request frames and sign
 * outgoing connection response frames
 *
 * Revision 1.20  2002/12/09 13:16:38  ctbto
 * Continue work on authentication (load local PEM keys, check CD1 data
 * signatures, log authentication information)
 *
 * Revision 1.19  2002/11/29 10:26:40  ctbto
 * Continue work on authentication
 *
 * Revision 1.18  2002/11/11 14:54:07  ctbto
 * Fix bug which causes core dump when receiving signed connection request
 * frame. Fix bug which causes wrong receiver keyID is sent back to sender.
 *
 * Revision 1.17  2002/10/30 09:47:17  ctbto
 * Fix bug that sent frames have wrong sending time
 *
 * Revision 1.16  2002/10/28 15:44:24  ctbto
 * Add check for 'could not save data' condition
 *
 * Revision 1.15  2002/10/15 14:06:28  ctbto
 * Split and move functions (include John's comments)
 *
 * Revision 1.14  2002/10/15 07:40:04  ctbto
 * Fix possible bug in listen.c while searching for a station object.
 * Use creator name for destination field when sending a CD-1.1 frame
 * instead of station name
 *
 * Revision 1.13  2002/10/10 09:54:38  ctbto
 * Include John's comments
 *
 * Revision 1.12  2002/09/24 14:37:43  ctbto
 * Count connection request warnings. Log valid incoming frames only if
 * MORE verbose mode is set. Do some code cleanup.
 *
 * Revision 1.11  2002/09/20 16:50:22  ctbto
 * Save outgoing data to log file
 *
 * Revision 1.10  2002/09/20 15:58:23  ctbto
 * Add responder type and creator name to config parameter. Add version to
 * startup messages
 *
 * Revision 1.9  2002/09/11 09:57:19  ctbto
 * Improve shutdown code. Fix bug when handling thread could not
 * initialize (e.g. file system full) number of stopped threads is
 * decremented twice ("connections limit reached").
 *
 * Revision 1.8  2002/09/05 16:20:52  ctbto
 * Include John's comments (rename functions, move functions to new files,
 * code changes, etc.)
 *
 * Revision 1.7  2002/08/26 12:22:46  ctbto
 * Add check if station is allowed to connect, add more comments, add
 * warning if a port which is marked as free can't be opened, change
 * name of unconnected files to _U_<listen_port>
 *
 * Revision 1.6  2002/07/23 08:51:14  ctbto
 * Add new parameters to read/writeWithTimeout
 *
 * Revision 1.5  2002/07/03 17:56:34  ctbto
 * Add different frame types to updateStatistic() calls. Add information
 * whether or not it is called from a connection request to saveData(). Add
 * check if authentication is switched on, frame is signed  and set
 * authentication state.
 *
 * Revision 1.4  2002/06/26 14:36:21  ctbto
 * Fix CRC check
 * Make setTrailer() more general useable and return frame size
 *
 * Revision 1.3  2002/06/21 12:15:32  ctbto
 * Add checks for max. number of connections, IPaddresses, and CRC.
 * Add new parameter to logEvent (additional information)
 *
 * Revision 1.2  2002/06/20 07:58:16  ctbto
 * Allow handling thread resources to be reclaimed. Change DEBUG output
 *
 * Revision 1.1.1.1  2002/06/18 07:28:05  ctbto
 * Import sources
 *
 */

#include <config.h>
#include <sys/ioctl.h>   /* Don't remove this include !!*/
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "cdrecv.h"
#include "command.h"
#include "events.h"
#include "logfiles.h"
#include "requests.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

#ifndef FIONREAD
#include <stropts.h>
#define FIONREAD I_NREAD  /* FIONREAD is used under linux */
#endif /* FIONREAD */


/**
 * @short find free port
 *
 * @description This function attempts to find an unused port. If one is
 * @description found, it tries to open a socket. If the open fails, the
 * @description next unused port is tried. If the port could be opened
 * @description the appropriate port map entry is assigned to the stations
 * @description redirect port info entry
 *
 * @param pConfigData      configuration information
 * @param pStation         station object
 *
 * @return 0 on success or 1 on failure
 */

static int
findPort (configParams_t *pConfigData, station_t *pStation)
{
   int stationSock;
   uint16_t ii;
   portMap_t *pPortInfo;
   in_addr_t inaddr;
   receiverParams_t *pReceiverParams;

   ST_DEBUG(9,"In findPortNum");

   /* Search for an unused port in port map */
   pReceiverParams = (receiverParams_t *)pConfigData->configExtension;
   pPortInfo = pReceiverParams->portInfoArray;

   for (ii = 0; ii < pReceiverParams->nbrOfPorts; ii++)
   {
      ST_DEBUG3(9, "port [%d], sd[%d]", (int)pPortInfo->localPort,
                (int)pPortInfo->socket);
      if (pPortInfo->socket == -1)
      {
         /* This should be a free port. Try to create a new socket */

         inaddr = inet_addr(pStation->portInfo->localIPaddr);
         ST_DEBUG2(9, "newaddr [%x]", (unsigned int)inaddr);
         /* inet_addr returns address in NBO!! */
         inaddr = ntohl(inaddr);
         stationSock = getListenSocket(inaddr, pPortInfo->localPort, 1);

         if (stationSock != -1)
         {
            /* Now the IP address of the incoming interface is copied
               the portmap structure */
            ST_DEBUG2(9, "port found[%d]", (int)pPortInfo->localPort);
            memcpy(pPortInfo->localIPaddr, pStation->portInfo->localIPaddr,
                   sizeof(pPortInfo->localIPaddr));
            pPortInfo->socket = stationSock;
            pStation->portInfo = pPortInfo;
            return(0);
         }
         logWarning("Can't open port marked as free", strerror(errno));
      }
      pPortInfo++;
   }
   /* The receiver could stat all ports here to make sure
      that no ports are marked as active which are not really active */

   return(1);
}

/**
 * @short Start handling thread
 *
 * @description This function listens on a predefined port for connection
 * @description requests, checks the requests and spawns a handling thread
 * @description to handle the connection if it is a valid request. Otherwise
 * @description a warning is generated and the new connection is closed.
 *
 * @param pConfigData  configuration data
 * @param pStation     station object
 *
 * @return 0 on success, 1 on recoverable error and -1 on non
 * @return recoverable error
 */

int
startHandlingThread(/*@dependent@*/configParams_t *pConfigData,
                    station_t *pStation)
{
   int rval;
   struct timespec when;
   struct timeval dateTime;
   receiverParams_t *pReceiverParams;

   /* Find available port for connection, open redirect socket */
   if (0 != findPort(pConfigData, pStation))
   {
      /* No port within range available */
      logCDevent (ET_E_CR, pStation, E_NO_PORT, NULL, NO_INFO);
      pStation->whenUnconnected.systemFailures++;
      pStation->lastUnsuccessfulCR = pStation->startRoW.tv_sec;
      pStation->state = STATION_UNCONNECTED;
      cleanUp(pStation);
      return(1);
   }

   /* Spawn handling thread to handle connection */
   if (0 != (rval = pthread_create(&(pStation->threadID), NULL,
                                   portHandle, pStation)))
   {
      /* At this place the old receiver would have stopped */
      logCDevent(ET_E_CR, pStation, E_THREAD, NULL, (uint32_t) rval);
      pStation->whenUnconnected.systemFailures++;
      pStation->lastUnsuccessfulCR = pStation->startRoW.tv_sec;
      cleanUp(pStation);
      return(1);
   }

   /* Wait until handling thread is ready. Thread has some second to
      signaling readiness. Otherwise it is assumed that there is a
      problem and the CRF is ignored */

   (void) gettimeofday(&dateTime, NULL);
   /*@-usedef@*/
   when.tv_sec = dateTime.tv_sec + MAX_HANDLING_STARTUP_TIME;
   when.tv_nsec = (long)dateTime.tv_usec;
   /*@=usedef@*/
   rval = 0;

   pReceiverParams = (receiverParams_t *)pConfigData->configExtension;
   /* Set a lock between this thread and the spawned station thread */
   if (0 != pthread_mutex_lock(&(pReceiverParams->readyMutex)))
   {
      shutdownApplication("Error while waiting for handling thread",
                          "pthread_mutex_lock failed");
      return(-1);
   }

   if (STATION_CONNECTING == pStation->state)
   {
      ST_DEBUG2(5,"wait on cond_timed [%d]", (int) pStation->state);
      rval = pthread_cond_timedwait(&(pReceiverParams->readyCondition),
                                      &(pReceiverParams->readyMutex), &when);
      ST_DEBUG2(5,"cond_wait waked up [%d]", rval);
   }

   if (0 != pthread_mutex_unlock(&(pReceiverParams->readyMutex)))
   {
      shutdownApplication("Error while waiting for handling thread",
                          "pthread_mutex_unlock failed");
      return(-1);
   }

   if (STATION_UNCONNECTED == pStation->state)
   {
      pStation->lastUnsuccessfulCR = pStation->startRoW.tv_sec;
      /* Handling thread has initialization error
       * event is logged in the handling thread */
      return(1);
   }

   if (ETIMEDOUT == rval)
   {
      /* Handling thread doesn't initialize fast enough */
      logCDevent(ET_E_CR, pStation, E_THREAD2SLOW, NULL, NO_INFO);
      pStation->whenUnconnected.systemFailures++;
      pStation->lastUnsuccessfulCR = pStation->startRoW.tv_sec;
      /* Set the shutdown flag to signal the handling thread to
         shut down. This is the safest way to handle this condition.
         Keep in mind that too slow threads detain the listener
         thread from processing requests */
      pStation->shutdownEnable = SD_FAST;
      return(1);
   }

   /* Handling thread has been started */
   return(0);
}

/**
 * @short Process Connection or Command Request
 *
 * @description This function processes a Connection or Command Request
 * @description If the request is a connection requests it spawns a handling
 * @description thread to handle the connection and sends a PAF or Connection
 * @description Response frame. Otherwise the command is executed and a
 * @description Command Response frame is sent.
 *
 * @param listenSock   Socket to listen on
 * @param pConfigData  configuration data
 * @param listenerObj  listener's pseudo station object
 *
 * @return None
 */

static void
processRequests(int listenSock, /*@dependent@*/configParams_t *pConfigData,
                station_t *listenerObj)
{
   rwFlags_t flags;
   int newSock,     /* Socket for new connection */
       rval,
       len;
   /*@dependent@*/
   station_t *pStation;      /* pointer to station object */
   uint16_t event;
   char clfType;
   /*@dependent@*/ char *buffer;
   uint32_t frameType;

   buffer = listenerObj->frameBuffer;
   flags.sockFlags = 0;
   flags.shutdownFlag = &(pConfigData->shutdownEnable);
   flags.firstTime = &(listenerObj->startRoW);
   flags.resetDataTimeout = -1;

   do
   {
      frameType = CONNECTION_REQUEST;

      /* Handle incoming data: Wait for new valid CR.
         All needed information is returned in a station object */
      if (-1 == (newSock = getNextValidRequest(listenSock, pConfigData,
                                               buffer, &pStation,
                                               listenerObj, NO)))
      {
         shutdownApplication("Error while waiting for connections",
                             "getNextValidRequest failed");
         break;
      }
      else if (-2 == newSock)
      {
         /* user shutdown request pending */
         ST_DEBUG(5, "Listener thread shutdown by user request");
         break;
      }

      /* Check for connection request */
      if (STATION_CONNECTING == pStation->state)
      {
         /* Connection Request, start handling thread */
         rval = startHandlingThread(pConfigData, pStation);

         if (1 == rval)
         {
            /* (Hopefully) temporary error condition */
            (void) close(newSock);
            continue;
         }
         else if (-1 == rval)
         {
            /* Non recoverable error */
            (void) close(newSock);
            break;
         }
         /* If redirect IP is provided overwrite local interface IP */
         if (pStation->pStationIP != NULL &&
             '\0' != pStation->pStationIP->redirectIP[0])
         {
            memcpy(pStation->portInfo->localIPaddr,
                   pStation->pStationIP->redirectIP, 16);
         }

         /* Handling thread is up now, could send response */
         ST_DEBUG3(5,"redirect to %.15s:%d",
                   pStation->portInfo->localIPaddr,
                   (int)pStation->portInfo->localPort);
      }
      else
      {
         frameType = COMMAND_REQUEST;
      }

      /* Build response */
      if (CD1 == pStation->type)
      {
         /* Set PAF frame */
         memset(buffer, 0, 32);
         (void) snprintf(buffer, LNG_CRF, "%.15s/%u",
                         pStation->portInfo->localIPaddr,
                         (unsigned int)pStation->portInfo->localPort);
         len = 32;
      }
      else
      {
         uint32_t trailerOffset;

         /* use the received buffer to build response frame */
         if (CONNECTION_REQUEST == frameType)
         {
            /* Set Header and payload for Connection Response */
            trailerOffset = 68;
            setHeader(buffer, CONNECTION_RESPONSE, (uint32_t) 68,
                      pConfigData->frameCreator, pStation->stationName,
                      (uint64_t) 0);
            setCRpayload(&buffer[36], pConfigData->frameCreator,
                         pConfigData->responderType,
                         pStation->portInfo->localIPaddr,
                         pStation->portInfo->localPort);
         }
         else
         {
            /* Execute command and
               set Header and payload for Command Response */
            trailerOffset = executeCommand(buffer, INIT_BUFFER_SIZE,
                                           listenerObj, pStation);
            setHeader(buffer, COMMAND_RESPONSE, trailerOffset,
                      pConfigData->frameCreator, pStation->stationName,
                      (uint64_t) 0);
         }

#define CLEANUP_STATION_IF_CONNECT_REQUEST() \
         if (CONNECTION_REQUEST == frameType) \
         { \
            /* Cancel waiting station handling thread */ \
            (void) pthread_cancel(pStation->threadID); \
            cleanUp(pStation); \
            pConfigData->pHistory->nbrOfThreadsStopped++; \
         }

#ifdef HAVE_AUTHENTICATION

         if (NULL == pConfigData->privateKey)
         {
            len = (int)setTrailer(buffer, trailerOffset, 0, NULL, 0);
         }
         else
         {
            char signature[SZ_SIGNATURE_MAX];
            size_t sigLen = sizeof(signature);

            /* Frame authentication is active */
            if (signData(pConfigData->privateKey, buffer, trailerOffset,
                         signature, &sigLen))
            {
               /* something unexpected has happen during signature
                  verification. Shutdown receiver?? */
               logError("Unexpected error while signing data",
                        "processRequests");

               pStation->whenUnconnected.systemFailures++;
               pStation->lastUnsuccessfulCR = pStation->startRoW.tv_sec;

               if (listenerObj->systemConf & STRICT_MODE_ENABLE)
               {
                  /* Do some clean up if connection request */
                  CLEANUP_STATION_IF_CONNECT_REQUEST();
                  (void) close(newSock);
                  continue;
               }
               /* Send unsigned frame */
               len = (int)setTrailer(buffer, trailerOffset, 0, NULL, 0);
            }
            else
            {
               len = (int)setTrailer(buffer, trailerOffset,
                                     pConfigData->privateKeyID, signature,
                                     sigLen);
            }
         }
#else

         len = (int)setTrailer(buffer, trailerOffset, 0, NULL, 0);

#endif /* HAVE_AUTHENTICATION */
      }

      /* Send response */
      listenerObj->checkPass = 0;
      flags.timeout = pConfigData->CRtimeOut;
      rval = writeWithTimeout(newSock, buffer, len, &flags);

      /* update end write time */
      (void) gettimeofday(&(listenerObj->endRoW), NULL);

      if (-1 == rval)
      {
         /* Interrupt or system failure */
         (void) close(newSock);
         shutdownApplication("Error sending response",
                             "writeWithTimeout failed");
         break;
      }
      else if (-2 == rval)
      {
         /* Fast user shutdown request; No data is saved */
         logCDevent(ET_W_CR, pStation, S_SHUTDOWN, listenerObj->portInfo,
                    NO_INFO);
         /* Do some clean up if connection request */
         CLEANUP_STATION_IF_CONNECT_REQUEST();
         (void) close(newSock);
         break;
      }

      if (rval != len)
      {
         /* timeout or connection closed */
         if ( 0 == flags.timeout)
         {
            pStation->whenUnconnected.systemFailures++;
            event = E_RESPONSE;
         }
         else
         {
            pStation->whenUnconnected.closedConnections++;
            event = E_CLOSED;
         }

         if (CD1 == pStation->type) clfType = BAD_PAF;
         else if (CONNECTION_REQUEST == frameType) clfType = BAD_CON_RES;
         else clfType = BAD_COMMAND_RES;

         logCDevent(ET_E_CR, pStation, event, listenerObj->portInfo, NO_INFO);
         pStation->lastUnsuccessfulCR = pStation->startRoW.tv_sec;
         updateStatistics(&(pStation->framesSent), len, YES, clfType);
         /* Data has to be saved in listeners log files */
         if (saveData(listenerObj, buffer, 0, len, event, YES, clfType))
         {
            /* There is nothing more to do. SaveData will shutdown the
               receiver if data could not be saved caused by a full disc
               or insufficient rights for creating needed directories */
            pStation->whenUnconnected.systemFailures++;
            logCDevent(ET_E_CR, pStation, E_SAVE_DATA, listenerObj->portInfo,
                       NO_INFO);
         }

         if (CD1 == pStation->type)
         {
            /* A Port Assignment frame has no station name information
               included, so the station name is added to the txt file
               manually */
            addName2TxtFile(listenerObj, pStation->stationName);
         }

         /* Do some clean up if connection request */
         CLEANUP_STATION_IF_CONNECT_REQUEST();
         (void) close(newSock);
         continue;
      }

      /* Set default for clf type */
      event = S_SEND_RESPONSE;

      if (CD1 == pStation->type) clfType = GOOD_PAF;
      else if (CONNECTION_REQUEST == frameType) clfType = GOOD_CON_RES;
      else
      {
         /* A command response is also good if command is not successful */
         event = S_COMMAND_RESPONSE;
         clfType = GOOD_COMMAND_RES;
      }

      if (listenerObj->systemConf & VERBOSE_LOGGING)
      {
         /* Log response */
         logCDevent(ET_S_CR, pStation, event, listenerObj->portInfo, NO_INFO);
      }

      updateStatistics(&(pStation->framesSent), len, YES, clfType);
      /* Data has to be saved in listeners log files */
      if (saveData(listenerObj, buffer, 0, len, event, YES, clfType))
      {
         /* Handling thread will raise the same error. SaveData will shutdown
            the receiver if data could not be saved caused by a full disc
            or insufficient rights for creating needed directories */
         pStation->whenUnconnected.systemFailures++;
         logCDevent(ET_E_CR, pStation, E_SAVE_DATA, listenerObj->portInfo,
                       NO_INFO);
      }

      if (CONNECTION_REQUEST == frameType)
      {
         if (CD1 == pStation->type)
         {
            /* A Port Assignment frame has no station name information
               included, so the station name is added to the txt file
               manually */
            addName2TxtFile(listenerObj, pStation->stationName);
         }

         if (listenerObj->systemConf & VERBOSE_LOGGING)
         {
            /* Log CR is accepted */
            logCDevent(ET_S_CR, pStation, S_CRF_ACCEPTED, listenerObj->portInfo,
                       NO_INFO);
         }
         pStation->whenUnconnected.connectsOK++;
         /* Redirecting make a lot of work,
            maybe it should be dropped in later versions  */
      }

#undef CLEANUP_STATION_IF_CONNECT_REQUEST
     (void) close(newSock);
   } while(1);
}

/**
 * @short Listening thread - start function
 *
 * @description This function is the starting point of the listener thread.
 * @description It initializes all the necessary variables and calls
 * @description the connection request handling function
 *
 * @param arg  void pointer to config parameters
 *
 * @return None
 */

/*@null@*/ void *
portListen(/*@dependent@*/void *arg)
{
   int listenSock;  /* Listener socket */
   uint16_t myPort;        /* port to listen */
   const char *myIPaddr;   /* address to listen */
   char comBuffer[LOGSIZE];  /* per Thread or on stack!! */
   /*@dependent@*/configParams_t *pConfigData = NULL;
   station_t *listenerObj;  /* pointer to pseudo station object */


   DEBUG(2, "+++ Listener thread started");

   /* Retrieve data passed to thread */
   pConfigData = (configParams_t *) arg;

   /* Listener uses the same signal mask as the main thread */

   /* Create and configure listener socket */
   listenSock = getListenSocket(INADDR_ANY, pConfigData->portNumber,
                                MAX_CONNECTIONS);
   if (-1 == listenSock)
   {
      shutdownApplication("Error configuring Listening thread",
                          "getListenSocket failed");
      pthread_exit(NULL);
      return(NULL); /*for splint */
   }

   if (-1 == getAddressInfo(listenSock, &myIPaddr, &myPort))
   {
      (void) close(listenSock);
      shutdownApplication("Error configuring Listening thread",
                          "getAddressInfo failed");
      pthread_exit(NULL);
      return(NULL); /*for splint */
   }

   /* All data in the connection request handling process is handled by
      a pseudo station object named "_U_<listen_port>" first. As soon as
      a station name is available, the real station object is used to do
      some checks and update statistics. But data sent and received during
      connection handling is *always* logged in files related to the
      pseudo station object. This is necessary to handle data from
      a station which is already connected */
   listenerObj = getListenerObject(pConfigData->portNumber);
   if (NULL == listenerObj )
   {
      (void) close(listenSock);
      shutdownApplication("Error configuring Listening thread",
                          "no memory for listeners station object");
      pthread_exit(NULL);
      return(NULL); /*for splint */
   }

   (void) snprintf(comBuffer, LOGSIZE, "Listener started, listen at %.15s:%hu",
                   myIPaddr, (unsigned short)myPort);
   logWrite(LOG_NOTICE, comBuffer);
   (void) snprintf(pConfigData->pHistory->listenOn, 21, "%.15s:%hu", myIPaddr,
                   (unsigned short)myPort);

   pConfigData->pHistory->nbrOfThreadsStarted++;

   /* Initialization finished, start to wait for connection requests */
   processRequests(listenSock, pConfigData, listenerObj);

   DEBUG(2, "+++ Listener thread exiting");
   /* nbrOfThreadsStopped is set in the main thread after pthread_join() */
   return(NULL); /* Only for splint */
}

