
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * Revision 1.61  2015-11-27 gerald
 * Follow change in CDlibVersion()
 *
 * Revision 1.60  2013-06-13 gerald
 * Add CD Receiver and libCD version to usage string. Add
 * libCD version when logging start message to syslog.
 * Add closing libCD. Write usage message to stderr
 *
 * Revision 1.59  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.58  2009-05-19 13:52:56  gerald
 * Change syslog IDs as requested by Henrik
 *
 * Revision 1.57  2008-09-16 18:04:12  gerald
 * Ignore suspend signal (CRTL-Z) and log a warning to
 * stdout when CRTL-Z is pressed
 *
 * Revision 1.56  2006/04/26 08:28:06  gerald
 * Move receiver specific config and station structure members
 * into config extension respectively station extension structure.
 * Move reading of receiver specific config parameters from libCD
 * to receiver specific function. Update comments
 *
 * Revision 1.55  2006/02/14 13:01:12  gerald
 * readConfigurationFile() does not longer need to know which
 * application has called it
 *
 * Revision 1.54  2005/09/21 09:22:11  gerald
 * Make use of CDtoolsVersion()
 *
 * Revision 1.53  2005/09/07 10:51:08  gerald
 * Move monitor.c and keychecker.c from cdrecv sources to libCD
 *
 * Revision 1.52  2005/08/11 09:21:12  gerald
 * Fix warning when compiled without authentication
 *
 * Revision 1.51  2004/11/12 14:21:45  gerald
 * Add app ID to initConfigParams(). Drop log messages to stdout
 *
 * Revision 1.50  2004/09/21 12:34:51  gerald
 * Change default value of some config params
 *
 * Revision 1.49  2004/07/06 13:18:16  gerald
 * Move joinHandlingThreads() to libCD
 *
 * Revision 1.48  2004/04/27 15:27:43  gerald
 * Fix missing "ctbt.acquisition..." string in syslog messages when
 * an initialization error is logged. The port number in the syslog
 * ID string is added after the configuration file is read
 *
 * Revision 1.47  2004/03/25 16:10:36  gerald
 * Fix wrong check for HAVE_AUTHENTICATION
 *
 * Revision 1.46  2004/03/22 12:12:12  gerald
 * Drop config parameter createBinInOffline and introduce more
 * general createBinFile (ON when receiver in online mode, OFF
 * when receiver in offline mode or sender)
 *
 * Revision 1.45  2004/03/08 15:35:42  gerald
 * Move init.[ch] logfiles.[ch] and buildWfdiscRecords() to libCD
 *
 * Revision 1.44  2004/02/13 14:08:38  gerald
 * Call processConfiguration() before writting startup message
 *
 * Revision 1.43  2003/12/03 14:39:03  gerald
 * Add station specific shutdown condition (setShutdownFlag()) and rename
 * receiverStarted to appStarted
 *
 * Revision 1.42  2003/11/27 11:24:07  gerald
 * Replace station name with connection ID
 *
 * Revision 1.41  2003/11/20 12:39:31  gerald
 * Move functions reading the configuration file and the configuration
 * structure itself from CDreceiver to libCD. Move utility function
 * shutdownReceiver() from CDreceiver to libCD and rename it to
 * shutdownApplication(). Fix splint warnings
 *
 * Revision 1.40  2003/10/15 08:58:41  gerald
 * Replace stopReceiver() with more generic stopProgram() function
 * Move stopProgram() and common defines/code to libcd (cd_common.[ch])
 *
 * Revision 1.39  2003/10/02 15:20:11  gerald
 * Combine station and IP file to one file (station file)
 *
 * Revision 1.38  2003/09/30 16:08:20  gerald
 * Re-name configuration parameter indexFilePath to cdrecvIndexFilePath
 * and maxPosDeviation to maxFutureTime
 *
 * Revision 1.37  2003/09/24 14:00:14  gerald
 * Fix wrong directory permissions
 *
 * Revision 1.36  2003/08/27 14:22:11  gerald
 * Renamed the configuration parameter realTimeWfdiscCreate to be
 * enableWfdiscFile. Fixed some typos of the word "world"
 *
 * Revision 1.35  2003/05/21 08:14:39  gerald
 * Fix Bug: Certification thread was started and didn't wait in
 * nanosleep() if no cerlCheckTime config parameter was provided
 * in the config file
 *
 * Revision 1.34  2003/05/14 15:30:09  gerald
 * Add certificate thread
 *
 * Revision 1.33  2003/04/17 13:58:03  gerald
 * Update LDAP root default and fix compilation problem when
 * compiling without LDAP
 *
 * Revision 1.32  2003/04/14 14:24:19  gerald
 * Fix spelling error
 *
 * Revision 1.31  2003/03/24 15:22:51  gerald
 * Move processXXparames and dumpXXparams functions to libcd
 *
 * Revision 1.30  2003/03/12 14:17:15  gerald
 * Remove trailing blanks.
 * Make creation of index file configurable
 *
 * Revision 1.29  2003/03/07 14:57:19  gerald
 * Add security requirements
 *
 * Revision 1.28  2003/02/25 09:39:49  ctbto
 * Create log dirs many levels deep. Make creation of binary files
 * configureable in offline mode. Rework readConfigurationFile() and
 * dumpConfiguration() and make these functions generic useable.
 *
 * Revision 1.27  2003/02/17 16:04:03  ctbto
 * Decouple verboseEnable config parameter and text file detail. Text
 * file detail is controlled now by config parameter enableTextFile.
 *
 * Revision 1.26  2003/02/12 09:10:33  ctbto
 * Fix warnings that occur when compiling with  -Wall
 *
 * Revision 1.25  2003/02/07 15:13:16  ctbto
 * Clean up. Move functions to other files
 *
 * Revision 1.24  2003/01/31 13:07:21  ctbto
 * Add ARM key cache directive. Fix bug that current directory is
 * changed after pre-loading keys. Add offline mode
 *
 * Revision 1.23  2003/01/13 14:54:27  ctbto
 * Fix bug that debugLevel is not loaded as first configuration parameter
 * to have debug feature available in readConfigurationFile()
 *
 * Rename some configuration parameter to have a common naming schema
 *
 * Fix bug that nbrOfThreads stopped is incorrect when Listener
 * thread is canceled by the main thread
 *
 * Revision 1.22  2002/12/16 13:29:47  ctbto
 * Add private key ID to configuration parameters
 *
 * Revision 1.21  2002/12/09 13:16:38  ctbto
 * Continue work on authentication (load local PEM keys, check CD1 data
 * signatures, log authentication information)
 *
 * Revision 1.20  2002/11/29 10:25:35  ctbto
 * Continue work on authentication (loading PEM keys)
 * Add possibility to compile cdreceiver without LDAP and
 * authentication. Add user extended flag (see configure -help)
 *
 * Revision 1.19  2002/11/11 15:02:56  ctbto
 * Move readKeyPreloadFile() to auth.c. Add #ifdef for authentication stuff
 * to allow user to compile when OpenSSL is not installed
 *
 * Revision 1.18  2002/10/28 15:35:49  ctbto
 * Fix bug: mandatory strings were not logged in dumpConfiguration()
 *
 * Revision 1.17  2002/10/18 12:05:56  ctbto
 * Correct some misspelling words
 *
 * Revision 1.16  2002/10/15 13:57:29  ctbto
 * Make configuration file mandatory, add shutdownMutex and move and
 * split functions (include John's comments)
 *
 * Revision 1.15  2002/10/15 07:20:01  ctbto
 * Log start and stop messages also to stdout
 * Change config parameter name file-prefix to file-path
 *
 * Revision 1.14  2002/09/26 10:28:48  ctbto
 * Change debug calls with level 1 to level 2 (level 1 is used to display
 * log messages only)
 *
 * Revision 1.13  2002/09/24 14:34:50  ctbto
 * Add MORE verbose logging to config parameter verboseEnable. Enhance
 * output of dumpConfiguration()
 *
 * Revision 1.12  2002/09/20 15:58:23  ctbto
 * Add responder type and creator name to config parameter. Add version to
 * startup messages
 *
 * Revision 1.11  2002/09/11 09:52:50  ctbto
 * Allow empty string for keyDirectory and certDirectory.
 * Improve shutdown code (wait until all handling threads have been
 * terminated before exit main thread).
 *
 * Revision 1.10  2002/09/05 16:20:52  ctbto
 * Include John's comments (rename functions, move functions to new files,
 * code changes, etc.)
 *
 * Revision 1.9  2002/09/02 11:31:47  ctbto
 * Use same data structure for CD1 and CD1.1 data information
 * Write more CD1.1 info to .txt file
 *
 * Revision 1.8  2002/08/26 10:00:31  ctbto
 * Change file name extension for raw CD_1.1 files to .BIN
 * Add more info to .txt files
 *
 * Revision 1.7  2002/07/23 09:07:18  ctbto
 * Add private key name, debug level and abs/relative wfdisc path names
 * to config params. Modify .txt output
 *
 * Revision 1.6  2002/07/04 15:54:41  ctbto
 * Add duplication of DFF frames. Add more text for DFF frames to .txt files
 *
 * Revision 1.5  2002/07/03 18:00:23  ctbto
 * Add information whether or not it is called from a connection request to
 * saveData(). Start work for .txt files. Enhance logging output.
 *
 * Revision 1.4  2002/06/26 14:56:08  ctbto
 * Add CD1.1 tbs entries
 *
 * Revision 1.3  2002/06/20 07:43:20  ctbto
 * Add mutex and condition for stopping monitor and certificate thread
 *
 * Revision 1.2  2002/06/19 10:01:07  ctbto
 * correct some splint warnings
 *
 * Revision 1.1.1.1  2002/06/18 07:28:05  ctbto
 * Import sources
 *
 */

#include <config.h>
#ifdef S_SPLINT_S
#include <netinet/in.h>  /* Need this to parse code with splint */
#endif /* S_SPLINT_S */
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "inittables.h"
#include "cdrecv.h"
#include "events.h"
#include "init.h"
#include "init_receiver.h"
#include "monitor.h"
#include "certificate.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif


/*
 * @short catch signals
 *
 * @description This function catches a signal and kills the main thread
 * @description with SIGKILL
 *
 * @param signal  signal
 *
 * @return None
 */

static void
catchSignal(int signal)
{
   char comBuffer[LOGSIZE];

   /* Don't use getpid() on LINUX here!! */
   (void) snprintf(comBuffer, LOGSIZE, "catchSignal: TID=%lu, SIG=%d",
                   (unsigned long)pthread_self(), signal);
   DEBUG(2, comBuffer);

   (void) pthread_kill(pthread_self(), SIGKILL);
}

/**
 * @short start required threads
 *
 * @description This function does all the necessary stuff to start the
 * @description required threads. This includes spawning the monitor, listener
 * @description or file reader and certificate thread and waiting for user
 * @description termination
 *
 * @param sigMask          signal mask for sigwait
 * @param pConfigData      configuration data
 * @param offlineMode      receivers operating mode
 *
 * @return    terminating signal number
 */

static int
startThreads(sigset_t *sigMask, configParams_t *pConfigData, int offlineMode)
{
   pthread_attr_t tAttr;     /* Thread attributes */
   pthread_t losthrdID;      /* ID listener or sender thread */
   pthread_t monID;          /* ID monitor thread */
#ifdef HAVE_AUTHENTICATION
   pthread_t certID;         /* ID certificate thread */
#endif /* HAVE_AUTHENTICATION */
   struct sigaction sigInfo;
   int sigVal,
       retVal;

   memset (&sigInfo, 0, sizeof (struct sigaction)); /* for splint */

   if (0 !=  pthread_attr_init(&tAttr))
   {
      stopProgram(APP_NAME, "can't initialize attribute", NULL);
   }

   if (0 !=  pthread_attr_setscope(&tAttr, PTHREAD_SCOPE_SYSTEM))
   {
      stopProgram(APP_NAME, "can't set scope", NULL);
   }

   if (0 == offlineMode)
   {
      /* Online mode */
      DEBUG(2, "+++ Online mode");

      /* start monitor thread if configured */
      if (0 != pConfigData->summaryRefreshTime &&
          0 != pthread_create(&monID, &tAttr, monitorThread, pConfigData))
      {
         stopProgram(APP_NAME, "can't start monitor thread", NULL);
      }

#ifdef HAVE_AUTHENTICATION
      /* start certificate thread if configured */
      if (0 != pConfigData->CRLcheckTime &&
          0 != pthread_create(&certID, &tAttr, certificateThread, pConfigData))
      {
         stopProgram(APP_NAME, "can't start certificate thread", NULL);
      }
#endif /* HAVE_AUTHENTICATION */

      /* start listener thread */
      if (0 != pthread_create(&losthrdID, &tAttr, portListen, pConfigData))
      {
         stopProgram(APP_NAME, "can't start listener thread", NULL);
      }
   }
   else
   {
      /* Offline mode, start reader thread */
      DEBUG(2, "+++ Offline mode");
      if (0 != pthread_create(&losthrdID, &tAttr, sendData, pConfigData))
      {
         stopProgram(APP_NAME, "can't start offline sender thread", NULL);
      }
   }

   /* Block all signals in main thread */
   (void) sigfillset(sigMask);
   (void) pthread_sigmask(SIG_BLOCK, sigMask, NULL);


   /* Wait until shutdown is signaled */
   sigVal = wait4shutdown(sigMask);

   /* Unblock all signals. This has no affect on LINUX if waiting
      on a pthread_join() call. Install signal handler */
   sigInfo.sa_handler = catchSignal;
   sigInfo.sa_flags	= 0;
   (void) sigaction(SIGINT, &sigInfo, NULL);
   (void) pthread_sigmask(SIG_UNBLOCK, sigMask, NULL);

   if (0 == offlineMode)
   {
      /* Online mode */

#ifdef HAVE_AUTHENTICATION
      /* Stop certificate Thread */
      if (0 != pConfigData->CRLcheckTime)
      {
         /* It's save to cancel the certificate thread */
         /*@-usedef@*/
         if (pthread_cancel(certID) != 0)
         /*@=usedef@*/
         {
            logError("can't cancel certificate thread", "startThreads");
         }
         /* Join the certificate thread
            Under Linux it could happen that joining the certificate thread
            is not possible when a segmentation fault has occurred. In this
            case the cdreceiver will wait at this point until a signal
            is received */
         DEBUG(9, "Before pthread_join certificate");
         retVal = pthread_join(certID, NULL);
         DEBUG(9, "After pthread_join certificate");
         pConfigData->pHistory->nbrOfThreadsStopped++;
         DEBUG2(5,"Joining certificates thread returns [%d]", retVal);
         logWrite(LOG_NOTICE, "Certificate thread stopped");
      }
#endif /* HAVE_AUTHENTICATION */

      /* Cancel Listener thread. Canceling this threads is save and
         has no side effects. The receiver should also work without
         canceling the listener thread, but this code path is not well
         tested */
      if (pthread_cancel(losthrdID) != 0)
      {
         /* This is nothing to concern. The listener thread could have
            been terminated before */
         logWarning("can't cancel listener thread", "startThreads");
      }
   }

   /* Set shutdown flag */
   if (SIGINT == sigVal) setShutdownFlag(SD_NORMAL);
   else setShutdownFlag(SD_FAST);

   /* Join the listener/sender thread
      Under Linux it could happen that joining the sender thread
      is not possible when a segmentation fault has occurred. In this
      case the cdreceiver will wait at this point until a signal
      is received */
   DEBUG(9, "Before pthread_join sender/listener");
   retVal = pthread_join(losthrdID, NULL);
   DEBUG(9, "After pthread_join sender/listener");
   pConfigData->pHistory->nbrOfThreadsStopped++;
   DEBUG2(5,"Joining listener/sender thread returns [%d]", retVal);

   if (0 == offlineMode)
   {
      /* Online mode */
      logWrite(LOG_NOTICE, "Listener thread stopped");
   }
   else
   {
      /* Offline mode */
      logWrite(LOG_NOTICE, "Offline sender thread stopped");
   }
   /* Wait for all threads to complete */
   joinHandlingThreads(pConfigData->pHistory->firstStation);

   /* Stop Monitor Thread. Don't use pthread_cancel() to stop monitor thread.
      That could lead to incomplete logging entries */
   if (0 == offlineMode && pConfigData->summaryRefreshTime != 0)
   {
      /*@-usedef@*/
      stopMonitor(monID, pConfigData);
      /*@=usedef@*/
      retVal = pthread_join(monID, NULL);
      DEBUG2(5,"Joining monitor thread returns [%d]", retVal);
      logWrite(LOG_NOTICE,"Monitor thread stopped");
   }

   if (0 !=  pthread_attr_destroy(&tAttr))
   {
      logError("can't destroy attribute", NULL);
   }

   return(sigVal);
}

/**
 * @short CD receiver main function
 *
 * @description This function handles to CLI arguments and calls
 * @description the appropriate functions for online or offline mode
 *
 * @param argc  Number or CLI arguments
 * @param argv  Vector with CLI arguments
 *
 * @return EXIT_FAILURE or EXIT_SUCCESS
 */

int
main (int argc, char *argv[])
{
   char comBuffer[LOGSIZE];  /* per Thread or on stack!! */
   char syslogID[48];
   char dateString[18];
   sigset_t  sigMask;        /* Signal mask value */
   int sigVal,
       offlineMode = 0,
       ii;
   configParams_t *pConfigData;
   receiverParams_t *pReceiverParams;

   /* Check if the CDreceiver don't run under super-user ID */
   if (0 == (ii = (int)getuid()))
   {
      /* If this would be allowed, the receiver would be able
         to read *ALL* files in the file system with a proper
         constructed configuration file */
      printf("Please start the CD-Receiver under a non super user ID\n");
      exit(EXIT_FAILURE);
   }

   /* initialize configuration data with default values */
   pConfigData = initConfigParams(APP_NAME, RECEIVER);

   /* Receiver port number is part of the syslog ID, but the configuration
      file has to be read first to get the configured port number. So
      a temporary syslog ID is used while reading the configuration file.
      Buffer is build with snprintf to avoid security holes! */
   (void) snprintf(syslogID, sizeof(syslogID), "ctbt.acquisition.cdrecv");
   logInit(syslogID, LOG_INFO);

   /* Check command line argument(s) */
   switch (argc)
   {
      case 2:
         /* The default is to create binary, wave disc,
            index and concise list files */
         pConfigData->systemConf |= CREATE_BIN_FILE |
                                    CREATE_CLF_FILE | CREATE_INDEX_FILE |
                                    ONLINE_MODE;
         /* Read the configuration file */
         (void) readConfigurationFile(APP_NAME, argv[1]);
         /* Add receiver specific object to station structure */
         pConfigData->sizeOfAppExtension = sizeof(cdrecv_t);
         /* Load receiver specific config variables */
         readReceiverConfigParams(APP_NAME, argv[1], pConfigData);
         break;
      case 3:
         /* The default is to create wave disc and clf files */
         pConfigData->systemConf |= CREATE_CLF_FILE;
         /* Read the configuration file */
         (void)readConfigurationFile (APP_NAME, argv[1]);
         /* Add receiver specific object to station structure */
         pConfigData->sizeOfAppExtension = sizeof(cdrecv_t);
         /* Load receiver specific config variables */
         readReceiverConfigParams(APP_NAME, argv[1], pConfigData);
         offlineMode = 1;
         /* Offline mode
            adjust some configuration parameters */
         /* Don't load a station file */
         pConfigData->authStationNameFile[0] = '\0';
         LOG2STDOUT("Offline mode, stationFile has no significance");
         /* Don't start monitor thread */
         pConfigData->summaryRefreshTime  = 0;
         /* Don't create index files is set in systemconf */
         pConfigData->indexFilePath[0] = '\0';
         /* Add extra suffix to file name if offline mode */
         pConfigData->offlineExt = EXTENSION_OFFLINE;
         pReceiverParams = (receiverParams_t *)pConfigData->configExtension;
         (void) strncpy(pReceiverParams->offlineFile, argv[2], LIBPAR_BUFSIZ);
         break;
      default:
         fprintf(stderr, "CD-Receiver %.10s (%.16s)\n",
                 VERSION, CDlibVersion());
         fprintf(stderr, "usage: %s parameter_file [offline_control_file]\n",
                 argv[0]);
         exit(EXIT_FAILURE);
   }

   /* Receiver port number is now available, so the real syslog ID
      can be set */
   closelog();
   (void) snprintf(syslogID, sizeof(syslogID), "ctbt.acquisition.cdrecv.%u",
                  (unsigned int)pConfigData->portNumber);
   logInit(syslogID, LOG_INFO);

   /* read further files if necessary and do some checks */
   processConfiguration(pConfigData, RECEIVER, APP_NAME);

   getDate(dateString, sizeof(dateString));
   (void) snprintf(comBuffer, LOGSIZE,
                   "CDreceiver %.10s (%.16s) started at %s by User %d",
                   VERSION, CDlibVersion(), dateString, ii);
   logWrite(LOG_NOTICE, comBuffer);

   /* if verboseEnable is set, dump configuration to syslog */
   if (pConfigData->systemConf & VERBOSE_LOGGING)
   {
       dumpConfigurationParameter();
       dumpReceiverConfigurationParameter(pConfigData);
   }

   /* The preferred way to handle signals is to block all signals in
      all threads and let the main thread waits for and processes
      signals. The following signals are used by the receiver:
      SIGHUP reserved for configuration reload
      SIGINT  orderly shutdown (CRTL C)
      SIGTERM reserved for immediate shutdown
      SIGUSR1 used internal by the receiver */
   (void) sigemptyset(&sigMask);

#if linux && HAVE_LIBPTHREAD
   /* unfortunately the preferred signal behavior will not work under the
      LINUX pthread implementation. If all signals are blocked, a synchronous
      signal is not seen in the main thread if the signal occurs in
      an other thread (don't mix occur with sent!!). This is not the
      POSIX behavior! If synchronous signals are not blocked in the main thread,
      a further blocked synchronous signal must be send to wake up sigwait().
      And than the (before sent) synchronous signal is delivered which cause
      the receiver to be terminated immediately. Very strange behavior!!
      For LINUX only a couple of synchronous signals are blocked for the
      worker threads. The main thread blocks all signals. This give the
      receiver a chance to handle synchronous signals as excepted. */
   (void) sigaddset(&sigMask, SIGHUP);
   (void) sigaddset(&sigMask, SIGINT);
   (void) sigaddset(&sigMask, SIGQUIT);
   (void) sigaddset(&sigMask, SIGTERM);
   (void) sigaddset(&sigMask, SIGUSR1);
#else /* linux && HAVE_LIBPTHREAD */
   (void) sigfillset(&sigMask);
#endif /* not linux && HAVE_LIBPTHREAD */

   (void) pthread_sigmask(SIG_BLOCK, &sigMask, NULL);

   pConfigData->mainID = pthread_self();
   /* Start required threads for online or offline mode */
   sigVal = startThreads(&sigMask, pConfigData, offlineMode);

   /* CD receiver is shutting down */
   getDate(dateString, sizeof(dateString));
   (void) snprintf(comBuffer, LOGSIZE, "CDreceiver stopped at %s by signal %d",
                   dateString, sigVal );
   logWrite(LOG_NOTICE, comBuffer);
   closeLibCD();
   closelog();
   exit(EXIT_SUCCESS);
}
