
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: offline.c,v $
 *
 * Revision 1.31  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.30  2009-06-04 07:44:20  gerald
 * Fix invalid offset in clf and txt files after reaching a
 * Disconnect Alert frame when a binary file is parsed in
 * offline mode with cdrecv
 *
 * Revision 1.29  2009-05-20 14:05:42  gerald
 * Fix timing problem in cdreceivers offline mode when an
 * alert frame is parsed
 *
 * Revision 1.28  2009-03-11 10:11:56  gerald
 * Fix splint warnings which occurred after upgrading
 * to new splint version
 *
 * Revision 1.27  2007/11/22 12:30:52  gerald
 * Fix splint warning
 *
 * Revision 1.26  2007/07/16 08:22:13  gerald
 * Fix too long shutdown time when a user requested shutdown is pending
 *
 * Revision 1.25  2007/05/10 09:09:45  gerald
 * Fix splint warnings
 *
 * Revision 1.24  2007/03/05 16:09:22  gerald
 * Fix stop parsing in CD Receivers offline mode when an
 * alert frame or an invalid frame has being parsed
 *
 * Revision 1.23  2006/04/26 08:28:06  gerald
 * Move receiver specific config and station structure members
 * into config extension respectively station extension structure.
 * Move reading of receiver specific config parameters from libCD
 * to receiver specific function. Update comments
 *
 * Revision 1.22  2005/05/11 09:33:18  gerald
 * Make use of new statistics structure
 *
 * Revision 1.21  2005/05/04 08:22:46  gerald
 * Drop redirectPortInfo structure
 *
 * Revision 1.20  2004/10/13 11:11:22  gerald
 * Change CD-1.0 type string from "cd1" to "cd10"
 * in the offline control file. Fix waiting receiver when an
 * offline control file parsing error occurs
 *
 * Revision 1.19  2004/09/13 08:37:01  gerald
 * Drop unused listener object and change some debug levels
 *
 * Revision 1.18  2004/09/06 09:32:18  gerald
 * Check station specific system configuration parameter when checking
 * if authentication is switched on instead of system wide parameter
 *
 * Revision 1.17  2004/03/09 13:57:59  gerald
 * Rename all variables which have tbs as part of the name and
 * rename WITHOUT_ defines
 *
 * Revision 1.16  2004/01/14 16:40:06  gerald
 * Fix compiler warnings when compiled without authentication
 *
 * Revision 1.15  2003/11/27 11:26:07  gerald
 * Add connection ID field to station structure
 *
 * Revision 1.14  2003/11/20 12:32:18  gerald
 * Move utility function shutdownReceiver() from CDreceiver to libCD and
 * rename it to shutdownApplication().  Fix splint warnings
 *
 * Revision 1.13  2003/10/17 14:01:17  gerald
 * Check for pending fast user shutdown request after a call to
 * writeWithTimeout()
 *
 * Revision 1.12  2003/08/27 14:26:15  gerald
 * Fixed some typos of the word "world"
 *
 * Revision 1.11  2003/04/17 14:48:04  gerald
 * Add attempt to load certificates from LDAP first
 *
 * Revision 1.10  2003/03/27 14:03:14  gerald
 * Fix more splint warnings
 *
 * Revision 1.9  2003/03/17 14:20:03  gerald
 * Rename NO_AUTHCHECK to AUTH_UNKNOWN
 *
 * Revision 1.8  2003/03/07 14:59:37  gerald
 * Signal readiness to main thread also when no control file exists
 *
 * Revision 1.7  2003/03/07 10:13:53  gerald
 * Include John's comments and fix some bugs (wait on startup until the
 * handling thread is in connected state, wait on shutdown until handling
 * thread has terminated, better behavior if a file is not found, etc.)
 *
 * Revision 1.6  2003/02/27 13:31:18  johannes
 * Fix compilation error when compile without openssl installed
 *
 * Revision 1.5  2003/02/17 15:53:23  ctbto
 * Add key loading if keys not pre-loaded. Use nanosleep() instead of sleep()
 *
 * Revision 1.4  2003/02/12 09:10:34  ctbto
 * Fix warnings that occur when compiling with  -Wall
 *
 * Revision 1.3  2003/02/07 15:11:13  ctbto
 * Check for response from handling thread after sending data
 *
 * Revision 1.2  2003/01/31 16:09:56  ctbto
 * Change debug output
 *
 * Revision 1.1  2003/01/31 13:15:45  ctbto
 * Initial revision
 *
 *
 */

#include <config.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include "cdrecv.h"
#include "cd_ldap.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

/**
 * @short connect handling thread
 *
 * @description This function connects to the data handling thread.
 *
 * @param sd        socket descriptor
 * @param pStation  pointer to station object
 *
 * @return 0 on success or -1 on error
 */

static int
connectHandlingThread(int sd, station_t *pStation)
{
   struct sockaddr_in sockAddr;
   struct timespec sleeptime;
   int tries = 3;

   /* Setup socket parameters */
   memset((char *) &sockAddr, 0, sizeof(sockAddr));
   sockAddr.sin_family = AF_INET;
   sockAddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
   DEBUG2(2,"Port is %hu",
         (unsigned short)pStation->portInfo->localPort);
   sockAddr.sin_port = htons(pStation->portInfo->localPort);

   if (connect(sd, (struct sockaddr *)&sockAddr,
               (socklen_t)sizeof(sockAddr)) < 0)
   {
      /* A check for EINPROGRESS could be useful here. This is a
         non blocking socket */
      if (errno != EINPROGRESS)
      {
         (void) close (sd);
         return (-1);
      }
   }

   /* In theory writeWithtimeout should fail if no connection could
      be established. In the real world and at least under Linux 2.4.16
      no blocking does occur when sending a small file. The data
      is buffered in the TCP buffers instead. So the cdreceiver has
      to wait until Handling thread changes its state */

   while (STATION_WAITING == pStation->state && tries > 0)
   {
      DEBUG2(5,"Wait for station thread [%d]", tries);
      /*@-usedef@*/
      sleeptime.tv_sec = 1;
      sleeptime.tv_nsec = 0;
      /*@=usedef@*/
      (void) nanosleep(&sleeptime, NULL);
      tries--;
   }
   if (tries > 0) return(0);
   (void) close (sd);
   return(-1);
}

/**
 * @short Start offline handling thread
 *
 * @description This function starts a handling thread for processing
 * @description the data offline
 *
 * @param pConfigData  pointer to configuration data
 * @param pStation     station object
 *
 * @return file descriptor on success, -1 on fatal error
 */

static int
startOffline(/*@dependent@*/configParams_t *pConfigData, station_t *pStation)
{
   int sendSock;           /* Sender socket */

   /* Create and configure sender socket */
   if (-1 == (sendSock = createSocket()))
   {
      shutdownApplication("Error configuring Sender thread",
                          "createSocket failed");
      return(-1);
   }

   /* Simulate connection request */
   pStation->state = STATION_CONNECTING;
   pStation->files.clfRecNum = 0;
   /* Set local port and IP address in the portinfo structure.
      In online mode, the listener thread would do all these
      stuff  */
   /* pStation->portInfo->localPort = 0; */
   memcpy(pStation->portInfo->localIPaddr, "127.0.0.1",
          sizeof(pStation->portInfo->localIPaddr));

#ifdef HAVE_AUTHENTICATION

   /* Check if certificates have to be loaded */
   if (HAVE_TO_LOAD_CERTIFCATES(pStation))
   {
      if (0 != loadCertificates(pStation,
                       (pConfigData->systemConf & VERBOSE_LOGGING)))
      {
         return(-1);
      }
   }

#endif /* HAVE_AUTHENTICATION */

   /* Start handling thread */
   if (0 != startHandlingThread(pConfigData, pStation))
   {
      LOG2STDOUT("Can't start handling thread");
      shutdownApplication("can't start handling thread",
                          "processOfflineFile");
      return(-1);
   }

   /* Connect to handling thread */
   if (-1 == connectHandlingThread(sendSock, pStation))
   {
      /* Unable to connect to the handling thread */
      LOG2STDOUT("Can't connect to the handling thread");
      shutdownApplication("Can't connect to the handling thread",
                          "processOfflineFile");
      return(-1);
   }
   return(sendSock);
}

static void
wait4handlingThread(station_t *pStation, int bytesSent)
{
   /* Give handling thread time to process the data. The sending
      has to wait now until all sent bytes are processed by the
      handling thread */
   if (NULL != pStation)
   {
      struct timespec sleeptime;

      DEBUG2(2, "state: %d", (int)pStation->state);

      while (STATION_CONNECTED == pStation->state &&
             (int)pStation->framesReceived.bytesTotal != bytesSent)
      {
         DEBUG3(5,"wait4handlingThread: Bytes sent [%d], received [%d]",
               bytesSent, (int)pStation->framesReceived.bytesTotal);
         /*@-usedef@*/
         sleeptime.tv_sec = 2;
         sleeptime.tv_nsec = 0;
         /*@=usedef@*/
         (void) nanosleep(&sleeptime, NULL);
      }

      DEBUG3(2," wait4handlingThread(2): Bytes sent [%d], received [%d]",
             bytesSent, (int)pStation->framesReceived.bytesTotal);
   }
}

/**
 * @short send file to handling thread
 *
 * @description This function sends the contents of a given file
 * @description via a socket to the handling thread
 *
 * @param name        filename
 * @param pConfigData pointer to configuration data
 * @param buffer      buffer for sent/received data
 * @param sendSock    socket descriptor
 * @param pStation     station object
 *
 * @return number of bytes sent on success or -1 on error
 */

static int
sendFile(const char *name, /*@dependent@*/configParams_t *pConfigData,
         /*@out@*/char *buffer, int * const sendSock, station_t *pStation)
{
   FILE *fp;
   rwFlags_t flags;
   int len,     /* bytes to send */
       rval;
   uint32_t lastBytesRead,
            bytesSent = 0,
            bytesReceived;

   LOG2STDOUT2("Processing [%.128s] ..... ", name);
   DEBUG2(2,"Processing [%.128s] ..... ", name);

   /* Open file to sent */
   if (NULL == (fp = fopen(name,"r")))
   {
      LOG2STDOUT2("Can't open file [%s]", name);
      /*@-mustdefine@*/
      return(0);
      /*@=mustdefine@*/
   }

   /* Set all timeouts and other parameters for writeWithTimeout */
   flags.sockFlags = 0;
   flags.shutdownFlag = &(pConfigData->shutdownEnable);
   flags.timeout = pConfigData->timeOut;
   flags.firstTime = NULL;
   flags.resetDataTimeout = pConfigData->timeOut;
   /* Remember offset from previous frame */
   lastBytesRead = pStation->framesReceived.bytesTotal;

   do
   {
     while ((len = (int)fread(buffer, 1, (size_t)BUFSIZ, fp)) > 0)
     {
        /* Sent data to handling thread. There is no need to
           take care for frame boundaries, the receiver is smart
           enough to handle data streams */

         DEBUG4(2,"lastBytesRead [%d], bytesSent [%d] Received.bytesTotal [%d]",
                (int)lastBytesRead, (int)bytesSent,
                (int)pStation->framesReceived.bytesTotal);
         DEBUG2(5,"Try to send [%d] bytes", len);
         rval = writeWithTimeout(*sendSock, buffer, len, &flags);
         DEBUG2(2,"Send [%d] bytes", rval);

         if (rval != len)
         {
            if (-1 == rval)
            {
               /* Interrupt or system failure */
               shutdownApplication("can't send data", "processOfflineFile");
               (void) fclose(fp);
               return (-1);
            }
            else if (-2 == rval)
            {
               /* User termination request */
               LOG2STDOUT2("User termination request while sending[%s]", name );
               (void) fclose(fp);
               return (-1);
            }
            else if (0 == flags.timeout)
            {
               /* timeout */
               LOG2STDOUT2("Send timeout line while sending[%s]", name );
               (void) fclose(fp);
               return(-1);
            }
            else if (SD_NONE != pConfigData->shutdownEnable)
            {
               /* User termination request */
               LOG2STDOUT2("User termination request while sending[%s]", name );
               (void) fclose(fp);
               return (-1);
            }
            else
            {
               bytesReceived = pStation->framesReceived.bytesTotal -
                               lastBytesRead;

               /* connection closed */
               DEBUG2(2, "Connection closed unexpected while sending [%s]",
                      name);

               if (bytesReceived != bytesSent)
               {
                  struct timespec sleeptime;

                  /* This can be caused by an invalid frame or an alert
                     frame in the binary data. Seek back to the last
                     byte read from the handling thread + 1 and restart the
                     handling thread */
                  (void)fseek(fp, (long)bytesReceived, SEEK_SET);
                  len = 0;
                  bytesSent = bytesReceived;
                  /* Wait some seconds to give the handling thread a chance
                     to shutdown */
                  /*@-usedef@*/
                  sleeptime.tv_sec = 2;
                  sleeptime.tv_nsec = 0;
                  /*@=usedef@*/
                  (void) nanosleep(&sleeptime, NULL);

                  if (-1 == (*sendSock = startOffline(pConfigData, pStation)))
                  {
                     shutdownApplication("Can't start handling thread",
                                         "startOffline failed");
                     return(-1);
                  }
                  /* Receiver checks the binary file size when the handling
                     thread is started, so the size of the binary file must
                     be set here otherwise it starts from 0 again when no
                     binary file is created */
                  pStation->files.sizeBinFile = bytesSent;
               }
               else
               {
                  (void) fclose(fp);
                  return(-1);
               }
            }
         }

         /* Increase number of sent bytes to know when the connection
             could be closed */
         bytesSent += (uint32_t)len;

         /* Check for response from handling thread to avoid
            receive buffer overrun. The handling thread would
            sent AckNack and Alert frames also in offline mode */
         do
         {
            rval = (int)recv(*sendSock, buffer, (size_t)BUFSIZ, 0);
            DEBUG2(2,"Receive [%d] bytes", rval);
         } while (rval > 0);

         if (feof(fp)) break;
     }

     /* Wait until the handling thread has processed the data or
        it has closed the connection */
     wait4handlingThread(pStation, (int)(bytesSent + lastBytesRead));
     bytesReceived = pStation->framesReceived.bytesTotal - lastBytesRead;

     /* Check if the connection was not closed by a timeout and
        if all bytes sent are received by the handling thread */

     if (pStation->whenConnected.timeouts == 0 &&
         bytesReceived != bytesSent)
     {
        /* This can be caused by an invalid frame or an alert
           frame in the binary data. Seek back to the last
           byte read from the handling thread + 1 and restart the
           handling thread */
        (void)close(*sendSock);
        (void)fseek(fp, (long)bytesReceived, SEEK_SET);
        len = 0;
        bytesSent = bytesReceived;

        if (-1 == (*sendSock = startOffline(pConfigData, pStation)))
        {
           shutdownApplication("Can't start handling thread",
                               "startOffline failed");
           return(-1);
        }
     }
     else break;
   } while (1);

   (void) fclose(fp);
   return((int)bytesSent);
}

/**
 * @short Process offline file
 *
 * @description This function processes the offline control file. It
 * @description starts a CD1 or CD1-1 handling thread, reads the input
 * @description file(s) and sends the data to the handling thread
 *
 * @param pConfigData  pointer to configuration data
 * @param buffer       buffer for data to sent
 *
 * @return 0 on success, 1 on error, -1 on fatal error
 */

static int
processOfflineFile(/*@dependent@*/configParams_t *pConfigData,
                   /*@reldef@*/char *buffer)
{
   int sendSock = -1;   /* Sender socket */
   int isStarted = 0,   /* Handling thread has started */
       ii;
   char *name,
        *value;
   station_t *pStation = NULL;
   int bytesSent = 0,
       tries = 3,
       rval = -1;

   DEBUG(5, "+++ Process offline file");

   /* Process lines which are no comment */
   for (ii = 0; getparline(ii, &name, &value); ii++)
   {
      DEBUG3(5, "Read [%s] [%s]", name, value);

      if ('\0' == *name)
      {
         LOG2STDOUT2("Invalid parameter line [%d]", ii);
         continue;
      }

      if (0 == ii)
      {
         /* read station name and type */
         if ('\0' == *value)
         {
            /* first line must contain station name and protocol */
            LOG2STDOUT2("Missing protocol [%d]", ii);
            /* Wakeup main thread */
            (void) pthread_kill(pConfigData->mainID, SIGINT);
            return(1);
         }

         /* Create station object */
         if (NULL == (pStation = getStation(name, name)))
         {
            /* No memory available */
            shutdownApplication("No memory available", "getStation failed");
            return(-1);
         }

         if (0 == strncmp("cd10", value, 5)) pStation->type = CD1;
         else if (0 == strncmp("cd11", value, 5)) pStation->type = CD1_1;
         else
         {
            /* Unknown protocol version */
            LOG2STDOUT2("unknown protocol version (cd10, cd11) [%s]", value);
            /* Wakeup main thread */
            (void) pthread_kill(pConfigData->mainID, SIGINT);
            return(1);
         }

         if (-1 == (sendSock = startOffline(pConfigData, pStation)))
         {
            shutdownApplication("Can't start handling thread",
                                "startOffline failed");
            return(-1);
         }

         isStarted = 1;
         continue;
      }

      /* Check for a station line */
      if (0 == isStarted)
      {
         LOG2STDOUT2("No station line before input file [%d]", ii);
         /* Wakeup main thread */
         (void) pthread_kill(pConfigData->mainID, SIGINT);
         return(1);
      }

      if (NULL == pStation) /* for splint */
      {
          shutdownApplication("Unexpected NULL pointer", "pStation");
          return(-1);
      }

      /* Send the file */
      rval = sendFile(name, pConfigData, buffer, &sendSock, pStation);

      /* Has an error occurred? */
      if (-1 == rval) break;

      bytesSent += rval;
      /* Check if station is in connected state */
      if (STATION_CONNECTED != pStation->state) break;
   }

   if (NULL == pStation) /* for splint */
   {
       shutdownApplication("Unexpected NULL pointer (2)", "pStation");
       return(-1);
   }

   /* Wait until handling thread has disconnected */
   while (STATION_CONNECTED == pStation->state && tries > 0)
   {
      struct timespec sleeptime;

      /*@-usedef@*/
      sleeptime.tv_sec = 2;
      sleeptime.tv_nsec = 0;
      /*@=usedef@*/
      (void) nanosleep(&sleeptime, NULL);
      tries--;
   }

   DEBUG3(2,"processOfflineFile: Bytes sent [%d], received [%d]", bytesSent,
          (int)pStation->framesReceived.bytesTotal);

   /* Wakeup main thread */
   (void) pthread_kill(pConfigData->mainID, SIGINT);
   return(0);
}

/**
 * @short Read offline file
 *
 * @description This function reads the offline control file.
 *
 * @param pConfigData  pointer to configuration data
 *
 * @return nothing
 */


static void
readOfflineFile(/*@dependent@*/configParams_t *pConfigData)
{
   char *buffer;           /* buffer for data to send */
   receiverParams_t *pReceiverParams;

   pReceiverParams = (receiverParams_t *)pConfigData->configExtension;
   if (NULL == pReceiverParams->offlineFile) return;

   DEBUG2(5, "+++ Read offline file [%s]", pReceiverParams->offlineFile);

   /* Process offline control file */
   if (setfpar("CDreceiver", pReceiverParams->offlineFile) != 0)
   {
      LOG2STDOUT2("can't read offline control file [%.80s]",
                  pReceiverParams->offlineFile);

      /* Wakeup main thread */
      (void) pthread_kill(pConfigData->mainID, SIGINT);

      return;
   }

   DEBUG(9, "Nach setfpar  ");

   /* Check if something is to do */
   if  (0 ==  getparlistlen())
   {
      LOG2STDOUT2("nothing to do in offline control file [%.80s]",
                  pReceiverParams->offlineFile);
      endpar();

      /* Wakeup main thread */
      (void) pthread_kill(pConfigData->mainID, SIGINT);

      return;
   }

   /* Allocate buffer; buffer is aligned proper by malloc */
   if (NULL == (buffer = (char *) malloc((size_t)BUFSIZ)))
   {
      endpar();
      shutdownApplication("Error configuring Sender thread",
                          "can't allocate memory");
      pthread_exit(NULL);
      return;
   }
   /* memset(buffer, 0, BUFSIZ); */

   DEBUG(5, "Before process offline file ");

   /* Start the real processing now */
   if (-1 == processOfflineFile(pConfigData, buffer))
   {
      endpar();
      (void)free(buffer);
      pthread_exit(NULL);
      return; /*for splint */
   }

   endpar();
   (void)free(buffer);
}

/**
 * @short sender thread - start function
 *
 * @description This function is the starting point of the sender
 * @description thread. It initializes all the necessary variables and
 * @description reads the file with the file names to send, simulates
 * @description a connection request, starts a handling thread and sent
 * @description data to this thread
 *
 * @param arg  void pointer to config parameters
 *
 * @return None
 */

/*@null@*/ void *
sendData(/*@dependent@*/void *arg)
{
   /*@dependent@*/configParams_t *pConfigData = NULL;


   DEBUG(2, "+++ Sender thread started");

   /* Retrieve data passed to thread */
   pConfigData = (configParams_t *) arg;

   /* Sender uses the same signal mask as the main thread */

   /* Write startup message to syslog */
   logWrite(LOG_NOTICE, "Sender thread started");
   pConfigData->pHistory->nbrOfThreadsStarted++;

   /* Read and process offline control file */
   readOfflineFile(pConfigData);

   DEBUG(2, "+++ Sender thread exiting");

   /* nbrOfThreadsStopped is set in the main thread after pthread_join() */
   return(NULL); /* Only for splint */
}
