
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Johannes Schreitl
 *
 * Revision 1.75  2015-11-27 gerald
 * Follow change in CDlibVersion()
 *
 * Revision 1.74  2015-09-06 gerald
 * Make loop through production lines more robust
 *
 * Revision 1.73  2015-03-07 gerald
 * Assign production line in initSenderStation()
 *
 * Revision 1.72  2013-06-13 gerald
 * Add CD Sender and libCD version to usage string. Add
 * libCD version when logging start message to syslog.
 * Add closing libCD. Write usage message to stderr
 *
 * Revision 1.71  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.70  2012-07-25 gerald
 * Add support for black list
 *
 * Revision 1.69  2012-06-27 gerald
 * Fix function comment
 *
 * Revision 1.68  2011-09-20 14:03:14  ctbto
 * Remove DB access code from CD Sender
 *
 * Revision 1.67  2009-05-19 13:52:56  gerald
 * Change syslog IDs as requested by Henrik
 *
 * Revision 1.66  2009-04-06 15:08:12  gerald
 * Fix incorrect start and end times in fixed interval mode
 *
 * Revision 1.65  2008-09-16 18:04:12  gerald
 * Ignore suspend signal (CRTL-Z) and log a warning to
 * stdout when CRTL-Z is pressed
 *
 * Revision 1.64  2008-06-03 14:51:26  gerald
 * Add new parameter for createDC and close DB connection before shutdown
 *
 * Revision 1.63  2007/05/10 09:23:40  gerald
 * Move database files back to cdsender to simplify build
 * dependencies for other applications (e. g. geotool)
 *
 * Revision 1.62  2007/01/09 12:12:28  gerald
 * Add monitor and certificate thread to sender
 *
 * Revision 1.61  2007/01/04 13:06:17  gerald
 * Change parameter of catIdxDir(). The passed start date is set to
 * the time of the last index file parsed to avoid sorting too many
 * index files when the sender runs for a long time
 *
 * Revision 1.60  2006/11/30 16:34:52  gerald
 * Envelop sender commands with EXPERIMENTAL define and
 * make auxiliary mode parameter consistent with the documentation
 *
 * Revision 1.59  2006/11/02 13:24:27  gerald
 * Treat sender start end and time as UTC time (times in file
 * names are UTC times)
 *
 * Revision 1.58  2006/10/04 12:41:09  gerald
 * Change data type of start date, end date and clfDate to time_t
 * Check in unconnected or stopped state if the clf file list
 * contains too old clf files. Comment updates
 *
 * Revision 1.57  2006/09/06 12:35:26  gerald
 * Check arguments for auxiliary station mode
 *
 * Revision 1.56  2006/03/28 09:04:34  gerald
 * Rename HAVE_LIBGODBC to HAVE_DB_ACCESS
 *
 * Revision 1.55  2006/03/20 14:05:03  gerald
 * Fix some typos and fix missing shutdown message when in continuous mode
 *
 * Revision 1.54  2006/03/08 09:58:31  gerald
 * Make date format for fixed interval mode consistent to cd2w
 *
 * Revision 1.53  2006/03/07 09:46:23  gerald
 * Use standard date formate instead of Julian for CD sender CLI
 *
 * Revision 1.52  2006/02/27 10:17:29  gerald
 * Add new params for catIdxDir()
 *
 * Revision 1.51  2006/02/14 10:25:04  gerald
 * Fix some typos and switch to sender specific config structure.
 * Using a sender specific config structure for sender specific
 * parameters requires to read the config file a second time to
 * load sender specific parameters.
 *
 * Revision 1.50  2006/01/31 15:52:16  gerald
 * Fix wrong check for HAVE_LIBODBC. This is set to true when any
 * ODBC library is installed. But Cdtools need libgODBC
 *
 * Revision 1.49  2005/10/12 12:37:36  gerald
 * Read list of clf to parse during start up from DB when DB connection is
 * configured
 *
 * Revision 1.48  2005/10/05 12:19:09  gerald
 * Change key loading behavior in sender. Now keys are loaded before the
 * handling thread is started or if it is re-started. This avoids loading
 * of certificates in different threads
 *
 * Revision 1.47  2005/09/21 09:22:11  gerald
 * Make use of CDtoolsVersion()
 *
 * Revision 1.46  2005/08/02 08:46:24  gerald
 * Open optional ODBC connection when ODBC is configured
 *
 * Revision 1.45  2005/07/14 13:04:51  gerald
 * Adjust debug level of some debug messages and move code for
 * creating of a station object to schedule.c
 *
 * Revision 1.44  2005/07/12 12:45:51  gerald
 * Make use of application extension object for sender specific
 * variables
 *
 * Revision 1.43  2005/07/07 09:44:38  gerald
 * Restart sender threads when in unconnected state caused by a stop
 * command or another event. Use control port for senders syslog ID
 * instead of receivers well known port
 *
 * Revision 1.42  2005/06/16 07:23:54  gerald
 * Change priority of a debug message
 *
 * Revision 1.41  2005/02/21 15:23:12  gerald
 * Start with sender specific commands
 *
 * Revision 1.40  2005/02/08 10:58:11  gerald
 * Fix wrong application string for cdsender
 *
 * Revision 1.39  2004/11/12 14:26:52  gerald
 * Remove log messages to stdout and add application ID to
 * initConfigParms()
 *
 * Revision 1.38  2004/09/16 15:23:58  gerald
 * Drop first command interface implementation and start a
 * command thread instead.
 *
 * Revision 1.37  2004/09/10 11:10:34  gerald
 * Add start and end date CLI parameter. Start work on command
 * interface.
 *
 * Revision 1.36  2004/07/06 14:38:56  gerald
 * Add end date processing to scheduler thread. Allocate DFinfo to
 * log nominal time of outgoing frames. Fix timing problem when
 * joining worker threads
 *
 * Revision 1.35  2004/04/27 15:27:43  gerald
 * Fix missing "ctbt.acquisition..." string in syslog messages when
 * an initialization error is logged. The port number in the syslog
 * ID string is added after the configuration file is read
 *
 * Revision 1.34  2004/03/26 08:39:18  gerald
 * Clean up code
 *
 * Revision 1.33  2004/03/08 15:55:09  gerald
 * Check all file paths during startup and create directories
 * if necessary
 *
 * Revision 1.32  2004/03/05 13:49:33  gerald
 * Fix random state changes
 *
 * Revision 1.31  2004/02/25 12:24:43  gerald
 * Fix wrong start date when re-parsing index files
 *
 * Revision 1.30  2004/02/18 15:30:31  gerald
 * Fix message and debug level
 *
 * Revision 1.29  2004/02/13 14:27:08  gerald
 * Update comments, add CD conform logging and better debugging output,
 * remove unused variables, change global CLF locking to station
 * specific locking, fix splint warnings and remove wakeup of worker
 * threads (worker threads now waiting for incoming data and check
 * for new data to send in given intervals)
 *
 * Revision 1.28  2004/02/03 10:21:08  gerald
 * Remove trailing blanks and change indents
 *
 * Revision 1.27  2004/01/19 16:07:32  karol
 * Simple continuous mode. After start sender read all data to send in
 * default time interval. After sending wait in continuous mode for new
 * data and send them.
 *
 * Revision 1.26  2003/12/09 10:53:09  gerald
 * Set history structure in initConfigParames()
 *
 * Revision 1.25  2003/11/28 10:37:05  gerald
 * Clean up Cdsender members in station structure
 *
 * Revision 1.24  2003/11/27 10:43:23  gerald
 * Allocate station structures with libCD function and
 * remove sender specific allocation function
 *
 * Revision 1.23  2003/11/20 13:11:10  gerald
 * Use station structure in CDsender and use stopProgram() instead of
 * stopSender().  Use more new functions from libCD.
 *
 * Revision 1.22  2003/11/20 08:31:05  karol
 * Remove watching thread, spawn of schedule thread.
 *
 * Revision 1.21  2003/11/14 10:29:58  karol
 * Common structure with cdreceiver (prodLineParam is replaced with
 * station struct).
 *
 * Revision 1.20  2003/11/07 10:58:57  gerald
 * Use common configuration structure and fix some bugs
 *
 * Revision 1.19  2003/10/30 09:04:26  karol
 * Schedule procedure of the sender. In future will be designed like thread.
 *
 * Revision 1.18  2003/10/28 10:23:56  karol
 * Test if not running like superuser
 * Remove some bugs
 *
 * Revision 1.17  2003/10/28 09:25:51  gerald
 * Fix core dump if no index file exists
 * Rename some configuration parameter to be consistent with CDreceiver
 * and documentation
 *
 * Revision 1.16  2003/10/17 14:07:37  gerald
 * Remove loading of "enableFsfFile" from the config file. It's not
 * described in the SDD
 *
 * Revision 1.15  2003/05/23 05:56:34  johannes
 * debug messaged in watch thread enhanced
 *
 * Revision 1.14  2003/05/23 05:11:20  johannes
 * fix signal handling for linux
 *
 * Revision 1.13  2003/05/22 13:45:27  johannes
 * signalhandling disabled
 *
 * Revision 1.12  2003/05/22 08:51:36  johannes
 * redesign of signal handling
 *
 * Revision 1.11  2003/05/21 11:55:44  johannes
 * Handling for SIGINT added
 *
 * Revision 1.10  2003/05/19 13:00:25  johannes
 * send data in frame borders
 *
 * Revision 1.9  2003/05/14 10:00:20  gerald
 * Fix architecture dependent bugs
 *
 * Revision 1.8  2003/05/12 12:32:38  johannes
 * multithreading for sending data added
 *
 * Revision 1.7  2003/05/08 07:03:20  johannes
 * changes for prototype sending CD1 frames
 *
 * Revision 1.6  2003/04/23 08:02:20  johannes
 * .
 *
 * Revision 1.4  2003/04/15 08:57:01  johannes
 * some printouts for test added
 *
 * Revision 1.1  2003/04/02 11:09:15  johannes
 * first issue
 *
 */

#include <config.h>

#ifdef S_SPLINT_S
 #include "../../splint_hints.h"
#endif /* S_SPLINT_S */

#include <netinet/in.h>  /* Need this to parse code with splint */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "events.h"
#include "cdsender.h"
#include "snd_command.h"
#include "init.h"
#include "inittables.h"
#include "init_sender.h"
#include "sending.h"
#include "command.h"
#include "monitor.h"
#include "certificate.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

#define JOIN_WORKER_THREADS joinHandlingThreads
#define SEVEN_DAYS   (60*60*24*7)

/**
 * @short Start or restart sending threads
 *
 * @description This function starts or restart a sending thread for
 * @description each known production line. No thread is started
 * @description when the sending thread is in the 'stopped' state.
 * @description periodically
 *
 * @param pConfigData     Configuration data
 *
 * @return Nothing
 */

static void
startSenderThreads(configParams_t *pConfigData)
{
   station_t *pStation;

   /* Start a worker (sending) thread for each production line
      when no sending thread exists */
   for (pStation = pConfigData->pHistory->firstStation;
        pStation != NULL;
        pStation = pStation->nextStation)
   {
      if (NULL == pStation->appExtension ||
          NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine)
      {
         /* Station without production line. This one is
            created from a key preload directive, we can ignore
            it */
         DEBUG2(9, "Station without production line found [%.17s]",
                pStation->conID);
         continue;
      }

      if (STATION_STOPPED == pStation->state)
      {
         /* Stopped station. No sending thread is created for
            such a station. A sending thread is created only if the
            state changes (by a start command) */
         DEBUG2(9, "Station stopped [%.17s]", pStation->conID);

         /* Check for clf files older than N days and remove these entries
            Removing entries is only allowed when no thread is started
            for that production line (STOPPED or UNCONNECTED state) */
         (void)removeTooOld(pStation, SEVEN_DAYS);
         continue;
      }

      if (STATION_UNCONNECTED != pStation->state)
      {
         /* Sending thread is already started */
         DEBUG3(9, "Thread exists [%.17s] [%d]", pStation->conID,
                (int) pStation->state);
         continue;
      }

      DEBUG2(9, "Try to start sender thread for [%s]", pStation->conID );

      /* Check if a shutdown request is pending. Don't check
         this before the debug messages. A printf is a possible
         thread switch! */
      if (SD_NONE != pConfigData->shutdownEnable) break;

#ifdef HAVE_AUTHENTICATION
      /* Check if certificates have to be loaded. Don't
         load the certificates in the handling threads,
         the functions might not be thread save */
      if (HAVE_TO_LOAD_CERTIFCATES(pStation))
      {
         /* No keys are loaded, load all keys for station */
         if (0 != loadCertificates(pStation,
                                  (pStation->systemConf & VERBOSE_LOGGING)))
         {
            return;
         }
      }
#endif /* HAVE_AUTHENTICATION */

      /* Start sender thread */
      pStation->state = STATION_WAIT2SEND;
      if (0 != pthread_create(&(pStation->threadID), NULL, sending, pStation))
      {
         shutdownApplication("can't start sending thread", "schedThread");
      }
      pConfigData->pHistory->nbrOfThreadsStarted++;
   }
}

/**
 * @short Schedule thread
 *
 * @description The scheduler thread is responsible for starting a
 * @description worker thread for each production line and for supplying
 * @description each worker thread with an initial clf file list.
 * @description In continuous mode the initial clf file list is updated
 * @description periodically
 *
 * @param arg  void pointer to config parameters
 *
 * @return Nothing
 */

/*@null@*/ static void *
schedThread(void *arg)
{

   station_t *pStation;
   configParams_t *pConfigData;  /* configuration parameter */
   struct _prodLine *pProdLine;  /* Current production line */
   int indexFilesParsed;
   char conID[LEN_CON_ID + 1];
   time_t startDate,             /* start date for parsing index files */
          endDate,               /* end date for parsing index files */
          nowTime;
   senderParams_t *pSenderParams;
#ifdef HAVE_AUTHENTICATION
   pthread_t comID;        /* ID command thread */
   station_t *pControl;
#endif /* HAVE_AUTHENTICATION */

   DEBUG(5, "+++ Scheduler thread started");
   /* Retrieve data passed to thread */
   pConfigData = (configParams_t *) arg;
   pConfigData->pHistory->nbrOfThreadsStarted++;

   /* set first production line */
   pSenderParams = (senderParams_t *)pConfigData->configExtension;
   pProdLine = pSenderParams->pProdLine;

   /* At least one production line has to exist */
   if (NULL == pProdLine)
   {
      stopProgram(APP_NAME, "Nothing to do", "schedThread");
      return(NULL); /*for splint */
   }

#ifdef HAVE_AUTHENTICATION
   /* Create a station object for the controller station */
   pControl = getControlObject(pConfigData);
   if (NULL != pControl)
   {
      DEBUG(2,"Control object created");

      /* starting of command thread */
      if (0 != pthread_create(&comID, NULL, commandThread, pConfigData))
      {
         pControl = NULL;
         stopProgram(APP_NAME, "can't start command thread", NULL);
      }
   }
#endif /* HAVE_AUTHENTICATION */

   /* Create a station object for each line in the production line file
      which does not contain a wildcard as data provider */
   for (; pProdLine != NULL; pProdLine = pProdLine->pNextprodLine)
   {
      if (NULL != pProdLine->pDataConsumer)
      {
         /* Check for wildcard in Data Provider name. No station object is
            created for a Data Provider with a wildcard now. This is done
            later when the index file is parsed */
         if ('*' == pProdLine->pDpName[0] && '\0' == pProdLine->pDpName[1])
            continue;

         /* Check if the station is on the black list or if one or more
            channel is on the black list. Keep in mind that, if a station
            is on the black list, the data can be processed later. No
            IRL file is created or modified, whereas the data of a blacklisted
            channel can't be processed later (in continuous mode).
            Blacklisted sites/channel are not supported yet. Stations with
            blacklisted site/channels must be created with the wild
            card later */
         if (pProdLine->isBlacklisted)
            continue;
         /* Check if channels/sites have been blacklisted here to
            distinguish between hole station and site/channel blacklists
         if (pProdLine->isBlacklisted && !HAS_CHANNEL_BLACKLISTED(pProdLine))
         {
            continue;
         } */

         /* Create new station object for each production line.
            The ID of the station object is DataproviderName_DataconsumerName
            and as the station name (used to find the right certificates)
            the Dataconsumer name is used */
         (void) snprintf(conID, sizeof(conID), "%s_%s", pProdLine->pDpName,
                          pProdLine->pDataConsumer->dcName);
         pStation = initSenderStation(conID, pProdLine->pDataConsumer->dcName,
                                      pProdLine);

         if (NULL == pStation)
         {
            stopProgram(APP_NAME, "can't create station object", conID);
            return(NULL); /* for splint */
         }
      }
   }

   /* Set parsing start and end time */
   nowTime = time(NULL);

   if (0 == pSenderParams->startDate)
   {
      /* continuous mode */
      startDate = nowTime - (pSenderParams->lookBackTimeSched * 3600);
      endDate = 0;
      DEBUG(5, "continuous mode or auxiliary mode");
   }
   else
   {
      /* fix interval mode  */
      startDate = pSenderParams->startDate;

      /* TODO: Check if end date is after nowTime */
      if (0 == pSenderParams->endDate) endDate = nowTime;
      else endDate = pSenderParams->endDate;
      DEBUG(5, "fix interval mode");
   }

   /* Parse index file(s). After parsing each station object
      should have assigned a list of clf files to process.
      If a DB connection is configured, clf files are
      searched in the DB only after start up. Keep in mind that
      no port information is saved in the DB, so result must not
      be the same when the clf files are read from DB */

   indexFilesParsed = catIdxDir(pConfigData->indexFilePath,
                                pConfigData->portNumber,
                                pSenderParams->pProdLine,
                                pSenderParams->nbrOfprodLines,
                                &startDate, endDate);

   if (indexFilesParsed < 0)
   {
      DEBUG2(5,"Search for clf files returns [%d]", indexFilesParsed);
      stopProgram(APP_NAME, "unable to parse Index files", "schedThread");
      /*@-mustfreefresh@*/
      return(NULL); /*for splint */
      /*@=mustfreefresh@*/
   }

   logWrite(LOG_NOTICE, "Scheduler started");

   /* Start or restart sending threads */
   startSenderThreads(pConfigData);

   /* Now all worker threads should start to send data. In continuous
      mode the scheduler thread will check in sleepTimeSched time
      intervals if new clf files are available */
   while (0 == endDate || endDate > time(NULL))
   {
      struct timespec sleeptime;
      int ii;

      /* No authentication or no control port configured */
      /*@-usedef@*/
      sleeptime.tv_sec = 1;
      sleeptime.tv_nsec = 0;
      /*@=usedef@*/

      /* sleep for a sleepTimeSched time interval */
      for (ii = (int)pSenderParams->sleepTimeSched;
           ii > 0 && SD_NONE == pConfigData->shutdownEnable; ii--)
      {
         /*@-usedef@*/
         (void) nanosleep(&sleeptime, NULL);
         /*@=usedef@*/
      }

      /* Check if a shutdown request is pending */
      if (SD_NONE != pConfigData->shutdownEnable) break;
      DEBUG(2, "Scheduler: check index file(s) for new entries");

      /* (Re-)parse index files. After parsing each station object
         will have an updated list of clf files to process.
         Modifications of the clf file list is protected by
         the stations mutex */
      indexFilesParsed = catIdxDir(pConfigData->indexFilePath,
                                   pConfigData->portNumber,
                                   pSenderParams->pProdLine,
                                   pSenderParams->nbrOfprodLines,
                                   &startDate, endDate);
      DEBUG2(5,"catIdxDir returns [%d]", indexFilesParsed);
      if (indexFilesParsed < 0)
      {
         shutdownApplication("unable to (re-)parse index files", NULL);
         break;
      }

      /* Restart sending threads */
      startSenderThreads(pConfigData);
   }

   /* Wait for all working threads to complete */
   JOIN_WORKER_THREADS(pConfigData->pHistory->firstStation);
   DEBUG(5, "+++ Scheduler thread terminating");

   /* Check if the main thread has a signal received before. Otherwise
      signal shutdown condition to main thread. In fixed interval mode
      no signal has been sent to the main thread. The SIGNAL should be
      sent only once.*/
   if (0 != pthread_mutex_lock(&(pConfigData->shutdownMutex)))
   {
      logError("can't set shutdown mutex", "schedThread");
   }

   if ((unsigned char)0 == pConfigData->shutdownEnable)
   {
      (void) pthread_kill(pConfigData->mainID, SIGUSR1);
   }

   if (0 != pthread_mutex_unlock(&(pConfigData->shutdownMutex)))
   {
      logError("can't unset shutdown mutex", "schedThread");
   }

#ifdef HAVE_AUTHENTICATION
   if (NULL != pControl)
   {
      int retVal;

      /* Wait for command thread to complete */
      DEBUG(9,"before joining command thread");
      retVal = pthread_join(comID, NULL);
      DEBUG2(9,"command thread joined [%d]", retVal);
   }
#endif /* HAVE_AUTHENTICATION */

   pthread_exit(NULL);
   /*@-mustfreefresh@*/
   return(NULL); /*for splint */
   /*@=mustfreefresh@*/
}


/*
 * @short Catch signals
 *
 * @description This function catches a signal and kills the main thread
 * @description with SIGKILL
 *
 * @param signal  signal
 *
 * @return None
 */

static void
catch_Signal(int signal)
{
   char comBuffer[LOGSIZE];

   /* Don't use getpid() on LINUX here!! */
   (void) snprintf(comBuffer, LOGSIZE, "catchSignal: TID=%lu, SIG=%d",
                   (unsigned long)pthread_self(), signal);
   DEBUG(2, comBuffer);

   (void) pthread_kill(pthread_self(), SIGKILL);
}


/**
 * @short Start required threads for CDsender
 *
 * @description This function does all the necessary stuff to start the
 * @description scheduler and monitor thread and waiting for user termination
 *
 * @param sigMask signal mask for sigwait
 * @param pConfigData   configuration data
 *
 * @return terminating signal number
 */

static int
startSThreads(sigset_t *sigMask, configParams_t *pConfigData)
{
   pthread_attr_t tAttr;     /* Thread attributes */
   struct sigaction sigInfo;
   pthread_t schedID;      /* ID schedule thread */
   pthread_t monID;        /* ID monitor thread */
#ifdef HAVE_AUTHENTICATION
   pthread_t certID;       /* ID certificate thread */
#endif /* HAVE_AUTHENTICATION */
   int sigVal,
       retVal;

   memset (&sigInfo, 0, sizeof (struct sigaction)); /* for splint */
   /* initialize attribute */
   if (0 !=  pthread_attr_init(&tAttr))
   {
      stopProgram(APP_NAME,"can't initialize attribute", NULL);
   }
   /* only PTHREAD_SCOPE_SYSTEM works on Linux! */
   if (0 !=  pthread_attr_setscope(&tAttr, PTHREAD_SCOPE_SYSTEM))
   {
      stopProgram(APP_NAME, "can't set scope", NULL);
   }

   /* start monitor thread if configured */
   if (0 != pConfigData->summaryRefreshTime &&
       0 != pthread_create(&monID, &tAttr, monitorThread, pConfigData))
   {
      stopProgram(APP_NAME, "can't start monitor thread", NULL);
   }

#ifdef HAVE_AUTHENTICATION
   /* start certificate thread if configured */
   if (0 != pConfigData->CRLcheckTime &&
       0 != pthread_create(&certID, &tAttr, certificateThread, pConfigData))
   {
      stopProgram(APP_NAME, "can't start certificate thread", NULL);
   }
#endif /* HAVE_AUTHENTICATION */

   /* starting of scheduler thread */
   if (0 != pthread_create(&schedID, &tAttr, schedThread, pConfigData))
   {
      stopProgram(APP_NAME, "can't start scheduler thread", NULL);
   }

   /* Block all signals in main thread */
   (void) sigfillset(sigMask);
   (void) pthread_sigmask(SIG_BLOCK, sigMask, NULL);

#if linux && HAVE_LIBPTHREAD
   /* (void) sigdelset(sigMask, SIGSEGV); */
#endif

   /* Wait until shutdown is signaled */
   sigVal = wait4shutdown(sigMask);

   /* Unblock all signals. This has no affect on LINUX if waiting
      on a pthread_join() call. Install signal handler */
   sigInfo.sa_handler = catch_Signal;
   sigInfo.sa_flags = 0;
   (void) sigaction(SIGINT, &sigInfo, NULL);
   (void) pthread_sigmask(SIG_UNBLOCK, sigMask, NULL);

#ifdef HAVE_AUTHENTICATION
   /* Stop certificate Thread */
   if (0 != pConfigData->CRLcheckTime)
   {
      /* It's save to cancel the certificate thread */
      /*@-usedef@*/
      if (pthread_cancel(certID) != 0)
      /*@=usedef@*/
      {
         logError("can't cancel certificate thread", "startSThreads");
      }
      /* Join the certificate thread
         Under Linux it could happen that joining the certificate thread
         is not possible when a segmentation fault has occurred. In this
         case the CD Sender will wait at this point until a signal
         is received */
      DEBUG(9, "Before pthread_join certificate");
      retVal = pthread_join(certID, NULL);
      DEBUG(9, "After pthread_join certificate");
      pConfigData->pHistory->nbrOfThreadsStopped++;
      DEBUG2(5,"Joining certificates thread returns [%d]", retVal);
      logWrite(LOG_NOTICE, "Certificate thread stopped");
   }
#endif /* HAVE_AUTHENTICATION */

   /* Set shutdown flag */
   if (SIGINT == sigVal) setShutdownFlag(SD_NORMAL);
   else setShutdownFlag(SD_FAST);

   /* Wait for scheduler thread to complete */
   DEBUG(9,"before joining scheduler thread");
   retVal = pthread_join(schedID, NULL);
   DEBUG2(9,"scheduler thread joined [%d]", retVal);

   /* Stop Monitor Thread. Don't use pthread_cancel() to stop monitor thread.
      That could lead to incomplete logging entries */
   if (pConfigData->summaryRefreshTime != 0)
   {
      /*@-usedef@*/
      stopMonitor(monID, pConfigData);
      /*@=usedef@*/
      retVal = pthread_join(monID, NULL);
      DEBUG2(5,"Joining monitor thread returns [%d]", retVal);
      logWrite(LOG_NOTICE,"Monitor thread stopped");
   }

   if (0 !=  pthread_attr_destroy(&tAttr))
   {
      logError("can't destroy attribute", NULL);
   }

   return(sigVal);
}

/**
 * @short Check argv vector for start and end date
 *
 * @description This function checks the argv vector for start and end date
 * @description or auxiliary station mode switch and puts the sender in fixed
 * @description interval mode if at least an end date is found, in auxiliary
 * @description station mode if the auxiliary switch is found or in
 * @description continuous mode otherwise
 *
 * @param pConfigData  Pointer to configuration object
 * @param argc         Number or CLI arguments
 * @param argv         Vector with CLI arguments
 *
 * @return nothing
 */

static void
checkArgs4Date(configParams_t *pConfigData, int argc, char *argv[])
{
   senderParams_t *pSenderParams;

   pSenderParams = (senderParams_t *)pConfigData->configExtension;

   if (argc > 2)
   {
      /* Check for auxiliary station mode */
      if (argc == 3 && 0 == strncmp(argv[2],"0/0/0 00:00:00",15))
      {
         /* Set auxiliary station operating mode */
         pSenderParams->senderMode |= OPER_MODE_AUX;
         DEBUG(2, "Auxiliary station mode");
      }
      else
      {
         /* Set start date */
         pSenderParams->startDate = ascDate2int(argv[2],
                                                "%4d/%2d/%2d %2d:%2d:%2d");
         if ((time_t)-1 == pSenderParams->startDate)
         {
            stopProgram(APP_NAME, "Invalid value in start date", NULL);
            return;
         }

         DEBUG2(5, "start date is %u", (unsigned int)pSenderParams->startDate);

         /* Set fixed interval operating mode */
         pSenderParams->senderMode |= OPER_MODE_FIXED;
      }

      if (argc == 4)
      {
         /* Set end date */
         pSenderParams->endDate = ascDate2int(argv[3],
                                              "%4d/%2d/%2d %2d:%2d:%2d");
         if ((time_t)-1 == pSenderParams->endDate)
         {
            stopProgram(APP_NAME, "Invalid value in end date", NULL);
            return;
         }
      }
   }
   else
   {
      /* Set continuous operating mode */
      pSenderParams->senderMode |= OPER_MODE_CONTINUOUS;
   }
}

/**
 * @short CD sender main function
 *
 * @description This function handles the CLI arguments and calls
 * @description the appropriate functions for data sending
 *
 * @param argc  Number or CLI arguments
 * @param argv  Vector with CLI arguments
 *
 * @return EXIT_FAILURE or EXIT_SUCCESS
 */

int
main (int argc, char *argv[])
{
   int	ii;
   char comBuffer[LOGSIZE + 1];
   char syslogID[48];
   char dateString[18];
   configParams_t *pConfigData;
   int sigVal;
   sigset_t  sigMask;        /* Signal mask value */

   /* Check if CDsender don't run under super-user ID (KT231003) */
   ii = (int)getuid();
   if (0 == ii)
   {
      /* If this would be allowed, the sender would be able
         to read *ALL* files in the file system with a proper
         constructed configuration file */
      printf("Please start the CD-Sender under a non super user ID\n");
      exit(EXIT_FAILURE);
   }

   /* Check command line argument(s) */
   if (argc < 2 || argc > 4)
   {
      fprintf(stderr, "CD-Sender %.10s (%.16s)\n", VERSION,
              CDlibVersion());
      fprintf(stderr, "usage: %s parameter_file [start date [end date]]\n",
              argv[0]);
      fprintf(stderr,
              "%s: %s config.par '2004/01/23 00:00:00' '2004/01/25 00:00:00'\n",
              "example", argv[0]);
      exit(EXIT_FAILURE);
   }

   /* get configuration data with default values */
   pConfigData = initConfigParams(APP_NAME, SENDER);

   /* Receiver port number is part of the syslog ID, but the configuration
      file has to be read first to get the configured port number. So
      a temporary syslog ID is used while reading the configuration file.
      Buffer is build with snprintf to avoid security holes! */
   (void) snprintf(syslogID, sizeof(syslogID), "ctbt.acquisition.cdsend");
   logInit(syslogID, LOG_INFO);

   /* Read the configuration file */
   (void) readConfigurationFile(APP_NAME, argv[1]);

   /* Add sender specific object to station structure */
   pConfigData->sizeOfAppExtension = sizeof(cdsend_t);

#ifdef EXPERIMENTAL
   /* Register application specific commands */
   pConfigData->appSpecificCtrl = processSenderCommands;
#endif

   /* Load sender specific config variables */
   readSenderConfigParams(APP_NAME, argv[1], pConfigData);

   /* Load data consumer file */
   readDCfile((senderParams_t *)pConfigData->configExtension, APP_NAME);

   /* Load production line file */
   readPLfile((senderParams_t *)pConfigData->configExtension, APP_NAME,
               NO_AUTO_CREATE);

   /* Check for fixed interval or auxiliary mode */
   checkArgs4Date(pConfigData, argc, argv);

   /* read further files if necessary and do some checks */
   processConfiguration(pConfigData, SENDER, APP_NAME);

   /* Control port number is now available, so the real syslog ID
      can be set */
   closelog();
   (void) snprintf(syslogID, sizeof(syslogID), "ctbt.acquisition.cdsend.%u",
                  (unsigned int)pConfigData->appControlPort);
   logInit(syslogID, LOG_INFO);

   getDate(dateString, sizeof(dateString));
   (void) snprintf(comBuffer, LOGSIZE,
                   "CDsender %.10s (%.16s) started at %s by User %d",
                   VERSION, CDlibVersion(), dateString, ii);
   logWrite(LOG_NOTICE, comBuffer);

   /* if verboseEnable is set, dump configuration to syslog */
   if (pConfigData->systemConf & VERBOSE_LOGGING)
   {
       dumpConfigurationParameter();
       dumpSenderConfigurationParameter(pConfigData);
   }

   /* The preferred way to handle signals is to block all signals in
      all threads and let the main thread waits for and processes
      signals. The following signals are used by the receiver:
      SIGHUP reserved for configuration reload
      SIGINT  orderly shutdown (CRTL C)
      SIGTERM reserved for immediate shutdown
      SIGUSR1 used internal by the receiver */
   (void) sigemptyset(&sigMask);

#if linux && HAVE_LIBPTHREAD
   /* unfortunately the preferred signal behavior will not work under the LINUX
      pthread implementation. If all signals are blocked, a synchronous signal
      is not seen in the main thread if the signal occurs in
      an other thread (don't mix occur with sent!!). This is not the
      POSIX behavior! If synchronous signals are not blocked in the main thread,
      a further blocked asynchronous signal must be send to wake up sigwait().
      And than the (before sent) synchronous signal is delivered which cause
      the receiver to be terminated immediately. Very strange behavior!!
      For LINUX only a couple of asynchronous signals are blocked for the worker
      threads. The main thread blocks all signals. This give the receiver a
      chance to handle synchronous signals as excepted. */
   (void) sigaddset(&sigMask, SIGHUP);
   (void) sigaddset(&sigMask, SIGINT);
   (void) sigaddset(&sigMask, SIGQUIT);
   (void) sigaddset(&sigMask, SIGTERM);
   (void) sigaddset(&sigMask, SIGUSR1);
#else /* linux && HAVE_LIBPTHREAD */
   (void) sigfillset(&sigMask);
#endif /* not linux && HAVE_LIBPTHREAD */

   (void) pthread_sigmask(SIG_BLOCK, &sigMask, NULL);
   pConfigData->mainID = pthread_self();

   /* Start required threads */
   sigVal = startSThreads(&sigMask, pConfigData);

   /* CD sender is shutting down */
   getDate(dateString, sizeof(dateString));
   (void) snprintf(comBuffer, LOGSIZE, "CDsender stopped at %s by signal %d",
                   dateString, sigVal );
   logWrite(LOG_NOTICE, comBuffer);
   closeLibCD();
   closelog();
   exit(EXIT_SUCCESS);
}
