
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Johannes Schreitl
 *
 * Revision 1.193 2015-04-21 gerald
 * Move duplicate code segments for channel filters
 * to libCD
 *
 * Revision 1.129 2014-07-28 gerald
 * Update comment
 *
 * Revision 1.128 2014-02-20 GK
 * Move function deleteWithThatFileObject() used by cd2wng
 * and cdsend to library. Fix gcc 4.7.2 warnings. Set ER_NOT_PARSED
 * when invalid CD1.1 frame header is detected
 *
 * Revision 1.127 2013-10-16 gerald
 * Fix warning on FreeBSD
 *
 * Revision 1.126 2013-07-17 gerald
 * Function header and other comments updated
 *
 * Revision 1.125 2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.124  2013-04-08 gerald
 * Fix frame acknowledgement when last frame sent is
 * not acknowledged. Fixes also sending to many frames
 * when restart and buggyAckNackFrames is OFF
 *
 * Revision 1.123  2012-06-27 gerald
 * Fix function comment
 *
 * Revision 1.122  2011-12-05 17:14:17  ctbto
 * Add possibility to switch on/off treatment of AckNack
 * frames in sender. Default is old behavior
 *
 * Revision 1.121  2011-09-20 14:03:14  ctbto
 * Remove DB access code from CD Sender
 *
 * Revision 1.120  2011-06-08 07:59:41  ctbto
 * Fix comment
 *
 * Revision 1.119  2011-03-02 15:27:11  gerald
 * Add support for EC signature size
 *
 * Revision 1.118  2011-02-10 10:12:37  gerald
 * Fix incorrect channel string count in sent CD1.1 frames when
 * a filter has been applied (CTH patch). Fix also incorrect return
 * value of filterFrame() when a filter has been applied
 *
 * Revision 1.117  2009-11-06 16:11:09  gerald
 * Add timely ordered sending mode to CD Sender
 *
 * Revision 1.116  2009-07-20 13:02:17  gerald
 * Move openIntRefList() from sender code to libCD
 * Add code for robust restart
 *
 * Revision 1.115  2009-06-09 08:45:47  gerald
 * Add comment how to reset bad frame destination
 *
 * Revision 1.114  2009-03-31 13:22:20  gerald
 * Fix splint warnings which occurred after upgrading
 * to new splint version
 *
 * Revision 1.113  2009-03-09 17:11:48  gerald
 * Mark socket as closed after closing the socket to avoid that
 * the socket is closed again in cleanUp()
 *
 * Revision 1.112  2009-01-19 17:31:31  gerald
 * Move saveIntRefList() and getSortLevel() to libcd and
 * fix missing shutdown in fixed interval mode
 *
 * Revision 1.111  2008-09-16 18:13:34  gerald
 * Fix bug which causes the sender to dump if the filter match
 * all channels, frame re-signing is switched on, the original
 * frame was not signed and the current frame buffer size is smaller
 * than the original frame size + size of a signature.
 * Stop sender if no memory is available for the temporary buffer
 * used by the channel filtering code
 *
 * Revision 1.110  2008-04-07 13:15:29  gerald
 * Fix bug PR-00446 (incorrect channel mask in sender generated
 * clf files when applying a filter)
 *
 * Revision 1.109  2007/09/13 15:52:19  gerald
 * Fix bug in auxiliary station mode which prevents sending
 * the same requested data twice
 *
 * Revision 1.108  2007/06/28 07:47:39  gerald
 * Save errno of last system error for logging in text
 * files (PR-0294)
 *
 * Revision 1.107  2007/05/10 12:32:35  gerald
 * Move code which connects to a remote port to libCD
 *
 * Revision 1.106  2007/03/05 16:10:17  gerald
 * Fix Sender crash when an AckNack frame is received which
 * exceeds the current frame buffer size (PR-0318)
 *
 * Revision 1.105  2007/02/19 14:31:29  gerald
 * Fix unaligned buffer for alert frames when compiled with
 * Solaris C compiler
 *
 * Revision 1.104  2007/02/12 10:17:05  gerald
 * Add virtual connect/disconnect (event) frame to clf file
 *
 * Revision 1.103  2007/01/24 09:09:53  gerald
 * Add computing and logging of send time span (rename
 * startRead to startRoW, endRead to endRoW, drop startWrite)
 *
 * Revision 1.102  2007/01/15 15:01:01  gerald
 * Channel name cache pointer has been replaced by a structure
 *
 * Revision 1.101  2007/01/11 15:44:44  gerald
 * Add channel name cache to parameter list when calling
 * parseAndCheckDFF() and parseAndCheckDataPayload()
 *
 * Revision 1.100  2007/01/09 12:14:40  gerald
 * Count number of successful connection requests
 *
 * Revision 1.99  2007/01/02 10:31:41  gerald
 * Change (re)start behavior of CD Sender; sort IRL only when
 * (re)starting the production line. After (re)start insert
 * new frames in the correct place
 *
 * Add an additional restart point if the connection has
 * broken or the data consumer has disconnect
 *
 * Revision 1.98  2006/11/02 13:25:14  gerald
 * Adjust debug level for some debug messages
 *
 * Revision 1.97  2006/10/11 16:07:37  gerald
 * Fix bug in sender that prevents the sender from sending
 * new arrived frames first when sending in LIFO mode
 *
 * Revision 1.96  2006/10/04 12:44:34  gerald
 * Fix sending too many data format flush alert frames when
 * a filter is applied (PR-0300). Do not write or read restart
 * information in auxiliary mode. Stop sending data and change
 * to stopped state when the first not requested frame is seen
 * in auxiliary mode
 *
 * Revision 1.95  2006/09/06 12:40:27  gerald
 * Use sender specific structure member from the application
 * extension structure for processing mode and operating mode.
 * Rework main sending loop to support true LIFO mode
 * after long outages and when new clf files are created
 * when backfilling. Check for new frames when sending
 * in FIFO or LIFO mode in sendIrlData() also in newly
 * created clf files
 *
 * Revision 1.94  2006/07/21 14:23:34  gerald
 * Check for NULL is necessary to wait for retry when no
 * clf file is available, e. g. the last N files are read from
 * the restart file, but no corresponding clf file exists
 *
 * Revision 1.93  2006/07/20 13:19:55  gerald
 * Update comments, make use of UNSORTED define and fix bug in
 * fixed interval mode which causes too early shutdown of worker
 * thread when the connection has been closed by peer.
 * Check for new frames in current clf file when sending frames
 * to fulfill FIFO requirements (No new clf files are checked!)
 * Parse all files in the clf file list before try to connect.
 * This fixes incorrect sorting when more than one clf file for
 * the current day exists and no restart information is found
 *
 * Revision 1.92  2006/07/04 09:24:00  gerald
 * Comment update, make use of UNSORTED define and fix bug in
 * fixed interval mode which causes too early shutdown of worker
 * thread when the connection has been closed by peer.
 *
 * Revision 1.91  2006/04/24 14:29:59  gerald
 * Fix core dump in sender when a filter is applied on a CD-1.1 frame,
 * the original frame has no frame signature, frame signing is
 * switched on and the filter string matches all channels in the frame
 *
 * Revision 1.90  2006/04/18 09:23:15  gerald
 * Rename defines for event types to avoid inadvertently use of
 * E_INTERNAL and fix typos
 *
 * Revision 1.89  2006/04/05 13:21:59  gerald
 * Rename parseFrame() to parseFrame2sent(), because libCD contains now
 * also a function parseFrame()
 *
 * Revision 1.88  2006/04/04 08:36:39  gerald
 * Debug messages improved
 *
 * Revision 1.87  2006/03/30 13:22:33  gerald
 * Send an alert frame for flushing format buffer before sending a DFF
 * if this is not the first DFF sent over this connection
 *
 * Revision 1.86  2006/03/28 09:11:57  gerald
 * Rename HAVE_LIBGODBC to HAVE_DB_ACCESS. Improve debug messages
 * and comments. Move unlinkClfElement(), haveDataToSend(),
 * checkClfFileObj() and getUnsentFrames() to libCD. These functions
 * are needed for the new Data Parsing Programming Interface.
 *
 * Revision 1.85  2006/03/20 14:17:21  gerald
 * Fix typos and improve comments. Fix missing close of socket
 * when no option response frame is received during the CD1.1
 * connection establishment process. Improve error and debug messages.
 * Add internal reference list header to sender structure.
 * Fix memory leak in sender when stop commands are received
 *
 * Revision 1.84  2006/03/09 15:10:23  gerald
 * Apply a filter only if the frame has no parsing errors
 *
 * Revision 1.83  2006/03/07 12:18:40  gerald
 * Continue work for Data Parsing Programming Interface. Rework
 * parseClfFile() and other functions to become more independent
 * from CDsender
 *
 * Revision 1.82  2006/02/24 11:04:50  gerald
 * Truncate restart file and improve logging messages
 *
 * Revision 1.81  2006/02/14 13:02:55  gerald
 * Fix typos
 *
 * Revision 1.80  2006/02/14 10:11:27  gerald
 * Make use of sender specific config structure
 *
 * Revision 1.79  2006/02/03 10:03:26  gerald
 * Fix wrong receiving time of alert frames
 * Fix wrong receiving start time of port assignment frames
 *
 * Revision 1.78  2005/11/25 13:41:42  gerald
 * Adjust debug levels of some messages, add more comment and
 * log if a restart point is found to syslog with LOG_NOTICE
 *
 * Revision 1.77  2005/11/11 14:42:29  gerald
 * Change filter string when a new filter request is pending
 *
 * Revision 1.76  2005/10/14 10:53:28  gerald
 * When no restart file is available for a production line send
 * only clf files created at the current day. Before that change
 * all clf files found between current day and lookBackTimeSched
 * were sent
 *
 * Revision 1.75  2005/10/05 12:19:09  gerald
 * Change key loading behavior in sender. Now keys are loaded before the
 * handling thread is started or if it is re-started. This avoids loading
 * of certificates in different threads
 *
 * Revision 1.74  2005/09/14 13:43:08  gerald
 * Unlink the frame from the sending history when it is re-scheduled
 * for sending. This should prevent that the sending list becomes
 * corrupted when the sending history contains old and new links and
 * the IRL is re-sorted. This can only happen when the data consumer
 * disconnects before all frames scheduled for sending, are sent
 *
 * Revision 1.73  2005/09/08 11:22:23  gerald
 * Fix some comments and make CD1 the implicit station type
 *
 * Revision 1.72  2005/09/07 11:34:17  gerald
 * Fix too high CPU time consumption when frames are loaded from
 * a restart file, but no clf file is available. Fix core dump
 * when all frames in the IRL are deleted by deleteWithThatFileObject()
 *
 * Revision 1.71  2005/08/05 13:22:43  gerald
 * Allocate portinfo structure when station object is allocated.
 *
 * Revision 1.70  2005/07/12 12:20:06  gerald
 * Move sender specific variables from station structure into
 * application extension structure. Shutdown sender when an
 * unexpected null pointer error occur instead of logging an
 * error message
 *
 * Revision 1.69  2005/07/07 09:55:40  gerald
 * Shutdown sender when an internal error occurs during
 * startup of worker thread. Enter state connecting when a
 * connection request is sent and change state to waiting
 * when waiting for response
 *
 * Revision 1.68  2005/07/07 09:10:32  gerald
 * Check for stop request in shutdown flag. This modification should
 * also speed up fast shutdown requests
 *
 * Revision 1.67  2005/07/01 12:49:52  gerald
 * Resend last N frames after re-connection
 *
 * Revision 1.66  2005/06/16 08:22:18  gerald
 * Set state to WAIT2SEND when unconnected. Wait to send
 * is the state when the sender has no connection to a data consumer but
 * a thread is started for that production line. This state is necessary to
 * determine if a thread for a given production line is really running.
 * This fixes also the ungraceful termination of sending threads when a
 * shutdown is initiated and the sending thread was in unconnected state.
 * Save unsent or un-acknowledged frames sent before the restart point and
 * last frame sent to the restart file. Add frames saved in the restart
 * file when reading the restart file. Check frame sent count and set frames
 * to acknowledged when frame sent count is exceeded. This fixes the
 * go back N frames feature when re-start is done and when sorted in
 * LIFO or FIFO mode
 *
 * Revision 1.65  2005/06/06 12:25:05  gerald
 * Sent state to "sent, but not acknowledged" when a frame has been sent.
 * Add more info to restart file when sending sorted by time (frames which
 * have not been sent or acknowledged; This information is not used in
 * the current release when reading the restart file)
 *
 * Revision 1.64  2005/05/20 09:05:31  gerald
 * Fix sending too many frames during restart when sender was in
 * LIFO or FIFO mode
 *
 * Revision 1.63  2005/05/17 15:23:25  gerald
 * Make use of new error/warning statistics structure
 *
 * Revision 1.62  2005/05/11 09:36:08  gerald
 * Call updateStatistics() instead of updateStationStatistics() and
 * pass new statistics structure instead of whole station structure
 * Fix a problem with the channel filtering. Previously the filter
 * for the first station in the production line file was applied
 * to all the stations in the production line file
 *
 * Revision 1.61  2005/04/29 11:03:14  gerald
 * Log restart info to syslog. Change Sender behavior: do not send DF
 * or DFF where the size in clf file is shorter than size seen in the frame.
 * Move retryWait() to libCD. Fix sporadic missing alert frame when the
 * sender shuts down
 *
 * Revision 1.60  2005/04/11 13:11:10  gerald
 * Fix unexpected and immediate sender termination
 * (message: "... CDsender initialization stopped")
 * when connection request is reset by peer
 *
 * Revision 1.59  2005/03/07 12:55:15  gerald
 * Test if a data request is pending after a frame is sent. If a request
 * is pending, move frames with requested nominal times to top of IRL.
 *
 * Revision 1.58  2005/02/17 15:17:24  gerald
 * Remove E_SIZE event. E_FRAMESIZE is the same event
 *
 * Revision 1.57  2005/02/08 10:59:07  gerald
 * Remove unused variable in checkEvent()
 *
 * Revision 1.56  2005/01/17 12:27:23  gerald
 * Fix type
 *
 * Revision 1.55  2004/12/14 13:35:48  gerald
 * Fix missing site filter and re-work filter code
 *
 * Revision 1.54  2004/11/26 12:07:55  gerald
 * Fix missing NIL at end of CD-1.1 channel name when filtering
 *
 * Revision 1.53  2004/11/12 14:30:45  gerald
 * Add channel filtering
 *
 * Revision 1.52  2004/11/08 17:16:47  gerald
 * Fix endless loop during re-start when a clf file newer than the
 * re-start file exists. Fix sorting only frames within the current
 * clf file when no re-start file exists.  Fix "unexpected NULL
 * pointer" error when no data has been sent before shutdown
 *
 * Revision 1.51  2004/10/27 12:22:20  gerald
 * Fix compilation warnings when compiled on Solaris.
 * Fix writing a restart file every time a frame is sent when
 * refListStoreInterval is set to zero
 *
 * Revision 1.50  2004/10/07 17:43:06  gerald
 * Fix missing frames after re-start when clf file has grown but
 * no new clf file has been created. Add FIFO/LIFO sorting before
 * sending frames from the internal reference list. Change re-start
 * file format to support sorting. Fix slow endless loop when binary
 * file is moved or removed during sending (untested!!)
 *
 * Revision 1.49  2004/09/27 12:09:39  gerald
 * Enhance sendIrlData() to allow sending when frames not sorted by
 * arrival time (first patch for sending sorted by nominal time)
 *
 * Revision 1.48  2004/09/22 15:32:23  gerald
 * Fix wrong timeouts when sending Option Request and
 * receiving Option Response
 *
 * Revision 1.47  2004/09/16 15:22:16  gerald
 * Move readCDframe() and adjustBufferSize() to libCD. Add
 * connection state to sendFrame(). Fix wrong connection state
 * when logging Option Request/Response Frame
 *
 * Revision 1.46  2004/09/13 12:27:15  gerald
 * Fix swapped parameters in isBadEvent()
 *
 * Revision 1.45  2004/09/10 11:19:21  gerald
 * Write debug messages to stderr instead of stdout, use station
 * objects frame buffer instead of own one. Call updateStatisctics()
 * when a frame is sent/received
 *
 * Revision 1.44  2004/09/03 11:57:21  gerald
 * Resend last N frames when restart (default is 10)
 *
 * Revision 1.43  2004/09/02 08:20:05  gerald
 * Add info about the corresponding DFF to internal reference
 * list and drop DFF frames from the list. Delete sent frames
 * not before a re-start point is written. Add more function
 * comments
 *
 * Revision 1.42  2004/08/23 10:51:13  gerald
 * Set frame type and size/auth offset for integrity check. Set 'not
 * parsed flag' when sending Data and Data Format Frames. Fix
 * sometimes wrong direction (incoming/outgoing) in log files.
 * Fix logging of sent Alert Frames. Set event in readCDFrame frame
 *
 * Revision 1.41  2004/07/07 14:12:27  gerald
 * Use cleanUp library function when stopping a worker thread instead
 * of doing clean up manually
 *
 * Revision 1.40  2004/07/06 14:41:30  gerald
 * Set nominal time info in DFinfo structure to have nominal
 * time available in logfiles
 *
 * Revision 1.39  2004/06/30 11:08:05  gerald
 * Clean up unused config param lockBackTimeWork and fix
 * wrong waiting time for re-connect in CDsender
 *
 * Revision 1.38  2004/06/30 08:27:33  gerald
 * Deactivate read retries for bin and clf file in standard
 * configuration. Configure with -DMAX_READ_RETRIES=<number>
 * to activate. Start work on read retries for clf file
 *
 * Revision 1.37  2004/06/28 16:13:25  gerald
 * Fix missing time for CRF and ORF frames in sender log files
 * Check if an Alert Frame is received before sending data
 * Check if an Alert Frame is received when an unexpected close
 * error is reported from writeWithtimeout before reporting
 * an error
 *
 * Revision 1.36  2004/06/21 09:54:47  gerald
 * Add wait time argument to retryWait(). Try to re-read
 * binary file some times when end of file is reached
 * unexpected (that seems to be a NFS problem under SOLARIS)
 *
 * Revision 1.35  2004/05/28 18:11:19  gerald
 * Add logging of incoming data
 *
 * Revision 1.34  2004/05/14 16:46:48  gerald
 * Add creation of log files to the sender. Incoming data in
 * connected state is not logged yet
 *
 * Revision 1.33  2004/04/16 10:51:17  gerald
 * Fix some splint warnings
 *
 * Revision 1.32  2004/04/06 14:49:43  gerald
 * Fix endless loop when last clf file does not exist
 *
 * Revision 1.31  2004/04/05 14:07:50  gerald
 * Start work on sending CD1.1 data
 *
 * Revision 1.30  2004/03/26 08:41:55  gerald
 * Fix wrong internal reference list file when in catch up or
 * fixed interval mode and a connection is closed or the sender is
 * stopped before the whole clf file is sent. This makes a format
 * change of the internal reference list file necessary
 *
 * Revision 1.29  2004/03/23 15:37:54  gerald
 * Fix unnecessary connection establishment when a new Alert Frame is
 * found in the clf file. Fix core dump (on Solaris) or endless loop
 * (on Linux) when a clf file can not be loaded
 *
 * Revision 1.28  2004/03/05 14:02:29  gerald
 * Add re-connect feature and improve CPU time consumption
 * Fix some re-start bugs (missing DFF or does not connect)
 * Some code clean-ups
 *
 * Revision 1.27  2004/02/25 12:59:42  gerald
 * Add re-start feature (save and restore internal reference list)
 * Improved debug and logging messages
 *
 * Revision 1.26  2004/02/13 15:03:38  gerald
 * Check for received alert frames, update comments, fix
 * splint warnings, add CD conform logging, improve debug
 * messages and wait for incoming data instead of wait
 * for broadcast from scheduler. Use station specific lock
 * instead of global lock when updating clf file list
 *
 * Revision 1.25  2004/02/03 10:29:09  gerald
 * Remove trailing blanks
 *
 * Revision 1.24  2004/01/19 16:18:28  karol
 * Simple continuous mode.
 *
 * Revision 1.23  2003/12/09 10:54:16  gerald
 * Rename some structure members to make them more generic useable
 *
 * Revision 1.22  2003/11/28 10:37:05  gerald
 * Clean up Cdsender members in station structure
 *
 * Revision 1.21  2003/11/27 10:44:03  gerald
 * Drop unnecessary mutex locks. Fix core dump when
 * frame size > BUFSIZE
 *
 * Revision 1.20  2003/11/20 13:11:10  gerald
 * Use station structure in CDsender and use stopProgram() instead of
 * stopSender().  Use more new functions from libCD.
 *
 * Revision 1.19  2003/11/20 08:32:56  karol
 * Remove watching thread, spawn ow schedule thread.
 *
 * Revision 1.18  2003/11/14 10:36:21  karol
 * Common structure with cdreceiver (prodLineParam struct is replaced
 * with station structure).
 *
 * Revision 1.17  2003/11/07 11:01:58  gerald
 * Use common configuration structure and defines
 *
 * Revision 1.16  2003/10/28 10:45:11  karol
 * Remove bug
 *
 * Revision 1.15  2003/10/28 09:27:36  gerald
 * Use CD library function for connect to CD receiver
 * Use connection request timeout value for connection request
 *
 * Revision 1.14  2003/05/21 13:05:38  johannes
 * minor fix to avoid compiler warning in Linux
 *
 * Revision 1.13  2003/05/21 13:02:47  johannes
 * Handling for SIGINT after each sent frame added
 *
 * Revision 1.12  2003/05/21 11:55:12  johannes
 * Handling for SIGINT added
 *
 * Revision 1.11  2003/05/20 07:09:30  johannes
 * minor corrections for displays
 *
 * Revision 1.10  2003/05/20 06:40:43  johannes
 * check for getsockopt: sockValue added
 *
 * Revision 1.9  2003/05/19 13:00:25  johannes
 * send data in frame borders
 *
 * Revision 1.8  2003/05/14 13:35:10  gerald
 * Add some checks
 *
 * Revision 1.7  2003/05/14 12:48:04  johannes
 * fix buffer overflow
 *
 * Revision 1.6  2003/05/14 10:00:20  gerald
 * Fix architecture depended bugs
 *
 * Revision 1.5  2003/05/13 07:13:20  johannes
 * fix for Linux asm/poll.h
 *
 * Revision 1.3  2003/05/12 12:32:38  johannes
 * multithreading for sending data added
 *
 * Revision 1.2  2003/05/08 07:03:20  johannes
 * changes for prototype sending CD1 frames
 *
 * Revision 1.1  2003/05/07 06:46:52  johannes
 * first issue
 *
 *
 */

#include <config.h>
 #ifdef S_SPLINT_S
#include "../../splint_hints.h"
#endif /* S_SPLINT_S */
#include <errno.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <netinet/in.h>  /* Need this to parse code with splint */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "cdsender.h"
#include "prepare.h"
#include "schedule.h"
#include "sending.h"
#include "misc.h"
#include "frames.h"
#include "events.h"
#include "readwrite.h"
#include "logfiles.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif


/**
 * @short Parse and check Port Assignment Frame
 *
 * @description This function parses and checks a received
 * @description Port Assignment Frame and places the parsing
 * @description result in the remote port info fields of a
 * @description portmap structure
 *
 * @param buffer        Received frame data
 * @param portInfo      Pointer to portinfo structure
 *
 * @return 0 for success , -1 on error
 */

static int
parsePAF(char *buffer, /*@out@*/portMap_t *portInfo)
{
   split_t result;
   unsigned int fields;

   /* init buffer */
   memset(portInfo->remoteIPaddr, 0, sizeof(portInfo->remoteIPaddr));
   /* init structure */
   memset(&result, 0, sizeof( result));
   fields = splitString(buffer, strlen(buffer), '/', &result);

   if (fields != 2 || result.lenString[0] > sizeof(portInfo->remoteIPaddr))
   {
      return (-1);
   }

   /* Copy IP address from received frame into port structure */
   strncpy(portInfo->remoteIPaddr, result.pString[0], result.lenString[0]);

   /* Check IP address */
   if ((in_addr_t)-1 == inet_addr(portInfo->remoteIPaddr)) return(-1);
   /* Copy port into port structure */
   portInfo->remotePort = (uint16_t)atoi(result.pString[1]);

   return (0);
}

/**
 * @short Set CD1.1 frame header and sign frame
 *
 * @description This function sets a CD1.1 frame header
 * @description and signs the frame when a private key is available
 *
 * @param pStation        pointer to station object
 * @param buffer          frame buffer
 * @param trailerOffset   offset of trailer in frame
 * @param actEvent        event object
 *
 * @return size of the frame to sent on success or -1 on error
 */

static int
setTrailerWithSig(station_t *pStation, char *buffer, uint32_t trailerOffset,
                   CDevent_t *actEvent)
{
   int sendSize;

#ifdef HAVE_AUTHENTICATION

   /* TODO: merge setTrailerWithSig and prepareTrailer */
   if (NULL == pStation->pConfigData->privateKey)
   {
      sendSize = (int)setTrailer(buffer, trailerOffset, 0, NULL, 0);
   }
   else
   {
      char signature[SZ_SIGNATURE_MAX];
      size_t sigLen = sizeof(signature);

      /* Frame authentication is active */
      if (signData(pStation->pConfigData->privateKey, buffer,
                   (size_t)trailerOffset, signature, &sigLen))
      {
         /* something unexpected has happen during signature
            verification. Shutdown application?? */
         logError("Unexpected error while signing data",
                  "processRequests");

         pStation->whenUnconnected.systemFailures++;
         pStation->lastUnsuccessfulCR = pStation->startRoW.tv_sec;
         logCDevent2(ET_L_ERROR, E_OOPS, actEvent, NO_INFO);

         if (pStation->systemConf & STRICT_MODE_ENABLE)
         {
            return(-1);
         }
         /* Send unsigned frame */
         sendSize = (int)setTrailer(buffer, trailerOffset, 0, NULL, 0);
      }
      else
      {
         sendSize = (int)setTrailer(buffer, trailerOffset,
                               pStation->pConfigData->privateKeyID,
                               signature, sigLen);
      }
   }
#else

   sendSize = (int)setTrailer(buffer, trailerOffset, 0, NULL, 0);

#endif /* HAVE_AUTHENTICATION */
   return(sendSize);
}

/**
 * @short Send Alert Frame
 *
 * @description This function sends an CD1_1 or CD1 Alert frame depending
 * @description on the station type
 *
 * @param pStation        pointer to station object
 * @param actEvent        event object
 *
 * @return nothing
 */

static void
sendAF(station_t *pStation, CDevent_t *actEvent)
{
   /*@dependent@*/ char *buffer = pStation->frameBuffer;
   uint16_t event;
   int sendSize;


   /* Implicit default is CD_1 */
   if (CD1 == pStation->type)
   {
      /* Send AF for connection termination */
      setCD1AF(buffer, TERMINATE);
      sendSize = SZ_CD1_ALERT;
   }
   else
   {
      /* Set Header and payload for CD1.1 alert frame */
      setHeader(buffer, ALERT, (uint32_t) 40,
                pStation->pConfigData->frameCreator,
                /*@-compdef@*//*@-nullderef@*/
                pStation->stationName,
                /*@=compdef@*//*@=nullderef@*/
                (uint64_t) 0);
      *((uint32_t *)&buffer[SZ_HDR]) = 0; /* message size */
      sendSize = setTrailerWithSig(pStation, buffer, (uint32_t)40, actEvent);

      if (-1 == sendSize)
      {
         /* Unexpected error while signing data,
            event is logged in the called function */
         return;
      }
   }

   /* Send frame */
   ST_DEBUG(2, "Send AF to close connection");

   /* statistic is updated and data is saved in sendFrame */
   pStation->checkPass = 0;
   event = sendFrame(pStation, buffer, sendSize, S_SEND_ALERT, GOOD_ALERT, 1);
   if (NO_ERROR != event)
   {
      if (E_SYSTEM == event) pStation->savedSystemErr = (uint32_t)errno;
      logCDevent2(ET_L_ERROR, event, actEvent, (uint32_t)errno);
   }

   /* No error is reported to the calling function, because
      connection is closed in all cases */
}

/**
 * @short Check if data is available
 *
 * @description This function checks if data is available
 * @description for a given socket
 *
 * @param sd        socked descriptor
 *
 * @return 1 if data is available, 0 otherwise
 */

static int
dataIsAvailable(int sd)
{
   int bytesRead;
   int cc = 0;


   /* Check if data could be read from socket */
   bytesRead = (int)recv(sd, &cc, 1, MSG_PEEK);
   DEBUG3(9, "Could read [%d][%d] bytes", bytesRead, errno);

   /* Error treatment is done in the sending part */
   if (1 == bytesRead) return(1);
   else return (0);
}

/**
 * @short Read data from socket
 *
 * @description This function tries to read a CD frame
 * @description from a given socket and stores the received data
 * @description in several files depending on the configuration.
 * @description Station has to be in connected state
 *
 * @param pStation             station object
 * @param actEvent             event object
 *
 * @return NO_ERROR or event
 */


static uint16_t
readData(station_t *pStation, CDevent_t *actEvent)
{
   rwFlags_t rflags;
   int len;
   char clfType;
   uint16_t event = NO_ERROR;

   /* Initialize receiving flags */
   rflags.sockFlags = 0;

   if (CD1 == pStation->type)
   {
      rflags.timeout = pStation->pConfigData->timeOut;
      rflags.resetDataTimeout = pStation->pConfigData->timeOut;
   }
   else
   {
      rflags.timeout = pStation->pConfigData->timeOut11;
      rflags.resetDataTimeout = pStation->pConfigData->timeOut11;
   }

   rflags.shutdownFlag = &(pStation->shutdownEnable);

   /* Try to read a data frame */
   /*@-compdef@*/
   len = readCDframe(pStation, actEvent, &rflags, &event, &clfType);
   /*@=compdef@*/
   if (len > 0)
   {
      /* Update statistics and save data */
      updateStatistics(&(pStation->framesReceived), len, NO, clfType);
      if (saveData(pStation, pStation->frameBuffer, len, 0, event, NO, clfType))
      {
         /* SaveData will shutdown the application if data
            could not be saved */
         pStation->whenConnected.systemFailures++;
         return(E_SAVE_DATA);
      }
   }

   ST_DEBUG3(2, "readCDframe returns [%d] [%c]", (int)event, clfType);
   return(event);
}

/**
 * @short              parse frame to sent
 *
 * @description        This function parses a frame which should be sent
 *
 * @param buffer       pointer to buffer containing the data frame
 * @param pStation     station object
 * @param clfType      type of frame to process
 * @param size         buffer size
 *
 * @return 0 on success or -1 on error
 */

static uint16_t
checkIfFrameTooShort(station_t *pStation, char clfType,
            uint32_t size)
{
   uint32_t frameLen;
   char *buff = pStation->frameBuffer;

   /* Do some consistency checks */
   if (NULL == pStation->DFinfo)
   {
      shutdownApplication("Unexpected NULL pointer (checkIfFrameTooShort)",
                          pStation->conID);
      return (E_INTERNAL);
   }

   /* Parse the frame to set Data Frame info structure */
   if (CD1_1 == pStation->type)
   {
      /* CD 1.1 data frame */
      uint32_t ii;

      /* parse header */
      (void)parseHeader(buff, &(pStation->header));
      if (pStation->header.frameType != DATA )
      {
         /* Clf file is out of sync with binary file. Best approach in
            this case is to send the frames and hope that the receiving
            software has a good re-acquire mode */
         logWarning("Frame is not a data frame (out of sync?)",
                    pStation->conID );
         return(NO_ERROR);
      }

      /* Compute the frame length using information stored in the buffer */
      frameLen = pStation->header.trailerOffset + 8 + 8;

      /* check if frame has a signature */
      if (size > frameLen)
      {
         memcpy(&ii, &buff[frameLen - 12], sizeof (ii));
         frameLen += ntohl(ii);
      }
   }
   else if (CD1 == pStation->type)
   {
      uint32_t frameType = ntohl(*((uint32_t *)buff));

      frameLen = ntohl(*((uint32_t *)&buff[4]));

      /* Check if frame is a DFF. NbrOfChannels must be modified in this case */
      if (GOOD_FORMAT == clfType)
      {
         pStation->checkPass = NO_ERROR;
         if (CD1_DATA_FORMAT != frameType)
         {
            /* Clf file is out of sync with binary file. Best approach in
               this case is to send the frames and hope that the receiving
               software has a good re-acquire mode */
            logWarning("Frame is not a data format frame (out of sync?)",
                       pStation->conID );
            return(NO_ERROR);
         }
      }
      else
      {
         if (CD1_DATA != frameType)
         {
            /* Clf file is out of sync with binary file. Best approach in
               this case is to send the frames and hope that the receiving
               software has a good re-acquire mode */
            logWarning("Frame is not a CD1 data frame (out of sync?)",
                       pStation->conID );
            return(NO_ERROR);
         }
      }
   }
   else
   {
      /* Shutdown the application */
      shutdownApplication("Unexpected station type", pStation->conID);
      return(E_INTERNAL);
   }

   /* Check if the frame size of the current frame is the same as
      the frame size logged in the clf file. Differences in frame
      size can be caused by a timeout during receiving of the frame */
   if (size < frameLen)
   {
      /* Policy is to drop the connection after sending that frame
         without sending an alert frame before */
      return (E_FRAME_TOO_SHORT);
   }
   return(NO_ERROR);
}

/**
 * @short              parse frame to sent
 *
 * @description        This function parses a frame which should be sent
 *
 * @param buff         pointer to buffer containing the data frame
 * @param pStation     station object
 * @param clfType      type of frame to process
 * @param size         buffer size
 *
 * @return 0 on success or -1 on error
 */

static uint16_t
parseFrame2sent(char *buff, station_t *pStation, char clfType,
                uint32_t size)
{
   uint32_t strictMode;

   /* Do some consistency checks */
   if (NULL == pStation->DFinfo)
   {
      shutdownApplication("Unexpected NULL pointer, function parseFrame2sent",
                          pStation->conID);
      return (E_INTERNAL);
   }

   strictMode = pStation->systemConf & STRICT_MODE_ENABLE;

   /* Parse the frame to set Data Frame info structure */
   if (CD1_1 == pStation->type)
   {
      /* CD 1.1 data frame */
      time_t   nowTime;

      nowTime = time(NULL);

      /* parse header */
      (void)parseHeader(buff, &(pStation->header));
      if (pStation->header.frameType != DATA )
      {
         /* Clf file is out of sync with binary file. Best approach in
            this case is to send the frames and hope that the receiving
            software has a good re-acquire mode */
         logWarning("Frame is not a data frame (out of sync?)(1)",
                    pStation->conID);
         /* Set not parsed error to prevent that the frame is
            filtered or the frame signature is checked later */
         pStation->checkPass |= ER_NOT_PARSED;
         return(NO_ERROR);
      }

      /* Parse and check data frame payload */
      pStation->checkPass |=  parseAndCheckDataPayload(
                                            pStation->DFinfo,
                                            strictMode, buff,
                                            pStation->pConfigData->maxDeviation,
                                            size, (long)nowTime,
                                            &(pStation->chCache));

      ST_DEBUG2(2, "DF_1_1 checkPass is: %d", (int)pStation->checkPass);
   }
   else if (CD1 == pStation->type)
   {
      if (GOOD_FORMAT == clfType)
      {
         pStation->checkPass = NO_ERROR;
         /* Check and parse frame. No events are logged in this function */
         pStation->checkPass |= parseAndCheckDFF(pStation->DFinfo, strictMode,
                                                 buff, &(pStation->chCache));

         ST_DEBUG2(2, "DFF checkPass is: %d", (int)pStation->checkPass);
         /* TODO: recheck the frame type here */
      }
      else
      {
         time_t   nowTime;

         /* CD 1.0 data frame */
         nowTime = time(NULL);

         /* Check and parse frame */
         pStation->checkPass |= parseAndCheckDF(pStation->DFinfo,
                                           strictMode, buff,
                                           pStation->pConfigData->maxDeviation,
                                           (long)nowTime,
                                           &(pStation->chCache));

         ST_DEBUG2(2, "DF checkPass is: %d", (int)pStation->checkPass);
         /* TODO: recheck the frame type here */
      }
   }
   else
   {
      /* Shutdown the application */
      shutdownApplication("Unexpected station type, function parseFrame2sent",
                          pStation->conID);
      return(E_INTERNAL);
   }
   return(NO_ERROR);
}

/**
 * @short              filter frame
 *
 * @description        This function extract a subset of channels from a
 * @description        data frame, according to the filter
 * @description        specified in the production line file
 *
 * @param buffer       pointer to buffer containing the data frame
 * @param pStation     station object
 * @param clfType      type of frame to process
 * @param size         buffer size
 * @param actEvent     event object
 *
 * @return number of channels to send
 */

static uint32_t
filterFrame(station_t *pStation, char clfType,
            uint32_t *size, CDevent_t *actEvent)
{
   char  sendChannel[MAX_CHANNELS]; /* Marker for data to sent */
   uint32_t  ii = 0;
   CDdata_t *DFinfo = pStation->DFinfo;
   uint32_t nbrOfChannels = 0;
   uint32_t nbrOfFilteredChannels;
   uint32_t tmpInt = 0;
   uint32_t frameLength = 0;
   cdsend_t *pCdSend;
   char *buff = pStation->frameBuffer;


   /* Do some consistency checks */
   if (NULL == DFinfo || NULL == pStation->appExtension ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pRefListHead ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine)
   {
      shutdownApplication("Unexpected NULL pointer (filterFrame)",
                          pStation->conID);
      return (0);
   }

   /* Access application extension data */
   pCdSend = ((cdsend_t *)(pStation->appExtension));
   memset(sendChannel, 0, sizeof(sendChannel)); /* for splint */

   /* Check if site/channel filters are set
      and if new filter request command is pending */
   if (0 == check4FilterProvided(pCdSend))
   {
      /* Frame is not parsed yet */
      pStation->checkPass |= ER_NOT_PARSED;
      return(1);
   }

   /* Parse the frame now to see if a filter rule matches */
   if (NO_ERROR != parseFrame2sent(buff, pStation, clfType, *size))
   {
      /* An error has occurred in parseFrame2sent and a shutdown
         is pending now */
      return (0);
   }

   /* Check if the frame is valid. Don't apply the filter
      if a parsing error has occurred or the sender might crash!
      Sent the frame as it is instead */
   if (HAS_DF_ERRORS(pStation->checkPass))
   {
      ST_DEBUG(5, "invalid frame, no filter applied");
      return(1);
   }

   /* Check if at least one channel matches the filter rules.
      We also send frames with warnings or errors, but only
      complete parsed channels */
   if (GOOD_FORMAT == clfType) nbrOfChannels = DFinfo->parsedDFFchannels;
   else nbrOfChannels = DFinfo->parsedChannels;

   ST_DEBUG2(2, "nbrOfChannels is: %d", (int)nbrOfChannels);

   /* Get filtered channel mask and number of filtered channels */
   nbrOfFilteredChannels = getFilteredChannelMask(pCdSend,
                                                  nbrOfChannels,
                                                  DFinfo,
                                                  sendChannel);

   /* Check if at least one channel passes the filter conditions */
   if (0 == nbrOfFilteredChannels)
   {
      /* Nothing to do */
      return(0);
   }

   /* Set data not parsed flag
      This is necessary to prevent logging of invalid channel
      information. The cause is that:

      o) The original buffer is overwritten and the frame/channel
         information does not match to the new buffer
      o) The old DFF information is needed to parse Data Frames
         So re-parsing the frame is not an option.
      */
   pStation->checkPass |= ER_NOT_PARSED;

   /* At least one channel matches the filter. Build new frame */
   if (CD1_1 == pStation->type)
   {
      uint32_t jj = 0;
      uint32_t newTrailerOffset;
      uint32_t frameTimeLength;
      uint32_t chanStringCount = 0;
      char     *newFrame = NULL;

      /* Build new frame. Maximum frame size is original
         buffer size plus size of a frame signature (frame might
         be unsigned, but frame signature might be switched on) */
      newFrame = (char *)malloc(*size + SZ_SIGNATURE_MAX);
      if (NULL == newFrame)
      {
         shutdownApplication("can't allocate buffer for filter (filterFrame)",
                              pStation->conID);
         return(0);
      }

      /* copy header */
      memcpy(newFrame, buff, 36);

      /* set new frame creator */
      memcpy(&newFrame[8], pStation->pConfigData->frameCreator, 8);
      /* memcpy(&newFrame[16], "0\0\0\0\0\0\0\0", 8); reset frame destination */

      /* set new number of (filtered) channels */
      /* CTH patch: number of filtered channels is needed afterwards (e. g.
         for channel string count, return value), so it can't be converted to
         network byte order. */
      tmpInt = htonl(nbrOfFilteredChannels);
      memcpy(&newFrame[36], &tmpInt, 4);

      /* frame time length */
      frameTimeLength = 0; /* initialization */
      /* now, calculate frameTimeLength as the maximum subFrameTimeLenght from
         the selected subchannels */
      for (ii = 0; ii< nbrOfChannels; ii++)
         if (SEND_CH == sendChannel[ii])
            if (frameTimeLength < DFinfo->channel[ii].subFrameTimeLng)
               frameTimeLength = DFinfo->channel[ii].subFrameTimeLng;
      frameTimeLength = htonl(frameTimeLength);
      memcpy(&newFrame[40], &frameTimeLength, 4);

      /* nominal time */
      memcpy(&newFrame[44], &buff[44], 20);

      /* channel string count */
      chanStringCount = nbrOfFilteredChannels * 10;
      chanStringCount = htonl(chanStringCount);
      memcpy(&newFrame[64], &chanStringCount, 4);

      /* channel string */
      for (ii = 0; ii < nbrOfChannels; ii++)
      {
         if (SEND_CH == sendChannel[ii])
         {
            /* should be copied from the raw buffer ?? */
            char channelString[10];

            memset(channelString, 0, sizeof(channelString));
            memcpy(channelString, DFinfo->channel[ii].siteName, 5);
            memcpy(&channelString[5], DFinfo->channel[ii].channelName, 3);
            memcpy(&newFrame[68 + 10 * jj], channelString, 10);
            jj++;
         }
      }
      /* if number of channels is not divisible by 2 then additional two null
         bytes will be added to the end of channel string in order to
         fulfil the requirement that channel string length must be
         divisible by 4*/
      if (0 != jj % 2)
      {
         memset(&newFrame[68 + 10 * jj], 0, 2);
         frameLength = 68 + 10*jj + 2;
      }
      else
         frameLength = 68 + 10*jj;

      /* load filtered subchannels */
      for (ii = 0; ii < nbrOfChannels; ii++)
      {
         if (SEND_CH == sendChannel[ii])
         {
            memcpy(&newFrame[frameLength],
                   &buff[DFinfo->channel[ii].channelOffset],
                   DFinfo->channel[ii].channelLength + 4);
            /* increment offset */
            frameLength += DFinfo->channel[ii].channelLength + 4;
            ST_DEBUG3(9, "channelsOffset[%d] is %d",
                      (int)ii, (int)DFinfo->channel[ii].channelOffset);
            ST_DEBUG3(9, "channelsLenght[%d] is %d",
                      (int)ii, (int)DFinfo->channel[ii].channelLength + 4);
         }
      }
      newTrailerOffset = frameLength;
      /* set new trailer offset */
      newTrailerOffset = htonl(newTrailerOffset);
      memcpy(&newFrame[4], &newTrailerOffset, 4);
      newTrailerOffset = ntohl(newTrailerOffset);
      ST_DEBUG2(2, "newTrailerOffset = %d", (int)newTrailerOffset);

      /* now sign new frame */
      *size = (uint32_t)setTrailerWithSig(pStation, newFrame,
                                          newTrailerOffset, actEvent);
      /* Check if the current frame buffer size
          sufficient to hold also a signature
         (see comment at start of the block) */
      if (pStation->frameBufferSize < (*size + SZ_SIGNATURE_MAX))
      {
         uint16_t event = NO_ERROR;

         if (-1 == adjustBufferSize(pStation,
                                    (*size + SZ_SIGNATURE_MAX), &event))
         {
            shutdownApplication("can't adjust buffer size (filterFrame)",
                                pStation->conID);
            /* free working memory */
            (void)free(newFrame);
            return(0);
         }
      }
      /* transfer new content to old buffer */
      memcpy(pStation->frameBuffer, newFrame, *size);
      /* free working memory */
      (void)free(newFrame);
      return(nbrOfFilteredChannels);
   }
   else if (CD1 == pStation->type)
   {
      /* check if frame is a DFF. NbrOfChannels must be modified in this case */
      if (GOOD_FORMAT == clfType)
      {

         /* Set new number of (filtered) channels */
         tmpInt = htonl(nbrOfFilteredChannels);
         memcpy(&buff[12], &tmpInt, 4);

         /* frame length without subchannels */
         frameLength = 20;

         /* load filtered subchannels */
         for(ii = 0; ii < nbrOfChannels; ii++)
         {
            if(SEND_CH == sendChannel[ii])
            {
               /* First channel will have same source and destination,
                  but this shout not cause problems for memcpy, and
                  the other frames will not overlap */
               memcpy(&buff[frameLength], &buff[20 + (ii * 28)], 28);
               /* increment offset */
               frameLength += 28;
               ST_DEBUG3(2, "channelsOffset[%d] is %d",
                        (int)ii, (int)(20 + (ii * 28)));
            }
         }

         /* padding rest of the frame with zero */
         memset(&buff[frameLength], 0, *size - frameLength);
         return (nbrOfFilteredChannels);
      }
      else
      {
         uint32_t desLength;
         uint32_t length = 0;
         char     *newFrame = NULL;

         /* build new frame */
         newFrame = (char *)malloc(*size); /* maximum size will be original
                                              buffer size */
         if (NULL == newFrame)
         {
            shutdownApplication(
                             "can't allocate buffer for filter (filterFrame)",
                              pStation->conID);
            return(0);
         }

         /* description length */
         desLength = DFinfo->desLength;
         ST_DEBUG2(2, "desLength is: %d", (int) DFinfo->desLength);

         /* frame length without subchannels */
         frameLength = 20 + desLength;

         /* copy header */
         memcpy(newFrame, buff, 20 + desLength);

         /* load filtered subchannels */
         for (ii = 0; ii < nbrOfChannels; ii++)
         {
            if (SEND_CH == sendChannel[ii])
            {
               /* channel length */
               length = DFinfo->channel[ii].channelLength + 4;
               /* Copy channel */
               memcpy(&newFrame[frameLength],
                      &buff[DFinfo->channel[ii].channelOffset],
                      length);
               /* increment offset */
               frameLength += length;
               ST_DEBUG3(9, "channelsOffset[%d] is %d", (int)ii,
                         (int)DFinfo->channel[ii].channelOffset);
               ST_DEBUG3(9, "channelsLenght[%d] is %d",
                        (int)ii, (int)DFinfo->channel[ii].channelLength + 4);
            }
         }

         /* set size for the new frame */
         ST_DEBUG2(2, "frameLength is: %d", (int)frameLength);
         *size = frameLength;
         frameLength = htonl(frameLength);
         /* set new frame length */
         memcpy(&newFrame[4], &frameLength, 4);

         /* transfer new content to old buffer */
         memcpy(buff, newFrame, *size);

         /* free working memory */
         (void)free(newFrame);
         return(nbrOfFilteredChannels);
      }
   }
   return(0);
}

/**
 * @short Check if a data request is pending
 *
 * @description This function checks if a data request is pending and
 * @description modifies the Internal Reference List do serve the data
 * @description request
 *
 * @param pStation             station object
 * @param pRefListHead         pointer to internal reference list header
 * @param pCurrElement         address of pointer to current element in IRL
 *
 * @return NO_ERROR or event
 */

static void
check4dataRequest(station_t *pStation, intRefListHead_t *pRefListHead,
                  intRefListParams_t **pCurrElement)
{
   intRefListParams_t *pActFrame;
   cdsend_t *pCdSend;  /* Application extension object */
   int framesFound = 0;
   uint32_t requestEndTime;

   /* Check for empty queue */
   if (NULL == *pCurrElement) return;

   if (NULL == pStation->appExtension)
   {
      /* This is an internal error */
      shutdownApplication("Internal error in check4dataRequest",
                          pStation->conID);
      /* abort(); */
      return;
   }

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   /* Check if a data request is currently being processed */
   if (NOT_SENT_REQUESTED == (*pCurrElement)->sendState) return;

   if (0 == pCdSend->requestStartTime) return;

   /* Data request pending. Copy frames which should be sent in this
      request at top of the IRL and sort them like requested */
   requestEndTime = pCdSend->requestStartTime + pCdSend->requestDuration;
   pActFrame = pRefListHead->pRefListStart;
   while (NULL != pActFrame)
   {
      intRefListParams_t *pFoundFrame;
      int rval;

      if (NOT_SENT != pActFrame->sendState
          ||
          (uint32_t)pActFrame->nominalTime < pCdSend->requestStartTime
          ||
          (uint32_t)pActFrame->nominalTime > requestEndTime)
      {
         /* ignore this frame */
         pActFrame = pActFrame->pNextElement;
         continue;
      }

      if (NULL == pActFrame) break; /* only for splint */

      /* Frame found */
      pFoundFrame = pActFrame;
      /* Set next element in IRL before moving to top of IRL!! */
      pActFrame = pActFrame->pNextElement;
      /* Move found frame to top of IRL */
      rval = moveFrame2top(pRefListHead, pFoundFrame);
      pFoundFrame->sendState = NOT_SENT_REQUESTED;
      framesFound++;
      /* Set next frame to send */
      *pCurrElement = pFoundFrame;

      if (0 != rval) RH_DEBUG2(2, "moveFrame2top returns %d", rval);
   }

   if (0 == framesFound)
   {
      pCdSend->requestStartTime = 0;
      pCdSend->requestDuration = 0;
      return;
   }

   /* TODO: Check if we should set an extra restart point */

   /* TODO: re-sort found frames */

   return;
}

/**
 * @short Parse all clf files available
 *
 * @description This functions parse all clf files available
 * @description and adds their contents unconditionally
 * @description (e.g. without checking for restart points)
 *
 * @param pStation      station object
 * @param pRefListHead  intRefList header
 * @param pClfFile      reference to pointer to current clf file
 * @param sortMode      requested sorting mode (LIFO, FIFO, UNSORTED)
 *
 * @return NO_ERROR or an error event
 */

static uint16_t
parseAllClfFiles(station_t *pStation, intRefListHead_t *pRefListHead,
                 clfFileParams_t **pClfFile, int sortMode)
{
   cdsend_t *pCdSend;  /* Application extension object */

   if (NULL == pClfFile)
   {
      /* Should not happen */
      return(E_INTERNAL);
   }

   if (NULL == pStation->appExtension)
   {
      /* Should not happen */
      logError("Internal Error in parseAllClfFiles", "Unexpected NULL pointer");
      return (E_INTERNAL);
   }

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   while (NULL != *pClfFile)
   {
      uint16_t event = NO_ERROR;
      int nbrOfNewEntries;

      pStation->type = (*pClfFile)->cdtype;
      nbrOfNewEntries = parseClfFile((*pClfFile)->pFile, pStation->type,
                                     pRefListHead, 0.0, 0,
                                     NEW_FILE, sortMode);

      RH_DEBUG3(2, "parseAllClfFiles: %d new entries in %s", nbrOfNewEntries,
	        (*pClfFile)->pFile);

      if (-1 == nbrOfNewEntries)
      {
         RH_DEBUG3(2, "Error while parsing file %s.%s", (*pClfFile)->pFile,
                   EXTENSION_LST);
      }

      /* Don't unlink the last clf file in the list!! It could grow */
      if (NULL != (*pClfFile)->pNextClfFileParams)
      {
         /* Close current clf file and check if clf file object
            should be removed */
         RH_DEBUG2(2, "Parsing of [%s] finished", (*pClfFile)->pFile);
         RH_DEBUG2(2, "intRefList has %u entries",
                   (unsigned int)pRefListHead->nbrOfEntries);
         checkClfFileObj(pRefListHead);

         /* don't unlink the clf entry in aux mode */
         if (pCdSend->sndOpMode & OPER_MODE_AUX)
         {
            RH_DEBUG2(2, "aux mode; don't unlink clf entry [%s]",
                         (*pClfFile)->pFile);

            /* get next element */
            /*@-onlytrans@*/
            *pClfFile = (*pClfFile)->pNextClfFileParams;
            /*@=onlytrans@*/

            continue;
         }

         /* get next element and unlink and destroy current element */
         /*@-dependenttrans@*/
         *pClfFile = unlinkClfElement(pStation, *pClfFile, &event);
         /*@=dependenttrans@*/
         if (NO_ERROR != event) return(event);
      }
      else break;
   }
   return(NO_ERROR);
}

#ifdef ROBUST_RESTART

/**
 * @short Back up internal restart file
 *
 * @description This function renames the restart file to a backup restart
 * @description file (if any)
 *
 * @param    refListStorePath    path where restart files are stored
 * @param    pRefListHead        Header of internal reference list
 * @param pRefListHead           pointer to internal reference list header
 *
 * @return  1 in case of an error or 0 otherwise
 */

int
backupRestart(const char *refListStorePath, intRefListHead_t *pRefListHead,
              uint16_t refListStoreInterval)
{
   FILE *fp;
   int fd;
   char current[1024],
        backup[1024];

   /* Check restart file pointer */
   if (NULL == pRefListHead->intRefListFp)
   {
      return(1);
   }

   /* Check if a backup file must be created */
   if (NULL == pRefListHead->pLastSent) return (0); /* No data sent before */
   if (0 == pRefListHead->framesSinceLastSave) return (0);
   if (0 == refListStoreInterval) return (0);
   /* No restart point is written if a data request is active */
   if (NOT_SENT_REQUESTED == pRefListHead->pLastSent->sendState) return (0);

   /* Check if the current restart file is empty. It is empty if
      this was the first call after starting the worker thread */
   if (0L == lseek(fileno(pRefListHead->intRefListFp), 0L, SEEK_END))
   {
      return(0);
   }

   /* Create file name for restart file */
   (void)snprintf(current, sizeof(current), "%s/.%s.%s",
                  refListStorePath, pRefListHead->plID, EXTENSION_IRL);
   /* Create file name for backup restart file */
   (void)snprintf(backup, sizeof(backup), "%s/.%s.%s",
                  refListStorePath, pRefListHead->plID, EXTENSION_IRL_BACKUP);

   /* Close current restart file */
   (void)fclose(pRefListHead->intRefListFp);

   RH_DEBUG3(2, "Rename restart file [%s] to [%s]", current, backup);

   /* Rename current file to backup file */
   if (rename(current, backup))
   {
      /* Don't try to handle an error here. It is too difficult.
         If something fails we open the old file anyway */
      logWarning("Can't make backup of restart file", backup);
   }

   /* Open the file with low level I/O functions to make sure
      that the file exists when open with fopen. Don't use "a+"
      for fopen here */
   fd = open(current, O_RDWR | O_CREAT, FILE_PERMS);
   if (-1 != fd) (void)close(fd);

   fp = fopen(current, "r+");
   if (NULL == fp)
   {
      logError("Can't create/open restart file", current);
      return(1);
   }

   /*@-mustfree@*/
   pRefListHead->intRefListFp = fp;
   /*@=mustfree@*/
   return(0);
}

#endif /* ROBUST_RESTART */

#define BUFFER_TIME (5*60)  /* Maximal buffer time for timely ordered mode */

/**
 * @short Check if frame can be send in timely ordered mode
 *
 * @description This function checks if frame can be send in timely
 * @description ordered mode.
 *
 * @param pRefListHead  pointer to IRL header
 * @param pActElement   pointer to current frame
 * @param pRefListHead  pointer to IRL header
 *
 * @return 0 if the frame can be sent, 1 otherwise
 */

static int
canNotSendTimely(intRefListHead_t *pRefListHead,
                 intRefListParams_t *pActElement)
{
   double nominalTime;
   uint64_t cm1 = 0;
   uint64_t cm2 = 0;

   if (NULL == pActElement)
   {
      RH_DEBUG(1, "dumpIRL: Warning!! No entry in list!");
      return(1);
   }

   nominalTime = pActElement->nominalTime;

   /* TODO: Check for complete parsed frame ?? */

   /* Check if the current frame contains all channels.
    * The frame might contain more than the known channels, because
    * the channel mask is learned
    */
   /*@-realcompare@*/
   while (pActElement != NULL && pActElement->nominalTime == nominalTime)
   {
      cm1 |= pActElement->cm1;
      cm2 |= pActElement->cm2;

      /* Don't go behind the last element in the queue */
      if (NULL == pActElement->pNextElement) break;
      pActElement = pActElement->pNextElement;
   }
   /*@=realcompare@*/

   /* The reference channel mask is set when the clf lines are parsed. Now
      check if all channels are present */
   if ((pRefListHead->cm1r & cm1) != pRefListHead->cm1r
        ||
       (pRefListHead->cm2r & cm2) != pRefListHead->cm2r)
   {
      return(1);
   }

   RH_DEBUG4(2, "Expected frame time in ordered mode [%f] [%f] [%u]",
             pRefListHead->frameEndTime, nominalTime,
             (unsigned int)pActElement->frameTimeLng);

   /* Check if there is no gap between the previous and the current frame
    */
   /*@-realcompare@*/
   if (pRefListHead->frameEndTime != 0.0 &&
       pRefListHead->frameEndTime != nominalTime)
      return(1);
   /*@=realcompare@*/

   /* All checks done, the frame can be sent. Don't forget to update
      the expected time of the next frame when the current frame has
      been sent */
   return(0);
}

/**
 * @short Send contents of internal reference list
 *
 * @description This function sends the contents of the internal
 * @description reference list via a socket to a CD receiver and deletes
 * @description the sent entry out of internal reference list
 *
 * @param pRefListHead         pointer to internal reference list
 * @param pStation             station object
 * @param actEvent             event object
 * @param pCurrentClf          reference to pointer to clf file list
 *
 * @return NO_ERROR or event
 */

static uint16_t
sendIrlData(intRefListHead_t *pRefListHead,
            station_t *pStation, CDevent_t *actEvent,
            clfFileParams_t **pCurrentClf)
{
   FILE *fp = NULL;
   intRefListParams_t *pCurrElement; /* current frame */
   char logBuffer[LOGSIZE],
        fileName[LIBPAR_BUFSIZ];  /* file name on bin file */
   uint16_t event = NO_ERROR;
   int sendDFF = 0,    /* Indication that DFF has been sent */
       sortType = getSortLevel(pStation);
   senderParams_t *pSenderParams;
   cdsend_t *pCdSend;  /* Application extension object */


   /* Do some checks */
   if (NULL == pRefListHead || NULL == pStation->frameBuffer
       ||
       NULL == pStation->appExtension)
   {
      logError("Internal Error in sendIrlData", "Unexpected NULL pointer");
      return (E_OOPS);
   }

   /* Don't forget to make sure that the IRL is sorted according to the
      current processing mode (LIFO, FIFO, AS_ARRIVED) */

   memset(fileName, 0, sizeof(fileName)); /* For splint */
   /* set sender specific config parameters */
   pSenderParams = (senderParams_t *)pStation->pConfigData->configExtension;

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   RH_DEBUG2(5, "Value of nbrOfEntries [%u]",
             (unsigned int)pRefListHead->nbrOfEntries);

   IF_RH_DEBUG(9) { dumpIRL(pRefListHead); }

   /* Sent all frames in the intReflist which has
      state equal NOT_SEND */
   pCurrElement = pRefListHead->pRefListStart;
   while (NULL != pCurrElement)
   {
      uint32_t frameLength,
               offsetBinFile,
               frameSize;
      int retries;
      char clfType;
      /*@dependent@*/ CDframeset_t *ackFrameSet = NULL;

      if (0 == sendDFF)
      {
         /* Check if a data request is pending. Only frames seen in
            the current IRL are considered. This is save, because the
            IRL contains all unsent frames currently known */
         check4dataRequest(pStation, pRefListHead, &pCurrElement);

         if (NULL == pCurrElement) break; /* For splint only */
      }

      if (pCurrElement->sendState != NOT_SENT &&
          pCurrElement->sendState != NOT_SENT_REQUESTED &&
          pCurrElement->sendState != HOLD_BACK)
      {
         /* Skip already sent element */
         pCurrElement = pCurrElement->pNextElement;
         continue;
      }

      if (NULL == pCurrElement) break; /* For splint only */

      /* Send only requested data in auxiliary station mode. Requested
         data is found on top of the IRL, so if data is seen which is not
         requested, all requested data have been sent */
      if (0 != (pCdSend->sndOpMode & OPER_MODE_AUX)
         &&
         pCurrElement->sendState != NOT_SENT_REQUESTED)
      {
         break;
      }

      /* Check if data is received. Do not call readData()
         when the frame buffer is filled with data to
         send or data to send will be overwritten when
         data are received. DON'T use a local buffer
         here which references the frame buffer because it could
         get re-allocated in readData() */
      /* Line idle timeout is considered only if at least one
         byte is received */
      if (dataIsAvailable(pStation->portInfo->socket))
      {
         /* Data is available, read it */
         event = readData(pStation, actEvent);

         if (S_ACKNAK == event &&
             0 == (pSenderParams->senderMode & BUGGY_ACKNACK_FRAMES))
         {
            /* Parse AckNack frame. Result is a frame set. The frame
               set  specific list is valid until the next AckNack
               frame for the same frame set is received.
               Only valid AckNack frames are treated further */
            ackFrameSet = parseAckNack(pStation->frameBuffer,
                                       pStation->header.trailerOffset,
                                       &(pCdSend->frameSets), &event);

            /* Ignore invalid AckNack frames in none strict mode */
            if (NO_ERROR != event &&
                0 != (pStation->systemConf & STRICT_MODE_ENABLE))
            {
               if (NULL != fp) (void) fclose(fp);
               return(event);
            }
         }

         /* Ignore AckNack frame when in buggy AckNack mode.
            TODO: parse option response frame */
         if (S_OPTION_RES == event || S_ACKNAK == event) event = NO_ERROR;

         if (NO_ERROR != event)
         {
            RH_DEBUG3(2, "Last offset was %u, framesize %u",
                      (unsigned int)(pCurrElement->offsetBinFile),
                      (unsigned int)(pCurrElement->frameSize));

            if (NULL != fp) (void) fclose(fp);
            return(event);
         }
      }

      /* Check for timely ordered mode. Check for arrived data
         has to be done before, or an incoming alert frame will
         not be processed while waiting for a missing frame */
      if (0 != (pCdSend->sndOpMode & PROCESS_MODE_ORDERED))
      {
         /* first check if the current frame has crossed the time boundary.
          * In this case sent it
          */
         if ((time_t)pCurrElement->nominalTime < (time(NULL) - BUFFER_TIME))
         {
            /* Check if the frame was hold back */
            if (HOLD_BACK != pCurrElement->sendState)
            {
               intRefListParams_t *pNextElement;
               int rval;

               /* Too old frame, delete it */
               RH_DEBUG2(2, "Drop too old frame in ordered mode [%d]",
                         (int)pCurrElement->nominalTime);
               pNextElement = pCurrElement->pNextElement;
               rval = deleteRefListFrame(pRefListHead, pCurrElement);
               if (0 != rval)
               {
                  RH_DEBUG2(2, "deleteRefListFrame returns %d", rval);
               }
               pCurrElement = pNextElement;
               continue;
            }
            RH_DEBUG2(2, "send frame which crosses the time boundary [%d]",
                      (int)pCurrElement->nominalTime);
            /* else send the frame now */
         }
         else if (0 != canNotSendTimely(pRefListHead, pCurrElement))
         {
            pCurrElement->sendState = HOLD_BACK;
            RH_DEBUG2(2, "Hold back frame in ordered mode [%d]",
                     (int)pCurrElement->nominalTime);
            /* Wait for retry time to expire or user shutdown */
            retryWait(pStation, (int)pSenderParams->sleepTimeWork);
            if (NULL != fp) (void) fclose(fp);
            return(NO_ERROR);
         }
         else
         {
            RH_DEBUG2(2, "can send frame in ordered mode [%d]",
                      (int)pCurrElement->nominalTime);
         }
         /* Send the frame now */
      }

#define P_LAST_ELEMENT_SENT  pRefListHead->pLastSent

      /* Check if a DFF has to be sent */
      if (CD1 == pStation->type && sendDFF == 0 &&
          (NULL == P_LAST_ELEMENT_SENT ||
           P_LAST_ELEMENT_SENT->DFFoffset != pCurrElement->DFFoffset ||
           P_LAST_ELEMENT_SENT->pFileObj != pCurrElement->pFileObj))
      {
         /* Send DFF */
         if (NULL == P_LAST_ELEMENT_SENT)
         {
            RH_DEBUG(2, "Insert DFF for the first frame");
         }
         else if (P_LAST_ELEMENT_SENT->DFFoffset != pCurrElement->DFFoffset)
         {
            RH_DEBUG(2, "Insert DFF because different DFFoffset");
            RH_DEBUG3(2, "LastSentDFFoffset= %ld, CurrentDFFoffset=%ld",
                      (long)P_LAST_ELEMENT_SENT->DFFoffset,
                      (long)pCurrElement->DFFoffset);
         }
         else if (P_LAST_ELEMENT_SENT->pFileObj != pCurrElement->pFileObj)
         {
            RH_DEBUG(2, "Insert DFF because different file pointer");
         }

         offsetBinFile = pCurrElement->DFFoffset;
         frameSize = pCurrElement->DFFsize;
         clfType = GOOD_FORMAT;
         sendDFF = 1;
      }
      else
      {
         /* Send list element */
         offsetBinFile = pCurrElement->offsetBinFile;
         frameSize = pCurrElement->frameSize;
         clfType = pCurrElement->frameType;
         sendDFF = 0;
      }

      if (NULL == fp
          ||
          P_LAST_ELEMENT_SENT == NULL /* This will cause an extra open */
          ||
          P_LAST_ELEMENT_SENT->pFileObj != pCurrElement->pFileObj)
      {

         /* File name has changed. Close last file and open new one */
         if (NULL != fp) (void)fclose(fp);

         /* Implicit default is CD1 */
         if (CD1 == pStation->type)
         {
            (void) snprintf(fileName, sizeof(fileName), "%s.%s",
                            pCurrElement->pFileObj->fileNameAndPath,
                            EXTENSION_BIN_CD1);
         }
         else
         {
            (void) snprintf(fileName, sizeof(fileName), "%s.%s",
                            pCurrElement->pFileObj->fileNameAndPath,
                            EXTENSION_BIN_CD11);
         }

         RH_DEBUG2(2, "+++ Open binary file [%s]", fileName);

         /* Open file to sent */
         fp = fopen(fileName, "rb");
         if (NULL == fp)
         {
            (void) snprintf(logBuffer, sizeof(logBuffer),
                            "[%.17s] can't open bin file", pStation->conID);
            logError(logBuffer, fileName);

            /* remove frames with that file from internal reference list */
            pCurrElement = deleteWithThatFileObject(pRefListHead,
                                                    pCurrElement);
            /* This is not treated as an error in the calling function */
            continue;
         }
      }

      /* Seek to the current position; optimization is done
         by fseek() */
      if (0 != fseek(fp, (long)offsetBinFile, SEEK_SET))
      {
         (void)snprintf(logBuffer, sizeof(logBuffer),
                        "[%.17s] Can't seek to offset %u in bin file",
                        pStation->conID,
                        (unsigned int)offsetBinFile);
         logWarning(logBuffer, fileName);
         /* remove frames with that file from internal reference list */
         pCurrElement = deleteWithThatFileObject(pRefListHead,
                                                 pCurrElement);
         /* TODO: behave like cd2wng here */
         /* This is not treated as an error in the calling function */
          continue;
      }

      /* check for a pending shutdown */
      if (pStation->shutdownEnable != SD_NONE)
      {
         (void) fclose(fp);
         RH_DEBUG(2, "user shutdown pending");
         return(S_SHUTDOWN);
      }

      RH_DEBUG3(5, "Offset %u, framesize %u", (unsigned int)(offsetBinFile),
                (unsigned int)(frameSize));

      /* Check if the stations frame buffer is large enough to hold
         the frame */
      if (pStation->frameBufferSize < frameSize)
      {
         if (-1 == adjustBufferSize(pStation, frameSize, &event))
         {
            (void) fclose(fp);
            return(event);
         }
      }

      frameLength = 0;
      retries = MAX_READ_RETRIES;
      /* Read the complete frame into buffer */
      while (frameLength < frameSize)
      {
         size_t len;

         len = fread(pStation->frameBuffer + frameLength, 1,
                     frameSize - frameLength, fp);

         if (0 == len)
         {
            /* This can be caused by file system caching (e.g.
               data to send and sender are located on different hosts and
               the sender uses an NFS share to access the data).
               This event is silently ignored when compiled with
               retries as long as max. retry count is not reached */
            RH_DEBUG3(2, "fread returns 0: frameLength %d, errno %d",
                      (int)frameLength, errno);
            RH_DEBUG3(2, "Offset %u, framesize %u",
                      (unsigned int)(offsetBinFile),
                      (unsigned int)(frameSize));

            if (retries <= 0)
            {
               (void)snprintf(logBuffer, sizeof(logBuffer),
                              "[%.17s] Unable to read frame at offset [%u]",
                              pStation->conID,
                              (unsigned int)(offsetBinFile));
               logWarning(logBuffer, fileName);
               /* This is not treated as an error */
               break;
            }
            clearerr(fp);
            retryWait(pStation, 5);
            retries--;
         }
         else retries = MAX_READ_RETRIES;

         frameLength += (uint32_t)len;
      }

      /* Set default values and authentication state */
      pStation->checkPass = NO_ERROR;
      pStation->authenPassFrame = AUTH_UNKNOWN; /* default value */
      pStation->authenSummary = AUTH_UNKNOWN; /* default value */
      if (NULL != pStation->DFinfo)
      {
         /* Don't reset number of parsedDFFchannels here.
            Don't forget to reset parsedDFFchannels
            when the filter is switched off via a command frame
            sometime !! */
         pStation->DFinfo->parsedChannels = 0;
         /* Sets the nominal time in the DF info structure to
            time read from clf file to have the nominal time
            logged in clf and txt files */
         pStation->DFinfo->nominalTime = pCurrElement->nominalTime;
      }

      /* Should we send partial frames?? */
      while (frameLength == frameSize)
      {
         uint16_t goodEvent = S_SEND_DATA;

         /* A frame is "good" by definition when it is sent */
         if (clfType >= 'A' && clfType <= 'Z') clfType += ' ';

         if (GOOD_FORMAT == clfType) goodEvent = S_SEND_DFF;

         /* Parse frame and check if frame size in frame is equal to
            frame size in clf file */
         event = checkIfFrameTooShort(pStation, clfType, frameLength);

         /* check for too short frame */
         if (E_FRAME_TOO_SHORT == event)
         {
            /* Don't send too short frames. If this was caused by a
               timeout during receiving of the frame the frame should be
               re-sent and no data is lost in that case. Behave as the
               frame has been sent.
               If frame was a Data Format frame all frames related
               to that DFF must be removed from the IRL later
               at the end of the while loop. */
         /* TODO: behave like cd2wng here */
            break;
         }
         else if (NO_ERROR != event)
         {
            /* If an internal error occurs sender should start with
               this frame again */
            (void)fclose(fp);
            return(event);
         }

         /* Filter the frame before sending. Keep in mind that
            pStation->frameBuffer might get re-allocated in filterFrame() */
         if (filterFrame(pStation, clfType, &frameLength, actEvent))
         {
            /* Check if a data format flush alert frame should be sent */
            if (1 == sendDFF
                &&
                CD1 == pStation->type && NULL != P_LAST_ELEMENT_SENT)
            {
               /* Don't use a char buffer here. The buffer must be aligned
                  to a 4 byte boundary and the Solaris compiler don't do
                  that!! */
               uint32_t AFbuffer[(SZ_CD1_ALERT/sizeof(uint32_t)) + 1];
               uint32_t savedCheckPass;

               /* Send alert frame before sending a DFF */
               setCD1AF((char *)AFbuffer, FLUSH_FORMAT);

               /* Send frame */
               RH_DEBUG(2, "Send AF to flush format");

               /* Save and reset check pass. This alert frame will pass
                  frame check */
               savedCheckPass =  pStation->checkPass;
               pStation->checkPass = 0;
               /* statistic is updated and data is saved in sendFrame */
               /*@-compdef@*/
               event = sendFrame(pStation, (char *)AFbuffer, SZ_CD1_ALERT,
                                 S_SEND_ALERT, GOOD_ALERT, 1);
               /*@=compdef@*/

               if (NO_ERROR != event)
               {
                  RH_DEBUG2(2, "Sendframe reports [%d]", (int) event);
                  (void)fclose(fp);
                  return(event);
               }
               /* Restore DFFs checkPass value */
               pStation->checkPass = savedCheckPass;
            }
            /* Send the frame now */
            event = sendFrame(pStation, pStation->frameBuffer,
                              (int)frameLength, goodEvent, clfType, 1);
            /* Remember the end time of the frame sent for ORDERED sending
               mode */
            if (0 == sendDFF)
                pRefListHead->frameEndTime = pCurrElement->nominalTime +
                         (pCurrElement->frameTimeLng / 1000);

            if (CD1_1 == pStation->type &&
                0 == (pSenderParams->senderMode & BUGGY_ACKNACK_FRAMES))
            {
               /* Remember the frame set name for processing of AckNack
                  frames if the receiver on the other side is capable sending
                  AckNack frames which do not harm too much. Otherwise
                  AckNack frames are ignored and there is no need to
                  remember the frame set name. If the frame set
                  is not known, a new frame set is created */
               if (NULL == getFrameSet(pCurrElement->aFrameset,
                                       &(pCdSend->frameSets),
                                       CREATE_FS | MAP_IRL_2FS))
               {
                  RH_DEBUG(2, "getFrameSet can't create frame set");
                  (void)fclose(fp);
                  return(E_NO_MEM);
               }
            }
         }
         else
         {
            /* TODO: add dropped frames by filter to statistics */
            RH_DEBUG(2, "frame dropped by filter rule");
         }

         if (NO_ERROR != event)
         {
            RH_DEBUG2(2, "Sendframe reports [%d]", (int) event);
            RH_DEBUG3(2, "Last offset was %u, framesize %u",
                      (unsigned int)(offsetBinFile),
                      (unsigned int)(frameSize));
            /* Temporary: should be done in sendFrame; check if an
                         alert frame is pending */
            if (dataIsAvailable(pStation->portInfo->socket))
            {
               uint16_t tmpEvent;

               /* Data is available, read it */
               tmpEvent = readData(pStation, actEvent);

               if (S_ACKNAK == tmpEvent &&
                   0 == (pSenderParams->senderMode & BUGGY_ACKNACK_FRAMES))
               {
                  /* Parse AckNack frame. Result is a frame set. The frame
                     set  specific list is valid until the next AckNack
                     frame for the same frame set is received.
                     Only valid AckNack frames are treated further */
                  ackFrameSet = parseAckNack(pStation->frameBuffer,
                                             pStation->header.trailerOffset,
                                             &(pCdSend->frameSets), &tmpEvent);

                  /* Ignore invalid AckNack frames in none strict mode */
                  if (NO_ERROR != tmpEvent &&
                      0 != (pStation->systemConf & STRICT_MODE_ENABLE))
                  {
                     if (NULL != fp) (void) fclose(fp);
                     return(tmpEvent);
                  }
               }

               /* Ignore AckNack frame when in buggy AckNack mode.
                  TODO: parse option response frame */
               if (S_OPTION_RES == tmpEvent || S_ACKNAK == tmpEvent)
                  tmpEvent = NO_ERROR;

               if (NO_ERROR != tmpEvent)
               {
                  (void) fclose(fp);
                  return(tmpEvent);
               }
            }

            /* If an error occurs sender should start with
               this frame again */
            (void) fclose(fp);
            return(event);
         }
         break; /* break while */
      } /* end while */

      /* If a DFF has been sent before, send list element now */
      if (0 != sendDFF)
      {
         /* If the DFF was too short to parse, remove all
            related frames from the IRL */
         continue;
      }

      /* Remember the sending sequence to be able to go back N frames
         when writing a restart point */
      if (NULL != P_LAST_ELEMENT_SENT)
      {
         P_LAST_ELEMENT_SENT->pHistNext = pCurrElement;
      }
      pCurrElement->pHistPrev = P_LAST_ELEMENT_SENT;
      /* Save the last element sent in intRefList header for re-start */
      P_LAST_ELEMENT_SENT = pCurrElement;
      pCurrElement->sendCount++;
      /* Check how often the frame has been sent to before avoid sending
         a frame which causes a disconnect too many times */
      if (pCurrElement->sendCount >= MAX_SENT_RETRIES)
      {
         pCurrElement->sendState = SENT_ACK;
      }
      else
      {
         pCurrElement->sendState = SENT_NOACK;
      }

      /*@-realcompare@*/
      if (0.0 == pRefListHead->restartNominalTime
          ||
          (LIFO == sortType &&
          pRefListHead->restartNominalTime > pCurrElement->nominalTime)
          ||
          (LIFO != sortType &&
           pRefListHead->restartNominalTime < pCurrElement->nominalTime))
      {
         /* Remember nominal time of oldest (LIFO) or newest (FIFO/unsorted)
            frame sent. This information in needed for re-start */
         pRefListHead->restartNominalTime = pCurrElement->nominalTime;
      }
      /*@=realcompare@*/

      /* Check if an AckNack Frame has been received */
      if (NULL != ackFrameSet)
      {
         /* Now acknowledged all frames which have been sent and which are
            not part of a gap and remove the send links from unacknowledged
            frames and move them to the top of the IRL */
         event = acknowledgeFrames(pRefListHead, ackFrameSet);

         ackFrameSet = NULL;
         if (NO_ERROR != event)
         {
            if (NULL != fp) (void) fclose(fp);
            return(event);
         }
      }

      /* Check if a history should be saved */
      pRefListHead->framesSinceLastSave++;
      if (0 != pSenderParams->refListStoreInterval
          &&
          0 == (pCdSend->sndOpMode & OPER_MODE_AUX)
          &&
          pRefListHead->framesSinceLastSave >
                           pSenderParams->refListStoreInterval)
      {
#ifdef ROBUST_RESTART
         /* Back up an existing previous written restart file */
         (void)backupRestart(pSenderParams->refListStorePath, pRefListHead,
                             pSenderParams->refListStoreInterval);
#endif /* ROBUST_RESTART */
         /* Write a re-start point to disk */
         saveIntRefList(getSortLevel(pStation),
                        pSenderParams->refListStoreInterval, pRefListHead);
      }

      /* Check if the sender runs in FIFO or LIFO mode and whether or
         not new data has arrived in the meantime. This check will
         not detect new frames in clf files created after the current
         clf file!! */
      if (UNSORTED != sortType &&
	  (0 != haveDataToSend(*pCurrentClf, pRefListHead->currentClfFile)))
      {
         /* Add all frames from the current position
            in the clf file to the end of the last clf file to the internal
            reference list */
         event = parseAllClfFiles(pStation, pRefListHead,
                                  pCurrentClf, sortType);
         if (NO_ERROR != event)
         {
            if (NULL != fp) (void) fclose(fp);
            return(event);
         }
	 /* There might be some new frames on top of the IRL, but
	    the sending pointer is somewhere in the IRL. Therefore
	    a check if the first element on the IRL was not send has
	    to be done here. */
	 if (pRefListHead->pRefListStart != NULL
	     && NOT_SENT == pRefListHead->pRefListStart->sendState)
	 {
            pCurrElement = pRefListHead->pRefListStart;
	 }
	 else
	 {
            /* Process next element in internal reference list */
            pCurrElement = pCurrElement->pNextElement;
	 }
      }
      else
      {
         /* Process next element in internal reference list */
         pCurrElement = pCurrElement->pNextElement;
      }

#undef P_LAST_ELEMENT_SENT

   } /* ( NULL != pCurrElement) */

   IF_RH_DEBUG(6) { dumpIRLsendSequence(pRefListHead); }

   /* Check fp. If deleteWithThatFileObject() deletes all frames, fp
      can became NULL */
   if (NULL != fp) (void) fclose(fp);

   if (pCdSend->sndOpMode & OPER_MODE_AUX)
   {
      pStation->shutdownEnable = SD_STOP;
      RH_DEBUG(2, "aux mode; shutdown sender thread");
      return(S_SHUTDOWN);
   }
   return (NO_ERROR);
}

/**
 * @short Check for "bad" event
 *
 * @description This function checks if an event has occurred
 * @description which aborts the current frame processing. Fatal
 * @description events causes the application to stop
 *
 * @param event          event
 * @param strictMode     strict mode flag
 *
 * @return 0 if "good" event or  1 if frame processing should be aborted
 */

static int
isBadEvent(uint16_t event, uint32_t strictMode)
{
   if (E_OOPS == event || E_NO_MEM == event)
   {
      /* These are unexpected events, default behaviour is
         is to shut down the application */
      shutdownApplication("Unexpected error while reading/processing frame",
                          NULL);
      return(1);
   }

   if (E_SYSTEM == event)
   {
      /* That could be an interrupted system call or a
         connection reset by peer, or ..
         These events are treated as an error */
      return(1);
   }

   if (E_TIMEOUT == event)
   {
       /* Check for heartbeat timeout. When at least one
          byte is received this is no heartbeat timeout per definition */
       /* if (0 == bytesRead && HEARTBEAT == nextTimeout) continue; */
      return(1);
   }

   if (E_CLOSED == event || E_FRAME_TYPE == event
       ||
       E_FRAMESIZE == event || E_FRAME_TOO_BIG == event)
   {
      /* This events are always an error */
      return(1);
   }

   /* Check if application is in strict mode */
   if (0 != strictMode
       &&
       (E_CRC == event || E_NO_KEYS == event || E_NO_FRAME_SIG == event
        ||
        E_NO_CERT == event || E_FRAME_AUTH == event))
   {
      /* Remaining events E_CRC, E_NO_KEYS, E_NO_FRAME_SIG,
         E_NO_CERT and E_FRAME_AUTH are treated as an error
         in strict mode */
       return(1);
   }
   return (0);
}

/**
 * @short Connect to CD receiver
 *
 * This function sends a CD-1.0 Connection Request Frame
 * to the well known port of a CD receiver, waits
 * for a Port Assignment Frame, closes the old (TCP)
 * connection and re-connects to the assigned port if
 * station type is CD-1.0 or sends a CD-1.1 Connection
 * Request Frame, waits for a Connection Response Frame,
 * closes the old (TCP) connection and re-connects to
 * the assigned port if station type is CD-1.1
 *
 * @param pStation    pointer to station object
 * @param rwFlags     read/write flags for socket
 * @param actEvent    event object
 *
 * @return 0 on success or -1 on error
 */

static int
connectToReceiver(station_t *pStation, /*@out@*/rwFlags_t *flags,
                  CDevent_t *actEvent)
{
   portMap_t *portInfo = pStation->portInfo;
   /*@dependent@*/ char *buffer;
   char clfType;
   int sendSize,
       retVal;
   uint16_t event = NO_ERROR;
   char dpName[LEN_STATION_NAME + 1];
   prodLine_t *pProdLine;

#if INIT_BUFFER_SIZE < LEN_CD1_PAF || INIT_BUFFER_SIZE < LNG_CRF
        /* Oops! Seems that INIT_BUFFER_SIZE has been defined wrong.
           Stop compilation with an error */
        #error Check the definition of INIT_BUFFER_SIZE
#endif

   if (NULL == pStation->appExtension)
   {
      shutdownApplication("Unexpected NULL pointer (connectToReceiver)",
                          pStation->conID);
      return (-1);
   }
   pProdLine = ((cdsend_t *)(pStation->appExtension))->pProdLine;

   /* Set remote address */
   if (NULL == pProdLine)
   {
      return (-1);
   }

   /* Set values for connection request */
   flags->sockFlags = 0;
   flags->shutdownFlag = &(pStation->shutdownEnable);
   flags->timeout = pStation->pConfigData->CRtimeOut;
   flags->resetDataTimeout = pStation->pConfigData->CRtimeOut;
   flags->firstTime = &(pStation->startRoW);

   /* Try to connect to the Receiver's well known port */
   if (-1 == connectTo(pStation, flags, actEvent,
                       pProdLine->pDataConsumer->cdIpAddr,
                       pProdLine->pDataConsumer->portNumber))
   {
      return (-1);
   }

   /* Use the frame buffer as sending or receive buffer */
   buffer = pStation->frameBuffer;

   /* Extract data provider name from connection id */
   getDPname(pStation->conID, dpName, sizeof(dpName));

   if (CD1 == pStation->type)
   {
      /* create CD1 connection request frame */
      setCD1ConRq(buffer, dpName);
      sendSize = LEN_CD1_CRF;
   }
   else
   {
      /* Set Header and payload for CD1.1 Connection Request */
      setHeader(buffer, CONNECTION_REQUEST, (uint32_t) 68,
                pStation->pConfigData->frameCreator,
                /*@-compdef@*/
                pProdLine->pDataConsumer->dcName,
                /*@=compdef@*/
                (uint64_t) 0);
      setCRpayload(&buffer[SZ_HDR], dpName,
                   pStation->pConfigData->responderType,
                   portInfo->localIPaddr,
                   portInfo->localPort);
      sendSize = setTrailerWithSig(pStation, buffer, (uint32_t)68, actEvent);

      if (-1 == sendSize)
      {
         /* Unexpected error while signing data */
         (void) close(portInfo->socket);
         portInfo->socket = -1;
         return(-1);
      }
   }

   /* Send connection request frame */
   retVal = writeWithTimeout(portInfo->socket, buffer, sendSize, flags);
   (void) gettimeofday(&(pStation->endRoW), NULL);
   ST_DEBUG2(9, "CRF sent, returns [%d]", retVal);

   if (retVal != sendSize)
   {
#define WHICH_EVENT(retVal, sor) \
      if ((retVal) == -1) /* System error */ \
      { \
         pStation->savedSystemErr = (uint32_t)errno; \
         logCDevent2(ET_L_ERROR, E_SYSTEM, actEvent, (uint32_t)errno); \
         pStation->whenUnconnected.systemFailures++; \
         event = E_SYSTEM; \
         IF_ST_DEBUG(2) \
         { \
            perror("Can't write to remote port"); \
         } \
      } \
      else if ((retVal) == -2) /* Fast or normal user requested shutdown */ \
      { \
         logCDevent2(ET_L_INFO, S_SHUTDOWN, actEvent, NO_INFO); \
         event = S_SHUTDOWN; \
      } \
      else if (0 == flags->timeout) /* Send timeout */ \
      { \
         logCDevent2(ET_L_ERROR, sor, actEvent, NO_INFO); \
         pStation->whenUnconnected.timeouts++; \
         event = sor; \
      } \
      else /* Connection closed unexpected */ \
      { \
         logCDevent2(ET_L_ERROR, E_CLOSED, actEvent, NO_INFO); \
         pStation->whenUnconnected.closedConnections++; \
         event = E_CLOSED; \
      }

      WHICH_EVENT(retVal, E_SEND_TIMEOUT);

#define SAVE_DATA(rcvSize, sendSize, event, clfType) \
   if (rcvSize > 0) \
   { \
     updateStatistics(&(pStation->framesReceived), rcvSize, YES, clfType); \
   } else { \
     updateStatistics(&(pStation->framesSent), sendSize, YES, clfType); \
   } \
   if (saveData(pStation, buffer, rcvSize, sendSize, event, YES, clfType)) \
   { \
      /* SaveData will shutdown the application if data could not be saved */ \
      pStation->whenUnconnected.systemFailures++; \
      logCDevent2(ET_L_ERROR, E_SAVE_DATA, actEvent, NO_INFO); \
      (void) close(portInfo->socket); \
      portInfo->socket = -1; \
      return(-1); \
   }

      if (retVal > 0)
      {
          SAVE_DATA(retVal, 0, event, BAD_CRF);
      }
      (void) close(portInfo->socket);
      portInfo->socket = -1;
      return(-1);
   }

   /* Reset parsing flags */
   pStation->checkPass = 0;

   /* Save sent data */
   SAVE_DATA(0, retVal, S_SEND_CRF, GOOD_CRF);

   /* Set state to wait for connection response */
   pStation->state = STATION_WAITING;

   /* Read the response frame now */
   flags->timeout = pStation->pConfigData->CRtimeOut;
   flags->firstTime = &(pStation->startRoW);

   if (CD1 == pStation->type)
   {
      /* CD1 data, read Port Assignment Frame */
      retVal = readWithTimeout(portInfo->socket, buffer, LEN_CD1_PAF, flags);
      (void) gettimeofday(&(pStation->endRoW), NULL);

      /* Set values for data integration test */
      pStation->header.frameType = ntohl(*((uint32_t *)buffer));
      pStation->header.trailerOffset = ntohl(*((uint32_t *)&buffer[4]));

      /* Close connection to well known port */
      (void) close(portInfo->socket);
      portInfo->socket = -1;

      if (LEN_CD1_PAF != retVal)
      {
         /* Interrupt, timeout, connection closed, shutdown pending
            or system failure */
         WHICH_EVENT(retVal, E_TIMEOUT);
         if (retVal > 0)
         {
            SAVE_DATA(retVal, 0, event, FRAME_UNKNOWN);
         }
         return(-1);
      }
      buffer[LEN_CD1_PAF] = '\0';
      ST_DEBUG2(5, "PAF received with value [%s]", buffer);

      /* parse PAF in buffer and update portInfo */
      if (0 != parsePAF(buffer, portInfo))
      {
         /* TODO: this overrides the portinfo values for logging */
         /* invalid content of PAF */
         logCDevent2(ET_L_ERROR, E_PAF, actEvent, NO_INFO);
         /* Save received data */
         SAVE_DATA(retVal, 0, E_PAF, BAD_PAF);
         return(-1);
      }
      /* Save received data */
      SAVE_DATA(retVal, 0, S_PAF, GOOD_PAF);

   }
   else
   {
      CDconnectRequest_t payload;
      uint32_t ipAddr;    /* IP address in NBO */

      /* CD1.1 data, read Connection Response Frame */
      retVal = readCDframe(pStation, actEvent, flags, &event, &clfType);

      /* Close connection to well known port */
      (void) close(portInfo->socket);
      portInfo->socket = -1;

      /* Check event */
      if (S_SHUTDOWN == event) return(-1);

      /* There could be data read from the socket */
      if (retVal > 0) { SAVE_DATA(retVal, 0, event, clfType); }

      /* Check if processing should be aborted */
      if (1 == isBadEvent(event, pStation->systemConf & STRICT_MODE_ENABLE))
      {
         return(-1);
      }

      /* check for connection response */
      if (clfType != GOOD_CON_RES && clfType != BAD_CON_RES)
      {
         logCDevent2(ET_L_ERROR, E_CONN_RES, actEvent, NO_INFO);
         return(-1);
      }

      ST_DEBUG(5, "Connection response received");
      /* Parse CD1.1 Connection Response payload */
      parseCRframe(&buffer[SZ_HDR], &payload);
      ipAddr = payload.IPaddress;
      (void)snprintf(portInfo->remoteIPaddr, sizeof(portInfo->remoteIPaddr),
                     "%d.%d.%d.%d",
                     (int)((ipAddr>>24)&0xFF),
                     (int)((ipAddr>>16)&0xFF),
                     (int)((ipAddr>>8)&0xFF),
                     (int)((ipAddr)&0xFF));
      portInfo->remotePort = payload.port;
   }

   ST_DEBUG3(2, "Redirect info is  [%.16s:%u]",
             portInfo->remoteIPaddr, (unsigned int) portInfo->remotePort);

   /* Reconnect to assigned data port of DataConsumer */
   ST_DEBUG(5, "Before createSocket");

   /* Try now to connect to the Receiver's given data port */
   if (-1 == connectTo(pStation, flags, actEvent,
                       portInfo->remoteIPaddr, portInfo->remotePort))
   {
      return (-1);
   }

   ST_DEBUG(5, "Production line established (connectToReceiver)");

   /* Connection has been established. Use connected state prefix */
   actEvent->is_CR = 0;

   /* Implicit default is CD1 */
   if (CD1 != pStation->type)
   {
      uint32_t stationNameLng;

      /* Set payload for CD1.1 Option Request */
      *((uint32_t *)&buffer[36]) = htonl(1); /* option count */
      *((uint32_t *)&buffer[40]) = htonl(1); /* option type */
      stationNameLng = (uint32_t)strlen(dpName);
      *((uint32_t *)&buffer[44]) = htonl(stationNameLng); /* option size */
      /* Set station (data provider) name */
      memset(&buffer[48], 0, (size_t)(PAD(stationNameLng)));
      (void)strncpy(&buffer[48], dpName, LEN_STATION_NAME);
      stationNameLng = PAD(stationNameLng) + 48;
      /* Set Header and trailer for CD1.1 Option Request */
      setHeader(buffer, OPTION_REQUEST, stationNameLng,
                pStation->pConfigData->frameCreator,
                /*@-compdef@*/
                pProdLine->pDataConsumer->dcName,
                /*@=compdef@*/
                (uint64_t) 0);
      sendSize = setTrailerWithSig(pStation, buffer, stationNameLng,
                                   actEvent);

      if (-1 == sendSize)
      {
         /* Unexpected error while signing data */
         (void) close(portInfo->socket);
         portInfo->socket = -1;
         return(-1);
      }

      flags->firstTime = &(pStation->startRoW);
      flags->timeout = pStation->pConfigData->timeOut11;
      flags->resetDataTimeout = pStation->pConfigData->timeOut11;
      pStation->checkPass = 0;
      /* TODO: use sendFrame() here */
      /* Send port assignment frame */
      retVal = writeWithTimeout(portInfo->socket, buffer, sendSize, flags);
      (void) gettimeofday(&(pStation->endRoW), NULL);
      ST_DEBUG2(9, "ORF sent, returns [%d]", retVal);

      if (retVal != sendSize)
      {
         WHICH_EVENT(retVal, E_SEND_TIMEOUT);
         if (retVal > 0)
         {
            SAVE_DATA(0, retVal, E_SEND_TIMEOUT, GOOD_OPTION_REQ);
         }
         (void) close(portInfo->socket);
         portInfo->socket = -1;
         return(-1);
      }

      SAVE_DATA(0, retVal, S_SEND_OPTION_REQ, GOOD_OPTION_REQ);
      flags->timeout = pStation->pConfigData->timeOut11;

      /* Read CD1.1 Option Response Frame */
      retVal = readCDframe(pStation, actEvent, flags, &event, &clfType);

      /* Check event */
      if (S_SHUTDOWN == event) return(-1);

      /* There could be data read from the socket */
      if (retVal > 0)
      {
         SAVE_DATA(retVal, 0, event, clfType);
      }

      /* Check if processing should be aborted */
      if (1 == isBadEvent(event, pStation->systemConf & STRICT_MODE_ENABLE))
      {
         (void) close(portInfo->socket);
         portInfo->socket = -1;
         return(-1);
      }

      /* check for option response */
      if (clfType != GOOD_OPTION_RES && clfType != BAD_OPTION_RES)
      {
         logCDevent2(ET_L_ERROR, E_NO_ORS, actEvent, NO_INFO);
         (void) close(portInfo->socket);
         portInfo->socket = -1;
         return(-1);
      }

      ST_DEBUG(5, "Option response received");
   }

#undef WHICH_CONNECT_EVENT
#undef WHICH_EVENT

   /* Write some additional info to txt and clf file */
   addInfo2Files(pStation, S_CONNECTED);

   /* Sending thread is now connected */
   logCDevent2(ET_L_INFO, S_CONNECTED, actEvent, NO_INFO);
   pStation->state = STATION_CONNECTED;
   pStation->whenUnconnected.connectsOK++;

   return(0);
}

/**
 * @short Re-schedule unacknowledged frames for sending
 *
 * @description This function scans the internal reference list for
 * @description frames with the 'sent but not acknowledged' state
 * @description and set to state of these frames to unsent and unlinks
 * @description this frame from the sending history list. This will
 * @description cause that the frames are being sent next time sendIrlData
 * @description is called. This is save, because each frame has a
 * @description private sent counter
 *
 * @param pRefListHead  intRefList header
 *
 * @return nothing
 */

static void
setUnsent(intRefListHead_t *pRefListHead)
{
   intRefListParams_t *pCurrentElement;

   /* Go to the internal reference list in sort order and
      set the state of all unacknowledged frames to unsent. This is
      save, because the sent count is left untouched */
   pCurrentElement = pRefListHead->pRefListStart;

   while (NULL != pCurrentElement)
   {
      /* check if the state of the current element is sent acknowledged */
      if (SENT_NOACK == pCurrentElement->sendState)
      {
         /* check if this is the last frame sent */
         if (pCurrentElement == pRefListHead->pLastSent)
         {
            pRefListHead->pLastSent = pCurrentElement->pHistPrev;
         }

         /* Set state to unsent */
         pCurrentElement->sendState = NOT_SENT;
         /* Unlink the element from the sending history to prevent
            that the sending list becomes corrupted when the sending history
            contains old and new links and the IRL is sorted */
         UNLINK_FROM_SEND_HIST(pCurrentElement);
      }
      pCurrentElement = pCurrentElement->pNextElement;
   }
}

/**
 * @short Restore internal reference list
 *
 * @description This functions restores the internal reference
 * @description list when a re-start file is found. If no file
 * @description is found, re-send starts at the first frame in
 * @description the first clf file in the clf file list.
 *
 * @param pStation      station object
 * @param pRefListHead  intRefList header
 *
 * @return -1 on error, 1 if a corrupted restart file has been found and
 *         0 otherwise
 */

static int
restoreIntRefList(station_t *pStation, intRefListHead_t *pRefListHead)
{
   char *pFile2Search,
        logBuffer[LOGSIZE],
        fnBufferLastSent[LIBPAR_BUFSIZ],  /* buffer history line */
        fnBufferLastParsed[LIBPAR_BUFSIZ];
   unsigned long tmpOffset,
                 tmpOffsetLastParsed;
   clfFileParams_t *pClfFile,  /* current clf file list entry */
                   *pRestartFile = NULL; /* file containing the restart point */
   double restartNominalTime; /* nominal time of oldest/newest frame sent */
   uint16_t event = 0;
   int sortMode = UNSORTED;
   cdsend_t *pCdSend;  /* Application extension object */
   int retVal;

   if (NULL == pStation || NULL == pRefListHead->intRefListFp ||
       NULL == pStation->appExtension ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pClfFilesList)
   {
      shutdownApplication("Internal error in restoreIntRefList",
                          pStation->conID);
      return(-1);
   }

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   ST_DEBUG(2, "Restore intRefList");

   /* Set default values */
   pRefListHead->pLastSent = NULL;
   pRefListHead->frameOffset = 0;
   pRefListHead->lastDFFoffset = (uint32_t)0;
   pRefListHead->lastDFFsize = (uint32_t)0;
   pRefListHead->framesSinceLastSave = 0;
   pRefListHead->restartNominalTime = 0.0;

   if (0L == lseek(fileno(pRefListHead->intRefListFp), 0L, SEEK_END))
   {
      /* Empty restart file */
      char today[10];

      (void)snprintf(logBuffer, sizeof(logBuffer),
           "[%.17s] Nothing to restore", pStation->conID);
      logWrite(LOG_INFO, logBuffer);

      if (0 != (pCdSend->sndOpMode & OPER_MODE_FIXED))
      {
         /* Don't drop clf files in fixed interval mode */
         return(0);
      }

      /* Get current date as string */
      getToday(today);

      pClfFile = pCdSend->pClfFilesList->pClfFileHead;
      /* Start sending all clf files found in clf file list which
         are created today. Remove all other clf files from the list */
      while (NULL != pClfFile)
      {
         char *startSearch;

         /* Search last slash */
         startSearch = rindex(pClfFile->pFile, (int) '/');
         /* Check if slash was present in string */
         if (NULL == startSearch) { startSearch = pClfFile->pFile; }

         startSearch = index(startSearch, (int) '.');
         /* Check if dot was present in string */
         if (NULL == startSearch)
         {
            logError("Internal error", "inconsistent clf file name");
            pClfFile = pClfFile->pNextClfFileParams;
            continue;
         }
         /* Skip dot */
         startSearch++;
         /* Check if the clf file has been created today */
         ST_DEBUG2(2, "start compare at %s", startSearch);
         if (0 == strncmp(startSearch, today, 8))
         {
            pClfFile = pClfFile->pNextClfFileParams;
            continue;
         }
         ST_DEBUG2(2, "Delete %s from clf file list", pClfFile->pFile);

         /* clf file not created today.
            Get next element and unlink and destroy current element */
         pClfFile = unlinkClfElement(pStation, pClfFile, &event);
         if (NO_ERROR != event) break;
      }
      return(0);
   }

   /* Set file pointer to begin of file */
   rewind(pRefListHead->intRefListFp);

   /* Load processMode, last clf file parsed and offset of the last parsed
      frame in that clf file during processing */
   if (3 != fscanf(pRefListHead->intRefListFp, RESTART_LINE_1_IN,
                  &sortMode, fnBufferLastParsed, &tmpOffsetLastParsed))
   {
      logWarning("Can't scan restart file", pStation->conID);
      return(1);
   }
   (void)snprintf(logBuffer, sizeof(logBuffer),
           "[%.17s] mode %d, last clf file parsed [%s] and offset %lu",
           pStation->conID, sortMode, fnBufferLastParsed, tmpOffsetLastParsed);
   logWrite(LOG_INFO, logBuffer);

   /* Load clf file of the last frame sent, offset last frame sent and
      nominal time of oldest (LIFO) or newest (FIFO) frame */
   if (3 != fscanf(pRefListHead->intRefListFp, RESTART_LINE_2_IN,
                   fnBufferLastSent, &tmpOffset, &restartNominalTime))
   {
      logWarning("Can't scan restart file, line 2", pStation->conID);
      return(1);
   }
   (void)snprintf(logBuffer, sizeof(logBuffer),
          "[%.17s] last frame sent filename [%s] and offset %lu; RNT %f",
          pStation->conID, fnBufferLastSent, tmpOffset, restartNominalTime);
   logWrite(LOG_INFO, logBuffer);

   /* Add unsent or un-acknowledged frames before the restart point to
      the Internal Reference List. The sender does not know the protocol
      type of these frames. If the protocol type of the data provider
      has changed, these frames have to be removed manually from the
      restart file */
   retVal = loadIrlEntries(pRefListHead);
   if (0 != retVal) return (retVal);

   RH_DEBUG3(2, "After loadIrlEntries: %u, %u",
             (unsigned int)pRefListHead->nbrOfEntries,
             (unsigned int)sizeof(intRefListParams_t));

   /* Re-schedule unacknowledged frames for sending (set state to unsent) */
   /* (void)setUnsent(pRefListHead); */

   /* Check if last mode was unsorted. In that case only information
      about the last frame sent is needed for correct restart.
      Otherwise the nominal time of the last frame sent and the information
      of the last parsed clf file is needed */
   if (UNSORTED == sortMode)
   {
      pFile2Search = fnBufferLastSent;
      restartNominalTime = 0.0;
   }
   else
   {
      tmpOffset = tmpOffsetLastParsed;
      pFile2Search = fnBufferLastParsed;
   }

   /* Check if the re-start point is in the list of clf files to process */
   for (pClfFile = pCdSend->pClfFilesList->pClfFileHead;
        NULL != pClfFile;
        pClfFile = pClfFile->pNextClfFileParams)
   {
      ST_DEBUG2(2, "check clf file name [%.128s]", pClfFile->pFile);

      if (0 == strncmp(pClfFile->pFile, pFile2Search,
                       sizeof(fnBufferLastSent)))
      {
         /* Restart point found */
         (void)snprintf(logBuffer, sizeof(logBuffer),
                        "[%.17s] restart file found in clflist: [%.128s]",
                        pStation->conID, pFile2Search);
         logWrite(LOG_INFO, logBuffer);
         pRestartFile = pClfFile;
      }
   }

   /* Start to scan the list from begin */
   pClfFile = pCdSend->pClfFilesList->pClfFileHead;

   if (NULL == pRestartFile)
   {
      /* If no restart point is found, send the whole clf list */
      (void)snprintf(logBuffer, sizeof(logBuffer),
                     "[%.17s] no restart file found in clflist",
                     pStation->conID);
      logWrite(LOG_INFO, logBuffer);
   }
   else
   {
      /* The next step depends on last mode used. If the internal reference
         list was sorted as the re-start point was saved, all clf files
         before the restart point have to be parsed for data frames with
         nominal time older or newer (depending on the sorting) as the
         saved nominal time. Frames found in that case have to be included
         in the internal reference list. Parsing the clf file(s) is not
         needed and therefore skipped if the last mode was unsorted */
      while (NULL != pClfFile)
      {
         if (pClfFile == pRestartFile) break;

         if (UNSORTED != sortMode)
         {
	    int nbrOfNewEntries;

            /* Parse clf file for frames which have older (or newer)
               nominal time as the last frame sent and add them to
               internal reference list */
            pStation->type = pClfFile->cdtype;
            nbrOfNewEntries = parseClfFile(pClfFile->pFile, pStation->type,
                              pRefListHead, restartNominalTime, 0,
                              OLD_FILE, sortMode);
            ST_DEBUG3(2, "restoreIntRefList: %d new entries in %s",
		      nbrOfNewEntries, pClfFile->pFile);
            /* Close current clf file and check if clf file object
               should be removed */
            checkClfFileObj(pRefListHead);
         }
         /* get next element and unlink and destroy current element */
         pClfFile = unlinkClfElement(pStation, pClfFile, &event);
         if (NO_ERROR != event) return(-1);
      }

      /* The last file parsed has to be treated specially. Ignore all
         frames before the restart point (unsorted) or check the nominal
         time (sorted). Add all frames after the restart point regardless of
         their nominal time and last sorting mode. */
      if (NULL != pClfFile)
      {
         pStation->type = pClfFile->cdtype;
         (void)parseClfFile(pClfFile->pFile, pStation->type, pRefListHead,
                            restartNominalTime, tmpOffset,
                            RESTART_FILE, sortMode);

         /* Don't unlink the last clf file in the list!! It could grow */
         if (NULL != pClfFile->pNextClfFileParams)
         {
            /* Close current clf file and check if clf file object
               should be removed */
            checkClfFileObj(pRefListHead);
            /* get next element and unlink and destroy current element */
            pClfFile = unlinkClfElement(pStation, pClfFile, &event);
            if (NO_ERROR != event) return(-1);
         }
         else pClfFile = pClfFile->pNextClfFileParams;
      }
   } /* NULL != pRestartFile */

   /* The last step is to check if clf files newer than the restart point
      exist. If no restart point is known, it is assumed that all files
      are newer. In both cases all frames from these files have to be added to
      the internal reference list. This is done in the sending() function */
   return(0);
}


/**
 * @short Check for data to receive
 *
 * @description This functions waits until a given time has passed.
 * @description During this time it checks periodically if data have
 * @description been received and calls readData() to read the data
 *
 * @param pStation  pointer to station object
 * @param actEvent  current logging object
 *
 * @return none
 */

static uint16_t
check4receivedData(station_t *pStation, CDevent_t *actEvent)
{
   struct timespec sleeptime;
   uint16_t ii;
   senderParams_t *pSenderParams;

   /* Wait for sleep time worker thread to expire or user shutdown */
   /*@-usedef@*/
   sleeptime.tv_sec = 1;
   sleeptime.tv_nsec = 0;
   /*@=usedef@*/

   /* set sender specific config parameters */
   pSenderParams = (senderParams_t *)pStation->pConfigData->configExtension;

   for (ii = pSenderParams->sleepTimeWork;
        ii > 0 && SD_NONE == pStation->shutdownEnable; ii--)
   {
      /* Check if data is received */
      if (dataIsAvailable(pStation->portInfo->socket))
      {
         uint16_t event;

         /* Data is available, read it */
         event = readData(pStation, actEvent);

         /* Option Response frames are ignored in this state */
         if (S_OPTION_RES == event) event = NO_ERROR;

         if (NO_ERROR != event) return(event);
      }
      ST_DEBUG(5, "Before sleepTimeWork wait");
      /*@-usedef@*/
      (void) nanosleep(&sleeptime, NULL);
      /*@=usedef@*/
   }

   if (ii > 0) return(S_SHUTDOWN);
   else return(NO_ERROR);
}

/**
 * @short Check the pending event
 *
 * @description This function checks and logs the pending event. Depending
 * @description on the event it either logs the event and sends an Alert
 * @description Frame and closes the connection or it logs only the event.
 * @description If the connection has to be closed, it will also compact
 * @description the internal reference list to avoid re-sending too many
 * @description frames when the connection is re-established.
 *
 * @param pStation           pointer to station object
 * @param event              pending event or NO_ERROR
 * @param actEvent           current logging object
 * @param pRefListHead       pointer to head of IRL
 *
 * @return 0 if the sender thread should be stopped, 1 otherwise
 */

static int
checkEvent(station_t *pStation, uint16_t event, CDevent_t *actEvent,
           intRefListHead_t *pRefListHead)
{
   senderParams_t *pSenderParams;

   RH_DEBUG2(2, "checkEvent [%d]", (int)event);

   /* Log the event */
   if (S_SHUTDOWN == event || S_CLOSED_BY_AF == event)
   {
      logCDevent2(ET_L_INFO, event, actEvent, NO_INFO);
   }
   else if (NO_ERROR != event)
   {
      if (E_SYSTEM == event) pStation->savedSystemErr = (uint32_t)errno;
      logCDevent2(ET_L_ERROR, event, actEvent, (uint32_t)errno);
   }

   /* set sender specific config parameters */
   pSenderParams = (senderParams_t *)pStation->pConfigData->configExtension;

   /* Check if the station is really connected. If no data
      was available, station is not connected */
   if (STATION_CONNECTED == pStation->state)
   {
      cdsend_t *pCdSend;  /* Application extension object */

      /* Access application extension data */
      pCdSend = (cdsend_t *)pStation->appExtension;

      /* Close connection
         E_SEND_TIMEOUT, E_TIMEOUT,  E_SYSTEM, E_INTERNAL:
         Send Alert frame, close connection and try to re-open connection
         E_CLOSED and S_CLOSED_BY_AF:
         Try to re-open connection without sending an Alert Frame
         */
      if (S_CLOSED_BY_AF != event && E_CLOSED != event)
      {
         /* Send an alert frame when no alert frame received before */
         sendAF(pStation, actEvent);
      }

      /* Check if a history information should be saved */
      if (0 != pSenderParams->refListStoreInterval
          &&
          NULL != pCdSend /* for splint */
          &&
          0 == (pCdSend->sndOpMode & OPER_MODE_AUX))
      {
#ifdef ROBUST_RESTART
         /* Back up an existing previous written restart file */
         (void)backupRestart(pSenderParams->refListStorePath, pRefListHead,
                             pSenderParams->refListStoreInterval);
#endif /* ROBUST_RESTART */
         /* Add an additional restart point */
         saveIntRefList(getSortLevel(pStation),
                        pSenderParams->refListStoreInterval, pRefListHead);
      }
      else
      {
         /* Compact the internal reference list now to avoid re-send of
            too many frames when re-connecting */
         compactIntRefList(pRefListHead);
      }

      /* Write some additional info to txt and clf file */
      addInfo2Files(pStation, S_DISCONNECTED);

      (void)close(pStation->portInfo->socket);
      pStation->portInfo->socket = -1;
      /* Station has now disconnected */
      logCDevent2(ET_L_INFO, S_DISCONNECTED, actEvent, NO_INFO);
      pStation->state = STATION_WAIT2SEND;
   }

   if (S_CLOSED_BY_AF == event || E_CLOSED == event
       ||
       E_SEND_TIMEOUT == event || E_TIMEOUT == event
       ||
       E_SYSTEM == event)
   {
      /* Wait for retry time to expire and try to
         re-connect or user shutdown */
      retryWait(pStation, (int)pSenderParams->sleepTimeRetry);
      return(1);
   }

   /* User shutdown request or unexpected event. Stop sender thread */
   return(0);
}

/**
 * @short Sending thread - main function
 *
 * @description This function handles a connection to a data consumer.
 * @description It takes care that data to sent is available when a
 * @description connection is established and re-establish the connection
 * @description if the connection is closed unexpected
 *
 * @param arg  void pointer to config parameters
 *
 * @return None
 */

/*@null@*/ void *
sending(void *arg)
{
   CDevent_t actEvent;    /* Event buffer */
   const configParams_t *pConfigData;
   station_t *pStation;
   rwFlags_t flags;
   intRefListHead_t *pRefListHead;
   uint16_t event = NO_ERROR;
   int sortType;
   cdsend_t *pCdSend;  /* Application extension object */
   senderParams_t *pSenderParams; /* Configuration extension object */
   clfFileParams_t *pCurrentClf;

   /* Do some internal checks first */
   if (NULL == arg)
   {
      shutdownApplication("Internal error in sending",
                          "No config arg available");
      pthread_exit(NULL);
      return(NULL);
   }

   pStation = (station_t *) arg;     /* pointer to actual station_t */
   pConfigData = pStation->pConfigData;

   if (NULL == pConfigData || NULL == pStation->appExtension ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pClfFilesList ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pRefListHead ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine->pDataConsumer)
   {
      shutdownApplication("Internal error in sending",
                          "unexpected NULL pointer (1)");
      /* Set the state to unconnected, close all files and detach the thread */
      cleanUp(pStation);
      if (NULL != pConfigData)
      {
         pConfigData->pHistory->nbrOfThreadsStopped++;
      }
      pthread_exit(NULL);
      return(NULL);
   }

   /* set sender specific config parameters */
   pSenderParams = (senderParams_t *)pConfigData->configExtension;

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   ST_DEBUG2(2, "Thread [%ld] for Prodline started", (long)pthread_self());

   /* Check if FIX Interval mode */
   if (0 != (pCdSend->sndOpMode & OPER_MODE_FIXED))
   {
      /* Check if something is to do */
      if (NULL == pCdSend->pClfFilesList->pClfFileHead)
      {
         ST_DEBUG(2, "No entries in index file");
         /* Set the state to unconnected, close all files and
            detach the thread */
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         pthread_exit(NULL);
         return(NULL); /*for splint */
      }
   }

   /* state is wait for send */
   pRefListHead = pCdSend->pRefListHead;
   memset(&flags, 0, sizeof(flags));

   /* Set default value for mode with buggy AckNack frames */
   pRefListHead->unAckedHighMark = NBR_RESENT_FRAMES;

   /* Set event values */
   actEvent.conID = pStation->conID;
   actEvent.portInfo = pStation->portInfo;
   actEvent.is_CR = 1;
   actEvent.type = UNKNOWN;
   actEvent.lastCRFreceived = pStation->lastCRFreceived;

   /* Initialize all needed output files. This will
      open old file(s) if one exists for the current day or
      create new file(s). It's necessary to open the files
      *before* data is read, because the text file contains
      additional information */
   /* Sync files if wfdisc records should be written by sender
      Keep in mind that channels have to be parsed if wfdisc
      records should be logged!!
   if (-1 == syncFiles(pStation, NO, REOPEN_OLD_FILES, 0))
   {
      pStation->systemFailures++;
      logCDevent2(ET_L_ERROR, E_FILES, &actEvent, NO_INFO);
      cleanUp(pStation);
      pConfigData->pHistory->nbrOfThreadsStopped++;
      pthread_exit(NULL);
      return(NULL);
   }
   */

   /* Reset frame end time for ORDERED sending mode */
   pRefListHead->frameEndTime = 0;

   /* Try to open or create internal reference list file */
   if (1 == openIntRefList(pSenderParams->refListStorePath, pRefListHead, 0))
   {
      /* Set the state to unconnected, close all files and detach the thread */
      cleanUp(pStation);
      pConfigData->pHistory->nbrOfThreadsStopped++;
      pthread_exit(NULL);
      return(NULL); /*for splint */
   }

   /* Check if a saved internal reference list exists and if
      one exists and sender is not in auxiliary mode, load it */
   if (0 == (pCdSend->sndOpMode & OPER_MODE_AUX))
   {
      int retVal;

      retVal = restoreIntRefList(pStation, pRefListHead);
      if (-1 == retVal)
      {
         /* Set the state to unconnected, close all files and detach
            the thread */
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         pthread_exit(NULL);
         return(NULL); /*for splint */
      }
#ifdef ROBUST_RESTART
      if (1 == retVal)
      {
         /* Corrupt restart file detected. Try to open the backup file */
         if (1 == openIntRefList(pSenderParams->refListStorePath,
                                 pRefListHead, 1))
         {
            cleanUp(pStation);
            pConfigData->pHistory->nbrOfThreadsStopped++;
            pthread_exit(NULL);
            return(NULL); /*for splint */
         }
         /* Read the backup file */
         (void)restoreIntRefList(pStation, pRefListHead);
      }
#endif /* ROBUST_RESTART */
   }

   pCurrentClf = pCdSend->pClfFilesList->pClfFileHead;

   /* If a restart point was found, all frames after the restart point
      which are not sent before are added to the IRL now */
   if (0 != haveDataToSend(pCurrentClf,
                           pRefListHead->currentClfFile))
   {
      event = parseAllClfFiles(pStation, pRefListHead,
                               &pCurrentClf,
                               UNSORTED);
      if (NO_ERROR != event)
      {
         /* Set the state to unconnected,
            close all files and detach the thread */
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         pthread_exit(NULL);
         return(NULL); /*for splint */
      }
   }

   /*  Make sure that the production line is in STOPPED state when
       the sending order is changed!! */
   sortType = getSortLevel(pStation);

   /* The IRL is unsorted, check if the list should be sorted */
   RH_DEBUG2(2, "sortType = %d", sortType);
   IF_RH_DEBUG(10) { dumpIRL(pRefListHead); }

   if (UNSORTED != sortType)
   {
      /* Sort internal reference list now. If the sorting
         order (processing mode) can be changed using a command,
         make sure that the production line is in STOPPED state when
         the sending order is changed or the sorting order is not
         correct! */
      sortIntRefList(pRefListHead, sortType);
      IF_RH_DEBUG(10) { dumpIRLsendSequence(pRefListHead); }
      RH_DEBUG(2, "sorting finished");
   }

   /* The (re)start has been made. The IRL should contain now all unsent
      frames and should be in the correct sorting order */
   while (SD_NONE == pStation->shutdownEnable)
   {
      int unsentFrames = 0;           /* number of unsent frames in IRL */

      /* Reset all events */
      event = NO_ERROR;

      /* If no clf file has exist when the production line has started,
         pCurrentClf will never have a successor. So an additive check
         is made */
      if (NULL == pCurrentClf)
      {
         pCurrentClf = pCdSend->pClfFilesList->pClfFileHead;
      }

      /* Parse all clf files if new frames are available */
      if (0 != haveDataToSend(pCurrentClf,
                              pRefListHead->currentClfFile))
      {
         /* Add new frames to the internal reference list in the correct
            order */
         event = parseAllClfFiles(pStation, pRefListHead,
                                  &pCurrentClf,
                                  sortType);
      }

      if (0 != (pCdSend->sndOpMode & PROCESS_MODE_ORDERED))
      {
         /* Remove all too old frames from the IRL */
      }

      /* Query number of unsent frames in IRL */
      unsentFrames = getUnsentFrames(pRefListHead);

      /* Check if all frames have been processed in fix interval mode */
      if (0 == unsentFrames
          &&
          0 != (pCdSend->sndOpMode & OPER_MODE_FIXED)
          &&
          NULL == pCurrentClf)
      {
         /* Not sure if this check is really necessary */
         break;
      }

      if (STATION_CONNECTED != pStation->state)
      {
         /* Check if something is to do in continuous mode. Fixed mode
            has been checked before. Don't open a connection if
            nothing is to sent, or the receiver will close the
            connection when the line idle timeout is reached.
            Also don't start to send before data is available in
            the clf files or the station type may be set wrong.
            Check for NULL is necessary to wait for retry when no
            clf file is available, e. g. the last N files are read from
            the restart file, but no corresponding clf file exists */
         if (NULL == pCurrentClf
             ||
             0 == unsentFrames)
         {
            ST_DEBUG(2, "No un-sent frames or no clf file found");

            if (0 == (pCdSend->sndOpMode & OPER_MODE_FIXED))
            {
               /* Wait for some time to expire or user shutdown */
               retryWait(pStation, (int)pSenderParams->sleepTimeWork);
               continue;
            }
         }
      }

      while (unsentFrames > 0 && NO_ERROR == event)
      {
         /* Check for pending shutdown request */
         if (SD_NONE != pStation->shutdownEnable)
         {
            event = S_SHUTDOWN;
            break;
         }

         /* Station type is the type of the *last* clf file seen!! */

         /* Check if connected. The connection should be
            established after the clf file is parsed to ensure
            that no command frame in the clf file causes
            an unnecessary connection establishment */
         if (STATION_CONNECTED != pStation->state)
         {
            /* Use connection request prefix */
            actEvent.is_CR = 1;
            actEvent.type = pStation->type;
            /* Try to establish a connection to a receiver */
            if (-1 == connectToReceiver(pStation, &flags, &actEvent))
            {
               pStation->state = STATION_WAIT2SEND;
               /* Wait for retry time to expire or user shutdown */
               retryWait(pStation, (int)pSenderParams->sleepTimeRetry);
               continue;
            }

            /* Set unacknowledged frames high water mark. This is done
               here to know the station type for sure */
            if (0 == (pSenderParams->senderMode & BUGGY_ACKNACK_FRAMES) &&
                CD1_1 == pStation->type)
            {
               pRefListHead->unAckedHighMark = MAX_UNACKED_FRAMES;
               ST_DEBUG(2, "using MAX_UNACKED_FRAMES");
            }

            /* Re-schedule unacknowledged frames for sending (set state
               to unsent) to resend unacknowledged and last N frames.
               Keep in mind that re-sending the last N frames is done
               when at least one new frame is found in the clf file */
            (void)setUnsent(pRefListHead);

            /* Ensure that a DFF is sent after connection establishment */
            pRefListHead->pLastSent = NULL;
         }

         /* There is something to send in the internal reference list */
         event = sendIrlData(pRefListHead, pStation, &actEvent,
                             &pCurrentClf);
         ST_DEBUG2(5, "sendIrlData returns [%d]", (int)event);

         if (0 == pSenderParams->refListStoreInterval)
         {
            /* Normally the internal reference list is
               compacted in saveIntRefList(), but no re-start
               point is written if refListStoreInterval is set to 0 */
           compactIntRefList(pRefListHead);
         }

         /* Check for system error, closed connection or user shutdown */
         if (NO_ERROR != event)
         {
            break; /* Break while loop */
         }

         /* Check if the application couldn't open the current clf file */
         if (NULL == pRefListHead->currentClfFile)
         {
            /* Close current clf file and check if clf file object
               should be removed */
            checkClfFileObj(pRefListHead);

            /* Get next clf element and unlink and destroy current element.
               NOTE: If this was the last clf file, no restart information
               will be saved due lack of 'last clf file parsed' information.
               This will cause re-sending too many frames on the next
               re-start */
#ifdef OLD_CODE
            pCurrentClf = unlinkClfElement(pStation, pCurrentClf, &event);
#endif
         }
         else
         {
            /* Continuous mode and end of current clf file reached.
               It is possible that this file will grow, so
               this file has to be processed again */
            ST_DEBUG(2, "End of current clf file reached (CM)");

           /* Break loop to wait for received data */
           break;
         }

         ST_DEBUG2(5, "number of clf files in list [%d]",
                   pCdSend->pClfFilesList->nbrOfClfParams);

         /* Query number of unsent frames in IRL */
         unsentFrames = getUnsentFrames(pRefListHead);
         ST_DEBUG2(2, "number of unsent frames in list [%d]", unsentFrames);

      } /* while (unsentFrames > 0 && NO_ERROR == event) */

      if (NO_ERROR != event)
      {
         /* Check the event and determine what action has
            to be taken */
         if (1 == checkEvent(pStation, event, &actEvent, pRefListHead))
            continue;

        /* Break from while if user requested shutdown, or another
           internal error */
         break;
      }

      /* No event. Check for fixed interval mode */
      if (0 != (pCdSend->sndOpMode & OPER_MODE_FIXED))
      {
         /* Fixed interval mode and all frames in IRL sent */
         ST_DEBUG(2, "All frames sent in fixed interval mode");
         break;
      }

      /* Check if worker is connected, otherwise wait for
         data to send */
      if (STATION_CONNECTED != pStation->state) continue;

      /* Currently there is nothing to send, but the sender
         has to check for received data. In this state the sender
         waits for received data with a timeout of "sleepTimeWork"
         and then checks for data to send again */
      event = check4receivedData(pStation, &actEvent);
      ST_DEBUG2(2, "After check4receivedData [%d]", (int)event);
      if (S_ACKNAK == event &&
          0 == (pSenderParams->senderMode & BUGGY_ACKNACK_FRAMES))
      {
         /*@dependent@*/ CDframeset_t *ackFrameSet = NULL;

         /* Parse AckNack frame. Result is a frame set. The frame
            set  specific list is valid until the next AckNack
            frame for the same frame set is received.
            Only valid AckNack frames are treated further */
         ackFrameSet = parseAckNack(pStation->frameBuffer,
                                    pStation->header.trailerOffset,
                                    &(pCdSend->frameSets), &event);

         /* Ignore invalid AckNack frames in none strict mode */
         if (NO_ERROR != event &&
             0 == (pStation->systemConf & STRICT_MODE_ENABLE))
         {
            event = NO_ERROR;
         }

         /* Check if an AckNack Frame has been received */
         if (NULL != ackFrameSet && event == NO_ERROR &&
             NULL != pRefListHead->pLastSent)
         {
            /* Now acknowledged all frames which have been sent and which are
               not part of a gap and remove the send links from unacknowledged
               frames and move them to the top of the IRL */
            event = acknowledgeFrames(pRefListHead, ackFrameSet);
            ackFrameSet = NULL;
         }
      }

      /* Ignore AckNack frame when in buggy AckNack mode.
         Ignore Option Response frame in this state */
      if (S_OPTION_RES == event || S_ACKNAK == event) event = NO_ERROR;

      if (NO_ERROR != event)
      {
         /* Check the event and determine what action has
            to be taken */
         if (1 == checkEvent(pStation, event, &actEvent, pRefListHead))
            continue;

        /* break from while if user requested shutdown, or another
           internal error */
         break;
      }
   } /* while (SD_NONE == pStation->shutdownEnable) */

   /* Check if state is unconnected, otherwise send alert frame */
   if (STATION_CONNECTED == pStation->state)
   {
      /* Send an alert frame when no alert frame is received before */
      sendAF(pStation, &actEvent);

      (void)close(pStation->portInfo->socket);
      pStation->portInfo->socket = -1;
      /* Station has now disconnected */
      logCDevent2(ET_L_INFO, S_DISCONNECTED, &actEvent, NO_INFO);
      pStation->state = STATION_WAIT2SEND;

      /* Write some additional info to txt and clf file */
      addInfo2Files(pStation, S_DISCONNECTED);

      /* Check if a history information should be saved */
      if (0 != pSenderParams->refListStoreInterval
          &&
          0 == (pCdSend->sndOpMode & OPER_MODE_AUX))
      {
         /* Add an additional restart point */
         saveIntRefList(getSortLevel(pStation),
                        pSenderParams->refListStoreInterval, pRefListHead);
      }
   }

   /* Clear internal reference list */
   clearIntRefList(pRefListHead);
   (void)fclose(pRefListHead->intRefListFp);
   pRefListHead->intRefListFp = NULL;
   /* Check if  current clf file has to be closed and check if clf file object
      has no more references and can be removed */
   checkClfFileObj(pRefListHead);
   /* Set the state to unconnected or stopped,
      close all files and detach the thread */
   cleanUp(pStation);
   pConfigData->pHistory->nbrOfThreadsStopped++;
   ST_DEBUG(2, "Sending thread exits");
   pthread_exit(NULL);
   /*@-mustfreeonly@*/
   return(NULL); /*for splint */
   /*@=mustfreeonly@*/
}
