/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: cdDataGen.c,v $
 *
 * Revision 1.9  2013-05-22 gerald
 * Follow changes in libcancomp
 *
 * Revision 1.8  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.7  2011-03-02 15:39:28  gerald
 * Update code for signing callback function
 *
 * Revision 1.6  2011-02-10 10:20:03  gerald
 * Typo fixed
 *
 * Revision 1.5  2009-11-16 14:37:41  gerald
 * Swap start time and days command line parameter
 *
 * Revision 1.4  2009-09-29 12:41:55  gerald
 * Fix wrong data type when compiled on Solaris
 *
 * Revision 1.3  2009-09-11 13:52:10  gerald
 * Set authflag to 1 if compiled with authentication
 *
 * Revision 1.2  2009-09-09 09:02:23  gerald
 * Fix wrong parsing of CLI arguments and some splint warnings
 *
 * Revision 1.1  2009-07-20 16:52:25  gerald
 * Add first release of cdDtaGen
 *
 *
 */

#include <config.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include "cd_common.h"
#include "events.h"
#include "readwrite.h"
#include "init.h"
#include "senderAPI.h"
#include "canada_compress.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

#define ONE_DAY  86400


/**
 * @short check for an error
 *
 * @description This function checks for an error and stops
 * @description the application with an error message if an
 * @description error is set
 *
 * @param error         error value
 * @param where         name of function which returns the error
 * @param addinfo       additional info
 *
 * @return nothing
 */

static void
check4error(int error, char *where, int addinfo)
{
   /* Check for error */
   if (0 != error)
   {
      fprintf(stderr, "error [%d] in [%.40s] [%d]\n", error, where, addinfo);
      exit(EXIT_FAILURE);
   }
}

#ifdef HAVE_AUTHENTICATION

/**
 * @short create data signature
 *
 * @description This function creates a signature for the given data
 *
 * @param data        data to sign
 * @param dataLen     length data to sign
 * @param signature   signature of data (returned)
 * @param pChannelRef pointer to application extension structure
 * @param sigLen      maximum length of signature IN/length of signature OUT
 *
 * @return  0 on success and -1 if an error has occurred
 */

static int
sampleSignData(const char *data, size_t dataLen, /*@out@*/ char *signature,
               /*@unused@*/channelRef_t *pChannelRef, size_t *sigLen)
{
   configParams_t *pConfigData = getConfigPointer();

   if (NULL != pConfigData->privateKey)
   {
      /* Add your own signing function here */
      return(signData(pConfigData->privateKey, data, dataLen,
                      signature, sigLen));
   }
   else
   {
      *signature = '\0'; /* for splint */
      return(-1);
   }
}

#endif /* HAVE_AUTHENTICATION */

static void
createData(station_t *pStation, int startTime, int days)
{
#define DURATION   10
#define SAMPRATE   40
#define NSAMP      (DURATION * SAMPRATE)
   int error = 0,
       ij,
       ik,
       im,
       ii;
   channelData_t channelData;
   channelRef_t *pChannelRef;
   channelDescr_t *pChannelDescr;
   unsigned char compressBuf[NSAMP * 4];
   uint32_t inBuf[NSAMP + 1],
            v0 = 0L;
   int lngCompressBuf = 0;
   int channels;
   unsigned long channelIDs[MAX_CHANNELS];
   unsigned long currentValues[MAX_CHANNELS];


   if (NULL == pStation || NULL == pStation->appExtension)
   {
      stopProgram("dataGen", "error in create data function",
                  "unexpected NULL pointer");
      exit(EXIT_FAILURE); /* for splint */
   }
   pChannelRef = (channelRef_t *)pStation->appExtension;
   compressBuf[0] = '\0'; /* for splint */

   /* initialise array with channel IDs */
   channels = (int)pChannelRef->nbrOfChannels;
   for (ii = 0; ii < channels; ii++)
   {
       sensor_t *pSensor;

       pChannelDescr = &(pChannelRef->channelDes[ii]);

       (void)fprintf(stdout,"Create data for %.9s/%.9s\n",
                     pChannelDescr->siteNameAPI,
                     pChannelDescr->channelNameAPI);
       (void)fflush(stdout);

       pSensor = getSensorDescr(pChannelDescr->siteNameAPI,
                                pChannelDescr->channelNameAPI);

       if (NULL == pSensor)
       {
          stopProgram("dataGen", "NULL sensor description", "createData");
          exit (EXIT_FAILURE);
       }
       /* Set channel IDs for channel with index ii */
       channelIDs[ii] = (unsigned long)pSensor->chanID;
       currentValues[ii] = 0L;
   }

   /* For all requested days */
   for (im = 0; im < days; im++)
   {
      (void)fprintf(stdout,"Day %d\n", im + 1);
      (void)fflush(stdout);

      /* Generate data for a given station for one day. Each channel subframe
         contains the channel ID and the start time as first two samples and a
         constant increasing value for the remaining data samples */
      for (ij = 0; ij < ONE_DAY/DURATION; ij++)
      {
         /* Go through all channels and generate channel data */
         for (ii = 0; ii < channels; ii++)
         {
            unsigned long curValue;

            /*@-usedef@*/
            /* Set channel start value */
            curValue = currentValues[ii];
            /* Set channel ID as first sample */
            inBuf[0] = channelIDs[ii];
            /*@=usedef@*/
            /* Set start time of CSF as the second sample */
            inBuf[1] = (unsigned int)(startTime & 0xFFFF);
            /*@-shiftimplementation@*/
            inBuf[2] = (unsigned int)((startTime >> 16) & 0xFFFF);
            /*@=shiftimplementation@*/
            for (ik = 3; ik < NSAMP; ik++)
            {
               inBuf[ik] = curValue++;  /* Set a constant increasing value as
                                           value for the remaining samples */
            }
            /* Compress the data now */
            error = canada_compress(compressBuf, inBuf, &lngCompressBuf,
                                    NSAMP, &v0);
            check4error(error, "canada_compress", ii);

            memset(&channelData, 0, sizeof(channelData));

            /* Set channel time stamp */
            channelData.timeStampAPI = (double)(startTime);

            /* Set length of samples data */
            channelData.channelDataSizeAPI = (uint32_t)lngCompressBuf;
            /* channelData is copied in setChannelData, so buffer can
               be re-used. */
            channelData.channelDataAPI = (char *)compressBuf;
            /* Subframe time length is in miliseconds */
            channelData.subframeTimeLengthAPI = DURATION * 1000;
            /* Set number of samples */
            channelData.nbrOfSamplesAPI = NSAMP;
            /* Set channel status size */
            channelData.channelStatusSizeAPI = 4;
            /* channelStatusData is copied in setChannelData, so buffer can
               be re-used. */
            channelData.channelStatusDataAPI = "\1\1\1\1";
            channelData.subframeCountAPI = 0;
            /*
            channelData.authKeyIdentifierAPI =
                        getConfigPointer()->privateKeyID;
            channelData.signatureSizeAPI = 40;
            channelData.signatureAPI = signature;
            */
            channelData.authKeyIdentifierAPI = 0;
            channelData.signatureSizeAPI = 0;
            /* signature is copied in setChannelData, so buffer can
               be re-used. */
            channelData.signatureAPI = "";

            /* Set Channel Data */
            error = setChannelData(pStation, ii, &channelData);
            /* Check for error */
            check4error(error, "setChannelData", ii);
            /* Remember the channel start value */
            currentValues[ii] = curValue;
         } /* End for (ii = 0; ii < channels; ii++) */
         startTime += DURATION;

         /* Write frame now */
         error = writeFrame(pStation, 0, "");
         /* Check for error */
         check4error(error, "writeFrame", -1);
      } /* End for (ij = 0; ij < ONE_DAY/DURATION; ij++) */
   } /* End for (im = 0; im < days; im++) */
}

/**
 * @short print usage message
 *
 * @description This function prints a usage message and exists with
 * @description EXIT_FAILURE
 *
 * @param pArgv          pointer to application name
 *
 * @return nothing
 */

static void
usage(char *pArgv0)
{
   fprintf(stderr,
      "usage: %s config_file sensor_file CD1|CD11 [starttime [days]]\n",
           pArgv0);
   exit(EXIT_FAILURE);
}

/**
 * @short sample sender API main function
 *
 * @description This function initializes the data source, sets
 * @description the channel description, reads data channels from the data
 * @description source, and writes new data frames.
 *
 * @param argc          argument count
 * @param argv          argument vector
 *
 * @return 0 on success or != 0 on error
 */

int
main (int argc, char *argv[])
{
   channelDescr_t channel;
   station_t *pStation = NULL;
   int ii,
       days = 1,
       sensorTableCount = 0,
       channels = 0;
   sensor_t *sensorDesc = NULL;
   time_t startTime;
   char lastStream[LEN_STATION_NAME+1];
   unsigned char cdProtocol = CD1;

   /* Check command line arguments */
   if (argc < 4)
   {
      usage(argv[0]);
   }

#define APP_NAME "dataGen"

   /* Load sensor description */
   sensorDesc = loadSensorFromFile(APP_NAME, argv[2], &sensorTableCount);
   DEBUG2(2, "%d lines loaded from sensor table", sensorTableCount);

   if (NULL != sensorDesc)
       setSensorDescr(sensorDesc, sensorTableCount, APP_NAME);
   else
   {
      stopProgram(APP_NAME, "Can't load sensor file", NULL);
      exit(EXIT_FAILURE); /* for splint */
   }
   memset(lastStream, 0, sizeof(lastStream));
   if (strncmp(argv[3], "CD11", 4) == 0)
   {
      cdProtocol = CD1_1;
   }
   else if (strncmp(argv[3], "CD1", 3) == 0)
   {
      cdProtocol = CD1;
   }
   else
   {
      usage(argv[0]);
   }


   /* Check for alternative start date */
   if (argc > 4)
   {
      startTime = ascDate2int(argv[4], "%4d/%2d/%2d %2d:%2d:%2d");
      if ((time_t)-1 == startTime)
      {
         stopProgram(APP_NAME, "Invalid value in start date", NULL);
      }
   }
   else
   {
      startTime = time(NULL);
   }
   /* Check for alternative number of days */
   if (argc > 5)
   {
      days = (int) strtol(argv[5], NULL, 10);
      if (days <= 0 || days > 24)
      {
         stopProgram(APP_NAME, "Invalid value for days (<= 0 or > 24)", NULL);
      }
   }

   /* Start on 00:00 */
   startTime -= startTime % ONE_DAY;

   /* Go back days - 1 days */
   startTime -= (days -1) * ONE_DAY;

   /* Now loop through the sensor table and create a station object
      for each stream found */
   for (ii = 0; ii < sensorTableCount; ii++)
   {
      IF_DEBUG (2)
      {
         DEBUG_HDR;
         (void) fprintf(DEBUG_FP,
            "actSensor: %.6s %.6s %.6s %.8s %ld %.6s %f %f %f %f %f %d %d\n",
                        sensorDesc[ii].stream,
                        sensorDesc[ii].intern_sta, sensorDesc[ii].extern_sta,
                        sensorDesc[ii].intern_chan,
                        sensorDesc[ii].chanID, sensorDesc[ii].insType,
                        sensorDesc[ii].nomCalib, sensorDesc[ii].nomCalibPer,
                        sensorDesc[ii].expSampleRateInDB,
                        sensorDesc[ii].startTime, sensorDesc[ii].endTime,
                        sensorDesc[ii].revision,
                        sensorDesc[ii].position);
          DEBUG_FOOTER;
      }

      /* Check if this is a new stream */
      if (strncmp(lastStream, sensorDesc[ii].stream, LEN_STATION_NAME) != 0)
      {
         /* Check if this is the first stream */
         if (0 != channels)
         {
            if (NULL == pStation) /* For splint */
            {
               stopProgram(APP_NAME, "Unexpected NULL pointer pStatione", NULL);
               exit(EXIT_FAILURE); /* for splint */
            }
            createData(pStation, (int)startTime, days);
            channels = 0;
            /* Delete channel information and free cache memory */
            (void)clearChannelInfo(pStation);
         }

         /* Allocate and set station object. A configuration file is needed
            for openBuffer */
         pStation = openBuffer(argv[1], sensorDesc[ii].stream, cdProtocol);
         DEBUG2(2, "+++++++ Opened station for %.9s", sensorDesc[ii].stream);

         if (NULL == pStation)
         {
            fprintf(stderr, "openBuffer failed\n");
            exit(EXIT_FAILURE);
         }

#ifdef HAVE_AUTHENTICATION
         /* Set callback for own signing function */
         if (NULL != pStation->appExtension)
            ((channelRef_t *)(pStation->appExtension))->callBackSign =
                                                        sampleSignData;
#endif /* HAVE_AUTHENTICATION */

         strncpy(lastStream, sensorDesc[ii].stream, LEN_STATION_NAME);
      }

      /* Set channel description for the channels of the stream */
      channel.siteNameAPI = sensorDesc[ii].extern_sta;
      channel.channelNameAPI = sensorDesc[ii].intern_chan;
      channel.locationNameAPI = "";
#ifdef WORDS_BIGENDIAN
      channel.datatypeAPI = "s4";
#else
      channel.datatypeAPI = "i4";
#endif
      channel.sensorTypeAPI = '\0';
      channel.optionFlagAPI = '\1';
#ifdef HAVE_AUTHENTICATION
      channel.authFlagAPI = '\1';
#else
      channel.authFlagAPI = '\0';
#endif
      channel.compressionAPI = '\1';
      channel.calibFactorAPI = 1;
      channel.calibPeriodAPI = 1;
      channels++;

      if (NULL == pStation) /* For splint */
      {
         stopProgram(APP_NAME, "Unexpected NULL pointer pStatione", NULL);
         exit(EXIT_FAILURE); /* for splint */
      }

      /* Initialize data channel information */
      if (initChannel(pStation, &channel) < 0)
      {
         fprintf(stderr, "can't initialize channel %.6s/%.9s\n",
                 sensorDesc[ii].extern_sta, sensorDesc[ii].intern_chan);
         exit(EXIT_FAILURE);
      }
   }

   /* Check if this is the first stream */
   if (0 != channels)
   {
      if (NULL == pStation) /* For splint */
      {
         stopProgram(APP_NAME, "Unexpected NULL pointer pStatione", NULL);
         exit(EXIT_FAILURE); /* for splint */
      }
      createData(pStation, (int)startTime, days);
      channels = 0;
      /* Delete channel information and free cache memory */
      (void)clearChannelInfo(pStation);
   }

   exit(EXIT_SUCCESS);
}
