
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * $Log: explain.c,v $
 *
 * Revision 1.7  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.6  2010-12-13 09:43:59  gerald
 * Fix usage message
 *
 * Revision 1.5  2010-11-30 18:15:39  gerald
 * Add channel status and quality flag conversion to human
 * readable form to explain
 *
 * Revision 1.4  2007/11/22 12:49:13  gerald
 * Add explanation of errno error code
 *
 * Revision 1.3  2005/02/08 11:03:32  gerald
 * Enhance explain tool to explain also nominal times
 *
 * Revision 1.2  2005/01/21 10:58:43  gerald
 * Fix splint warnings and add explanation of CD events
 *
 * Revision 1.1  2004/11/12 15:06:15  gerald
 * Add new helper tool 'explain'
 *
 */

#include <config.h>
#ifdef S_SPLINT_S
#include <netinet/in.h>  /* Need this to parse code with splint */
#endif /* S_SPLINT_S */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "cd_common.h"
#include "events.h"
#include "logfiles.h"
#include "inittables.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

/**
 * @short Print usage message
 *
 * @description This function prints a usage meassage
 *
 * @param argv  Vector with CLI arguments
 *
 * @return Nothing
 */

static void
usage(char *argv[])
{
   printf("usage: %s [-[f|c|e|n|s] value \n", argv[0]);
   printf(" -f interprets value as frame parsing code (default)\n");
   printf(" -c interprets value as channel parsing code\n");
   printf(" -e interprets value as event code\n");
   printf(" -n interprets value as nominal time\n");
   printf(" -s interprets value as sysconfig value\n");
   printf(" -E interprets value as errno value\n");
   printf(" -S interprets value as CD11 channel status value\n");
   printf(" -q interprets value as CD quality flag\n");
   exit(EXIT_FAILURE);
}

/**
 * @short Check command line arguments
 *
 * @description This function checks the command line arguments
 *
 * @param argc  Number of CLI arguments
 * @param argv  Vector with CLI arguments
 *
 * @return Nothing
 */

static void
checkCommandLine(int argc, char *argv[])
{
   /* Check if 3 arguments are provided */
   if (argc < 2 || argc > 3) usage(argv);
}

/**
 * @short Explain main function
 *
 * @description This function handles the CLI arguments and calls
 * @description the appropriate functions
 *
 * @param argc  Number of CLI arguments
 * @param argv  Vector with CLI arguments
 *
 * @return EXIT_FAILURE or EXIT_SUCCESS
 */

int
main (int argc, char *argv[])
{
   long value = 0;
   double dvalue = 0.0;
   char *endptr;
   int base = 10;

   /* Load idcsyslog functions to avoid undefined references when compiling */
   if (argc == 100) (void) getDebugLevel();

   /* Check command line argument(s) */
   (void) checkCommandLine(argc, argv);

   if (argv[argc-1][1] == 'x') base = 16;

   /* Check if value is a  number */
   if (argv[1][1] == 'n')
   {
      dvalue = strtod(argv[argc-1], &endptr);
      if (*endptr != '\0') usage(argv);
   }
   else
   {
      value = strtol(argv[argc-1], &endptr, base);
      if (*endptr != '\0') usage(argv);
   }

   if (argc == 3)
   {
      if (argv[1][1] == 's')
      {
         /* Dump value as a sysConf value in human readable form */
         fprintf(stdout, "systemConf value [%s] is:\n", argv[argc-1]);
         decodeSystemConf((uint32_t)value);
      }
      else if (argv[1][1] == 'c')
      {
         fprintf(stdout, "channel parsing code [%s] is:\n", argv[argc-1]);
         /* Dump value as a channel parsing code in human readable form */
         printChannelParsingCodeAsText((uint32_t)value, stdout);
      }
      else if (argv[1][1] == 'e')
      {
         /* Dump value as a channel parsing code in human readable form */
         fprintf(stdout, "event code [%s] is: %.64s\n", argv[argc-1],
                 getCDevent((uint16_t)value));
      }
      else if (argv[1][1] == 'n')
      {
         /* Dump value as a nominal time in human readable form */
         fprintf(stdout, "nominal time [%s] is: ", argv[argc-1]);
         double2ascDate(dvalue, stdout);
         fprintf(stdout, "\n");
      }
      else if (argv[1][1] == 'E')
      {

         /* Dump value as a errno in human readable form */
         fprintf(stdout, "errno [%s] is: ", argv[argc-1]);
         fprintf(stdout, "%s",
                 NULL == strerror(value) ? "unknown" : strerror(value));
         fprintf(stdout, "\n");
      }
      else if (argv[1][1] == 'S')
      {

         /* Dump value as CD11 channel status code in human readable form */
         fprintf(stdout, "CD11 channel status [%s] is:\n ", argv[argc-1]);
         printCD11ChannelStatusAsText((uint32_t)value, stdout);
         fprintf(stdout, "\n");
      }
      else if (argv[1][1] == 'q')
      {

         /* Dump value as CD quality flag in human readable form */
         fprintf(stdout, "CD quality flag [%s] is:\n ", argv[argc-1]);
         printCDqualityFlagAsText((uint32_t)value, stdout);
         fprintf(stdout, "\n");
      }
      else
      {
         fprintf(stdout, "frame parsing code [%s] is:\n", argv[argc-1]);
         /* Dump value as a frame parsing code in human readable form */
         printFrameParsingCodeAsText((uint32_t)value, stdout);
      }
   }
   else
   {
      fprintf(stdout, "frame parsing code [%s] is:\n", argv[argc-1]);
      /* Dump value as a frame parsing code in human readable form */
      printFrameParsingCodeAsText((uint32_t)value, stdout);
   }

   exit(EXIT_SUCCESS);
}
