
/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Gerald Klinkl
 *
 * Revision 1.85  2016-04-12 GK
 * Fix comment
 *
 * Revision 1.84  2015-11-25 gerald
 * Fix incorrect end time calculation when loading
 * quality records, which could cause incorrect split
 * of quality records
 *
 * Revision 1.83  2015-09-10 gerald
 * Fix typo
 *
 * Revision 1.82  2015-04-21 gerald
 * Move duplicate code segments for channel filters
 * to libCD
 *
 * Revision 1.81  2014-10-21 gerald
 * Fix check for maximum number of channels
 *
 * Revision 1.80  2014-07-07 gerald
 * Fix time used for frame signature check
 *
 * Revision 1.79  2014-05-26 gerald
 * Fix incorrect skipping of frames when maxDeviation has been set
 * and frame was too close to real time
 *
 * Revision 1.78  2014-03-17 gerald
 * Remove all related Data Frames from IRL when Data Format
 * frame was corrupted. Fix incorrect constant data detection
 * if not at begin of frame
 *
 * Revision 1.77  2014-02-05 gerald
 * Fix comments and do not parse frames when frame header
 * is corrupted (see ISI-206). Move deleteWithThatFileObject()
 * to libCD
 *
 * Revision 1.76  2014-02-05 gerald
 * Fix some statistics logged to the wrong place (whenUnconnected
 * instead of whenConnected). Write commit info to fields in (error)
 * statistic structure which are not used in cd2wng before. Commit
 * the whole window for one station instead of each day separately.
 *
 * Revision 1.75  2013-11-01 gerald
 * Fix 'constant data' detection off by one sample.
 * Supersede 'constant data' by 'no input from sensor'.
 * Do not longer trigger a commit if a single channel is synced.
 * Commit by station instead (one per day). Adjust debug
 * level for some debug messages.
 *
 * Revision 1.74  2013-10-16 gerald
 * Re-work update the frameproduct table to minimize
 * incorrect values for binary frame files
 *
 * Revision 1.73  2013-10-12 gerald
 * Update frameproduct when closing a binary file only when station
 * type is CD-1.1 or a DFF is processed
 *
 * Revision 1.72  2013-08-20 gerald
 * Follow API changes in getChMaskPosition()
 * Fix to many time based syncs when a thread is waiting for data
 *
 * Revision 1.71  2013-07-12 gerald
 * Move WFDISC_WCS30 and WFDISC_WCS30_IN define to libcd/logfiles.h
 * Write 7 significant digits for ncalib to wfdisc
 * (see ISI-126)
 *
 * Revision 1.70  2013-07-05 gerald
 * Move 'getODBCProvider' to db_access.c
 *
 * Revision 1.69  2013-06-26 gerald
 * Replace 'sensor noise' by 'no input from sensor'
 *
 * Revision 1.68  2013-06-17 gerald
 * Add function to query ODBC provider library
 *
 * Revision 1.67  2013-06-06 gerald
 * Do not longer stop when updateReceiverFrameProduct() does not
 * find the binary file (Fix for ISI-56)
 *
 * Revision 1.66  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.65  2013-02-25 gerald
 * Add per thread time based update of
 * restart/wfdisc/cdquality information
 *
 * Revision 1.64  2012-12-20 gerald
 * Fix compiler warning on ARM
 *
 * Revision 1.63  2012-10-09 gerald
 * Fix comment
 *
 * Revision 1.62  2012-07-20 gerald
 * Fix internal error when cdquality records overlap, which
 * could happen when a database problem prevents the deletion
 * cdquality segments in a previous run of cd2wng. Improve
 * logging of debug messages
 *
 * Revision 1.61  2012-06-18 gerald
 * Use more safe but slower approach for dealing with consecutive
 * wfdisc records during loading (delete and re-create). This fixes
 * WFDISCSCTEX constraint violations when the wfdisc table contains
 * records which should have been already merged
 *
 * Revision 1.60  2012-04-20 gerald
 * Remove all segments which are scheduled for release first to
 * avoid a constrain violation if HAVE_WFDISCSCTEX is set (fixes
 * ISI-47)
 *
 * Revision 1.59  2012-04-01 gerald
 * Fix merging frames with frame time length not a multiple
 * of a second (Fixes ISI-29)
 *
 * Revision 1.58  2012-03-19 gerald
 * Add frame authentication state and more detailed
 * authentication state for channels to quality flags
 *
 * Revision 1.57  2011-12-05 17:26:51  ctbto
 * Fix too long code lines and set default value for
 * "unacknowledged" high water mark to NBR_RESENT_FRAMES to
 * maintain backwards compatibility for restart tests
 *
 * Revision 1.56  2011-11-02 15:44:15  ctbto
 * Fix zero file size for waveform files in frameproduct
 *  table and tweak some debug message levels
 *
 * Revision 1.55  2011-09-27 08:49:48  ctbto
 * Fix compilation without database access if
 * sqltypes.h is not installed. Use different
 * variable types for recycleIds depending on
 * IDC ODBC access library used. Clean up old
 * comments
 *
 * Revision 1.54  2011-06-08 07:56:09  ctbto
 * Clean up old comment
 *
 * Revision 1.53  2011-05-16 09:21:30  gerald
 * Fix function header comment
 *
 * Revision 1.52  2011-04-29 09:58:20  gerald
 * Take frame product type ID from variable instead of build in
 * types and do not longer create entries for CD text files in the
 * frameproduct table
 *
 * Revision 1.51  2011-04-27 15:30:23  gerald
 * Remember current waveform file size for frameproduct table.
 * Rework updateFrameProduct to get rid of dependencies to internal
 * reference list. Update waveform file size when syncing wfdisc
 * records
 *
 * Revision 1.50  2011-02-10 10:14:07  gerald
 * Skip too short frames (<1s) when merging. This change fixes the
 * FP exceptions when cd2ng try to merge too short frames
 *
 * Revision 1.49  2011-01-17 14:58:49  gerald
 * Reset sensor noise and constant data flag when below minimum
 * sample rate for detection. Do not longer check absolute wfdisc path
 * configuration parameter in cd2wng. Relative wfdisc paths never
 * worked correct in cd2wng and causes a lot of troubles in other
 * tools (e.g. cdqual).  If the path length is too short symbolic
 * links should be used instead. If this is not an option, the
 * filePath configuration can be used to simulate absWfdiscPath=NO,
 * e.g. filePath=./
 *
 * Revision 1.48  2010-12-13 16:37:11  gerald
 * Save waveform file name in frameproduct table. Replace dfile
 * with file ID from frameproduct table. Add ngoodsamp. Fix missing
 * quality records when waveform segment is stored at the end of
 * the waveform file (needed when segment offset is stored in the
 * waveform file)
 *
 * Revision 1.47  2010-11-30 19:07:07  gerald
 * Switch to cdquality table. Split cdquality records based
 * on different channel attributes. Map CD1 to CD1_1 attributes.
 * Start work on waveform file ID
 *
 * Revision 1.46  2010-11-11 16:43:20  gerald
 * Make sure that number of frames since last save is increased
 * when a shutdown is initiated and this was the first frame
 * processed. Add more comments. Add check if frame nominal
 * time belongs to future time. Do not longer write channels without
 * channel info to the waveform file in database mode
 *
 * Revision 1.45  2010-10-27 11:43:24  gerald
 * Rename structure member segSumData to segMean and
 * segSumSquareData to segM2, and compute RMS as the
 * root square count of the differences between mean and count.
 * Do not write quality records in DB mode if channel not known
 * in DB (PES-526)(fix for the quality with no related wfdisc record
 * problem in DB mode)
 *
 * Revision 1.44  2010-10-18 09:45:40  gerald
 * Return more detailed error code in case of an error and fix
 * follow up errors. Add second DB connection to get rid of
 * ODBCGetNextId() timeouts during start up of cd2wng.
 * This change requires also a fix in libgODBC
 *
 * Revision 1.43  2010-07-07 14:22:23  gerald
 * Fix incorrect quality record end time which was off by one sample
 *
 * Revision 1.42  2010-05-11 15:25:59  gerald
 * Fix missing authentication state in quality table, add more
 * detailed log messages when a frame is skipped and fix some
 * statistic counters
 *
 * Revision 1.41  2010-04-16 08:57:35  gerald
 * Typo fixed
 *
 * Revision 1.40  2010-03-04 12:34:25  gerald
 * Move getJdateAsNumber() from cd2wng code to libcd
 *
 * Revision 1.39  2010-01-28 15:13:27  gerald
 * Compare the frame sample rate with the contents of the static tables
 * and skip channel if the frame sample rate deviates too much.
 * Note that the real sample rate is not checked, because to compute
 * the (ne) real sample rate the CSF has to be merged first!
 *
 * Revision 1.38  2010-01-12 10:30:35  gerald
 * Fix possible shutdown timing problem when getNextID times
 * out to avoid clf entries with frame ID -1
 *
 * Revision 1.37  2009-09-11 13:51:02  gerald
 * Comment updated
 *
 * Revision 1.36  2009-07-20 17:01:16  gerald
 * Step back to use of real segment endtime internally. Start time
 * of last sample is computed now when the wfdisc is written and
 * when a wfdisc is loaded the (expected) time of one sample is added.
 * This fix the shutdown when a one sample frame is processed (error
 * message: start == end time)
 *
 * Revision 1.35  2009-07-20 13:07:01  gerald
 * Move openIntRefList() from cd2w code to libCD
 *
 * Revision 1.34  2009-06-25 11:35:07  gerald
 * Fix compiling problem when no ODBC libraries
 * are available
 *
 * Revision 1.33  2009-06-23 16:48:09  gerald
 * Fix update code for frameproduct table and add
 * channel mask to clf table. The position in the
 * channel mask is taken from the position field in the
 * channame table
 *
 * Revision 1.32  2009-06-17 08:48:51  gerald
 * Fix the 'releasing cached elements' problem when
 * the maxWfdiscMemory parameter exceeds the merging
 * window. It makes no sense to load wfdisc records
 * before the computed start time of the window.
 *
 * Revision 1.31  2009-06-15 16:16:23  gerald
 * Start merging window at current time - maxWfdiscMemoryDuration
 * in continuous mode and treat all data with an older nominal
 * time as obsolete
 *
 * Revision 1.30  2009-06-09 15:53:05  gerald
 * Do not merge data older than current time - maxWfdiscMemoryDuration
 *
 * Revision 1.29  2009-06-09 10:56:42  gerald
 * Add monitor extension
 *
 * Revision 1.28  2009-06-04 07:55:36  gerald
 * Fix wfdisc record endtime off by one sample and
 * improve log messages for gaps/overlaps
 *
 * Revision 1.27  2009-05-20 15:47:14  gerald
 * Fix start date of initial merge window (which starts
 * one day too early). Log channel gaps/overlaps now as
 * an info syslog message instead of a debug message.
 * Skip CD-1.1 channels which starts not within one sample of
 * nominal time (and deviate from the nominal time more than
 * one second)
 *
 * Revision 1.26  2009-05-18 09:42:01  gerald
 * Fix typo in debug message
 *
 * Revision 1.25  2009-05-11 13:17:07  gerald
 * Fix not merging consecutive segments when a segment scheduled
 * for release is between two consecutive segments
 * Add more information to error messages when stop with
 * internal error
 *
 * Revision 1.24  2009-04-29 16:06:24  gerald
 * Write clf records to database (only). A clf record is only
 * written if the frame is merged into a waveform segment.
 * No clf record is written for invalid data or data which
 * does not match the filter criteria. Note: This
 * feature should be removed after the squal re-design
 *
 * Revision 1.23  2009-04-24 14:03:54  gerald
 * Fix incorrect waveform file size (shown as incorrect offset for too
 * long segments) after re-started
 * Fix missing instrument type in file mode when no sensor
 * information is available
 * Fix "Internal error: assumption for NEW_SEG failed" when two
 * consecutive quality records are loaded which have different
 * quality flags
 * Fix incorrect waveform header when cd2wng is not stopped with
 * crtl-C. Waveform header files are not longer re-written when
 * cd2wng re-starts
 * Do not report digitizer noise when constant data is detected
 * for the whole record
 * Check if the sum of the square counts exceed the double
 * data type and split the record if needed
 * Add file name to quality records required for correct re-start
 * Decrease amount of debugging text
 * Change data type of sumcounts and sumsquarecounts to
 * double to avoid possible integer overflows when computing
 * the square sum within one frame
 *
 * Revision 1.22  2009-03-31 15:50:09  gerald
 * Fix data type for square values
 *
 * Revision 1.21  2009-03-26 11:34:32  gerald
 * Add check for invalid site or channel names and skip channels
 * channels with invalid names when loading the waveform header
 *
 * Revision 1.20  2009-03-23 17:44:03  gerald
 * Fix splint warnings which occurred after upgrading
 * to new splint version. Fix 'uncompression failure'
 * when data is in S3. Fix incorrect timeliness check
 *
 * Revision 1.19  2009-02-18 16:28:17  gerald
 * Change algorithm for const data detection, add check for
 * wrong data type and check for channel dangerous warnings
 * before try to merge
 *
 * Revision 1.18  2009-02-12 18:40:11  gerald
 * Add saving/updating/loading of quality records in DB mode.
 * Check for invalid values before writing wfdisc records.
 * Create quality files and wfdisc files only when necessary
 * (not in DB mode). Fix wrong average timeliness value. Rename
 * some quality records fields
 *
 * Revision 1.17  2009-01-19 17:31:30  gerald
 * Move saveIntRefList() and getSortLevel() to libcd and
 * fix misleading internal error code when the DB is
 * not available
 *
 * Revision 1.16  2008-12-22 11:51:36  gerald
 * Omit creation of DB quality records when creation of
 * quality records is switched off
 *
 * Revision 1.15  2008-12-18 17:03:09  gerald
 * Write and read quality records from/to quality files
 *
 * Revision 1.14  2008-12-10 11:54:22  gerald
 * Add creation of quality records when data is merged.
 * Re-work statistic output
 *
 * Revision 1.13  2008-10-30 15:52:13  gerald
 * Fix too many records after re-start if the internal used
 * huge record spanws more than one requested wfdisc periode
 *
 * Revision 1.12  2008-09-16 18:12:30  gerald
 * Store waveform data also in S3. Update wfdisc
 * records for the whole merging window when
 * the merging window is shifted by one day. Fix
 * sign when converting negative s2 or s3 numbers to s4
 *
 * Revision 1.11  2008-08-12 16:33:57  gerald
 * Add and update clf receiving time to frameproduct table
 * Save restart file only when syncing channels and sync all
 * channels when a day switch occurs
 *
 * Revision 1.10  2008/07/28 12:59:45  gerald
 * Read intern and extern site names from DB, insert
 * binary files and text files into frameproduct,
 * check for relative wfdisc paths, check if all channels
 * are loaded from the waveform header when reading
 * wfdisc files. Write internal site name to wfdisc records.
 * Fix corrupted waveform headers when re-started twice
 * (or more) and Make site and channel name lengths consistenti
 * to wfdisc format
 *
 * Revision 1.9  2008/07/07 09:37:55  gerald
 * Fix restart problem when a channel was missing for a whole
 * day, but was present in older waveform files. Fix duplicate
 * wfids caused by an unprotected recycleIds() call. Fix wrong
 * offset when merging two segments start <-> end (MERGE_RIGHT)
 * Fix orphaned segments for channels without a channel ID in DB
 * mode. Fix DB error when creating new wfids caused by an unprotected
 * commit.
 *
 * Revision 1.8  2008-06-03 15:24:47  gerald
 * Start work for DB restart. Fix loading only one channel
 * and wrong channel offset if the channel was not present before
 * restart in file restart mode
 *
 * Revision 1.7  2008/05/08 15:59:22  gerald
 * Add DB support for writing wfdisc records. Use sensor
 * info for channel ID, calibration value, calibration period
 * and instrument type in file and DB mode and omit channels where
 * no sensor info is available in DB mode
 *
 * Revision 1.6  2008/04/07 14:38:17  gerald
 * Fix Solaris compiler warning
 *
 * Revision 1.5  2008/04/07 12:49:41  gerald
 * Change algorithm which splits internal used huge wfdisc record
 * into smaller pieces to avoid wfdisc records which are not
 * aligned to slot boundaries. This avoids wrong slot computation
 * caused by rounding errors during re-starts
 *
 * Revision 1.4  2008/03/20 17:06:24  gerald
 * Fix wrong offset after an overlapped segment, fix endless loop
 * when shifting for one day and the stations does not send all
 * channels in the same frame. Complete restart code for file mode
 *
 * Revision 1.3  2008/01/14 14:03:53  gerald
 * Use one second slots and start work for continuous mode
 *
 * Revision 1.2  2007/12/11 10:29:13  gerald
 * Decrease slot size to 1 second
 *
 * Revision 1.1  2007/12/10 13:12:21  gerald
 * Add CD2W_NG directory
 *
 *
 */

#include <config.h>
 #ifdef S_SPLINT_S
#include "../../splint_hints.h"
#endif /* S_SPLINT_S */
#include <errno.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <regex.h>

#include <netinet/in.h>  /* Need this to parse code with splint */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <float.h>

#include "cd2w.h"
#include "merge.h"
#include "prepare.h"
#include "schedule.h"
#include "misc.h"
#include "frames.h"
#include "readwrite.h"
#include "events.h"
#include "logfiles.h"
#include "canada_compress.h"
#ifdef HAVE_DB_ACCESS
#include "db_access.h"
#endif /* HAVE_DB_ACCESS */

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

#define NO_DURATION    1

static uint32_t validChannelTimeLengths[] = { NO_DURATION, 6, 10, 20, 30 };
/* static uint32_t validSampleRates[] = { 4, 40, 100, 250 }; */

#define CHECK_INT(value, array) \
        checkInt((value),(array),(sizeof(array)/sizeof(uint32_t)))

/**
 * @short              get nearest value
 *
 * @description        This function returns the nearest value
 * @description        to the given value based on an array of valid values
 *
 * @param value        value to check
 * @param intArray     array which holds valid values
 * @param arrayLength  length of array
 *
 * @return Nearest value
 */

static uint32_t
checkInt(const uint32_t value, const uint32_t intArray[], size_t arrayLength)
{
   size_t ii,
          index = 0;
   unsigned int diff,
                nearest = (unsigned int)-1;

   DEBUG2(2,"Value = %d", (int)value);
   for (ii = 0; ii < arrayLength; ii++)
   {
      diff = (unsigned int)abs((int)(intArray[ii] - value));
      if (diff < nearest)
      {
          nearest = diff;
          index = ii;
      }
   }

   return(intArray[index]);
}

/**
 * @short              Create channel header
 *
 * @description        This function creates a channel header if a channel
 * @description        is seen the first time. Additionally it predefines
 * @description        the offset in the waveform file for each expected
 * @description        channel sub-frame
 *
 * @param pStation        station object
 * @param channelTime     time the new channel encompasses
 * @param sampleRate      sample rate of new channel
 * @param lengthWaveFile  pointer to computed length of waveform file (in/out)
 * @param siteName        site name taken from frame
 * @param channelName     channel name taken from frame
 * @param posInChMask     index in channel mask
 *
 * @return 0 on success or -1 on failure
 */

/*@null@*/static channelHead_t *
createChannelHeader(station_t *pStation, uint32_t channelTime,
                    unsigned int sampleRate, uint32_t *lengthWaveFile,
                    const char *siteName, const char *channelName,
                    int posInChMask, uint16_t sampleSize)
{
   uint32_t expChannelTimeLng,
            slots;
   int ik;
   channelHead_t *pChannelHead;


   pChannelHead = calloc(1, sizeof(channelHead_t));
   if (NULL == pChannelHead)
   {
      shutdownApplication("Can't allocate channel header",
                          "createChannelHeader");
      return(NULL);
   }

   /* Get the nearest valid channel time length */
   expChannelTimeLng = CHECK_INT(channelTime / 1000,
                                 validChannelTimeLengths);

   ST_DEBUG2(2, "expChannelTimeLng = %d", (int)expChannelTimeLng);

   /* Compute waveform offset based on seconds */
   slots = ONE_DAY;

   /* Compute maximal offset for this channel */
   for (ik = 0; ik < MAX_MERGE_DAYS; ik++)
   {
      /* Check sample rate */
      /* sampleRate =  CHECK_INT(sampleRate, validSampleRates); */

      /* Set Channel offset. Channel offset might be different depending
         on the number of non merge-able segments for a channel per day */
      pChannelHead->day[ik].channelWavOffset = lengthWaveFile[ik];

      /* Set new (computed) length of waveform file */
      lengthWaveFile[ik] += GET_CHANNEL_SIZE(sampleRate, slots, sampleSize);
   }
   ST_DEBUG3(2, "sampleRate = %u channelWavOffset = %u", sampleRate,
             (unsigned int)pChannelHead->day[0].channelWavOffset);

   pChannelHead->nbrOfSlots = slots;
   pChannelHead->expSampleRate = (uint32_t)sampleRate;
   pChannelHead->expChannelTimeLng = (uint32_t)expChannelTimeLng;
   pChannelHead->channelIndex = (uint32_t)posInChMask;

   /* Try to get more information about the channel */
   /*@-mustfreeonly@*/
   pChannelHead->sensorInfo = getSensorDescr(siteName, channelName);
   /*@-mustfreeonly@*/

   return(pChannelHead);
}

/**
 * @short log segment attributes
 *
 * @description This function logs main attributes of a given segment
 * @description from the waveform file for a given day
 *
 * @param scName          site and channel name
 * @param actSegment      pointer to waveform segment
 * @param pass            wfdisc or quality segment
 *
 * @return nothing
 */

static void
logSegment(const char *scName, channelWaveSeg_t *actSegment, int pass)
{
   double timeDiff = actSegment->segEndTime -
                     actSegment->segStartTime;

   DEBUG_HDR;
   (void) fprintf(DEBUG_FP, "%.17s %d %d %d %f %f %f %d %c\n",
                  scName,
                  (int)actSegment->firstSlot,
                  (int)actSegment->lastSlot,
                  (int)actSegment->segState,
                  actSegment->segStartTime,
                  actSegment->segEndTime,
                  /* Compute *REAL* sample rate */
                  (double)(actSegment->samplesInSeg)/timeDiff,
                  (int)actSegment->samplesInSeg,
                  WFDISC_PASS == pass ? 'W' : 'Q');
    DEBUG_FOOTER;
}

/**
 * @short load channel descriptions
 *
 * @description This function loads the channel descriptions
 * @description from the waveform file for a given day
 *
 * @param pStation        pointer to station structure
 * @param loadDate        date of wfdisc file to load
 * @param pWaveFormHead   pointer to waveform header
 * @param dayInWindow     offset in day window
 *
 * @return 0 on success or -1 on error
 */

int
loadChannelDescription(station_t *pStation, uint32_t loadDate,
                       waveFormHead_t *pWaveFormHead, int dayInWindow)
{
   char siteName[LEN_SITE_NAME + 1],
        channelName[LEN_CHANNEL_NAME + 1];
   int items,
       rFlag,
       ii,
       posInChMask;
   FILE *fp;
   unsigned int channelOffset,
        expSampleRate,
        dataType;
   channelHead_t *pChannelHead;
   char lineBuffer[LEN_WAVEFILE_HEAD_LINE + 1];

   /* Open waveform file for given day */
   fp = openOld(pStation, (time_t)loadDate, CREATE_USR_FILE);

   /* Check if a waveform file has been written */
   if (NULL == fp)
   {
      /* No corresponding waveform file. Application should stop
         when reaching this condition */
      return(-1);
   }

   /* Set file pointer to begin of file */
   rewind(fp);
   lineBuffer[LEN_WAVEFILE_HEAD_LINE] = '\0';

   /* Read channel descriptions */
   for (ii = 0; ii < MAX_CHANNELS; ii++)
   {
      /* Read lines into buffer before trying to parse the line. This is
         necessary to process empty lines caused by (temporary) missing
         channels for a day */
      items = (int)fread(lineBuffer, 1, LEN_WAVEFILE_HEAD_LINE, fp);

      /* Check if End of File is reached */
      if (items == -1)
      {
         if (feof(fp))
         {
            ST_DEBUG2(5, "End of file reached, line [%d]", ii);
            break;
         }
         /* Otherwise clear error and continue */
         logWarning("temporary error reading channel record", pStation->conID);
         clearerr(fp);
         /* Simply decrease ii by one might cause an infinitive loop, so
            it is omitted here */
         continue;
      }

      if ('\0' == lineBuffer[0])
      {
         /* No more channel headers */
         ST_DEBUG2(2, "No more channels, line [%d]", ii);
         break;
      }
      items = sscanf(lineBuffer, WAVEFILE_HEAD_LINE_IN,
                     siteName, channelName, &channelOffset, &expSampleRate,
                     &dataType);
      ST_DEBUG2(2, "wItems=%d", items);

      /* Check if this is a correct channel description */
      if (items != 5)
      {
         ST_DEBUG2(3, "No more channels 2, line [%d]", ii);
         break;
      }

      IF_DEBUG (3)
      {
         (void) fprintf(DEBUG_FP,
                       "(SD:%.2hu)[%.17s]"
                        "%.6s %.8s %u %u %u\n",
                        pStation->debugLevel, pStation->conID,
                        siteName, channelName, channelOffset, expSampleRate,
                        dataType);
          DEBUG_FOOTER;
      }

      /* Learn channel or get position in channel mask */
      posInChMask = getChMaskPosition(siteName, channelName,
                                      &(pStation->chCache), &rFlag,
                             (pStation->systemConf & STRICT_MODE_ENABLE));

      /* Check for invalid characters in site and channel name */
      if (rFlag)
      {
         logWarning("Ignoring invalid site or channel name in waveform header",
                     lineBuffer);
         continue;
      }

      if (-1 == posInChMask)
      {
         /* No memory available, shutdown is done in getChMaskPosition() */
         (void)(fclose(fp));
         return(-1);
      }

      pChannelHead = pWaveFormHead->channelHeads[posInChMask];

      /* Check if the channel header exists, otherwise create one */
      if (NULL == pChannelHead)
      {
         ST_DEBUG3(2, "index = %d, expected sample rate = %d",
                   (int)posInChMask, (int)expSampleRate);

         ST_DEBUG3(2, "Allocate channel header for %.6s/%.8s - loadChannelDesc",
                   siteName, channelName);

         /* Set the standard offset for the channel for all days first */
         pChannelHead = createChannelHeader(pStation, NO_DURATION,
                                            expSampleRate,
                                            pWaveFormHead->lengthWaveFile,
                                            siteName, channelName,
                                            posInChMask,
                                            pWaveFormHead->sampleSize);
         if (NULL == pChannelHead)
         {
            (void)(fclose(fp));
            return(-1);
         }

         ST_DEBUG2(2, "New waveform file length: %u",
                   (unsigned int)pWaveFormHead->lengthWaveFile[0]);

         /* Save channel header in waveform header for further use */
         pWaveFormHead->channelHeads[posInChMask] = pChannelHead;
      }
      else if (0 == pChannelHead->day[dayInWindow].channelWavOffset)
      {
         /* This channel is known but not initialized in this day.
            The waveform file size has to be adjusted now. Increase
            (computed) length of waveform file by size of channel data */
         pWaveFormHead->lengthWaveFile[dayInWindow] +=
                     GET_CHANNEL_SIZE(pChannelHead->expSampleRate,
                                      pChannelHead->nbrOfSlots,
                                      pWaveFormHead->sampleSize);
      }

      /* Set the real offset for that channel and day. The real
         offset might differ from the computed 'standard' offset
         depending on the number of non merge-able segments for
         a channel per day */
      pChannelHead->day[dayInWindow].channelWavOffset = (uint32_t)channelOffset;
      /* The channel is present in the waveform file header. Don't write it
         again */
      pChannelHead->day[dayInWindow].headerIsWritten = 1;
   }

   /* Don't try to get the waveform file size from the file system,
      because the reported file size is not the real one as long
      as the last byte has not been written! The waveform file
      size is increased if a too long segment is detected during
      loading of wfdisc records */
   (void)(fclose(fp));

   /* Last step is to set the number of header lines used */
   pWaveFormHead->headerLinesWritten[dayInWindow] = (uint32_t)ii;
   return(0);
}

/**
 * @short              Check and adjust integer buffer
 *
 * @description        This function checks the if the current size of an
 * @description        integer buffer is sufficient and it enlarges the
 * @description        buffer if not
 *
 * @param intBuffer    reference to buffer holding 4 byte integers
 * @param size         requested buffer size in n integers
 * @param currSize     reference to current size of buffer
 *
 * @return -1 on error, 1 if buffer is adjusted and 0 otherwise
 */

static int
checkAndAdjustInt(unsigned char **intBuffer, uint32_t size,
                  uint32_t *currSize)
{
   /* Check if the uncompress buffer must be enlarged */
   if (*currSize < size)
   {
      unsigned char *buffer;

      DEBUG3(2, "allocate buffer for [%d] samples (old [%d])",
             (int)size, (int)*currSize);

      /* realloc() is not used here, because splint has problems to
         understand realloc() */
      buffer = (unsigned char *)calloc((size_t)size, sizeof(uint32_t));
      if (NULL == buffer)
      {
         DEBUG2(1, "Cannot allocate memory with size [%d]", (int)size);
         shutdownApplication("Can't allocate memory segment",
                             "checkAndIntAdjust");
         return(-1);
      }
      else
      {
         /*@-unqualifiedtrans@*/
         /* Free the old buffer */
         if (NULL != *intBuffer) (void)free(*intBuffer);

         *intBuffer = buffer;
         /*@=unqualifiedtrans@*/
         *currSize = size;
         return(1);
      }
   }
   return (0);
}

/**
 * @short              Convert data format type from any to s4
 *
 * @description        This function converts the data format type
 * @description        from ca, i2/i4, s2/s3 to s4
 *
 * @param channelData   data to convert/uncompress
 * @param uncompressBuf buffer for converted/uncompressed data
 * @param nbrOfSamples  number of samples
 * @param dataFormat    original data format
 * @param lenOfSamples  length of original data in bytes
 * @param qualFlag      pointer to quality flag (out)
 * @param pMean         pointer to mean of samples (out)
 * @param pM2           pointer to square sum of differences (out)
 *
 * @return -1 on error, 1 if unsupported data type, 0 otherwise
 */

static int
any2s4(unsigned char *channelData, unsigned char *uncompressBuf,
       uint32_t nbrOfSamples, char *dataFormat, uint32_t lenOfSamples,
       char isCompressed, uint16_t *qualFlag, double *pMean, double *pM2)
{
   uint32_t ii;
   int32_t min,
           max,
           ri,
           last = 0,
           cCount;
   unsigned char *startBuf = uncompressBuf;
   double M2 = 0.0,
          mean = 0.0;


/* DEBUG3(2, "Convert %.2s to s4 [%d]", dataFormat, (int)nbrOfSamples); */

   if ('\0' != isCompressed || (dataFormat[0] == 'c' && dataFormat[1] == 'a'))
   {
      int ret;
      uint32_t v0 = 0L;

      /* Decompress channel data. Keep in mind that result of
         decompression has data type s4 (network byte order),
         regardless which data type has been compressed! */
      ret = canada_uncompress((unsigned char *)channelData,
                              (uint32_t*)uncompressBuf, (int *) &lenOfSamples,
                              (int)nbrOfSamples, &v0);

      if (CANCOMP_SUCCESS != ret)
      {
         /* Canadian uncompress failed */
          return(-1);
      }
   }
   else if (dataFormat[0] == 's' && dataFormat[1] == '4')
   {
      if ((nbrOfSamples * S4) != lenOfSamples) return(-1);

      /* Nothing to convert. Copy data 1:1 to output buffer */
      memcpy(uncompressBuf, channelData, (size_t)(nbrOfSamples * S4));
   }
   else if (dataFormat[0] == 's' && dataFormat[1] == '3')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         if ((*channelData & 0x80) != (unsigned char)0)
             *uncompressBuf++ = (unsigned char) 0xFF;
         else *uncompressBuf++ = '\0';

         *uncompressBuf++ = *channelData++;
         *uncompressBuf++ = *channelData++;
         *uncompressBuf++ = *channelData++;
      }
   }
   else if (dataFormat[0] == 's' && dataFormat[1] == '2')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         if ((*channelData & 0x80) != (unsigned char)0)
         {
            *uncompressBuf++ = (unsigned char) 0xFF;
            *uncompressBuf++ = (unsigned char) 0xFF;
         }
         else
         {
            *uncompressBuf++ = '\0';
            *uncompressBuf++ = '\0';
         }

         *uncompressBuf++ = *channelData++;
         *uncompressBuf++ = *channelData++;
      }
   }
   else if (dataFormat[0] == 'i' && dataFormat[1] == '2')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         if ((*(channelData + 1) & 0x80) != (unsigned char)0)
         {
            *uncompressBuf++ = (unsigned char) 0xFF;
            *uncompressBuf++ = (unsigned char) 0xFF;
         }
         else
         {
            *uncompressBuf++ = '\0';
            *uncompressBuf++ = '\0';
         }

         *uncompressBuf++ = *(channelData + 1);
         *uncompressBuf++ = *channelData++;
         channelData++;
      }
   }
   else if (dataFormat[0] == 'i' && dataFormat[1] == '4')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         *uncompressBuf++ = *(channelData + 3);
         *uncompressBuf++ = *(channelData + 2);
         *uncompressBuf++ = *(channelData + 1);
         *uncompressBuf++ = *channelData++;
         channelData++;
         channelData++;
         channelData++;
      }
   }
   else return(1);

   /* Now check for no input from sensor and constant data */
   ri = ((int32_t)*startBuf) & 0xFF;
   /*@-shiftimplementation@*//* left shift is NOT undefined! */
   ri = (ri << 8) | (((int32_t)*(startBuf+1)) & 0xFF);
   ri = (ri << 8) | (((int32_t)*(startBuf+2)) & 0xFF);
   ri = (ri << 8) | (((int32_t)*(startBuf+3)) & 0xFF);

   /* cCount must start with 0 at first iteration, then
      it has to start with 1 */
   cCount = 0;
   max = min = last = ri;
   mean = (double)ri;

   /* Compare first value with all sample values */
   for (ii = 0; ii < nbrOfSamples; ii++)
   {
      double delta;

      ri = ((int32_t)*startBuf++) & 0xFF;
      ri = (ri << 8) | (((int32_t)*startBuf++) & 0xFF);
      ri = (ri << 8) | (((int32_t)*startBuf++) & 0xFF);
      ri = (ri << 8) | (((int32_t)*startBuf++) & 0xFF);
      /*@=shiftimplementation@*/

      IF_DEBUG(99)
      {
         DEBUG_HDR;
         (void) fprintf(DEBUG_FP, "B 0x%02x 0x%02x 0x%02x 0x%02x -> %d\n",
                        (unsigned int)*(startBuf-4) & 0xFF,
                        (unsigned int)*(startBuf-3) & 0xFF,
                        (unsigned int)*(startBuf-2) & 0xFF,
                        (unsigned int)*(startBuf-1) & 0xFF,
                        (int)ri);
          DEBUG_FOOTER;
      }

      /* Compute mean and variance as described in
         http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
         On-line_algorithm */
                                       /* n = n + 1 */
      delta = (double)ri - mean;       /* delta = x - mean */
      mean = mean + delta / (ii + 1);  /* mean = mean + delta/n */
      M2 = M2 + delta * (ri - mean);   /* M2 = M2 + delta*(x - mean) */

      if (ri > max) max = ri;
      else if (ri < min) min = ri;

      if (ri == last)
      {
         cCount++;
         if (cCount >= CONST_DATA_THRESHOLD)
         {
            *qualFlag |= QUA_CONST_DATA;
         }
      }
      else { last = ri; cCount = 1; }
   }

   *pMean = mean;
   *pM2 = M2;

   if (max != min && abs((int)(max - min)) <= DIGITIZER_NOISE_THRESHOLD)
   {
       DEBUG4(99, "M=%d - %d = %d", (int)max, (int)min, abs((int)(max - min)));
       /* 'No input from sensor' supersedes 'constant data' */
       *qualFlag &= ~QUA_CONST_DATA;
       *qualFlag |= QUA_NO_INPUT;
   }

   return(0);
}

/**
 * @short              Convert data format type from any to s3
 *
 * @description        This function converts the data format type
 * @description        from any (ca, i2/i4, s2/s3) to s3. If the data
 * @description        is received in s4/i4/ca than the MSB is cut off
 *
 * @param channelData   data to convert/uncompress
 * @param uncompressBuf buffer for converted/uncompressed data
 * @param nbrOfSamples  number of samples
 * @param dataFormat    original data format
 * @param lenOfSamples  length of original data in bytes
 * @param qualFlag      pointer to quality flag (out)
 * @param pMean         pointer to mean of samples (out)
 * @param pM2           pointer to square sum of differences (out)
 *
 * @return -1 on error, 1 if unsupported data type, 0 otherwise
 */

static int
any2s3(unsigned char *channelData, unsigned char *uncompressBuf,
       uint32_t nbrOfSamples, char *dataFormat, uint32_t lenOfSamples,
       char isCompressed, uint16_t *qualFlag, double *pMean, double *pM2)
{
   uint32_t ii;
   int32_t min,
           max,
           ri,
           last = 0,
           cCount;
   unsigned char *startBuf = uncompressBuf;
   double M2 = 0.0,
          mean = 0.0;


/* DEBUG3(2, "Convert %.2s to s3 [%d]", dataFormat, (int)nbrOfSamples); */

   if ('\0' != isCompressed || (dataFormat[0] == 'c' && dataFormat[1] == 'a'))
   {
      int ret;
      uint32_t v0 = 0L;
      unsigned char *p3;

      /* Decompress channel data. Keep in mind that result of
         decompression has data type s4 (network byte order),
         regardless which data type has been compressed! */
      ret = canada_uncompress((unsigned char *)channelData,
                              (uint32_t*)uncompressBuf, (int *) &lenOfSamples,
                              (int)nbrOfSamples, &v0);

      if (CANCOMP_SUCCESS != ret)
      {
         /* Canadian uncompress failed */
          return(-1);
      }
      /* Compressed data might be not be in s4 after decompression in
         CD-1.1 ?? */
      p3 = uncompressBuf;
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         uncompressBuf++; /* Skip most significant byte */
         *p3++ = *uncompressBuf++;
         *p3++ = *uncompressBuf++;
         *p3++ = *uncompressBuf++;
      }
   }
   else if (dataFormat[0] == 's' && dataFormat[1] == '3')
   {
      /* length of samples is padded to 4 */
      if (PAD(nbrOfSamples * S3) != lenOfSamples) return(-1);

      /* Nothing to convert. Copy data 1:1 to output buffer */
      memcpy(uncompressBuf, channelData, (size_t)(nbrOfSamples * S3));
   }
   else if (dataFormat[0] == 's' && dataFormat[1] == '4')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         channelData++; /* Skip most significant byte */
         /* Not sure what should I do with the sign ?? */
         *uncompressBuf++ = *channelData++;
         *uncompressBuf++ = *channelData++;
         *uncompressBuf++ = *channelData++;
      }
   }
   else if (dataFormat[0] == 's' && dataFormat[1] == '2')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         if ((*channelData & 0x80) != (unsigned char)0)
             *uncompressBuf++ = (unsigned char) 0xFF;
         else *uncompressBuf++ = '\0';

         *uncompressBuf++ = *channelData++;
         *uncompressBuf++ = *channelData++;
      }
   }
   else if (dataFormat[0] == 'i' && dataFormat[1] == '2')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         if ((*(channelData + 1) & 0x80) != (unsigned char)0)
             *uncompressBuf++ = (unsigned char) 0xFF;
         else *uncompressBuf++ = '\0';

         *uncompressBuf++ = *(channelData + 1);
         *uncompressBuf++ = *channelData++;
         channelData++;
      }
   }
   else if (dataFormat[0] == 'i' && dataFormat[1] == '4')
   {
      for (ii = 0; ii < nbrOfSamples; ii++)
      {
         /* Skip most significant byte.
            Not sure if I should check the sign;
            *uncompressBuf++ = *(channelData + 3); */
         *uncompressBuf++ = *(channelData + 2);
         *uncompressBuf++ = *(channelData + 1);
         *uncompressBuf++ = *channelData++;
         channelData++;
         channelData++;
         channelData++;
      }
   }
   else return(1);

   /* Now check for no input from sensor and constant data. Samples are
      in s3 format (MSB first)! */
   if ((*startBuf & 0x80) != (unsigned char)0) { ri = (int32_t)0xFF; }
   else { ri = 0; }
   /*@-shiftimplementation@*//* left shift is NOT undefined! */
   ri = (ri << 8) | (((int32_t)*(startBuf)) & 0xFF);
   ri = (ri << 8) | (((int32_t)*(startBuf+1)) & 0xFF);
   ri = (ri << 8) | (((int32_t)*(startBuf+2)) & 0xFF);

   /* cCount must start with 0 at first iteration, then
      it has to start with 1 */
   cCount = 0;
   max = min = last = ri;
   mean = (double)ri;

   /* Compare first value with all sample values */
   for (ii = 0; ii < nbrOfSamples; ii++)
   {
      double delta;

      if ((*startBuf & 0x80) != (unsigned char)0) { ri = (int32_t)0xFF; }
      else { ri = 0; }
      ri = (ri << 8) | (((int32_t)*startBuf++) & 0xFF);
      ri = (ri << 8) | (((int32_t)*startBuf++) & 0xFF);
      ri = (ri << 8) | (((int32_t)*startBuf++) & 0xFF);
      /*@=shiftimplementation@*/

      IF_DEBUG(99)
      {
         DEBUG_HDR;
         (void) fprintf(DEBUG_FP, "B 0x%02x 0x%02x 0x%02x 0x%02x -> %d\n",
                        *(startBuf-3) & 0x80 ? (unsigned int)0xFF : 0,
                        (unsigned int)*(startBuf-3) & 0xFF,
                        (unsigned int)*(startBuf-2) & 0xFF,
                        (unsigned int)*(startBuf-1) & 0xFF,
                        (int)ri);
          DEBUG_FOOTER;
      }

      /* Compute mean and variance as described in
         http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
         On-line_algorithm */
                                       /* n = n + 1 */
      delta = (double)ri - mean;       /* delta = x - mean */
      mean = mean + delta / (ii + 1);  /* mean = mean + delta/n */
      M2 = M2 + delta * (ri - mean);   /* M2 = M2 + delta*(x - mean) */

      if (ri > max) max = ri;
      else if (ri < min) min = ri;

      if (ri == last)
      {
         cCount++;
         if (cCount >= CONST_DATA_THRESHOLD)
         {
            *qualFlag |= QUA_CONST_DATA;
         }
      }
      else { last = ri; cCount = 1; }
   }

   *pMean = mean;
   *pM2 = M2;

   if (max != min && abs((int)(max - min)) <= DIGITIZER_NOISE_THRESHOLD)
   {
       DEBUG4(99, "M=%d - %d = %d", (int)max, (int)min, abs((int)(max - min)));
       /* 'No input from sensor' supersedes 'constant data' */
       *qualFlag &= ~QUA_CONST_DATA;
       *qualFlag |= QUA_NO_INPUT;
   }

   return(0);
}

/**
 * @short              Write binary data to file
 *
 * @description        This function writes binary data to a given file
 * @description        pointer
 *
 * @param fp           file pointer
 * @param buffer       pointer to buffer containing the data frame
 * @param len          buffer size
 *
 * @return 0 on success or -1 on error
 */

static int
writeBinData(FILE *fp, const unsigned char *buffer, int len, const char *conID)
{
#define NBR_OF_RETRIES          10
   ssize_t retVal = 0;
   int lTotal = 0,
       retries = NBR_OF_RETRIES;
   struct timespec sleeptime;

   if (NULL == fp) return(0);

   /* Try to save the binary frame now */
   do
   {
      retVal = write(fileno(fp), buffer + lTotal, (size_t)(len - lTotal));

      if ((int)retVal != (len - lTotal))
      {
         char logBuffer[LOGSIZE];

         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "%.17s Problem writing to bin file [%d] [%d]",
                          conID, (len-lTotal), (int)retVal);
         logWarning(logBuffer, "writeBinData");

         if (-1 == retVal)
         {
            /* Interrupts and EAGAIN are treated as recoverable errors
               until the retry count is reached */
            if (retries <= 0 || (errno != EINTR && errno != EAGAIN)) break;

            /* Reset errno and set retVal to zero to avoid
               endless loop */
            errno = 0;
            retVal = 0;
            retries--;
         }
         else
         {
            retries = NBR_OF_RETRIES;
         }
         /* Sleep a while, maybe that's a temporary condition */
         /*@-usedef@*/
         sleeptime.tv_sec = 1;
         sleeptime.tv_nsec = 0;
         (void) nanosleep(&sleeptime, NULL);
         /*@=usedef@*/
      }
      lTotal += (int)retVal;
   }
   while (lTotal < len);

   if (lTotal < len)
   {
      /* In most cases this is a permission or file system
         problem. Application is not able to save data, so a shutdown
         is initiated */
      shutdownApplication("can't write to binary/waveform file", conID);
      clearerr(fp);
      return(-1);
   }
   return(0);
}
/**
 * @short              Create channel segment
 *
 * @description        This function creates a channel segment
 * @description        after 'leftSeg'
 *
 * @param pSegHead     segment header for channel
 * @param leftSeg      left segment
 *
 * @return waveform segment or NULL in case of error
 */

/*@null@*//*@dependent@*/
channelWaveSeg_t *
createSegment(segmentHead_t *pSegHead, /*@null@*/channelWaveSeg_t *leftSeg)
{
   channelWaveSeg_t *pSegment;
   int ii;

   pSegment = calloc(1, sizeof(channelWaveSeg_t));
   if (NULL == pSegment)
   {
      shutdownApplication("Can't allocate segment", "createSegment");
      return(NULL);
   }

   pSegment->segState = SEG_INIT;

   if (NULL == leftSeg)
   {
      /* First segment */
      /*@-mustfreeonly@*/
      pSegment->nextSeg = pSegHead->firstSeg;
      /*@=mustfreeonly@*/
      pSegHead->firstSeg = pSegment;
   }
   else
   {
      /* Insert after 'left segment' */
      /*@-mustfreeonly@*/
      pSegment->nextSeg = leftSeg->nextSeg;
      /*@=mustfreeonly@*/
      leftSeg->nextSeg = pSegment;
      pSegment->prevSeg = leftSeg;
   }

   /* Check if it is save to link back */
   if (NULL != pSegment->nextSeg)
   {
      pSegment->nextSeg->prevSeg = pSegment;
   }

   /* Check if this is the first (and/or last) segment */
   if (leftSeg == pSegHead->lastSeg)
      pSegHead->lastSeg = pSegment;

   pSegHead->nbrOfSegments++;

   /* Initialize WFID's */
   for (ii = 0; ii < MAX_SUB_RECORDS; ii++)
        pSegment->subrecord.rid[ii] = -1L;

   return(pSegment);
}

/**
 * @short              Destroy channel segment
 *
 * @description        This function destroys a given channel segment
 *
 * @param pSegHead     segment header for channel
 * @param pSegment     segment to destroy
 *
 * @return nothing
 */

static void
destroySegment(segmentHead_t *pSegHead, channelWaveSeg_t *pSegment)
{
   channelWaveSeg_t *pRelSeg; /* Segment to release */

   if (NULL == pSegment)
   {
      shutdownApplication("NULL segment", "destroySegment");
      return;
   }

   if (SEG_RELEASED == pSegment->segState)
   {
      /* Release segments which have a SEG_CACHED flag set is OK here!! */
      shutdownApplication("segment already released", "destroySegment");
      return;
   }

   if (NULL == pSegment->prevSeg) /* Must be the first segment */
   {
      /* First segment */
      pRelSeg = pSegHead->firstSeg;
      if (NULL == pRelSeg)
      {
         /* Internal error: empty list */
         shutdownApplication("segment list empty", "destroySegment");
         return;
      }
      pSegHead->firstSeg = pRelSeg->nextSeg;
   }
   else
   {
      /* Not the first segment */
      pRelSeg = pSegment->prevSeg->nextSeg; /* This is for splint */
      if (NULL == pRelSeg)
      {
         /* Internal error: corrupted list */
         shutdownApplication("segment list corrupted", "destroySegment");
         return;
      }
      pSegment->prevSeg->nextSeg = pRelSeg->nextSeg;
   }

   /* Check if it is save to link back */
   if (NULL != pRelSeg->nextSeg)
   {
      pRelSeg->nextSeg->prevSeg = pRelSeg->prevSeg;
   }

   /* Check if this is the first (and/or last) segment */
   if (pRelSeg == pSegHead->lastSeg)
      pSegHead->lastSeg = pRelSeg->prevSeg;

   pSegHead->nbrOfSegments--;
   pRelSeg->segState = SEG_RELEASED;
   (void)free(pRelSeg);
}

/**
 * @short              Get adjacent, overlapped or create new segment
 *
 * @description        This function checks if startSlot and endSlot are
 * @description        inside an existing segment, adjacent to an existing
 * @description        segment or a new segment has to be created.
 *
 * @param SegHead      segment header for channel
 * @param startSlot    start slot of new segment
 * @param endSlot      end slot of new segment
 * @param result       result of check
 *  .|  seg_1  |.........|  seg_2  |...........
 *  ............|  new  |............|  new  |. NEW_SEG
 *  ...........|   new   |..................... MERGE_LEFT
 *  ...........|  new  |...........|  new  |... MERGE_LEFT
 *  .........|  new  |...........|  new  |..... MERGE_LEFT,OVERLAP_LEFT
 *  .............|  new  |..................... MERGE_RIGHT
 * |  new  |........|  new  |.................. MERGE_RIGHT,OVERLAP_RIGHT
 * .........|     new       |.................. MERGE_LEFT,OVERLAP_RIGHT+LEFT
 *  .|   new   |.........|   new  |............ DUPLICATE
 *  ..|  new   |..........|  new |............. DUPLICATE
 * ..................|      new       |........ DUPLICATE,OVERLAP_COMPLETE*
 * ..........|          new           |........ DUPLICATE,OVERLAP_COMPLETE**
 * .|             new                 |........ DUPLICATE,OVERLAP_COMPLETE***
 *
 * @return       existing or new waveform segment or NULL in case of error
 */

/*@null@*//*@dependent@*/
channelWaveSeg_t *
getSegment(segmentHead_t *pSegHead, uint32_t startSlot, uint32_t endSlot,
           /*@out@*/int *result)
{
   channelWaveSeg_t *actSegment,
                    *bestResult = NULL;
   int overlapState = 0;

   DEBUG3(5, "Try to find segment %d - %d", (int)startSlot, (int)endSlot);

   *result = NEW_SEG;
   for (actSegment = pSegHead->lastSeg; actSegment != NULL;
                                    actSegment = actSegment->prevSeg)
   {
      IF_DEBUG(4)
      {
         (void) fprintf(DEBUG_FP, "Check Segment %d - %d [%hu %f-%f]\n",
		(int)actSegment->firstSlot, (int)actSegment->lastSlot,
	 	(unsigned short)actSegment->segState, actSegment->segStartTime,
		actSegment->segEndTime);

          DEBUG_FOOTER;
      }
      if (SEG_RELEASED == actSegment->segState)
      {
         shutdownApplication("Released segment in segment list",
                             "getSegment");
         return(NULL);
      }

      /* Check if the segment is scheduled for release. Such a segment
         is released when the wfdisc records are updated. Don't merge
         such a segment. It is already merged!! */
      if (0 != (SEG_SCHEDULED4RELEASE & actSegment->segState))
      {
         /* Skip segment */
         continue;
      }

      /* Check if there is a filled slot on the left side; left side
         overrides right one */
      if (actSegment->lastSlot == (startSlot - 1))
      {
         /* Overwrite right segment if any. Left segment has precedence */
         *result = MERGE_LEFT | overlapState;
         return(actSegment);
      }

      /* Check if start slot of the data is after the current
         segments end slot */
      if (actSegment->lastSlot < startSlot)
      {
         /* Break loop. Don't set bestResult here!! */
         break;
      }

      /* Check if the start slot of the new data is
         inside the current segment */
      if (actSegment->lastSlot >= startSlot &&
          actSegment->firstSlot <= startSlot)
      {
         /* Now check if the end slot of the new data is inside the
            current element */
         if (actSegment->lastSlot >= endSlot)
         {
             *result = DUPLICATE_SEG;
             return(actSegment);
         }
         /* This is partial duplicate data, try to merge it left */
         *result = MERGE_LEFT | OVERLAP_LEFT | overlapState;
         return(actSegment);
      }

      /* Check if there is a filled slot on the right side */
      if (actSegment->firstSlot == (endSlot + 1))
      {
         *result = MERGE_RIGHT;
         bestResult = actSegment;
      }
      /* Check if the end slot of the new data is
         inside the current segment */
      else if (actSegment->firstSlot <= endSlot &&
               actSegment->lastSlot >= endSlot)
      {
         /* This is a partial overlap on the right side */
         overlapState = OVERLAP_RIGHT;
         *result = MERGE_RIGHT | OVERLAP_RIGHT;
         bestResult = actSegment;
      }
      /* Check if the current segment is complete embedded in the
         new data */
      else if (actSegment->firstSlot >= startSlot &&
               actSegment->lastSlot <= endSlot)
      {
         /* The current segment is complete overlapped */
         overlapState |= OVERLAP_COMPLETE;
         /* TODO: ONLY FOR TESTING */
         *result = DUPLICATE_SEG | OVERLAP_COMPLETE;
         return(actSegment);
      }
   }

   /* Check if a right neighbor segment has been found. This condition
      needs no check for overlapping segments */
   if (NEW_SEG != *result) return(bestResult);

   /* No merge-able segment found. Create new one */
   *result = NEW_SEG | overlapState;
   return(createSegment(pSegHead, actSegment));
}

/**
 * @short              Get next, not scheduled for released, segment
 *
 * @description        This function returns the next, not scheduled for
 * @description        released element, if any
 *
 * @param actSegment   segment to start the search
 *
 * @return       next waveform segment or NULL
 */

/*@null@*//*@dependent@*/
static channelWaveSeg_t *
getNextSegment(channelWaveSeg_t *actSegment)
{
   DEBUG(5, "Try to find next Segment");

   actSegment = actSegment->nextSeg;

   for ( ; actSegment != NULL; actSegment = actSegment->nextSeg)
   {
      IF_DEBUG(4)
      {
         (void) fprintf(DEBUG_FP, "Check Segment %d - %d [%hu %f-%f]\n",
		(int)actSegment->firstSlot, (int)actSegment->lastSlot,
	 	(unsigned short)actSegment->segState, actSegment->segStartTime,
		actSegment->segEndTime);

          DEBUG_FOOTER;
      }

      /* Check for released segments */
      if (SEG_RELEASED == actSegment->segState)
      {
         shutdownApplication("Released segment in segment list",
                             "getSegment");
         return(NULL);
      }

      /* Check if the segment is scheduled for release. Such a segment
         is released when the wfdisc records are updated. Don't merge
         such a segment. It is already merged!! */
      if (0 == (SEG_SCHEDULED4RELEASE & actSegment->segState))
      {
         /* This is the next element */
         break;
      }
   }

   return(actSegment);
}

/* Translation table to translate CD-1.0 into CD-1.1 status */
static translate_t cd10tocd11status[] = {
   { (uint32_t)(1<<23), STS_VAULT_DOOR_OPEN },
   { (uint32_t)(1<<22), STS_AUTH_SEAL_BROKEN },
   { (uint32_t)(1<<21), STS_EQUIP_MOVED },
   { (uint32_t)(1<<20), STS_CLOCK_DIFF_TOO_LARGE },
   { (uint32_t)(1<<31), STS_DEAD_SENSOR_CHANNEL },
   { (uint32_t)(1<<30), STS_ZEROED_DATA },
   { (uint32_t)(1<<29), STS_CLIPPED },
   { (uint32_t)(1<<28), STS_CALIBRATION_UNDERWAY }
};

/**
 * @short              Retrieve channel status
 *
 * @description        This function retrieves the channel status into
 * @description        4 byte unsigned integer and map CD1 status to CD1_1
 *
 * @param type          station type
 * @param status        pointer to first byte if channel status
 *
 * @return channel status
 */

static uint32_t
getAttribs(unsigned char type, const char *status)
{
   int offs;
   uint32_t chanStatus,
            cd11status;

   if (NULL == status)
      return(0);

   /* Compute offset and clipped bit mask */
   if (CD1 == type)
   {
      offs = 0;
   }
   else
   {
      offs = 1;
   }

   if (CD1_1 == type)
   {
       chanStatus = ((((uint32_t)(status[offs])) & 0xFF) +
                    ((((uint32_t)(status[offs + 1])) & 0xFF) << 8) +
                    ((((uint32_t)(status[offs + 2])) & 0xFF) << 16) +
                    ((((uint32_t)(status[offs + 3])) & 0xFF) << 24));

      return(chanStatus);
   }

   chanStatus = (((((uint32_t)(status[offs])) & 0xFF) << 24) +
                ((((uint32_t)(status[offs + 1])) & 0xFF) << 16) +
                ((((uint32_t)(status[offs + 2])) & 0xFF) << 8) +
                (((uint32_t)(status[offs + 3])) & 0xFF));

   /* Translate CD-1.0 status into CD-1.1 status */
   cd11status = translateInt(chanStatus, cd10tocd11status,
                             sizeof(cd10tocd11status));
   DEBUG3(3, "Map CD-1.0 status 0x%x into CD-1.1 status 0x%x",
          (unsigned int)chanStatus, (unsigned int)cd11status);
   return(cd11status);
}

/**
 * @short              Get DB handle if any
 *
 * @description        This function returns the DB handle if any
 *
 * @param pConfigData  Pointer to configuration data
 * @param which        Standard or ID handle
 *
 * @return database handle or NULL
 */

/*@null@*//*@dependent@*/ void *
getHdbc(const configParams_t *pConfigData, int which)
{
   senderParams_t *pCd2wConfig;

   if (NULL == pConfigData->configExtension)
   {
      /* No config extension */
      shutdownApplication("No config extension", "getHdbc");
      return(NULL);
   }
   pCd2wConfig = (senderParams_t *)(pConfigData->configExtension);

   if (DB_HANDLE == which)
       return(pCd2wConfig->sndAPIAppPointer);

   if (ID_HANDLE == which)
       return(pCd2wConfig->sndAPIAppPointer2);

   return(NULL);
}


#ifdef HAVE_DB_ACCESS

/**
 * @short              Delete database record found in sub-records
 *
 * @description        This function releases all Db records found in the
 * @description        sub-records
 *
 * @param pStation     station object
 * @param pSegment     waveform segment
 * @param what         record type (wfdisc or quality record)
 *
 * @return 0 on success or -1 on error
 */

static int
releaseDBrecord(station_t *pStation, channelWaveSeg_t *pSegment, int what)
{
   void *hdbc;
   int sR;
#ifndef HAVE_LIBIDCODBC
   int ids[MAX_SUB_RECORDS];
#else
   long ids[MAX_SUB_RECORDS];
#endif

   hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);

   if (NULL == hdbc)
   {
       /* No DB connection */
       ST_DEBUG(2, "No DB connection (releaseDBrecord)");
       return(0);
   }

   for (sR = 0; sR < MAX_SUB_RECORDS; sR++)
   {

      if (-1 == pSegment->subrecord.rid[sR]) break;

      /* Delete wfdisc or quality record in DB and give back record ID
         for re-use */
      if (-1 == deleteRecordInDB((SQLHDBC)hdbc, pSegment->subrecord.rid[sR],
                                  what))
      {
          return(-1);
      }

#ifndef HAVE_LIBIDCODBC
      ids[sR] = (int)pSegment->subrecord.rid[sR];
#else
      ids[sR] = pSegment->subrecord.rid[sR];
#endif
      pSegment->subrecord.rid[sR] = -1L;
   }
   if (sR > 0)
   {
      configParams_t *pConfigData = (configParams_t *)(pStation->pConfigData);

      if (-1 == recycleIds(ids, sR, &(pConfigData->appMutex), what))
      {
         shutdownApplication("Can't recycle id's", "releaseDBrecord");
         return(-1);
      }
   }
   return(0);
}

#ifdef USE_INSERT_SUB_REC

/**
 * @short              Add sub-record to segment
 *
 * @description        This function adds a sub-record to a given segment
 *
 * @param pSegHead     segment header
 * @param pSegment     waveform segment
 * @param rid          record id (if loaded from DB) or -1
 * @param offsetSub    offset in sub-record read from DB
 * @param samplesInSub number of samples in sub-record read from DB
 *
 * @return nothing
 */

static void
insertSubRec(segmentHead_t *pSegHead, channelWaveSeg_t *pSegment, long rid,
             uint32_t offsetSub, uint32_t samplesInSub)
{
   int ii;
   channelWaveSeg_t *newSegment;

   if (-1L == rid) return;

   /* search for an empty slot */
   for (ii = 0; ii < MAX_SUB_RECORDS; ii++)
   {
       if (-1L == pSegment->subrecord.rid[ii])
       {
          pSegment->subrecord.rid[ii] = rid;
          pSegment->subrecord.offsetSub[ii] = offsetSub;
          pSegment->subrecord.samplesInSub[ii] = samplesInSub;
          return;
       }
   }

   DEBUG2(2, "Create empty slot for rid [%ld]", rid);

   /* OUCH!! No empty slot found. That happens when other waveform
      lengths were used in the previous run of cd2w. Create an
      empty segment and schedule it for release */
   newSegment = createSegment(pSegHead, NULL);
   if (NULL == newSegment)
   {
      /* Shutdown is done in createSegment */
      return;
   }
   newSegment->segState = SEG_SCHEDULED4RELEASE | SEG_CACHED;
   newSegment->subrecord.rid[0] = rid;
}

#endif /* USE_INSERT_SUB_REC */

#endif  /* HAVE_DB_ACCESS */

/**
 * @short              Write channel info to waveform file
 *
 * @description        This function writes the channel info to the
 * @description        waveform file if it has not been written for
 * @description        the given day
 *
 * @param pStation           station object
 * @param pDay               day header
 * @param pChannelHead       channel header
 * @param siteName           site name
 * @param channelName        channel name
 * @param linesWritten       number of lines written to the waveform header
 *
 * @return  -1 on error or 0 on success
 */

static int
writeChannelInfo(station_t *pStation, dayHeader_t *pDay,
                 channelHead_t *pChannelHead, const char *siteName,
                 const char *channelName, uint16_t sampleSize,
                 uint32_t linesWritten)
{
   if (1 == pDay->headerIsWritten || NULL == pStation->files.fpUsrFile)
      return(0);

   ST_DEBUG4(2,"Write channel header [%.6s/%.8s %u]", siteName,
             channelName, (unsigned int)linesWritten);

   /* Move file pointer to computed offset in waveform file */
   if (-1 == fseek(pStation->files.fpUsrFile,
            (long)(linesWritten * LEN_WAVEFILE_HEAD_LINE), SEEK_SET))
   {
      /* Should never ever happen */
      shutdownApplication("can't seek in waveform file", pStation->conID);
      return(-1);
   }

   /* Use the internal station name if possible.
      Re-start is able to deal with internal and
      external channel names */
   fprintf(pStation->files.fpUsrFile, WAVEFILE_HEAD_LINE,
           NULL == pChannelHead->sensorInfo ? siteName :
           pChannelHead->sensorInfo->intern_sta, channelName,
           (unsigned int)pDay->channelWavOffset,
           (unsigned int)pChannelHead->expSampleRate,
           (int)sampleSize);

   (void)fflush(pStation->files.fpUsrFile);
   pDay->headerIsWritten = 1;

   return(0);
}

/**
 * @short          Compute mean and M2
 *
 * @description    This function computes the mean and M2 of
 * @description    two segments using the parallel algorithm described in
 * @description    en.wikipedia.org/wiki/Algorithms_for_calculating_variance
 *
 * @param pSegment      waveform segment description structure
 * @param segMean       mean value of segment to merge
 * @param segM2         M2 value of segment to merge
 * @param samplesInSeg  number of samples in segment to merge
 * @param mean          new mean of both segments (out)
 * @param M2            new M2 value of both segments (out)
 *
 * @return  -1 on error , 0  on success
 */

static int
getMeanAndM2(channelWaveSeg_t *pSegment, double segMean, double segM2,
             uint32_t samplesInSeg, /*@out@*/double *mean, /*@out@*/double *M2)
{
   double delta,
          tmp,
          M2ab;

   delta = pSegment->segMean - segMean;
   *mean = 0;
   *M2 = 0;

   tmp = pSegment->segMean * pSegment->samplesInSeg + segMean * samplesInSeg;

   /* Do some overflow tests here */
   if (isnan(tmp))
      return (-1);

   *mean = tmp / (pSegment->samplesInSeg + samplesInSeg);

   M2ab = pSegment->segM2 + segM2;

   /* Do some overflow tests here */
   if (isnan(M2ab))
      return (-1);

   tmp = delta * delta * ((pSegment->samplesInSeg * samplesInSeg) /
         (pSegment->samplesInSeg + samplesInSeg));

   /* Do some overflow tests here */
   if (isnan(tmp))
      return (-1);

   tmp += M2ab;

   /* Do some overflow tests here */
   if (isnan(tmp))
      return (-1);

   *M2 = tmp;
   return(0);
}

/**
 * @short              Try to merge the new waveform data
 *
 * @description        This function tries to merge new waveform data or
 * @description        values read from a wfdisc record for a given
 * @description        channel. If new data is merged, the data
 * @description        is saved to the waveform file immediately.
 * @description        Wfdisc records have to be merged before new data is
 * @description        merged or the cached flag won't be set correct
 *
 * @param channel            channel containing data or wfdisc info
 * @param nominalTime        nominal time of frame or 0 when loading wfdisc
 * @param channelTime        channel time span
 * @param pChannelHead       channel header
 * @param slot               first slot occupied by channel data
 * @param pStation           station object
 * @param pWaveFormHead      waveform header
 * @param currentDay         day the waveform data belongs to
 * @param mergeDay           day in merging window
 * @param isRestartWfdisc    set when a wfdisc is merged during re-start
 * @param clipped            clipped status when a wfdisc is merged
 * @param wfdiscEndTime      end time when a wfdisc is merged
 * @param wfid               wfdisc ID (if loaded from DB) or -1
 * @param clfTime            start sending/receiving time in clf file
 * @param event              pointer to event (out)
 *
 * @return  -1 on error , 0  data uncompress problem
 *           or result of merge attempt (1 ->left merge,
 *           2 -> right merge, 3 -> , 4 -> duplicate
 */

int
tryToMerge(CDdataChannel_t *channel, double nominalTime,
           uint32_t channelTime, channelHead_t *pChannelHead, uint32_t slot,
           station_t *pStation, waveFormHead_t *pWaveFormHead,
           uint32_t currentDay, int mergeDay, int isRestartWfdisc,
           char clipped, double wfdiscEndTime, long wfid, double clfTime,
           uint16_t *event)
{
   uint16_t segState = SEG_NORMAL;
   channelWaveSeg_t *pSegment;
   int result,
       retVal,
       pass,
       tryRight;
   uint32_t samples,
            offset,
            addOneSlot,
            slotsOccupied;
   double tDiff,
          dataEndTime,
          segEndTime,
          segStartTime;
   uint32_t attributes;    /* Channel status */
   dayHeader_t *pDay;
   uint16_t qualFlag = 0;
   segmentHead_t *pSegHeader;
   double segMean = 0.0,
          segM2 = 0.0,
          meanMerged = 0.0,
          M2merged = 0.0;

   *event = NO_ERROR;
   /* Compute number of occupied slots */
   slotsOccupied = (uint32_t)(channelTime / 1000);

   /* Check if the channel time contains also fractions of a second.
      According to the F&P this should not happen, but some stations
      behave like this (see JIRA ISI-29) */
   addOneSlot = (channelTime % 1000) > 0 ? (uint32_t)1 : (uint32_t)0;
   if (addOneSlot > 0)
   {
      DEBUG4(1, "Fraction seconds detected [%.6s/%.8s] [%d]. Adding one slot",
            channel->siteName, channel->channelName, (int)channelTime);
      slotsOccupied++;
   }

   pDay = &(pChannelHead->day[mergeDay]);
   pSegHeader = &(pDay->wfdiscSeg);

   /* Get one of the neighbor segments, the (first) overlapped
      or an empty segment if no neighbors are found.
      If two neighbors are found the left one is returned */
   pSegment = getSegment(pSegHeader, slot, slot + slotsOccupied - 1,
                         &result);

   if (NULL == pSegment)
   {
      DEBUG4(1, "Attempt to merge data [%.6s/%.8s] [%d] failed",
             channel->siteName, channel->channelName, result);
      /* Shutdown is done in getSegment() */
      *event = E_NO_MEM;
      return (-1);
   }

   IF_DEBUG(3)
   {
      (void) fprintf(DEBUG_FP,
             "(SD:%.2hu)[%.17s]"
             "Merge data [%.6s/%.8s] [%d] [%d] [%f] [%u] [%d] [%d] [%d]\n",
             pStation->debugLevel, pStation->conID,
             channel->siteName, channel->channelName, result, (int)slot,
             nominalTime, (unsigned int)channelTime, (int)channel->timeStamp,
             (int)pSegment->firstSlot, (int)pSegment->lastSlot);

      DEBUG_FOOTER;
   }

   /* Check if the new segment overlaps complete one or more existing
      segments */
   if (0 != (result & OVERLAP_COMPLETE))
   {
      DEBUG5(2, "Complete overlapped slot(s) detected [%.6s/%.8s] [%d] [%d]",
             channel->siteName, channel->channelName,
             (int)slot, (int)(slot + slotsOccupied - 1));
      /* TODO: treat that condition like other overlaps ?? */
      pDay->duplicates++;
      return(4);
   }
   if (0 != (result & (OVERLAP_LEFT | OVERLAP_RIGHT)))
   {
      DEBUG5(2,
             "Left or right overlapped slot(s) detected [%.6s/%.8s] [%f] [%d]",
             channel->siteName, channel->channelName,
             nominalTime, (int)(slot + slotsOccupied - 1));
      /* Mark the slot as 'not merge able', create new segment if necessary
         and put the data at the end of the waveform file. This is not
         necessarily a real overlap. Sometimes a station sends a too long
         frame in terms of frame nominal time and it looks like an overlap,
         but in terms of channel nominal time not (last frame ends ...41,
         current frame starts ....40, but channel time is ...40.99999) ! */
      segState = SEG_TOO_LONG;
      /* Reset OVERLAP_* and DUPLICATE states */
      result &= (MERGE_RIGHT | MERGE_LEFT);
   }

   /* Check for duplicate segments */
   if (DUPLICATE_SEG == result)
   {
      /* Alternate data is received or received data is identical
         to an existing segment */
      DEBUG5(2, "Duplicate channel data detected [%.6s/%.8s] [%f] [%d]",
             channel->siteName, channel->channelName,
             nominalTime, (int)slot);

      /* TODO: check if data is alternate or existing data */
      pDay->duplicates++;
      return(4);
   }

   /* Check channel time only if this is no re-start */
   if (0 == isRestartWfdisc)
   {
      /* Now check if the channel time length is the same */
      if (channelTime > (pChannelHead->expChannelTimeLng * 1000))
      {
         /* It is not required that the channel time length is always
            the same. Don't trust this value too much, but this is
            'nice to have' statistics */
         DEBUG5(2, "%.6s/%.8s, channel time length too high %d > %d",
                channel->siteName, channel->channelName, (int)channelTime,
                (int)pChannelHead->expChannelTimeLng);

         /* Increase too long counter */
         pDay->tooLong++;
      }
      else if (channelTime < (pChannelHead->expChannelTimeLng * 1000))
      {
         /* It is not required that the channel time length is always
            the same. Don't trust this value too much, but this is
            'nice to have' statistics */
         DEBUG5(2, "%.6s/%.8s, channel time length too short %d > %d",
                channel->siteName, channel->channelName, (int)channelTime,
                (int)pChannelHead->expChannelTimeLng);
         /* Increase too short counter */
         pDay->tooShort++;
      }

      /* Set authentication flag for quality record */
      switch (channel->authentPass)
      {
         case AUTH_PASSED:
              qualFlag = 0;
              break;
         case AUTH_FAILED:
              pStation->whenConnected.authenticationFailures++;
              qualFlag = QUA_AUTH_FAILED;
              break;
         case AUTH_NO_SIG:
              qualFlag = QUA_AUTH_NOT_SIGNED;
              break;
         case AUTH_NO_KEY:
              qualFlag = QUA_AUTH_NO_CERT;
              break;
         default:
              qualFlag = QUA_NOT_AUTHENTICATED;
              break;
      }

      if (CD1_1 == pStation->type)
      {
         /* Check frame authentication only for CD-1.1 stations */
         switch (pStation->authenPassFrame)
         {
            case AUTH_PASSED:
                 break;
            case AUTH_FAILED:
                 pStation->whenConnected.authenticationFailures++;
                 qualFlag |= QUA_FR_AUTH_FAILED;
                 break;
            case AUTH_NO_SIG:
                 qualFlag |= QUA_FR_AUTH_NOT_SIGNED;
                 break;
            case AUTH_NO_KEY:
                 qualFlag |= QUA_FR_AUTH_NO_CERT;
                 break;
            default:
                 qualFlag |= QUA_FR_NOT_AUTHENTICATED;
                 break;
         }
      }
      /* CD-1.0 does not support frame signing  */
      else qualFlag |= QUA_FR_AUTH_NA;
   }

   /* Check if the channel boundary in the waveform file is not exceeded */
   if ((slot + slotsOccupied - 1) >= ONE_DAY)
   {
      /* This data can't be merged, because it would exceed the channel
         boundary. Place too long segments on the computed offset is only
         allowed if the offset of all following segments is adjusted and
         already written data is moved to the adjusted offset. But this is
         an extreme time expensive operation. To avoid this, the data
         segment is saved at the end of the file  */
      DEBUG5(2,
           "Too long data segment for channel detected [%.6s/%.8s] [%f] [%d]",
             channel->siteName, channel->channelName,
             nominalTime, (int)slot);
      segState = SEG_TOO_LONG;
   }

   /* Enough information to write the data in the waveform file */
   samples = channel->nbrOfSamples;

   if (isRestartWfdisc)
   {
      /* Take offset from the wfdisc when a wfdisc is merged and
         check if this a non merge-able segment indicated by a too
         high offset */
      offset = channel->channelOffset;

      if (offset >= (pDay->channelWavOffset +
           (pChannelHead->expSampleRate * ONE_DAY * pWaveFormHead->sampleSize)))
      {
         uint32_t endOfRecord;

         segState = SEG_TOO_LONG;

         /* Check if the size of the waveform file has to be adjusted.
            This could happen if too long segments are stored at the end
            of the computed waveform file (size). Don't try to get the file
            size from the file system, because the reported file size
            is not the required one as long as the last byte of a channel
            has not been written! */
         endOfRecord = (offset + pWaveFormHead->sampleSize *
                                 channel->nbrOfSamples);

         /* File size of waveform file has not been changed since
            the waveform file in the frameproduct table has been created.
            No need to update curLenWF here */

         if (endOfRecord > pWaveFormHead->lengthWaveFile[mergeDay])
         {
            ST_DEBUG4(2, "Adjusting file size %.6s/%.8s (%ld)",
                      channel->siteName, channel->channelName,
                      (long)endOfRecord);
            pWaveFormHead->lengthWaveFile[mergeDay] = endOfRecord;
         }
      }
   }
   else
   {
      if (SEG_TOO_LONG == segState)
      {
         /* This is overlapped data */
         offset = pWaveFormHead->lengthWaveFile[mergeDay];
         pWaveFormHead->lengthWaveFile[mergeDay] +=
                                       samples * pWaveFormHead->sampleSize;
         ST_DEBUG4(2, "Adjusting file size (2) %.6s/%.8s (%ld)",
                   channel->siteName, channel->channelName,
                   (long)pWaveFormHead->lengthWaveFile[mergeDay]);
      }
      else
      {
         offset = GET_CHANNEL_SIZE(pChannelHead->expSampleRate, slot,
                      pWaveFormHead->sampleSize) + pDay->channelWavOffset;
      }
   }

   /* Create a new segment if the found segment is marked
      as 'not merge-able' or if it is not allowed to merge the segment */
   if (0 != (SEG_DONT_MERGE & pSegment->segState) ||
       SEG_TOO_LONG == segState)
   {
      /* Create new segment if necessary and store the data at the
         end of the waveform file */
      if (NEW_SEG != result)
      {
         pSegment = createSegment(pSegHeader, MERGE_RIGHT == result ?
                                  pSegment->prevSeg : pSegment);
         if (NULL == pSegment)
         {
            /* Shutdown is done in createSegment() */
            *event = E_NO_MEM;
            return(-1);
         }
         result = NEW_SEG;
      }
   }

   pDay->filledSlots += slotsOccupied;

   /* Write data only to waveform file when data is merged. If a
      wfdiscs is merged during re-start, data have been saved to
      the waveform file before */
   if (0 == isRestartWfdisc)
   {
      senderParams_t *pCd2wParams;
      pCd2wParams = (senderParams_t *)pStation->pConfigData->configExtension;

      /* Check if the output files have to be switched */
      if (-1 == syncFiles(pStation, NO, 0, (time_t)currentDay))
      {
         pStation->whenConnected.systemFailures++;
         /* Shutdown is done in syncFiles() but not for all conditions */
         *event = E_FILES;
         return(-1);
      }

      /* Uncompress the data if a waveform file should be written or
         if a quality check of the data is requested */
      if (NULL != pStation->files.fpUsrFile ||
          (pCd2wParams->senderMode & CREATE_QUAL_RECORDS) != 0)
      {
         /* Number of samples has not be 20 in CD1.1. In this
            case the time series should be padded to a multiple
            of 20 before decompression */
         if ((char)1 == channel->compression &&
              CD1_1 == pStation->type)
         {
            samples = PAD20(samples);
         }

         /* Check if the uncompress buffer must be enlarged */
         if (-1 == checkAndAdjustInt(&(pWaveFormHead->uncompressBuf),
                                     samples + 1, /* +1 for canadian */
                                     &(pWaveFormHead->uncompressBufLng)))
         {
            /* Shutdown is done in checkAndAdjustInt() */
            *event = E_NO_MEM;
            return(-1);
         }

         /* Check timeliness of data */
         if ((int)(clfTime - channel->timeStamp + slotsOccupied/2)
              > TIMELY_THRESHOLD)
         {
            qualFlag |= QUA_NOT_TIMELY;
         }

         if (3 == pWaveFormHead->sampleSize)
         {
            /* Uncompress or convert channel data (i2/i4, s2/s3 to s4) to s4 */
            retVal = any2s3((unsigned char *)channel->channelData,
                             pWaveFormHead->uncompressBuf,
                             samples,
                             channel->dataFormat, channel->dataSize,
                             channel->compression, &qualFlag, &segMean,
                             &segM2);
         }
         else
         {
            /* Uncompress or convert channel data (i2/i4, s2/s3 to s4) to s3 */
            retVal = any2s4((unsigned char *)channel->channelData,
                             pWaveFormHead->uncompressBuf,
                             samples,
                             channel->dataFormat, channel->dataSize,
                             channel->compression, &qualFlag, &segMean,
                             &segM2);
         }
         ST_DEBUG4(3, " %.6s/%.8s: M2 %f",
                  channel->siteName, channel->channelName, segM2);

         if (-1 == retVal || 1 == retVal)
         {
            if (-1 == retVal)
            {
               /* Canadian uncompress failed */
               channel->checkPass |= WA_SAMPLES_CAN;
            }
            else
            {
               /* Unsupported data type */
               channel->checkPass |= AWA_ITYPE;
            }
            pStation->whenConnected.warnings++;
            ST_DEBUG4(1, "Uncompress failed, channel %.6s/%.8s, NT: %f",
                      channel->siteName, channel->channelName, nominalTime);
            /* Free segment or an empty segment will stay in the list */
            if (NEW_SEG == result)
                destroySegment(pSegHeader, pSegment);
            return(0);
         }

         /* SOH channels might have many constant data and no input from sensor
            detections due to slowly changing values for temperature,
            wind, battery level etc.. To avoid unreal detections, a minimum
            sample rate can be set and all detections below this threshold
            are reset here */
         if (pChannelHead->expSampleRate <= pCd2wParams->detectionThreshold)
         {
            qualFlag &= ~(QUA_CONST_DATA | QUA_NO_INPUT);
         }
      }

      /* Write data only to waveform files if requested. It is
         possible that only the quality of the data is checked */
      if (NULL != pStation->files.fpUsrFile)
      {
         uint32_t endOfRecord;

         /* Check if the channel header has been written to
            the waveform file */
         if (0 == pDay->headerIsWritten)
         {
            if (-1 == writeChannelInfo(pStation, pDay, pChannelHead,
                                       channel->siteName, channel->channelName,
                                       pWaveFormHead->sampleSize,
                                pWaveFormHead->headerLinesWritten[mergeDay]))
            {
               /* Shutdown is done in writeChannelInfo() */
               *event = E_FILES;
               return(-1);
            }
            /* Increase number of lines used in the waveform header */
            pWaveFormHead->headerLinesWritten[mergeDay]++;
         }

         /* Move file pointer to computed offset in waveform file */
         if (-1 == lseek(fileno(pStation->files.fpUsrFile), (off_t)offset,
                                SEEK_SET))
         {
            /* Should never ever happen */
            shutdownApplication("can't seek in waveform file (2)",
                                pStation->conID);
            *event = E_FILES;
            return(-1);
         }

         /* Increase current length of waveform file if necessary */
         endOfRecord = (offset + pWaveFormHead->sampleSize *
                                 channel->nbrOfSamples);

         if (endOfRecord > pWaveFormHead->curLenWF[mergeDay])
             pWaveFormHead->curLenWF[mergeDay] = endOfRecord;

         IF_DEBUG(3)
         {
            (void) fprintf(DEBUG_FP,
                   "(SD:%.2hu)[%.17s]"
                   "Save data [%s/%s] [%f] [%ld] [%d] (%ld) to %.24s (%d)\n",
                   pStation->debugLevel, pStation->conID,
                   channel->siteName, channel->channelName, nominalTime,
                   (long)offset,
                   (int)channel->nbrOfSamples * (int)pWaveFormHead->sampleSize,
                   (long)channel->dataOffset,
                   pStation->files.savedFileName, mergeDay);

            DEBUG_FOOTER;
         }

         /* Write data to waveform file */
         if (-1 == writeBinData(pStation->files.fpUsrFile,
                                pWaveFormHead->uncompressBuf,
                    (int)channel->nbrOfSamples * (int)pWaveFormHead->sampleSize,
                                pStation->conID))
         {
            /* Application has been shut down in writeBinData() */
            *event = E_FILES;
            return(-1);
         }
      }
   }

   /* Data is saved. No try to merge the waveform segments */
   if (isRestartWfdisc)
   {
      /* Push clipped state to attributes */
      if ('c' == clipped)
      {
         attributes = STS_CLIPPED;
      }
      else
         attributes = 0;

#ifdef STRANGE_ENDTIME
      dataEndTime = wfdiscEndTime;
#else
      /* Take end time from wfdisc record. The wfdisc definition of
         the end time is the *START* time of the last sample, so
         one sample (time) has to be added to the end time. If the
         expected sample rate is not precise enough, this function
         must be enhanced to pass also the real sample rate when
         a wfdisc record is loaded (during saving of the wfdisc
         the real sample rate is used to compute the start time of
         the last sample */
      dataEndTime = wfdiscEndTime +  1.0/pChannelHead->expSampleRate;
#endif
   }
   else
   {
      /* Retrieve channel status and map to CD1_1 of CD1 station */
      attributes = getAttribs(pStation->type, channel->channelStatus);
      /* Apply attribute mask to channel attributes */
      attributes &= pWaveFormHead->channelAttrMaskHd;

      /* Compute end time of data channel */
#ifdef STRANGE_ENDTIME
      dataEndTime = channel->timeStamp + ((double)channelTime/1000.0) -
                    1.0/pChannelHead->expSampleRate;
#else
      dataEndTime = channel->timeStamp + ((double)channelTime/1000.0);
#endif
   }

   retVal = 0; /* for splint */
   /* First try to merge the wfdisc, then try to merge the quality
      records */
   for (pass = WFDISC_PASS; pass <= QUALITY_PASS; pass++)
   {
      channelWaveSeg_t *pNextSeg = NULL;;

      if (QUALITY_PASS == pass)
      {
         /* Don't merge quality records when wfdiscs are loaded
            during re-start. Quality records are merged when they
            are loaded directly */
         if (isRestartWfdisc) return(retVal);

         pSegHeader = &(pDay->qualSeg);

         /* Get one of the neighbor segments, the (first) overlapped
            or an empty segment if no neighbors are found.
            If two neighbors are found the left one is returned */
         pSegment = getSegment(pSegHeader, slot, slot + slotsOccupied - 1,
                               &result);

         if (NULL == pSegment)
         {
            DEBUG4(1, "Attempt to merge quality data [%.6s/%.8s] [%d] failed",
                   channel->siteName, channel->channelName, result);
            /* Shutdown is done in getSegment() */
            *event = E_NO_MEM;
            return (-1);
         }

         IF_DEBUG(3)
         {
            (void) fprintf(DEBUG_FP,
                   "(SD:%.2hu)[%.17s]"
                   "Merge quality data [%.6s/%.8s] [%d] [%d] [%f] [%d] [%d]\n",
                   pStation->debugLevel, pStation->conID,
                   channel->siteName, channel->channelName, result, (int)slot,
                   nominalTime, (int)pSegment->firstSlot,
                   (int)pSegment->lastSlot);

            DEBUG_FOOTER;
         }
         /* Check for duplicate and complete overlapped segments */
         if (0 != (result & OVERLAP_COMPLETE) ||
             0 != (result & DUPLICATE_SEG))
         {
             DEBUG5(2, "Duplicate quality data detected [%.6s/%.8s] [%f] [%d]",
                    channel->siteName, channel->channelName,
                    nominalTime, (int)slot);
            return(retVal);
         }

         /* Partial overlaps are normally detected before in the wfdisc
            pass, but if there was a DB problem in the previous run
            of cd2wng it could happen that an overlap is detected.
            Logging a warning and continue seems to be the best
            approach here */
         if (0 != (result & (OVERLAP_LEFT | OVERLAP_RIGHT)))
         {
            char logBuf[96];

            (void)snprintf(logBuf, sizeof(logBuf),
                   "Left or right overlapped slot(s) detected [%.6s/%.8s] [%f] [%d]",
                   channel->siteName, channel->channelName,
                   nominalTime, (int)(slot + slotsOccupied - 1));
            logWarning(logBuf, "quality pass");
            /* See comment for wfdisc above */
            segState = SEG_TOO_LONG;
         }

         /* Create a new segment if the found segment is marked
            as 'not merge-able' or if it is not allowed to merge the segment */
         if (0 != (SEG_DONT_MERGE & pSegment->segState) ||
             SEG_TOO_LONG == segState)
         {
            /* Reset OVERLAP_* and DUPLICATE states */
            result &= (MERGE_RIGHT | MERGE_LEFT);
            /* Create new segment if necessary and store the data at the
               end of the waveform file */
            if (NEW_SEG != result)
            {
               pSegment = createSegment(pSegHeader, MERGE_RIGHT == result ?
                                        pSegment->prevSeg : pSegment);
               if (NULL == pSegment)
               {
                  /* Shutdown is done in createSegment() */
                  *event = E_NO_MEM;
                  return(-1);
               }
               result = NEW_SEG;
            }
         }
      }

      if (NEW_SEG == result)
      {
label_new_seg:
         /* New Segment. No neighbors or unusual length */
         pSegment->segState = segState;
         pSegment->segStartTime = channel->timeStamp;
         pSegment->segEndTime = dataEndTime;
         pSegment->samplesInSeg = channel->nbrOfSamples;
         pSegment->offsetWaveform = offset;
         pSegment->firstSlot = slot;
         pSegment->lastSlot = slot + slotsOccupied - 1;
         pSegment->calib = channel->calibFactor;
         pSegment->calper = channel->calibPeriod;
         pSegment->attributes = attributes;
         pSegment->segStatusBitMask = pWaveFormHead->channelAttrMaskHd;

         if (0 != isRestartWfdisc)
         {
            /* Don't set the cached flag for records which are only
               loaded during re-start */
            pSegment->subrecord.rid[0] = wfid;
            pSegment->subrecord.offsetSub[0] = offset;
            pSegment->subrecord.samplesInSub[0] = channel->nbrOfSamples;
         }
         else
         {
            pSegment->segState |= SEG_CACHED;
         }
         if (WFDISC_PASS == pass)
         {
            retVal = 0;
            continue;
         }
         /* Quality pass */
         pSegment->segQual = qualFlag;
         pSegment->segMinTimeliness =
                   pSegment->segMaxTimeliness =
                   pSegment->segAvgTimeliness = clfTime - channel->timeStamp;
         pSegment->segMean = segMean;
         pSegment->segM2 = segM2;

         return(retVal);
      }

      /* The new data segment has at least one neighbor */
      if (MERGE_LEFT == result)
      {
         /* |existing segment|new segment|
            Get end time of the segment which is merged with the new data */
         segEndTime = pSegment->segEndTime;
         /* And start time of new data */
         segStartTime = channel->timeStamp;
      }
      else if (MERGE_RIGHT == result)
      {
label_merge_right:
         /* |new segment|existing segment|
            Get end time of new data */
         segEndTime = dataEndTime;
         /* And start time of the segment which is merged with the new data */
         segStartTime = pSegment->segStartTime;
      }
      else
      {
         char logBuf[64];

         (void)snprintf(logBuf, sizeof(logBuf),
                   "Unexpected result [%.6s/%.8s] [%d]",
                   channel->siteName, channel->channelName, result);
         shutdownApplication(logBuf, "tryToMerge");
         *event = E_INTERNAL;
         return(-1);
      }

      /* Compute difference between the two segments */
#ifdef STRANGE_ENDTIME
      tDiff = segStartTime - segEndTime - 1.0/pChannelHead->expSampleRate;
#else
      tDiff = segStartTime - segEndTime;
#endif
      tryRight = 0;
      /*@-realcompare@*/
      /* Check if the end and start times match */
      if (fabs(tDiff) > (1.0/pChannelHead->expSampleRate))
      {
         char logBuf[128];


         if (tDiff < 0)
         {
            /* Overlaps are checked before using the nominal
               time of the frame. This check uses the channel
               time and the overlap should not be greater
               than one second, otherwise treat it as duplicate data.
               Overlaps below one second are merged, but should not
               occur, because these data deviates from the F&P definition */
            if (WFDISC_PASS == pass)
               pDay->overlaps++;

            if (tDiff < -1.0)
            {
               /* Should be detected before in the calling function */
               (void)snprintf(logBuf, sizeof(logBuf),
       "Too large overlap detected, skip data [%.8s/%.4s] %f %.6f %.6f [%f]",
                   channel->siteName, channel->channelName, nominalTime, tDiff,
                   1.0/pChannelHead->expSampleRate, segEndTime);
               logWrite(LOG_INFO, logBuf);

               if (WFDISC_PASS == pass)
               {
                  retVal = 4;
                  continue;
               }
               return(retVal);
            }
            (void)snprintf(logBuf, sizeof(logBuf),
                  "Overlap detected [%.8s/%.4s] %f %.6f %.6f [%f]",
                   channel->siteName, channel->channelName, nominalTime, tDiff,
                   1.0/pChannelHead->expSampleRate, segEndTime);
            logWrite(LOG_INFO, logBuf);
         }
         else
         {
            if (WFDISC_PASS == pass)
               pDay->gaps++;

            /* try to merge it on the right side */
            tryRight = 1;
            (void)snprintf(logBuf, sizeof(logBuf),
                  "Too large gap, split segments [%.8s/%.4s] %f %.6f %.6f [%f]",
                   channel->siteName, channel->channelName, nominalTime, tDiff,
                   1.0/pChannelHead->expSampleRate, segEndTime);
            logWrite(LOG_INFO, logBuf);
         }
      }

      if (WFDISC_PASS == pass)
      {
         /* Check if calip and calper are the same */
         if (pSegment->calib != channel->calibFactor ||
             pSegment->calper != channel->calibPeriod)
         {
            DEBUG3(2, "Calib/calper are different: %.6f, %.6f",
                    channel->calibFactor, channel->calibPeriod);
            DEBUG3(2, "Calib/calper are different: %.6f, %.6f",
                    pSegment->calib, pSegment->calper);
            tryRight = 1;
         }

         /* Check clipped status */
         if ((attributes & STS_CLIPPED) !=
             (pSegment->attributes & STS_CLIPPED))
         {
            DEBUG(2, "clipped bit mask is different");
            tryRight = 1;
         }
      }
      else /* QUALITY_PASS */
      {
         /* Check quality bits */
         if (pSegment->segQual != qualFlag)
         {
            DEBUG(2, "quality flag is different");
            tryRight = 1;
         }
         /* Add check for channel status bits here */
         /* Check for double overflow */
         if (getMeanAndM2(pSegment, segMean, segM2, samples,
                          &meanMerged, &M2merged))
         {
            DEBUG(2, "double overflow, split record");
            tryRight = 1;
         }

         /* Check status bits */
         if (pSegment->attributes != attributes)
         {
            DEBUG3(2, "status bits are different: %u <> %u",
                   (unsigned int)pSegment->attributes,
                   (unsigned int)attributes);
            tryRight = 1;
         }
      }

      if (1 == tryRight)
      {
          /* Check if the data could be merged on the right side */
          if (MERGE_RIGHT == result || NULL == pSegment->nextSeg ||
              0 != (SEG_DONT_MERGE & pSegment->nextSeg->segState) ||
              pSegment->nextSeg->firstSlot != (slot + slotsOccupied - 1 + 1))
             /*
              pSegment->nextSeg->firstSlot != (slot + 1) ||
              pSegment->nextSeg->calib != channel->calibFactor ||
              pSegment->nextSeg->calper != channel->calibPeriod ||
              (pSegment->attributes & STS_CLIPPED) !=
                                      (attributes & STS_CLIPPED))
                                      */
          {
             DEBUG3(2, "No chance: [%.8s/%.4s]", channel->siteName,
                    channel->channelName);
             /* No chance to merge the new data */
             pSegment = createSegment(pSegHeader,
                                      MERGE_RIGHT == result ?
                                      pSegment->prevSeg : pSegment);
             if (NULL == pSegment)
             {
                /* Shutdown is done in createSegment() */
                *event = E_NO_MEM;
                return(-1);
             }
             goto label_new_seg;
          }
          /* There is a chance to merge on the right side. Try it */
          pSegment = pSegment->nextSeg;
          result = MERGE_RIGHT;
          goto label_merge_right;
      }
      /*@=realcompare@*/

      /* At this point it is allowed to merge the new data with an existing
         segment */
      if (QUALITY_PASS == pass)
      {
         /* No need to treat quality flag, it is the same.
            Statistic data can be merge now, there is no
            difference between left and right merge for statistics */
         /*@-realcompare@*/
         if ((clfTime - channel->timeStamp) < pSegment->segMinTimeliness)
            pSegment->segMinTimeliness = clfTime - channel->timeStamp;

         if ((clfTime - channel->timeStamp) > pSegment->segMaxTimeliness)
            pSegment->segMaxTimeliness = clfTime - channel->timeStamp;
         /*@=realcompare@*/

         pSegment->segAvgTimeliness = ((pSegment->segAvgTimeliness *
                                        pSegment->samplesInSeg) +
                                        ((clfTime - channel->timeStamp)
                                         * channel->nbrOfSamples)) /
                                        (pSegment->samplesInSeg +
                                         channel->nbrOfSamples);

         pSegment->segMean = meanMerged;
         pSegment->segM2 = M2merged;
      }

      if (MERGE_RIGHT == result)
      {
         /* Merge with the following segment. A right merge
            does not occur when a wfdisc is loaded, because
            wfdisc are read sorted by time */
         pSegment->segState = SEG_NORMAL | SEG_CACHED;
         pSegment->segStartTime = channel->timeStamp;
         pSegment->samplesInSeg += channel->nbrOfSamples;
         pSegment->firstSlot = slot;
         pSegment->offsetWaveform = offset;
         /* Nothing more to do */
         if (WFDISC_PASS == pass)
         {
            retVal = 2;
            continue;
         }
         return(retVal);
      }
      if (MERGE_LEFT == result)
      {
         pSegment->segState = SEG_NORMAL | SEG_CACHED;
         /* Merge with the previous segment */
         if (0 == isRestartWfdisc)
         {
            /* Set the cached flag only for records which are not
               loaded during re-start */
            pSegment->segState = SEG_NORMAL | SEG_CACHED;
         }
         pSegment->segEndTime = dataEndTime;
         pSegment->samplesInSeg += channel->nbrOfSamples;
         pSegment->lastSlot = slot + slotsOccupied - 1;
#ifdef HAVE_DB_ACCESS
         if (isRestartWfdisc)
         {
            channelWaveSeg_t *newSegment;

            DEBUG2(3, "Consecutive segment scheduled for release: [%ld]",
                   wfid);
            /* The best approach here is to delete all but the first records
               in the DB. Otherwise it could happen, that the wrong wfdisc
               is getting updated when too short consecutive wfdisc records
               are loaded (which should have been already merged or if the
               maximum wfdisc record size have changed). First record has been
               created as NEW_SEG before. Current record is merged with
               this record now, so it is save to delete the record in
               the DB. Keep in mind that the wfdisc ID might change!! */
             newSegment = createSegment(pSegHeader, pSegment);
             if (NULL == newSegment)
             {
                /* Shutdown is done in createSegment() */
                *event = E_NO_MEM;
                return(-1);
             }
             /* The DB record is deleted when the wfdiscs are updated next
                time */
             newSegment->segState = SEG_SCHEDULED4RELEASE | SEG_CACHED;
             newSegment->subrecord.rid[0] = wfid;
             /* Setting the next values is not real necessary, but nice to
                have in the logs when running in debug mode */
             newSegment->subrecord.offsetSub[0] = offset;
             newSegment->subrecord.samplesInSub[0] = channel->nbrOfSamples;
             newSegment->segStartTime = channel->timeStamp;
             newSegment->segEndTime = dataEndTime;
             newSegment->samplesInSeg = channel->nbrOfSamples;
             newSegment->offsetWaveform = offset;
             newSegment->firstSlot = slot;
             newSegment->lastSlot = slot + slotsOccupied - 1;
             /* Update sub-records only if these values are read from a wfdisc.
                Sub-records must be read in timely order
             insertSubRec(pSegHeader, pSegment, wfid, offset,
                          channel->nbrOfSamples);
             */
         }
#endif  /* HAVE_DB_ACCESS */
      }

      /* Now check if two existing segments have to be merged.
         Keep in mind that there could be segments scheduled for release
         between the current and the next segment:
         aaaaa....................   NEW_SEG,    chain => a
         aaaaa...............bbbbb   NEW_SEG     chain => a, b
         aaaaa.....ccccc.....bbbbb   NEW_SEG     chain => a, c, b
         aaaaa<<<<<ccccc.....bbbbb   MERGE_LEFT and merge with c, c is
         aaaaaaaaaaaaaaa.....bbbbb    scheduled for release; chain => a, C, b
         aaaaaaaaaaaaaaa<<<<<bbbbb   MERGE_LEFT  and merge with b, b is
         aaaaaaaaaaaaaaaaaaaaaaaaa    scheduled for release; chain => a, C, B
       */
      pNextSeg = getNextSegment(pSegment);

      if (NULL == pNextSeg ||
          0 != (SEG_DONT_MERGE & pNextSeg->segState) ||
          pNextSeg->firstSlot != (slot + slotsOccupied - 1 + 1))
      {
         /* Nothing more to do */
         if (WFDISC_PASS == pass)
         {
            retVal = 1;
            continue;
         }
         return(retVal);
      }

      DEBUG5(2, "Attempt to merge segments [%s/%s] [%d] [%d]",
             channel->siteName, channel->channelName, (int)pSegment->lastSlot,
             (int)pNextSeg->firstSlot);

      /* Compute difference between the two segments */
#ifdef STRANGE_ENDTIME
      tDiff = pNextSeg->segStartTime - pSegment->segEndTime
              - 1.0/pChannelHead->expSampleRate;
#else
      tDiff = pNextSeg->segStartTime - pSegment->segEndTime;
#endif

      /*@-realcompare@*/
      /* Check the gap between the two segments */
      if (fabs(tDiff) > (1.0/pChannelHead->expSampleRate))
      {
         char logBuf[128];


         if (tDiff < 0)
         {
            /* Overlaps are checked before using the nominal
               time of the frame. This check uses the channel
               time and the overlap should not be greater
               than one second, otherwise treat it as duplicate data.
               Overlaps below one second are merged, but should not
               occur, because these data deviates from the F&P definition */
            if (WFDISC_PASS == pass)
               pDay->overlaps++;

            if (tDiff < -1.0)
            {
               /* Should be detected before in the calling function */
               (void)snprintf(logBuf, sizeof(logBuf),
       "Too large overlap detected, skip data (2)[%.8s/%.4s] %f %.6f %.6f [%f]",
                   channel->siteName, channel->channelName, nominalTime, tDiff,
                   1.0/pChannelHead->expSampleRate, segEndTime);
               logWrite(LOG_INFO, logBuf);
               if (WFDISC_PASS == pass)
               {
                  retVal = 4;
                  continue;
               }
               return(retVal);
            }
            (void)snprintf(logBuf, sizeof(logBuf),
                  "Overlap detected (2)[%.8s/%.4s] %f %.6f %.6f [%f]",
                   channel->siteName, channel->channelName, nominalTime, tDiff,
                   1.0/pChannelHead->expSampleRate, segEndTime);
            logWrite(LOG_INFO, logBuf);
         }
         else
         {
             /* Nothing more to do */
             if (WFDISC_PASS == pass)
             {
                pDay->gaps++;
                retVal = 1;
                continue;
             }
             return(retVal);
         }
      }

      if (WFDISC_PASS == pass)
      {
         /* Check if calip and calper are the same */
         if (pNextSeg->calib != channel->calibFactor ||
             pNextSeg->calper != channel->calibPeriod)
         {
            DEBUG3(2, "Calib/calper are different: %.6f, %.6f",
                    channel->calibFactor, channel->calibPeriod);
            DEBUG3(2, "Calib/calper are different: %.6f, %.6f",
                    pNextSeg->calib, pNextSeg->calper);
            /* Nothing more to do */
            if (WFDISC_PASS == pass)
            {
               retVal = 1;
               continue;
            }
            return(retVal);
         }
         /*@=realcompare@*/

         /* Check clipped status */
         if ((attributes & STS_CLIPPED) !=
             (pNextSeg->attributes & STS_CLIPPED))
         {
            DEBUG(2, "clipped bit mask is different");
            /* Nothing more to do */
            if (WFDISC_PASS == pass)
            {
               retVal = 1;
               continue;
            }
            return(retVal);
         }
      }
      else /* QUALITY_PASS */
      {
         if (pSegment->segQual != qualFlag)
         {
            DEBUG(2, "quality flag is different");
            return(retVal);
         }
         /* Check for double overflow */
         if (getMeanAndM2(pSegment, pNextSeg->segMean, pNextSeg->segM2,
                          pNextSeg->samplesInSeg, &meanMerged, &M2merged))
         {
            DEBUG(2, "double overflow, no merge");
            return(retVal);
         }
      }

      /* Merge to consecutive segments */
      pSegment->segState = SEG_NORMAL | SEG_CACHED;
      pSegment->segEndTime = pNextSeg->segEndTime;
      pSegment->samplesInSeg += pNextSeg->samplesInSeg;
      pSegment->lastSlot = pNextSeg->lastSlot;

      /* Don't update the sub-records here */

      /* Setting the 'scheduled for release' flag is done here to
         have the same path in the application for file and DB mode.
         If you don't want to have this, check if the merged segment
         has at least a wfdisc ID and release the segment with
         destroySegment(pSegHeader, pSegment->nextSeg) otherwise
         The 'scheduled for released' flag this is checked when the wfdisc's
         are updated */
      pNextSeg->segState = SEG_SCHEDULED4RELEASE | SEG_CACHED;
      if (WFDISC_PASS == pass)
      {
         retVal = 3;
         continue;
      }

      /* Quality pass */
      /* Merge the statistics of the two segments */
      /*@-realcompare@*/
      if (pNextSeg->segMinTimeliness < pSegment->segMinTimeliness)
         pSegment->segMinTimeliness = pNextSeg->segMinTimeliness;

      if (pNextSeg->segMaxTimeliness > pSegment->segMaxTimeliness)
         pSegment->segMaxTimeliness = pNextSeg->segMaxTimeliness;
      /*@=realcompare@*/

      pSegment->segAvgTimeliness = ((pSegment->segAvgTimeliness *
                                     pSegment->samplesInSeg) +
                                     (pNextSeg->segAvgTimeliness *
                                      pNextSeg->samplesInSeg)) /
                                     (pSegment->samplesInSeg +
                                      pNextSeg->samplesInSeg);

      pSegment->segMean = meanMerged;
      pSegment->segM2 = M2merged;
      return(retVal);
   }
   /* This point should not be reached. The function should return in
      the for loop above! */
   DEBUG4(1, "Internal error (tryToMerge): invalid exit point [%.6s/%.8s] [%d]",
          channel->siteName, channel->channelName, result);
   shutdownApplication("Invalid exit point", "tryToMerge");
   *event = E_INTERNAL;
   return (-1);
}

/**
 * @short load quality record
 *
 * @description This function loads all quality records for a given
 * @description day. This function must be called before data of a
 * @description new day is merged and after the wfdisc records are read.
 *
 * @param pStation        pointer to station structure
 * @param loadDate        date of quality record file to load
 * @param pWaveFormHead   pointer to waveform header
 *
 * @return NO_ERROR on success or event which caused the error
 */

static uint16_t
loadQualityRecordsFromFile(station_t *pStation, uint32_t loadDate,
                          waveFormHead_t *pWaveFormHead)
{
   int items,
       result,
       dayInWindow,
       posInChMask;
   double endTime;
   FILE *fp;
   channelHead_t *pChannelHead;
   CDdataChannel_t *channel;
   channelWaveSeg_t *pSegment;
   segmentHead_t *pSegHeader;
   dayHeader_t *pDay;

   DEBUG(3, "Load quality records");

   /* Open Wfdisc file for given day */
   fp = openOld(pStation, (time_t)loadDate, CREATE_APP_FILE);

   /* Check if a quality record file has been written */
   if (NULL == fp)
   {
      return(NO_ERROR);
   }

   channel = &(pWaveFormHead->tmpChannel);

   /* Set file pointer to begin of file */
   rewind(fp);

   /* Read quality records records */
   do
   {
      unsigned int startSlot,
                   channelStatus,
                   channelStatusMask;
      unsigned short int segQual;
      uint32_t slotsOccupied;
      double segMinTimeliness,
             segMaxTimeliness,
             segAvgTimeliness;
      unsigned long samplesInSegment,
                    fOff;
      long chanId;
      double segMean,
             segM2;

      /* Load one record */
      items = fscanf(fp, QUAL_12_IN,
                     channel->siteName, channel->channelName,
                     &(channel->timeStamp), /*record ID, */ &chanId,
                     &endTime, &samplesInSegment,
                     &startSlot, &segQual, &channelStatus, &channelStatusMask,
                     /* Timeliness threshold, */
                     &segMinTimeliness, &segAvgTimeliness,
                     &segMaxTimeliness, &segMean, /* rmsDiff */
                     &segM2, &fOff /*, wffid, ngoodsamp, lddate */);

      ST_DEBUG2(2, "qItems=%d", items);

      /* Check if End of File is reached */
      if (items == -1)
      {
         if (feof(fp))
         {
            break;
         }
         /* Otherwise clear error and continue */
         logWarning("temporary error reading quality record", pStation->conID);
         clearerr(fp);
         continue;
      }

      /* Check if this is a correct quality record */
      if (items != 16)
      {
         logWarning("Ignoring invalid quality record", pStation->conID);
         continue;
      }

      IF_DEBUG (3)
      {
         (void) fprintf(DEBUG_FP,  "(SD:%.2hu)[%.17s]"QUAL_12,
                     pStation->debugLevel, pStation->conID,
                     channel->siteName, channel->channelName,
                     channel->timeStamp, -1L, chanId,
                     endTime, samplesInSegment,
                     startSlot, segQual, channelStatus, channelStatusMask,
                     TIMELY_THRESHOLD,
                     segMinTimeliness, segMaxTimeliness,
                     segAvgTimeliness, segMean,
                     0.0, /* rmsDiff */
                     segM2, fOff, -1L, 0L, "-" /* lddate */);
          DEBUG_FOOTER;
      }

      /* Compute day in time window */
      dayInWindow = (int)(loadDate - pWaveFormHead->startTimeWindow)/ONE_DAY;
      if (dayInWindow >= MAX_MERGE_DAYS)
      {
         shutdownApplication("Invalid dayInWindow",
                             "loadQualityRecordsFromFile");
         (void)(fclose(fp));
         return(E_INTERNAL);
      }

      /* The channel mask contain an entry for each channel
       * found in the quality records. The problem is, that the waveform
       * file contains external channel names, whereas the quality records
       * contain internal channel names. Therefore the channel mask has to
       * be searched using the external channel name if necessary. It is
       * required that cd2w has at least the same information when it is
       * re-started as it has in the previous run!!
       */
      posInChMask = checkChMaskPosition(channel->siteName,
                                        channel->channelName,
                                        &(pStation->chCache));
      if (-1 == posInChMask)
      {
         /* TODO: make this an error in non strict mode but don't stop.
                  Pre 2.1.45 can write quality records if no sensor
                  info was available */
         ST_DEBUG3(1, "No channel index for %.6s/%.8s - loadQualityFromFile",
                   channel->siteName, channel->channelName);
         shutdownApplication("No channel index found while reading file",
                             "loadQualityRecordsFromFile");
         (void)(fclose(fp));
         return(E_INTERNAL);
      }

      pChannelHead = pWaveFormHead->channelHeads[posInChMask];

      /* Check if the channel header exists, otherwise stop
         application. Channels are created in loadChannelDescription()
         and the quality record can't refer to non existing channels */
      if (NULL == pChannelHead)
      {
         ST_DEBUG3(2, "No channel header for %.6s/%.8s - loadQualityFromFile",
                   channel->siteName, channel->channelName);
         shutdownApplication("No channel header found while reading file",
                             "loadQualityRecordsFromFile");
         (void)(fclose(fp));
         return(E_INTERNAL);
      }

      if (startSlot > ONE_DAY)
      {
         /* TODO: This value might be off by one slot!!
            If the channel is sorted by time, that segment must have
            a left overlap with the preceding or a right overlap
            with the following segment  */
         startSlot = (unsigned int)(((uint32_t)channel->timeStamp) - loadDate);
	 ST_DEBUG(2, "too high start slot");
      }

      /* Compute time this wfdisc record encompasses in milliseconds */
      /*
      channelTime = ((uint32_t)(samplesInSegment) / pChannelHead->expSampleRate
                                * 1000);
      channel->nbrOfSamples = (uint32_t)samplesInSegment;

      retVal = tryToMerge(channel, 0, channelTime, pChannelHead, startSlot,
                          pStation, pWaveFormHead, 0, dayInWindow, 1,
                          '-', endTime, -1L, 0.0);
      ST_DEBUG2(2, "tryToMerge (quality) returns %d", retVal);
      */

      /* Compute number of occupied slots */
      slotsOccupied = (uint32_t)(samplesInSegment)/pChannelHead->expSampleRate;

      pDay = &(pChannelHead->day[dayInWindow]);
      pSegHeader = &(pDay->qualSeg);

      /* Get a segment. There is no need to merge it with other
         segments because a quality record has always the largest
         possible size and this day has no quality records processed
         so far */
      pSegment = getSegment(pSegHeader, (uint32_t)startSlot,
                            startSlot + slotsOccupied - 1, &result);

      if (NULL == pSegment)
      {
         /* Shutdown is started in the called function */
         DEBUG4(1, "Attempt to get segment [%.6s/%.8s] [%d] failed",
                channel->siteName, channel->channelName, result);
         return (E_NO_MEM);
      }

      IF_DEBUG(3)
      {
         (void) fprintf(DEBUG_FP,
                "(SD:%.2hu)[%.17s]"
                "Get q-segment [%.6s/%.8s] [%d] [%d] [%u] [%d] [%d]\n",
                pStation->debugLevel, pStation->conID,
                channel->siteName, channel->channelName, result,
                (int)startSlot, (unsigned int)slotsOccupied,
                (int)pSegment->firstSlot, (int)pSegment->lastSlot);

         DEBUG_FOOTER;
      }

      /* Check if a new quality segment has to be created */
      if (NEW_SEG != result)
      {
         /* This happens when the quality flag is different or the
            value range for the square counts is exceeded */
         DEBUG3(2, "Must create new q-segment: [%.8s/%.4s]",
                    channel->siteName, channel->channelName);
         pSegment = createSegment(pSegHeader, pSegment);
         if (NULL == pSegment)
         {
            /* Shutdown is down in createSegment() */
            return(E_NO_MEM);
         }
      }

      pSegment->segState = SEG_NORMAL;
      pSegment->segStartTime = channel->timeStamp;
      /* Internal used end time is the real one, but
         end time in quality record is the start time of
         the last sample. Keep also in mind that the real sample
         rate is used when the quality record has been written, but
         the expected is used here. This might cause a very small
         difference when the segments are compared (in the debug output) */
      pSegment->segEndTime = endTime + 1.0/pChannelHead->expSampleRate;
      pSegment->samplesInSeg = (uint32_t)samplesInSegment;
      pSegment->offsetWaveform = (uint32_t)fOff;
      pSegment->firstSlot = (uint32_t)startSlot;
      pSegment->lastSlot = (uint32_t)(startSlot + slotsOccupied - 1);
      pSegment->calib = 1.0;
      pSegment->calper = 0.0;
      pSegment->attributes = (uint32_t)channelStatus;
      pSegment->segStatusBitMask = (uint32_t)channelStatusMask;
      pSegment->segQual = (uint16_t)segQual;
      pSegment->segMinTimeliness = segMinTimeliness;
      pSegment->segMaxTimeliness = segMaxTimeliness;
      pSegment->segAvgTimeliness = segAvgTimeliness;
      pSegment->segMean = segMean;
      pSegment->segM2 = segM2;
   }
   while (items > 0 && pStation->shutdownEnable == SD_NONE);

   (void)(fclose(fp));
   return(NO_ERROR);
}

/**
 * @short load wfdisc
 *
 * @description This function loads all wfdisc records for a given
 * @description day. This function must be called before a new day
 * @description is merged or the channel offset might be incorrect!!
 *
 * @param pStation        pointer to station structure
 * @param loadDate        date of wfdisc file to load
 * @param pWaveFormHead   pointer to waveform header
 *
 * @return NO_ERROR on success or event which caused the error
 */

static uint16_t
loadWfdiscRecordsFromFile(station_t *pStation, uint32_t loadDate,
                          waveFormHead_t *pWaveFormHead)
{
   char dir[LEN_WFDISC_DIR],
        dfile[33],
        datatype[3],
        clipped;
   int items,
       isFirst = 1,
       dayInWindow,
       posInChMask;
   long samplesInSegment,
        offset,
        chanId;
   double sampleRate,  /* Real sample Rate */
          endTime;
   FILE *fp;
   channelHead_t *pChannelHead;
   CDdataChannel_t *channel;
   uint16_t event = NO_ERROR;

   /* Compute day in time window */
   dayInWindow = (int)(loadDate - pWaveFormHead->startTimeWindow)/ONE_DAY;
   if (dayInWindow >= MAX_MERGE_DAYS)
   {
      shutdownApplication("Invalid dayInWindow",
                          "loadWfdiscRecordsFromFile");
      return(E_INTERNAL);
   }

   /* Open Wfdisc file for given day */
   fp = openOld(pStation, (time_t)loadDate, CREATE_WFDISC_FILE);

   /* Check if a wfdisc file has been written */
   if (NULL == fp)
   {
      return(NO_ERROR);
   }

   channel = &(pWaveFormHead->tmpChannel);

   /* Set file pointer to begin of file */
   rewind(fp);

   /* Read wfdisc records */
   do
   {
      uint32_t startSlot,
               channelTime;
      int retVal;

      items = fscanf(fp, WFDISC_WCS30_IN,
                     channel->siteName, channel->channelName,
                     &(channel->timeStamp), /* wfid */ &chanId,
                     /* jdate */ &endTime, &samplesInSegment, &sampleRate,
                     &(channel->calibFactor), &(channel->calibPeriod),
                     /* instype */ /* segtype */ datatype,
                     &clipped, dir, dfile, &offset /* commid */ /* lddate */);

      ST_DEBUG2(2, "Items=%d", items);

      /* Check if End of File is reached */
      if (items == -1)
      {
         if (feof(fp))
         {
            break;
         }
         /* Otherwise clear error and continue */
         logWarning("temporary error reading wfdisc record", pStation->conID);
         clearerr(fp);
         continue;
      }

      /* Check if this is a correct wfdisc record */
      if (items != 14)
      {
         logError("Ignoring invalid wfdisc record", pStation->conID);
         continue;
      }

      IF_DEBUG (3)
      {
         (void) fprintf(DEBUG_FP,
                        "(SD:%.2hu)[%.17s]"
                    "%.6s %.8s %f %ld %f %ld %f %f %f %.2s %c %s %s %ld\n",
                        pStation->debugLevel, pStation->conID,
                        channel->siteName, channel->channelName,
                        channel->timeStamp, /* wfid */ chanId,
                        /* jdate */ endTime, samplesInSegment, sampleRate,
                        channel->calibFactor, channel->calibPeriod,
                        /* instype */ /* segtype */ datatype,
                        clipped, dir, dfile, offset /* commid */ /* lddate */);
          DEBUG_FOOTER;
      }

      /* This are some common bugs in records which should not occur
         (theoretically, at least) */
      if (samplesInSegment == 0)
      {
         logError("0 samples in record, ignoring invalid wfdisc",
                    pStation->conID);
         continue;
      }
      /*@-realcompare@*/
      if (channel->timeStamp == endTime)
      {
         logError("start == end time, ignoring invalid wfdisc record",
                    pStation->conID);
         continue;
      }
      /*@=realcompare@*/

      /* At least one wfdisc record exists. There must be a
         corresponding waveform file and the channel information
         in the waveform file has to be read before the first wfdisc
         record is merged */
      if (isFirst)
      {
         /* Read channel info from waveform file and create all
            channels found in the waveform file. This step is necessary
            to have the correct channel start offset in all possible cases */
         if (loadChannelDescription(pStation, loadDate, pWaveFormHead,
                                    dayInWindow))
         {
            shutdownApplication("Can't load channel description",
                                pStation->conID);
            (void)(fclose(fp));
            return(E_FILES);
         }
         isFirst = 0;
      }

      /* Now the channel mask should contain an entry for each channel
       * found in the wfdisc records. The problem is, that the waveform
       * file contains external channel names, whereas the wfdisc records
       * contain internal channel names. Therefore the channel mask has to
       * be searched using the external channel name if necessary. It is
       * required that cd2w has at least the same information when it is
       * re-started as it has in the previous run!!
       */
      posInChMask = checkChMaskPosition(channel->siteName,
                                        channel->channelName,
                                        &(pStation->chCache));
      if (-1 == posInChMask)
      {
         ST_DEBUG3(1, "No channel index for %.6s/%.8s - loadWfdiscFromFile",
                   channel->siteName, channel->channelName);
         shutdownApplication("No channel index found while reading file",
                             "loadWfdiscRecordsFromFile");
         (void)(fclose(fp));
         return(E_INTERNAL);
      }

      pChannelHead = pWaveFormHead->channelHeads[posInChMask];

      /* Check if the channel header exists, otherwise stop
         application. Channels are created in loadChannelDescription()
         and the wfdisc record can't refer to non existing channels */
      if (NULL == pChannelHead)
      {
         ST_DEBUG3(2, "No channel header for %.6s/%.8s - loadWfdiscFromFile",
                   channel->siteName, channel->channelName);
         shutdownApplication("No channel header found while reading file",
                             "loadWfdiscRecordsFromFile");
         (void)(fclose(fp));
         return(E_INTERNAL);
      }

      /* At this point the channel is initialized and the wfdisc
         can be loaded. First the slot is computed from the offset.
         If the computed slot is higher than slots per day, than
         the slot is computed using the channel start time and the
         segment is marked as not merge-able. */
      startSlot = ((uint32_t)(offset) -
                   pChannelHead->day[dayInWindow].channelWavOffset) /
                   (pChannelHead->expSampleRate * pWaveFormHead->sampleSize);

      if (startSlot > ONE_DAY)
      {
         /* TODO: This value might be off by one slot!!
            If the channel is sorted by time, that segment must have
            a left overlap with the preceding or a right overlap
            with the following segment  */
         startSlot = ((uint32_t)channel->timeStamp) - loadDate;
      }

      /* Compute time this wfdisc record encompasses in milliseconds */
      channelTime = ((uint32_t)(samplesInSegment) / pChannelHead->expSampleRate
                                * 1000);
      channel->channelOffset = (uint32_t)offset;
      channel->nbrOfSamples = (uint32_t)samplesInSegment;

      retVal = tryToMerge(channel, 0, channelTime, pChannelHead, startSlot,
                          pStation, pWaveFormHead, 0, dayInWindow, 1,
                          clipped, endTime, -1L, 0.0, &event);
      ST_DEBUG2(2, "tryToMerge (wfdisc) returns %d", retVal);
      if (NO_ERROR != event) return(event);
   }
   while (items > 0 && pStation->shutdownEnable == SD_NONE);

   (void)(fclose(fp));
   return(NO_ERROR);
}

/**
 * @short              parse frame to sent
 *
 * @description        This function parses a frame
 *
 * @param buffer       pointer to buffer containing the data frame
 * @param pStation     station object
 * @param clfType      type of frame to process
 * @param size         buffer size
 * @param nowTime      current time
 *
 * @return NO_ERROR on success or event code on error
 */

static uint16_t
parseFrameInBuffer(char *buff, station_t *pStation, char clfType,
                uint32_t size, time_t nowTime)
{
   uint32_t strictMode;
   /* TODO: move to libCD (same as parseFrame2sent) */

   /* Do some consistency checks */
   if (NULL == pStation->DFinfo)
   {
      shutdownApplication("Unexpected NULL pointer, function parseFrame2sent",
                          pStation->conID);
      return (E_INTERNAL);
   }

   strictMode = pStation->systemConf & STRICT_MODE_ENABLE;

   /* Parse the frame to set Data Frame info structure */
   if (CD1_1 == pStation->type)
   {
      /* TODO: add more checks */
      /* CD 1.1 data frame. Parse header */
      (void)parseHeader(buff, &(pStation->header));
      if (pStation->header.frameType != DATA )
      {
         /* Clf file is out of sync with binary file. Don't process
            this frame and but don't set an error or cd2wng will shut
            down when this happens. This error should not occur here,
            because it is checked also before. */
         logWarning("Frame is not a data frame (out of sync?)(1)",
                    pStation->conID);
         /* Set not parsed error to prevent that the frame is
            processed or the frame signature is checked later */
         pStation->checkPass |= ER_NOT_PARSED;
         return(NO_ERROR);
      }

      /* Parse and check data frame payload */
      pStation->checkPass |=  parseAndCheckDataPayload(
                                            pStation->DFinfo,
                                            strictMode, buff,
                                            pStation->pConfigData->maxDeviation,
                                            size, (long)nowTime,
                                            &(pStation->chCache));

      ST_DEBUG2(5, "DF_1_1 checkPass is: %d", (int)pStation->checkPass);
   }
   else if (CD1 == pStation->type)
   {
      if (GOOD_FORMAT == clfType)
      {
         pStation->checkPass = NO_ERROR;
         /* Check and parse frame. No events are logged in this function */
         pStation->checkPass |= parseAndCheckDFF(pStation->DFinfo, strictMode,
                                                 buff, &(pStation->chCache));

         ST_DEBUG2(5, "DFF checkPass is: %d", (int)pStation->checkPass);
      }
      else
      {
         /* CD 1.0 data frame. Check and parse frame */
         pStation->checkPass |= parseAndCheckDF(pStation->DFinfo,
                                           strictMode, buff,
                                           pStation->pConfigData->maxDeviation,
                                           (long)nowTime,
                                           &(pStation->chCache));

         ST_DEBUG2(5, "DF checkPass is: %d", (int)pStation->checkPass);
      }
   }
   else
   {
      /* Shutdown the application */
      shutdownApplication("Unexpected station type (parseFrameInBuffer)",
                          pStation->conID);
      return(E_INTERNAL);
   }
   return(NO_ERROR);
}


/**
 * @short              Write quality record
 *
 * @description        This function writes a quality record. Quality records
 * @description        are not split like wfdisc records!
 *
 * @param pStation          station object
 * @param pSensorinfo       sensor information
 * @param pSegment          internal used (huge) wfdisc segment
 * @param lddate            loading date
 * @param wffId             waveform file ID
 *
 * @return 0 on success or -1 on error
 */


static int
writeQualityRecord(station_t *pStation, const sensor_t *pSensorinfo,
                   channelWaveSeg_t *pSegment,
                   /*@unused@*/uint32_t expSampleRate,
                   const char *lddate, long wffId)
{
   void *hdbc = NULL;
   uint32_t samplesInSeg;
   double sampleRate;  /* Real sample Rate */
#ifdef STRANGE_ENDTIME
   double startTime;
#endif /* STRANGE_ENDTIME */
   double RMSdiff = 0.0;


   /* Check if this is a valid segment pointer */
   if (NULL == pSegment)
   {
      shutdownApplication("Unexpected NULL pSegment pointer",
                          "writeQualityRecord");
      return(-1);
   }

   samplesInSeg = pSegment->samplesInSeg;
#ifdef STRANGE_ENDTIME
   /* The next three!! lines are required to get rid of rounding errors.
      Otherwise the sample rate will be inaccurate in small segments,
      e.g. 40.0000001 instead of 40.0000000. Don't remove the DEBUG
      macro or the code might get "optimized" */
   startTime = pSegment->segStartTime - 1.0/expSampleRate;
   DEBUG4(99,"endtime=%.7f starttime=%.7f sa=%.7f",
          pSegment->segEndTime, startTime, 1.0/expSampleRate);

   /* Compute *REAL* sample rate */
   sampleRate = ((double)samplesInSeg/ (pSegment->segEndTime - startTime));
#else
   /* Compute *REAL* sample rate */
   sampleRate = ((double)samplesInSeg/
                        (pSegment->segEndTime - pSegment->segStartTime));
#endif

#ifdef HAVE_DB_ACCESS
   /* Check if there is a DB handle available */
   hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);

   if (NULL != hdbc && 0 == (pSegment->segState & SEG_CACHED))
   {
      /* Segment has not been modified since last run */
      DEBUG4(3, "%.6s/%.8s: segment has not been modified %f",
             pSensorinfo->intern_sta, pSensorinfo->intern_chan,
             pSegment->segStartTime);
      return(0);
   }
#endif  /* HAVE_DB_ACCESS */

   /* Check if a quality file has to be written or if a DB handle
      is available */
   if (NULL == pStation->files.fpAppFile && NULL == hdbc)
   {
      /* Nothing to do */
      return(0);
   }

   /* Algorithm requires samples -1
      variance = M2/(n - 1)
      see en.wikipedia.org/wiki/Algorithms_for_calculating_variance
      on-line algorithm */
   if (samplesInSeg > 1)
      RMSdiff = sqrt(pSegment->segM2/(double)(samplesInSeg - 1));

#ifdef HAVE_DB_ACCESS
   if (NULL != hdbc)
   {
      int mode = UPDATE_WFDISC;
      int ret;
      cdQuality_t qr;

      DEBUG(2, "DB mode");
      /* Database mode */

      /* Check if this is a know channel and ignore the channel
         otherwise in DB mode */
      if (-1L != pSensorinfo->chanID)
      {
         /* This quality record has to be inserted or updated in the DB */
         if (-1L == pSegment->subrecord.rid[0])
         {
            SQLHDBC idHdbc = (SQLHDBC)getHdbc(pStation->pConfigData, ID_HANDLE);

            pSegment->subrecord.rid[0] =
                      getNextID(idHdbc, QUALID_STRING,
                                (configParams_t *)pStation->pConfigData);

            /* Check the ID */
            if (-1L == pSegment->subrecord.rid[0])
            {
               shutdownApplication("can't get qualid from DB", pStation->conID);
               return(-1);
            }
            mode = INSERT_WFDISC;
         }
         qr.q_site = pSensorinfo->intern_sta;
         qr.q_channel = pSensorinfo->intern_chan;
         qr.q_startTime = pSegment->segStartTime;
         qr.q_qualID = (int32_t)pSegment->subrecord.rid[0];
         qr.q_chanID = (int32_t)pSensorinfo->chanID;
         qr.q_endTime = pSegment->segEndTime - 1.0/sampleRate;
         qr.q_nSamp = samplesInSeg;
         qr.q_firstSlot = pSegment->firstSlot;
         qr.q_qualityFlags = pSegment->segQual;
         qr.q_channelStatus = pSegment->attributes;
         qr.q_channelStatusMask = pSegment->segStatusBitMask;
         qr.q_tThresh = TIMELY_THRESHOLD;
         qr.q_minTime = pSegment->segMinTimeliness;
         qr.q_avgTime = pSegment->segAvgTimeliness;
         qr.q_maxTime = pSegment->segMaxTimeliness;
         qr.q_meanCounts = pSegment->segMean;
         qr.q_rmsDiff = RMSdiff;
         qr.q_M2 = pSegment->segM2;
         qr.q_fOff = pSegment->offsetWaveform;
         qr.q_wffId = (int32_t)wffId;

         /* Insert record in DB or update record */
         ret = updateQualityInDB((SQLHDBC)hdbc, mode, &qr);
         if (-1 == ret) return(-1);
      }
   }
#endif  /* HAVE_DB_ACCESS */

   if (NULL == hdbc && /* for splint */NULL != pStation->files.fpAppFile)
   {
      /* Write quality record, return code is checked in the
         calling function */
      (void)fprintf(pStation->files.fpAppFile, QUAL_12,
                    pSensorinfo->intern_sta,
                    pSensorinfo->intern_chan,
                    pSegment->segStartTime,
                    -1L,
                    pSensorinfo->chanID,
                    pSegment->segEndTime - 1.0/sampleRate,
                    (unsigned long)samplesInSeg,
                    (unsigned int)pSegment->firstSlot,
                    (unsigned short)pSegment->segQual,
                    (unsigned int) pSegment->attributes,
                    (unsigned int) pSegment->segStatusBitMask,
                    TIMELY_THRESHOLD,
                    pSegment->segMinTimeliness,
                    pSegment->segAvgTimeliness,
                    pSegment->segMaxTimeliness,
                    pSegment->segMean,
                    RMSdiff,
                    pSegment->segM2,
                    (unsigned long)pSegment->offsetWaveform,
                    -1L,
                    -1L,
                    lddate);
   }
   return(0);
}

#define MIN_WFDISC_DURATION     10  /* Minimal wfdisc duration */

/**
 * @short              Write Wfdisc record
 *
 * @description        This function writes a Wfdisc record. The internal used
 * @description        (huge) wfdisc record is split in smaller records if its
 * @description        duration is longer than requested by the user
 *
 * @param pStation          station object
 * @param pSensorinfo       sensor information
 * @param pSegment          internal used (huge) wfdisc segment
 * @param expSampleRate     expected sample rate
 * @param lddate            loading date
 * @param dir               path to waveform file
 * @param dfile             name of waveform file
 *
 * @return 0 on success or -1 on error
 */


static int
writeWfdisc(station_t *pStation, const sensor_t *pSensorinfo,
            channelWaveSeg_t *pSegment, uint32_t expSampleRate,
            const char *lddate, const char *dir, const char *dfile,
            uint16_t sampleSize)
{
   char clipped[2],
        dType[3];
   uint32_t samplesInReqSegment,  /* Samples in requested wfdisc duration */
            offset,
            rem,
            samplesInSeg,
            samplesUsed;
   double sampleRate,  /* Real sample Rate */
          startTime,
          endTime;
   void *hdbc = NULL;
#ifdef HAVE_DB_ACCESS
   int sR = 0;
   SQLHDBC idHdbc;
#endif  /* HAVE_DB_ACCESS */


   /* Check if this is a valid segment pointer */
   if (NULL == pSegment)
   {
      shutdownApplication("Unexpected NULL pSegment pointer", pStation->conID);
      return(-1);
   }

#ifdef HAVE_DB_ACCESS
   /* Check if there is a DB handle available */
   hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);
   idHdbc = (SQLHDBC)getHdbc(pStation->pConfigData, ID_HANDLE);

   if (NULL != hdbc && 0 == (pSegment->segState & SEG_CACHED))
   {
      /* Segment has not been modified since last run */
      DEBUG4(3, "%.6s/%.8s: segment has not been modified %f",
             pSensorinfo->intern_sta, pSensorinfo->intern_chan,
             pSegment->segStartTime);
      return(0);
   }
#endif  /* HAVE_DB_ACCESS */

   /* Check if a wfdisc file has to be written or if a DB handle
      is available */
   if (NULL == pStation->files.fpWfdiscFile && NULL == hdbc)
   {
      /* Nothing to do */
      return(0);
   }

   samplesInSeg = pSegment->samplesInSeg;
#ifdef STRANGE_ENDTIME
   /* The next three!! lines are required to get rid of rounding errors.
      Otherwise the sample rate will be inaccurate in small segments,
      e.g. 40.0000001 instead of 40.0000000. Don't remove the DEBUG
      macro or the code might get "optimized" */
   startTime = pSegment->segStartTime - 1.0/expSampleRate;
   DEBUG4(99,"endtime=%.7f starttime=%.7f sa=%.7f",
          pSegment->segEndTime, startTime, 1.0/expSampleRate);

   /* Compute *REAL* sample rate */
   sampleRate = ((double)samplesInSeg/ (pSegment->segEndTime - startTime));
#else
   /* Compute *REAL* sample rate */
   sampleRate = ((double)samplesInSeg/
                        (pSegment->segEndTime - pSegment->segStartTime));
#endif

   (void)strncpy(clipped, "-", sizeof(clipped));
   (void)strncpy(dType, "s3", sizeof(dType));

   /* Check clipped status */
   if (0 != (pSegment->attributes & STS_CLIPPED))
   {
       clipped[0] = 'c';
   }

   /* Set data type */
   dType[1] = (char)sampleSize + '0';

#ifdef HAVE_DB_ACCESS
   for (sR = 0; sR < MAX_SUB_RECORDS; sR++)
   {
      IF_DEBUG (3)
      {
         (void) fprintf(DEBUG_FP,
                        "(SD:%.2hu)[%.17s]"
                        "sR=%d %ld %ld %ld\n",
                        pStation->debugLevel, pStation->conID,
                        sR,
                        pSegment->subrecord.rid[sR],
                        (long)pSegment->subrecord.offsetSub[sR],
                        (long)pSegment->subrecord.samplesInSub[sR]);
          DEBUG_FOOTER;
      }
    }
    sR = 0;
#endif /* HAVE_DB_ACCESS */

   /* Split internal used wfdisc record into wfdisc records of requested
      duration */
   samplesUsed = 0;
   startTime = pSegment->segStartTime;
   offset = pSegment->offsetWaveform;

   /* Compute the number of samples which could be stored in the waveform
      segment. The wfdisc record has to be split if it is too large for
      a single segment. In this case the first record is shorten, and
      the second record is started at the next waveform segment
      on a requested time boundary (e.g.  every 4 hours) */
   rem = pSegment->firstSlot % WFDISC_DURATION;
   if (0 != rem)
   {
      DEBUG4(3, "%.6s/%.8s: wfdisc does not start at requested time %d",
             pSensorinfo->intern_sta, pSensorinfo->intern_chan, (int)rem);
   }

   /* It is necessary for a correct restart that a wfdisc record contains
      COMPLETE slots only! For this reason the expected sample rate is
      used in the computation */
   samplesInReqSegment = (uint32_t)((WFDISC_DURATION - rem) * expSampleRate);

   /* Merge a too short wfdisc record with the next record */
   if (samplesInReqSegment < (expSampleRate * MIN_WFDISC_DURATION))
   {
       samplesInReqSegment += (uint32_t)(WFDISC_DURATION * expSampleRate);
   }

   while (samplesInSeg > samplesInReqSegment)
   {
      ST_DEBUG4(2, "o=%ld, ns = %d, sr=%f", (long)offset,
                (int)samplesInReqSegment, sampleRate);
      /* Don't split too short wfdisc records */
      if ((samplesInSeg - samplesInReqSegment) <
                          (expSampleRate * MIN_WFDISC_DURATION))
      {
         break;
      }

      samplesUsed += samplesInReqSegment;
      /* Compute wfdisc record end time */
      endTime = pSegment->segStartTime + samplesUsed/sampleRate;

      /*@-realcompare@*/
      if (pSegment->segStartTime == endTime)
      {
         /* Should not happen here. Happened when a channel subframe was merged
            into the wrong segment which has a too high time difference to the
            nominal time of the frame. But this case is checked now before when
            the data is merged */
         logSegment(pSensorinfo->intern_sta, pSegment, WFDISC_PASS);
         shutdownApplication("start == end time when saving wfdisc record",
                             pStation->conID);
         return(-1);
      }
      /*@=realcompare@*/

      if (samplesInReqSegment == 0)
      {
         logSegment(pSensorinfo->intern_sta, pSegment, WFDISC_PASS);
         shutdownApplication("0 samples when saving wfdisc record",
                             pStation->conID);
         return(-1);
      }

      /* Note: Merging the same data once in file mode and once in DB
         mode might not result in exact the same wfdisc records if the
         records are split as a result of the max. 4 hour policy.
         Except the first record of a 4 hour block, the start and end time
         and the sample rate might deviate slightly. The reason for the
         difference is that all records in file mode must be re-written,
         whereas only modified or new records are written in DB mode.
         But during merging the real sample rate gets more accurate by
         time when the size of (the internal used huge) segment grows.
         Records which are written later (all records in file mode)
         do have a more precise sample rate (which might also result in
         different start and end times) */
#ifdef HAVE_DB_ACCESS
      if (NULL != hdbc)
      {
         DEBUG2(2, "DB mode: sR=%d", sR);
         /* Database mode */
         if (pSegment->subrecord.offsetSub[sR] != offset ||
             pSegment->subrecord.samplesInSub[sR] != samplesInReqSegment)
         {
            int ret;
            int mode = UPDATE_WFDISC;

            /* Check if this is a know channel and ignore the channel
               otherwise in database mode */
            if (-1L != pSensorinfo->chanID)
            {
               /* This wfdisc record has to be inserted or updated in the DB.
                  Check if it has a wfid and request one otherwise */
               if (-1L == pSegment->subrecord.rid[sR])
               {
                  pSegment->subrecord.rid[sR] =
                            getNextID(idHdbc, WFID_STRING,
                                      (configParams_t *)pStation->pConfigData);

                  /* Check the ID */
                  if (-1L == pSegment->subrecord.rid[sR])
                  {
                     shutdownApplication("can't get wfid from DB",
                                          pStation->conID);
                     return(-1);
                  }
                  /* No ID before. This is a new record */
                  mode = INSERT_WFDISC;
               }

               /* Insert record in DB or update record */
               ret = updateWfdiscInDB((SQLHDBC)hdbc, mode,
                                      pSensorinfo->intern_sta,
                                      pSensorinfo->intern_chan,
                                      startTime, pSegment->subrecord.rid[sR],
                                      pSensorinfo->chanID,
                                      getJdateAsNumber(startTime),
                                      /* End time is defined as the start
                                         time of the last sample */
                                      endTime - 1.0/sampleRate,
                                      (long)samplesInReqSegment,
                                      sampleRate, pSegment->calib,
                                      pSegment->calper,
                                      pSensorinfo->insType, /* segtype, */
                                      dType, clipped, dir, dfile,
                                      (long)offset);
               if (-1 == ret) return(-1);
            }
            /* Remember the values for the next run */
            pSegment->subrecord.offsetSub[sR] = offset;
            pSegment->subrecord.samplesInSub[sR] = samplesInReqSegment;
         }

         /* Increment sub record count */
         sR++;
         if (sR >= MAX_SUB_RECORDS)
         {
            /* If this happens check the define of WFDISC_DURATION */
            shutdownApplication("Invalid sub record count", pStation->conID);
            return(-1);
         }
      }
#endif  /* HAVE_DB_ACCESS */

      if (NULL == hdbc && /* for splint */NULL != pStation->files.fpWfdiscFile)
      {
         /* Write wfdisc record, return code is checked in the
            calling function */
         (void)fprintf(pStation->files.fpWfdiscFile, WFDISC_WCS30A,
                       pSensorinfo->intern_sta,
                       pSensorinfo->intern_chan,
                       startTime,
                       -1L, /* wfid */
                       pSensorinfo->chanID,
                       getJdateAsNumber(startTime),
                       /* End time is defined as the start
                          time of the last sample */
                       endTime - 1.0/sampleRate,
                       (long)samplesInReqSegment,
                       sampleRate,
                       pSegment->calib,
                       pSegment->calper,
                       pSensorinfo->insType,
                       "-",   /* segtype */
                       dType,    /* datatype */
                       clipped,  /* clip */
                       dir,
                       dfile,
                       (long)offset,
                       -1L,  /* commid */
                       lddate );
      }
      /* Increase offset and start time */
      offset += (samplesInReqSegment * sampleSize);
      startTime = endTime;
      samplesInSeg -= samplesInReqSegment;

      if (0 != rem)
      {
         /* Make sure that next segment has the requested duration */
         samplesInReqSegment = (uint32_t)(WFDISC_DURATION * expSampleRate);
         rem = 0;
      }
   }

   samplesUsed += samplesInSeg;

#ifndef STRANGE_ENDTIME
   /*@-realcompare@*/
   if (pSegment->segStartTime == pSegment->segEndTime)
   {
      logSegment(pSensorinfo->intern_sta, pSegment, WFDISC_PASS);
      shutdownApplication("(2) start == end time when saving wfdisc record",
                          pStation->conID);
      return(-1);
   }
   /*@=realcompare@*/
#endif

   if (samplesInSeg == 0)
   {
      logSegment(pSensorinfo->intern_sta, pSegment, WFDISC_PASS);
      shutdownApplication("(2) 0 samples when saving wfdisc record",
                          pStation->conID);
      return(-1);
   }

   /* Write remaining wfdisc record,
      return code is checked in the calling function */

#ifdef HAVE_DB_ACCESS
   if (NULL != hdbc)
   {
      ST_DEBUG4(2, "(2) o=%ld, ns=%d, sR=%d", (long)offset,
                (int)samplesInSeg, sR);

      /* Database mode */
      if (pSegment->subrecord.offsetSub[sR] != offset ||
          pSegment->subrecord.samplesInSub[sR] != samplesInSeg)
      {
         int ret;
         int mode = UPDATE_WFDISC;

         /* Check if this is a know channel and ignore the channel
            otherwise in DB mode. */
         if (-1L != pSensorinfo->chanID)
         {
            /* This wfdisc record has to be inserted or updated in the DB */
            if (-1L == pSegment->subrecord.rid[sR])
            {
               pSegment->subrecord.rid[sR] =
                         getNextID(idHdbc, WFID_STRING,
                                   (configParams_t *)pStation->pConfigData);

               /* Check the ID */
               if (-1L == pSegment->subrecord.rid[sR])
               {
                  shutdownApplication("can't get wfid from DB",
                                       pStation->conID);
                  return(-1);
               }
               mode = INSERT_WFDISC;
            }

            /* Insert record in DB or update record*/
            ret = updateWfdiscInDB((SQLHDBC)hdbc, mode,
                                    pSensorinfo->intern_sta,
			            pSensorinfo->intern_chan,
                                    startTime, pSegment->subrecord.rid[sR],
                                    pSensorinfo->chanID,
                                    getJdateAsNumber(startTime),
                                    /* End time is defined as the start
                                       time of the last sample */
#ifdef STRANGE_ENDTIME
                                    pSegment->segEndTime,
#else
                                    pSegment->segEndTime - 1.0/sampleRate,
#endif
                                    (long)samplesInSeg, sampleRate,
                                    pSegment->calib,  pSegment->calper,
                                    pSensorinfo->insType, /* segtype, */
                                    dType, clipped,
                                    dir, dfile, (long)offset);
            if (-1 == ret) return(-1);
         }
         /* Remember the values for the next run */
         pSegment->subrecord.offsetSub[sR] = offset;
         pSegment->subrecord.samplesInSub[sR] = samplesInSeg;
      }
   }

#endif  /* HAVE_DB_ACCESS */

   if (NULL == hdbc && /* for splint */NULL != pStation->files.fpWfdiscFile)
   {
      (void)fprintf(pStation->files.fpWfdiscFile, WFDISC_WCS30A,
                    pSensorinfo->intern_sta,
                    pSensorinfo->intern_chan,
                    startTime,
                    -1L, /* wfid */
                    pSensorinfo->chanID,
                    getJdateAsNumber(startTime),
                    /* End time is defined as the start
                       time of the last sample */
#ifdef STRANGE_ENDTIME
                    pSegment->segEndTime,
#else
                    pSegment->segEndTime - 1.0/sampleRate,
#endif
                    (long)samplesInSeg,
                    sampleRate,
                    pSegment->calib,
                    pSegment->calper,
                    pSensorinfo->insType,
                    "-",   /* segtype */
                    dType,    /* datatype */
                    clipped,  /* clip */
                    dir,
                    dfile,
                    (long)offset,
                    (long)-1,  /* commid */
                    lddate );
   }
   DEBUG4(3, "%.6s/%.8s: Difference in end times is %f",
          pSensorinfo->intern_sta, pSensorinfo->intern_chan,
          pSegment->segStartTime +
                    samplesUsed/sampleRate - pSegment->segEndTime);
   return(0);
}

/**
 * @short              Split SC name into site and channel
 *
 * @description        This function splits a string containing a site
 * @description        and channel name separated by a slash into a
 * @description        site and channel name. Site name  must have space for
 * @description        LEN_SITE_NAME+1 characters. Channel name must have
 * @description        space for LEN_CHANNEL_NAME+1 characters
 *
 * @param siteChan     site and channel name
 * @param site         site name (out)
 * @param channel      channel name (out)
 *
 * @return nothing
 */

static void
splitSC(const char *siteChan, /*@out@*/char *site, /*@out@*/char *channel)
{
   int ij;

   *site = *channel = '\0'; /* for splint */

   /* Copy site name */
   for (ij = 0; ij < LEN_SITE_NAME; ij++)
   {
      if ('\0' == *siteChan || '/' == *siteChan) { *site = '\0'; break; }
      *site++ = *siteChan++;
   }
   if ('/' == *siteChan) siteChan++;

   /* Copy channel name */
   for (ij = 0; ij < LEN_CHANNEL_NAME; ij++)
   {
       if ('\0' == *siteChan) { *channel = '\0'; break; }
       *channel++ = *siteChan++;
   }
}

/**
 * @short Update binary file in frame product table
 *
 * @description This function updates the binary file size and end time
 * @description in frame product table for the given file ID.
 *
 * @param hdbc         database handle
 * @param pLastSend    pointer to the entry in the IRL last sent
 * @param cdtype       CD type of station (CD-1.0 or CD-1.1)
 * @param hint         size of the binary file (if known) or 0
 *
 * @return 0 on success or -1 on error
 */

#ifdef HAVE_DB_ACCESS
static int
updateReceiverFrameProduct(/*@null@*/HDBC hdbc,
                           /*@null@*/intRefListParams_t *pLastSent,
                           unsigned char cdtype, uint32_t hint)
{
   if (NULL == hdbc) return(0);

   if (NULL == pLastSent)
   {
      DEBUG(3, "nothing to do in updateReceiverFrameProduct");
      return(0);
   }

   /* Check if the file object owns a frame ID. This is not an
      error condition! */
   if (-1 == pLastSent->pFileObj->fid)
   {
      DEBUG(2, "frid is -1");
      return(0);
   }

   /*@-compdestroy@*/
   /*@-mustfreefresh@*/
   if (0 == hint)
   {
      struct stat statBuf;
      char buffer[QUERY_BUFFER_SIZE];

      memset(&statBuf, 0, sizeof(struct stat)); /* for splint */

      (void)snprintf(buffer, sizeof(buffer), "%s.%s",
                     pLastSent->pFileObj->fileNameAndPath,
                     CD1 == cdtype ? "10bin" : "11bin");
      DEBUG2(2,"Stat [%s]", buffer);
      if (0 != stat(buffer, &statBuf))
      {
         /* File could have been moved or removed if there was a
            long outage and a file pointer might not be available for
            'hint' value. Older versions of cd2wng have stopped here,
            but it seems that in such a case update of the frameproduct
            table is not necessary (ISI-56). It is recommended to use
            time based synchronization to avoid cases where the frameproduct
            table is not updated a long time. Old comment also stated this
            might be caused by a CD protocol type change, but this should
            not have an impact here. This should be noticed before */
         logWarning("UpdateReceiverFrameproduct: can't stat binary file",
                    buffer);
         return(0);
      }
      return(updateFrameProduct(hdbc, pLastSent->pFileObj->fid,
                                (uint32_t)statBuf.st_size, pLastSent->clfTime));
   }
   /*@=compdestroy@*/
   /*@=mustfreefresh@*/
   return(updateFrameProduct(hdbc, pLastSent->pFileObj->fid,
                             (uint32_t)hint, pLastSent->clfTime));

}
#endif /* HAVE_DB_ACCESS */

/**
 * @short              Sync wfdisc records
 *
 * @description        This function syncs wfdisc records and logs some
 * @description        statistics. It syncs all channels for
 * @description        for the given number of days starting on the first
 * @description        day of the day window. Wfdisc record files are
 * @description        purged before the updated wfdisc records are written
 * @description        This function also updates the re-start information
 *
 * @param pStation     station object
 * @param days         number of days to consider
 * @param syncFlag     write only statistics
 *
 * @return NO_ERROR on success or error event on error
 *
 * |.........|.........|.........|.........|.........| requested size
 *     aaaaaaaaaaaaaaaaaaaa  bbbbbbbbbb        cc      internal wfdiscs
 *     a1a1a1a2a2a2a2a2a3a3  b1b1b3b3b3        c1      subrecords (used in DB)
 * |.........|.........|.........|.........|.........|
 *                         nn                          new data arrived
 *     aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa        cc      internal wfdiscs
 *     a1a1a1a2a2a2a2a2a3a3a3..........        c1      delete b1,b2,b3 (in DB)
 *     a1a1a1a2a2a2a2a2a3a3a3a3a3a4a4a4        c1      upd a3, ins a4 (in DB)
 * |.........|.........|.........|.........|.........|
 *                                           nn        new data arrived
 *     aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa      cccc      internal wfdiscs
 *     a1a1a1a2a2a2a2a2a3a3a3a3a3a4a4a4      c1c1      update c1 (in DB)
 * |.........|.........|.........|.........|.........|
 *                                       nnnn          new data arrived
 *     aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  cccccccc      internal wfdiscs
 *     a1a1a1a2a2a2a2a2a3a3a3a3a3a4a4a4  c1c2c2c2      upd c1, ins c2 (in DB)
 * |.........|.........|.........|.........|.........|
 * nn                                                  new data arrived
 * dd  aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  cccccccc      internal wfdiscs
 * d1  a1a1a1a2a2a2a2a2a3a3a3a3a3a5a4a4  c1c2c2c2      insert d1 (in DB)
 * |.........|.........|.........|.........|.........|
 *   nn                                                new data arrived
 * dddddddddddddddddddddddddddddddddddd  cccccccc      internal wfdiscs
 * d1..................................  c1c2c2c2      del a1-4 (in DB)
 * d1d1d1d1d1d2d2d2d2d2d3d3d3d3d3d4d4d4  c1c2c2c2      upd d1, ins d2-4 (in DB)
 *
 *     Optimized version: when merge right with existing segment update
 *     last SR from left, delete first SR of right element and append
 *     remaining SR of right into left segment (not implemented yet)
 * |.........|.........|.........|.........|.........|
 *     a1a1a1a2a2a2a2a2a3a3  b1b1b2b2b2b2b2b3b3b3  c1  subrecords (used in DB)
 * |.........|.........|.........|.........|.........|
 *                         nn                          new data arrived
 *     aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa  cc  internal wfdiscs
 *     a1a1a1a2a2a2a2a2a3a3......b2b2b2b2b2b3b3b3  c1  delete b1
 * |.........|.........|.........|.........|.........|
 *     a1a1a1a2a2a2a2a2a3a3a3a3a3a4a4a4a4a4a5a5a5  c1  update a3, a4=b2, a5=b3
 */

static uint16_t
syncWfdiscs(station_t *pStation, int days, int syncFlag)
{
   cdsend_t *pCdSend;
   waveFormHead_t *pWaveFormHead;
   channelHead_t *pChannelHead;
   chName_t *pChannelCache;
   uint16_t ii;
   int dayInWindow;
   char lddate[18],
        dir[LEN_WFDISC_DIR],
        dfile[33];
   channelWaveSeg_t *actSegment;
   sensor_t sensorinfo,
            *pSensorinfo;
   channelWaveSeg_t *segment2delete;
   senderParams_t *pCd2wParams;
#ifdef HAVE_DB_ACCESS
   void *hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);
#endif /* HAVE_DB_ACCESS */


   /* Do not check shutdown flag in this function or some
      wfdisc or quality records will not get saved! */

   /* Do some consistency checks */
   if (NULL == pStation->appExtension)
   {
      shutdownApplication("Unexpected NULL pointer (syncWfdiscs)",
                          pStation->conID);
      return(E_INTERNAL);
   }

   pCdSend = (cdsend_t *)(pStation->appExtension);
   pCd2wParams = (senderParams_t *)pStation->pConfigData->configExtension;

   if (NULL == pCdSend || NULL == pCdSend->pRefListHead)
   {
      shutdownApplication("Unexpected NULL pointer (2)(syncWfdiscs)",
                          pStation->conID);
      return(E_INTERNAL);
   }

   if (SYNC_WFDISCS == syncFlag)
   {
      /* Reset timer even if nothing has to be synced! */
      if (0 != pCd2wParams->syncTimeInterval)
      {
         struct timeval nowTime;

         (void) gettimeofday(&nowTime, NULL);
         pCdSend->nextSyncTime = nowTime.tv_sec +
                                 (time_t)pCd2wParams->syncTimeInterval;
      }

      /* Check if frames have to be synced */
      if (0 == pCd2wParams->refListStoreInterval &&
          0 == pCd2wParams->syncTimeInterval)
      {
         logWarning("syncing wfdiscs/restart is switched off!", "syncWfdiscs");
         return(NO_ERROR);
      }

      /* Check if frames has been processed */
      if (0 == pCdSend->pRefListHead->framesSinceLastSave)
      {
         ST_DEBUG(2, "syncWfdiscs: no frames processed, nothing to do");
         return(NO_ERROR);
      }
   }

   /* Check if the waveform header is already initialized, otherwise
      return because there is nothing to do (for statistics).
      Check also if the station has a channel name cache */
   if (NULL == pCdSend->addAppExtension ||
       0 == pStation->chCache.nrOfChannels)
   {
      ST_DEBUG(2, "syncWfdiscs: nothing to do");
      return(NO_ERROR);
   }


   pWaveFormHead = (waveFormHead_t *)pCdSend->addAppExtension;

   ST_DEBUG(2, "syncWfdiscs: start syncing");

   if (ONLY_STATISTICS == syncFlag)
   {
      char logBuffer[LOGSIZE];  /* per Thread or on stack!! */

      /* Log number of obsolete frames and frames dropped by filter rule */
     (void)snprintf(logBuffer, sizeof(logBuffer),
                    "%.17s: Frame statistics [%u/%u]",
                    pStation->conID,
                    (unsigned int)pWaveFormHead->obsoleteFrames,
                    (unsigned int)pWaveFormHead->framesDropped);
      logWrite(LOG_INFO, logBuffer);
   }

   /* Check number of days */
   if (days > MAX_MERGE_DAYS) days = MAX_MERGE_DAYS;

   /* Sync wfdisc records on per day base */
   for (dayInWindow = 0; dayInWindow < days; dayInWindow++)
   {
      int hasWfdiscRecs = 0;

      pChannelCache = pStation->chCache.firstCh;

      /* First check if there are records available for this day */
      for (ii = 0; ii < pStation->chCache.nrOfChannels; ii++,
                                  pChannelCache = pChannelCache->nextChannel)
      {
         dayHeader_t *pDay;
         segmentHead_t *pSegHead;

         if (NULL == pChannelCache)
         {
            shutdownApplication("Invalid channel cache pointer (syncWfdiscs)",
                                pStation->conID);
            return(E_INTERNAL);
         }

         /* Access and check the channel header */
         pChannelHead = pWaveFormHead->channelHeads[ii];
         if (NULL == pChannelHead)
         {
            /* This is not an error. That happens short after start up
               if the channel has not been created */
            continue; /* with next channel */
         }

         /* Set day to process */
         pDay = &(pChannelHead->day[dayInWindow]);

         if (NULL == pDay)
         {
             shutdownApplication("Invalid day pointer (syncWfdiscs)",
                                 pStation->conID);
             return(E_INTERNAL);
         }

         if (0 == pDay->filledSlots)
         {
            /* Nothing to do for this day */
            continue; /* with next channel */
         }

         pSegHead = &(pDay->wfdiscSeg);
         /* Check if there are records for this day available */
         if (0 == pSegHead->nbrOfSegments)
         {
            /* Do not exit the loop here. It could happen that the receiver
               is running a very long time and the station stops to send
               a channel for some time */
            continue; /* with next channel */
         }

         /* Check if only statistic should be written */
         if (ONLY_STATISTICS == syncFlag)
         {
            char logBuffer[LOGSIZE];  /* per Thread or on stack!! */

            /* "site/channel-day", "segments", "gap", "olap", "dup",
               "2long", "2short", "2Old", "missing" */
           (void)snprintf(logBuffer, sizeof(logBuffer),
                          "%10s-%02d: %d/%d/%d/%d/%d/%d/%d\n",
                          pChannelCache->scName[0], dayInWindow,
                          (int)pSegHead->nbrOfSegments,
                          (int)pDay->gaps, (int)pDay->overlaps,
                          (int)pDay->duplicates, (int)pDay->tooLong,
                          (int)pDay->tooShort,
                          (int)((pChannelHead->nbrOfSlots - pDay->filledSlots)/
                                 pChannelHead->expChannelTimeLng));
            logWrite(LOG_INFO, logBuffer);
         }
         else
         {
            hasWfdiscRecs = 1;
            /* Break out here requires the checks for NULL pointers
               have to be done again! */
            break;
         }
      }

      if (ONLY_STATISTICS == syncFlag)
      {
         ST_DEBUG2(2, "Only statistics, nothing to sync for day %d",
                   (int)dayInWindow);
         continue; /* with next day */
      }

      if (0 == hasWfdiscRecs)
      {
         ST_DEBUG2(2, "Nothing to sync for day %d", (int)dayInWindow);
         continue; /* with next day */
      }

      /* There might be something to sync. That does not mean that really
         data has to be written to the wfdisc/cdquality files/records,
         especially when the sync has been triggered by a day switch!! */
      memset(lddate, 0, sizeof(lddate)); /* for splint */

      /* Set wfdisc loading date */
      (void) snprintf(lddate, sizeof(lddate), "%u.%05u",
                      (unsigned int)time(NULL), 0);

      /* Check if the output files have to be switched */
      if (-1 == syncFiles(pStation, NO, 0,
             (time_t)(pWaveFormHead->startTimeWindow + dayInWindow * ONE_DAY)))
      {
         /* Shutdown is initiated in syncFiles */
         pStation->whenConnected.systemFailures++;
         return(E_FILES);
      }

      /* Set dir and dfile for wfdiscs. Absolute wfdisc paths
         are not longer used to simplify configuration. Symbolic
         links or a proper filePath should be used instead */
      (void)strncpy(dir, pStation->files.savedFilePath, LEN_WFDISC_DIR);
      BUILD_FILE_NAME(dfile, sizeof(dfile), pStation->conID,
                      pStation->files.savedFileName, "mwf", "");

      /* Purge the wfdisc file before write to it. If CD2W is in
         DB mode, fpWfdiscFile should be NULL */
      if (NULL != pStation->files.fpWfdiscFile)
      {
         /* Flush to be on the safe side */
         (void)fflush(pStation->files.fpWfdiscFile);

         /* Truncate the wfdisc file before write to it */
         if (-1 == ftruncate(fileno(pStation->files.fpWfdiscFile), 0L))
         {
            shutdownApplication("can't truncate wfdisc file",
                                pStation->files.savedFileName);
            return(E_FILES);
         }
      }
      /* Purge the quality file before write to it. If CD2W is in
         DB mode, fpAppFile should be NULL */
      if (NULL != pStation->files.fpAppFile)
      {
         /* Flush to be on the safe side */
         (void)fflush(pStation->files.fpAppFile);

         /* Truncate the wfdisc file before write to it */
         if (-1 == ftruncate(fileno(pStation->files.fpAppFile), 0L))
         {
            shutdownApplication("can't truncate quality file",
                                pStation->files.savedFileName);
            return(E_FILES);
         }
      }

#ifdef HAVE_DB_ACCESS
      /* Create waveform file in frameproduct table. This must be done here
         because the cdquality records contain a reference to this file. */
      if (NULL != hdbc)
      {
         long wffId;
         char pathAndFile[128];
         float loadDate = (float)pWaveFormHead->startTimeWindow +
                                 (dayInWindow * ONE_DAY);

         (void)snprintf(pathAndFile, sizeof(pathAndFile), "%s/%s",
                        dir, dfile);
         wffId = pWaveFormHead->waveFileID[dayInWindow];
         if (-1L == wffId)
         {
            /* Query frameproduct table for waveform file and create
               an entry if it does not exist */
            wffId = getFrameProduct((SQLHDBC)hdbc, pStation, pathAndFile,
                                    CD_MWF_FILE, CREATE_ID, loadDate, NULL);
            /* Save the waveform file ID */
            pWaveFormHead->waveFileID[dayInWindow] = wffId;
         }
         else if (pWaveFormHead->curLenWF[dayInWindow] > 0)
         {
            /* Waveform file might grow in size. Error is logged
               and shutdown is started in updateFrameProduct() in case
               of error. Don't use lengthWaveFile here, because this
               is the computed length if data is available for the whole
               day, but not all data might have been received so far.
               It is also possible that no data has been processed for
               this station indicated by current length of waveform
               file is zero. The frameproduct table must not be updated
               in this case  */
            (void)updateFrameProduct((SQLHDBC)hdbc, wffId,
                                     pWaveFormHead->curLenWF[dayInWindow],
                                     loadDate + ONE_DAY - .00001);
         }
      }
#endif /* HAVE_DB_ACCESS */

      pChannelCache = pStation->chCache.firstCh;

      /* NOTE: Protect syncing with mutex lock if no duplicate segments
               should be reported after a DB problem (some of the
               records can get updated, but the restart info not)
         TODO: Use two wfdisc files in file mode and switch the file after
               the new one has been written. Otherwise you could end up
               loosing wfdisc records when a file system error happens! */

      /* Process all channels */
      for (ii = 0; ii < pStation->chCache.nrOfChannels; ii++,
                                  pChannelCache = pChannelCache->nextChannel)
      {
         dayHeader_t *pDay;
         segmentHead_t *pSegHead;
         int pass;

         ST_DEBUG4(3, "Syncing channel %.17s, index = %d, day %d",
                   pChannelCache == NULL ? "null" : pChannelCache->scName[0],
                   (int)ii, dayInWindow);

         if (NULL == pChannelCache)
         {
            shutdownApplication("Invalid channel cache pointer (syncWfdiscs)",
                                pStation->conID);
            return(E_INTERNAL);
         }

         /* Check channel header */
         pChannelHead = pWaveFormHead->channelHeads[ii];
         if (NULL == pChannelHead)
         {
            /* This is not an error. That happens short after start up
               if the channel has not been created */
            continue; /* with next channel */
         }

         /* Set day to process */
         pDay = &(pChannelHead->day[dayInWindow]);
         if (NULL == pDay)
         {
             shutdownApplication("Invalid day pointer (syncWfdiscs)",
                                 pStation->conID);
             return(E_INTERNAL);
         }

         /* Process wfdisc records first */
         pSegHead = &(pDay->wfdiscSeg);
         ST_DEBUG2(3, "%d wfdisc segments", (int)pSegHead->nbrOfSegments);

         /* Check if there are records for this day available */
         if (0 == pSegHead->nbrOfSegments)
         {
            /* Do not exit the loop here. It could happen that the receiver
               is running a very long time and the station stops to send
               a channel for some time */
            continue; /* with next channel */
         }

         DEBUG5(99, "SR: %d, TM: %d, STR: %d, SN: %d\n",
                (int)pChannelHead->expSampleRate,
                (int)pChannelHead->expChannelTimeLng,
                (int)pWaveFormHead->startTimeWindow,
                (int)pChannelHead->nbrOfSlots);

         /* Check if data about the channel is available */
	 if (NULL == pChannelHead->sensorInfo)
	 {
            if (NULL != getHdbc(pStation->pConfigData, DB_HANDLE))
            {
               /* Don't continue here or cd2w will stop when the day is closed.
	          Omitting is done when the wfdisc is written later */
               if (0 != (pStation->systemConf & STRICT_MODE_ENABLE))
	       {
                  shutdownApplication(
                        "No sensor info available and in strict mode",
                         pChannelCache->scName[0]);
                  return(E_CONFIG);
	       }
               logWarning("No sensor info available, omit wfdisc in DB mode",
                           pChannelCache->scName[0]);
	    }

            /* Set up default sensor info structure */
            memset(&sensorinfo, 0, sizeof(sensorinfo));
            sensorinfo.chanID = -1;
            sensorinfo.insType[0] = '-';

            splitSC(pChannelCache->scName[0], sensorinfo.intern_sta,
                    sensorinfo.intern_chan);

	    pSensorinfo = &sensorinfo;
	 }
	 else
	 {
	    pSensorinfo = pChannelHead->sensorInfo;
	 }
         /*@-branchstate@*/
         /*@=branchstate@*/

         /* First process the wfdisc, then process the quality records */
         for (pass = WFDISC_PASS; pass <= QUALITY_PASS; pass++)
         {
            segment2delete = NULL;

#define CD2W_MODE (pCd2wParams->senderMode)

#define HAVE_WFDISCSCTEX 1
#ifdef HAVE_WFDISCSCTEX
            /* The code below requires that the same wfdisc/quality record
               could be created twice. This is not the case when a
               composite index has been created on 'sta', 'chan' and 'time'.
               In this case an additional iteration through the segment
               chain has to be done to remove all segments which are
               scheduled for release first to avoid a constrain violation.
               In principle it is more safe to create the new segment before
               the old one is released if another thread is allowed
               to commit the connection. Inserts in the database are not
               optimized, so it could happen that, when two segments are
               merged and split into 4 hour records afterward, some of the
               segments are identical to before (the merge) but still
               scheduled for release (and therefor deleted) and recreated
               afterwards */
            for (actSegment = pSegHead->firstSeg; actSegment != NULL;)
            {
               if (0 != (actSegment->segState & SEG_SCHEDULED4RELEASE))
               {
                  /* Check for orphaned wfdisc ID's. This happens when two
                     segments are merged together. Freeing these segments
                     after being merged is not a good idea, because this
                     will cause that these wfdisc records are not deleted
                     in the database, which is done here!! */
                  DEBUG3(3, "Release segment %d for channel %.8s",
                         (int)actSegment->firstSlot, pChannelCache->scName[0]);

                  IF_DEBUG (4)
                  {
                     logSegment(pChannelCache->scName[0], actSegment, pass);
                  }
                  /* Check if this is a know channel and ignore the channel
                     otherwise. This will waste some wfIDs, but it gives
                     the opportunity to write the records later in the wfdisc
                     table (with a special channel ID, or to write the records
                     in the quarantine table (but this will require more effort,
                     because this has also to consider what happens when the
                     channel gets a channel ID when cd2w is running!!) */
#ifdef HAVE_DB_ACCESS
	          if (-1L != pSensorinfo->chanID)
	          {
                     if ((WFDISC_PASS == pass &&
                          (CD2W_MODE & CREATE_WFDISC_RECORDS) != 0)
                         ||
                         (QUALITY_PASS == pass &&
                          (CD2W_MODE & CREATE_QUAL_RECORDS) != 0))
                     {
                       if (-1 == releaseDBrecord(pStation, actSegment, pass))
                       { return(E_SAVE_RECORD); }
                     }
                  }
#endif /* HAVE_DB_ACCESS */

                  segment2delete =  actSegment;
                  actSegment = segment2delete->nextSeg;

                  /* Free segment which is now part of a bigger segment */
                  ST_DEBUG3(3, "Destroy segment %.17s %f",
                               pChannelCache->scName[0],
                               segment2delete->segStartTime);

                  destroySegment(pSegHead, segment2delete);
                  segment2delete = NULL;
                  continue;
               }
               /* Check next segment */
               if (NULL != actSegment) /* for splint */
                  actSegment = actSegment->nextSeg;
            }
#endif /* HAVE_WFDISCSCTEX */

            /* Create or update records for current channel */
            for (actSegment = pSegHead->firstSeg; actSegment != NULL;
                 actSegment = actSegment->nextSeg)
            {
               if (NULL != segment2delete)
               {
                  /* Free segment which is now part of a bigger segment */
                  ST_DEBUG3(3, "Destroy segment %.17s %f",
                               pChannelCache->scName[0],
                               segment2delete->segStartTime);

                  destroySegment(pSegHead, segment2delete);
                  segment2delete = NULL;
               }

               IF_DEBUG (4)
               {
                  logSegment(pChannelCache->scName[0], actSegment, pass);
               }

               if (0 != (actSegment->segState & SEG_SCHEDULED4RELEASE))
               {
                  /* Check for orphaned wfdisc ID's. This happens when two
                     segments are merged together. Freeing these segments
                     after being merged is not a good idea, because this
                     will cause that these wfdisc records are not deleted
                     in the database, which is done here!! */
                  DEBUG3(3, "Release segment %d for channel %.8s",
                         (int)actSegment->firstSlot, pChannelCache->scName[0]);
                  /* Check if this is a know channel and ignore the channel
                     otherwise. This will waste some wfIDs, but it gives
                     the opportunity to write the records later in the wfdisc
                     table (with a special channel ID, or to write the records
                     in the quarantine table (but this will require more effort,
                     because this has also to consider what happens when the
                     channel gets a channel ID when cd2w is running!!) */
#ifdef HAVE_DB_ACCESS
	          if (-1L != pSensorinfo->chanID)
	          {
                     if ((WFDISC_PASS == pass &&
                          (CD2W_MODE & CREATE_WFDISC_RECORDS) != 0)
                         ||
                         (QUALITY_PASS == pass &&
                          (CD2W_MODE & CREATE_QUAL_RECORDS) != 0))
                     {
                       if (-1 == releaseDBrecord(pStation, actSegment, pass))
                       { return(E_SAVE_RECORD); }
                     }
                  }
#endif /* HAVE_DB_ACCESS */
                  actSegment->segState &= ~SEG_CACHED;

                  /* Segment MUST be delete after the new actSegment
                     is computed! */
                  segment2delete = actSegment;
               }
               else
               {
                  /* TODO: replace dir and dfile with waveform file ID */
                  if (WFDISC_PASS == pass)
                  {
                     if ((CD2W_MODE & CREATE_WFDISC_RECORDS) != 0)
                     {
                        /* Write wfdisc record(s) now */
                        if (-1 == writeWfdisc(pStation, pSensorinfo,
                                              actSegment,
                                              pChannelHead->expSampleRate,
                                              lddate, dir, dfile,
                                              pWaveFormHead->sampleSize))
                        { return(E_SYNC_WFDISCS); }
                     }
                  }
                  else
                  {
                     if ((CD2W_MODE & CREATE_QUAL_RECORDS) != 0)
                     {
                        /* Write quality record now */
                        if (-1 == writeQualityRecord(pStation, pSensorinfo,
                                              actSegment,
                                              pChannelHead->expSampleRate,
                                              lddate,
                               pWaveFormHead->waveFileID[dayInWindow]))
                        { return(E_SYNC_WFDISCS); }
                     }
                  }
                  /* Either wfdisc or quality segment has been saved if
                     the cached flag was set. Reset cached flag */
                  actSegment->segState &= ~SEG_CACHED;
               }
            } /* end for all segments of one channel */

#undef CD2W_MODE

            if (NULL != segment2delete)
            {
               /* Free segment which is now part of a bigger segment */
                ST_DEBUG3(2, "Destroy segment %.17s %f",
                             pChannelCache->scName[0],
                             segment2delete->segStartTime);

               destroySegment(pSegHead, segment2delete);
               segment2delete = NULL;
            }

            if (WFDISC_PASS == pass)
            {
               pSegHead = &(pDay->qualSeg);
               ST_DEBUG2(2, "%d qual segments", (int)pSegHead->nbrOfSegments);
            }
         } /* end update wfdisc and quality records */

         /* No need to flush in DB mode */
         if (NULL != pStation->files.fpWfdiscFile)
         {
            /* flush buffer and check if an error has occurred */
            if (0 != fflush(pStation->files.fpWfdiscFile) ||
                0 != ferror(pStation->files.fpWfdiscFile))
            {
               /* In most cases this is a permission or file system
                  problem. CD2W is not able to save wfdisc records,
                  so a shutdown is initiated */
               shutdownApplication("can't write to wfdisc file",
                                   pStation->conID);
               clearerr(pStation->files.fpWfdiscFile);
               return(E_FILES);
            }
         }
         if (NULL != pStation->files.fpAppFile)
         {
            /* flush buffer and check if an error has occurred */
            if (0 != fflush(pStation->files.fpAppFile) ||
                0 != ferror(pStation->files.fpAppFile))
            {
               /* In most cases this is a permission or file system
                  problem. CD2W is not able to save quality records,
                  so a shutdown is initiated */
               shutdownApplication("can't write to quality file",
                                   pStation->conID);
               clearerr(pStation->files.fpAppFile);
               return(E_FILES);
            }
         }
      } /* end for all channels of a station */
   } /* end for all days (merge window) */

#ifdef HAVE_DB_ACCESS
   /* All wfdisc records for one channel and one day have been
      written/updated. Commit the changes now. Keep in mind that
      only one connection is used by all threads and an another thread
      could have commit the connection in the meantime. This is
      no problem as long as no thread tries to roll back the changes. */
   if (NULL != hdbc)
   {
      ST_DEBUG(2, "commit changes");
      /* Save number of commits in connectsOK, it's not used
         in cd2wng otherwise */
      pStation->whenConnected.connectsOK++;
      if (-1 == commitChanges(hdbc, (configParams_t *)pStation->pConfigData))
      {
          shutdownApplication("can't commit changes", pStation->conID);
          return(E_SYNC_WFDISCS);
      }
      ST_DEBUG2(4, "After commit for %lx", (unsigned long)hdbc);
   }
#endif /* HAVE_DB_ACCESS */

   /* Wfdiscs are written. It is save now to update the restart information */
   if (0 == (pCdSend->sndOpMode & OPER_MODE_AUX)
       && ONLY_STATISTICS != syncFlag)
   {
       /* Write a re-start point to disk */
       saveIntRefList(getSortLevel(pStation), pCd2wParams->refListStoreInterval,
                      pCdSend->pRefListHead);
#ifdef HAVE_DB_ACCESS
       /* Update frame product table */
       if (-1 == updateReceiverFrameProduct(hdbc,
                                    pCdSend->pRefListHead->pLastSent,
                                    pStation->type, 0))
       {
          /* Shutdown is started in updateFrameProduct() */
           return(E_SAVE_RECORD);
       }
#endif /* HAVE_DB_ACCESS */
   }

   return(NO_ERROR);
}

/**
 * @short              Log cd2w specific information
 *
 * @description        This function logs cd2w specific information to
 * @description        syslog. Normally this function is called from
 * @description        the monitor thread via a call back.
 *
 * @param pStation     station object
 *
 * @return -1 on failure, 0 otherwise
 */

int cd2wMonitor(station_t *pStation)
{
   return((int)syncWfdiscs(pStation, MAX_MERGE_DAYS, ONLY_STATISTICS));
}

/**
 * @short              Close day
 *
 * @description        This function releases all segments for one day
 * @description        for a given channel
 *
 * @param pSegHead     pointer to day header (wfdisc or quality)
 * @param forceFlag    do not complain about cached segments
 *
 * @return nothing
 */

static void
closeDay(segmentHead_t *pSegHead, int forceFlag)
{
   channelWaveSeg_t *actSegment,
                    *nextSeg;

   for (actSegment = pSegHead->firstSeg; actSegment != NULL; )
   {
      if (SEG_RELEASED == actSegment->segState)
      {
         shutdownApplication("Attempt to release released segment",
                             "closeDay");
         return;
      }
      if (0 == forceFlag)
      {
         /* Force flag is set when an error has occurred. It can
            happen that the segments are cached in this
            case, so do not complain about cached flag here */
         if (0 != (SEG_SCHEDULED4RELEASE & actSegment->segState))
         {
            shutdownApplication("Attempt to release scheduled4released segment",
                                "closeDay");
            return;
         }
         if (0 != (actSegment->segState & SEG_CACHED))
         {
            shutdownApplication("Attempt to release cached segment",
                                "closeDay");
            return;
         }
      }
      actSegment->segState = SEG_RELEASED;
      nextSeg = actSegment->nextSeg;
      (void)free(actSegment);
      pSegHead->nbrOfSegments--;
      actSegment = nextSeg;
   }
   pSegHead->firstSeg = NULL;
   /*@-onlytrans@*/
   pSegHead->lastSeg = pSegHead->firstSeg;
   /*@=onlytrans@*/
}

/**
 * @short              Log channel offsets for the whole window
 *
 * @description        This function logs channel offsets for the whole window
 *
 * @param pStation           station object
 * @param pWaveFormHead      waveform header
 *
 * @return 0 on success or -1 if an error has occured
 */

static int
logChannelOffsets(station_t *pStation, waveFormHead_t *pWaveFormHead)
{
   channelHead_t *pChannelHead;
   int ii;
   uint16_t ij;
   chName_t *pChannelCache;    /* Channel Cache */

   /* Shift the length of waveform files by one day */
   for (ii = 0; ii < MAX_MERGE_DAYS; ii++)
   {
       ST_DEBUG4(2, "day = %u lengthWaveFile = %d (%d)", (unsigned int)ii,
                 (int)pWaveFormHead->lengthWaveFile[ii],
                 (int)pWaveFormHead->headerLinesWritten[ii]);

   }

   pChannelCache = pStation->chCache.firstCh;

   /* Log channel offset for all channels */
   for (ij = 0; ij < pStation->chCache.nrOfChannels; ij++,
                               pChannelCache = pChannelCache->nextChannel)
   {
      /* Check index first */
      if (ij >= MAX_CHANNELS)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
               "Internal error: Bad channel cache index [%d], %.17s - %.15s",
               (int)ij, pStation->conID,
               NULL == pChannelCache ? "null" : pChannelCache->scName[0]);
         shutdownApplication(logBuf, "logChannelOffsets");
         return(-1);
      }

      if (NULL == pChannelCache)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
               "Internal error: Bad channel cache ptr, index [%d], %.17s",
               (int)ij, pStation->conID);
         shutdownApplication(logBuf, "logChannelOffsets");
         return(-1);
      }

      /* Get channel header */
      pChannelHead = pWaveFormHead->channelHeads[ij];

      if (NULL == pChannelHead)
      {
         /* This could be a filtered channel or the channel is not yet
            initialized */
         continue;
      }
      ST_DEBUG2(2,"Log channel %.18s", pChannelCache->scName[0]);

      for (ii = 0; ii < MAX_MERGE_DAYS; ii++)
      {
         ST_DEBUG3(2, "day = %u channelWavOffset = %d", (unsigned int)ii,
                (int)pChannelHead->day[ii].channelWavOffset);
      }
   }

   return(0);
}

/**
 * @short          Close merging window
 *
 * @description   This function closes the merging window. This means
 * @description   that the segments for all days and channels are synced
 * @description   (wfdiscs updated) and all segments are released.
 * @description   This function will also ensure that all channel
 * @description   headers are written to the waveform files
 *
 * @param pStation           station object
 * @param pWaveFormHead      waveform header
 * @param actEvent           pointer to event buffer
 * @param event              event which causes worker thread to stop
 *
 * @return 0 on success or -1 if an error has occurred
 */

static int
closeWindow(station_t *pStation, waveFormHead_t *pWaveFormHead,
            CDevent_t *actEvent, uint16_t event)
{
   channelHead_t *pChannelHead;
   int ii,
       forceFlag = 0;
   uint16_t ij;
   chName_t *pChannelCache;    /* Channel Cache */

   ST_DEBUG2(1, "closeWindow: start closing [%u]", (unsigned int)event);

   if (NO_ERROR == event || S_SHUTDOWN == event)
   {
      /* Sync the wfdisc records first. If an error has occurred do not
         sync the wfdiscs (and quality) records or it can happen
         that the quality and wfdisc records get out of sync if it was
         a problem with the ID generation  */
      ST_DEBUG(2, "close window sync");
      event = syncWfdiscs(pStation, MAX_MERGE_DAYS, SYNC_WFDISCS);
      if (NO_ERROR != event)
      {
         if (E_SYSTEM == event) pStation->savedSystemErr = (uint32_t)errno;
         logCDevent2(ET_L_ERROR, event, actEvent, (uint32_t)errno);
         return (-1);
      }
   }
   else
      forceFlag = 1;


   /* Close all channels for this station */
   pChannelCache = pStation->chCache.firstCh;
   for (ij = 0; ij < pStation->chCache.nrOfChannels; ij++,
                               pChannelCache = pChannelCache->nextChannel)
   {
      /* Check index first */
      if (ij >= MAX_CHANNELS)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
               "Internal error: Bad channel cache index [%d], %.17s - %.15s",
               (int)ij, pStation->conID,
               NULL == pChannelCache ? "null" : pChannelCache->scName[0]);
         shutdownApplication(logBuf, "closeWindow");
         return(-1);
      }

      if (NULL == pChannelCache)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
               "Internal error: Bad channel cache ptr, index [%d], %.17s",
               (int)ij, pStation->conID);
         shutdownApplication(logBuf, "closeWindow");
         return(-1);
      }

      /* Get channel header */
      pChannelHead = pWaveFormHead->channelHeads[ij];

      if (NULL == pChannelHead)
      {
         /* This could be a filtered channel or the channel is not yet
            initialized */
         continue;
      }

      /* Close all days in the window */
      for (ii = 0; ii < MAX_MERGE_DAYS; ii++)
      {
          if (0 != pChannelHead->day[ii].filledSlots)
          {
             ST_DEBUG3(2, "close day %d for [%.15s]", ii,
                       pChannelCache->scName[0]);
             closeDay(&(pChannelHead->day[ii].wfdiscSeg), forceFlag);
             closeDay(&(pChannelHead->day[ii].qualSeg), forceFlag);
          }
      }
   }

   return(0);
}

/**
 * @short              Load wfdisc and quality records
 *
 * @description        This function loads wfdisc and quality records
 * @description        fro a given day in the day window if configured.
 * @description        The records are loaded either from file or from database.
 * @description        The channel offsets and the waveform file size is
 * @description        loaded from the corresponding waveform file (if a
 * @description        wfdisc file exists for the requested day)
 *
 * @param pStation        station object
 * @param pWaveFormHead   waveform header
 * @param dayInWindow     day in window to load
 *
 * @return NO_ERROR on success or event which caused the error
 */

static uint16_t
loadRecords(station_t *pStation, waveFormHead_t *pWaveFormHead,
            uint32_t dayInWindow)
{
   senderParams_t *pCd2wParams;  /* Applications config extension */
   uint16_t event = NO_ERROR;
   uint32_t loadDate;


   /* set sender specific config parameters */
   pCd2wParams = (senderParams_t *)pStation->pConfigData->configExtension;
   loadDate = pWaveFormHead->startTimeWindow + (dayInWindow * ONE_DAY);

#define CD2W_MODE (pCd2wParams->senderMode)
#ifdef HAVE_DB_ACCESS

   /* Check if there is a DB handle available */
   if (NULL == getHdbc(pStation->pConfigData, DB_HANDLE))
   {
      if ((CD2W_MODE & CREATE_WFDISC_RECORDS) != 0)
      {
          event = loadWfdiscRecordsFromFile(pStation, loadDate, pWaveFormHead);
          if (NO_ERROR != event) return(event);
      }
      if ((CD2W_MODE & CREATE_QUAL_RECORDS) != 0)
      {
         event = loadQualityRecordsFromFile(pStation, loadDate, pWaveFormHead);
      }
   }
   else
   {
      if ((CD2W_MODE & CREATE_WFDISC_RECORDS) != 0)
      {
         event = loadWfdiscRecordsFromDB(pStation, loadDate, pWaveFormHead);
         if (NO_ERROR != event) return(event);
      }
      if ((CD2W_MODE & CREATE_QUAL_RECORDS) != 0)
      {
         event = loadQualityRecordsFromDB(pStation, loadDate, pWaveFormHead,
                                          dayInWindow);
      }
   }

#else  /* HAVE_DB_ACCESS */

   if ((CD2W_MODE & CREATE_WFDISC_RECORDS) != 0)
   {
      event = loadWfdiscRecordsFromFile(pStation, loadDate, pWaveFormHead);
      if (NO_ERROR != event) return(event);
   }
   if ((CD2W_MODE & CREATE_QUAL_RECORDS) != 0)
   {
      event = loadQualityRecordsFromFile(pStation, loadDate, pWaveFormHead);
   }

#endif /* do not HAVE_DB_ACCESS */
#undef CD2W_MODE

   return(event);
}

/**
 * @short              Shift for one day
 *
 * @description        This function shifts the day window for one day
 * @description        for all channels. Keep in mind that shifting the
 * @description        window only for one channel will not work, because
 * @description        that will cause a corrupted file size if all channels
 * @description        are stored in one file and not all channels are known
 * @description        yet
 *
 * @param pStation           station object
 * @param pWaveFormHead      waveform header
 *
 * @return NO_ERROR on success or event code if an error has occurred
 */

static uint16_t
shiftOneDay(station_t *pStation, waveFormHead_t *pWaveFormHead)
{
   channelHead_t *pChannelHead;
   int ii;
   uint16_t ij,
            event;
   chName_t *pChannelCache;    /* Channel Cache */

   /* Sync the wfdisc records first */
   ST_DEBUG(2, "shift window sync");
   event = syncWfdiscs(pStation, MAX_MERGE_DAYS, SYNC_WFDISCS);
   if (NO_ERROR != event)
   {
      return (event);
   }

   for (ii = 1; ii < MAX_MERGE_DAYS; ii++)
   {
       /* Shift the length of waveform files by one day */
       pWaveFormHead->lengthWaveFile[ii - 1] =
                                  pWaveFormHead->lengthWaveFile[ii];
       /* Shift header lines used in waveform files by one day */
       pWaveFormHead->headerLinesWritten[ii - 1] =
                                  pWaveFormHead->headerLinesWritten[ii];
       /* Shift waveform file IDs by one day */
       pWaveFormHead->waveFileID[ii - 1] = pWaveFormHead->waveFileID[ii];
       /* Shift current length waveform file by one day */
       pWaveFormHead->curLenWF[ii - 1] = pWaveFormHead->curLenWF[ii];
   }
   /* Reset length of newest waveform file */
   pWaveFormHead->lengthWaveFile[MAX_MERGE_DAYS - 1] =
                                 MAX_CHANNELS * LEN_WAVEFILE_HEAD_LINE;
   /* Reset header lines used of newest waveform file */
   pWaveFormHead->headerLinesWritten[MAX_MERGE_DAYS - 1] = 0;
   /* Reset waveform file ID */
   pWaveFormHead->waveFileID[MAX_MERGE_DAYS - 1] = -1L;
   /* Reset current length waveform file */
   pWaveFormHead->curLenWF[MAX_MERGE_DAYS - 1] = 0;

   /* Shift all channels values now by one day. It is necessary to shift all
      known channels now and not only the channels present in the current
      frame!! */
   pChannelCache = pStation->chCache.firstCh;
   for (ij = 0; ij < pStation->chCache.nrOfChannels; ij++,
                               pChannelCache = pChannelCache->nextChannel)
   {
      /* Check index first */
      if (ij >= MAX_CHANNELS)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
               "Internal error: Bad channel cache index [%d], %.17s - %.15s",
               (int)ij, pStation->conID,
               NULL == pChannelCache ? "null" : pChannelCache->scName[0]);
         shutdownApplication(logBuf, "shiftOneDay");
         return(E_INTERNAL);
      }

      if (NULL == pChannelCache)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
               "Internal error: Bad channel cache ptr, index [%d], %.17s",
               (int)ij, pStation->conID);
         shutdownApplication(logBuf, "shiftOneDay");
         return(E_INTERNAL);
      }

      /* Get channel header */
      pChannelHead = pWaveFormHead->channelHeads[ij];

      if (NULL == pChannelHead)
      {
         /* This could be a filtered channel or the channel is not yet
            initialized */
         continue;
      }

      /* Close the oldest day in the window */
      if (0 != pChannelHead->day[0].filledSlots)
      {
          ST_DEBUG2(2, "close day 0 for [%.15s]", pChannelCache->scName[0]);
          closeDay(&(pChannelHead->day[0].wfdiscSeg), 0);
          closeDay(&(pChannelHead->day[0].qualSeg), 0);
      }

      ST_DEBUG2(2,"Shift channel %.18s", pChannelCache->scName[0]);

      /* Shift for one day */
      for (ii = 1; ii < MAX_MERGE_DAYS; ii++)
      {
         /*@-usereleased@*/
         pChannelHead->day[ii - 1] =  pChannelHead->day[ii];
         /*@=usereleased@*/
         ST_DEBUG3(2, "day = %u channelWavOffset = %d", (unsigned int)ii - 1,
                (int)pChannelHead->day[ii - 1].channelWavOffset);
      }

      /* Clear last day */
      memset(&(pChannelHead->day[MAX_MERGE_DAYS - 1]), 0, sizeof(dayHeader_t));

      /* Don't set the channel offset here. If a previous non existing
         channel gets an offset here it is possible that these offset is
         already used in the waveform file. To be on the safe side it is
         necessary to try to load the wfdisc files for the new days first!! */
   }

   /* Set new window start time */
   pWaveFormHead->startTimeWindow += ONE_DAY;

   IF_DEBUG (2)
   {
      (void) fprintf(DEBUG_FP,
                   "(SD:%.2hu)[%.17s]"
                   "New start at: [%d] -> ",
                    pStation->debugLevel, pStation->conID,
                     (int)pWaveFormHead->startTimeWindow);
      double2ascDate((double)pWaveFormHead->startTimeWindow, DEBUG_FP);
      (void)fprintf(DEBUG_FP, "\n");
      DEBUG_FOOTER;
   }

   /* Try to load the wfdisc records for the new day. The channel offsets and
      the waveform file size is loaded from the corresponding waveform
      file (if a wfdisc file exists for the new day). Also try to
      load the quality records */
   event = loadRecords(pStation, pWaveFormHead, MAX_MERGE_DAYS - 1);

   /* Check if something unexpected has happened during loading of
      wfdisc or quality records */
   if (NO_ERROR != event)
   {
       return(event);
   }

   /* Now check if all channels have a waveform offset. Channels known
      but not present in the waveform file do not have a waveform offset */
   pChannelCache = pStation->chCache.firstCh;
   for (ij = 0; ij < pStation->chCache.nrOfChannels; ij++,
                               pChannelCache = pChannelCache->nextChannel)
   {
      /* Check index first */
      if (ij >= MAX_CHANNELS)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
             "Internal error: Bad channel cache index (2) [%d], %.17s - %.15s",
               (int)ij, pStation->conID,
               NULL == pChannelCache ? "null" : pChannelCache->scName[0]);
         shutdownApplication(logBuf, "shiftOneDay");
         return(E_INTERNAL);
      }

      if (NULL == pChannelCache)
      {
         char logBuf[96];

         (void)snprintf(logBuf, sizeof(logBuf),
               "Internal error: Bad channel cache ptr (2), index [%d], %.17s",
               (int)ij, pStation->conID);
         shutdownApplication(logBuf, "shiftOneDay");
         return(E_INTERNAL);
      }

      /* Get channel header */
      pChannelHead = pWaveFormHead->channelHeads[ij];

      if (NULL == pChannelHead)
      {
         /* This could be a filtered channel or the channel is not yet
            initialized */
         continue;
      }

      if (0 != pChannelHead->day[MAX_MERGE_DAYS - 1].channelWavOffset)
      {
         /* This channel has been initialized from the waveform file.
            It can happen that this channel is not received till the end
            of the day and the channel index has changed (i. e. the channel
            sequence is different compared to the time the when the
            channel was last seen). */
         /*
         if (-1 == syncFiles(pStation, NO, 0,
                             (time_t)pWaveFormHead->startTimeWindow +
                                     ((MAX_MERGE_DAYS - 1) * ONE_DAY)))
         {
            pStation->whenConnected.systemFailures++;
            return(E_INTERNAL);
         }

         if (NULL != pStation->files.fpUsrFile)
         {
            char siteName[LEN_SITE_NAME + 1];
            char channelName[LEN_CHANNEL_NAME + 1];

            splitSC(pChannelCache->scName[0], siteName, channelName);
            (void)writeChannelInfo(pStation,
                                   &(pChannelHead->day[MAX_MERGE_DAYS - 1]),
                                   pChannelHead, siteName, channelName,
                                   pWaveFormHead->sampleSize);
         }
         */
         continue;
      }

      /* Preset required values;
         This is a new file so set the channel offsets to the
         current maximum file size */
      pChannelHead->day[MAX_MERGE_DAYS - 1].channelWavOffset =
                        pWaveFormHead->lengthWaveFile[MAX_MERGE_DAYS - 1];
      /* Increase (computed) length of waveform file by size of channel data.
         Compute length of waveform file for ALL channels */
      pWaveFormHead->lengthWaveFile[MAX_MERGE_DAYS - 1] +=
                     GET_CHANNEL_SIZE(pChannelHead->expSampleRate,
                                      pChannelHead->nbrOfSlots,
                                      pWaveFormHead->sampleSize);
      /* Channel offset in older files might be different depending
         on the number of non merge-able segments for a channel per day */
      ST_DEBUG3(2,"Set channel offset %.18s: %d", pChannelCache->scName[0],
                (int)pChannelHead->day[MAX_MERGE_DAYS - 1].channelWavOffset);
   }

   ST_DEBUG2(2, "New waveform file length: %u",
             (unsigned int)pWaveFormHead->lengthWaveFile[MAX_MERGE_DAYS - 1]);

   return(NO_ERROR);
}

/**
 * @short              Check calibration value
 *
 * @description        This function checks if the calibration value
 * @description        is inside a given tolerance and raises a warning
 * @description        otherwise. It also replaces the calibration value and
 * @description        calibration period sent in the frame with their
 * @description        nominal counterparts from the instrument table
 *
 * @param pStation     station object
 * @param channel      data channel
 * @param sensorInfo   sensor and instrument values
 *
 * @return number of channels to send
 */

static void
setAndCheckCalib(station_t *pStation, CDdataChannel_t *channel,
                 sensor_t *sensorInfo)
{
   float diff;
   float calib = channel->calibFactor;

   /*@-realcompare@*/
   if (0.0 == sensorInfo->nomCalib)
   {
      shutdownApplication("Invalid calibration factor in encountered",
                          channel->siteName);
      return;
   }

   /* Use nominal values instead of values from the frame. Values
      from the frame are considered as "bad" by default */
   channel->calibFactor = sensorInfo->nomCalib;
   channel->calibPeriod = sensorInfo->nomCalibPer;

   /* Check if a calibration factor and period was set and if the
      calibration factor sent is the same as in the sensor info */
   if (0.0 == channel->calibFactor || -1 == channel->calibPeriod ||
       calib == sensorInfo->nomCalib)
   {
      return;
   }

   diff = (sensorInfo->nomCalib - calib)/sensorInfo->nomCalib;
   if (fabs(diff) > CALIB_TOLERANCE)
   {
      pStation->whenConnected.warnings++;

      /* calibration info deviates from nominal value more than 5% */
      if (pStation->systemConf & MORE_VERBOSE_LOGGING)
      {
         logCDevent(ET_W_CON, pStation, W_CALIB_INFO, NULL, 0);
      }
   }
   /*@=realcompare@*/
}

/**
 * @short              Add a clf record
 *
 * @description        This function adds a clf record to the clf table.
 * @description        It computes the channel mask required by squal,
 * @description        adds the revision from the stream and writes
 * @description        the current clf line to the clf table
 *
 * @param pStation      station object
 * @param DFinfo        data frame structure
 * @param hdbc          database handle
 * @param pCurrElement  current clf entry
 * @param clfType       type of frame to process
 * @param nbrOfChannels number of channels parsed
 * @param channelCnt    number of channels in the channame table (for stream)
 *
 * @return 0 on success or -1 on error
 */

#ifdef HAVE_DB_ACCESS

static int
addClfRecord(station_t *pStation, CDdata_t *DFinfo, void *hdbc,
             intRefListParams_t *pCurrElement, char clfType,
             uint32_t nbrOfChannels, uint16_t channelCnt)
{
   uint32_t ii;
   chName_t *pChCache;
#define LNG_CHANNEL_MASK  64  /* Length of the channel mask in DB. Change
                                 this to 100 if clf table is gone */
   char channelMask[LNG_CHANNEL_MASK + 1];
   char logBuf[96];
   int revision = 0;

   /* Set 'missing' as default value */
   memset(channelMask, (int)'0', LNG_CHANNEL_MASK);

   /* Loop through the parsed channels and set the channel mask
      present flag. Not parsed or invalid channels will be treated
      as 'not present' in the frame */
   for (ii = 0; ii < nbrOfChannels; ii++)
   {
      int posInChMask;
      CDdataChannel_t *channel;    /* Channel Data */

      channel = &(DFinfo->channel[ii]);
      posInChMask = channel->posInChMask;

      pChCache = getChannelEntryByIndex((uint16_t)posInChMask,
                                        &pStation->chCache);

      /* Check if the channel has a sensor info assigned */
      if (NULL == pChCache || NULL == pChCache->chSensorInfo)
      {
         (void)snprintf(logBuf, sizeof(logBuf),
               "Configuration error: No Sensor info [%d], %.17s - %.7s/%.9s",
                posInChMask, pStation->conID,
                channel->siteName, channel->channelName);
          /* Most likely is that the channel is not in the channame table */
          if (0 != (pStation->systemConf & STRICT_MODE_ENABLE))
          {
             shutdownApplication(logBuf, "addClfRecord");
             return(-1);
          }
          logWarning(logBuf, "addClfRecord");
          continue;
      }
      else if (pChCache->chSensorInfo->position >= LNG_CHANNEL_MASK)
      {
         /* Channel position exceeds the channel mask length */
         (void)snprintf(logBuf, sizeof(logBuf),
               "Configuration error: position too high [%d], %.17s - %.7s/%.9s",
                (int)pChCache->chSensorInfo->position, pStation->conID,
                channel->siteName, channel->channelName);
          if (0 != (pStation->systemConf & STRICT_MODE_ENABLE))
          {
             shutdownApplication(logBuf, "addClfRecord");
             return(-1);
          }
          logWarning(logBuf, "addClfRecord");
          continue;
      }
      channelMask[pChCache->chSensorInfo->position] = '1';
      revision = pChCache->chSensorInfo->revision;
   }

   /* Trim channel mask */
   if (channelCnt > LNG_CHANNEL_MASK)
   {
      (void)snprintf(logBuf, sizeof(logBuf),
            "Internal error: channelCnt too high [%d], %.17s", (int)channelCnt,
             pStation->conID);
      shutdownApplication(logBuf, "addClfRecord");
      return(-1);
   }
   channelMask[channelCnt] = '\0';

   /* Insert a new record into the DB */
   if (-1 == insertClfInDB(hdbc, pCurrElement, clfType,
                           DFinfo->frameTimeLength / 1000,
                           revision, channelMask))
   {
      return(-1);
   }
   return(0);
}
#endif /* HAVE_DB_ACCESS */

/**
 * @short              Insert segment
 *
 * @description        This function extracts a subset of channels from a
 * @description        data frame, according to the filter specified in the
 * @description        production line file and merges timely and valid
 * @description        channel data into a waveform segment
 *
 * @param pStation     station object
 * @param clfType      type of frame to process
 * @param framesize    size of frame in stations frame buffer
 * @param event        event code of merging process (out)
 * @param cd2wMode     cd2w configuration info
 * @param pCurrElement pointer to current element in processing queue
 * @param lookBackTime maximum delay for frames for being processed
 * @param sampleRateTolerance   sample rate tolerance in percent
 *
 * @return number of channels processed or -1
 */

static int
insertSegment(station_t *pStation, char clfType, uint32_t frameSize,
              /*@out@*/uint16_t *event, const uint32_t cd2wMode,
              intRefListParams_t *pCurrElement, time_t lookBackTime,
              uint16_t sampleRateTolerance)
{
   char  sendChannel[MAX_CHANNELS]; /* Marker for data to sent */
   uint32_t  ii = 0;
   CDdata_t *DFinfo = pStation->DFinfo;
   uint32_t nbrOfChannels = 0;
   uint32_t nbrOfFilteredChannels;
   char *buff = pStation->frameBuffer;
   cdsend_t *pCdSend;
   waveFormHead_t *pWaveFormHead;
   uint32_t currentDay;    /* (nominal time) day of data currently processed */
   time_t nowTime;


   /* Do some consistency checks */
   if (NULL == DFinfo || NULL == pStation->appExtension ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pRefListHead ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine)
   {
      shutdownApplication("Unexpected NULL pointer (insertSegment)",
                          pStation->conID);
      *event = E_INTERNAL;
      return (-1);
   }

   /* Access application extension data */
   memset(sendChannel, 0, sizeof(sendChannel)); /* for splint */
   pCdSend = (cdsend_t *)(pStation->appExtension);

   /* Set and check waveform header */
   pWaveFormHead = (waveFormHead_t *)pCdSend->addAppExtension;
   if (NULL == pWaveFormHead)
   {
      shutdownApplication("Unexpected NULL pointer (waveform header)",
                          pStation->conID);
      *event = E_INTERNAL;
      return(-1);
   }

   *event = NO_ERROR;

   /* Check if site/channel filters are set
      and if new filter request command is pending */
   (void) check4FilterProvided(pCdSend);

   /* Parse the frame now to see if a filter rule matches */
   nowTime = time(NULL);
   *event = parseFrameInBuffer(buff, pStation, clfType, frameSize, nowTime);
   if (NO_ERROR != *event)
   {
      return (-1);
   }

   /* Update statistic data. Use frames received values for CD2WNG */
   updateStatistics(&(pStation->framesReceived), (int)frameSize, NO, clfType);

   /* Check if the frame is valid. Don't apply the filter
      if a parsing error has occurred or cd2wng might crash!
      Skip frame instead */
   if (HAS_DF_ERRORS(pStation->checkPass))
   {
      char logBuf[128];

      (void)snprintf(logBuf, sizeof(logBuf),
                     "Skip frame with with fpc 0x%x %f",
                     (unsigned int)pStation->checkPass, DFinfo->nominalTime);
      logWarning(logBuf, pStation->conID);
      pStation->whenConnected.invalidFrames++;
      return(0);
   }

#ifdef HAVE_AUTHENTICATION
   /* Verify frame signature */
   if (0 != (pStation->systemConf & FRAME_AUTH))
   {
      senderParams_t *pCd2wParams;

      pCd2wParams = (senderParams_t *)pStation->pConfigData->configExtension;
      if ((pCd2wParams->senderMode & CREATE_QUAL_RECORDS) != 0)
      {
         CDtrailer_t trailer;
         parseTrailer(&buff[pStation->header.trailerOffset], &trailer);
         /* End of reading is required to check validity period of
            frame signing certificate. CD2WNG uses the frames nominal
            time for checking */
         pStation->endRoW.tv_sec = (time_t)pCurrElement->nominalTime;
         pStation->endRoW.tv_usec = 0;
         ii = (uint32_t)checkFrameAuthentication(pStation, pStation->keyCache,
                             pStation->pStationIP == NULL ?  NULL :
                             pStation->pStationIP->keyCacheSta,
                             (pStation->systemConf & STRICT_MODE_ENABLE),
                             buff, (int)frameSize, &trailer, event);
         if (2 == ii)
         {
            /* A warning has occurred during frame signature check */
            pStation->whenConnected.warnings++;
            logCDevent(ET_W_CON, pStation, *event, NULL, NO_INFO);
         }
         else if (1 == ii)
         {
            /* An error has occurred during frame signature check */
            if (E_OOPS == *event || E_NO_MEM == *event)
               pStation->whenConnected.systemFailures++;
            else
               pStation->whenConnected.authenticationFailures++;

            shutdownApplication("Internal error while verifying data",
                                 "checkFrameAuthentication failed");
            return(-1);
         }
      }
   }

   /* Check the signature of the channel data if
      channel authentication is switched on */
   if (0 != (pStation->systemConf & CHANNEL_AUTH))
   {
      senderParams_t *pCd2wParams;

      pCd2wParams = (senderParams_t *)pStation->pConfigData->configExtension;
      if ((pCd2wParams->senderMode & CREATE_QUAL_RECORDS) != 0)
      {
         /* Authenticate data only if quality records are enabled.
            Wfdiscs are not longer split if the authentication status
            of the CSFs are different. This information has been moved
            to the quality records */
         if (NO_ERROR != (*event = verifyChannelData(pStation)))
         {
            /* Shutdown is initiated in verifyChannelData() */
            pStation->whenConnected.systemFailures++;
            return(-1);
         }
      }
   }
#endif /* HAVE_AUTHENTICATION */

   /* We also process frames with warnings or errors, but only
      complete parsed channels */
   if (GOOD_FORMAT == clfType) nbrOfChannels = DFinfo->parsedDFFchannels;
   else nbrOfChannels = DFinfo->parsedChannels;

   ST_DEBUG2(3, "nbrOfChannels is: %d", (int)nbrOfChannels);

   /* The frame has been parsed. There is nothing more to do
      for DFFs */
   if (GOOD_DATA != clfType && BAD_DATA != clfType)
      return ((int)nbrOfChannels);

   /* Get filtered channel mask and number of filtered channels */
   nbrOfFilteredChannels = getFilteredChannelMask(pCdSend,
                                                  nbrOfChannels,
                                                  DFinfo,
                                                  sendChannel);

   /* Check if at least one channel passes the filter conditions */
   if (0 == nbrOfFilteredChannels)
   {
      /* Nothing to do */
      ST_DEBUG(3, "No channel left after filtering");
      pWaveFormHead->framesDropped++;
      return(0);
   }

   /* Check if the frame nominal time belongs to future time and
      skip frame if nominal time deviates too much */
   if ((time_t)DFinfo->nominalTime -
       (time_t)pStation->pConfigData->maxDeviation > nowTime)
   {
      char logBuf[128];

      (void)snprintf(logBuf, sizeof(logBuf),
            "Skip frame with with nominal time %f > now %lu + maxFutureTime",
                     DFinfo->nominalTime, (unsigned long int)nowTime);
      logWarning(logBuf, pStation->conID);
      pStation->whenConnected.invalidFrames++;
      return(0);
   }

   /* Compute start of day time for this data */
   currentDay = (uint32_t)DFinfo->nominalTime -
                ((uint32_t)DFinfo->nominalTime % 86400);

   if (0 == pWaveFormHead->startTimeWindow)
   {
      /* Compute initial start time of window. To store the waveform
         segments a moving window of MAX_MERGE_DAYS days
         is used. Each time new day is encountered, the window
         is moved for one day. To give late arrived data a chance
         being merged, the start day for the window is the day of
         the data being processed minus the time spawn of the window */
      pWaveFormHead->startTimeWindow = currentDay -
                                       (ONE_DAY * (MAX_MERGE_DAYS - 1));

      /* Now check if the start time of the window is not too long in
         the past */
      if (0 != lookBackTime)
      {
         uint32_t notBeforeTime = (uint32_t)nowTime;

         /* Window starts at midnight (00:00:00) */
         notBeforeTime -= (notBeforeTime % 86400);

         /* Start the waveform window not before
            current time - maxWfdiscMemoryDuration */
         notBeforeTime -= lookBackTime;

         /* It makes no sense to has a start time before the current
            one because this means that the window must be shifted before
            new arrived data is merged into it */
         if (notBeforeTime > pWaveFormHead->startTimeWindow)
            pWaveFormHead->startTimeWindow = notBeforeTime;
         else
            ST_DEBUG(2, "Not using lookBackTime");
      }

      IF_ST_DEBUG (2)
      {
         S_DEBUG_HDR;
         (void) fprintf(DEBUG_FP, "Start at: [%d] -> ",
                        (int)pWaveFormHead->startTimeWindow);
         double2ascDate((double)pWaveFormHead->startTimeWindow, DEBUG_FP);
         (void)fprintf(DEBUG_FP, "\n");
         DEBUG_FOOTER;
      }
      /* Set initial waveform file length */
      for (ii = 0; ii < MAX_MERGE_DAYS; ii++)
      {
          /* Set length of waveform files */
          pWaveFormHead->lengthWaveFile[ii] =
                                        MAX_CHANNELS * LEN_WAVEFILE_HEAD_LINE;
          /* Reset number of header lines used in waveform files */
          pWaveFormHead->headerLinesWritten[ii] = 0;
          pWaveFormHead->waveFileID[ii] = -1;

      }
      /* Try to load records. Initial waveform file lengths
         have to be set before for ALL channels!! */
      for (ii = 0; ii < MAX_MERGE_DAYS; ii++)
      {
         /* Try to load the wfdisc and quality records for the given day
            in the day window. The channel offsets and the waveform file
            size is loaded from the corresponding waveform file (if a
            wfdisc file exists for the new day) */
         *event = loadRecords(pStation, pWaveFormHead, ii);

         /* Check if something unexpected has happened during loading of
            wfdisc or quality records */
         if (NO_ERROR != *event)
         {
            return(-1);
         }
      }
      IF_ST_DEBUG (2)
      {
         ST_DEBUG(2, "After loadRecords:");
         (void)logChannelOffsets(pStation, pWaveFormHead);
      }
   }

   /* Check if the channel data belongs to the current days
      waveform slots */
   if ((uint32_t)DFinfo->nominalTime < pWaveFormHead->startTimeWindow)
   {
      char logBuf[96];

      /* TODO: log this warning only by enableSyslog = MORE */
      (void)snprintf(logBuf, sizeof(logBuf), "Skip too old frame %d",
                     (int)DFinfo->nominalTime);
      logWarning(logBuf, pStation->conID);
      ST_DEBUG3(2, "Obsolete data %f < %d", DFinfo->nominalTime,
                (int)pWaveFormHead->startTimeWindow);
      pWaveFormHead->obsoleteFrames++;

      return(0);
   }

#ifdef HAVE_DB_ACCESS
   /* Check if cd2wng should create clf records in the database.
      This function is DEPRECATED and should be removed after the
      squal re-design!! */
   if (0 != (cd2wMode & CREATE_CLF_RECORDS) &&
       0 != pWaveFormHead->channelCnt)
   {
      void *hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);

      /* UGLY: Keep in mind that another thread can also try to
         update the clf table when two or more production lines
         process the same stream when a filter is used. A clf
         record is only written if the frame is merged into
         a waveform segment. No clf record is written for
         invalid data or data which does not match the
         filter criteria */
      if (NULL != hdbc && NULL != pStation->DFinfo)
      {
         if (-1 == addClfRecord(pStation, DFinfo, hdbc, pCurrElement,
                                clfType, nbrOfChannels,
                                pWaveFormHead->channelCnt))
         {
            *event = E_SAVE_RECORD;
            return(-1);
         }
      }
   }
#endif /* HAVE_DB_ACCESS */

   /* At least one channel matches the filter. Continue and process
      filtered sub channels */
   for (ii = 0; ii < nbrOfChannels; ii++)
   {
       if (SEND_CH == sendChannel[ii])
       {
          int posInChMask,
              ij;
          unsigned int sampleRate;
          uint32_t channelTime = 1000,
                       slot;
          channelHead_t *pChannelHead;
          CDdataChannel_t *channel;    /* Channel Data */
          int retVal;
          sensor_t *sensorInfo;

          channel = &(DFinfo->channel[ii]);
          posInChMask = channel->posInChMask;

          /* Check index first. Frames with invalid channel names are
             skipped before. If the complete frame should not be skipped
             in this case, either do not reset WA_INVALID_NAME and check for
             invalid channel name before check for posInChMask or skip
             channel if posInChMask is -1 */
          if (-1 == posInChMask || posInChMask >= MAX_CHANNELS)
          {
             char logBuf[96];

             (void)snprintf(logBuf, sizeof(logBuf),
                   "Internal error: Bad channel index [%d], %.17s - %.7s/%.9s",
                    posInChMask, pStation->conID,
                    channel->siteName, channel->channelName);
              shutdownApplication(logBuf, "insertSegment");
              *event = E_INTERNAL;
              return(-1);
          }

          /* Run some additional checks to avoid invalid segments */
          if (((channel->checkPass & (WA_SUB_TIME_STAMP | WA_COMPRESSSW |
                         AWA_STEIM_COMPRESS | WA_SAMPLES_CAN )) != 0) ||
              ((channel->checkPass & WA_NOT_IN_SAMPLE) != 0 &&
               /* Check for WA_NOT_IN_SAMPLE alone is too restrictive and will
                  cause skipping too many hydro station channels */
                        fabs(channel->timeStamp - DFinfo->nominalTime) > 1.0))
          {
             char logBuf[128];

            (void)snprintf(logBuf, sizeof(logBuf),
                           "Skip channel %.6s/%.8s with with cpc 0x%x %f %f",
                           channel->siteName, channel->channelName,
             (unsigned int)channel->checkPass, DFinfo->nominalTime,
                           channel->timeStamp);
             logWarning(logBuf, pStation->conID);
             pStation->whenConnected.warnings++;
             continue;
          }

          /* Check if this is a valid channel */
          if (0 == channel->nbrOfSamples)
          {
             /* This should not happen because only valid frames should
                be merged (check for HAS_DF_ERRORS is done before) */
              shutdownApplication("Internal error: bad sample number",
                                  "insertSegment");
              *event = E_INTERNAL;
              return(-1);
          }

          /* Set channel time span */
          if (channel->subFrameTimeLng != 0)
          {
             channelTime = channel->subFrameTimeLng;
          }
          else if (DFinfo->frameTimeLength != 0)
          {
              channelTime = DFinfo->frameTimeLength;
          }

          /* Run check to avoid FP exception when the frame is
             shorter than one second */
          if (channelTime < (uint32_t)1000)
          {
             char logBuf[128];

            (void)snprintf(logBuf, sizeof(logBuf),
                  "Skip channel %.6s/%.8s with with channel time < 1000 %d",
                  channel->siteName, channel->channelName, (int)channelTime);
             logWarning(logBuf, pStation->conID);
             pStation->whenConnected.warnings++;
             continue;
          }

          /* Compute (expected) sample rate. Do not confuse this
             sample rate with the real sample rate which is computed
             based on real channel start and end times of more than
             one frame. This value must be an integer to compute
             the correct channel size and this requirement is fulfilled
             if only one frame is considered */
          sampleRate = (unsigned int)
                       ((channel->nbrOfSamples * 1000)/channelTime);
          /* If the computed sample rate is not an integer value use the
             following formula:
             round(channel->nbrOfSamples/(double)(channelTime/1000.0)))
           */

          pChannelHead = pWaveFormHead->channelHeads[posInChMask];

          if (NULL != pChannelHead)
          {
             sensorInfo = pChannelHead->sensorInfo;
          }
          else
          {
             /* Sensor information is normally set in createChannelHeader()
                or loadRecords(), but no channel should be created if the
                sample rate deviates too much */
             sensorInfo = getSensorDescr(channel->siteName,
                                         channel->channelName);
          }

          /* Check if the expected sample rate does not deviate too much
             from the value found in the sensor information. Data should
             only be merged if the sample rate is within the defined range,
             because most of the processing applications use the values
             from the static tables for their computations. Considering
             only one frame will detect only inconsistencies between the
             static tables and the expected sample rate, because considering
             only one frame will always show the expected sample rate and not
             the real sample rate */
          if (NULL != sensorInfo && 0 != sampleRateTolerance)
          {
             float fSr = ((float)(channel->nbrOfSamples) /
                         ((float)channelTime / 1000));
             float fSrDiff,
                   fSrMaxDiff;

             fSrDiff = fSr - sensorInfo->expSampleRateInDB;
             fSrMaxDiff = sensorInfo->expSampleRateInDB / 100 *
                          (float)sampleRateTolerance;

             if (fabs(fSrDiff) > fSrMaxDiff)
             {
                char logBuf[128];

                (void)snprintf(logBuf, sizeof(logBuf),
                   "Sample rate differs to DB (%f/%f), skip %.6s/%.8s",
                               fSr,
                               sensorInfo->expSampleRateInDB,
                               channel->siteName, channel->channelName);
                logWarning(logBuf, pStation->conID);
                pStation->whenConnected.warnings++;
                continue;
             }
          }

          /* Comment out next block if channels without channel
             info should be processed in database mode. It is
             recommended to use a 'special' channel ID for such channels. */
          if (NULL == sensorInfo &&
              NULL != getHdbc(pStation->pConfigData, DB_HANDLE))
          {
             char logBuf[128];

             (void)snprintf(logBuf, sizeof(logBuf),
                "No channel info available for %.6s/%.8s, skip channel",
                            channel->siteName, channel->channelName);
             logWarning(logBuf, pStation->conID);
             pStation->whenConnected.warnings++;
             continue;
          }

          /* Check if the channel header exists, otherwise create one */
          if (NULL == pChannelHead)
          {
             ST_DEBUG4(2, "Nbr of samples %d, channel time = %dms, index = %d",
                       (int)channel->nbrOfSamples, (int)channelTime,
                       (int)posInChMask);

             ST_DEBUG4(2, "Allocate channel header for %.6s/%.8s, NT: %f",
                       channel->siteName, channel->channelName,
                       DFinfo->nominalTime);

             pChannelHead = createChannelHeader(pStation, channelTime,
                                                sampleRate,
                                                pWaveFormHead->lengthWaveFile,
                                                channel->siteName,
                                                channel->channelName,
                                                posInChMask,
                                                pWaveFormHead->sampleSize);
             if (NULL == pChannelHead)
             {
                /* shutdown is initiated in createChannelHeader() */
                *event = E_NO_MEM;
                return(-1);
             }

             IF_ST_DEBUG (2)
             {
                ST_DEBUG(2, "After createChannelHeader:");
                (void)logChannelOffsets(pStation, pWaveFormHead);
             }

             /* Save channel header in waveform header for further use */
             pWaveFormHead->channelHeads[posInChMask] = pChannelHead;
          }
          else
          {
             /* Check if the channel time has been set, which is not the
                case when the channel has been created during loading
                of wfdisc records */
             if (NO_DURATION == pChannelHead->expChannelTimeLng)
             {
                 /* Get the nearest valid channel time length */
                 pChannelHead->expChannelTimeLng =
                      CHECK_INT(channelTime / 1000, validChannelTimeLengths);
             }
          }

          /* Check if the channel data is newer than the current window */
          while ((uint32_t)DFinfo->nominalTime >=
                          (pWaveFormHead->startTimeWindow + MAX_WINDOW))
          {
             ST_DEBUG3(2, "Shift for one day %f >= %d", DFinfo->nominalTime,
                       (int)pWaveFormHead->startTimeWindow + MAX_WINDOW);

             /* Shift for one day */
             *event = shiftOneDay(pStation, pWaveFormHead);
             if (NO_ERROR != *event)
             {
                return(-1);
             }

             /* Check for a pending shutdown here!! */
             if (pStation->shutdownEnable != SD_NONE)
             {
                 *event = S_SHUTDOWN;
                 return((int)ii);
             }
          }

          /* Check if the expected sample rate is the same. Waveform
             segments can only be merged if the expected sample rate
             does not change, because the size of a channel is pre-computed
             and only one file is created per day. If a channel changes its
             sample rate, make sure that no restart records are loaded for
             this station (sample rate is loaded from the waveform file)
             with the (old) sample rate. Otherwise this channel
             will not get merged!! */
          if ((uint32_t)sampleRate != pChannelHead->expSampleRate)
          {
             char logBuf[128];

             (void)snprintf(logBuf, sizeof(logBuf),
                   "Sample rate differs (%d/%d), skip %.6s/%.8s",
                            (int)sampleRate,
                            (int)pChannelHead->expSampleRate,
                            channel->siteName, channel->channelName);
             logWarning(logBuf, pStation->conID);
             continue;
          }

          /* To compute the slot the frame nominal time is used
             instead of the channel nominal time, because the channel
             nominal time will drift caused by inaccurate sample rates */
          slot = ((uint32_t)DFinfo->nominalTime -
                                    pWaveFormHead->startTimeWindow);

          for (ij = 0; ij < MAX_MERGE_DAYS; ij++)
          {

             if (slot >= pChannelHead->nbrOfSlots)
             {
                slot -= pChannelHead->nbrOfSlots;
                continue;
             }
             break;
          }
          if (slot >= pChannelHead->nbrOfSlots)
          {
             shutdownApplication("Internal error: bad slot", "insertSegment");
             *event = E_INTERNAL;
             return(-1);
          }

          /* Check if the channel has an assigned nominal calibration
             information and check if the value in the frame deviates
             a given percentage from the nominal value */
          if (NULL != pChannelHead->sensorInfo)
          {
             setAndCheckCalib(pStation, channel, pChannelHead->sensorInfo);
          }

          retVal = tryToMerge(channel, DFinfo->nominalTime,
                              channelTime, pChannelHead, slot,
                              pStation, pWaveFormHead, currentDay, ij, 0,
                              '\0', 0.0, -1L, pCurrElement->clfTime, event);
          ST_DEBUG2(2, "tryToMerge returns %d", retVal);

          if (NO_ERROR != *event) return(-1);

          /* Do not check for a pending shutdown while
             not all channels are processed or you will
             end up with missing channels when the restart information
             is written. Keep in mind that the restart information
             include frames, not channels */
          /*
          if (pStation->shutdownEnable != SD_NONE)
          {
             *event = S_SHUTDOWN;
             return((int)ii);
          }
          */
       }
   }

   return((int)nbrOfFilteredChannels);
}

/**
 * @short Check if a data request is pending
 *
 * @description This function checks if a data request is pending and
 * @description modifies the Internal Reference List do serve the data
 * @description request
 *
 * @param pStation             station object
 * @param pRefListHead         pointer to internal reference list header
 * @param pCurrElement         address of pointer to current element in IRL
 *
 * @return NO_ERROR or event
 */

static void
check4dataRequest(station_t *pStation, intRefListHead_t *pRefListHead,
                  intRefListParams_t **pCurrElement)
{
   intRefListParams_t *pActFrame;
   cdsend_t *pCdSend;  /* Application extension object */
   int framesFound = 0;
   uint32_t requestEndTime;
   /* TODO: move to libCD */


   /* Check for empty queue */
   if (NULL == *pCurrElement) return;

   if (NULL == pStation->appExtension)
   {
      /* This is an internal error */
      shutdownApplication("Internal error in check4dataRequest",
                          pStation->conID);
      /* abort(); */
      return;
   }

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   /* Check if a data request is currently being processed */
   if (NOT_SENT_REQUESTED == (*pCurrElement)->sendState) return;

   if (0 == pCdSend->requestStartTime) return;

   /* Data request pending. Copy frames which should be sent in this
      request at top of the IRL and sort them like requested */
   requestEndTime = pCdSend->requestStartTime + pCdSend->requestDuration;
   pActFrame = pRefListHead->pRefListStart;
   while (NULL != pActFrame)
   {
      intRefListParams_t *pFoundFrame;
      int rval;

      if (NOT_SENT != pActFrame->sendState
          ||
          (uint32_t)pActFrame->nominalTime < pCdSend->requestStartTime
          ||
          (uint32_t)pActFrame->nominalTime > requestEndTime)
      {
         /* ignore this frame */
         pActFrame = pActFrame->pNextElement;
         continue;
      }

      if (NULL == pActFrame) break; /* only for splint */

      /* Frame found */
      pFoundFrame = pActFrame;
      /* Set next element in IRL before moving to top of IRL!! */
      pActFrame = pActFrame->pNextElement;
      /* Move found frame to top of IRL */
      rval = moveFrame2top(pRefListHead, pFoundFrame);
      pFoundFrame->sendState = NOT_SENT_REQUESTED;
      framesFound++;
      /* Set next frame to send */
      /*@-onlytrans@*/
      *pCurrElement = pFoundFrame;
      /*@=onlytrans@*/

      if (0 != rval) RH_DEBUG2(2, "moveFrame2top returns %d", rval);
   }

   if (0 == framesFound)
   {
      pCdSend->requestStartTime = 0;
      pCdSend->requestDuration = 0;
      return;
   }

   /* TODO: Check if we should set an extra restart point */

   /* TODO: re-sort found frames */

   return;
}

/**
 * @short Parse all clf files available
 *
 * @description This functions parse all clf files available
 * @description and adds their contents unconditionally
 * @description (e.g. without checking for restart points)
 *
 * @param pStation      station object
 * @param pRefListHead  intRefList header
 * @param pClfFile      reference to pointer to current clf file
 * @param sortMode      requested sorting mode (LIFO, FIFO, UNSORTED)
 *
 * @return NO_ERROR or an error event
 */

static uint16_t
parseAllClfFiles(station_t *pStation, intRefListHead_t *pRefListHead,
                 clfFileParams_t **pClfFile, int sortMode)
{
   cdsend_t *pCdSend;  /* Application extension object */
#ifdef HAVE_DB_ACCESS
   void *hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);
#endif
   /* TODO: move to libCD if binary file must not longer be written
            to the frameproduct table */

   if (NULL == pClfFile)
   {
      /* Should not happen */
      return(E_INTERNAL);
   }

   if (NULL == pStation->appExtension)
   {
      /* Should not happen */
      logError("Internal Error in parseAllClfFiles", "Unexpected NULL pointer");
      return (E_INTERNAL);
   }

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   while (NULL != *pClfFile)
   {
      uint16_t event = NO_ERROR;
      int nbrOfNewEntries;

      pStation->type = (*pClfFile)->cdtype;
      nbrOfNewEntries = parseClfFile((*pClfFile)->pFile, pStation->type,
                                     pRefListHead, 0.0, 0,
                                     NEW_FILE, sortMode);

      RH_DEBUG3(2, "parseAllClfFiles: %d new entries in %s", nbrOfNewEntries,
                (*pClfFile)->pFile);

      if (-1 == nbrOfNewEntries)
      {
         RH_DEBUG3(2, "Error while parsing file %s.%s", (*pClfFile)->pFile,
                   EXTENSION_LST);
      }

      /* Don't unlink the last clf file in the list!! It could grow */
      if (NULL != (*pClfFile)->pNextClfFileParams)
      {
         /* Close current clf file and check if clf file object
            should be removed */
         RH_DEBUG2(2, "Parsing of [%s] finished", (*pClfFile)->pFile);
         RH_DEBUG2(2, "intRefList has %u entries",
                   (unsigned int)pRefListHead->nbrOfEntries);
         checkClfFileObj(pRefListHead);

         /* don't unlink the clf entry in aux mode */
         if (pCdSend->sndOpMode & OPER_MODE_AUX)
         {
            RH_DEBUG2(2, "aux mode; don't unlink clf entry [%s]",
                         (*pClfFile)->pFile);

            /* get next element */
            /*@-onlytrans@*/
            *pClfFile = (*pClfFile)->pNextClfFileParams;
            /*@=onlytrans@*/

            continue;
         }

#ifdef HAVE_DB_ACCESS
         /* Create or update an entry for the binary file in the
            frameproduct table. This is necessary to write entries
            to the clf table later and for the archiver. Doing
            this if another clf file is available is the preferred
            method. Fid could have been created before when the binary
            file was read. Keep in mind that the complete binary file
            size could be stored before the file is processed (if the
            station is catching up). That might have some impact on
            archiving. Keep also in mind that, if there is a long outage,
            the binary file might have been archived in the meantime.
            Strictly speaking, there is no clean approach to write the
            file size of the binary file to the frameproduct table. There
            is always the chance that non data frames have been received
            short before the day changes and these types of frames are not
            processed by cd2wng! */

         if (NULL != hdbc)
         {
            long fid;
            double clftime = 0.0;
            char fileName[LIBPAR_BUFSIZ];  /* file name on bin file */
            int flag;

            /* Implicit default is CD1 */
            (void)snprintf(fileName, sizeof(fileName), "%s.%s",
                           (*pClfFile)->pFile,
                           CD1 == (*pClfFile)->cdtype ?
                               EXTENSION_BIN_CD1 : EXTENSION_BIN_CD11);


            if (NULL != pRefListHead->pRefListEnd)
               clftime =  pRefListHead->pRefListEnd->clfTime;

            ST_DEBUG(2, "Checking frame product (2)");
            /* The receive time is only correct if the IRL is not
               sorted (mode = AS_ARRIVED)!! */
            fid = getFrameProduct(hdbc, pStation, fileName,
                                  CD1 == (*pClfFile)->cdtype ?
                                      CD_B10_FILE : CD_B11_FILE,
                                  IGNORE_NONEXIST, clftime, &flag);

            /* Check if an ID is available */
            if (-1L == fid)
            {
               return(E_SAVE_RECORD);
            }

            /* Check if the file was already known. In this case
               the file size has to be updated */
            /*@-compdestroy@*/
            /*@-mustfreefresh@*/
            if (0 == flag)
            {
               struct stat statBuf;

               memset(&statBuf, 0, sizeof(struct stat)); /* for splint */
               if (0 != stat(fileName, &statBuf))
               {
                  /* File could have been moved or removed if there was a
                     long outage  Older versions of cd2wng have stopped here,
                     but file size is corrected by updateFrameProduct.pl
                     now. Old comment also stated this might be caused by
                     a CD protocol type change, but this should
                     not have an impact here. */
                     logWarning("parseAllClfFiles: can't stat binary file",
                                 fileName);
               }
               else
               {
                  if (-1L == updateFrameProduct(hdbc, fid,
                                                (uint32_t)statBuf.st_size,
                                                clftime))
                  {
                     return(E_SAVE_RECORD);
                  }
               }
            }
            /*@=compdestroy@*/
            /*@=mustfreefresh@*/
         }
#endif /* HAVE_DB_ACCESS */

         /* get next element and unlink and destroy current element */
         /*@-dependenttrans@*/
         *pClfFile = unlinkClfElement(pStation, *pClfFile, &event);
         /*@=dependenttrans@*/
         if (NO_ERROR != event) return(event);
      }
      else break;
   }
   return(NO_ERROR);
}

/**
 * @short Process content of internal reference list
 *
 * This function processes the content of the internal
 * reference list and deletes the processed entries from
 * internal reference list. Errors related to invalid frames
 * are treated in this version
 *
 * @param pRefListHead         pointer to internal reference list
 * @param pStation             station object
 * @param actEvent             event object
 * @param pCurrentClf          reference to pointer to clf file list
 *
 * @return NO_ERROR or event
 */

static uint16_t
processIrlData(intRefListHead_t *pRefListHead, station_t *pStation,
               CDevent_t *actEvent, clfFileParams_t **pCurrentClf)
{
   FILE *fp = NULL;
   intRefListParams_t *pCurrElement; /* current frame */
   char logBuffer[LOGSIZE],
        fileName[LIBPAR_BUFSIZ];  /* file name on bin file */
   uint16_t event = NO_ERROR;
   int sendDFF = 0,    /* Indication that DFF has been sent */
       sortType = getSortLevel(pStation);
   senderParams_t *pCd2wParams;
   cdsend_t *pCdSend;  /* Application extension object */
#ifdef HAVE_DB_ACCESS
   void *hdbc = getHdbc(pStation->pConfigData, DB_HANDLE);
#endif


   /* Do some checks */
   if (NULL == pRefListHead || NULL == pStation->frameBuffer
       ||
       NULL == pStation->appExtension)
   {
      logError("Internal Error in processIrlData", "Unexpected NULL pointer");
      return (E_OOPS);
   }

   /* Don't forget to make sure that the IRL is sorted according to the
      current processing mode (LIFO, FIFO, AS_ARRIVED) */

   memset(fileName, 0, sizeof(fileName)); /* For splint */
   /* set sender specific config parameters */
   pCd2wParams = (senderParams_t *)pStation->pConfigData->configExtension;

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   RH_DEBUG2(5, "Value of nbrOfEntries [%u]",
             (unsigned int)pRefListHead->nbrOfEntries);

   /* Process all frames in the intReflist which has
      state equal NOT_SEND */
   pCurrElement = pRefListHead->pRefListStart;
   while (NULL != pCurrElement)
   {
      uint32_t frameLength,
               offsetBinFile,
               frameSize;
      int retries;
      char clfType;
      int doSync;

      if (0 == sendDFF)
      {
         /* Check if a data request is pending. Only frames seen in
            the current IRL are considered. This is save, because the
            IRL contains all unprocessed frames currently known */
         check4dataRequest(pStation, pRefListHead, &pCurrElement);

         if (NULL == pCurrElement) break; /* For splint only */
      }

      if (pCurrElement->sendState != NOT_SENT &&
          pCurrElement->sendState != NOT_SENT_REQUESTED)
      {
         /* Skip already processed element */
         pCurrElement = pCurrElement->pNextElement;
         continue;
      }

      if (NULL == pCurrElement) break; /* For splint only */

      /* Process only requested data in auxiliary station mode. Requested
         data is found on top of the IRL, so if data is seen which is not
         requested, all requested data have been processed */
      if (0 != (pCdSend->sndOpMode & OPER_MODE_AUX)
         &&
         pCurrElement->sendState != NOT_SENT_REQUESTED)
      {
         break;
      }

#define P_LAST_ELEMENT_SENT  pRefListHead->pLastSent

      /* NOTE: if cd2wng/sender should handle station type changes
               correct it needs to know the CD type of the current
               frame. This could be achieved by storing the CD type in
               the file object setting the CD type before each frame
               is processed. But in this case the statistics might be
               incorrect */
      /* Check if a DFF has to be parsed */
      if (CD1 == pStation->type && sendDFF == 0 &&
          (NULL == P_LAST_ELEMENT_SENT ||
           P_LAST_ELEMENT_SENT->DFFoffset != pCurrElement->DFFoffset ||
           P_LAST_ELEMENT_SENT->pFileObj != pCurrElement->pFileObj))
      {
         /* Process DFF */
         if (NULL == P_LAST_ELEMENT_SENT)
         {
            RH_DEBUG(2, "load DFF for the first frame");
         }
         else if (P_LAST_ELEMENT_SENT->DFFoffset != pCurrElement->DFFoffset)
         {
            RH_DEBUG(2, "load DFF because different DFFoffset");
            RH_DEBUG3(2, "LastSentDFFoffset= %ld, CurrentDFFoffset=%ld",
                      (long)P_LAST_ELEMENT_SENT->DFFoffset,
                      (long)pCurrElement->DFFoffset);
         }
         else if (P_LAST_ELEMENT_SENT->pFileObj != pCurrElement->pFileObj)
         {
            RH_DEBUG(2, "load DFF because different file pointer");
         }

         offsetBinFile = pCurrElement->DFFoffset;
         frameSize = pCurrElement->DFFsize;
         clfType = GOOD_FORMAT;
         sendDFF = 1;
      }
      else
      {
         /* Process data frame */
         offsetBinFile = pCurrElement->offsetBinFile;
         frameSize = pCurrElement->frameSize;
         clfType = pCurrElement->frameType;
         sendDFF = 0;
      }

      if (NULL == fp
          ||
          P_LAST_ELEMENT_SENT == NULL /* This will cause an extra open */
          ||
          P_LAST_ELEMENT_SENT->pFileObj != pCurrElement->pFileObj)
      {
         /* File name has changed. Close last file and open new one */
         if (NULL != fp)
         {
            /* Keep in mind that P_LAST_ELEMENT_SENT->pFileObj is NOT
               updated after a DFF has been read. This causes an extra
               open if a file switch has occurred and station type is
               CD-1.0 because P_LAST_ELEMENT_SENT->pFileObj and
               pCurrElement->pFileObj are different! Keep also in mind
               that it happens that the next clf file is not created yet, but
               there are no more data frames in the current clf file. In
               this case the binary file pointer is closed and when an
               Alert or AckNack Frame is received not followed by a Data
               Frame, the size of the binary file can't be determined correct
               here!  */

            (void)fclose(fp);
         }

         /* Implicit default is CD1 */
         if (CD1 == pStation->type)
         {
            (void) snprintf(fileName, sizeof(fileName), "%s.%s",
                            pCurrElement->pFileObj->fileNameAndPath,
                            EXTENSION_BIN_CD1);
         }
         else
         {
            (void) snprintf(fileName, sizeof(fileName), "%s.%s",
                            pCurrElement->pFileObj->fileNameAndPath,
                            EXTENSION_BIN_CD11);
         }

         RH_DEBUG2(2, "+++ Open binary file [%s]", fileName);

         /* Open file to sent */
         fp = fopen(fileName, "rb");
         if (NULL == fp)
         {
            /* Station might have changed the protocol type. See NOTE above.
               CD type is set correct again if a new clf file has been parsed.
             */
            (void) snprintf(logBuffer, sizeof(logBuffer),
                            "[%.17s] can't open bin file", pStation->conID);
            logError(logBuffer, fileName);

            /* Remove related frames from the internal reference list */
            pCurrElement = removeFrameWith(pRefListHead, pCurrElement,
                                           THAT_FILE_OBJECT);
            /* This is not treated as an error in the calling function */
            continue;
         }

#ifdef HAVE_DB_ACCESS
         if (NULL != hdbc)
         {
            /* Check if the frameproduct table has information about
               the new opened binary file, otherwise create a binary file
               in the frameproduct table. fid is used for clf table,
               too */
            if (-1 == pCurrElement->pFileObj->fid)
            {
               SQLHDBC idHdbc;

               idHdbc = (SQLHDBC)getHdbc(pStation->pConfigData, ID_HANDLE);
               if (NULL == idHdbc) /* for splint */
               {
                  (void) fclose(fp);
                  return(E_SAVE_RECORD);
               }
               ST_DEBUG(2, "Checking frame product");
               /* The receive time is only correct if the IRL is not
                  sorted (mode = AS_ARRIVED)!! */
               pCurrElement->pFileObj->fid = getFrameProduct(hdbc,
                                                             pStation,
                                                             fileName,
                             CD1 == pStation->type ? CD_B10_FILE : CD_B11_FILE,
                                                             CREATE_ID,
                                                  pCurrElement->clfTime, NULL);
               /* Check if an ID is available */
               if (-1L == pCurrElement->pFileObj->fid)
               {
                  (void) fclose(fp);
                  return(E_SAVE_RECORD);
               }
               /*
                  Keep in mind that at txt file entry has been created
                  in checkFrameProduct too, but no one is created in
                  getFrameProduct
               (void)strncpy(logBuffer, fileName, sizeof(logBuffer));
               replaceExtension(logBuffer, "txt", sizeof(logBuffer));
               if (-1L = getFrameProduct(hdbc, pStation, logBuffer,
                                         CD_TXT_FILE, IGNORE_NONEXIST,
                                         pCurrElement->clfTime);
               {
                  (void) fclose(fp);
                  return(E_SAVE_RECORD);
               }
               */
            }
         }
#endif /* HAVE_DB_ACCESS */
      }

      /* Seek to the current position; optimization is done
         by fseek() */
      if (0 != fseek(fp, (long)offsetBinFile, SEEK_SET))
      {
         (void)snprintf(logBuffer, sizeof(logBuffer),
                        "[%.17s] Can't seek to offset %u in bin file",
                        pStation->conID,
                        (unsigned int)offsetBinFile);
         logWarning(logBuffer, fileName);

         if (0 != sendDFF)
         {
            (void)snprintf(logBuffer, sizeof(logBuffer),
                  "[%.17s] Delete frames with same DFF offset [%u] from IRL",
                           pStation->conID,
                           (unsigned int)(offsetBinFile));
            logWarning(logBuffer, fileName);

            /* Current frame is a DFF, remove it and all
               related frames from the IRL */
            pCurrElement = removeFrameWith(pRefListHead, pCurrElement,
                                           THAT_DFF);

            /* Make sure that next frame is checked for DFF */
            sendDFF = 0;
         }
         else
         {
            /* Remove single frame from sending queue.
               Error has been logged before */
            pCurrElement = removeFrameFromIRL(pRefListHead, pCurrElement);
         }
         /* This is not treated as an error */
          continue;
      }

      /* check for a pending shutdown */
      if (pStation->shutdownEnable != SD_NONE)
      {
         (void) fclose(fp);
         RH_DEBUG(2, "shutdown pending");
         return(S_SHUTDOWN);
      }

      RH_DEBUG3(5, "Offset %u, framesize %u", (unsigned int)(offsetBinFile),
                (unsigned int)(frameSize));

      /* Check if the stations frame buffer is large enough to hold
         the frame */
      if (pStation->frameBufferSize < frameSize)
      {
         if (-1 == adjustBufferSize(pStation, frameSize, &event))
         {
            (void) fclose(fp);
            return(event);
         }
      }

      frameLength = 0;
      retries = MAX_READ_RETRIES;
      /* Read the complete frame into buffer */
      while (frameLength < frameSize)
      {
         size_t len;

         len = fread(pStation->frameBuffer + frameLength, 1,
                     frameSize - frameLength, fp);

         if (0 == len)
         {
            /* This can be caused by file system caching (e.g.
               data to send and sender are located on different hosts and
               the sender uses an NFS share to access the data).
               This event is silently ignored when compiled with
               retries as long as max. retry count is not reached */
            RH_DEBUG3(2, "fread returns 0: frameLength %d, errno %d",
                      (int)frameLength, errno);
            RH_DEBUG3(2, "Offset %u, framesize %u",
                      (unsigned int)(offsetBinFile),
                      (unsigned int)(frameSize));

            if (retries <= 0)
            {
               (void)snprintf(logBuffer, sizeof(logBuffer),
                              "[%.17s] Unable to read frame at offset [%u]",
                              pStation->conID,
                              (unsigned int)(offsetBinFile));
               logWarning(logBuffer, fileName);
               /* This is not treated as an error */
               break;
            }
            clearerr(fp);
            /* NOTE: If wfdisc records should be synced based on time here
               use retryWaitWithSync() instead */
            retryWait(pStation, 5);
            retries--;
         }
         else retries = MAX_READ_RETRIES;

         frameLength += (uint32_t)len;
      } /* TODO: make a function out of this block */

      if (frameLength != frameSize)
      {
         /* Partial frames are not parsed.
            This is not treated as an error */
         if (0 != sendDFF)
         {
            (void)snprintf(logBuffer, sizeof(logBuffer),
                  "[%.17s] Delete frames with same DFF offset [%u] from IRL",
                           pStation->conID,
                           (unsigned int)(offsetBinFile));
            logWarning(logBuffer, fileName);

            /* Current frame is a DFF, remove it and all
               related frames from the IRL */
            pCurrElement = removeFrameWith(pRefListHead, pCurrElement,
                                           THAT_DFF);

            /* Make sure that next frame is checked for DFF */
            sendDFF = 0;
         }
         else
         {
            /* Remove frame from IRL.
               Error has been logged before */
            pCurrElement = removeFrameFromIRL(pRefListHead, pCurrElement);
         }
         continue;
      }

      /* Set default values and authentication state */
      pStation->checkPass = NO_ERROR;
      pStation->authenPassFrame = AUTH_UNKNOWN; /* default value */
      pStation->authenSummary = AUTH_UNKNOWN; /* default value */
      if (NULL != pStation->DFinfo)
      {
         /* Don't reset number of parsedDFFchannels here.
            Don't forget to reset parsedDFFchannels
            when the filter is switched off via a command frame
            sometime !! */
         pStation->DFinfo->parsedChannels = 0;
         /* Sets the nominal time in the DF info structure to
            time read from clf file to have the nominal time
            logged in clf and txt files */
         pStation->DFinfo->nominalTime = pCurrElement->nominalTime;
      }

      /* A frame is "good" by definition when it is sent */
      if (clfType >= 'A' && clfType <= 'Z') clfType += ' ';

      /* Parse frame header and check if frame size in frame equal to
         frame size in clf file */
      event = checkIfFrameValid(pStation->frameBuffer, clfType, frameLength,
                                pStation->type);

      if (NO_ERROR != event)
      {
         if (E_INTERNAL == event)
         {
            /* If an internal error occurs cd2w should start with
               this frame again */
            shutdownApplication("Unexpected station type", pStation->conID);
            (void)fclose(fp);
            return(event);
         }
         /* Don't process too short or out of sync frames.
            It is too dangerous to parse these frames. Log
            a warning and skip the frame instead  */
         if (E_OUT_OF_SYNC == event)
         {
            pStation->checkPass |= WA_UNKNOWN_TYPE;
         }
         else if (E_FRAME_TOO_BIG == event)
         {
            pStation->checkPass |= ER_FRAME2BIG;
         }
         else
         {
            pStation->checkPass |= ER_SIZE;
         }
         /* Log also old warning as before */
         logWarning("Frame type or size invalid, skipped (out of sync?)",
                    pStation->conID);
         /* Log the event */
         logCDevent(ET_W_CON, pStation, event, NULL, (uint32_t)frameLength);

         if (0 != sendDFF)
         {
            (void)snprintf(logBuffer, sizeof(logBuffer),
                  "[%.17s] Delete frames with same DFF offset [%u] from IRL",
                           pStation->conID,
                           (unsigned int)(offsetBinFile));
            logWarning(logBuffer, fileName);

            /* If the current frame is a DFF, remove it and all
               related frames from the IRL */
            pCurrElement = removeFrameWith(pRefListHead, pCurrElement,
                                           THAT_DFF);

            /* Make sure that next frame is checked for DFF */
            sendDFF = 0;
         }
         else
         {
            /* Remove frame from IRL.
               Error has been logged before */
            pCurrElement = removeFrameFromIRL(pRefListHead, pCurrElement);
         }
         continue;
      }

      /* Try to merge the frame now. DFF's are only parsed in the
         called function */
      if (0 == insertSegment(pStation, clfType, frameLength, &event,
                             pCd2wParams->senderMode,
                             pCurrElement,
                     (time_t)(pCd2wParams->lookBackTimeWorker * ONE_DAY),
                             pCd2wParams->sampleRateTolerance))
      {
         RH_DEBUG(3, "frame skipped");
      }

      RH_DEBUG2(2, "insertSegment reports [%d]", (int) event);
      /* If a shutdown is pending and this is the first frame
         processed make sure that framesSinceLastSave is increased.
         Make also sure that ALL channels are processed when checking
         for S_SHUTDOWN which mean do not check for S_SHUTDOWN in
         insertSegment when at least one channel is processed! */
      /* --> if (NO_ERROR != event) */
      if (NO_ERROR != event && S_SHUTDOWN != event)
      {
         RH_DEBUG3(5, "Last offset was %u, framesize %u",
                   (unsigned int)(offsetBinFile),
                   (unsigned int)(frameSize));

         /* If an error occurs cd2w should start with
            this frame again */
         (void) fclose(fp);
         return(event);
      }

      /* If a DFF has been processed before, process list element now */
      if (0 != sendDFF) continue;

      /* Remember the sending sequence to be able to go back N frames
         when writing a restart point */
      if (NULL != P_LAST_ELEMENT_SENT)
      {
         P_LAST_ELEMENT_SENT->pHistNext = pCurrElement;
      }
      pCurrElement->pHistPrev = P_LAST_ELEMENT_SENT;
      /* Save the last element sent in intRefList header for re-start */
      P_LAST_ELEMENT_SENT = pCurrElement;
      pCurrElement->sendCount++;
      /* Set state to acknowledged */
      pCurrElement->sendState = SENT_ACK;

      /*@-realcompare@*/
      if (0.0 == pRefListHead->restartNominalTime
          ||
          (LIFO == sortType &&
          pRefListHead->restartNominalTime > pCurrElement->nominalTime)
          ||
          (LIFO != sortType &&
           pRefListHead->restartNominalTime < pCurrElement->nominalTime))
      {
         /* Remember nominal time of oldest (LIFO) or newest (FIFO/unsorted)
            frame sent. This information in needed for re-start */
         pRefListHead->restartNominalTime = pCurrElement->nominalTime;
      }
      /*@=realcompare@*/

      pRefListHead->framesSinceLastSave++;

      /* Check for a pending shutdown */
      if (pStation->shutdownEnable != SD_NONE)
      {
         if (NULL != fp) (void) fclose(fp);
         RH_DEBUG(2, "shutdown pending");
         return(S_SHUTDOWN);
      }

      doSync = 0;
      /* Check if the restart file, wfdisc records/file and
         cdquality records/file (if any) should be synced based
         on frames processed */
      if (0 != pCd2wParams->refListStoreInterval
          &&
          pRefListHead->framesSinceLastSave >
                           pCd2wParams->refListStoreInterval)
      {
         ST_DEBUG(2, "frame based sync");
         doSync = 1;
         /* Save number of frame based syncs in invalidStations, it's not used
            in cd2wng otherwise */
         pStation->whenConnected.invalidStations++;
      }

      /* Check if the restart file, wfdisc records/file and
         cdquality records/file (if any) should be synced based
         based on time passed */
      if (0 == doSync && 0 != pCd2wParams->syncTimeInterval)
      {
         struct timeval nowTime;

         (void) gettimeofday(&nowTime, NULL);
         if (pCdSend->nextSyncTime <= nowTime.tv_sec)
         {
            ST_DEBUG(2, "time based sync");
            doSync = 1;
            /* Save number of time based syncs in isConnected, it's not used
               in cd2wng otherwise */
            pStation->whenConnected.isConnected++;
         }
      }

      if (1 == doSync)
      {
         /* Update wfdisc records and re-start information */
         event = syncWfdiscs(pStation, MAX_MERGE_DAYS, SYNC_WFDISCS);
         if (NO_ERROR != event)
         {
            if (NULL != fp) (void) fclose(fp);
            return(event);
         }
      }

      /* Check if the sender runs in FIFO or LIFO mode and whether or
         not new data has arrived in the meantime. This check will
         not detect new frames in clf files created after the current
         clf file!! */
      if (UNSORTED != sortType &&
          (0 != haveDataToSend(*pCurrentClf, pRefListHead->currentClfFile)))
      {
         /* Add all frames from the current position
            in the clf file to the end of the last clf file to the internal
            reference list */
         event = parseAllClfFiles(pStation, pRefListHead,
                                  pCurrentClf, sortType);
         if (NO_ERROR != event)
         {
            if (NULL != fp) (void) fclose(fp);
            return(event);
         }
         /* There might be some new frames on top of the IRL, but
            the sending pointer is somewhere in the IRL. Therefore
            a check if the first element on the IRL was not send has
            to be done here. */
         if (pRefListHead->pRefListStart != NULL
             && NOT_SENT == pRefListHead->pRefListStart->sendState)
         {
            pCurrElement = pRefListHead->pRefListStart;
         }
         else
         {
            /* Process next element in internal reference list */
            pCurrElement = pCurrElement->pNextElement;
         }
      }
      else
      {
         /* Process next element in internal reference list */
         pCurrElement = pCurrElement->pNextElement;
      }

#undef P_LAST_ELEMENT_SENT

   } /* ( NULL != pCurrElement) */

   IF_RH_DEBUG(6) { dumpIRLsendSequence(pRefListHead); }

   /* Check fp. If deleteWithThatFileObject() deletes all frames, fp
      can became NULL */
   if (NULL != fp) (void) fclose(fp);

   if (pCdSend->sndOpMode & OPER_MODE_AUX)
   {
      pStation->shutdownEnable = SD_STOP;
      RH_DEBUG(2, "aux mode; shutdown sender thread");
      return(S_SHUTDOWN);
   }
   return (NO_ERROR);
}

/**
 * @short Restore internal reference list
 *
 * @description This functions restores the internal reference
 * @description list when a restart file is found. If no restart file
 * @description is found, re-sending starts at the first frame in
 * @description the clf file for the current day. If a restart file
 * @description file is found, but the corresponding restart point
 * @description is not found in the clf file list (too old restart point)
 * @description than the whole clf file list is sent
 *
 * @param pStation      station object
 * @param pRefListHead  intRefList header
 *
 * @return 0 if an error or no list restored, otherwise 1
 */

static int
restoreIntRefList(station_t *pStation, intRefListHead_t *pRefListHead)
{
   char *pFile2Search,
        logBuffer[LOGSIZE],
        fnBufferLastSent[LIBPAR_BUFSIZ],  /* buffer history line */
        fnBufferLastParsed[LIBPAR_BUFSIZ];
   unsigned long tmpOffset,
                 tmpOffsetLastParsed;
   clfFileParams_t *pClfFile,  /* current clf file list entry */
                   *pRestartFile = NULL; /* file containing the restart point */
   double restartNominalTime; /* nominal time of oldest/newest frame sent */
   uint16_t event = 0;
   int sortMode = UNSORTED;
   cdsend_t *pCdSend;  /* Application extension object */

   if (NULL == pStation || NULL == pRefListHead->intRefListFp ||
       NULL == pStation->appExtension ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pClfFilesList)
   {
      shutdownApplication("Internal error in restoreIntRefList",
                          pStation->conID);
      return(0);
   }

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   ST_DEBUG(2, "Restore intRefList");

   /* Set default values */
   pRefListHead->pLastSent = NULL;
   pRefListHead->frameOffset = 0;
   pRefListHead->lastDFFoffset = (uint32_t)0;
   pRefListHead->lastDFFsize = (uint32_t)0;
   pRefListHead->framesSinceLastSave = 0;
   pRefListHead->restartNominalTime = 0.0;

   if (0L == lseek(fileno(pRefListHead->intRefListFp), 0L, SEEK_END))
   {
      /* Empty restart file */
      char today[10];

      (void)snprintf(logBuffer, sizeof(logBuffer),
           "[%.17s] Nothing to restore", pStation->conID);
      logWrite(LOG_INFO, logBuffer);

      if (0 != (pCdSend->sndOpMode & OPER_MODE_FIXED))
      {
         /* Don't drop clf files in fixed interval mode */
         return(0);
      }

      /* Get current date as string */
      getToday(today);

      pClfFile = pCdSend->pClfFilesList->pClfFileHead;
      /* Start sending all clf files found in clf file list which
         are created today. Remove all other clf files from the list */
      while (NULL != pClfFile)
      {
         char *startSearch;

         /* Search last slash */
         startSearch = rindex(pClfFile->pFile, (int) '/');
         /* Check if slash was present in string */
         if (NULL == startSearch) { startSearch = pClfFile->pFile; }

         startSearch = index(startSearch, (int) '.');
         /* Check if dot was present in string */
         if (NULL == startSearch)
         {
            logError("Internal error", "inconsistent clf file name");
            pClfFile = pClfFile->pNextClfFileParams;
            continue;
         }
         /* Skip dot */
         startSearch++;
         /* Check if the clf file has been created today */
         ST_DEBUG2(2, "start compare at %s", startSearch);
         if (0 == strncmp(startSearch, today, 8))
         {
            pClfFile = pClfFile->pNextClfFileParams;
            continue;
         }
         ST_DEBUG2(2, "Delete %s from clf file list", pClfFile->pFile);

         /* clf file not created today.
            Get next element and unlink and destroy current element */
         pClfFile = unlinkClfElement(pStation, pClfFile, &event);
         if (NO_ERROR != event) break;
      }
      return(0);
   }

   /* Set file pointer to begin of file */
   rewind(pRefListHead->intRefListFp);

   /* Load processMode, last clf file parsed and offset of the last parsed
      frame in that clf file during processing */
   if (3 != fscanf(pRefListHead->intRefListFp, RESTART_LINE_1_IN,
                  &sortMode, fnBufferLastParsed, &tmpOffsetLastParsed))
   {
      logWarning("Can't scan restart file", pStation->conID);
      return(0);
   }
   (void)snprintf(logBuffer, sizeof(logBuffer),
           "[%.17s] mode %d, last clf file parsed [%s] and offset %lu",
           pStation->conID, sortMode, fnBufferLastParsed, tmpOffsetLastParsed);
   logWrite(LOG_INFO, logBuffer);

   /* Load clf file of the last frame sent, offset last frame sent and
      nominal time of oldest (LIFO) or newest (FIFO) frame */
   if (3 != fscanf(pRefListHead->intRefListFp, RESTART_LINE_2_IN,
                   fnBufferLastSent, &tmpOffset, &restartNominalTime))
   {
      logWarning("Can't scan restart file, line 2", pStation->conID);
      return(0);
   }
   (void)snprintf(logBuffer, sizeof(logBuffer),
          "[%.17s] last frame sent filename [%s] and offset %lu; RNT %f",
          pStation->conID, fnBufferLastSent, tmpOffset, restartNominalTime);
   logWrite(LOG_INFO, logBuffer);

   /* Add unsent or un-acknowledged frames before the restart point to
      the Internal Reference List. The sender does not know the protocol
      type of these frames. If the protocol type of the data provider
      has changed, these frames have to be removed manually from the
      restart file */
   (void)loadIrlEntries(pRefListHead);

   /* Re-schedule unacknowledged frames for sending (set state to unsent) */
   /* (void)setUnsent(pRefListHead); */

   /* Check if last mode was unsorted. In that case only information
      about the last frame sent is needed for correct restart.
      Otherwise the nominal time of the last frame sent and the information
      of the last parsed clf file is needed */
   if (UNSORTED == sortMode)
   {
      pFile2Search = fnBufferLastSent;
      restartNominalTime = 0.0;
   }
   else
   {
      tmpOffset = tmpOffsetLastParsed;
      pFile2Search = fnBufferLastParsed;
   }

   /* Check if the re-start point is in the list of clf files to process */
   for (pClfFile = pCdSend->pClfFilesList->pClfFileHead;
        NULL != pClfFile;
        pClfFile = pClfFile->pNextClfFileParams)
   {
      ST_DEBUG2(2, "check clf file name [%.128s]", pClfFile->pFile);

      if (0 == strncmp(pClfFile->pFile, pFile2Search,
                       sizeof(fnBufferLastSent)))
      {
         /* Restart point found */
         (void)snprintf(logBuffer, sizeof(logBuffer),
                        "[%.17s] restart file found in clflist: [%.128s]",
                        pStation->conID, pFile2Search);
         logWrite(LOG_INFO, logBuffer);
         pRestartFile = pClfFile;
      }
   }

   /* Start to scan the list from begin */
   pClfFile = pCdSend->pClfFilesList->pClfFileHead;

   if (NULL == pRestartFile)
   {
      /* If no restart point is found, send the whole clf list */
      (void)snprintf(logBuffer, sizeof(logBuffer),
                     "[%.17s] no restart file found in clflist",
                     pStation->conID);
      logWrite(LOG_INFO, logBuffer);
   }
   else
   {
      /* The next step depends on last mode used. If the internal reference
         list was sorted as the re-start point was saved, all clf files
         before the restart point have to be parsed for data frames with
         nominal time older or newer (depending on the sorting) as the
         saved nominal time. Frames found in that case have to be included
         in the internal reference list. Parsing the clf file(s) is not
         needed and therefore skipped if the last mode was unsorted */
      while (NULL != pClfFile)
      {
         if (pClfFile == pRestartFile) break;

         if (UNSORTED != sortMode)
         {
            int nbrOfNewEntries;

            /* Parse clf file for frames which have older (or newer)
               nominal time as the last frame sent and add them to
               internal reference list */
            pStation->type = pClfFile->cdtype;
            nbrOfNewEntries = parseClfFile(pClfFile->pFile, pStation->type,
                              pRefListHead, restartNominalTime, 0,
                              OLD_FILE, sortMode);
            ST_DEBUG3(2, "restoreIntRefList: %d new entries in %s",
                      nbrOfNewEntries, pClfFile->pFile);
            /* Close current clf file and check if clf file object
               should be removed */
            checkClfFileObj(pRefListHead);
         }
         /* get next element and unlink and destroy current element */
         pClfFile = unlinkClfElement(pStation, pClfFile, &event);
         if (NO_ERROR != event) break;
      }

      /* The last file parsed has to be treated specially. Ignore all
         frames before the restart point (unsorted) or check the nominal
         time (sorted). Add all frames after the restart point regardless of
         their nominal time and last sorting mode. */
      if (NULL != pClfFile)
      {
         pStation->type = pClfFile->cdtype;
         (void)parseClfFile(pClfFile->pFile, pStation->type, pRefListHead,
                            restartNominalTime, tmpOffset,
                            RESTART_FILE, sortMode);

         /* Don't unlink the last clf file in the list!! It could grow */
         if (NULL != pClfFile->pNextClfFileParams)
         {
            /* Close current clf file and check if clf file object
               should be removed */
            checkClfFileObj(pRefListHead);
            /* get next element and unlink and destroy current element */
            pClfFile = unlinkClfElement(pStation, pClfFile, &event);
            if (NO_ERROR != event) return(0);
         }
         else pClfFile = pClfFile->pNextClfFileParams;
      }
   } /* NULL != pRestartFile */

   /* The last step is to check if clf files newer than the restart point
      exist. If no restart point is known, it is assumed that all files
      are newer. In both cases all frames from these files have to be added to
      the internal reference list. This is done in the sending() function */
   return(1);
}

/**
 * @short Wait for retry or sync time to expire or user shutdown
 *
 * This function waits until the retry or sync time has
 * expired as long as no user shutdown is pending. If time based
 * synchronization is used, this function checks also for
 * the next sync time and issues a sync if necessary
 *
 * @param pStation          pointer to station object
 * @param WaitTime          number of seconds to wait
 * @param syncTimeInterval  number of seconds to wait
 * @param pCdSend           sender object
 *
 * @return NO_ERROR on success or error event on error
 */

static uint16_t
retryWaitWithSync(station_t *pStation, int WaitTime,
                  uint16_t syncTimeInterval, cdsend_t *pCdSend)
{
   struct timespec sleeptime;
   int ii;
   time_t actTime = 0L;
   struct timeval nowTime;
   uint16_t retCode = NO_ERROR;

   /* Wait for retry time to expire or user shutdown */
   /*@-usedef@*/
   sleeptime.tv_sec = 1;
   sleeptime.tv_nsec = 0;
   /*@=usedef@*/

   /* Reset timer even if nothing has to be synced! */
   if (0 != syncTimeInterval)
   {
      (void) gettimeofday(&nowTime, NULL);
      actTime = nowTime.tv_sec;
   }

   /* pStation->pConfigData->retryTime; */
   for (ii = WaitTime;
        ii > 0 && SD_NONE == pStation->shutdownEnable; ii--)
   {
      ST_DEBUG(5, "Before retry wait with sync");
      /*@-usedef@*/
      (void) nanosleep(&sleeptime, NULL);
      /*@=usedef@*/
      actTime++;
      if (0 != syncTimeInterval && pCdSend->nextSyncTime <= actTime)
      {
         ST_DEBUG(2, "time based sync (retryWait)");
         /* Syncing the wfdiscs might take some time. This
          * is not considered in waitTime! */
         retCode = syncWfdiscs(pStation, MAX_MERGE_DAYS, SYNC_WFDISCS);
         /* Save number of time based syncs in isConnected, it's not used
            in cd2wng otherwise */
         pStation->whenConnected.isConnected++;
         if (NO_ERROR != retCode) return (retCode);
         (void) gettimeofday(&nowTime, NULL);
         actTime = nowTime.tv_sec;
      }
   }
   return(retCode);
}

/**
 * @short Merge thread - main function
 *
 * @description This function handles a stations data.
 * @description It takes care that data to process is available when
 * @description calling processIrlData
 *
 * @param arg  void pointer to config parameters
 *
 * @return None
 */

/*@null@*/void *
merge(void *arg)
{
   CDevent_t actEvent;    /* Event buffer */
   const configParams_t *pConfigData;
   station_t *pStation;
   intRefListHead_t *pRefListHead;
   uint16_t event = NO_ERROR;
   int sortType;
   cdsend_t *pCdSend;  /* Application extension object */
   senderParams_t *pCd2wParams; /* Configuration extension object */
   clfFileParams_t *pCurrentClf;

   /* Do some internal checks first */
   if (NULL == arg)
   {
      shutdownApplication("Internal error in processing",
                          "No config arg available");
      pthread_exit(NULL);
      return(NULL);
   }

   pStation = (station_t *) arg;     /* pointer to actual station_t */
   pConfigData = pStation->pConfigData;

   if (NULL == pConfigData || NULL == pStation->appExtension ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pClfFilesList ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pRefListHead ||
       NULL == ((cdsend_t *)(pStation->appExtension))->pProdLine->pDataConsumer)
   {
      shutdownApplication("Internal error in processing",
                          "unexpected NULL pointer (1)");
      /* Set the state to unconnected, close all files and detach the thread */
      cleanUp(pStation);
      if (NULL != pConfigData)
      {
         pConfigData->pHistory->nbrOfThreadsStopped++;
      }
      pthread_exit(NULL);
      return(NULL);
   }

   /* set sender specific config parameters */
   pCd2wParams = (senderParams_t *)pConfigData->configExtension;

   /* Access application extension data */
   pCdSend = (cdsend_t *)pStation->appExtension;

   /* Check if the waveform header is already initialized, otherwise
      allocate a waveform header */
   if (NULL == pCdSend->addAppExtension)
   {
      /* Allocate wave form header */
      pCdSend->addAppExtension = calloc(1, sizeof(waveFormHead_t));
      if (NULL == pCdSend->addAppExtension)
      {
         shutdownApplication("Can't allocate waveform header", "merge");
         return(NULL);
      }
      /* Set waveform sample size */
      if ('3' == pCd2wParams->waveformType[1])
          ((waveFormHead_t *)pCdSend->addAppExtension)->sampleSize = 3;
      else
         ((waveFormHead_t *)pCdSend->addAppExtension)->sampleSize = 4;

      ((waveFormHead_t *)pCdSend->addAppExtension)->channelCnt =
                         getChannelCount(pStation->stationName);
      DEBUG3(2, "Channel cnt for [%.8s] is %d", pStation->stationName,
             (int)((waveFormHead_t *)pCdSend->addAppExtension)->channelCnt);
      ((waveFormHead_t *)pCdSend->addAppExtension)->channelAttrMaskHd =
                                pCd2wParams->channelStatusMask;
   }

   ST_DEBUG2(2, "Thread [%ld] for Prodline started", (long)pthread_self());

   /* Check if FIX Interval mode */
   if (0 != (pCdSend->sndOpMode & OPER_MODE_FIXED))
   {
      /* Check if something is to do */
      if (NULL == pCdSend->pClfFilesList->pClfFileHead)
      {
         ST_DEBUG(2, "No entries in index file");
         /* Set the state to unconnected, close all files and
            detach the thread */
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         pthread_exit(NULL);
         return(NULL); /*for splint */
      }
   }

   /* state is wait for send */
   pRefListHead = pCdSend->pRefListHead;

   /* The following line is not really necessary, but has been added
      to avoid breaking any existing test case. A value of 0 for
      unacknowledged frames (which should be the default value) is
      also fine for cd2wng */
   pRefListHead->unAckedHighMark = NBR_RESENT_FRAMES;

   /* Set event values */
   actEvent.conID = pStation->conID;
   actEvent.portInfo = pStation->portInfo;
   actEvent.is_CR = 0;
   actEvent.type = UNKNOWN;
   actEvent.lastCRFreceived = 0;

   /* Do not initialize output files here. Cd2wng does not
      write binary, text or clf files and it uses a slightly
      different file naming schema! */

   /* Try to open or create internal reference list file */
   if (1 == openIntRefList(pCd2wParams->refListStorePath, pRefListHead, 0))
   {
      /* Set the state to unconnected, close all files and detach the thread */
      cleanUp(pStation);
      pConfigData->pHistory->nbrOfThreadsStopped++;
      pthread_exit(NULL);
      return(NULL); /*for splint */
   }

   /* Check if a saved internal reference list exists and if
      one exists and sender is not in auxiliary mode, load it */
   if (0 == (pCdSend->sndOpMode & OPER_MODE_AUX))
   {
      (void)restoreIntRefList(pStation, pRefListHead);
   }
   else
   {
      /* Set default values */
      pRefListHead->pLastSent = NULL;
      pRefListHead->frameOffset = 0;
      pRefListHead->lastDFFoffset = (uint32_t)0;
      pRefListHead->lastDFFsize = (uint32_t)0;
      pRefListHead->framesSinceLastSave = 0;
      pRefListHead->restartNominalTime = 0.0;
   }

   pCurrentClf = pCdSend->pClfFilesList->pClfFileHead;

   /* If a restart point was found, all frames after the restart point
      which are not processed before are added to the IRL now */
   if (0 != haveDataToSend(pCurrentClf,
                           pRefListHead->currentClfFile))
   {
      event = parseAllClfFiles(pStation, pRefListHead,
                               &pCurrentClf,
                               UNSORTED);
      if (NO_ERROR != event)
      {
         /* Set the state to unconnected,
            close all files and detach the thread */
         cleanUp(pStation);
         pConfigData->pHistory->nbrOfThreadsStopped++;
         pthread_exit(NULL);
         return(NULL); /*for splint */
      }
   }

   /*  Make sure that the production line is in STOPPED state when
       the processing order is changed!! */
   sortType = getSortLevel(pStation);

   /* The IRL is unsorted, check if the list should be sorted */
   RH_DEBUG2(2, "sortType = %d", sortType);
   IF_RH_DEBUG(10) { dumpIRL(pRefListHead); }

   if (UNSORTED != sortType)
   {
      /* Sort internal reference list now. If the sorting
         order (processing mode) can be changed using a command,
         make sure that the production line is in STOPPED state when
         the processing order is changed or the sorting order is not
         correct! */
      sortIntRefList(pRefListHead, sortType);
      IF_RH_DEBUG(10) { dumpIRLsendSequence(pRefListHead); }
      RH_DEBUG(2, "sorting finished");
   }

   RH_DEBUG2(2, "syncTimeInterval=%hu",
             (unsigned short int)pCd2wParams->syncTimeInterval);
   if (0 != pCd2wParams->syncTimeInterval)
   {
      /* This could trigger syncing of wfdiscs also if no frame is
         processed, but that does not hurt. This condition is checked
         when the wfdiscs are synced.
       */
      struct timeval nowTime;

      (void) gettimeofday(&nowTime, NULL);
      pCdSend->nextSyncTime = nowTime.tv_sec +
                              (time_t)pCd2wParams->syncTimeInterval;
   }
   else
   {
      /* Reset sync timer */
      pCdSend->nextSyncTime = 0;
   }

   /* The (re)start has been made. The IRL should contain now all unsent
      frames and should be in the correct sorting order */
   while (SD_NONE == pStation->shutdownEnable)
   {
      int framesNotprocessed = 0;           /* number of unsent frames in IRL */

      /* Reset all events */
      event = NO_ERROR;

      /* If no clf file has exist when the production line has started,
         pCurrentClf will never have a successor. So an additive check
         is made */
      if (NULL == pCurrentClf)
      {
         pCurrentClf = pCdSend->pClfFilesList->pClfFileHead;
      }

      /* Parse all clf files if new frames are available */
      if (0 != haveDataToSend(pCurrentClf,
                              pRefListHead->currentClfFile))
      {
         /* Add new frames to the internal reference list in the correct
            order */
         event = parseAllClfFiles(pStation, pRefListHead, &pCurrentClf,
                                  sortType);
         if (NO_ERROR != event)
         {
            break; /* Break while loop */
         }
      }

      /* Query number of unsent frames in IRL */
      framesNotprocessed = getUnsentFrames(pRefListHead);

      /* Check if all frames have been processed in fix interval mode */
      if (0 == framesNotprocessed
          &&
          0 != (pCdSend->sndOpMode & OPER_MODE_FIXED)
          &&
          NULL == pCurrentClf)
      {
         /* Not sure if this check is really necessary */
         break;
      }

      /* Check if something is to do in continuous mode. Fixed mode
         has been checked before. Don't start to send before data is
         available in the clf files or the station type may be set wrong.
         Check for NULL is necessary to wait for retry when no
         clf file is available, e. g. the last N files are read from
         the restart file, but no corresponding clf file exists */
      if (NULL == pCurrentClf
          ||
          0 == framesNotprocessed)
      {
         ST_DEBUG(2, "No un-sent frames or no clf file found");

         if (0 == (pCdSend->sndOpMode & OPER_MODE_FIXED))
         {
            /* Wait for some time to expire or user shutdown */
            event = retryWaitWithSync(pStation,
                                     (int)pCd2wParams->sleepTimeWork,
                                     pCd2wParams->syncTimeInterval,
                                     /*@-compmempass@*/
                                     pCdSend);
                                     /*@=compmempass@*/
            if (NO_ERROR != event)
            {
               break; /* Break while loop */
            }
            continue;
         }
      }

      while (framesNotprocessed > 0 && NO_ERROR == event)
      {
         /* Check for pending shutdown request */
         if (SD_NONE != pStation->shutdownEnable)
         {
            event = S_SHUTDOWN;
            break;
         }

         /* Station type is the type of the *last* clf file seen!! */

         /* There is something to process in the internal reference list */
         event = processIrlData(pRefListHead, pStation, &actEvent,
                                &pCurrentClf);
         ST_DEBUG2(5, "processIrlData returns [%d]", (int)event);

         if (0 == pCd2wParams->refListStoreInterval)
         {
            /* Normally the internal reference list is
               compacted in saveIntRefList(), but no re-start
               point is written if refListStoreInterval is set to 0 */
           compactIntRefList(pRefListHead);
         }

         /* Check for system error, closed connection or user shutdown */
         if (NO_ERROR != event)
         {
            break; /* Break while loop */
         }

         /* Check if the application couldn't open the current clf file */
         if (NULL == pRefListHead->currentClfFile)
         {
            /* Close current clf file and check if clf file object
               should be removed */
            checkClfFileObj(pRefListHead);
         }
         else
         {
            /* Continuous mode and end of current clf file reached.
               It is possible that this file will grow, so
               this file has to be processed again in continuous mode */
            ST_DEBUG(2, "End of current clf file reached");

           /* Break loop to wait for received data */
           break;
         }

         ST_DEBUG2(5, "number of clf files in list [%d]",
                   pCdSend->pClfFilesList->nbrOfClfParams);

         /* Query number of unsent frames in IRL */
         framesNotprocessed = getUnsentFrames(pRefListHead);
         ST_DEBUG2(2, "number of not processed frames in list [%d]",
                       framesNotprocessed);

      } /* while (framesNotprocessed > 0 && NO_ERROR == event) */

      if (NO_ERROR != event)
      {
         break; /* Break while loop */
      }

      /* No event. Check for fixed interval mode */
      if (0 != (pCdSend->sndOpMode & OPER_MODE_FIXED))
      {
         /* Fixed interval mode and all frames in IRL sent */
         ST_DEBUG(2, "All frames sent in fixed interval mode");
         break;
      }
   } /* while (SD_NONE == pStation->shutdownEnable) */

   if (NO_ERROR != event)
   {
      /* Log the event and stop processing */
      if (S_SHUTDOWN == event)
      {
         logCDevent2(ET_L_INFO, event, &actEvent, NO_INFO);
      }
      else if (NO_ERROR != event)
      {
         if (E_SYSTEM == event) pStation->savedSystemErr = (uint32_t)errno;
         logCDevent2(ET_L_ERROR, event, &actEvent, (uint32_t)errno);
      }
   }

   /* Handling thread has been stopped */
   logCDevent2(ET_L_INFO, S_DISCONNECTED, &actEvent, NO_INFO);

   /* Sync Wfdisc (save segments), release segments (which is necessary if
      shutdown has been caused by a stop command), make sure that all
      channel headers have been written, update frameproduct table
      and save restart point */
   (void)closeWindow(pStation, pCdSend->addAppExtension, &actEvent, event);

   /* Clear internal reference list */
   clearIntRefList(pRefListHead);
   /* Check if  current clf file has to be closed and check if clf file object
      has no more references and can be removed */
   checkClfFileObj(pRefListHead);
   if (NULL != pRefListHead->intRefListFp)
      (void)fclose(pRefListHead->intRefListFp);
   pRefListHead->intRefListFp = NULL;
   /* Set the state to unconnected or stopped,
      close all files and detach the thread */
   cleanUp(pStation);
   pConfigData->pHistory->nbrOfThreadsStopped++;
   ST_DEBUG(2, "Processing thread exits");
   pthread_exit(NULL);
   /*@-mustfreeonly@*/
   return(NULL); /*for splint */
   /*@=mustfreeonly@*/
}
