/*******************************************************/
/*  International Data Centre                          */
/*  Comprehensive Test Ban Treaty Organization         */
/*  Vienna                                             */
/*  Austria                                            */
/*******************************************************/

/**
 * @author    Martin Sieder
 *
 * Revision 1.34  2015-11-27 gerald
 * Follow change in CDlibVersion()
 *
 * Revision 1.33  2014-10-17 gerald
 * Fix check for maximum number of channels
 *
 * Revision 1.32  2013-06-13 gerald
 * Add CDrep and libCD version to usage string.
 * Add closing libCD
 *
 * Revision 1.31  2013-04-09 gerald
 * Support new idcsyslog public header filename (from cbase 1.1.0)
 *
 * Revision 1.30  2012-11-26 gerald
 * Add time of last GPS time synchronisation to extended
 * channel output
 *
 * Revision 1.29  2012-06-27 gerald
 * Fix function comment
 *
 * Revision 1.28  2012-02-20 gerald
 * Set application type to other to load certificates for
 * data provider (bugfix for ISI-38). This bugfix requires
 * libCD >= 2.3.1 (cdrep.c)
 *
 * Revision 1.27  2011-06-09 15:37:01  ctbto
 * Fix missing authentication checks in cdrep
 *
 * Revision 1.26  2010-05-11 15:27:10  gerald
 * Write question mark when no data type is found in
 * the CSF. This is necessary for parsing the data
 * afterwards
 *
 * Revision 1.25  2010-01-21 14:39:39  martins
 * Enhance cdrep:
 * Introduce the new option -g
 * This new option forces cdrep to write the stdout output to the ehl files.
 * The content which will be written to the ehl files depends on the other
 * specified options.
 *
 * Examples:
 * cdrep config_file -e -g
 * writes event stream to ehl files, same output as option -h
 *
 * cdrep config_file -f -g
 * writes frame stream to ehl files
 *
 * cdrep config_file -f -c -g
 * writes frame stream and channel stream to ehl files
 *
 * cdrep config_file -e -f -d -g
 * writes event stream, frame stream and extended channel stream to ehl files
 *
 * Revision 1.24  2009-09-14 14:13:07  martins
 * Enhance channel stream mode: subframe time length for CD1.0 frames
 * Instead of "NA" the value of DFinfo->frameTimeLength is printed to stdout.
 *
 * Revision 1.23  2009-08-13 14:28:05  martins
 * Fix bug which causes the exit of cdrep in storeStatusDataOfAllPresentChannels().
 * If a channels position in the channel mask has the value -1,
 * it is not to be considered as an error. The channel has to be skipped.
 *
 * Revision 1.22  2009-06-23 15:29:22  martins
 * Fixed typo in change log of revision 1.20
 *
 * Revision 1.21  2009-06-23 14:50:37  martins
 * Enhance cdrep:
 * 1) Allow combination of command options
 * The command line option can now be specified in various combinations.
 * e.g. cdrep <config-file> -c -e
 * Except the options -c and -d all other options can be combined.
 *
 * 2) The extended channel stream mode
 * The extended channel stream mode can be switched on with the new option -d.
 * In addition to the output of the channel stream mode the filename and path of
 * the binary file and the offset of the channel from the beginning of
 * the binary file are written to stdout.
 *
 * 3) Remove the option -a
 * The option -a is removed. To get the output of the option -a the option
 * combination -d -e -f has to be specified
 *
 * Revision 1.20  2009-02-23 09:54:45  martins
 * Enhance cdrep: added new option -a
 * Option -a: writes event stream, frame stream and channel stream to stdout
 *
 * Revision 1.19  2008-12-15 11:35:54  martins
 * Fix bug which causes the exit of cdrep in resetChannelLeveLStates().
 * If a channel specific application extension object is not present,
 * it is not to be considered as an error. The channel has to be skipped.
 *
 * Revision 1.18  2008-11-11 15:30:06  martins
 * Added signal handling. SIGINT can be used to initiate an regular shutdown
 * of dcrep.
 * Added syslog message for shutdown.
 *
 * Revision 1.17  2008-11-11 11:40:37  martins
 * If cdrep exits with "EXIT_FAILURE" an error message to syslog is written.
 * Changed the existing debug messages to logWrite().
 *
 * Revision 1.16  2008-10-09 11:02:10  gerald
 * Wrong format character in debug message fixed
 *
 * Revision 1.15  2008-10-06 11:32:15  martins
 * Frame level events:
 * The frame level events are now generated based on an event time window.
 * During this event time window the received frames are counted and classified
 * (timely or late). If the event time window is expired the frame level state is
 * processed according to the current state and the classification of the data
 * received within the expired event time window.
 *
 * Revision 1.14  2008-09-29 11:32:59  martins
 * Fix bug which causes the exit of cdrep because of too many
 * open files.
 * Set pRefListHead->pLastSent while procssing a data format frame.
 *
 * Revision 1.13  2008-09-11 14:46:47  martins
 * Channel level events:
 * added the new feature " initial state after restart"
 *
 * Revision 1.12  2008-08-12 16:18:21  gerald
 * Make site and channel name lengths consistent to wfdisc
 * format and upgrade to new station cache format
 *
 * Revision 1.11  2008/04/07 14:38:17  gerald
 * Fix Solaris compiler warning
 *
 * Revision 1.10  2008/03/20 17:54:24  gerald
 * Typos fixed
 *
 * Revision 1.9  2008/03/19 16:23:10  martins
 * Channel level events:
 * Channel level events for missing channels are triggered after some time of
 * not receiving data from it. Currently the value of the time window for the
 * creation of channel level events is set to 180 seconds.
 *
 * Bandwidth events:
 * The bandwidth events are created after the bandwidth was averaged over some
 * time. The current time span for the creation of bandwidth events is set to
 * 60 seconds.
 *
 * Frame stream mode:
 * In the frame stream mode, the frame time length in seconds is added to the
 * output.
 *
 * Revision 1.8  2007/11/22 12:37:21  gerald
 * Add parameter for requested file creation time to syncFiles()
 *
 * Revision 1.7  2007/10/10 11:12:55  martins
 * Renamed two command line options:
 * -f [frame stream mode] (old value -c)
 * -c [channel stream mode] (old value -s)
 *
 * Revision 1.6  2007/09/14 08:44:11  martins
 * Syslog init changed to "ctbt.acquisition.cdrep".
 * Detection of channel events considers that some
 * stations send the channel data using more than one
 * data frame with the same nominal time.
 *
 * Revision 1.5  2007/08/22 08:12:20  gerald
 * Fix splint warnings
 *
 * Revision 1.4  2007/07/23 13:03:48  martins
 * Added the connection level related bandwidth events
 * SUFFICIENT_BW | INSUFFICIENT_BW.
 * Removed the frame level events TIMELY_IBW | BACKFILLING_IBW |
 * NO_TIMELY_DATA_IBW.
 *
 * Revision 1.3  2007/07/16 12:57:46  martins
 * Use the original frame parsing code from the clf file for the detection
 * of the frame level events.
 * Process all types of frames in the clf stream mode.
 * Change the output of the channel stream mode according to the description.
 *
 * Revision 1.2  2007/07/05 11:57:34  martins
 * Fix bug in the detection of 'insufficient bandwidth".
 * Remove the 'report unchanged state' processing.
 * Fix bug in the cmd line parsing on solaris.
 *
 * Revision 1.1  2007/06/28 08:38:40  gerald
 * Rename report.c cdrep.c
 *
 * Revision 1.7  2007/06/28 07:49:34  gerald
 * Add new event 'No timely data arrived' and fix bug in the end
 * of backfilling detection
 *
 * Revision 1.3  2007/03/12 13:17:27  gerald
 * Add more features to report tool (connect/disconnect
 * event, info events and more)
 *
 * Revision 1.2  2007/02/19 14:32:05  gerald
 * Add generation of .ehl files
 *
 * Revision 1.1  2007/02/12 09:48:43  gerald
 * Start work on event log file
 *
 *
 */

#include <config.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <libgen.h>
#include "cd_common.h"
#include "events.h"
#include "fileAPI.h"
#include "logfiles.h"

#ifdef HAVE_IDCSYSLOG_H
 #include "idcsyslog.h"
#else
 #include "logging.h" /* Deprecated */
#endif

#define TIMELY_DATA_WINDOW 300 /* Data from a station sending continuously
                                  is considered to be timely if it’s received
                                  less than five minutes after data are
                                  recorded at a station.*/
#define NO_DATA_TIMEOUT 60   /* Timeout value for creating the event NO_DATA.
                                This value is sufficient to cover the various
                                Data frame receiving rates of the different
                                types of IMS stations. */
#define CHANNEL_INFO_EVENT_INTERVAL 300  /* Interval in seconds for creating
                                            the Channel Info event */

#define CH_EVENT_WINDOW_SIZE 180 /* time span of the "moving window" for the
                                   creation of channel level events */
#define CH_HISTORY_WINDOW_SIZE 360 /* max time span for the chmask history */

#define BWDTH_EVENT_WINDOW_SIZE 60 /* time span of the "moving window" for the
                                      creation of bandwidth events */
#define FR_EVENT_WINDOW_SIZE 60 /* time span of the "moving window" for the
                                   creation of frame level events */

#define PROTOCOL_TYPE_CD1     "1.0"
#define PROTOCOL_TYPE_CD11    "1.1"

#define LNG_4_CHK_OF_CD11_STATUS 5  /* for CD1.1 the first 5 Bytes of the
                                     * status information shall be checked */

/**
 * command line arguments of the report tool
 */
#define CLI_MIN_ARGUMENTS     2
#define CLI_MAX_ARGUMENTS     8
#define CLI_START_OF_OPTIONS  3

/**
 * run modes of the report tool
 */
#define     NO_OPTION_SET              0
#define     CR_EHL_FILES               (uint16_t)(1<<0)
#define     EV_STREAM_2_STDOUT         (uint16_t)(1<<1)
#define     CLF_INFO_2_STDOUT          (uint16_t)(1<<2)
#define     CH_INFO_2_STDOUT           (uint16_t)(1<<3)
#define     CH_INFO_EXT1_2_STDOUT      (uint16_t)(1<<4)
#define     WR_STDOUT_2_EHL_FILES      (uint16_t)(1<<5)

/**
 * error codes of run mode detection
 */
#define     RUN_MODE_OK                 0
#define     ERR_NO_CFG_FILE            -1
#define     ERR_TO_MANY_ARGS           -2
#define     ERR_INV_CFG_FILE           -3
#define     ERR_INV_OPTION             -4
#define     ERR_INV_OPTION_FORMAT      -5
#define     ERR_INV_OPTION_COMBINATION -6

/**
 * frame level events, which do not cause change in the frame level state
 */
#define         DATA_FRAME_STATUS_CHANGED    ((unsigned char)'w')

/**
 * channel level events, which do not cause change in the channel level state
 */
#define         CHANNEL_INFO                 ((unsigned char)'i')
#define         CHANNEL_STATUS_CHANGED       ((unsigned char)'s')

/**
 * initial state for bandwith and channel events
 */
#define         INITIAL                      ((unsigned char)'1')

/**
 * defines for bandwidth event generation
 */
#define     BDWDTH_EVENTS_PROCESSED         0
#define     BDWDTH_EVENTS_NOT_PROCESSED     1
#define     BDWDTH_WINDOW_NOT_EXPIRED       2
#define     BDWDTH_WINDOW_EXPIRED           3

/**
 * check types for unlinking channel mask cache elements
 */
#define     CHK_TYPE_CLF_TIME       0
#define     CHK_TYPE_NOW_TIME       1

/**
 * frame level event timer states
 */
#define     FRLVL_NO_TIMER          0
#define     FRLVL_TIMER_ACTIVE      1
#define     FRLVL_TIMER_EXPIRED     2

/**
 * frame level event data classification
 */
#define     DC_UNDEFINED                  0
#define     DC_ONLY_TIMELY_FRAMES         1
#define     DC_ONLY_LATE_FRAMES           2
#define     DC_TIMELY_AND_LATE_FRAMES     3


/**
 * structure for a channel mask item
 *
 * The channel mask item data structure contains the channel mask value,
 * the clf time of the first frame with this nominal time, the frame time
 * length in seconds, the validation check result of the channel,
 * the channel status size and the channel status data.
 */

typedef struct _chMaskAndStatus {
   char        channelMaskItem; /**< channel mask item for a particular
                                     channel */
   double      clfTime; /**< clf time of the first frame with this nominal
                             time. */
   double      frTimeLngSec; /**< frame time length in seconds */
   uint32_t    checkPass; /**< channel has passed validation check */
   uint32_t    chStatusSize; /**< CD1: 4; CD1.1: 5 */
   char        chStatusData[LNG_4_CHK_OF_CD11_STATUS]; /**< status data */
} chMaskAndStatus_t;


/**
 * structure for a channel mask history entry
 *
 * The channel mask history entry data structure contains the nominal time,
 * the clf time of the first frame received with this nominal time, the now
 * time value when cdrep has detected the first frame with this nominal time,
 * the event processing state of this entry, the channel mask and status of all
 * channels contained within frame with this nominal time and the pointer to
 * the next channel mask history entry.
 */

typedef struct _chMaskHistory {
   double         nominalTime;   /**< nominal time which is corresponding
                                      to the channel data */
   double         clfTime;       /**< clf time of the first frame with this
                                      nominal time. */
   double         nowTime;       /**< at that time the first frame with this
                                      nominal time was detected by the
                                      reporting tool. */
   int            evProcessingState;
#define CH_EVENTS_PROCESSED 0
#define CH_EVENTS_NOT_PROCESSED 1
   chMaskAndStatus_t chMaskAndStatus[MAX_CHANNELS]; /**< channel mask and
                                 channel status data for this nominal time */
   struct _chMaskHistory *nextChMaskHistory; /**< pointer to next chmask
                                                  history element */
} chMaskHistory_t;


/**
 * structure for a station specific channel mask history cache
 *
 * The channel mask history cache data structure contains the number of channel
 * mask history elements, and the pointer to the first channel mask history
 * element.
 */

typedef struct _channelHistoryCache {
   uint32_t       nrOfHistoryElements;   /**< number of channel mask history
                                              elements */
   struct _chMaskHistory *firstChMaskHistory; /**< pointer to first chmask
                                                   history element */
} channelHistoryCache_t;


/**
 * structure for a frame level event log data entry
 *
 * The frame level event log data structure contains the frame level state,
 * the frame reception time of the last data frame received, the frame parsing
 * code of this frame, the nominal time, clf time and fpc of a frame which will
 * raise an event, the states & values for the bandwidth state,
 * the time of the last 'i' event (channel info), the pointer to the channel
 * history cache and the nominal time use for the next channel event.
 */

typedef struct _frLvlEvLogData {
   unsigned char  frLvlState;  /**< frame level state, values see below.
                                    States which are represented by
                                    a char will raise an event */
#define         UNKNOWN_WAIT                 ((unsigned char)'1')
#define         TIMELY                       ((unsigned char)'t')
#define         TIMELY_WAIT                  ((unsigned char)'2')
#define         BACKFILLING                  ((unsigned char)'b')
#define         BACKFILLING_WAIT             ((unsigned char)'3')
#define         LATE                         ((unsigned char)'o')
#define         LATE_WAIT                    ((unsigned char)'4')
#define         NO_DATA                      ((unsigned char)'N')
   double         dataRcvTime; /**< Data Frame reception time for event log.
                                    Necessary to check for NO_DATA events */
   uint32_t       fpc;         /**< Frame parsing code, for event log.
                                    Necessary to check if the data frame
                                    status has changed */
   double         frLvlEvClfTime; /**< clf time to be used for the next frame
                                       level event */
   double         frLvlEvNomTime; /**< nominal time to be used for the next
                                       frame level event */
   uint32_t       frLvlEvFpc; /**< frame parsing code to be used for the next
                                   frame level event */
   double         frLvlEvTimer; /**< start time of the timer for frame level
                                     events */
   uint32_t       nrOfTimelyFramesArrived; /**< counter of the number of timely
                                              frames, arrived during the current
                                              timer period */
   uint32_t       nrOfLateFramesArrived; /**< counter of the number of timely
                                            frames, arrived during the current
                                            timer period */
   /* for bandwidth events */
   unsigned char bandwdthState;  /**< bandwidth state:
                                      note: this is a value related to the
                                      connection level. To avoid to introduce
                                      new connection level data it is stored
                                      in the frame level related data */
#define         SUFFICIENT_BW               ((unsigned char)'a')
#define         INSUFFICIENT_BW             ((unsigned char)'A')
   unsigned char  bandwdthFirstFrame;  /**< First frame seen ? */
#define         BW_FIRST_FRAME_NOT_SEEN   ((unsigned char)'0')
#define         BW_FIRST_FRAME_SEEN       ((unsigned char)'1')
   double   bandwdthClfTimeSpanTotal;   /**< total of the frame reception time
                                             of the last N data frames, received
                                             withing the current bandwidth event
                                             creation window */
   double   bandwdthFrameTimeTotal; /**< total of the frame time length of the
                                         last N data frames, received withing
                                         the current bandwidth event creation
                                         window */
   double   bandwdthClfTime;  /**< clf time of the first data frame of the
                                   current bandwidth event creation window*/
   double   bandwdthNowTime;  /**< now time of the first data frame of the
                                   current bandwidth event creation window*/
   uint32_t bandwdthNrOfFrames;/**< number of frames received within the
                                   current bandwidth event creation window*/
   /* for channel level events */
   struct _channelHistoryCache chHistoryCache; /**< channel history cache */
   double         nominalTime4ChEvents; /**< nominal Time for event log */
   double         chInfoEventTime; /**< channel Info event time.
                                        Necessary to check if the timer to create
                                        channel info events is expired */
   /* statistics */
   uint32_t       dataFramesReceived;  /**< number of received data frames */
} frLvlEvLogData_t;


/**
 * structure for a channel level event log data  entry
 *
 * The channel level event log data structure contains the channel level state,
 * the channel parsing code, the earliest data time, the latest data time,
 * channel status size, and a pointer to the channel status data.
 */

typedef struct _chLvlEvLogData {
   unsigned char  chState;  /**< event log channel state, values see below */
   unsigned char  newChState;  /**< new event log channel state,
                                    values see below */
#define         CHANNEL_PRESENT     ((unsigned char)'c')
#define         CHANNEL_MISSING     ((unsigned char)'C')
   int            chStatusChanged;  /**< checkPass or Status changed,
                                         values see below */
#define CH_STATUS_UNCHANGED      0
#define CH_PARSING_CODE_CHANGED  1
#define CH_STATUS_SIZE_CHANGED   2
#define CH_STATUS_DATA_CHANGED   3
#define CH_STATUS_NOT_SAVED      4
   int            updateChDataTimes; /**< This switch is used to skip the update
                                          of the channel data times in some
                                          special cases */
#define CH_DO_DATA_TIMES_UPD     0
#define CH_SKIP_DATA_TIMES_UPD   1
   uint32_t       checkPass;   /* channel parsing code */
   uint32_t       chStatusSize;  /* Always 4 for CD1 */
   /*@null@*/
   char           *chStatusData;
   double         earliestNominalTime; /* earliest data time of the channel */
   double         latestNominalTime; /* latest data time of the channel */
   double         lastFrameRecNominalTime; /* last data time of the channel,
                                           taken from the last frame received */
   double         eventClfTime; /* clf time of the channel, taken from the frame
                                   which contains the channel, or the last frame
                                   received which has the same nominal time.
                                   (if channel is missing) */
   double frameTimeLengthSec; /* time in seconds the last frame received which
                                 contained this channel encompasses */
   double nomTimeOfExpiredEvWindow; /* the nominal time of the (n) data frames
                                       of the expired channel level event
                                       window */
} chLvlEvLogData_t;


/**
 * prototypes
 */
static void
unlinkAllChMaskHistoryElements(frLvlEvLogData_t *pFrLvlData);
static void
updateChannelDataTimes(station_t *pStation,
                       frLvlEvLogData_t *pFrLvlData);
static void
updateChannelData(station_t *pStation,
                  chMaskHistory_t *pChMaskHistoryElem,
                  frLvlEvLogData_t *pFrLvlData,
                  double   evTime4MissingChannels);
static void
checkNewFrameLevelStateAndRaiseEvent(station_t *pStation,
                                     frLvlEvLogData_t *pFrLvlData,
                                     unsigned char newFrLvlState,
                                     uint16_t runMode);
static void
resetDataClassification(frLvlEvLogData_t *pFrLvlData);
static int
getDataClassificationSummary(frLvlEvLogData_t *pFrLvlData);

/**
 * @short init event log frame level extension object
 *
 * @description This function initializes the data structure
 * @description used to store event log frame level related
 * @description data. The memory to store the data is allocated
 * @description and the data pointer is returned to the calling
 * @description function.
 *
 * @return pointer to the event log frame level extension object, or
 * @return NULL in case of error.
 */

/*@null@*/static frLvlEvLogData_t *
initEvLogFrLvlData(void)
{
   frLvlEvLogData_t *pFrLvlData;

   pFrLvlData = calloc(1, sizeof(frLvlEvLogData_t));
   if (NULL == pFrLvlData)
      return (NULL);

   /* initialize frame level state */
   pFrLvlData->frLvlState = UNKNOWN_WAIT;

   /* initialize bandwidth state */
   pFrLvlData->bandwdthState = INITIAL;

   /* first data frame not seen yet */
   pFrLvlData->bandwdthFirstFrame = BW_FIRST_FRAME_NOT_SEEN;

   return (pFrLvlData);
}

/**
 * @short reset event log frame level extension object
 *
 * @description This function perform the reset of the data structure
 * @description used to store event log frame level related
 * @description data. The memory to store the data is allocated
 * @description and the data pointer is returned to the calling
 * @description function.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrLvlData     pointer to the current frame level data structure
 *
 * @return nothing
 */

static void
resetEvLogFrLvlData(station_t *pStation,
                   /*@null@*/ frLvlEvLogData_t *pFrLvlData)
{
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* delete the elements of the channel mask history cache
    * because the frame level data structure will be reset now */
   unlinkAllChMaskHistoryElements(pFrLvlData);

   /* reset all the values of the frame level structure */
   (void) memset(pFrLvlData, 0, sizeof(frLvlEvLogData_t));

   /* set the frame level data to INITIAL */
   pFrLvlData->frLvlState = UNKNOWN_WAIT;

   /* initialize bandwidth state */
   pFrLvlData->bandwdthState = INITIAL;

   /* first data frame not seen yet */
   pFrLvlData->bandwdthFirstFrame = BW_FIRST_FRAME_NOT_SEEN;

   return;
}

/**
 * @short init event log extension object of particular channel
 *
 * @description This function initializes the data structure
 * @description used to store event log channel level related
 * @description data for a particular channel.
 * @description The memory to store the data is allocated
 * @description and the data pointer is returned to the calling
 * @description function.
 *
 * @return pointer to the event log  extension object of particular channel,
 * @return or NULL in case of error.
 */

/*@null@*/static chLvlEvLogData_t *
initEvLogChLvlData(void)
{
   chLvlEvLogData_t *pChLvlData;

   pChLvlData = calloc(1,sizeof(chLvlEvLogData_t));
   if (NULL == pChLvlData)
      return (NULL);

   pChLvlData->chState = INITIAL;
   pChLvlData->newChState = INITIAL;
   pChLvlData->chStatusChanged = CH_STATUS_NOT_SAVED;

   return (pChLvlData);
}


/**
 * @short reset event log extension object of particular channel
 *
 * @description This function resets the data structure
 * @description used to store event log channel level related
 * @description data for a particular channel.
 *
 * @param pStation       pointer to the current station structure
 * @param pChLvlData     pointer to the current channel level data structure
 *
 * @return nothing
 */

static void
resetEvLogChLvlData(station_t *pStation,
                    /*@null@*/ chLvlEvLogData_t *pChLvlData)
{
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pChLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pChLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   pChLvlData->chState = INITIAL;
   pChLvlData->newChState = INITIAL;
   pChLvlData->chStatusChanged = CH_STATUS_NOT_SAVED;
   pChLvlData->updateChDataTimes = CH_DO_DATA_TIMES_UPD;
   pChLvlData->checkPass = 0;
   pChLvlData->earliestNominalTime = 0;
   pChLvlData->latestNominalTime = 0;
   pChLvlData->lastFrameRecNominalTime = 0;
   pChLvlData->eventClfTime = 0;
   pChLvlData->frameTimeLengthSec = 0;
   pChLvlData->nomTimeOfExpiredEvWindow = 0;

   /* note: chStatusSize and *chStatusData are not reset, because it does not
    * make sense to free the memory and allocate it again.
    * CH_STATUS_NOT_SAVED denotes, that there is no channel status data saved.
    */

   return;
}


/**
 * @short reset event log channel level states of all channels
 *
 * @description This function loops through the entrys
 * @description of the channel name cache and resets the channel
 * @description level data structure of each particular channel.
 *
 * @param pStation       pointer to the current station structure
 *
 * @return nothing
 */

static void
resetChannelLeveLStates(station_t *pStation)
{
   chName_t *actChEntry;
   chNameCache_t *chNameCache;
   chLvlEvLogData_t *pChLvlData;

   /* check the impossible */
   if (NULL == pStation)
   {
      logError("Internal application error", "NULL == pStation");
      exit(EXIT_FAILURE);
   }

   chNameCache = &(pStation->chCache);

   /* loop through the entries of the stations channel name cache */
   for (actChEntry = chNameCache->firstCh; actChEntry != NULL;
        actChEntry = actChEntry->nextChannel)
   {
      pChLvlData = (chLvlEvLogData_t *)actChEntry->chCacheExt;

      /* check if channel specific application extension object is existing */
      if (NULL == (pChLvlData))
      {
         /* No channel specific application extension object found.
          * Skip the channel. This is possible if no timely frame has been
          * seen, since the connection was established.
          * The channel specific application extension object will be allocated
          * during the channel level processing of a timely data frame.
          * If no timely frame has been seen, there is no channel specific
          * application extension object.*/
         DEBUG3(5,"no extension object found: channel skipped [%.17s] %.17s",
                pStation->conID, actChEntry->scName[0]);

         continue;
      }

      /* reset the channel specific event log data structure */
      resetEvLogChLvlData(pStation,
                          pChLvlData);
   }
   return;
}


/**
 * @short create new channel mask history element
 *
 * @description This function creates a new channel mask history element.
 * @description The channel mask history element is used to store
 * @description data of all channels for a particular nominal time.
 * @description The memory to store the data is allocated, the number of
 * @description history elements is increased and the data pointer is
 * @description returned to the calling function.
 *
 * @param chHistoryCache   pointer to the channel history cache
 *
 * @return pointer to the new channel mask history element
 * @return or NULL in case of error.
 */

/*@null@*/static chMaskHistory_t *
newChMaskHistoryElement(channelHistoryCache_t *chHistoryCache)
{
   chMaskHistory_t *pChMaskHistory;
   uint32_t ii;

   pChMaskHistory = calloc(1,sizeof(chMaskHistory_t));
   if (NULL == pChMaskHistory)
      return (NULL);

   pChMaskHistory->evProcessingState = CH_EVENTS_NOT_PROCESSED;

   /* set chmask items to "CHANNEL MISSING" */
   for (ii = 0; ii < MAX_CHANNELS; ii++)
   {
      pChMaskHistory->chMaskAndStatus[ii].channelMaskItem = '0';
   }

   /* increase number of elements in the cache */
   chHistoryCache->nrOfHistoryElements++;

   DEBUG2(5,"newChMaskHistory: nrOfHistoryElements: %d",
          chHistoryCache->nrOfHistoryElements);

   return (pChMaskHistory);
}


/**
 * @short delete channel mask history element
 *
 * @description This function deletes a channel mask history element.
 * @description The memory of this element is freed and the number of
 * @description history elements in the cache is decreased.
 *
 * @param chHistoryCache   pointer to the channel history cache
 * @param pChMaskHistElem  pointer to the current channel mask history element
 *
 * @return nothing
 */

static void
delChMaskHistoryElement(channelHistoryCache_t *chHistoryCache,
                        chMaskHistory_t *pChMaskHistElem)
{
   /* check the impossible */
   if (NULL == pChMaskHistElem)
   {
      logError("Internal application error", "pChMaskHistElem == NULL");
      exit(EXIT_FAILURE);
   }

   DEBUG2(5,"delChMaskHistoryElement: nominalTime %f",
          pChMaskHistElem->nominalTime);

   free (pChMaskHistElem);

   /* decrease number of elements in the cache */
   chHistoryCache->nrOfHistoryElements--;

   DEBUG2(5,"delChMaskHistoryElement: nrOfHistoryElements: %d",
          chHistoryCache->nrOfHistoryElements);
}


/**
 * @short store the present channels into the chmask history element
 *
 * @description This function stores information on all channels which are
 * @description present in the current data frame. The channel present sign,
 * @description frame reception time and the frame time length in seconds
 * @description are stored in the channel mask history element which is
 * @description corresponding to the nominal time of the current data
 * @description frame.
 * @description If the current data frame will trigger channel level events
 * @description the stored frame reception time will be the event time.
 *
 * @param pChMaskHistElem  pointer to the current channel mask history element
 * @param pChMask          pointer to the channel mask of the current frame
 * @param frameTimeLength  time in ms this frame encompasses
 * @param clftime          frame reception time of the current frame
 *
 * @return nothing
 */

static void
storePresentChannels(chMaskHistory_t *pChMaskHistElem,
                     char *pChMask,
                     uint32_t frameTimeLength,
                     double clftime)
{
   uint32_t ii;
   double frameTimeLengthSec;

   /* compute time in seconds, this frame encompasses */
   frameTimeLengthSec = (double) (frameTimeLength / 1000);

   /* check the impossible */
   if (NULL == pChMaskHistElem)
   {
      logError("Internal application error", "pChMaskHistElem == NULL");
      exit(EXIT_FAILURE);
   }

   /* check the impossible */
   if (NULL == pChMask)
   {
      logError("Internal application error", "pChMask == NULL");
      exit(EXIT_FAILURE);
   }

   /* store the information on the present channels of this data frame */
   for (ii = 0; ii < MAX_CHANNELS; ii++)
   {
      /* if a channel is present the value in the channel mask is set to '1'
       * and the store the frame reception time (necessary to create the
       * channel level event */
      if ('1' == pChMask[ii])
      {
         pChMaskHistElem->chMaskAndStatus[ii].channelMaskItem = '1';
         pChMaskHistElem->chMaskAndStatus[ii].clfTime = clftime;
         pChMaskHistElem->chMaskAndStatus[ii].frTimeLngSec =
                                                 frameTimeLengthSec;
      }
   }
}


/**
 * @short store status data of all present channels
 *
 * @description This function loops through the channels of the current frame
 * @description and stores the status information of all present channels
 * @description in the channel mask history element which is
 * @description corresponding to the nominal time of the current data frame.
 *
 * @param pStation         pointer to the current station structure
 * @param pChMaskHistElem  pointer to the current channel mask history element
 *
 * @return nothing
 */

static void
storeStatusDataOfAllPresentChannels(station_t *pStation,
                                    chMaskHistory_t *pChMaskHistElem)
{
   uint32_t ii;
   uint32_t maskPos;
   const CDdataChannel_t *channel;  /* Channel Data */
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* check the impossible */
   if (NULL == pChMaskHistElem)
   {
      logError("Internal application error", "pChMaskHistElem == NULL");
      exit(EXIT_FAILURE);
   }

   channel = pStation->DFinfo->channel;

   /* loop through the channels of the current frame and store the
    * status information of all present channels */
   for (ii = 0; ii < pStation->DFinfo->parsedChannels; ii++)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
            "Site/Channel[%.6s/%.8s] posInChMask[%d] chId[%d] parsedCh[%d]",
                      channel->siteName,
                      channel->channelName,
                      channel->posInChMask,
                      ii,
                      pStation->DFinfo->parsedChannels);
      DEBUG(6, logBuffer);

      /* Check index first */
      if (MAX_CHANNELS <= channel->posInChMask)
      {
         (void) snprintf(logBuffer, sizeof(logBuffer),
               "Site/Channel[%.6s/%.8s] posInChMask[%d] chId[%d] parsedCh[%d]",
                         channel->siteName,
                         channel->channelName,
                         channel->posInChMask,
                         ii,
                         pStation->DFinfo->parsedChannels);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
      }
      /* Check if invalid Site/channel name.
       * Skip the channel if the mask position has the value -1 */
      else if (-1 == channel->posInChMask)
      {
         (void) snprintf(logBuffer, sizeof(logBuffer),
               "Site/Channel[%.6s/%.8s] posInChMask[%d] chId[%d] parsedCh[%d]",
                         channel->siteName,
                         channel->channelName,
                         channel->posInChMask,
                         ii,
                         pStation->DFinfo->parsedChannels);
         logWarning(
               "storeStatusDataOfAllPresentChannels: skip invalid channel",
               logBuffer);

         channel++;
         continue;
      }

      maskPos = channel->posInChMask;
      /* determine the size of the channel status data to be stored.
       * For CD1 use the value of channel->channelStatusSize.
       * For CD1.1 use LNG_4_CHK_OF_CD11_STATUS.
       * For safety reasons, check if the status sizes exceeds
       * LNG_4_CHK_OF_CD11_STATUS. */
      if ((CD1_1 == pStation->type) &&
          (LNG_4_CHK_OF_CD11_STATUS <= channel->channelStatusSize))
         pChMaskHistElem->chMaskAndStatus[maskPos].chStatusSize =
                                          LNG_4_CHK_OF_CD11_STATUS;
      /* CD1 or channel->channelStatusSize less than LNG_4_CHK_OF_CD11_STATUS */
      else
         pChMaskHistElem->chMaskAndStatus[maskPos].chStatusSize =
                                           channel->channelStatusSize;

      /* save channel status data */
      memcpy(pChMaskHistElem->chMaskAndStatus[maskPos].chStatusData,
             channel->channelStatus,
             (size_t)pChMaskHistElem->chMaskAndStatus[maskPos].chStatusSize);

      /* save channel parsing code */
      pChMaskHistElem->chMaskAndStatus[maskPos].checkPass = channel->checkPass;

      channel++;
   }
}


/**
 * @short get chmask history element with expired window by clf time
 *
 * @description This function loops through the elements of the channel mask
 * @description history cache. If a channel mask history element is already
 * @description unprocessed the size of the event creation window is checked.
 * @description The check is performed using the clf time of the history element
 * @description and the clf time of the current data frame.
 * @description If the event creation window for the particular element is
 * @description expired, the pointer to the channel mask history element is
 * @description returned to the calling function.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param clfTime        clf time of the current data frame
 *
 * @return pointer to the expired channel mask history element or NULL if no
 * @return expired channel mask history element found.
 */

/*@null@*/static chMaskHistory_t *
getExpiredWindowByClfTime(frLvlEvLogData_t *pFrLvlData,
                          double clfTime)
{
   chMaskHistory_t *actElement;
   channelHistoryCache_t *chHistoryCache;
   double delta_t;

   DEBUG(5,"getExpiredWindowByClfTime");

   chHistoryCache = &(pFrLvlData->chHistoryCache);

   /* search for an unprocessed  channel mask history element with an
    * expired event creation window */
   for (actElement = chHistoryCache->firstChMaskHistory; actElement != NULL;
        actElement = actElement->nextChMaskHistory)
   {
      /* check if the current history element is still unprocessed */
      if (CH_EVENTS_PROCESSED == actElement->evProcessingState)
         continue;

      /* compute the current event window size for this element */
      delta_t = clfTime - actElement->clfTime;

      DEBUG5(5,"chkEventWindow: nominal time %f:  %f - %f clfTimeDelta_t: %f",
             actElement->nominalTime,
             clfTime,
             actElement->clfTime,
             delta_t);

      /* check if the event creation window for this history element
       * is expired */
      if (CH_EVENT_WINDOW_SIZE < delta_t)
         return actElement;
   }
   return NULL;
}


/**
 * @short get chmask history element with expired window by now time
 *
 * @description This function loops through the elements of the channel mask
 * @description history cache. If a channel mask history element is already
 * @description unprocessed the size of the event creation window is checked.
 * @description The check is performed using the now time of the history element
 * @description and the current time.
 * @description If the event creation window for the particular element is
 * @description expired, the pointer to the channel mask history element is
 * @description returned to the calling function.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param nowTime        current time
 *
 * @return pointer to the expired channel mask history element or NULL if no
 * @return expired channel mask history element found.
 */

/*@null@*/static chMaskHistory_t *
getExpiredWindowByNowTime(frLvlEvLogData_t *pFrLvlData,
                          double nowTime)
{
   chMaskHistory_t *actElement;
   channelHistoryCache_t *chHistoryCache;
   double delta_t;

   DEBUG(5,"getExpiredWindowByNowTime");

   chHistoryCache = &(pFrLvlData->chHistoryCache);

   /* search for an unprocessed  channel mask history element with an
    * expired event creation window */
   for (actElement = chHistoryCache->firstChMaskHistory; actElement != NULL;
        actElement = actElement->nextChMaskHistory)
   {
      /* check if the current history element is still unprocessed */
      if (CH_EVENTS_PROCESSED == actElement->evProcessingState)
         continue;

      /* compute the current event window size for this element */
      delta_t = nowTime - actElement->nowTime;

      DEBUG5(5,"chkEventWindow: nominal time %f: %f - %f nowTimeDelta_t: %f",
             actElement->nominalTime,
             nowTime,
             actElement->nowTime,
             delta_t);

      /* check if the event creation window for this history element
       * is expired */
      if (CH_EVENT_WINDOW_SIZE < delta_t)
         return actElement;
   }
   return NULL;
}


/**
 * @short get the next unprocessed chmask history element
 *
 * @description This function loops through the elements of the channel mask
 * @description history cache. If a channel mask history element is already
 * @description unprocessed the pointer to the channel mask history element is
 * @description returned to the calling function.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 *
 * @return Pointer to the next unprocessed channel mask history element
 * @return or NULL if no unprocessed channel mask history element found.
 */

/*@null@*/static chMaskHistory_t *
getUnprocessedChMaskHistoryELement(frLvlEvLogData_t *pFrLvlData)
{
   chMaskHistory_t *actElement;
   channelHistoryCache_t *chHistoryCache;

   chHistoryCache = &(pFrLvlData->chHistoryCache);

   /* search for an unprocessed  channel mask history element. Don't care about
    * the event processing window */
   for (actElement = chHistoryCache->firstChMaskHistory; actElement != NULL;
        actElement = actElement->nextChMaskHistory)
   {
      /* check if the current history element is still unprocessed */
      if (CH_EVENTS_PROCESSED == actElement->evProcessingState)
         continue;

      DEBUG2(5,"unprocessed ChMaskHistoryELem found: nominal time %f",
             actElement->nominalTime);

      return actElement;
   }
   return NULL;
}


/**
 * @short unlink all processed and expired channel mask history elements
 *
 * @description This function loops through the elements of the channel mask
 * @description history cache. If a channel mask history element is already
 * @description processed the size of the channel mask history window is
 * @description checked.
 * @description The check is performed using either the clf time value or the
 * @description now time value of the history element.
 * @description If the channel mask history window for the particular element is
 * @description expired, the pointer to the channel mask history element is
 * @description unlinked from the channel mask history cache. The data of the
 * @description unlinked channel mask history element is freed.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param checkType      clf time or now time
 * @param checkTime      time value for the check of expired history windows
 *
 * @return nothing
 */

static void
unlinkExpiredChMaskHistoryElements(frLvlEvLogData_t *pFrLvlData,
                                   uint16_t  checkType,
                                   double checkTime)
{
   channelHistoryCache_t *chHistoryCache;
   chMaskHistory_t *actElement,
                   *previousElement = NULL,
                   *nextElement = NULL;
   double delta_t;
   char logBuffer[LOGSIZE];

   chHistoryCache = &(pFrLvlData->chHistoryCache);

   DEBUG2(5,"unlinkExpiredChMaskHistoryElements: nr of elements in cache: %d",
          chHistoryCache->nrOfHistoryElements);

   /* search for processed  channel mask history elements with an
    * expired event channel mask history window. Delete these elements. */
   for (actElement = chHistoryCache->firstChMaskHistory; actElement != NULL;
        actElement = nextElement)
   {
      /* save the pointer to the next element in the cache */
      nextElement = actElement->nextChMaskHistory;

      DEBUG2(5,"next element: %lx", (unsigned long)nextElement);

      /* check if the current history element is already processed */
      if (CH_EVENTS_NOT_PROCESSED == actElement->evProcessingState)
      {
         /* save the pointer to the previous element in the list */
         previousElement = actElement;
         continue;
      }

      DEBUG2(5,"check type: %d", checkType);

      /* compute window size, depending on the specified check type */
      if (CHK_TYPE_CLF_TIME == checkType)
      {
         /* compute the current history windows size for this element */
         delta_t = checkTime - actElement->clfTime;

         DEBUG5(5,"chkHistoryWindow: nominal time %f: %f - %f clfTimeDelta_t: %f",
                actElement->nominalTime,
                checkTime,
                actElement->clfTime,
                delta_t);
      }
      else if (CHK_TYPE_NOW_TIME == checkType)
      {
         /* compute the current history windows size for this element */
         delta_t = checkTime - actElement->nowTime;

         DEBUG5(5,"chkHistoryWindow: nominal time %f: %f - %f nowTimeDelta_t: %f",
                actElement->nominalTime,
                checkTime,
                actElement->nowTime,
                delta_t);
      }
      else
      {
         /* error! unknown check type */
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "unknown checkType: %d",
                         checkType);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
      }

      /* check if this history element is out of the boundaries of the
       * channel mask history size */
      if (CH_HISTORY_WINDOW_SIZE > delta_t)
      {
         /* save the pointer to the previous element in the list */
         previousElement = actElement;
        continue;
      }

      /* unlink the current chmask history element from the cache.
       * Note that it has to be checked if the current element is the first
       * element of the list  */
      if (chHistoryCache->firstChMaskHistory == actElement)
      {
         DEBUG2(5,"deleting first element: %lx", (unsigned long)actElement);

         /* unlink the current element for deletion, the next elemt in the
          * cache is now the new first element */
         chHistoryCache->firstChMaskHistory = actElement->nextChMaskHistory;

         DEBUG2(5,"new first element: %lx",
               (unsigned long)chHistoryCache->firstChMaskHistory);
      }
      else
      {
         DEBUG2(5,"delete element: %lx", (unsigned long)actElement);
         if (NULL == previousElement)
         {
            /* an cache element, which is not the first element must
             * have a predecessor */
            (void) snprintf(logBuffer, sizeof(logBuffer),
                            "previousElement == NULL, actElement = %lx",
                            (unsigned long)actElement);
            logError("Internal application error", logBuffer);
            exit(EXIT_FAILURE);
         }
         /* unlink the current element for deletion */
         previousElement->nextChMaskHistory = actElement->nextChMaskHistory;
      }

      /* delete the current channel mask history element */
      delChMaskHistoryElement(chHistoryCache, actElement);
   }
}


/**
 * @short unlink all channel mask history elements
 *
 * @description This function loops through the elements of the channel mask
 * @description history cache.
 * @description For each element of the channel mask histoy cache
 * @description the pointer to the channel mask history element is
 * @description unlinked from the channel mask history cache. The data of the
 * @description unlinked channel mask history element is freed.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 *
 * @return nothing
 */

static void
unlinkAllChMaskHistoryElements(frLvlEvLogData_t *pFrLvlData)
{
   chMaskHistory_t *actElement,
                   *nextElement = NULL;
   channelHistoryCache_t *chHistoryCache;

   chHistoryCache = &(pFrLvlData->chHistoryCache);

   DEBUG2(5,"unlinkAllChMaskHistoryElements: nr of elements in cache: %d",
          chHistoryCache->nrOfHistoryElements);

   /* delete all channel mask history elements */
   for (actElement = chHistoryCache->firstChMaskHistory; actElement != NULL;
        actElement = nextElement)
   {
      /* save the pointer to the next element in the cache */
      nextElement = actElement->nextChMaskHistory;

      /* delete channel mask history element */
      delChMaskHistoryElement(chHistoryCache, actElement);
   }
}


/**
 * @short store bandwidth data of the current frame
 *
 * @description This function stores the bandwidth data of the current data frame.
 * @description The clfTimeSpan of the current frame is added to clfFtimeTotal
 * @description value for the current bandwidth window. The fram time length in
 * @description seconds is added to the frame time length total for the current
 * @description bandwitdth window. These values are needed to compute the averaged
 * @description bandwith rate.
 *
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param pFrameDes        pointer to the clf file line, stored in an irl
 *                         entry
 * @param frameTimeLength  time in ms this frame encompasses
 *
 * @return noting
 */

static void
storeBandwidthData(frLvlEvLogData_t *pFrLvlData,
                   intRefListParams_t *pFrameDes,
                   uint32_t frameTimeLength)
{
   double frameTimeLengthSec;

   frameTimeLengthSec = (double) (frameTimeLength / 1000);

   DEBUG3(5,"frameTimeLengthSec: %f clfTimeSpan: %f",
          frameTimeLengthSec,
          pFrameDes->clfTimeSpan);

   /* add values of the current frame, to the current total */
   pFrLvlData->bandwdthClfTimeSpanTotal += pFrameDes->clfTimeSpan;
   pFrLvlData->bandwdthFrameTimeTotal += frameTimeLengthSec;

   DEBUG3(5,"bandwdthFrameTimeTotal: %f bandwdthClfTotal: %f ",
          pFrLvlData->bandwdthFrameTimeTotal,
          pFrLvlData->bandwdthClfTimeSpanTotal);
}


/**
 * @short Compute bandwidth state
 *
 * @description This function computes the bandwidth state for the station.
 * @description During the time span of bandwidth event window the bandwitdh
 * @description data of all data frames received was added.
 * @description The bandwidth state of the station represents the averaged
 * @description bandwidth of all data frames received in the expoired
 * @description bandwidth event window.
 * @description The bandwidth is sufficient if the averaged clf time span is
 * @description lower than the half of the averaged frame time length.
 *
 * @param pFrLvlData       pointer to the current frame level data structure
 *
 * @return bandwidth state SUFFICIENT_BW | INSUFFICIENT_BW
 */

static unsigned char
computeBandwidthState(frLvlEvLogData_t *pFrLvlData)
{
   double avFrameTimeLength;
   double avClfTimeSpan;

   /* compute average value of frame time length and clfTimeSpan */
   avFrameTimeLength = (double) (pFrLvlData->bandwdthFrameTimeTotal /
                                             pFrLvlData->bandwdthNrOfFrames);

   avClfTimeSpan = (double) (pFrLvlData->bandwdthClfTimeSpanTotal /
                                             pFrLvlData->bandwdthNrOfFrames);

   DEBUG3(5,"avFrameTimeLength: %f avClfTimeSpan: %f",
          avFrameTimeLength,
          avClfTimeSpan);

   /* check if there was sufficient bandwidth to receive data. */
   if ((long) avClfTimeSpan > (long)(avFrameTimeLength / 2))
      return(INSUFFICIENT_BW);
   else
      return(SUFFICIENT_BW);
}


/**
 * @short check if the current frame is a timely frame
 *
 * @description This function checks if the current frame was recieved
 * @description in a timely manner.
 * @description Timely data must be received less than five minutes
 * @description after the data are recorded at a station.
 * @description Frames which do not fulfill the condition are classified
 * @description as "late".
 *
 * @param pFrameDes        pointer to the clf file line, stored in an irl
 *                         entry
 * @param timelyDataWindow max delay in seconds for a frame to be timely
 *
 * @return frame level state TIMELY | LATE
 */

static unsigned char
checkIfDataFrameIsTimely(intRefListParams_t *pFrameDes,
                         long timelyDataWindow)
{
   double clfTimeMax;

   /* compute latest clf time for the current data frame */
   clfTimeMax = pFrameDes->nominalTime + timelyDataWindow;

   /* check if the data frame was received less than "timelyDataWindow"
      seconds after the data was recorded at the station */
   if (pFrameDes->clfTime < clfTimeMax)
   {
      DEBUG4(5,"timely data:  frameset: %.6s clfTimeMax: %f clfTime: %f",
             pFrameDes->aFrameset,
             clfTimeMax,
             pFrameDes->clfTime);

      return(TIMELY);
   }
   else
   {
      DEBUG4(5,"late data:  frameset: %.6s clfTimeMax: %f clfTime: %f",
             pFrameDes->aFrameset,
             clfTimeMax,
             pFrameDes->clfTime);

      return(LATE);
   }
}


/**
 * @short Compute frame level state
 *
 * @description This function computes the new frame level state.
 * @description The old frame level state is a wait state and
 * @description the frame level event timer has expired. According
 * @description to the classification of the data, received within the
 * @description the timer window, the new frame level state is computed.
 * @description The data classification is reset.
 *
 * @param pFrLvlData       pointer to the current frame level data structure
 *
 * @return new frame level state
 */

static unsigned char
computeFrameLevelState(frLvlEvLogData_t *pFrLvlData)
{
   int dataClassification = DC_UNDEFINED;
   unsigned char newFrLvlState = UNKNOWN_WAIT;
   char logBuffer[LOGSIZE];

   /* If in a wait state, compute the new frame level state
    * according to the data classification. */
   switch (pFrLvlData->frLvlState)
   {
      /* check all wait states */
      case UNKNOWN_WAIT:
      case TIMELY_WAIT:
      case LATE_WAIT:
      case BACKFILLING_WAIT:
         /* get data classification summary*/
         dataClassification = getDataClassificationSummary(pFrLvlData);

         /* check data classification summary*/
         switch (dataClassification)
         {
         case DC_ONLY_TIMELY_FRAMES:
            newFrLvlState = TIMELY;
            break;
         case DC_ONLY_LATE_FRAMES:
            newFrLvlState = LATE;
            break;
         case DC_TIMELY_AND_LATE_FRAMES:
            /* If the old frame level state is BACKFILLING_WAIT,
             * don't change to BACKFILLING, stay in the old state */
            if (BACKFILLING_WAIT == pFrLvlData->frLvlState)
               newFrLvlState = BACKFILLING_WAIT;
            else
               newFrLvlState = BACKFILLING;
            break;
         }
         break;

      default:
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "frame level state, not a wait state: %c",
                         pFrLvlData->frLvlState);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
   }

   DEBUG4(5,"compute frlvl state: old state %c data class %d new state %c",
          pFrLvlData->frLvlState,
          dataClassification,
          newFrLvlState);

   /* reset the data classification counters */
   resetDataClassification(pFrLvlData);

   return newFrLvlState;
}


/**
 * @short Compute channel level state
 *
 * @description This function computes the current channel level state.
 * @description The function gets the maskpos of the channel and the channel
 * @description mask of a particular frame.
 * @description With the maskpos, the arrayindex in the channel mask is given,
 * @description and the channel state can be addressed.
 *
 * @param pChMaskHistoryElem  pointer to the expired channel mask
 *                            history element
 * @param maskPos          the current channels position in the channel mask
 *
 * @return channel level state (CHANNEL_PRESENT | CHANNEL_MISSING)
 */

static unsigned char
computeChannelLevelState(chMaskHistory_t *pChMaskHistoryElem,
                         uint16_t maskPos)
{
   char logBuffer[LOGSIZE];

   /* Check index first */
   if (maskPos >= MAX_CHANNELS)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "maskPos[%d] >= MAX_CHANNELS",
                      maskPos);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* check and return the current channel level state for the particular
    * nominal time */
   if ('1' == pChMaskHistoryElem->chMaskAndStatus[maskPos].channelMaskItem)
      return(CHANNEL_PRESENT);
   else
      return(CHANNEL_MISSING);
}


/**
 * @short save channel status
 *
 * @description This function stores the channel status information.
 *
 * @param checkPass       channel parsing code
 * @param statusSize      channel status size
 * @param channelStatus   pointer to the channel status data
 * @param pChLvlData      pointer to the channel level event log data
 *                        for the particular data channel
 *
 * @return 0 ok, otherwise error
 */

static int
saveChannelStatus(uint32_t checkPass, uint32_t statusSize,
                  const char *channelStatus, chLvlEvLogData_t *pChLvlData)
{
   /* save channel parsing code */
   pChLvlData->checkPass =  checkPass;

   /* check buffer for channel status data */
   if ((NULL == pChLvlData->chStatusData) ||
       (statusSize > pChLvlData->chStatusSize))
   {
      pChLvlData->chStatusData = (char *)realloc(pChLvlData->chStatusData,
                                                 (size_t)statusSize);
   }

   if (NULL == pChLvlData->chStatusData)
   {
      pChLvlData->chStatusSize = 0;
      DEBUG(5,"realloc for channel status data failed");
      return 1;
   }

   /* save channel status data */
   memcpy(pChLvlData->chStatusData, channelStatus, (size_t)statusSize);
   pChLvlData->chStatusSize = statusSize;

   return 0;
}


/**
 * @short initialize the timer for frame level events
 *
 * @description This function initializes the start time which will be
 * @description taken to calculate if the timer for the creation of
 * @description frame level events is expired.
 * @description The value is the clf time of the data frame, which will.
 * @description raise the frame level event.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param clfTime        clf time of data frame
 *
 * @return nothing
 */

static void
initFrlLvlEventTimer(frLvlEvLogData_t *pFrLvlData,
                     double clfTime)
{
   pFrLvlData->frLvlEvTimer = clfTime;
   return;
}


/**
 * @short store the event attributes of the current data frame
 *
 * @description This function stores the event attributes of the current
 * @description data frame in the frame level data structure.
 * @description The clf time, the nominal time and the frame parsing code
 * @description of the current data frame ara attributes of the frame level
 * @description event which will be caused by the current data frame.
 * @description These values need to be stored, because the event will
 * @description be raised on expiration of the frame level event window timer.
 * @description The stored values are the atttributes of the frame level
 * @description event caused by the current data frame.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param pFrameDes      pointer to the clf file line, stored in an irl entry
 *
 * @return nothing
 */

static void
storeEventAttributesOfCurrentDf(frLvlEvLogData_t *pFrLvlData,
                               intRefListParams_t *pFrameDes)
{
   pFrLvlData->frLvlEvClfTime = pFrameDes->clfTime;
   pFrLvlData->frLvlEvNomTime= pFrameDes->nominalTime;
   pFrLvlData->frLvlEvFpc = pFrameDes->fpc;
   return;
}


/**
 * @short set the frame level event attributes
 *
 * @description This function takes the stored frame level event attributes
 * @description and writes it to the specified "dummy irl element".
 * @description Necessary because due to the window mechanism the original
 * @description irl element of the data frame which causes the new frame
 * @description level event is already released and the function which
 * @description writes the event still expects an irl entry.
 * @description The stored frame level event attributes are reset.
 *
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param pFrameDes        pointer to the "dummy" irl entry
 *
 * @return nothing
 */

static void
setFrLvlEventAttributes(frLvlEvLogData_t *pFrLvlData,
                        intRefListParams_t *pFrameDes)
{
   /* set the event attributes in the dummy irl element */
   pFrameDes->clfTime = pFrLvlData->frLvlEvClfTime;
   pFrameDes->nominalTime = pFrLvlData->frLvlEvNomTime;
   pFrameDes->fpc = pFrLvlData->frLvlEvFpc;

   /* reset the event attributes in the frame level date structure */
   pFrLvlData->frLvlEvClfTime = 0;
   pFrLvlData->frLvlEvNomTime = 0;
   pFrLvlData->frLvlEvFpc = 0;
   return;
}


/**
 * @short store the channel mask & channel status of the current data frame
 *
 * @description This function stores the channel mask and the channel status
 * @description of the current data frame.
 * @description Channel events are processed for timely data only, so do
 * @description nothing if it is a late frame.
 * @description Search for an unprocessed  channel mask history element
 * @description with the current nominal time. If there is no history element
 * @description for the current nominal time, create a new one.
 * @description Store the channel mask of the current frame and the channel
 * @description staus data.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrameDes        pointer to the clf file line, stored in an
 *                         irl entry
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param timelyDataWindow max delay in seconds for a frame to be timely
 *
 * @return nothing
 */

static void
storeChMaskAndStatus(station_t *pStation,
                     intRefListParams_t *pFrameDes,
                     frLvlEvLogData_t *pFrLvlData,
                     long timelyDataWindow)
{
   unsigned char isTimelyFrame; /* frame classified as TIMELY or BACKFILLING */
   int elemFound = 0;   /* default: not found */
   chMaskHistory_t *actElement,
                   *lastElement = NULL;
   channelHistoryCache_t *chHistoryCache;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* compute the frame level state of the current frame */
   isTimelyFrame = checkIfDataFrameIsTimely(pFrameDes,
                                            timelyDataWindow);

   /* check if the current frame contains timely data */
   if (TIMELY != isTimelyFrame)
      return;  /* old data, don't process channel events */

   chHistoryCache = &(pFrLvlData->chHistoryCache);

   /* search for a channel mask history element with the
    * current nominal time */
   for (actElement = chHistoryCache->firstChMaskHistory; actElement != NULL;
        actElement = actElement->nextChMaskHistory)
   {
      /* check if the nominal time of this element matches the
       * current nominal time */
      if (actElement->nominalTime == pFrameDes->nominalTime)
      {
         elemFound = 1;  /* chmask history element found */
         break;
      }
      lastElement = actElement;
   }

   /* check if a history with the current nominal time exists */
   if (1 == elemFound)
   {
      /* check if the events for this element are still unprocessed */
      if (CH_EVENTS_NOT_PROCESSED != actElement->evProcessingState)
      {
         DEBUG2(5,"storeChMaskAndStatus events are processed: nominalTime %f",
                actElement->nominalTime);
         return;
      }

      DEBUG2(5,"add current channel mask to the history: nominalTime %f",
             actElement->nominalTime);
   }
   else  /* history element for the current nominal time does not
            exist, create one */
   {
      /* get new chmask history element */
      actElement = newChMaskHistoryElement(chHistoryCache);

      /* check for errors */
      if (NULL == actElement)
      {
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "newChMaskHistory failed [%.17s]",
                         pStation->conID);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
      }

      /* initialize the new element */
      actElement->clfTime = pFrameDes->clfTime;
      actElement->nominalTime = pFrameDes->nominalTime;
      actElement->nowTime = time(NULL);

      DEBUG4(5,"newChMaskHistory: nominalTime %f: clfTime: %f nowTime: %f",
             actElement->nominalTime,
             actElement->clfTime,
             actElement->nowTime);

      /* link the element to the cache */
      if (NULL == lastElement) chHistoryCache->firstChMaskHistory = actElement;
      /*@-mustfreeonly@*/
      else lastElement->nextChMaskHistory = actElement;
      /*@=mustfreeonly@*/
   }

   /* store the present channels of the current data frame into the corresponding
    * channel mask item */
   storePresentChannels(actElement,
                        pStation->DFinfo->channelMask,
                        pStation->DFinfo->frameTimeLength,
                        pFrameDes->clfTime);

   /* store the channel status info of the present channels */
   storeStatusDataOfAllPresentChannels(pStation, actElement);
}


/**
 * @short write the event "connection established" to ehl file
 *
 * @description This function logs a connection established line
 * @description The event will be written into the current ehl file.
 *
 * @param fp             file pointer to the ehl file
 * @param pFrameDes      pointer to the clf file line, stored in an irl entry
 * @param type           CD data type (CD1 or CD1_1)
 *
 * @return nothing
 */

static void
writeConnEst2Ehl(/*@null@*/FILE *fp,
                 intRefListParams_t *pFrameDes,
                 unsigned char type)
{
   const char *sProtType;

   /* check fp of the current ehl file */
   if (NULL == fp)
   {
      logError("Internal application error",
               "unexpected NULL pointer, writeConnEst2Ehl");
      exit(EXIT_FAILURE);
   }

   /* check CD data type (CD1 or CD1_1) */
   if (CD1_1 == type)
      sProtType = PROTOCOL_TYPE_CD11;
   else if (CD1 == type)
      sProtType = PROTOCOL_TYPE_CD1;
   else
   {
      /* protcol type "UNKNOWN" */
      logError("Internal application error",
               "protocol type <UNKNOWN>");
      exit(EXIT_FAILURE);
   }

   /* Log connection established event line to the current ehl file */
   (void) fprintf(fp, "k %f %.3s %f %f\n",
                  pFrameDes->clfTime,
                  sProtType,
                  pFrameDes->clfTimeSpan,
                  pFrameDes->nominalTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(fp) || 0 != ferror(fp))
   {
      /* In most cases this is a permission or file system
         problem. report is not able to save data, so a shutdown
         is initiated */
      logError("Internal application error",
               "can't write to ehl file");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short write the event "connection established" to stdout
 *
 * @description This function logs a connection established line
 * @description The event will be written to stdout.
 *
 * @param conID          ID of the station object
 * @param pFrameDes      pointer to the clf file line, stored in an irl entry
 * @param type           CD data type (CD1 or CD1_1)
 *
 * @return nothing
 */

static void
writeConnEst2Stdout(const char *conID,
                    intRefListParams_t *pFrameDes,
                    unsigned char type)
{
   const char *sProtType;

   /* check CD data type (CD1 or CD1_1) */
   if (CD1_1 == type)
      sProtType = PROTOCOL_TYPE_CD11;
   else if (CD1 == type)
      sProtType = PROTOCOL_TYPE_CD1;
   else
   {
      /* protcol type "UNKNOWN" */
      logError("Internal application error",
               "protocol type <UNKNOWN>");
      exit(EXIT_FAILURE);
   }

   /* Log connection established event line to stdout */
   (void) fprintf(stdout, "%.17s k %f %.3s %f %f\n",
                  conID,
                  pFrameDes->clfTime,
                  sProtType,
                  pFrameDes->clfTimeSpan,
                  pFrameDes->nominalTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(stdout) || 0 != ferror(stdout))
   {
      logError("Internal application error",
               "can't write to stdout");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short write the event "connection terminated" to ehl file
 *
 * @description This function logs a connection terminated line
 * @description The event will be written into the current ehl file.
 *
 * @param fp             file pointer to the ehl file
 * @param pFrameDes      pointer to the clf file line, stored in an irl entry
 *
 * @return nothing
 */

static void
writeConnTerm2Ehl(/*@null@*/FILE *fp,
                  intRefListParams_t *pFrameDes)
{
   /* check fp of the current ehl file */
   if (NULL == fp)
   {
      logError("Internal application error",
               "unexpected NULL pointer, writeConnTerm2Ehl");
      exit(EXIT_FAILURE);
   }

   /* Log connection terminated event line to the current ehl file */
   (void) fprintf(fp, "l %f\n",
                  pFrameDes->clfTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(fp) || 0 != ferror(fp))
   {
      /* In most cases this is a permission or file system
         problem. report is not able to save data, so a shutdown
         is initiated */
      logError("Internal application error",
               "can't write to ehl file");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short write the event "connection terminated" to stdout
 *
 * @description This function logs a connection terminated line
 * @description The event will be written to stdout.
 *
 * @param conID          ID of the station object
 * @param pFrameDes      pointer to the clf file line, stored in an irl entry
 *
 * @return nothing
 */

static void
writeConnTerm2Stdout(const char *conID,
                     intRefListParams_t *pFrameDes)
{
   /* Log connection terminated event line to stdout */
   (void) fprintf(stdout, "%.17s l %f\n",
                  conID,
                  pFrameDes->clfTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(stdout) || 0 != ferror(stdout))
   {
      logError("Internal application error",
               "can't write to stdout");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Write a bandwidth event to the ehl file
 *
 * @description This function logs a particular bandwidth event.
 * @description The event will be written into the current ehl file.
 *
 * @param fp             file pointer to the ehl file
 * @param bdwdthState    current bandwidth state
 * @param eventTime      event time (clf time or now time)
 *
 * @return nothing
 */

static void
writeBdwdthEvent2Ehl(/*@null@*/FILE *fp,
                     unsigned char bdwdthState,
                     double eventTime)
{
   /* check fp of the current ehl file */
   if (NULL == fp)
   {
      logError("Internal application error",
               "unexpected NULL pointer, writeBdwdthEvent2Ehl");
      exit(EXIT_FAILURE);
   }

   /* Log bandwidth event line to the current ehl file */
   (void) fprintf(fp, "%c %f\n",
                  bdwdthState,
                  eventTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(fp) || 0 != ferror(fp))
   {
      /* In most cases this is a permission or file system
         problem. report is not able to save data, so a shutdown
         is initiated */
      logError("Internal application error",
               "can't write to ehl file");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Write a bandwidth event to stdout
 *
 * @description This function logs a particular bandwidth event.
 * @description The event will be written to stdout.
 *
 * @param conID          ID of the station object
 * @param bdwdthState    current bandwidth state
 * @param eventTime      event time (clf time or now time)
 *
 * @return nothing
 */

static void
writeBdwdthEvent2Stdout(const char *conID,
                       unsigned char bdwdthState,
                       double eventTime)
{
   /* Log frame level event line to stdout */
   (void) fprintf(stdout, "%.17s %c %f\n",
                  conID,
                  bdwdthState,
                  eventTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(stdout) || 0 != ferror(stdout))
   {
      logError("Internal application error",
               "can't write to stdout");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Write a frame level event to ehl file
 *
 * @description This function logs a particular frame level event.
 * @description The event will be written into the current ehl file.
 *
 * @param fp             file pointer to the ehl file
 * @param frLvlState     current frame level state (frame level event)
 * @param pFrameDes      pointer to the clf file line, stored in an irl entry
 *
 * @return nothing
 */

static void
writeFrameEvent2Ehl(/*@null@*/FILE *fp,
                    unsigned char frLvlState,
                    intRefListParams_t *pFrameDes)
{
   /* check fp of the current ehl file */
   if (NULL == fp)
   {
      logError("Internal application error",
               "unexpected NULL pointer, writeFrameEvent2Ehl");
      exit(EXIT_FAILURE);
   }

   /* Log frame level event line to the current ehl file */
   (void) fprintf(fp, "%c %f %f %x\n",
                  frLvlState,
                  pFrameDes->clfTime,
                  pFrameDes->nominalTime,
                  (unsigned int)pFrameDes->fpc);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(fp) || 0 != ferror(fp))
   {
      /* In most cases this is a permission or file system
         problem. report is not able to save data, so a shutdown
         is initiated */
      logError("Internal application error",
               "can't write to ehl file");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Write a frame level event to stdout
 *
 * @description This function logs a particular frame level event.
 * @description The event will be written to stdout.
 *
 * @param conID          ID of the station object
 * @param frLvlState     current frame level state (frame level event)
 * @param pFrameDes      pointer to the clf file line, stored in an irl entry
 *
 * @return nothing
 */

static void
writeFrameEvent2Stdout(const char *conID,
                       unsigned char frLvlState,
                       intRefListParams_t *pFrameDes)
{
   /* Log frame level event line to stdout */
   (void) fprintf(stdout, "%.17s %c %f %f %x\n",
                  conID,
                  frLvlState,
                  pFrameDes->clfTime,
                  pFrameDes->nominalTime,
                  (unsigned int)pFrameDes->fpc);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(stdout) || 0 != ferror(stdout))
   {
      logError("Internal application error",
               "can't write to stdout");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Log channel level event line to ehl file
 *
 * @description This function logs a particular channel level event.
 * @description The event will be written into the current ehl file.
 *
 * @param fp             file pointer to the ehl file
 * @param chLvlEvent     current channel level event
 * @param pFrLvlData     pointer to the frame level data
 * @param pChLvlData     pointer to the channel level appl extension object
 * @param scName         siteName/channelname
 * @param type           CD data type (CD1 or CD1_1)
 *
 * @return nothing
 */

static void
writeChannelEvent2Ehl(/*@null@*/FILE *fp,
                      unsigned char chLvlEvent,
                      const frLvlEvLogData_t *pFrLvlData,
                      const chLvlEvLogData_t *pChLvlData,
                      const char *scName,
                      unsigned char type)
{
   uint16_t ii;
   uint16_t sizeToLog;

   /* check fp of the current ehl file */
   if (NULL == fp)
   {
      logError("Internal application error",
               "unexpected NULL pointer, writeChannelEvent2Ehl");
      exit(EXIT_FAILURE);
   }

   if (CHANNEL_MISSING == chLvlEvent)
   {
      /* Log channel level event line to the current ehl file */
      (void) fprintf(fp, "%c %f %.17s\n",
                     chLvlEvent,
                     pChLvlData->eventClfTime,
                     scName);
   }
   else  /* CHANNEL_PRESENT | CHANNEL_STATUS_CHANGED */
   {
      /* determine the size of the channel status data to be logged.
       * For CD1 use the value of StatusSize, for CD1.1 use
       * LNG_4_CHK_OF_CD11_STATUS. For safety reasons, check if the status
       * sizes exceeds LNG_4_CHK_OF_CD11_STATUS. */
      if ((CD1_1 == type) &&
          (LNG_4_CHK_OF_CD11_STATUS <= pChLvlData->chStatusSize))
         sizeToLog = LNG_4_CHK_OF_CD11_STATUS;
      /* CD1 or StatusSize less than LNG_4_CHK_OF_CD11_STATUS */
      else
         sizeToLog = pChLvlData->chStatusSize;

      /*  Log channel level event line to the current ehl file */
      (void) fprintf(fp, "%c %f %.17s %f %x ",
                     chLvlEvent,
                     pChLvlData->eventClfTime,
                     scName,
                     pFrLvlData->nominalTime4ChEvents,
                     (unsigned int)pChLvlData->checkPass);

      for (ii = 0; ii < sizeToLog; ii++)
      {
         fprintf(fp, "%02x",
                 ((unsigned int)pChLvlData->chStatusData[ii] & 0xFF));
      }
      fprintf(fp, "\n");
   }

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(fp) || 0 != ferror(fp))
   {
      /* In most cases this is a permission or file system
         problem. CD report is not able to save data, so a shutdown
         is initiated */
      logError("Internal application error",
               "can't write to ehl file");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Log channel level event line to stdout
 *
 * @description This function logs a particular channel level event.
 * @description The event will be written to stdout.
 *
 * @param conID          ID of the station object
 * @param chLvlEvent     current channel level event
 * @param pFrLvlData     pointer to the frame level data
 * @param pChLvlData     pointer to the channel level appl extension object
 * @param scName         siteName/channelname
 * @param type           CD data type (CD1 or CD1_1)
 *
 * @return nothing
 */

static void
writeChannelEvent2Stdout(const char *conID,
                         unsigned char chLvlEvent,
                         const frLvlEvLogData_t *pFrLvlData,
                         const chLvlEvLogData_t *pChLvlData,
                         const char *scName,
                         unsigned char type)
{
   uint16_t ii;
   uint16_t sizeToLog;

   if (CHANNEL_MISSING == chLvlEvent)
   {
      /*  Log channel level event line to stdout */
      (void) fprintf(stdout, "%.17s %c %f %.17s\n",
                     conID,
                     chLvlEvent,
                     pChLvlData->eventClfTime,
                     scName);
   }
   else  /* CHANNEL_PRESENT | CHANNEL_STATUS_CHANGED */
   {
      /* determine the size of the channel status data to be logged.
       * For CD1 use the value of StatusSize, for CD1.1 use
       * LNG_4_CHK_OF_CD11_STATUS. For safety reasons, check if the status
       * sizes exceeds LNG_4_CHK_OF_CD11_STATUS. */
      if ((CD1_1 == type) &&
          (LNG_4_CHK_OF_CD11_STATUS <= pChLvlData->chStatusSize))
         sizeToLog = LNG_4_CHK_OF_CD11_STATUS;
      /* CD1 or StatusSize less than LNG_4_CHK_OF_CD11_STATUS */
      else
         sizeToLog = pChLvlData->chStatusSize;

      /*   Log channel level event line to stdout */
      (void) fprintf(stdout, "%.17s %c %f %.17s %f %x ",
                     conID,
                     chLvlEvent,
                     pChLvlData->eventClfTime,
                     scName,
                     pFrLvlData->nominalTime4ChEvents,
                     (unsigned int)pChLvlData->checkPass);

      for (ii = 0; ii < sizeToLog; ii++)
      {
         fprintf(stdout, "%02x",
                 ((unsigned int)pChLvlData->chStatusData[ii] & 0xFF));
      }
      fprintf(stdout, "\n");
   }

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(stdout) || 0 != ferror(stdout))
   {
      logError("Internal application error",
               "can't write to stdout");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Log channel level event CHANNEL_INFO event to ehl file
 *
 * @description This function logs the channel level event
 * @description CHANNEL_INFO event.
 * @description The event will be written into the current ehl file.
 *
 * @param fp             file pointer to the ehl file
 * @param eventTime      current time (clfTime, or nowTime)
 * @param scName         pointer to current siteName/channelName
 * @param pChLvlData     pointer to the channel level event log data structure
 *
 * @return nothing
 */

static void
writeChannelInfoEvent2Ehl(/*@null@*/FILE *fp,
                          double eventTime,
                          const char *scName,
                          const chLvlEvLogData_t *pChLvlData)
{
   /* check fp of the current ehl file */
   if (NULL == fp)
   {
      logError("Internal application error",
               "unexpected NULL pointer, writeChannelInfoEvent2Ehl");
      exit(EXIT_FAILURE);
   }

   /* Log channel level event line N_DATA_FRAMES RECEIVED to the current
      ehl file */
   (void) fprintf(fp, "%c %f %.17s %f %f %f\n",
                  CHANNEL_INFO,
                  eventTime,
                  scName,
                  pChLvlData->earliestNominalTime,
                  pChLvlData->latestNominalTime,
                  pChLvlData->lastFrameRecNominalTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(fp) || 0 != ferror(fp))
   {
      /* In most cases this is a permission or file system
         problem. report is not able to save data, so a shutdown
         is initiated */
      logError("Internal application error",
               "can't write to ehl file");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Log channel level event CHANNEL_INFO event to stdout
 *
 * @description This function logs the channel level event
 * @description CHANNEL_INFO event.
 * @description The event will be written to stdout.
 *
 * @param conID          ID of the station object
 * @param eventTime      current time (clfTime, or nowTime)
 * @param scName         pointer to current siteName/channelName
 * @param pChLvlData     pointer to the channel level event log data structure
 *
 * @return nothing
 */

static void
writeChannelInfoEvent2Stdout(const char *conID,
                             double eventTime,
                             const char *scName,
                             const chLvlEvLogData_t *pChLvlData)
{
   /* Log channel level event line N_DATA_FRAMES RECEIVED to stdout */
   (void) fprintf(stdout, "%.17s %c %f %.17s %f %f %f\n",
                  conID,
                  CHANNEL_INFO,
                  eventTime,
                  scName,
                  pChLvlData->earliestNominalTime,
                  pChLvlData->latestNominalTime,
                  pChLvlData->lastFrameRecNominalTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(stdout) || 0 != ferror(stdout))
   {
      logError("Internal application error",
               "can't write to stdout");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Write the frame level event "NO_DATA" to ehl file
 *
 * @description This function logs an "NO_DATA" event line which only
 * @description consists of the event code and the time when the event
 * @description occured.
 * @description The event will be written into the current ehl file.
 *
 * @param fp             file pointer to the ehl file
 * @param eventTime      event time
 *
 * @return nothing
 */

static void
writeNoDataEvent2Ehl(/*@null@*/FILE *fp,
                     double eventTime)
{
   /* check fp of the current ehl file */
   if (NULL == fp)
   {
      logError("Internal application error",
               "unexpected NULL pointer, writeNoDataEvent2Ehl");
      exit(EXIT_FAILURE);
   }

   /* Log NO_DATA event line to the current ehl file */
   (void) fprintf(fp, "N %f\n",
                  eventTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(fp) || 0 != ferror(fp))
   {
      /* In most cases this is a permission or file system
         problem. report is not able to save data, so a shutdown
         is initiated */
      logError("Internal application error",
               "can't write to ehl file");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short Write the frame level event "NO_DATA" to stdout
 *
 * @description This function logs an "NO_DATA" event line which only
 * @description consists of the event code and the time when the event
 * @description occurred.
 * @description The event will be written to stdout.
 *
 * @param conID          ID of the station object
 * @param eventTime      event time
 *
 * @return nothing
 */

static void
writeNoDataEvent2Stdout(const char *conID,
                        double eventTime)
{
   /* Log NO_DATA event line to stdout */
   (void) fprintf(stdout, "%.17s N %f\n",
                  conID,
                  eventTime);

   /* flush buffer and check if an error has occurred */
   if (0 != fflush(stdout) || 0 != ferror(stdout))
   {
      logError("Internal application error",
               "can't write to stdout");
      exit(EXIT_FAILURE);
   }
}


/**
 * @short channel Info: write CD11 channel info line to stdout or ehl file
 *
 * @description This function writes the information on a particular CD11
 * @description channel subframe to stdout.
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param channel        pointer to the data channel info
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
writeCD11ChannelInfo2StdoutOrEhl(station_t *pStation,
                                 const CDdataChannel_t *channel,
                                 const intRefListParams_t *pFrameDes,
                                 uint16_t runMode)

{
   uint16_t ii;
   uint16_t sizeToLog;
   int compression;
   uint32_t chOffsetBinFile;
   FILE *fp;

   /* Check if the information has to be written to stdout, or to the
    * ehl files */
   if (runMode & WR_STDOUT_2_EHL_FILES)
   {
      fp = pStation->files.fpUsrFile;
   }
   else
   {
      fp = stdout;

      /* conID if output to stdout */
      (void) fprintf(fp,
                     "%.17s ",
                     pStation->conID);
   }

   /* Determine the size of the channel status data to be logged.
    * For CD1.1 use LNG_4_CHK_OF_CD11_STATUS if the first byte of the
    * has the value '1'. For safety reasons,
    * check if the status sizes exceeds LNG_4_CHK_OF_CD11_STATUS. */
   if (('\1' == channel->channelStatus[0]) &&
      (LNG_4_CHK_OF_CD11_STATUS <= channel->channelStatusSize))
      sizeToLog = LNG_4_CHK_OF_CD11_STATUS;
   /* StatusSize less than LNG_4_CHK_OF_CD11_STATUS, or unknown format */
   else
      sizeToLog = channel->channelStatusSize;

   /* Protocol type, site/channel/location */
   (void) fprintf(fp,
                  "%.3s %.6s/%.8s/%.2s ",
                  PROTOCOL_TYPE_CD11,
                  channel->siteName,
                  channel->channelName,
                  channel->locationName);

   /* channel parsing code,
    * channel offset: offset of the first byte of the channel relative to the
    *                 start of the frame (not part of the channel subframe,
    *                 added by the reporting tool)
    * channel length,
    * authentication offset */
   (void) fprintf(fp,
                  "%x %u %u %u ",
                  (unsigned int)channel->checkPass,
                  (unsigned int)channel->channelOffset,
                  (unsigned int)channel->channelLength,
                  (unsigned int)channel->authOffset);

   /* determine compression flag */
   compression = (int) channel->compression;

   /* Canadian values are swapped during parsing for CD1.1! */
   switch (channel->compression)
   {
      case 1:
         compression = 2; /* swapped during parsing */
         break;
      case 2:
         compression = 1; /* swapped during parsing */
         break;
   }

   /* authentication switch: 1 —> on,
    *                        0 —> off
    * compression flag: 0 –> off,
    *                   1 –> Canadian before signing
    *                   2 –> Canadian after signing
    * sensor type: 0 –> seismic,
    *              1 –> hydroacustic,
    *              2 –> infrasonic,
    *              3 –> weather,
    *              > 3 other */
   (void) fprintf(fp,
                  "%d %d %d ",
                  (int)channel->authSwitch,
                  compression,
                  (int)channel->sensorType);

   /* option flag  0 = unused
    *              1 = calibration factor and calibration period provided */
   (void) fprintf(fp,
                  "%d ",
                  (int)channel->option);

   /* Log calibration information if available */
   if ((char) 1 == channel->option)
   {
      /* calibration factor,
       * calibration period */
      (void) fprintf(fp,
                     "%f %f ",
                     channel->calibFactor,
                     channel->calibPeriod);
   }
   else
   {
      /* no calibration information available */
      (void) fprintf(fp,
                     "NA NA ");
   }

   /* uncompressed data format type or data format type */
   if ('\0' == channel->dataFormat[0])
   {
      /* Byte 15 of the channel description is binary null,
         print question mark to avoid empty data type */
      (void) fprintf(fp, "? ");
   }
   else
   {
      (void) fprintf(fp, "%.2s ", channel->dataFormat);
   }

   /* time stamp (nominal time of channel),
    * subframe time length,
    * number of samples */
   (void) fprintf(fp,
                  "%f %u %u ",
                  channel->timeStamp,
                  (unsigned int)channel->subFrameTimeLng,
                  (unsigned int)channel->nbrOfSamples);

   /* channel status size */
   (void) fprintf(fp,
                  "%d ",
                  (int)sizeToLog);

   /* channel status information */
   for (ii = 0; ii < sizeToLog; ii++)
   {
      fprintf(fp, "%02x",
              ((unsigned int)channel->channelStatus[ii] & 0xFF));
   }

   /* channel data size,
    * subframe count */
   (void) fprintf(fp,
                  " %u %u ",
                  (unsigned int)channel->dataSize,
                  (unsigned int)channel->subframeCount);

   /* authentication key identifier,
    * authentication size: authentication size (0 if not signed or
    *                                           40 if signed) */
   (void) fprintf(fp,
                  "%u %u ",
                  (unsigned int)channel->authKeyIdentifier,
                  (unsigned int)channel->signatureSize);

   /* Check if the extension option -d was specified on the cmd line */
   if (runMode & CH_INFO_EXT1_2_STDOUT)
   {
      (void) fprintf(fp,
                     "%s.%s ",
                     pFrameDes->pFileObj->fileNameAndPath,
                     EXTENSION_BIN_CD11);

      /* Compute offset of the channel from the beginning of the binary file */
      chOffsetBinFile = pFrameDes->offsetBinFile + channel->channelOffset;
      DEBUG3(5,"pFrameDes->offsetBinFile %u + channel->channelOffset %u",
             pFrameDes->offsetBinFile,
             channel->channelOffset);

      (void) fprintf(fp,
                     "%u ",
                     (unsigned int)chOffsetBinFile);

      /* Output time of last GPS synchronization and clock
         differential if CD-1.1 frame */
      if (channel->channelStatusSize == 32 &&
          '\1' == *(channel->channelStatus))
      {
         /* time of last GPS synchronization is terminated by a space
            character */
         (void) fprintf(fp,
                        "%.3f %lu",
                        ascDate2double(channel->channelStatus + 8),
         (unsigned long)ntohl(*((uint32_t *)(channel->channelStatus + 28))));
      }
      else
      {
         DEBUG(5, "No CD-1.1 Option 1 status");
         (void) fprintf(fp, "NA NA");
      }
   }

   fprintf(fp, "\n");
}


/**
 * @short channel Info: write CD1 channel info line to stdout or ehl file
 *
 * @description This function writes the information on a particular CD1
 * @description channel subframe to stdout.
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param channel        pointer to the data channel info
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
writeCD1ChannelInfo2StdoutOrEhl(station_t *pStation,
                                const CDdataChannel_t *channel,
                                const intRefListParams_t *pFrameDes,
                                uint16_t runMode)
{
   uint32_t ii;
   uint32_t chOffsetBinFile;
   FILE *fp;

   /* Check if the information has to be written to stdout, or to the
    * ehl files */
   if (runMode & WR_STDOUT_2_EHL_FILES)
   {
      fp = pStation->files.fpUsrFile;
   }
   else
   {
      fp = stdout;

      /* conID if output to stdout */
      (void) fprintf(fp,
                     "%.17s ",
                     pStation->conID);
   }

   /* Protocol type, site/channel/location */
   (void) fprintf(fp,
                  "%.3s %.6s/%.8s ",
                  PROTOCOL_TYPE_CD1,
                  channel->siteName,
                  channel->channelName);

   /* channel parsing code,
    * channel offset: offset of the first byte of the channel relative to the
    *                 start of the frame (not part of the channel subframe,
    *                 added by the reporting tool)
    * packet length ,
    * authentication offset: 0 (CD-1.0)*/
   (void) fprintf(fp,
                  "%x %u %u 0 ",
                  (unsigned int)channel->checkPass,
                  (unsigned int)channel->channelOffset,
                  (unsigned int)channel->channelLength);

   /* authentication switch: 1 —> on,
    *                        0 —> off
    * compression flag: 0 –> off,
    *                   1 –> Canadian before signing
    *                   2 –> Canadian after signing
    * sensor type: “NA” for CD-1.0 */
   (void) fprintf(fp,
                  "%d %d NA ",
                  (int)channel->authSwitch,
                  (int)channel->compression);

   /* option flag  0 = unused
    *              1 = calibration factor and calibration period provided */
   (void) fprintf(fp,
                  "%d ",
                  (int)channel->option);

   /* Log calibration information if available */
   if ((char) 1 == channel->option)
   {
      /* CSS 3.0 calibration factor,
       * CSS 3.0 calibration period */
      (void) fprintf(fp,
                     "%f %f ",
                     channel->calibFactor,
                     channel->calibPeriod);
   }
   else
   {
      /* no calibration information available */
      (void) fprintf(fp,
                     "NA NA ");
   }

   /* uncompressed data format type or data format type.
    * (not part of the CD-1.0 channel subframe) */
   if ('\0' == channel->dataFormat[0])
   {
      /* Byte 15 of the channel description is binary null,
         print question mark to avoid empty data type */
      (void) fprintf(fp, "? ");
   }
   else
   {
      (void) fprintf(fp, "%.2s ", channel->dataFormat);
   }

   /* time stamp (nominal time of channel),
    * subframe time length: "DFinfo->frameTimeLength" for CD1.0 ,
    * number of samples */
   (void) fprintf(fp,
                  "%f %u %u ",
                  channel->timeStamp,
                  (unsigned int)pStation->DFinfo->frameTimeLength,
                  (unsigned int)channel->nbrOfSamples);

   /* channel status size */
   (void) fprintf(fp,
                  "%d ",
                  (int)channel->channelStatusSize);

   /* channel status information */
   for (ii = 0; ii < channel->channelStatusSize; ii++)
   {
      fprintf(fp, "%02x",
              ((unsigned int)channel->channelStatus[ii] & 0xFF));
   }

   /* channel data size: (not part of the CD-1.0 channel subframe,
    *                     added by the reporting tool),
    * subframe count:    “NA” for CD-1.0 */
   (void) fprintf(fp,
                  " %u NA ",
                  (unsigned int)channel->dataSize);

   /* authentication key identifier: -1 for CD1.0
    * authentication size: authentication size (0 if not signed or
    *                                           40 if signed) */
   (void) fprintf(fp,
                  "-1 %u ",
                  (unsigned int)channel->signatureSize);

   /* Check if the extension option -d was specified on the cmd line */
   if (runMode & CH_INFO_EXT1_2_STDOUT)
   {
      (void) fprintf(fp,
                     "%s.%s ",
                     pFrameDes->pFileObj->fileNameAndPath,
                     EXTENSION_BIN_CD1);

      /* Compute offset of the channel from the beginning of the binary file */
      chOffsetBinFile = pFrameDes->offsetBinFile + channel->channelOffset;
      DEBUG3(5,"pFrameDes->offsetBinFile %u + channel->channelOffset %u",
             pFrameDes->offsetBinFile,
             channel->channelOffset);

      (void) fprintf(fp,
                     "%u NA NA",
                     (unsigned int)chOffsetBinFile);
   }

   fprintf(fp, "\n");
}


/**
 * @short check if frame level event shall be reported
 *
 * @description This function checks if the new frame level shall
 * @description shall be logged the ehl file, or to the event log
 * @description stream.
 * @description Some of the frame level events do not raise events
 * @description because they are <wait states>.
 * @description <Wait states> are intermediate states which do not cause
 * @description a frame level event.
 *
 * @param frLvlState     frame level state of the current frame
 *
 * @return 0             don't raise a frame level event
 * @return 1             a frame level event shall be raised
 */

static int
checkIfFrLvlStateRaisesAnEvent(unsigned char frLvlState)
{
   int returnCode;
   char logBuffer[LOGSIZE];

   /* check the new frame level state */
   switch (frLvlState)
   {
      case UNKNOWN_WAIT:
      case TIMELY_WAIT:
      case LATE_WAIT:
      case BACKFILLING_WAIT:
         returnCode = 0; /* don't raise a frame level event */
         break;

      case TIMELY:
      case LATE:
      case BACKFILLING:
         returnCode = 1; /* raise a frame level event */
         break;
      default:
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "unexpected frame level state: %c",
                         frLvlState);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
   }
   return returnCode;
}


/**
 * @short check frame level event timer
 *
 * @description This function checks frame level event timer.
 * @description The timer may be not active, active or expired.
 * @description If the timer was expired, reset the start time value.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param checkTime      clf time of the current frame, or current time
 *
 * @return NO_TIMER | TIMER_ACTIVE | TIMER_EXPIRED
 */

static int
checkFrameLevelEventTimer(frLvlEvLogData_t *pFrLvlData,
                          double checkTime)
{
   double delta_t;

   /* check if timer active */
   if (0 == pFrLvlData->frLvlEvTimer)
      return (FRLVL_NO_TIMER);

   /* compute the current event window size for this element */
   delta_t = checkTime - pFrLvlData->frLvlEvTimer;

   DEBUG4(5,"chkFrlvlEventTimer: %f - %f delta_t: %f",
          checkTime,
          pFrLvlData->frLvlEvTimer,
          delta_t);

   /* check if the event creation window is expired */
   if (FR_EVENT_WINDOW_SIZE > delta_t)
      return (FRLVL_TIMER_ACTIVE);  /* timer is still active */

   /* frame level event timer is expired:
    * reset the start time value */
   pFrLvlData->frLvlEvTimer = 0;

   return (FRLVL_TIMER_EXPIRED);
}


/**
 * @short Update data classification counters
 *
 * @description This function checks if the current frame is TIMELY or
 * @description LATE. The corresponding data classification counter
 * @description is updated. This is necessary in the wait states, because
 * @description the state change depends on the data classification summary
 * @description of all frames, received within the timer window.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param isTimelyFrame  TIMELY or LATE
 *
 * @return nothing
 */

static void
updateDataClassification(frLvlEvLogData_t *pFrLvlData,
                         unsigned char isTimelyFrame)
{
   /* update the "data classification" counters */
   if (TIMELY == isTimelyFrame)
   {
      pFrLvlData->nrOfTimelyFramesArrived++;
   }
   else
   {
      pFrLvlData->nrOfLateFramesArrived++;
   }

   DEBUG3(5,"nr of timely frames: %d, nr of late frames: %d ",
          pFrLvlData->nrOfTimelyFramesArrived,
          pFrLvlData->nrOfLateFramesArrived);
   return;
}


/**
 * @short get data classification summary
 *
 * @description This function checks the data classification counters
 * @description and returns the data classification summary
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 *
 * @return DC_ONLY_TIMELY_FRAMES | DC_ONLY_LATE_FRAMES |
 * @return DC_TIMELY_AND_LATE_FRAMES
 */

static int
getDataClassificationSummary(frLvlEvLogData_t *pFrLvlData)
{
   int dataClassification = DC_UNDEFINED;

   /* compute data classification */
   if (0 == pFrLvlData->nrOfTimelyFramesArrived)
      dataClassification = DC_ONLY_LATE_FRAMES;
   else if (0 == pFrLvlData->nrOfLateFramesArrived)
      dataClassification = DC_ONLY_TIMELY_FRAMES;
   else
      dataClassification = DC_TIMELY_AND_LATE_FRAMES;

   DEBUG4(5,"nr of timely frames: %d  nr of late frames: %d dataClass: %d",
         pFrLvlData->nrOfTimelyFramesArrived,
         pFrLvlData->nrOfLateFramesArrived,
         dataClassification);

   return dataClassification;
}


/**
 * @short reset the data classification counters
 *
 * @description This function resets the data classification counters.
 *
 * @param pFrLvlData     pointer to the current frame level data structure
 *
 * @return nothing
 */

static void
resetDataClassification(frLvlEvLogData_t *pFrLvlData)
{
   /* reset data classification counters */
   pFrLvlData->nrOfTimelyFramesArrived = 0;
   pFrLvlData->nrOfLateFramesArrived = 0;
   return;
}


/**
 * @short check bandwidth events by now time
 *
 * @description This function checks if the time window for the check for
 * @description bandwidth events has expired. This check is performed with
 * @description the "now time values".
 * @description If the time window has expired, the bandwidth state of the
 * @description n data frames received in the window is computed. If the
 * @description bandwidth state has changed, the corresponding event is raised
 * @description The corresponding frame level event
 * @description will be written to the station related ehl file or to stdout
 * @description (depending on the current run mode).
 *
 * @param pStation         pointer to the current station structure
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param nowTime          current time
 * @param runMode          run mode of the report tool
 *
 * @return BDWDTH_EVENTS_PROCESSED | BDWDTH_EVENTS_NOT_PROCESSED
  */

static int
check4BandwidthEventsByNowTime(station_t *pStation,
                               frLvlEvLogData_t *pFrLvlData,
                               double nowTime,
                               uint16_t runMode)
{
   unsigned char bdwdthState; /* bandwidth state computed for the n data frames
                                 of the expired event window */
   double        currentWindowSize;


   /* If the station is not in the state "STATION_CONNECTED" do nothing */
   if(STATION_CONNECTED != pStation->state)
      return BDWDTH_EVENTS_NOT_PROCESSED;

   currentWindowSize = nowTime - pFrLvlData->bandwdthNowTime;

   DEBUG5(5,"station: [%.8s] current bandw window size %f nowTime %f, bwNow %f",
          pStation->stationName,
          currentWindowSize,
          nowTime,
          pFrLvlData->bandwdthNowTime);

   /* the bandwidth event shall be created periodically, using a time window
    * during this time window the bandwidth rates are averaged and this
    * averaged bandwidth rate is used to compute the current bandwidth state
    * check if the event window has expired  */
   if (BWDTH_EVENT_WINDOW_SIZE > currentWindowSize)
      return BDWDTH_EVENTS_NOT_PROCESSED;

   /* event window expired, check for bandwidth events
    * using the bandwidth rates of past n received frames */
   bdwdthState = computeBandwidthState(pFrLvlData);

   DEBUG3(5,"old bandwdthState: %c, new bandwdthState: %c",
          pFrLvlData->bandwdthState,
          bdwdthState);

   /* check if the bandwidth state is equal to the stored value */
   if (bdwdthState == pFrLvlData->bandwdthState)
         return BDWDTH_EVENTS_PROCESSED;

   /* The bandwidth state has changed.
    * Log the occurred event, depending on the run mode */
   if (runMode & CR_EHL_FILES)
      writeBdwdthEvent2Ehl(pStation->files.fpUsrFile,
                           bdwdthState,
                           nowTime);
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeBdwdthEvent2Ehl(pStation->files.fpUsrFile,
                           bdwdthState,
                           nowTime);
   else if (runMode & EV_STREAM_2_STDOUT)
      writeBdwdthEvent2Stdout(pStation->conID,
                              bdwdthState,
                              nowTime);

   /* store the current frame level state */
   pFrLvlData->bandwdthState = bdwdthState;

   return BDWDTH_EVENTS_PROCESSED;
}


/**
 * @short check bandwidth events by clf time
 *
 * @description This function checks if the time window for the check for
 * @description bandwidth events has expired. This check is performed with
 * @description the "clf time values".
 * @description If the time window has expired, the bandwidth state of the
 * @description n data frames received in the window is computed. If the
 * @description bandwidth state has changed, the corresponding event is raised
 * @description The corresponding frame level event
 * @description will be written to the station related ehl file or to stdout
 * @description (depending on the current run mode).
 *
 * @param pStation         pointer to the current station structure
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param clfTime          clf time of the current data frame
 * @param runMode          run mode of the report tool
 *
 * @return BDWDTH_EVENTS_PROCESSED | BDWDTH_EVENTS_NOT_PROCESSED
  */

static int
check4BandwidthEventsByClfTime(station_t *pStation,
                               frLvlEvLogData_t *pFrLvlData,
                               double clfTime,
                               uint16_t runMode)
{
   unsigned char bdwdthState; /* bandwidth state computed for the n data frames
                                 of the expired event window */
   uint16_t      windowExpired = BDWDTH_WINDOW_NOT_EXPIRED;
   double        currentWindowSize;


   currentWindowSize = clfTime - pFrLvlData->bandwdthClfTime;

   DEBUG5(5,"station: [%.8s] current bandw window size %f clfTime %f, bwClf %f",
          pStation->stationName,
          currentWindowSize,
          clfTime,
          pFrLvlData->bandwdthClfTime);

   /* The first bandwidth event is created when the first data frame is seen.
    * Until then, bandwidth events shall be created periodically,
    * using a time window. During this time window the bandwidth rates are
    * averaged and this averaged bandwidth rate is used to compute the current
    * bandwidth state */
   if (BWDTH_EVENT_WINDOW_SIZE < currentWindowSize)
      windowExpired = BDWDTH_WINDOW_EXPIRED;

   /* If this is the first data frame, the bandwidth state is computed
    * although the event window is not expired */
   if (BW_FIRST_FRAME_NOT_SEEN == pFrLvlData->bandwdthFirstFrame)
   {
      windowExpired = BDWDTH_WINDOW_EXPIRED;

      /* first data frame after connect seen */
      pFrLvlData->bandwdthFirstFrame = BW_FIRST_FRAME_SEEN;
   }

   DEBUG4(5,"station: [%.8s] pFrLvlData->bandwdthFirstFrame %c windowExp %d",
          pStation->stationName,
          pFrLvlData->bandwdthFirstFrame,
          windowExpired);

   /* check if the bandwidth state shall be computed */
   if (BDWDTH_WINDOW_NOT_EXPIRED == windowExpired)
      return BDWDTH_EVENTS_NOT_PROCESSED;

   /* event window expired, or first frame seen.
    * check for bandwidth events using the bandwidth rates of past n received
    * frames */
   bdwdthState = computeBandwidthState(pFrLvlData);

   DEBUG3(5,"old bandwdthState: %c, new bandwdthState: %c",
          pFrLvlData->bandwdthState,
          bdwdthState);

   /* check if the bandwidth state is equal to the stored value */
   if (bdwdthState == pFrLvlData->bandwdthState)
         return BDWDTH_EVENTS_PROCESSED;

   /* The bandwidth state has changed.
    * Log the occurred event, depending on the run mode */
   if (runMode & CR_EHL_FILES)
      writeBdwdthEvent2Ehl(pStation->files.fpUsrFile,
                           bdwdthState,
                           clfTime);
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeBdwdthEvent2Ehl(pStation->files.fpUsrFile,
                           bdwdthState,
                           clfTime);
   else if (runMode & EV_STREAM_2_STDOUT)
      writeBdwdthEvent2Stdout(pStation->conID,
                              bdwdthState,
                              clfTime);

   /* store the current frame level state */
   pFrLvlData->bandwdthState = bdwdthState;

   return BDWDTH_EVENTS_PROCESSED;
}


/**
 * @short check bandwidth events
 *
 * @description This function checks for bandwidth events.
 * @description If there are still no frames received within the current
 * @description event window store the bandwidth data of the current frame
 * @description and return.
 * @description Otherwise check if the window for the bandwidth event
 * @description generation has expired. Use the clf time of the current frame
 * @description for this check!
 * @description Depending on the processind state of the bandwidth event window
 * @description the bandwidth data of the current frame is stored.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrameDes        pointer to the clf file line, stored in an
 *                         irl entry.
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param nowTime          current time
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
check4BandwidthEvents(station_t *pStation,
                      intRefListParams_t *pFrameDes,
                      frLvlEvLogData_t *pFrLvlData,
                      double nowTime,
                      uint16_t runMode)
{
   int bwdtEventsProcessed;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* check number of received frames, if there are no frames received yet
    * initialize the values needed for teh event generation. */
   if (0 == pFrLvlData->bandwdthNrOfFrames)
   {
      /* initialize the vales for the event generation check.
       * store clf time of the current frame and the current time */
      pFrLvlData->bandwdthClfTime = pFrameDes->clfTime;
      pFrLvlData->bandwdthNowTime = nowTime;

      /* increase frame counter */
      pFrLvlData->bandwdthNrOfFrames++;

      /* store bandwidth data of the current frame */
      storeBandwidthData(pFrLvlData,
                         pFrameDes,
                         pStation->DFinfo->frameTimeLength);
      return;
   }

   /* check if it is time to process the current bandwidth state */
   bwdtEventsProcessed = check4BandwidthEventsByClfTime(pStation,
                                                        pFrLvlData,
                                                        pFrameDes->clfTime,
                                                        runMode);

   /* check if the bandwidth events of the current window are processed */
   if (BDWDTH_EVENTS_NOT_PROCESSED == bwdtEventsProcessed)
   {
      /* store bandwidth data of the current frame */
      storeBandwidthData(pFrLvlData,
                         pFrameDes,
                         pStation->DFinfo->frameTimeLength);

      /* increase frame counter */
      pFrLvlData->bandwdthNrOfFrames++;
   }
   else
   {
      /*  bandwidth events are processed for the current window.
       *  reset all counters and start the next window.
       *  Store the bandwidth data of the current window and initialize
       *  the window time stamps */
      pFrLvlData->bandwdthNrOfFrames = 0;
      pFrLvlData->bandwdthClfTimeSpanTotal = 0;
      pFrLvlData->bandwdthFrameTimeTotal = 0;
      pFrLvlData->bandwdthClfTime = 0;
      pFrLvlData->bandwdthNowTime = 0;

      /*  store bandwidth values of the current
       * data frame, store initial time time stamps for the
       * next event window */
      storeBandwidthData(pFrLvlData,
                         pFrameDes,
                         pStation->DFinfo->frameTimeLength);

      /* store clf time of the current frame and the current time */
      pFrLvlData->bandwdthClfTime = pFrameDes->clfTime;
      pFrLvlData->bandwdthNowTime = nowTime;

      /* increase frame counter */
      pFrLvlData->bandwdthNrOfFrames++;
   }
}


/**
 * @short check the current data frame for frame level events
 *
 * @description This function checks the current data frame for
 * @description frame level events. If the current frame will raise
 * @description a change in the frame level state the corresponding
 * @description wait_state is entered and the frame level event window
 * @description timer is initialized. Within the event window all received
 * @description frames are classified (TIMELY or LATE) and counted.
 * @description If the event timer is expired, the new frame level state
 * @description is computed according to the classification of the
 * @description received data (only timely frames, only late frames, or
 * @description timely and late frames intermixed).
 * @description The corresponding frame level event
 * @description will be written to the station related ehl file or to stdout
 * @description (depending on the current run mode).
 * @description If there is no state change, or no event window timer
 * @description is expired, no frame level event will be reported.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrameDes        pointer to the clf file line, stored in an
 *                         irl entry
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param timelyDataWindow max delay in seconds for a frame to be timely
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
check4FrameLevelEvents(station_t *pStation,
                       intRefListParams_t *pFrameDes,
                       frLvlEvLogData_t *pFrLvlData,
                       long timelyDataWindow,
                       uint16_t runMode)
{
   unsigned char isTimelyFrame; /* frame state: TIMELY or LATE */
   unsigned char newFrLvlState = pFrLvlData->frLvlState; /* frame level state,
                                                         default unchanged */
   int   frLlvlTimerState;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* check if the current frame is timely or late */
   isTimelyFrame = checkIfDataFrameIsTimely(pFrameDes,
                                            timelyDataWindow);

   DEBUG4(5,"frameSet: %.6s, clfTime: %f , isTimelyFrame: %c ",
          pFrameDes->aFrameset,
          pFrameDes->clfTime,
          isTimelyFrame);

   /* check if frame level event timer is active
    * or expired */
   frLlvlTimerState = checkFrameLevelEventTimer(pFrLvlData,
                                                pFrameDes->clfTime);

   DEBUG4(5,"clfTime: %f old frLvlState : %c, timer state: %d ",
          pFrameDes->clfTime,
          pFrLvlData->frLvlState,
          frLlvlTimerState);

   /* check the stored frame level state */
   switch (pFrLvlData->frLvlState)
   {
      case UNKNOWN_WAIT:
         /* update data classification */
         updateDataClassification(pFrLvlData,
                                  isTimelyFrame);

         /* note: in this state, it is possible that the frame level
          * event timer is not active, because it is the initial state */

         /* check if frame level event timer is active */
         if (FRLVL_NO_TIMER == frLlvlTimerState)
         {
            /* store attributes of the current data frame which
             * are needed for the event creation */
            storeEventAttributesOfCurrentDf(pFrLvlData,
                                            pFrameDes);

            /* start timer for frame level event creation */
            initFrlLvlEventTimer(pFrLvlData,
                                 pFrameDes->clfTime);
         }
         /* check if frame level event timer is expired */
         else if (FRLVL_TIMER_EXPIRED == frLlvlTimerState)
         {
            /* compute the new frame level state */
            newFrLvlState = computeFrameLevelState(pFrLvlData);
         }
         break;

      case TIMELY:
         /* check if the current frame is a late frame */
         if (LATE == isTimelyFrame)
         {
            /* update data classification */
            updateDataClassification(pFrLvlData,
                                     isTimelyFrame);

            /* store attributes of the current data frame which
             * are needed for the event creation */
            storeEventAttributesOfCurrentDf(pFrLvlData,
                                            pFrameDes);

            /* start timer for frame level event creation */
            initFrlLvlEventTimer(pFrLvlData,
                                 pFrameDes->clfTime);

            /* set the new frame level state */
            newFrLvlState = TIMELY_WAIT;
         }
         break;

      /* both states can be treated the same */
      case TIMELY_WAIT:
      case LATE_WAIT:
         /* update data classification */
         updateDataClassification(pFrLvlData,
                                  isTimelyFrame);

         /* check if frame level event timer is active */
         if (FRLVL_NO_TIMER == frLlvlTimerState)
         {
            (void) snprintf(logBuffer, sizeof(logBuffer),
                            "Station [%.8s] no frame level timer in state: %c",
                            pStation->stationName,
                            pFrLvlData->frLvlState);
            logError("Internal application error", logBuffer);
            exit(EXIT_FAILURE);
         }
         /* check if frame level event timer is expired */
         else if (FRLVL_TIMER_EXPIRED == frLlvlTimerState)
         {
            /* compute the new frame level state */
            newFrLvlState = computeFrameLevelState(pFrLvlData);
         }
         break;

      case LATE:
         /* check if the current frame is a timely frame */
         if (TIMELY == isTimelyFrame)
         {
            /* update data classification */
            updateDataClassification(pFrLvlData,
                                     isTimelyFrame);

            /* store attributes of the current data frame which
             * are needed for the event creation */
            storeEventAttributesOfCurrentDf(pFrLvlData,
                                            pFrameDes);

            /* start timer for frame level event creation */
            initFrlLvlEventTimer(pFrLvlData,
                                 pFrameDes->clfTime);

            /* set the new frame level state */
            newFrLvlState = LATE_WAIT;
         }
         break;

      case BACKFILLING:
         /* update data classification */
         updateDataClassification(pFrLvlData,
                                  isTimelyFrame);

         /* store attributes of the current data frame which
          * are needed for the event creation */
         storeEventAttributesOfCurrentDf(pFrLvlData,
                                         pFrameDes);

         /* start timer for frame level event creation */
         initFrlLvlEventTimer(pFrLvlData,
                              pFrameDes->clfTime);

         /* set the new frame level state */
         newFrLvlState = BACKFILLING_WAIT;
         break;

      case BACKFILLING_WAIT:
         /* update data classification */
         updateDataClassification(pFrLvlData,
                                  isTimelyFrame);

         /* check if frame level event timer is active */
         if (FRLVL_NO_TIMER == frLlvlTimerState)
         {
            /* note: in this state, it is possible that the frame level
             * event timer is not active, because if timely and late frames
             * are received intermixed there is no state change to
             * BACKFILLING, the frame level state remains the same. */

            /* store attributes of the current data frame which
             * are needed for the event creation */
            storeEventAttributesOfCurrentDf(pFrLvlData,
                                            pFrameDes);

            /* start timer for frame level event creation */
            initFrlLvlEventTimer(pFrLvlData,
                                 pFrameDes->clfTime);

            /* set the new frame level state */
            newFrLvlState = BACKFILLING_WAIT;
         }
         /* check if frame level event timer is expired */
         else if (FRLVL_TIMER_EXPIRED == frLlvlTimerState)
         {
            /* compute the new frame level state */
            newFrLvlState = computeFrameLevelState(pFrLvlData);
         }
         break;

      case NO_DATA:
         /* update data classification */
         updateDataClassification(pFrLvlData,
                                  isTimelyFrame);

         /* store df attributes needed for event creation */
         storeEventAttributesOfCurrentDf(pFrLvlData,
                                         pFrameDes);

         /* start timer for frame level event creation */
         initFrlLvlEventTimer(pFrLvlData,
                              pFrameDes->clfTime);

         /* set the new frame level state */
         newFrLvlState = UNKNOWN_WAIT;
         break;

      default:
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "Station [%.8s] unexpected frame level state: %c",
                         pStation->stationName,
                         pFrLvlData->frLvlState);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
   }

   /* check the new frame level state, and raise the corresponding frame level
    * event if necessary */
   checkNewFrameLevelStateAndRaiseEvent(pStation,
                                        pFrLvlData,
                                        newFrLvlState,
                                        runMode);
}


/**
 * @short check if event window for frame level events expired
 *
 * @description This function checks the current frame level state
 * @description of the station. If the frame level state is in a
 * @description wait_state, the frame level event timer is checked.
 * @description All frames received within the expired event window
 * @description have been classified (TIMELY or LATE) and counted.
 * @description If the frame level event timer is expired the new frame level
 * @description state is computed according to the classification of the
 * @description received data (only timely frames, only late frames, or
 * @description timely and late frames intermixed).
 * @description The corresponding frame level event
 * @description will be written to the station related ehl file or to stdout
 * @description (depending on the current run mode).
 * @description If there is no state change, or no event window timer
 * @description is expired, no frame level event will be reported.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param nowTime          current time
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
check4FrameLevelEventsNoDataReceived(station_t *pStation,
                                     frLvlEvLogData_t *pFrLvlData,
                                     double nowTime,
                                     uint16_t runMode)
{
   unsigned char newFrLvlState = pFrLvlData->frLvlState; /* new frame level state,
                                                            default unchanged */
   int   frLlvlTimerState;
   char logBuffer[LOGSIZE];

   /* check if frame level event timer is active
    * or expired */
   frLlvlTimerState = checkFrameLevelEventTimer(pFrLvlData,
                                                nowTime);

   DEBUG4(5,"nowTime: %f old frLvlState : %c, timer state: %d ",
          nowTime,
          pFrLvlData->frLvlState,
          frLlvlTimerState);

   /* check the stored frame level state */
   switch (pFrLvlData->frLvlState)
   {
      /* all wait states can be treated the same */
      case UNKNOWN_WAIT:
      case TIMELY_WAIT:
      case LATE_WAIT:
      case BACKFILLING_WAIT:
         /* check if frame level event timer is expired */
         if (FRLVL_TIMER_EXPIRED == frLlvlTimerState)
         {
            /* compute the new frame level state */
            newFrLvlState = computeFrameLevelState(pFrLvlData);
         }
         break;

      /* all other states remain unchanged */
      case TIMELY:
      case LATE:
      case BACKFILLING:
      case NO_DATA:
         /* frame level state unchanged */
         newFrLvlState = pFrLvlData->frLvlState;
         break;

      default:
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "Station [%.8s] unexpected frame level state: %c",
                         pStation->stationName,
                         pFrLvlData->frLvlState);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
   }

   /* check the new frame level state, and raise the corresponding frame level
    * event if necessary */
   checkNewFrameLevelStateAndRaiseEvent(pStation,
                                        pFrLvlData,
                                        newFrLvlState,
                                        runMode);
}


/**
 * @short check for unprocessed frame level events
 *
 * @description This function checks the current frame level state
 * @description of the station because the station is disconnected
 * @description now, or the time span since the reception of the last
 * @description data frame exceeds the NO_DATA_TIMEOUT
 * @description and it is possible that the frame level state
 * @description was in a wait state (unprocessed frame level event pending).
 * @description All frames received since the start of the timer
 * @description for frame level events have been classified (TIMELY or LATE)
 * @description and counted.
 * @description The new frame level state is computed according to the
 * @description classification of the received data (only timely frames, only
 * @description late frames, or timely and late frames intermixed).
 * @description The corresponding frame level event
 * @description will be written to the station related ehl file or to stdout
 * @description (depending on the current run mode).
 * @description If there is no state change no frame level event will
 * @description be reported.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param clfTime          clf time of the frame (data frame, or virtual
 *                         disconnect frame)
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
check4UnprocessedFrameLevelEvents(station_t *pStation,
                                  frLvlEvLogData_t *pFrLvlData,
                                  double clfTime,
                                  uint16_t runMode)
{
   unsigned char newFrLvlState = pFrLvlData->frLvlState; /* new frame level state,
                                                            default unchanged */
   int   frLlvlTimerState;
   char logBuffer[LOGSIZE];

   /* check if frame level event timer is active
    * or expired */
   frLlvlTimerState = checkFrameLevelEventTimer(pFrLvlData,
                                                clfTime);

   DEBUG4(5,"clfTime: %f old frLvlState : %c, timer state: %d ",
          clfTime,
          pFrLvlData->frLvlState,
          frLlvlTimerState);

   /* check the stored frame level state */
   switch (pFrLvlData->frLvlState)
   {
      /* all wait states can be treated the same */
      case UNKNOWN_WAIT:
      case TIMELY_WAIT:
      case LATE_WAIT:
      case BACKFILLING_WAIT:
         /* check if frame level event timer is active,
          * or expired */
         if (FRLVL_TIMER_ACTIVE == frLlvlTimerState ||
             FRLVL_TIMER_EXPIRED == frLlvlTimerState)
         {
            /* compute the new frame level state */
            newFrLvlState = computeFrameLevelState(pFrLvlData);
         }
         break;

      /* all other states remain unchanged */
      case TIMELY:
      case LATE:
      case BACKFILLING:
      case NO_DATA:
         /* frame level state unchanged */
         newFrLvlState = pFrLvlData->frLvlState;
         break;

      default:
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "Station [%.8s] unexpected frame level state: %c",
                         pStation->stationName,
                         pFrLvlData->frLvlState);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
   }

   /* check the new frame level state, and raise the corresponding frame level
    * event if necessary */
   checkNewFrameLevelStateAndRaiseEvent(pStation,
                                        pFrLvlData,
                                        newFrLvlState,
                                        runMode);
}

/**
 * @short check the new frame level state and raise event
 *
 * @description This function checks the new frame level state
 * @description of the station. If the frame level state has changed
 * @description it is necessary to check if the new frame level state will
 * @description raise a frame level event.
 * @description A dummy irl element is provided, bevause the functions which
 * @description process the event, expect an irl element which contains the
 * @description clf line of the frame which raises the event.
 * @description The dummy irl element contains the necessary event attributes
 * @description only.
 * @description The corresponding frame level event
 * @description will be written to the station related ehl file or to stdout
 * @description (depending on the current run mode).
 * @description If there is no state change no frame level event will
 * @description be reported.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param newFrLvlState    new frame level state
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
checkNewFrameLevelStateAndRaiseEvent(station_t *pStation,
                                     frLvlEvLogData_t *pFrLvlData,
                                     unsigned char newFrLvlState,
                                     uint16_t runMode)
{
   int   logStateChange;
   intRefListParams_t frameDes; /* dummy irl element */

   /* check if the frame level state has changed */
   if (newFrLvlState == pFrLvlData->frLvlState)
      return;  /* nothing to log for the current data frame */

   DEBUG4(5,"frLvlEvClfTime: %f old frLvlState: %c, new frLvlState: %c",
          pFrLvlData->frLvlEvClfTime,
          pFrLvlData->frLvlState,
          newFrLvlState);

   /* store the new frame level state */
   pFrLvlData->frLvlState = newFrLvlState;

   /* check if the state change raises an event */
   logStateChange = checkIfFrLvlStateRaisesAnEvent(newFrLvlState);
   if (0 == logStateChange)
   {
      DEBUG2(5,"new frLvlState: %c does not raise an event",
             newFrLvlState);
      return;
   }

   /* The writeFrameEvent2xxx() functions expect an irl element
    * with the clf line of the data frame which causes the new event
    * create a "dummy irl" element with the stored event attributes */
   setFrLvlEventAttributes(pFrLvlData,
                           &frameDes);

   /* The frame level state has changed.
    * Log the occurred event, depending on the run mode */
   if (runMode & CR_EHL_FILES)
      writeFrameEvent2Ehl(pStation->files.fpUsrFile,
                          newFrLvlState,
                          &frameDes);
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeFrameEvent2Ehl(pStation->files.fpUsrFile,
                          newFrLvlState,
                          &frameDes);
   else if (runMode & EV_STREAM_2_STDOUT)
      writeFrameEvent2Stdout(pStation->conID,
                             newFrLvlState,
                             &frameDes);
}


/**
 * @short check if the data frame status has changed
 *
 * @description This function checks the data frame status
 * @description (frame parsing code fpc) of the current data frame
 * @description If the fpc is different to the stored fpc the
 * @description frame level event "data frame status changed" is
 * @description raised and will be written to the station related ehl file
 * @description or to stdout (depending on the current run mode)
 * @description If there is no change in the data frame status
 * @description no frame level event will be reported.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrameDes        pointer to the clf file line, stored in an
 *                         irl entry
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
check4FrameStatusChange(station_t *pStation,
                       intRefListParams_t *pFrameDes,
                       frLvlEvLogData_t *pFrLvlData,
                       uint16_t runMode)
{
   unsigned char frLvlEvent;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* check if a fpc is stored in the frame level data structure */
   if (0 == pFrLvlData->fpc)
   {
      pFrLvlData->fpc = pFrameDes->fpc; /* store the frame parsing code */
      return;
   }

   /* check if the frame parsing code has changed */
   if (pFrLvlData->fpc == pFrameDes->fpc)
      return; /* data frame status not changed */

   /* save the changed data frame status */
   pFrLvlData->fpc = pFrameDes->fpc; /* store the frame parsing code */

   /* set the corresponding frame level event */
   frLvlEvent = DATA_FRAME_STATUS_CHANGED;

   /* The frame level state has changed.
    * Log the occurred event, depending on the run mode */
   if (runMode & CR_EHL_FILES)
      writeFrameEvent2Ehl(pStation->files.fpUsrFile,
                          frLvlEvent,
                          pFrameDes);
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeFrameEvent2Ehl(pStation->files.fpUsrFile,
                          frLvlEvent,
                          pFrameDes);
   else if (runMode & EV_STREAM_2_STDOUT)
      writeFrameEvent2Stdout(pStation->conID,
                             frLvlEvent,
                             pFrameDes);
}


/**
 * @short check for channel level status change
 *
 * @description This function checks if the channel parsing code
 * @description of the current data channel and/or the channel status
 * @description are different to the data stored for the particular
 * @description data channel. If the checked values are different
 * @description the corresponding channel level event will be written
 * @description to the ehl file.
 * @description The check of the channel status data will be treated specially
 * @description for CD1.1 stations. The check for a change in the channel
 * @description status information will be performed in the length of
 * @description LNG_4_CHK_OF_CD11_STATUS Bytes. This has to be done, because
 * @description of the other information consists of the value of the
 * @description <time of last GPS synchronization>. A change of this
 * @description time stamp, shall not cause a channel level event.
 *
 * @param checkPass       channel parsing code
 * @param statusSize      channel status size
 * @param channelStatus   pointer to the channel status data
 * @param pChLvlData      pointer to the current channel level data
 * @param type            CD data type (CD1 or CD1_1)
 *
 * @return CH_STATUS_UNCHANGED | CH_PARSING_CODE_CHANGED |
 * @return CH_STATUS_SIZE_CHANGED | CH_STATUS_DATA_CHANGED
 */

static int
check4ChannelStatusChange(uint32_t checkPass,
                          uint32_t statusSize,
                          const char *channelStatus,
                          chLvlEvLogData_t *pChLvlData,
                          unsigned char type)
{
   uint16_t sizeToCheck;

   /* check the impossible */
   if (NULL == pChLvlData->chStatusData)
   {
      logError("Internal application error",
               "pChLvlData->chStatusData == NULL");
      exit(EXIT_FAILURE);
   }

   /* determine the size of the channel status data to be checked.
    * For CD1 use the value of statusSize, for CD1.1 use
    * LNG_4_CHK_OF_CD11_STATUS. For safety reasons, check if the status
    * sizes exceeds LNG_4_CHK_OF_CD11_STATUS. */
   if ((CD1_1 == type) && (LNG_4_CHK_OF_CD11_STATUS <= statusSize))
      sizeToCheck = LNG_4_CHK_OF_CD11_STATUS;
   /* CD1 or statusSize less than LNG_4_CHK_OF_CD11_STATUS */
   else
      sizeToCheck = statusSize;

   /* check channel parsing code, and channel status data for differences
    * compared to the stored values */
   if (checkPass != pChLvlData->checkPass)
      return CH_PARSING_CODE_CHANGED;
   else if (statusSize != pChLvlData->chStatusSize)
      return CH_STATUS_SIZE_CHANGED;
   else if (0 != memcmp((const void *)channelStatus,
                        (const void *)pChLvlData->chStatusData,
                        (size_t)sizeToCheck))
      return CH_STATUS_DATA_CHANGED;
   else
      return CH_STATUS_UNCHANGED;
}


/**
 * @short check for channel level events
 *
 * @description This function checks the stored channel level data for
 * @description new channel level events. The channel level data contains
 * @description the updated channel state/info of the channels computed
 * @description during the processing of the in the 1-n data frames
 * @description with the same nominal time from the current expired window
 * @description for the creation of channel level events.
 * @description If a new channel level event is detected,
 * @description the corresponding channel level event
 * @description will be written to the station related ehl file or
 * @description to stdout (depending on the current run mode).
 *
 * @param pStation       pointer to the current station structure
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
check4ChannelLevelEvents(station_t *pStation,
                         /*@null@*/ frLvlEvLogData_t *pFrLvlData,
                         uint16_t runMode)
{
   unsigned char newChLvlState;
   unsigned char chLvlEvent;        /* event to be reported */
   chName_t *actChEntry;
   chNameCache_t *chNameCache;
   chLvlEvLogData_t *pChLvlData;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   chNameCache = &(pStation->chCache);

   /* loop through the entries of the stations channel name cache */
   for (actChEntry = chNameCache->firstCh; actChEntry != NULL;
        actChEntry = actChEntry->nextChannel)
   {
      pChLvlData = (chLvlEvLogData_t *)actChEntry->chCacheExt;

      /* check if channel specific application extension object is existing */
      if (NULL == (pChLvlData))
      {
         /* Fatal error: no channel specific application extension
            object found. */
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "NULL == (pChLvlData) [%.17s] %.17s 1",
                         pStation->conID,
                         actChEntry->scName[0]);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
      }

      /* store the new channel level state */
      newChLvlState = pChLvlData->newChState;

      /* reset the newChState value in the extension object.
       * This is done to know if there is a new state computed
       * for the particular channel */
      pChLvlData->newChState = INITIAL;

      /* check if the channel level state has changed */
      if (newChLvlState == pChLvlData->chState)
      {
         /* if old and new channel level state == CHANNEL_PRESENT
          * perform the check for change of the channel status info.
          * In any other case return without reporting. */
         if (CHANNEL_PRESENT != pChLvlData->chState)
            continue;

         if (CH_STATUS_UNCHANGED == pChLvlData->chStatusChanged)
            continue;

         /* the channel status info or checkPass has changed:
          * set the value of the current event. Note that there are channel
          * level events, which do not represent an own channel level state */
         chLvlEvent = CHANNEL_STATUS_CHANGED;
      }
      else
      {
         /* The new channel level state is different to the stored channel level
          * state.
          * TODO: if the new state is channel missing AND the old state = present
          *       check if the current data end time is higher than the nominal time of the
          *       expired event window.
          *       In this case the channel is counted as present, because of the channel data of this
          *       channel have been received with the last data frame received.
          *       The Channel level state is counted as "unchanged" -> continue. */
         if ((CHANNEL_MISSING == newChLvlState) &&
             (CHANNEL_PRESENT == pChLvlData->chState))
         {
            double currentDataEndTime;

            /* determine the current data end time of the particular channel */
            currentDataEndTime = pChLvlData->lastFrameRecNominalTime +
                                                pChLvlData->frameTimeLengthSec;

            DEBUG3(5,"currentDataEndTime: %f nomTimeOfExpiredEvWindow: %f",
                   currentDataEndTime,
                   pChLvlData->nomTimeOfExpiredEvWindow);

 /* TODO: It is necessary to prevent the update of the channel data times in this case!
  *       because to store this nominal time or the current channel is simply a wrong
  *       information */
            /* If the channel data of the current have been received continue and
             * count the channel level state as unchanged */
            if (currentDataEndTime > pChLvlData->nomTimeOfExpiredEvWindow)
               continue;
         }

         /* Set the value of the current event. Note that there are channel
          * level events, which do not represent an own channel level state */
         chLvlEvent = newChLvlState;
      }

      if (CHANNEL_PRESENT == newChLvlState)
      {
         /* Either the channel has been seen for the first time,
          * or it has been missing before and now is back again,
          * or the channel status information has changed.
          * Initialize the data for the CHANNEL_INFO event.
          * Save the nominal time of the particular channel in order
          * to preset pChLvlData->earliestNominalTime,
          * pChLvlData->latestNominalTime and
          * pChLvlData->currentNominalTime with the nominal
          * time of the data frame which caused the event */
         pChLvlData->earliestNominalTime = pFrLvlData->nominalTime4ChEvents;
         pChLvlData->latestNominalTime = pFrLvlData->nominalTime4ChEvents;
         pChLvlData->lastFrameRecNominalTime = pFrLvlData->nominalTime4ChEvents;
         DEBUG3(5,"init time stamps of ch: [%.17s] with [%f]",
                actChEntry->scName[0],
                pFrLvlData->nominalTime4ChEvents);
      }

      /* store the current channel level state */
      pChLvlData->chState = newChLvlState;

      /* Either the channel level state, or the channel status data
       * has changed. Log the occurred event, depending on the run mode */
      if (runMode & CR_EHL_FILES)
         writeChannelEvent2Ehl(pStation->files.fpUsrFile,
                               chLvlEvent,
                               pFrLvlData,
                               pChLvlData,
                               actChEntry->scName[0],
                               pStation->type);
      else if (runMode & EV_STREAM_2_STDOUT &&
               runMode & WR_STDOUT_2_EHL_FILES)
         writeChannelEvent2Ehl(pStation->files.fpUsrFile,
                               chLvlEvent,
                               pFrLvlData,
                               pChLvlData,
                               actChEntry->scName[0],
                               pStation->type);
      else if (runMode & EV_STREAM_2_STDOUT)
         writeChannelEvent2Stdout(pStation->conID,
                                  chLvlEvent,
                                  pFrLvlData,
                                  pChLvlData,
                                  actChEntry->scName[0],
                                  pStation->type);
   }
}


/**
 * @short perform check for CHANNEL_INFO events
 *
 * @description This function checks for the condition to create the
 * @description channel level event CHANNEL_INFO event.
 * @description This event will be logged periodically in the interval
 * @description of CHANNEL_INFO_EVENT_INTERVAL seconds for all currently
 * @description present channels.
 * @description The channel level event CHANNEL_INFO event will be
 * @description written to the current ehl file or to stdout
 * @description (depending on the current run mode).
 *
 * @param pStation       pointer to the current station structure
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param checkTime      timestamp to check the event interval
 *                       clfTime, if called during data frame processing or
 *                       nowTime, if called in case of "currently no data
 *                                frame present"
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
check4ChannelInfoEvents(station_t *pStation,
                        /*@null@*/ frLvlEvLogData_t *pFrLvlData,
                        double checkTime,
                        uint16_t runMode)
{
   chName_t *actChEntry;
   chNameCache_t *chNameCache;
   chLvlEvLogData_t *pChLvlData;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* If the station is not in the state "STATION_CONNECTED" do nothing */
   if(STATION_CONNECTED != pStation->state)
      return;

   /* check if the time interval for the creation of the CHANNEL_INFO event
    * has been exceeded
    * TODO: change code style of the if (... */
   if (((long)checkTime - (long)pFrLvlData->chInfoEventTime) <
                                               CHANNEL_INFO_EVENT_INTERVAL)
      return;

   DEBUG3(5,"checkTime: %f chInfoEventTime: %f",
          checkTime,
          pFrLvlData->chInfoEventTime);

   /* update CHANNEL_INFO event time stored in the frame level data */
   pFrLvlData->chInfoEventTime = checkTime;

   chNameCache = &(pStation->chCache);

   /* create the channel info event for all present data channels */
  for (actChEntry = chNameCache->firstCh; actChEntry != NULL;
        actChEntry = actChEntry->nextChannel)
   {
      /* locate the channel specific application extension object */
      pChLvlData = ((chLvlEvLogData_t *)actChEntry->chCacheExt);

      /* check if channel specific application extension object is existing */
      if (NULL == (pChLvlData))
      {
         /* no channel specific application extension object found.
          * skip the channel. This is possible if the new frame with the
          * different nominal time contains additional channels. The channel
          * channel specific application extension object will be allocated
          * during the processing of the new data frame. */
         DEBUG3(5,"channel skipped [%.17s] %.17s",
                pStation->conID, actChEntry->scName[0]);

         continue;
      }

      /* check if the channel was  not present in the last data frames with the
       * same nominal time, received within the event time window for the
       * creation of channel level events. */
      if (CHANNEL_PRESENT != pChLvlData->chState)
         continue;

      /* create CHANNEL_INFO event, depending on the current run mode */
      if (runMode & CR_EHL_FILES)
         writeChannelInfoEvent2Ehl(pStation->files.fpUsrFile,
                                   pFrLvlData->chInfoEventTime,
                                   actChEntry->scName[0],
                                   pChLvlData);
      else if (runMode & EV_STREAM_2_STDOUT &&
               runMode & WR_STDOUT_2_EHL_FILES)
         writeChannelInfoEvent2Ehl(pStation->files.fpUsrFile,
                                   pFrLvlData->chInfoEventTime,
                                   actChEntry->scName[0],
                                   pChLvlData);
      else if (runMode & EV_STREAM_2_STDOUT)
         writeChannelInfoEvent2Stdout(pStation->conID,
                                      pFrLvlData->chInfoEventTime,
                                      actChEntry->scName[0],
                                      pChLvlData);
   }
}


/**
 * @short perform check for NO_DATA events
 *
 * @description This function checks the condition to create the frame
 * @description level event NO_DATA.
 * @description The check for the NO_DATA event will be performed only, if the
 * @description station is connected.
 * @description If the time span, since the last data frame was received,
 * @description exceeds the NO_DATA_TIMEOUT, the frame level state changes to
 * @description NO_DATA and the frame level event NO_DATA will be written to
 * @description the station related ehl file or to stdout.
 * @description (depending on the current run mode).
 *
 *
 * @param pStation       pointer to the current station structure
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param checkTime      current time if getNextFrameDescription() returns a
 *                       NULL pointer or value of the clfTime of the
 *                       current frame
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
check4NoDataEvents(station_t *pStation,
                   /*@null@*/ frLvlEvLogData_t *pFrLvlData,
                   double checkTime,
                   uint16_t runMode)
{
   double delta_t;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* If the station is not in the state "STATION_CONNECTED" do nothing */
   if(STATION_CONNECTED != pStation->state)
      return;

   /* If the frame level state is already "NO_DATA" do nothing */
   if (NO_DATA == pFrLvlData->frLvlState)
      return;

   delta_t = checkTime - pFrLvlData->dataRcvTime;

   DEBUG4(5,"check4NoDataEvents: %f - %f delta_t: %f",
          checkTime,
          pFrLvlData->dataRcvTime,
          delta_t);

   /* Check, if the time span, since the last DF/DFF has arrived exceeds
      the value of the NO_DATA_TIMEOUT.
      In this case the frame level state stored in pFrLvlData->frLvlState
      is set to NO_DATA and the event NO_DATA will be created */
   if (NO_DATA_TIMEOUT > delta_t)
      return;

   DEBUG4(5,"checkTime: %f old frLvlState: %c, new frLvlState: %c",
          checkTime,
          pFrLvlData->frLvlState,
          NO_DATA);

   /* NO_DATA event detected, set frame level state to NO_DATA */
   pFrLvlData->frLvlState = NO_DATA;

   /* log the occurred event, depending on the run mode.
      To be sure to log the NO_DATA event with the proper time stamp
      log the event at the time of the last received frame + the value
      of the NO_DATA_TIMEOUT */
   if (runMode & CR_EHL_FILES)
      writeNoDataEvent2Ehl(pStation->files.fpUsrFile,
                           (pFrLvlData->dataRcvTime + NO_DATA_TIMEOUT));
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeNoDataEvent2Ehl(pStation->files.fpUsrFile,
                           (pFrLvlData->dataRcvTime + NO_DATA_TIMEOUT));
   else if (runMode & EV_STREAM_2_STDOUT)
      writeNoDataEvent2Stdout(pStation->conID,
                              (pFrLvlData->dataRcvTime + NO_DATA_TIMEOUT));
}


/**
 * @short perform check for NO_DATA events, new data frame arrived
 *
 * @description This function checks the condition to create the frame
 * @description level event NO_DATA in the special case that a new data
 * @description has arrived.
 * @description If the time span, since the last data frame was received,
 * @description exceeds the NO_DATA_TIMEOUT, the frame condition to
 * @description change the frame level state tu NO_DATA is fulfilled.
 * @description If the current frame level state is a "wait state" the
 * @description unprocessed frame level event is processed.
 * @description The frame level state changes to
 * @description NO_DATA and the frame level event NO_DATA will be written to
 * @description the station related ehl file or to stdout.
 * @description (depending on the current run mode).
 *
 *
 * @param pStation       pointer to the current station structure
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param clfTime        value of the clfTime of the current frame
 *                       (data frame or virtual disconnect frame)
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
check4NoDataEventsNewFrameArrived(station_t *pStation,
                                  /*@null@*/ frLvlEvLogData_t *pFrLvlData,
                                  double clfTime,
                                  uint16_t runMode)
{
   double delta_t;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* If the frame level state is already "NO_DATA" do nothing */
   if (NO_DATA == pFrLvlData->frLvlState)
      return;


   delta_t = clfTime - pFrLvlData->dataRcvTime;

   DEBUG4(5,"check4NoDataEventsNewFrameArrived: %f - %f delta_t: %f",
          clfTime,
          pFrLvlData->dataRcvTime,
          delta_t);

   /* Check, if the time span, since the last DF/DFF has arrived exceeds
      the value of the NO_DATA_TIMEOUT. */
   if (NO_DATA_TIMEOUT > delta_t)
      return;

   /* The NO_DATA condition is fulfilled, check if the current frame level state
    * a "wait state".
    * If the current frame level state is a "wait state" there is an unprocessed
    * frame level event pending, which has to be processed before changing to
    * the frame level state NO_DATA */
   check4UnprocessedFrameLevelEvents(pStation,
                                     pFrLvlData,
                                     clfTime,
                                     runMode);

   DEBUG4(5,"clfTime: %f old frLvlState: %c, new frLvlState: %c",
          clfTime,
          pFrLvlData->frLvlState,
          NO_DATA);

   /* NO_DATA event detected, set frame level state to NO_DATA */
   pFrLvlData->frLvlState = NO_DATA;

   /* log the occurred event, depending on the run mode.
      To be sure to log the NO_DATA event with the proper time stamp
      log the event at the time of the last received frame + the value
      of the NO_DATA_TIMEOUT */
   if (runMode & CR_EHL_FILES)
      writeNoDataEvent2Ehl(pStation->files.fpUsrFile,
                           (pFrLvlData->dataRcvTime + NO_DATA_TIMEOUT));
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeNoDataEvent2Ehl(pStation->files.fpUsrFile,
                           (pFrLvlData->dataRcvTime + NO_DATA_TIMEOUT));
   else if (runMode & EV_STREAM_2_STDOUT)
      writeNoDataEvent2Stdout(pStation->conID,
                              (pFrLvlData->dataRcvTime + NO_DATA_TIMEOUT));
}


/**
 * @short check the events when no irl element is present
 *
 * @description This function checks for unprocessed events if no irl
 * @description element is present.
 * @description The generation of the bandwidth events, the frame level events
 * @description and the channel level events is based on event window mechanism.
 * @description If there is no irl element present, it has still to be checked
 * @description if there are unprocessed bandwidth, frame level, channel level.
 * @description or NO_DATA events.
 *
 * @param pStation         pointer to the current station structure
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param nowTime          current time
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
check4EventsNoFrameReceived(station_t *pStation,
                           frLvlEvLogData_t *pFrLvlData,
                           double nowTime,
                           uint16_t runMode)
{
   chMaskHistory_t *actElement;
   int bwdtEventsProcessed;

    /* Step 01: check for bandwidth events.
     * if the bandwidth events of the current window are processed
     * reset all counters */
    bwdtEventsProcessed = check4BandwidthEventsByNowTime(pStation,
                                                         pFrLvlData,
                                                         nowTime,
                                                         runMode);
    if (BDWDTH_EVENTS_PROCESSED == bwdtEventsProcessed)
    {
       /*  bandwidth events are processed for the current window.
        *  reset all counters and start the next window. */
       pFrLvlData->bandwdthNrOfFrames = 0;
       pFrLvlData->bandwdthClfTimeSpanTotal = 0;
       pFrLvlData->bandwdthFrameTimeTotal = 0;
       pFrLvlData->bandwdthClfTime = 0;
       pFrLvlData->bandwdthNowTime = 0;
    }

    /* Step 02: check for frame level events
     * Check if the frame level state is a wait state.
     * If the frame level event timer is expired compute
     * the new frame level state */
    check4FrameLevelEventsNoDataReceived(pStation,
                                         pFrLvlData,
                                         nowTime,
                                         runMode);

    /* Step 03: get first chmask history element with an expired
     * "event creation window". If a window is expired update
     * the channel data and check for channel level events */
    actElement = getExpiredWindowByNowTime(pFrLvlData,
                                           (double)nowTime);
    if (NULL != actElement)
    {
       /* Step 04: update channel data
       * Note: the current value of nowTime is used
       * as the event time for "CHANNEL_MISSING" events */
       updateChannelData(pStation,
                         actElement,
                         pFrLvlData,
                         (double)nowTime);

       /* Step 05: check for channel level events */
       check4ChannelLevelEvents(pStation,
                                pFrLvlData,
                                runMode);

       /* Step 06: update data times of all present channels*/
       updateChannelDataTimes(pStation,
                              pFrLvlData);
    }

    /* Step 07: check for creation of the CHANNEL_INFO events */
    check4ChannelInfoEvents(pStation,
                            pFrLvlData,
                            (double)nowTime,
                            runMode);

    /* Step 08: unlink expired history elements */
    unlinkExpiredChMaskHistoryElements(pFrLvlData,
                                       CHK_TYPE_NOW_TIME,
                                       (double)nowTime);

    /* Step 01: check for NO_DATA event */
     check4NoDataEvents(pStation,
                        pFrLvlData,
                        nowTime,
                        runMode);

}


 /**
 * @short update channel data times
 *
 * @description This function updates the data times of all channels,
 * @description with the state "CHANNEL_PRESENT" which are stored in
 * @description the channel name cache.
 * @description If a channel is missing, or CH_SKIP_DATA_TIMES_UPD is set for
 * @description the particular channel, the channel data times of this
 * @description particular channel are not updated.
 * @description The values to be updated, will be stored in the channel
 * @description specific application extension object.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrLvlData     pointer to the current frame level data structure
 *
 * @return nothing
 */

static void
updateChannelDataTimes(station_t *pStation,
                       frLvlEvLogData_t *pFrLvlData)
{
   chName_t *actChEntry;
   chNameCache_t *chNameCache;
   chLvlEvLogData_t *pChLvlData;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation)
   {
      logError("Internal application error", "NULL == pStation");
      exit(EXIT_FAILURE);
   }

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   chNameCache = &(pStation->chCache); /* station related channel name cache */

   /* loop through the channels of the channel name cache cache */
   for (actChEntry = chNameCache->firstCh; actChEntry != NULL;
        actChEntry = actChEntry->nextChannel)
   {
      /* locate the channel specific application extension object */
      pChLvlData = ((chLvlEvLogData_t *)actChEntry->chCacheExt);

      /* check if channel specific application extension object is existing */
      if (NULL == (pChLvlData))
      {
         /* Fatal error: no channel specific application extension
            object found. */
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "NULL == (pChLvlData) [%.17s] %.17s 2",
                         pStation->conID,
                         actChEntry->scName[0]);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
      }

      /* check if the channel is present */
      if (CHANNEL_PRESENT != pChLvlData->chState)
         continue;

      /* check if the channel channel data times shall be updated with the
       * nominal time of the current history element */
      if (CH_DO_DATA_TIMES_UPD != pChLvlData->updateChDataTimes)
         continue;

      DEBUG3(5,"update time stamps of ch: [%.17s] with [%f]",
             actChEntry->scName[0],
             pFrLvlData->nominalTime4ChEvents);
      /* update the event log data times of this particular channel */
      if ((long)pFrLvlData->nominalTime4ChEvents <
                                (long)pChLvlData->earliestNominalTime)
         pChLvlData->earliestNominalTime = pFrLvlData->nominalTime4ChEvents;
      else if ((long)pFrLvlData->nominalTime4ChEvents >
                                           (long)pChLvlData->latestNominalTime)
         pChLvlData->latestNominalTime = pFrLvlData->nominalTime4ChEvents;

      pChLvlData->lastFrameRecNominalTime = pFrLvlData->nominalTime4ChEvents;
      DEBUG2(5,"pChLvlData->earliestNominalTime: [%f]",
             pChLvlData->earliestNominalTime);
      DEBUG2(5,"pChLvlData->latestNominalTime: [%f]",
             pChLvlData->latestNominalTime);
      DEBUG2(5,"pChLvlData->lastFrameRecNominalTime: [%f]",
             pChLvlData->lastFrameRecNominalTime);
   }
}


/**
 * @short update channel data
 *
 * @description This function updates the channel information of expected
 * @description channels, after the time window for the event creation is
 * @description expired.
 * @description The function loop through all elements of the channel name
 * @description cache.
 * @description For each channel the current channel level state is computed.
 * @description If a channel is missing the flag CH_SKIP_DATA_TIMES_UPD is set.
 * @description For present channel it is necessary to check if the channel
 * @description status has changed.
 *
 * @param pStation         pointer to the current station structure
 * @param pChMaskHistory   pointer to the expired channel mask history element
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param evTime4MissingChannels   event time for missing channels (either the
 *                                 clftime of the frame which has triggered the
 *                                 event creation, or now time)
 *
 * @return nothing
 */

static void
updateChannelData(station_t *pStation,
                  chMaskHistory_t *pChMaskHistoryElem,
                  frLvlEvLogData_t *pFrLvlData,
                  double   evTime4MissingChannels)
{
   chName_t *actChEntry;
   chNameCache_t *chNameCache;
   int retVal;
   chLvlEvLogData_t *pChLvlData;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation)
   {
      logError("Internal application error", "NULL == pStation");
      exit(EXIT_FAILURE);
   }

   /* check the impossible */
   if (NULL == pChMaskHistoryElem)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pChMaskHistoryElem == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* check the impossible */
   if (CH_EVENTS_PROCESSED == pChMaskHistoryElem->evProcessingState)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
               "pChMaskHistoryElem events already processed [%.8s] nominal %f",
               pStation->stationName,
               pChMaskHistoryElem->nominalTime);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   chNameCache = &(pStation->chCache);

   /* loop through the entries of the stations channel name cache */
   for (actChEntry = chNameCache->firstCh; actChEntry != NULL;
        actChEntry = actChEntry->nextChannel)
   {
      pChLvlData = (chLvlEvLogData_t *)actChEntry->chCacheExt;

      /* check if channel specific application extension object is existing */
      if (NULL == (pChLvlData))
      {
         /* initialize channel specific application extension object */
         pChLvlData = initEvLogChLvlData();

         /* init ok ? */
         if (NULL == pChLvlData)
         {
            (void) snprintf(logBuffer, sizeof(logBuffer),
                     "initEvLogChLvlData() failed [%.17s] %.17s",
                     pStation->conID, actChEntry->scName[0]);
            logError("Internal application error", logBuffer);
            exit(EXIT_FAILURE);
         }
         /* save pointer to the channel specific application extension object */
         /*@-mustfreeonly@*/
         actChEntry->chCacheExt = pChLvlData;
         /*@=mustfreeonly@*/
      }

      /* compute the new channel level state for the particular channel */
      pChLvlData->newChState = computeChannelLevelState(pChMaskHistoryElem,
                                                        actChEntry->maskPos);

      /* store the nominal time of the current history element. This is
       * necessary to be able to perform the special check in case of a
       * missing channel */
      pChLvlData->nomTimeOfExpiredEvWindow = pChMaskHistoryElem->nominalTime;

      /* Check if the channel is missing */
      if (CHANNEL_PRESENT != pChLvlData->newChState)
      {
         /* Store the event time:
          * If a channel is missing the "event time" is the value of the clf time
          * which has triggered the expiration of the channel level event
          * creation window */
         pChLvlData->eventClfTime = evTime4MissingChannels;

         /* Set the flag to skip the update of the channel data times. This is
          * necessary, because the data of this channel & nominal time can
          * already be received. (if the previous frames had contained channel
          * data with a longer frame time length) */
         pChLvlData->updateChDataTimes = CH_SKIP_DATA_TIMES_UPD;

         DEBUG3(5,"ch %.17s: skip upd of channel data times nominal time: %f",
                actChEntry->scName[0],
                pChLvlData->nomTimeOfExpiredEvWindow);
         continue;
      }

      /* The channel is present in the (n) data frames of the current
       * chmask history window, set the flag to perform the update of
       * the channel data times for this channel. */
      pChLvlData->updateChDataTimes = CH_DO_DATA_TIMES_UPD;

      /* Store the event time, if a channel is present, the value of the event
       * time is the value of the clftime of the frame which contains the
       * channel */
      pChLvlData->eventClfTime =
                    pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].clfTime;

      /* store the frame time length in seconds for the particular channel */
      pChLvlData->frameTimeLengthSec =
         pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].frTimeLngSec;

      /* store the nominal time of the frame(s) which has contained
       * the channels */
      pFrLvlData->nominalTime4ChEvents = pChMaskHistoryElem->nominalTime;

      /* The current frame contains the channel:
       * Check if there is a saved status info */
      if (CH_STATUS_NOT_SAVED == pChLvlData->chStatusChanged)
      {
         /* save the channel parsing code, channel status data */
         retVal = saveChannelStatus(
          pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].checkPass,
          pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].chStatusSize,
          pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].chStatusData,
          pChLvlData);

         /* check result */
         if (0 != retVal)
         {
            (void) snprintf(logBuffer, sizeof(logBuffer),
                            "Save channel status failed [%d]",
                            retVal);
            logError("Internal application error", logBuffer);
            exit(EXIT_FAILURE);
         }

         /* The channel status info is saved for the first time, it
          * is not necessary to generate a "channel status changed" event */
         pChLvlData->chStatusChanged = CH_STATUS_UNCHANGED;

         continue;
      }

      /* check channel data for status change */
      pChLvlData->chStatusChanged = check4ChannelStatusChange(
         pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].checkPass,
         pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].chStatusSize,
         pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].chStatusData,
         pChLvlData,
         pStation->type);

      /* if channel status unchanged, do nothing */
      if (CH_STATUS_UNCHANGED == pChLvlData->chStatusChanged)
         continue;

      DEBUG2(5,"channel status changed [%d]", pChLvlData->chStatusChanged);

      /* save the channel parsing code, channel status data */
      retVal = saveChannelStatus(
         pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].checkPass,
         pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].chStatusSize,
         pChMaskHistoryElem->chMaskAndStatus[actChEntry->maskPos].chStatusData,
         pChLvlData);

      /* check result */
      if (0 != retVal)
      {
         (void) snprintf(logBuffer, sizeof(logBuffer),
                         "Save channel status failed [%d]",
                         retVal);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
      }
   }

   /* set state to  CH_EVENTS_PROCESSED */
   pChMaskHistoryElem->evProcessingState = CH_EVENTS_PROCESSED;

   DEBUG2(5,"pChMaskHistoryElem nominalTime: %f set state: CH_EVENTS_PROCESSED",
          pChMaskHistoryElem->nominalTime);
}


/**
 * @short event log; process VIRT_CONNECT
 *
 * @description This function processes a clf line with the type VIRT_CONNECT.
 * @description (option -h  ehl files | -e event log to stdout)
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
evLogProcVirtConnect(station_t *pStation,
                     intRefListParams_t *pFrameDes,
                     /*@null@*/ frLvlEvLogData_t *pFrLvlData,
                     uint16_t runMode)
{
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* Reset the frame level state and the channel level states.
    * Necessary to be able to deal with the situation that the
    * VIRT_DISCONNECT frame can not be seen. */
   resetEvLogFrLvlData(pStation,
                       pFrLvlData);
   resetChannelLeveLStates(pStation);

   /* set the station state to STATION_CONNECTED */
   pStation->state = STATION_CONNECTED;

   /* Store the frame reception time of the VIRT_CONNECT frame.
    * This time stamp will be used at the next check for the
    *  condition to create the NO_DATA event */
   pFrLvlData->dataRcvTime = pFrameDes->clfTime;

   /* Store the frame reception time of the first VIRT_CONNECT frame
    * seen, after the startup of the report tool.
    * This time stamp will be used to be the initial time stamp for
    * the time interval to create the CHANNEL_INFO events */
   if ((double)0 == pFrLvlData->chInfoEventTime)
      pFrLvlData->chInfoEventTime = pFrameDes->clfTime;

   /* log the occured event, depending on the run mode */
   if (runMode & CR_EHL_FILES)
      writeConnEst2Ehl(pStation->files.fpUsrFile,
                       pFrameDes,
                       pStation->type);
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeConnEst2Ehl(pStation->files.fpUsrFile,
                       pFrameDes,
                       pStation->type);
   else if (runMode & EV_STREAM_2_STDOUT)
      writeConnEst2Stdout(pStation->conID,
                          pFrameDes,
                          pStation->type);
}


/**
 * @short event log; process VIRT_DISCONNECT
 *
 * @description This function processes a clf line with the type
 * @description VIRT_DISCONNECT.
 * @description (option -h  ehl files | -e event log to stdout)
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param pFrLvlData     pointer to the current frame level data structure
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
evLogProcVirtDisconnect(station_t *pStation,
                        intRefListParams_t *pFrameDes,
                        /*@null@*/ frLvlEvLogData_t *pFrLvlData,
                        uint16_t runMode)
{
   chMaskHistory_t *actElement;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   /* Step 01: check for frame level events.
    * don't care if the frame level event timer is expired, because the station
    * is disconnected now!
    * Check if there is an unprocessed frame level event */
   check4UnprocessedFrameLevelEvents(pStation,
                                     pFrLvlData,
                                     pFrameDes->clfTime,
                                     runMode);

   /* Step 02: check for channel level events in a loop.
    * don't care if the event creation window of a specific channel mask
    * history element has expired, because the station is
    * disconnected now!
    * Process all unprocessed channel mask history cache elements */
   while (NULL != (actElement = getUnprocessedChMaskHistoryELement(pFrLvlData)))
   {
      /* Step 03: update channel data
       * Note: the clf time of the VIRT_DISCONNECT is used
       * as the event time for "CHANNEL_MISSING" events */
      updateChannelData(pStation,
                        actElement,
                        pFrLvlData,
                        pFrameDes->clfTime);

      /* Step 04: check for channel level events */
      check4ChannelLevelEvents(pStation,
                               pFrLvlData,
                               runMode);

      /* Step 05: update data times of all present channels*/
      updateChannelDataTimes(pStation,
                             pFrLvlData);
   }

   /* Step 06: check for NO_DATA events
    * This has to be done, because it is possible that the time span between
    * the time the last data frame was received, or since the connection has
    * been established and the time the station is disconnected exceeds the
    * value of the NO_DATA_TIMEOUT. This check is necessary to detect the
    * so called "past" events  */
   check4NoDataEvents(pStation,
                      pFrLvlData,
                      pFrameDes->clfTime,
                      runMode);

   /* Step 07: set the station state after the check for NO_DATA events
    * and so on ...
    * Because the check for NO_DATA events will only be performed
    * if the station is connected. */
   pStation->state = STATION_UNCONNECTED;

   /* log the occurred event, depending on the run mode */
   if (runMode & CR_EHL_FILES)
      writeConnTerm2Ehl(pStation->files.fpUsrFile,
                        pFrameDes);
   else if (runMode & EV_STREAM_2_STDOUT &&
            runMode & WR_STDOUT_2_EHL_FILES)
      writeConnTerm2Ehl(pStation->files.fpUsrFile,
                        pFrameDes);
   else if (runMode & EV_STREAM_2_STDOUT)
      writeConnTerm2Stdout(pStation->conID,
                           pFrameDes);

   /* reset the frame level state */
   resetEvLogFrLvlData(pStation,
                       pFrLvlData);

   /* reset the channel level states */
   resetChannelLeveLStates(pStation);
}


/**
 * @short event log; process data frame
 *
 * @description This function processes a clf line with the type GOOD DATA |
 * @description BAD_DATA.
 * @description (option -h  ehl files | -e event log to stdout)
 *
 * @param pStation         pointer to the current station structure
 * @param pFrameDes        pointer to irl entry
 * @param pFrLvlData       pointer to the current frame level data structure
 * @param timelyDataWindow max delay in seconds for a frame to be timely
 * @param nowTime          current time
 * @param runMode          run mode of the report tool
 *
 * @return nothing
 */

static void
evLogProcDataFrame(station_t *pStation,
                   intRefListParams_t *pFrameDes,
                   /*@null@*/ frLvlEvLogData_t *pFrLvlData,
                   long timelyDataWindow,
                   time_t nowTime,
                   uint16_t runMode)
{
   chMaskHistory_t *actElement;
   char logBuffer[LOGSIZE];


   /* check the impossible */
   if (NULL == pFrLvlData)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pFrLvlData == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   DEBUG4(5,"evLogProcDataFrame frameSet: %.6s, clfTime: %f , nominalTime: %f ",
          pFrameDes->aFrameset,
          pFrameDes->clfTime,
          pFrameDes->nominalTime);

   /* Step 01: check connection level state, because it is possible that
    * the VIRT_CONNECT frame cannot be seen by the reporting
    * tool. In this case, the connection level state changes to
    * STATION_CONNECTED when the first data frame is seen.
    * Store the frame reception time of the data frame.
    * This time stamp will be used at the next check for the
    * condition to create the NO_DATA event */
   if(STATION_CONNECTED != pStation->state)
   {
      pStation->state = STATION_CONNECTED;
      pFrLvlData->dataRcvTime = pFrameDes->clfTime;
   }

   DEBUG3(5,"nominalTime = %ld, nowTime - timelyDataWindow = %ld",
          (long)pFrameDes->nominalTime,
          (long)(nowTime - timelyDataWindow));

   /* Step 02: Store the frame reception time of the first data frame
    * seen, after the startup of the report tool.
    * This has to be done, if the VIRT_CONNECT frame has not been seen.
    * This time stamp will be used to be the initial time stamp for
    * the time interval to create the CHANNEL_INFO events */
   if ((double)0 == pFrLvlData->chInfoEventTime)
      pFrLvlData->chInfoEventTime = pFrameDes->clfTime;

   /* Step 03: perform check for NO_DATA events.
    * In the case, that the time span between the last and
    * the current data frame exceeds the NO_DATA_TIMEOUT, the NO_DATA event
    * has to be logged prior to the next event caused by a data frame.
    * If the current frame level state is a wait state, the unprocessed
    * frame level event shall be processed first. */
   check4NoDataEventsNewFrameArrived(pStation,
                                     pFrLvlData,
                                     pFrameDes->clfTime,
                                     runMode);

   /* Step 04: check for bandwidth events
    * check if the time span for the bandwidth event generation has expired.
    * Use the clf time of the current data frame for this check */
   check4BandwidthEvents(pStation,
                         pFrameDes,
                         pFrLvlData,
                         nowTime,
                         runMode);

   /* Step 05: check for frame level events */
   check4FrameLevelEvents(pStation,
                          pFrameDes,
                          pFrLvlData,
                          timelyDataWindow,
                          runMode);

   /* Step 06: check for data frame status change */
   check4FrameStatusChange(pStation,
                          pFrameDes,
                          pFrLvlData,
                          runMode);

   /* Step 07: Store the frame reception time of the current data frame.
    * This time stamp will be used at the next check for the condition to
    * create the NO_DATA event */
   pFrLvlData->dataRcvTime = pFrameDes->clfTime;

   /* Step 08: store current channel mask and status data of all channels which
    * are present in the current data frame */
   storeChMaskAndStatus(pStation,
                        pFrameDes,
                        pFrLvlData,
                        timelyDataWindow);

   /* Step 09: get first chmask history element with an expired
    * "event creation window". If a window is expired update the channel data
    * and check for channel level events */
   actElement = getExpiredWindowByClfTime(pFrLvlData,
                                          pFrameDes->clfTime);
   if (NULL != actElement)
   {
      /* Step 10: update channel data
      * Note: the clf time of the current data frame is used
      * as the event time for "CHANNEL_MISSING" events */
      updateChannelData(pStation,
                        actElement,
                        pFrLvlData,
                        pFrameDes->clfTime);

      /* Step 11: check for channel level events */
      check4ChannelLevelEvents(pStation,
                               pFrLvlData,
                               runMode);

      /* Step 12: update data times of all channels, which are contained
       * in the (n) data frames of the current window.*/
      updateChannelDataTimes(pStation,
                             pFrLvlData);

   }

   /* Step 13: check for creation of the CHANNEL_INFO events */
   check4ChannelInfoEvents(pStation,
                           pFrLvlData,
                           pFrameDes->clfTime,
                           runMode);

   /* Step 14: unlink expired history elements */
   unlinkExpiredChMaskHistoryElements(pFrLvlData,
                                      CHK_TYPE_CLF_TIME,
                                      pFrameDes->clfTime);
}


/**
 * @short channel info process Data Frame
 *
 * @description This function writes the information on the channel subframes
 * @description of the data frame related to the current irl entry to stdout.
 * @description (option: -c writes channel stream to stdout,
 * @description (option: -d writes channel stream to stdout)
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
chInfoProcDataFrame(station_t *pStation,
                    intRefListParams_t *pFrameDes,
                    uint16_t runMode)

{
   uint32_t ii;
   const CDdataChannel_t *channel;  /* Channel Data */
   uint32_t channelsToProcess;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   channelsToProcess = pStation->DFinfo->parsedChannels;
   channel = pStation->DFinfo->channel;

   /* loop through the channels of the current frame */
   for (ii = 0; ii < channelsToProcess; ii++)
   {
      /* There are different informations to be written to stdout
       * for CD1 and CD1.1 channels */
      if (CD1_1 == pStation->type)
      {
         writeCD11ChannelInfo2StdoutOrEhl(pStation,
                                          channel,
                                          pFrameDes,
                                          runMode);
      }
      else /* CD1 */
      {
         writeCD1ChannelInfo2StdoutOrEhl(pStation,
                                         channel,
                                         pFrameDes,
                                         runMode);
      }

      if (runMode & WR_STDOUT_2_EHL_FILES)
      {
         /* flush buffer and check if an error has occurred */
         if (0 != fflush(pStation->files.fpUsrFile) ||
             0 != ferror(pStation->files.fpUsrFile))
         {
            logError("Internal application error", "can't write to ehl file");
            exit(EXIT_FAILURE);
         }
      }
      else
      {
         /* flush buffer and check if an error has occurred */
         if (0 != fflush(stdout) || 0 != ferror(stdout))
         {
            logError("Internal application error", "can't write to stdout");
            exit(EXIT_FAILURE);
         }
      }

      channel++;
   }
 }


/**
 * @short write clf info according to the run mode
 *
 * @description This function writes a clf line to stdout or to the ehl files.
 * @description (option -f; write clf stream to stdout)
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param pClfLine       pointer to clf line which has to be written
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
clfInfoWrite2EhlOrStdout(station_t *pStation,
                         char *pClfLine,
                         uint16_t runMode)
{

   /* check if the information has to be written to the ehl files */
   if (runMode & WR_STDOUT_2_EHL_FILES)
   {
      FILE *fp = pStation->files.fpUsrFile;

      (void) fprintf(fp,
                     "%s",
                     pClfLine);

      /* flush buffer and check if an error has occurred */
      if (0 != fflush(fp) || 0 != ferror(fp))
      {
         /* In most cases this is a permission or file system
            problem. report is not able to save data, so a shutdown
            is initiated */
         logError("Internal application error",
                  "can't write to ehl file");
         exit(EXIT_FAILURE);
      }
   } else {
      /* write the information to stdout with the station name as prefix */

      (void) fprintf(stdout,
                     "%.17s %s",
                     pStation->conID,
                     pClfLine);

      /* flush buffer and check if an error has occurred */
      if (0 != fflush(stdout) || 0 != ferror(stdout))
      {
         logError("Internal application error", "can't write to stdout");
         exit(EXIT_FAILURE);
      }
   }
}


/**
 * @short clf info; process VIRT_CONNECT
 *
 * @description This function writes a clf line with the type VIRT_CONNECT
 * @description to stdout.
 * @description (option -f; write clf stream to stdout)
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
clfInfoProcVirtConnect(station_t *pStation,
                       intRefListParams_t *pFrameDes,
                       uint16_t runMode)
{
   char virtual = '*';
   char logBuffer[LOGSIZE];

   /* There is different information logged for CD1 and CD1.1 frames */
   if (CD1_1 == pStation->type)
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                     "%.38s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f || %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD11,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      virtual,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      0,
                      pFrameDes->clfTimeSpan,
                      (double) 0);
   }
   else /* CD1 */
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                      "%.6s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f || %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD1,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      virtual,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      0,
                      pFrameDes->clfTimeSpan,
                      (double) 0);
   }

   /* write the clf line to stdout or ehl files
    * (depends on the run mode) */
   clfInfoWrite2EhlOrStdout(pStation,
                            logBuffer,
                            runMode);
}


/**
 * @short clf info process VIRT_DISCONNECT
 *
 * @description This function writes a clf line with the type VIRT_DISCONNECT
 * @description to stdout.
 * @description (option -f; write clf stream to stdout)
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
clfInfoProcVirtDisconnect(station_t *pStation,
                          intRefListParams_t *pFrameDes,
                          uint16_t runMode)
{
   char virtual = '*';
   char logBuffer[LOGSIZE];

   /* There are different informations logged for CD1 and CD1.1 frames */
   if (CD1_1 == pStation->type)
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                      "%.38s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f || %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD11,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      virtual,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      0,
                      pFrameDes->clfTimeSpan,
                      (double) 0);
   }
   else /* CD1 */
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                      "%.6s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f || %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD1,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      virtual,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      0,
                      pFrameDes->clfTimeSpan,
                      (double) 0);
   }

   /* write the clf line to stdout or ehl files
    * (depends on the run mode) */
   clfInfoWrite2EhlOrStdout(pStation,
                            logBuffer,
                            runMode);
}


/**
 * @short clf info process Data Frame
 *
 * @description This function writes a clf line with the type GOOD_DATA |
 * @description BAD_DATA to stdout.
 * @description (option -f; write clf stream to stdout)
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
clfInfoProcDataFrame(station_t *pStation,
                     intRefListParams_t *pFrameDes,
                     uint16_t runMode)
{
   char incoming = '>';
   double frameTimeLengthSec;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   frameTimeLengthSec = (double) (pStation->DFinfo->frameTimeLength / 1000);

   /* There is different information logged for CD1 and CD1.1 frames */
   if (CD1_1 == pStation->type)
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                     "%.38s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f |%.100s| %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD11,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      incoming,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      (unsigned int)pFrameDes->fpc,
                      pFrameDes->clfTimeSpan,
                      pStation->DFinfo->channelMask,
                      frameTimeLengthSec);
   }
   else /* CD1 */
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                      "%.6s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f |%.100s| %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD1,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      incoming,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      (unsigned int)pFrameDes->fpc,
                      pFrameDes->clfTimeSpan,
                      pStation->DFinfo->channelMask,
                      frameTimeLengthSec);
   }

   /* write the clf line to stdout or ehl files
    * (depends on the run mode) */
   clfInfoWrite2EhlOrStdout(pStation,
                            logBuffer,
                            runMode);
}


/**
 * @short clf info process Data Format Frame
 *
 * @description This function writes a clf line with the type DATA_FORMAT
 * @description to stdout.
 * @description Due to the fact, that data format frames are not stored in the irl
 * @description only the frame number, DFFoffset and the DFFsize can be provided.
 * @description (option -f; write clf stream to stdout)
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
clfInfoProcDataFormatFrame(station_t *pStation,
                           intRefListParams_t *pFrameDes,
                           uint16_t runMode)
{
   char incoming = '>';
   double frameTimeLengthSec;
   char logBuffer[LOGSIZE];

   /* check the impossible */
   if (NULL == pStation->DFinfo)
   {
      (void) snprintf(logBuffer, sizeof(logBuffer),
                      "pStation->DFinfo == NULL Station [%.8s]",
                      pStation->stationName);
      logError("Internal application error", logBuffer);
      exit(EXIT_FAILURE);
   }

   frameTimeLengthSec = (double) (pStation->DFinfo->frameTimeLength / 1000);

   (void) snprintf(logBuffer,
                   sizeof(logBuffer),
                   "%.6s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f |%.100s| %f\n",
                   pFrameDes->aFrameset,
                   EXTENSION_BIN_CD1,
                   pStation->pConfigData->offlineExt,
                   (unsigned int)pFrameDes->offsetBinFile,
                   (int)pFrameDes->frameSize,
                   incoming,
                   pFrameDes->frameType,
                   pFrameDes->clfTime,
                   pFrameDes->nominalTime,
                   (unsigned int)pFrameDes->fpc,
                   pFrameDes->clfTimeSpan,
                   pStation->DFinfo->channelMask,
                   frameTimeLengthSec);

   /* write the clf line to stdout or ehl files
    * (depends on the run mode) */
   clfInfoWrite2EhlOrStdout(pStation,
                            logBuffer,
                            runMode);
}


/**
 * @short clf info process Frame
 *
 * @description This function writes a clf line with the types ...
 * @description .... to stdout.
 * @description (option -f; write clf stream to stdout)
 * @description If the option -g was specified too, the information is written
 * @description to the ehl files.
 *
 * @param pStation       pointer to the current station structure
 * @param pFrameDes      pointer to irl entry
 * @param runMode        run mode of the report tool
 *
 * @return nothing
 */

static void
clfInfoProcFrame(station_t *pStation,
                 intRefListParams_t *pFrameDes,
                 uint16_t runMode)
{
   char incoming = '>';
   char *pCM = "";
   char logBuffer[LOGSIZE];


   /* There is different information logged for CD1 and CD1.1 frames */
   if (CD1_1 == pStation->type)
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                     "%.38s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f |%.100s| %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD11,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      incoming,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      (unsigned int)pFrameDes->fpc,
                      pFrameDes->clfTimeSpan,
                      pCM,
                      (double) 0);
   }
   else /* CD1 */
   {
      (void) snprintf(logBuffer,
                      sizeof(logBuffer),
                      "%.6s %.6s%.5s %u %d %c%c %f %f 0x%x %.6f |%.100s| %f\n",
                      pFrameDes->aFrameset,
                      EXTENSION_BIN_CD1,
                      pStation->pConfigData->offlineExt,
                      (unsigned int)pFrameDes->offsetBinFile,
                      (int)pFrameDes->frameSize,
                      incoming,
                      pFrameDes->frameType,
                      pFrameDes->clfTime,
                      pFrameDes->nominalTime,
                      (unsigned int)pFrameDes->fpc,
                      pFrameDes->clfTimeSpan,
                      pCM,
                      (double) 0);
    }

   /* write the clf line to stdout or ehl files
    * (depends on the run mode) */
   clfInfoWrite2EhlOrStdout(pStation,
                            logBuffer,
                            runMode);
}


/**
 * @short Print command usage
 *
 * @description This function prints all available options and
 * @description explanation to stdout
 *
 * @return  nothing
 */

static void
printUsage(char * cmdNameWithPath)
{
   char * cmdName;

   cmdName = basename(cmdNameWithPath);

   fprintf(stderr, "CD-rep %.10s (%.16s)\n", VERSION, CDlibVersion());
   fprintf(stderr,
      "usage: %s config_file [-h -e -f [-c |-d -g] [timely_data_window]]\n",
           cmdName);
   fprintf(stderr, "-h creates ehl files (default option)\n");
   fprintf(stderr, "-e writes event stream to stdout\n");
   fprintf(stderr, "-f writes frame stream to stdout\n");
   fprintf(stderr, "-c writes channel stream to stdout\n");
   fprintf(stderr, "-d writes extended channel stream to stdout\n");
   fprintf(stderr, "-g writes stdout stream to ehl files\n");
}


/**
 * @short check cmd line options if config file specified
 *
 * @description This function checks the specified cmd line
 * @description options if a config file is specified.
 *
 * @param argc     Number of CLI arguments
 * @param argv     Vector with CLI arguments
 *
 * @return RUN_MODE_OK or error code
 */
static int checkIfCfgFileSpecified(int argc,
                                   char *argv[])
{
   /* Please note: This funcion is called before debugging is possible */
   /* First check the command line arguments.
      At least a config file has to be specified */
   if (CLI_MIN_ARGUMENTS > argc)
   {
      return ERR_NO_CFG_FILE;
   }
   /* cdrep config_file (without option) */
   else if ('-' == argv[1][0])
   {
      /* error option instead of config file specified */
      return ERR_INV_CFG_FILE;
   }
   return RUN_MODE_OK;
}


/**
 * @short check cmd line options of the report tool
 *
 * @description This function checks the specified cmd line
 * @description options and tries to determine the run mode.
 *
 * @param argc     Number of CLI arguments
 * @param argv     Vector with CLI arguments
 * @param runMode  detected run mode (out)
 *
 * @return RUN_MODE_OK or error code
 */
static int checkCmdOptions(int argc,
                           char *argv[],
                           /*@out@*/uint16_t *runMode)
{
   int i;

   /* cdrep config_file [-option1 ... -optionn [timely_data_window]] */
   *runMode = NO_OPTION_SET; /* initialize; for splint */

   DEBUG2(5,"argc: %d", argc)

   /* First check the number of command line arguments.
      At least a config file has to be specified, or the max number of
      cli arguments shall not be exceeded */
   if (CLI_MIN_ARGUMENTS > argc)
   {
      DEBUG(5,"error, no config file specified");
      return ERR_NO_CFG_FILE;
   }
   else if (CLI_MAX_ARGUMENTS < argc)
   {
      DEBUG(5,"error, to many args");
      return ERR_TO_MANY_ARGS;
   }
   /* Now determine the desired run mode.
      Loop through the specified arguments, start with the first option after
      the config file, which is the third argument */
   for (i = CLI_START_OF_OPTIONS; i <= argc; i++)
   {
      /* Check if the argument is an option (to do so it has to start with '-')
         If the argument is not an option the command line option
         parsing is finished */
      if ('-' != argv[i-1][0])
      {
         DEBUG3(5,"end of options argv[%d]: %s", (i-1), argv[i-1]);
         return RUN_MODE_OK;
      }
      /* Check if the argument is an option (to do so it has consist of only
       * one option char */
      if ('\0' != argv[i-1][2])
      {
         DEBUG3(5,"invalid option format argv[%d]: %s", (i-1), argv[i-1]);
         return ERR_INV_OPTION_FORMAT;
      }
      /* Check the specified option */
      switch (argv[i-1][1])
      {
      /* check for valid options */
      case 'c':
         /* write channel subframe info to stdout (channel stream mode) */
         *runMode |= CH_INFO_2_STDOUT;
         DEBUG(5,"option -c CH_INFO_2_STDOUT detected");
         break;
      case 'd':
         /* write channel subframe info to stdout ext1
          * ( extended channel stream mode) */
         *runMode |= CH_INFO_EXT1_2_STDOUT;
         DEBUG(5,"option -d CH_INFO_EXT1_2_STDOUT detected");
         break;
      case 'e':
         /* write event stream to stdout (event stream mode)*/
         *runMode |= EV_STREAM_2_STDOUT;
         DEBUG(5,"option -e EV_STREAM_2_STDOUT detected");
         break;
      case 'f':
         /* write clf info to stdout (frame stream mode) */
         *runMode |= CLF_INFO_2_STDOUT;
         DEBUG(5,"option -f CLF_INFO_2_STDOUT detected");
         break;
      case 'h':
         /* create ehl files, which is the default option */
         *runMode |= CR_EHL_FILES;
         DEBUG(5,"option -h CR_EHL_FILES detected");
         break;
      case 'g':
         /* write stdout stream to ehl files */
         *runMode |= WR_STDOUT_2_EHL_FILES;
         DEBUG(5,"option -g WRITE_TO_EHL_FILES detected");
         break;
      default:
         DEBUG3(5,"invalid option argv[%d]: %s", (i-1), argv[i-1]);
         return ERR_INV_OPTION;
     }
   }
   DEBUG(5,"all options processed");

   /* check if option -c and -d are both specified, because this is an
    * invalid option combination */
   if ((*runMode & CH_INFO_2_STDOUT) &&
       (*runMode & CH_INFO_EXT1_2_STDOUT))
   {
      DEBUG(5,"invalid option combination -c -d");
      return ERR_INV_OPTION_COMBINATION;
   }

   return RUN_MODE_OK;
}


/*
 * @short catch signals
 *
 * @description This function catches a signal and sets the shutdown
 * @description flag in the config structure
 *
 * @param signal  signal
 *
 * @return None
 */

static void
catchSignal(int signal)
{
   char logBuffer[LOGSIZE];

   /* Don't use getpid() on LINUX here!! */
   (void) snprintf(logBuffer, LOGSIZE, "catchSignal: SIG=%d", signal);
   DEBUG(2, logBuffer);

   /* Command all handing threads to terminate */
   setShutdownFlag(SD_FAST);
}


/**
 * @short reporting tool main function
 *
 * @description This function initializes the file reader API
 * @description and reads all index files for a given port in a
 * @description given directory, builds a list with clf files,
 * @description parses the clf files and tries to find events in
 * @description the clf files. Possible events are:
 * @description Data frame seen the first time from a station
 * @description TODO: ?? description of the options???
 *
 *
 * @param argc          argument count
 * @param argv          argument vector
 *
 * @return 0 on success or != 0 on error
 */

int
main(int argc, char *argv[])
{
   configParams_t *pConfigData;
   senderParams_t *pSenderParams;
   station_t *pStation;
   intRefListParams_t *pFrameDes,
                      *element2free;
   time_t startDate,             /* start date for parsing index files */
          nowTime;
   int retVal;
   uint16_t event = NO_ERROR;
   uint32_t strictMode;
   FILE *fp;
   char stationNameBuffer[LEN_STATION_NAME + 1];
   long timelyDataWindow = TIMELY_DATA_WINDOW;
   uint16_t runMode = NO_OPTION_SET;   /* initialize */
   int errCode;
   int newDFF;
   char logBuffer[LOGSIZE];
   struct sigaction sigInfo;
   char dateString[18];

   /* Check if at least a config file is specified in the command
    * line options */
   if (RUN_MODE_OK != (errCode = checkIfCfgFileSpecified(argc, argv)))
   {
      printUsage(argv[0]);
      exit(EXIT_FAILURE);
   }

   /* Init syslog */
   logInit("ctbt.acquisition.cdrep", LOG_INFO);

   /* Initialize file reader API. No configuration file is
      read if the first parameter is NULL */
   pConfigData = initFileReaderAPI(argv[1], "report tool");
   /* Set application type 'other' to load certificates for
      data providers (fixes ISI-38, requires libCD >= 2.3.1) */
   pConfigData->pHistory->appID = OTHER;
   pSenderParams = (senderParams_t *)pConfigData->configExtension;
   if (NULL == pSenderParams)  /* for splint */
   {
      logError("Internal application error",
                "pSenderParams is NULL");
      exit(EXIT_FAILURE);
   }

   /* Check the command line options and determine the run mode */
   if (RUN_MODE_OK != (errCode = checkCmdOptions(argc, argv, &runMode)))
   {
      printUsage(argv[0]);
      exit(EXIT_FAILURE);
   }
   /* Check if an option specified, otherwise set the default option */
   else if (NO_OPTION_SET == runMode){
      runMode = CR_EHL_FILES;
      DEBUG(5,"no option specified, set default option: CR_EHL_FILES");
   }
   /* Check if an output option specified, otherwise set the default option */
   else if (WR_STDOUT_2_EHL_FILES == runMode){
      runMode |= CR_EHL_FILES;
      DEBUG(5,"no output option specified, set default option: CR_EHL_FILES");
   }

   DEBUG2(5,"run mode: %d", runMode);

   /* Default: switch off the creation of files */
   pConfigData->systemConf &= ~(CREATE_INDEX_FILE | CREATE_ANY_FRAME_FILE);

   /* Install signal handler */
   memset (&sigInfo, 0, sizeof (struct sigaction)); /* for splint */
   sigInfo.sa_handler = catchSignal;
   sigInfo.sa_flags  = 0;
   (void) sigaction(SIGINT, &sigInfo, NULL);

   /* Check if an alternative value for the timely data window is specified.
      This is only necessary if the desired run mode is to create
      ehl files, or to write the event log stream to stdout */
   if ((runMode & CR_EHL_FILES) ||
       (runMode & EV_STREAM_2_STDOUT))
   {
      /* Check if more than the prog name and a config file
       * is specified as command option and if the last cmd line
       * argument is not an option */
      if ((CLI_MIN_ARGUMENTS < argc) &&
          ('-' != argv[argc-1][0]))
      {
         DEBUG2(5,"timelyDataWindow value from cmdline: %s", argv[argc-1]);
         timelyDataWindow = strtol(argv[argc-1], (char **)NULL, 10);
         DEBUG2(5,"timelyDataWindow value converted: %ld", timelyDataWindow);
      }
      /* check the time data window value. if an illegal value was specified,
       * the converted value is set to 0. */
      if (0 == timelyDataWindow)
      {
         DEBUG(5,"invalid timelyDataWindow value, shall be > 0");
         printUsage(argv[0]);
         exit(EXIT_FAILURE);
      }
   }

   DEBUG2(5,"timelyDataWindow value  %ld", timelyDataWindow);

   /* Check, if the desired run mode is set to create ehl files.
      In this case switch on the creation of user specific output files */
   if (runMode & CR_EHL_FILES ||
       runMode & WR_STDOUT_2_EHL_FILES)
   {
      pConfigData->systemConf |= CREATE_USR_FILE;
      /*@-mustfreeonly@*/
      /*@-memtrans@*/
      pConfigData->usrFileExt = "ehl";
      /*@=memtrans@*/
      /*@=mustfreeonly@*/
   }

   /* If you don't like the values read from the configuration file,
      modify it here. Keep in mind that enableSyslog and debugLevel
      have to be set with setDebugLevel() and setSyslogFilterLevel() */
   /* Switch off DB access */
   pSenderParams->dbConnectName[0] = '\0';
   pSenderParams->dbAccount[0] = '\0';

   /* enable strict mode */
   strictMode = pConfigData->systemConf & STRICT_MODE_ENABLE;

   /* Determine the initial start date for the check for new clf files */
   nowTime = time(NULL);
   startDate = nowTime - (pSenderParams->lookBackTimeSched * 3600);

   /* All you have to do is to check for new clf files in the index file
      using catIdxDir(). CatIdxDir() will update a list of clf files each
      time it is called() */
   while ((unsigned char)0 == pConfigData->shutdownEnable)
   {
      /* Parse index files for a given port and a given directory */
      retVal = catIdxDir(pConfigData->indexFilePath,
                         pConfigData->portNumber,
                         pSenderParams->pProdLine,
                         pSenderParams->nbrOfprodLines,
                         &startDate, 0);

      if (retVal < 0)
      {
         (void) snprintf(logBuffer, sizeof(logBuffer),
              "Search for clf files returns [%d]. unable to parse Index files",
                         retVal);
         logError("Internal application error", logBuffer);
         exit(EXIT_FAILURE);
      }

      /*
       * Check for all production lines if new frames are arrived
       */
      pStation = getFirstStation();
      while (NULL != pStation)
      {
         intRefListHead_t *pRefListHead;
         cdsend_t *pCdSend;            /* Application extension object */
         frLvlEvLogData_t *pFrLvlData; /* event log frame level data */

#ifdef HAVE_AUTHENTICATION
         /* Check if certificates have to be loaded */
         if (HAVE_TO_LOAD_CERTIFCATES(pStation))
         {
            /* No keys are loaded, load all keys for station */
            if (0 != loadCertificates(pStation,
                         (pStation->systemConf & VERBOSE_LOGGING)))
            {
               exit(EXIT_FAILURE);
            }
         }
#endif /* HAVE_AUTHENTICATION */

         /* get the station related application extension object */
         pCdSend = (cdsend_t *)pStation->appExtension;

         /* check if the station has a production line */
         if (NULL == pCdSend ||  /* for splint only */
             NULL == pCdSend->pProdLine)
         {
            /* Station without production line. This one is
               created from a key preload directive, we can ignore
               it */
            ST_DEBUG(2, "Station without production line found");
            pStation = pStation->nextStation;
            continue;
         }

         /* check the pointer to the head of the internal reference list */
         if (NULL == pCdSend->pRefListHead)
         {
            (void) snprintf(logBuffer, sizeof(logBuffer),
                            "pRefListHead is NULL [%.17s]",
                            pStation->conID);
            logError("Internal application error", logBuffer);
            exit(EXIT_FAILURE);
         }

         /* Set data provider (station) name */
         getDPname(pStation->conID, stationNameBuffer,
                   sizeof(stationNameBuffer));
         DEBUG2(2,"Station [%.8s]", stationNameBuffer);

         /* Check, if the desired run mode is set to create ehl files */
         if (runMode & CR_EHL_FILES ||
             runMode & WR_STDOUT_2_EHL_FILES)
         {
            /* In this case check if new files have to be created or old ones
               have to be opened */
            if (-1 == syncFiles(pStation, NO, 0, 0))
            {
               (void) snprintf(logBuffer, sizeof(logBuffer),
                               "syncFiles failed [%.17s]",
                               pStation->conID);
               logError("Internal application error", logBuffer);
               exit(EXIT_FAILURE);
            }
         }

         pFrLvlData = (frLvlEvLogData_t *)pCdSend->addAppExtension;
         /* check if the event log specific additional application
            extension object already exists.
            This has to be done if the run mode is set to CR_EHL_FILES, or
            EV_STREAM_2_STDOUT */
         if (((runMode & CR_EHL_FILES) || (runMode & EV_STREAM_2_STDOUT)) &&
             (NULL == pFrLvlData))
         {
            /* initialize the event log specific additional application
               extension object */
            pFrLvlData = initEvLogFrLvlData();

            /* init ok ? */
            if (NULL == pFrLvlData)
            {
               (void) snprintf(logBuffer, sizeof(logBuffer),
                               "initEvLogFrLvlData() failed [%.17s]",
                               pStation->conID);
               logError("Internal application error", logBuffer);
               exit(EXIT_FAILURE);
            }

            /* save the pointer to the event log specific additional
               application extension object */
            /*@-mustfreeonly@*/
            pCdSend->addAppExtension = pFrLvlData;
            /*@=mustfreeonly@*/
         }

         pRefListHead = ((cdsend_t *)(pStation->appExtension))->pRefListHead;
         /* Check for a particular production line if new frames
            have arrived. The pLastSent field of the reference list header
            is used here to store the last frame read (or better to know
            which was the last DFF parsed (only CD1.0)) */
         fp = NULL;
         while (NULL != (pFrameDes = getNextFrameDescription(pStation)))
         {
            /* check frame type */
            switch (pFrameDes->frameType)
            {
               case  VIRT_CONNECT:
                  /* process the irl element depending on the run mode */
                  if ((runMode & CR_EHL_FILES) ||
                      (runMode & EV_STREAM_2_STDOUT))
                        evLogProcVirtConnect(pStation,
                                             pFrameDes,
                                             pFrLvlData,
                                             runMode);

                  if (runMode & CLF_INFO_2_STDOUT)
                        clfInfoProcVirtConnect(pStation,
                                               pFrameDes,
                                               runMode);
                  /* Schedule current irl element read for release */
                  element2free = pFrameDes;
                  DEBUG4(5,"%lx %lx %c",
                        (unsigned long)pFrameDes,
                        (unsigned long)pRefListHead->pLastSent,
                        pFrameDes->frameType);
                  break;
               case  VIRT_DISCONNECT:
                  /* process the irl element depending on the run mode */
                  if ((runMode & CR_EHL_FILES) ||
                      (runMode & EV_STREAM_2_STDOUT))
                        evLogProcVirtDisconnect(pStation,
                                                pFrameDes,
                                                pFrLvlData,
                                                runMode);

                  if (runMode & CLF_INFO_2_STDOUT)
                        clfInfoProcVirtDisconnect(pStation,
                                                  pFrameDes,
                                                  runMode);
                  /* Schedule current irl element read for release */
                  element2free = pFrameDes;
                  DEBUG4(5,"%lx %lx %c",
                        (unsigned long)pFrameDes,
                        (unsigned long)pRefListHead->pLastSent,
                        pFrameDes->frameType);
                  break;
               case  GOOD_DATA:
               case  BAD_DATA:
                  /* Read frame from binary file */
                  /*@-nullstate@*/
                  retVal = readFrameByIrl(pStation,
                                          pFrameDes,
                                          pRefListHead->pLastSent,
                                          &fp,
                                          &event,
                                          &newDFF);
                  /*@=nullstate@*/
                  ST_DEBUG4(2, "readFrameByIrl returns %d [%d] %d",
                            retVal, newDFF, (int) event);

                  /* check result of readFrameByIrl() */
                  if (NO_ERROR == event)
                  {
                     ST_DEBUG3(2, "Frame received at: %f, needs %fs",
                                  pFrameDes->clfTime,
                                  pFrameDes->clfTimeSpan);

                     /* reset frame parsing code */
                     pStation->checkPass = ER_NOT_PARSED;

                     /* Parse the frame */
                     parseFrame(pStation,
                                strictMode,
                                newDFF,
                                (long)pFrameDes->clfTime,
                                pFrameDes->frameSize);

                     /* process the frame element depending on the run mode */
                     if ((runMode & CR_EHL_FILES) ||
                         (runMode & EV_STREAM_2_STDOUT))
                           evLogProcDataFrame(pStation,
                                              pFrameDes,
                                              pFrLvlData,
                                              timelyDataWindow,
                                              nowTime,
                                              runMode);

                     if (runMode & CLF_INFO_2_STDOUT)
                           clfInfoProcDataFrame(pStation,
                                                pFrameDes,
                                                runMode);

                     if (runMode & CH_INFO_2_STDOUT)
                           chInfoProcDataFrame(pStation,
                                               pFrameDes,
                                               runMode);

                     if (runMode & CH_INFO_EXT1_2_STDOUT)
                           chInfoProcDataFrame(pStation,
                                               pFrameDes,
                                               runMode);

                     ST_DEBUG3(9,"1:%lx %lx",
                               (unsigned long)pFrameDes,
                               (unsigned long)pRefListHead->pLastSent);
                     /* Schedule last irl element read for release */
                     element2free = pRefListHead->pLastSent;
                     /* Set new last element read and
                        remember the last element read one station cycle */
                     pRefListHead->pLastSent = pFrameDes;
                  }
                  else if (E_SYSTEM == event)
                  {
                     /* There was a problem reading the data and/or
                        the data format frame. If the data format frame
                        couldn't be read, all following readFrameByIrl()
                        calls until a new DFF is found, will fail */
                     fprintf(stderr, "Can't read frame at position %u\n",
                             (unsigned int)pFrameDes->offsetBinFile);
                     /* Schedule current irl element read for release */
                     element2free = pFrameDes;
                     DEBUG3(5,"2:%lx %lx",
                            (unsigned long)pFrameDes,
                            (unsigned long)pRefListHead->pLastSent);
                  }
                  else
                  {
                     fprintf(stderr, "Unexpected event Offset: %u , Event:%d\n",
                             (unsigned int)pFrameDes->offsetBinFile,
                             (int) event);
                     /* Schedule current irl element read for release */
                     element2free = pFrameDes;
                     DEBUG3(5,"3:%lx %lx",
                            (unsigned long)pFrameDes,
                            (unsigned long)pRefListHead->pLastSent);
                  }
               break;
               case  GOOD_FORMAT:
               case  BAD_FORMAT:
               case  GOOD_VIRT_DFF:
               case  BAD_VIRT_DFF:
                  /* process data format frame */
                  /* Read data format frame from binary file */
                  /*@-nullstate@*/
                  retVal = readDffByIrl(pStation,
                                        pFrameDes,
                                        pRefListHead->pLastSent,
                                        &fp,
                                        &event);
                  /*@=nullstate@*/

                  ST_DEBUG3(2, "readDffByIrl returns %d %d",
                            retVal, (int) event);

                  /* check result of readDffByIrl() */
                  if (NO_ERROR == event)
                  {
                     ST_DEBUG3(2, "Dff received at: %f, needs %fs",
                                  pFrameDes->clfTime,
                                  pFrameDes->clfTimeSpan);

                     /* reset frame parsing code */
                     pStation->checkPass = ER_NOT_PARSED;

                     /* Parse data format frame. Data is available in
                        the stations format frame buffer. No events are
                        logged in this function */
                     pStation->checkPass = parseAndCheckDFF(
                                                   pStation->DFinfo,
                                                   strictMode,
                                                   pStation->formatBuffer,
                                                   &(pStation->chCache));
                     ST_DEBUG2(2, "parseAndCheckDFF returns %d",
                     (int)pStation->checkPass);

                     /* check run mode */
                     if (runMode & CLF_INFO_2_STDOUT)
                     {
                        clfInfoProcDataFormatFrame(pStation,
                                                   pFrameDes,
                                                   runMode);
                     }

                     ST_DEBUG3(9,"1:%lx %lx",
                               (unsigned long)pFrameDes,
                               (unsigned long)pRefListHead->pLastSent);
                     /* Schedule last irl element read for release */
                     element2free = pRefListHead->pLastSent;
                     /* Set new last element read and
                        remember the last element read one station cycle */
                     pRefListHead->pLastSent = pFrameDes;
                  }
                  else if (E_SYSTEM == event)
                  {
                     /* There was a problem reading the data and/or
                        the data format frame. If the data format frame
                        couldn't be read, all following readFrameByIrl()
                        calls until a new DFF is found, will fail */
                     fprintf(stderr, "Can't read frame at position %u\n",
                             (unsigned int)pFrameDes->offsetBinFile);
                     /* Schedule current irl element read for release */
                     element2free = pFrameDes;
                     DEBUG3(5,"2:%lx %lx",
                            (unsigned long)pFrameDes,
                            (unsigned long)pRefListHead->pLastSent);
                  }
                  else
                  {
                     fprintf(stderr, "Unexpected event Offset: %u , Event:%d\n",
                             (unsigned int)pFrameDes->offsetBinFile,
                             (int) event);
                     /* Schedule current irl element read for release */
                     element2free = pFrameDes;
                     DEBUG3(5,"3:%lx %lx",
                            (unsigned long)pFrameDes,
                            (unsigned long)pRefListHead->pLastSent);
                  }
                  break;
               case  GOOD_ALERT:
               case  BAD_ALERT:
               case  GOOD_ACKNAK:
               case  BAD_ACKNAK:
               case  GOOD_OPTION_REQ:
               case  BAD_OPTION_REQ:
               case  GOOD_OPTION_RES:
               case  BAD_OPTION_RES:
                  /* process frame */

                  /* check run mode */
                  if (runMode & CLF_INFO_2_STDOUT)
                  {
                     clfInfoProcFrame(pStation,
                                      pFrameDes,
                                      runMode);
                  }

                  /* Schedule current irl element read for release */
                  element2free = pFrameDes;
                  DEBUG4(5,"%lx %lx %c",
                        (unsigned long)pFrameDes,
                        (unsigned long)pRefListHead->pLastSent,
                        pFrameDes->frameType);
                  break;
            default:
               DEBUG2(5,"unexpected frame type: %c, frame skipped ...",
                      pFrameDes->frameType);
               /* Schedule current irl element read for release */
               element2free = pFrameDes;
               DEBUG4(5,"4:%lx %lx %ld", (unsigned long)pFrameDes,
                      (unsigned long)pRefListHead->pLastSent,
                      (long)(nowTime - (long)pFrameDes->nominalTime));
            }

            DEBUG2(5,"-> %lx",(unsigned long)element2free);
            if (element2free == pRefListHead->pLastSent)
            {
               (void) snprintf(logBuffer, sizeof(logBuffer),
                       "An unrecoverable has occurred 1[%d], stop program",
                       (int) event);
               logError("Internal application error", logBuffer);
               exit(EXIT_FAILURE);

            }
            /* Free the reflist element */
            if (NULL != element2free) freeRefListElement(element2free);

            /* E_SYSTEM is a problem opening the file or reading the frame,
               it is treated as recoverable */
            if (event != NO_ERROR && event != E_SYSTEM)
            {
               (void) snprintf(logBuffer, sizeof(logBuffer),
                       "An unrecoverable has occurred 2[%d], stop program",
                       (int) event);
               logError("Internal application error", logBuffer);
               exit(EXIT_FAILURE);
            }
         }

         /* Currently no irl element present, check the run mode.
          * If the run mode is set to CR_EHL_FILES, or
          * EV_STREAM_2_STDOUT,it is necessary to peform the check
          * for events which are not processed yet (bandwidth events,
          * frame level events, channel level events, NO_DATA) */
         if ((runMode & CR_EHL_FILES) ||
             (runMode & EV_STREAM_2_STDOUT))
         {
            check4EventsNoFrameReceived(pStation,
                                        pFrLvlData,
                                       (double) nowTime,
                                        runMode);
         }

         /* check if binary file needs to be closed */
         if (NULL != fp)
         {
            ST_DEBUG(2, "Binary file needs to be closed");
            (void)fclose(fp);
         }
         /* get next station */
         pStation = pStation->nextStation;
      }
      (void)sleep((unsigned int)pSenderParams->sleepTimeWork);
      nowTime = time(NULL);
   }

   /* CD report is shutting down */
   getDate(dateString, sizeof(dateString));
   (void) snprintf(logBuffer, LOGSIZE, "CDreport stopped at %s",
                   dateString);
   logWrite(LOG_NOTICE, logBuffer);
   closeLibCD();
   closelog();
   exit(EXIT_SUCCESS);
}
